--[[
			Advanced Tools for Mission Editor - ATME
			Author : Sunski34
			
			Testers : Snowsniper/Sunski34

			Oriented Object lua core module
			
			V.2.0.0 - Beta Version - 01/2020
				-> Radio transmission buffering when setting the same frequency fixed
				-> ATME bug reporting when using dynamic link for lua files fixed
				-> ATME bug when setting route with landing on ship or farp fixed
				-> ATME.coordLocalToDCS function : bad altitude fixed
				-> Add ATME.getAzimuth360 function to reduce azimuth in 0 / 360 interval
				-> Add ATME.getAzimuth180 function to reduce azimuth in -180 / 180 interval
				-> Rename ATME.C_Area class into ATME.C_ZonesMgr for future 3D zones
				-> Remove ATME.C_Group.getGroupsInAreaForCoalition function
				-> Remove ATME.C_Group.getGroupsInZone2DForCoalition function
				-> Remove ATME.C_Group.getGroupsInDCSZoneForCoalition function
				-> Remane ATME.C_Group.getReadyToBoardGroupsForCoalition function into ATME.C_Group.findReadyToBoardGroupsForCoalition
				-> Remove ATME.C_Group.getGroupsInAreaForAll function
				-> Remove ATME.C_Group.getGroupsInZone2DForAll function
				-> Remove ATME.C_Group.getGroupsInDCSZoneForAll function -> Use ATME.C_Group.findInZone2DForAll function
				-> Rename ATME.C_Group.getReadyToBoardGroupsForAll function into ATME.C_Group.findReadyToBoardGroupsForAll
				-> Add new ATME.C_FindContext class to find groups, airbases, static objects or units in zone2D, zone3D or zones managers
				-> Rename isInArea function of ATME.C_Group, ATME.C_AIUnit and ATME.C_Player into isInZones
				-> Rename getCategory function into getCategoryName for ATME.C_Airbase and ATME.C_StaticObject classes
				-> Remove getLinkedUnitName function of ATME.C_ZonesMgr class
				-> Add new addUnitRelative function in ATME.C_ZonesMgr class to manage moving zones
				-> Add new getZoneLinkedUnitName function to ATME.C_ZonesMgr class
				-> Add new altitude optional parameter to ATME.C_Circle constructor when using DCSZone name
				-> Remove ATME.C_Ring class
				-> Add changeThickness function to ATME.C_Circle to make rings from circles
				-> Add getThickness function to ATME.C_Circle
				-> Add new zone 3D classes : ATME.C_Sphere, ATME.C_Cylinder
				-> Add new zone 3D virtual classes : ATME.C_Zone, ATME.C_Zone3D, ATME.C_MovableZone3D
				-> Add new zone2D vitual ATME.C_MovableZone2D class
				-> Remove isInSphere and isNear functions in ATME.C_Group, ATME.C_AIUnit and TAME.C_Player classes
				-> Add new isInZone3D function to ATME.C_Group, ATME.C_AIUnit and TAME.C_Player classes
				-> Add new addReferencePolarWaypoint function to ATME.C_Route class to define waypoints from a reference
				-> Add new getLandingAirbase function to ATME.C_Route class
				-> Remove getBox function from all classes derived from ATME.C_Zone
				-> Add new optionnal parameter to ATME.C_Rectangle to define hippodrome (width sides are half circles)
				-> Add new ATME.C_Box class to manage 3D boxes. Able to define a 3D hippodrome (width sides are half vertical cylinder).
				-> Modify parameters of addInitialWaypointsFrom and addWaypointsFrom functions of ATME.C_Route class
				-> Add new function addFromRoute into ATME.C_Route to copy waypoints from a route to another
				-> Add new function getInitialZone into ATME.C_ZonesMgr to get the initial zone in a zones manager
				-> Now waypointId can be a waypoint name or an index of a waypoint for addTaskToWaypoint and addOptionToWaypoint 
				   functions of ATME.C_Route
				-> ATME manages module name when calling user module handlers or callbacks, so functions below have new use definitions
				-> constructor of ATME.C_F10Menu don't have a module parameter
				-> getUserPrivateDatas function of appropriate classes don't have a module parameter
				-> getF10MenuRoot function of ATME.C_Player don't have a module parameter
				-> Add new flag stopCondition in setPatrol function of ATME.C_Group class
				-> Add new optional parameter to display functions to clear the view before display new message
				-> Add new functions getAttributes and hasAttributes to ATME.C_AIUnit and ATME.C_Player classes
				-> Rename DCS task ENGAGE_TARGETS_IN_ZONE into ENGAGE_TARGETS_IN_ZONE_DCS
				-> unload function of ATME.C_AIUnit and ATME.C_Player classes now returns true if onload is ok, else false. 
				   But still use ATME.getError function if necessary
				-> Rename onModuleStartHandler into onStartMissionHandler. Boolean parameter remove, called when all units and groups are already initialized
				-> When starting mission onCreateAIUnitHandler, onCreatePlayerHandler and onCreateGroupHandler are called 
				   when all units and groups are already initialized and just after onStartMissionHandler.
				-> ATME.convertToDMS now convert only in degrees, minutes, seconds and cents of seconds.
				-> Add new ATME.convertToDMd function to convert in degrees, minutes, cents of minutes
				-> Rename ATME.LLtoString into ATME.LLToDMSString to format a string in hight (cents of second) or low precision (just seconds).
				-> Add new ATME.LLToDMdString to format a string in minutes and cents of minutes.
				-> Rename ATME.convertToHMSd into ATME.convertToHMS
				-> Rename setFrequency of ATME.C_AIUnit class into transmissionSetFrequency
				-> Add new setFrequency function to ATME.C_AIUnit class for unit radio frequency settings (AWACS and Tankers)
				-> Add new beaconOn and beaconOff to ATME.C_AIUnit class for unit beacons management (Actualy TACAN, TACAN_AA and ICLS)
				-> remove function from ATME.C_IndexList now returns the removed item
				-> get function of ATME.C_IndexList now returns two value : item, id
				-> Add new getFirst, getLast, getNext and getPrevious functions to ATME.C_IndexList. They all return two values : item, id
				-> Add new optionnal parameter to ATME.run function. Now races (instance of ATME.C_Race) are disable by default.
				-> Remove ATME.staticObjects structure, replace by new ATME.C_StaticCloningContext classe. Useful to spawn different static objects.
				-> Remove setGroupTracking and resetGroupTracking functions from ATME.C_ZonesMgr. TNow the same functions are used for units or groups
				-> Remove setWaypointsTracking, resetWaypointsTracking, isWaypointsTracking and isFullWaypointsTracking functions from ATME.C_Group.
				-> Add new addTrackingCallback and removeTrackingCallback functions to ATME.C_Group for waypoints tracking
				-> Add new addNewRouteCallback and removeNewRouteCallback functions to ATME.C_Group when group's route has been modified by setRoute function
				-> Rename ATME.C_MultiSegment2D to ATME.C_Segment2D. Now this class managed single 2D segment instances.
				-> Remove setTracking and resetTracking functions from ATME.C_Segment2D, replaced by addTrackingCallback and removeTrackingCallback for units and groups
				-> Remove setTracking and resetTracking functions from ATME.C_ZonesMgr, replaced by addTrackingCallback and removeTrackingCallback for units and groups
				-> Remove ZONES_MGR_GROUP_ENTERS, ZONES_MGR_GROUP_LEAVES, ZONES_MGR_UNIT_ENTERS, ZONES_MGR_UNIT_LEAVES, CROSSED_SEGMENTS and TRACKING_WAYPOINTS events
				-> Rename ATME.C_CoreEvent into ATME.C_Event
				-> Remove createAbsoluteAlarm and createFlagRelativeAlarm function from ATME.C_Module class
				-> Remove getAlarmByName and removeAlarm function from ATME.C_Module class
				-> Add new constructor on ATME.C_Alarm to create absolute and flag relative alarms
				-> Add new ATME.C_Alarm.getByName function to get an alarm by its name
				-> Remove all createNamed... functions and all get...ByName functions of ATME.C_Module class. Transfered to appropriate classes (see below)
				-> Add createNamed and getByName static functions to ATME.C_Vector3D, ATME.C_Line2D, ATME.C_Segment2D, ATME.C_Task, ATME.C_ZonesMgr, ATME.C_CloningContext,
				   ATME.C_StaticCloningContext, ATME.C_FindContext classes.
				-> Remove addEventCallback and removeEventCallback functions from ATME.C_Module class
				-> Add addEventCallback and removeEventCallback functions to ATME.C_AIUnit and ATME.C_Player classes
				-> Remove ATME.C_EventsMgr class, replace by appropriate callbacks (addEventCallback and removeEventCallback, see above)
				-> onTimerHandler of ATME.C_Module class don't have any parameter. Events are now managed with callbacks (see above)
				-> Add new ATME.C_Alarm.createNamed function to create a named alarm. Now alarms can be nonamed
				-> Remove getModuleName from ATME.C_Alarm
				-> Add new getBissectrix and getSegment2D functions into ATME.C_Line2D
				-> Now getPerpendicular function of ATME.C_Line2D has a optionnal parameter. Default, perpendicular pass through origin
				-> Add new "MARK_ADDED" and "MARK_REMOVED" events
				-> Remove addWeaponEvents and removeWeaponEvents from ATME.C_Player and AMTE.C_AIUnit classes
				-> Rename ATME.isObjectClass into ATME.isObject. Now class parameter is optionnal
				-> Add new isActivated function in ATME.C_Player class
				-> Add new isInZones, isInDCSZone, isInZone2D, isInZone3D functions to ATME.C_Airbase and ATME.C_StaticObject classes
				-> Add new getAttributes, hasAttributes functions to ATME.C_Airbase and ATME.C_StaticObject classes
				-> ATME.C_Airbase instances and DCS zones are now references like static objets or units
				-> Remove ATME.getPlayers, ATME.getGroups, ATME.getAirbases, ATME.getAIUnits, ATME.getStaticObjects functions
				-> Add new getCategoryName function to ATME.C_AIUnit and ATME.C_Player classes
				-> When returning false, Find context callback avoid record item in the found items list (useful for specific filtering)
				-> Rename ATME.arePointsVisible function into ATME.areVisible
				-> Add ATME.C_F10Menu instance for first parameter in menu item callback fct(menu, args)
				-> Add new getUserPrivateDatas function to ATME.C_Segment2D, ATME.C_Line2D, ATME.C_FindContext, ATME.C_StaticContext,
				   ATME.C_Circle, ATME.C_Box, ATME.C_PickupZone, ATME.C_Cylinder, ATME.C_Polygon, ATME.C_Sphere, ATME.C_Rectangle,
				   ATME.C_F10Menu, ATME.C_Vector3D, C_Airbase
				-> Add new getPlayer function to ATME.C_F10Menu class
				-> Add new isEngineStarted function to ATME.C_AIUnit. Always true for ground units or ships
				-> Add new ATME.effectBigSmoke function
				-> Add new ATME.C_Segments2D.createSegments function to create several segments from points or centers of DCS zone
				-> Add new addReadyToBoardCallback and removeReadyToBoardCallback functions to ATME.C_Group
				-> Add new ATME.getOnRoadPoint function
				-> Add getParking function to ATME.C_Airbase class
				-> Add new ATME.C_Weapon class for weapon management
				-> Add new ATME.C_SceneryObject class for scenery object management (map's objects)
				-> output and outputVar functions of ATME.C_Module have their level parameter optionnel. If nil, level = 1
				-> Remove ATME.setDebugLevel function
				-> Add new ATME.logDisplay function to display level 1 logs.
				   New rules : 
						Level 1 -> display if module in debug mode and ATME.logDisplay called when started, else log only in file
						Level 2 -> Always log in file
						Level 3 -> log in file if in debug mode only
				-> Add new ATME.C_Cargo class to manage cargos
				-> Add new ATME.C_CargoCloningContext for cargo spawning
				-> Add new onCreateCargoHandler and onDeleteCargoHandler functions in ATME.C_Module
				->add getDCSID for Group, unit, player, static cargo et airbase
				

			DO NOT MODIFY ANYTHING - For any idea or bug, please send a message on the forum
			
			Notes :
			-----
			ATME function initObjectsInMission is based on existing MIST function to initialize mission context
			
			All comments in that file are in french because of ATMECore software debugging. If you want 
			informations about ATME, classes, functions, events or handlers, please refer to english documentation.
]]--


local versionATME = "V2.0.0"


----------------------------------------------------------------------------------------------------------
-- DEFINITION DES VALEURS CONSTANTES & VARIABLES GLOBALES
----------------------------------------------------------------------------------------------------------

-- Données locales à ATME y compris descripteurs de classe créés ultérieurement
local ATMEDatas = {
	exitATME = false,					-- Flag de sortie d'ATME, en cas d'erreur. cf ATMEFct.errorOutputCore
	debugSwitch = true,				-- Activation ou non du mode affichage debug pour le module Core
	error = nil,						-- Erreur : code en cas de retour nil sur une fonction.
	files = {},							-- si le fichier MissionScripting.lua de DCS ..\Scripts a été modifié,
										-- tableau de gestion des fichiers
	language = "FR",					-- Langue utilisée par défaut
	racesActivation = false,			-- Activation des courses, par défaut pas de course
	isRunning = false,					-- isRunning est à false jusqu'au lancement par ATME.run après chargement des modules
	theatreName = "",
	logDisplay = false,					-- Log niveau 1 affiché ou non à l'écran
	currentModuleName = "",				-- Nom du module courant : Mis à jour lors des appels aux events ou handlers d'un module
	localMembers = {},
	--newPlayer = false,					-- Contournement bug DCS pour les loop radio transmissions (bug DCS à priori corrigé, à confirmer), retourner le code pour supprimer les commentaires si besoin
	--autoRestartDCS = false				-- Contournement bug DCS pour les loop radio transmissions (bug DCS à priori corrigé, à confirmer), à remettre en oeuvre si besoin, cause inopérant
	allInitialGroupNames = {},			-- Contient les nom des groupes initiaux : de la mission ou premier spawn de création directe hors copie.
	allGroupsInMission = {},			-- Fixe les groupes définis dans la mission
	allAIUnitsInMission = {},			-- Fixe les unités AI définis dans la mission
	allStaticObjectsInMission = {},		-- Fixe les objets définis dans la mission
	allCargosInMission = {},			-- Fixe les cargos définis dans la mission
	allPlayerUnitsInMission = {},		-- Fixe les unités joueur définis dans la mission
	allStaticAirbasesInMission = {},	-- Fixe et pas de spawn sur les airfields et FARP
	allAIUnitsAndPlayers = {},			-- Liste de toutes les unités actives joueurs et AI.
	players = {},						-- Liste des joueurs actifs
	oldNbPlayers = 0,
	nbPlayers = 0,
	groups = {},						-- Groupes actifs
	AIUnits = {},						-- AIUnit actifs
	spawnedGroups = {},					-- Table de gestion des id XXX max de groupes avec suffixe "Spawned #XXX", une entrée pour un nom de groupe donné
	spawnedUnits = {},					-- Table de gestion des id XXX max de unités avec suffixe "Spawned #XXX", une entrée pour un nom d'unité donné
	spawnedStaticObjects = {},			-- Table de gestion des id XXX max d'objets statiques avec suffixe "Spawned #XXX", une entrée pour un nom d'object statique donné
	spawnedCargos = {},					-- Table de gestion des id XXX max des cargos avec suffixe "Spawned #XXX", une entrée pour un nom de cargos donné
	repeatSpawnGroupId = 1,            	-- Nombre total de groupes spawnés en répétition automatique : permet de gérer l'ID repeat des groupes et unités
	staticObjects = {},					-- Objets statiques actifs
	cargos = {},						-- Cargos actifs
	races = {},							-- Courses créées
	airbasesById = {},					-- Liste des airbases du théatre indexés par ID (y compris FARP et ship)
	airbases = {
		["SHIP"] = {},
		["HELIPAD"] = {},
		["AIRBASE"] = {},
	},									-- Liste des airbases (y compris FARP et ship)
	modules = {},						-- Liste nommée des modules (utilisée pour ATME & les classes disposant de la variable modules)
	modulesLocalMembers = {},			-- Liste des localMembers des modules, indexés par le nom du module.
	modulesNames = {},					-- Liste indexée stockant les noms des modules de la mission
	refModules = {},					-- Table référencant les noms de fichier temporaires des modules dans la mission
	readyToBoardGroups = {},			-- Liste des groupes prêts à embarquer
	respawnedGroups = {},				-- Liste des noms de groupes à respawn auto ayant été détruits
	groupsWithNewRouteSet = {},			-- Liste des groupes ayant eu leur route modifiée
	spawnedGroupsWaypoints = {},		-- Liste pour mémoriser les WP des groupes spawnés.
	groupsOnBoard = {},					-- Liste des groupes embarqués, en cas de destruction du transport
	groupsToRespawn = {}, 				-- Liste des groupes à respawner au prochain cycle timer
	staticObjectsToBirth = {},			-- Liste des objets statiques spawnés et non encore créés suite à S_EVENT_BIRTH
	cargosToBirth = {},					-- Liste des cargos spawnés et non encore créés suite à S_EVENT_BIRTH
	spawnGroupDatasRepeat = {},			-- Liste des données pour les groupes datas à multiple spawn
	spotsToReassign = {},				-- Spots à réassigner après changement de cibles
	missionsDelayed = {},				-- Missions (taches) à créer avec un délai
	idDCSRadioTransmissions = 0,		-- Compteur de création des radios transmissions
	locked = false,						-- Verrouillage de toutes les fréquences
	lockedRadioFrequencies = {},		-- Table des radios fréquences verrouillées : Emission en cours, pas de coupure
	radioFrequencyId = {},				-- Table de l'ID pour une fréquence radio donnée
	patrolsToUpdate = {}, 				-- Table de définition des patrouilles en tracking
	markId = 0,							-- Identifiant pour les marqueurs de la carte F10
	totalCharge = 0,					-- Valeur pour le calcul de charge dans le main timer pour lisser (traitements sécables)
	sounds = {},						-- Sons créés dans ATME.
	radioFrequencies = {},				-- Liste des radio-fréquences indexées par fréquence et modulation
	unitsReadyToCreate = {},			-- Unités ayant déclenché un S_EVENT_BIRTH pendant un addGroup. Parfois le groupe DCS n'est
										-- alors pas encore initialisé. Aussi, elles sont ensuite intégrées à ATME dès la fin d'exécution
										-- de addgroup.
	groupsReadyToBirth = {},			-- Pour les groupes aircraft, le S_EVENT_BIRTH est déclenché bien après le addGroup. Cette table
										-- permet la mémorisation des informations qui sera utilisée lors du S_EVENT_BIRTH
	unitsReadyToBirth = {},				-- Pour les groupes aircrafts, les unités associées sont mémorisées pour éviter les doublons
	newCreatedGroups = {},				-- Liste de gestion des groupes créés et/ou spawnés entre deux cycles timer. Ces listes sont
										-- mises à jour dans la fonction lors de la création d'un groupe pour ATMEDatas.newCreatedGroups. 
										-- Cette table est aussi utilisée pour le spawn avec une key isSpawned mise à jour.
	newCreatedCargos = {},				-- Idem pour les cargos
	newCreatedStatics = {},				-- Idem pour les statics
	newPreparedGroups = {},				-- Liste des groupes préparés, nécessaire lors de l'initialisation
	newPreparedCargos = {},				-- Idem pour les cargos
	newPreparedStatics = {},			-- idem pour les statics
	unitsToDisable = {},
	groupsToDisable = {},
	staticsToDisable = {},
	cargosToDisable = {},
	startOffset = 0,					-- Délai d'attente au démarrage avant initialisation
	startMissionOk = 0,					-- Vaut 0 avant initialisation, 1 lorsque tout est ok et avant onStartMissionHandler, et 2 après.
	marks = {},							-- Marques créées dans la mission, indexées par ID. ID géré par ATME
	weapons = {},
	sceneryObjects = {},
	idRoute = 0,						-- Identifiant de route, utilisé lors d'un setRoute pour ne pas traiter l'ancienne route empilée dans stackTimer
}

-- Tables des fonctions locales à ATME core

local ATMEAIIndFct = {}
local ATMEFct = {}
local ATMETimers = {}

local ATMEConst = {
	smokeColors = {
		["GREEN"] = trigger.smokeColor.Green,
		["RED"] = trigger.smokeColor.Red,
		["WHITE"] = trigger.smokeColor.White,
		["ORANGE"] = trigger.smokeColor.Orange,
		["BLUE"] = trigger.smokeColor.Blue,
		["RANDOM"] = -1,
	},
	flareColors = {
		["GREEN"] = trigger.flareColor.Green,
		["RED"] = trigger.flareColor.Red,
		["WHITE"] = trigger.flareColor.White,
		["YELLOW"] = trigger.flareColor.Yellow,
		["RANDOM"] = -1,
	},
	flareColorsIndex = {
		"GREEN",
		"RED",
		"WHITE",
		"YELLOW"
	},
	smokeColorsIndex = {
		"GREEN",
		"RED",
		"WHITE",
		"ORANGE",
		"BLUE",
	},
	bigSmokeType = {
		["SMALL"] = 5,
		["MEDIUM"] = 6,
		["LARGE"] = 7,
		["HUGE"] = 8,
	},
	groupCategories = {
		["plane"] = Group.Category.AIRPLANE,
		["helicopter"] = Group.Category.HELICOPTER,
		["vehicle"] = Group.Category.GROUND,
		["ship"] = Group.Category.SHIP,
	},
	JTACStatus = {		-- Status pour test JTAC dans C_Task_getTaskDatas
		["plane"] = 2,
		["helicopter"] = 2,
		["vehicle"] = 1,
		["ship"] = 0,
	},
	categoryNames = {
		[Group.Category.AIRPLANE] = "AIRPLANE",
		[Group.Category.HELICOPTER] = "HELICOPTER",
		[Group.Category.GROUND] = "GROUND",
		[Group.Category.SHIP] = "SHIP",
	},
	categoryNameFilters = {
		["AIRPLANE"] = "plane",
		["HELICOPTER"] = "helicopter",
		["GROUND"] = "vehicle",
		["SHIP"] = "ship",
	},
	canFly = {
		["AIRPLANE"] = true,
		["HELICOPTER"] = true,
		["GROUND"] = false,
		["SHIP"] = false,
	},
	altRouteSet = {
		["plane"] = false,
		["helicopter"] = false,
		["vehicle"] = true,
		["ship"] = true,
	},
	fctRadius = {
		["AIRPLANE"] = function (_action)
			if action == "Fly Over Point" then
				return 500
			else
				return 3500
			end
		end,
		["HELICOPTER"] = function (_action)
			if action == "Fly Over Point" then
				return 300
			else
				return 1000
			end
		end,
		["GROUND"] = function (_action)
			if action == "On Road" then
				return 10
			else
				return 20
			end
		end,
		["SHIP"] = function (_action)
			return 50
		end,
	},
	coalitionNames = {
		[coalition.side.RED] = "RED",
		[coalition.side.BLUE] = "BLUE",
		[coalition.side.NEUTRAL] = "NEUTRAL",
		[3] = "NEUTRAL",							-- Bug DCS si l'aéroport est à la fois bleu et rouge en début de mission
	},
	coalitionIds = {
		["RED"] = coalition.side.RED,
		["BLUE"] = coalition.side.BLUE,
		["NEUTRAL"] = coalition.side.NEUTRAL,
	},
	coalitionIdsBlueRed = {
		["RED"] = coalition.side.RED,
		["BLUE"] = coalition.side.BLUE,
	},
	coalitionEnemy = {
		["BLUE"] = "RED",
		["RED"] = "BLUE",
	},
	countries = {
		["RUSSIA"] = 0,
		["UKRAINE"] = 1,
		["USA"] = 2,
		["TURKEY"] = 3,
		["UK"] = 4,
		["FRANCE"] = 5,
		["GERMANY"] = 6,
		["AGGRESSORS"] = 7,
		["CANADA"] = 8,
		["SPAIN"] = 9,
		["THE_NETHERLANDS"] = 10,
		["BELGIUM"] = 11,
		["NORWAY"] = 12,
		["DENMARK"] = 13,
		["ISRAEL"] = 15,
		["GEORGIA"] = 16,
		["INSURGENTS"] = 17,
		["ABKHAZIA"] = 18,
		["SOUTH_OSETIA"] = 19,
		["ITALY"] = 20,
		["AUSTRALIA"] = 21,
		["SWITZERLAND"] = 22,
		["AUSTRIA"] = 23,
		["BELARUS"] = 24,
		["BULGARIA"] = 25,
		["CHEZH_REPUBLIC"] = 26,
		["CHINA"] = 27,
		["CROATIA"] = 28,
		["EGYPT"] = 29,
		["FINLAND"] = 30,
		["GREECE"] = 31,
		["HUNGARY"] = 32,
		["INDIA"] = 33,
		["IRAN"] = 34,
		["IRAQ"] = 35,
		["JAPAN"] = 36,
		["KAZAKHSTAN"] = 37,
		["NORTH_KOREA"] = 38,
		["PAKISTAN"] = 39,
		["POLAND"] = 40,
		["ROMANIA"] = 41,
		["SAUDI_ARABIA"] = 42,
		["SERBIA"] = 43,
		["SLOVAKIA"] = 44,
		["SOUTH_KOREA"] = 45,
		["SWEDEN"] = 46,
		["SYRIA"] = 47,
		["YEMEN"] = 48,
		["VIETNAM"] = 49,
		["VENEZUELA"] = 50,
		["TUNISIA"] = 51,
		["THAILAND"] = 52,
		["SUDAN"] = 53,
		["PHILIPPINES"] = 54,
		["MOROCCO"] = 55,	
		["MEXICO"] = 56,	
		["MALAYSIA"] = 57,		
		["LIBYA"] = 58,	
		["JORDAN"] = 59,
		["INDONESIA"] = 60,	
		["HONDURAS"] = 61,		
		["ETHIOPIA"] = 62,		
		["CHILE"] = 63,		
		["BRAZIL"] = 64,		 
		["BAHRAIN"] = 65,
		["THIRDREICH"] = 66,
		["YUGOSLAVIA"] = 67,
		["USSR"] = 68,
		["ITALIAN_SOCIAL_REPUBLIC"] = 69,
		["ALGERIA"] = 70,
		["KUWAIT"] = 71,
		["QATAR"] = 72,
		["OMAN"] = 73,
		["UNITED_ARAB_EMIRATES"] = 74,
	},
	personnelCarrierAttributes = {
		["UH-1H"] = {maxUnits = 8, minDistance = 25,},
		["Mi-8MT"] = {maxUnits = 16, minDistance = 28,},
		["SA342M"] = {maxUnits = 2, minDistance = 22,},
		["SA342L"] = {maxUnits = 2, minDistance = 22,},
		["SA342Mistral"] = {maxUnits = 2, minDistance = 22,},
		["AAV7"] = {maxUnits = 25, minDistance = 20,},
		["M-113"] = {maxUnits = 11, minDistance = 20,},
		["LAV-25"] = {maxUnits = 6, minDistance = 20,},
		["M1126 Stryker ICV"] = {maxUnits = 9, minDistance = 20,},
		["M-2 Bradley"] = {maxUnits = 6, minDistance = 20,},
		["BTR-80"] = {maxUnits = 9, minDistance = 20,},
		["BTR_D"] = {maxUnits = 12, minDistance = 20,},
		["MTLB"] = {maxUnits = 10, minDistance = 20,},
		["BMD-1"] = {maxUnits = 4, minDistance = 20,},
		["BMP-1"] = {maxUnits = 8, minDistance = 20,},
		["BMP-2"] = {maxUnits = 7, minDistance = 20,},
		["BMP-3"] = {maxUnits = 7, minDistance = 20,},
		["UAZ-469"] = {maxUnits = 6, minDistance = 20,},
		["Tigr_233036"] = {maxUnits = 6, minDistance = 20,},
		["GAZ-3307"] = {maxUnits = 16, minDistance = 20,},
		["GAZ-3308"] = {maxUnits = 16, minDistance = 20,},
		["Ural-4320T"] = {maxUnits = 20, minDistance = 20,},
		["Ural-4320-31"] = {maxUnits = 20, minDistance = 20,},
		["M 818"] = {maxUnits = 20, minDistance = 20,},
		["KAMAZ Truck"] = {maxUnits = 20, minDistance = 20,},
	},
	cargos = {
		["BASIC"] = {
			type = "uh1h_cargo",
			shape = "ab-212_cargo",
		},
		["TRUNKS SMALL"] = {
			type = "trunks_small_cargo",
			shape = "trunks_small_cargo",
		},
		["TRUNKS LONG"] = {
			type = "trunks_long_cargo",
			shape = "trunks_long_cargo",
		},
		["TETRAPOD"] = {
			type = "tetrapod_cargo",
			shape = "tetrapod_cargo",
		},
		["PIPE SMALL"] = {
			type = "pipes_small_cargo",
			shape = "pipes_small_cargo",
		},
		["PIPE BIG"] = {
			type = "pipes_big_cargo",
			shape = "pipes_big_cargo",
		},
		["OIL TANK"] = {
			type = "oiltank_cargo",
			shape = "oiltank_cargo",
		},
		["M117"] = {
			type = "m117_cargo",
			shape = "m117_cargo",
		},
		["SMALL CONTAINER"] = {
			type = "iso_container_small",
			shape = "iso_container_small_cargo",
		},
		["F BAR"] = {
			type = "f_bar_cargo",
			shape = "f_bar_cargo",
		},
		["FUEL TANK"] = {
			type = "fueltank_cargo",
			shape = "fueltank_cargo",
		},
		["BW CONTAINER"] = {
			type = "container_cargo",
			shape = "bw_container_cargo",
		},
		["CONTAINER"] = {
			type = "iso_container",
			shape = "iso_container_cargo",
		},
		["BARRELS"] = {
			type = "barrels_cargo",
			shape = "barrels_cargo",
		},
		["AMMO BOX"] = {
			type = "ammo_cargo",
			shape = "ammo_box_cargo",
		},
	},
	staticObjectCategories = {
		-- Bug DCS sur certains objets dont le nom est toujours 'static'
		-- LTAvehicles/Effects ne fonctionnent pas en spawn.
		
		["SHIP"]= {
			category = "Ships",
		},
		["AIRPLANE"]= {
			category = "Planes",
		},
		["HELICOPTER"]= {
			category = "Helicopters",
		},
		["WAREHOUSE"]= {
			category = "Warehouses",
		},
		["GROUND"]= {
			category = "Unarmed",
		},
		["FORTIFICATION"]= {
			category = "Fortifications",
		},
	},
	backStaticObjectCategories = {
		["Ships"] = "SHIP",
		["Planes"] = "AIRPLANE",
		["Helicopters"] = "HELICOPTER",
		["Unarmed"] = "GROUND",
		["Warehouses"] = "WAREHOUSE",
		["Fortifications"] = "FORTIFICATION",
	},
	supportedLanguages = {
		["FR"] = "French",
		["EN"] = "English",
		["RUS"] = "Russian",
		["GER"] = "German",
	},
	specialAircraftTypes = {
		["E-3A"] = "awacs",
		["E-2C"] = "awacs",
		["KC-135"] = "tankers",
		["KC135MPRS"] = "tankers",
		["KC130"] = "tankers",
		["A-10A"] = "A10",
		["A-10C"] = "A10",
	},
	blueCallsigns = {
		["aircrafts"] = {
			["ENFIELD"] = {
				name = "Enfield",
				id = 1,
				type = "standard",
			},
			["SPRINGFIELD"] = {
				name = "SpringField",
				id = 2,
				type = "standard",
			},
			["UZI"] = {
				name = "Uzi",
				id = 3,
				type = "standard",
			},
			["COLT"] = {
				name = "Colt",
				id = 4,
				type = "standard",
			},
			["DODGE"] = {
				name = "Dodge",
				id = 5,
				type = "standard",
			},
			["FORD"] = {
				name = "Ford",
				id = 6,
				type = "standard",
			},
			["CHEVY"] = {
				name = "Chevy",
				id = 7,
				type = "standard",
			},
			["PONTIAC"] = {
				name = "Pontiac",
				id = 8,
				type = "standard",
			},
			["HAWG"] = {
				name = "Hawg",
				id = 9,
				type = "A10",
			},
			["BOAR"] = {
				name = "Boar",
				id = 10,
				type = "A10",
			},
			["PIG"] = {
				name = "Pig",
				id = 11,
				type = "A10",
			},
			["TUSK"] = {
				name = "Tusk",
				id = 12,
				type = "A10",
			},
			["TEXACO"] = {
				name = "Texaco",
				id = 1,
				type = "tankers",
			},
			["ARCO"] = {
				name = "Arco",
				id = 2,
				type = "tankers",
			},
			["SHELL"] = {
				name = "Shell",
				id = 3,
				type = "tankers",
			},
			["OVERLORD"] = {
				name = "Overlord",
				id = 1,
				type = "awacs",
			},
			["MAGIC"] = {
				name = "Magic",
				id = 2,
			},
			["WIZARD"] = {
				name = "Wizard",
				id = 3,
				type = "awacs",
			},
			["FOCUS"] = {
				name = "Focus",
				id = 4,
				type = "awacs",
			},
			["DARKSTAR"] = {
				name = "Darkstar",
				id = 5,
				type = "awacs",
			},
		},
		["jtacs"] = {
			["AXEMAN"] = {
				name = "Axeman",
				id = 1,
			},
			["DARKNIGHT"] = {
				name = "Darknight",
				id = 2,
			},
			["WARRIOR"] = {
				name = "Warrior",
				id = 3,
			},
			["POINTER"] = {
				name = "Pointer",
				id = 4,
			},
			["EYEBALL"] = {
				name = "Eyeball",
				id = 5,
			},
			["MOONBEAM"] = {
				name = "Moonbeam",
				id = 6,
			},
			["WHIPLASH"] = {
				name = "Whiplash",
				id = 7,
			},
			["FINGER"] = {
				name = "Finger",
				id = 8,
			},
			["PINPOINT"] = {
				name = "Pinpoint",
				id = 9,
			},
			["FERRET"] = {
				name = "Ferret",
				id = 10,
			},
			["SHABA"] = {
				name = "Shaba",
				id = 11,
			},
			["PLAYBOY"] = {
				name = "Playboy",
				id = 12,
			},
			["HAMMER"] = {
				name = "Hammer",
				id = 13,
			},
			["JAGUAR"] = {
				name = "Jaguar",
				id = 14,
			},
			["DEATHSTAR"] = {
				name = "Deathstar",
				id = 15,
			},
			["ANVIL"] = {
				name = "Anvil",
				id = 16,
			},
			["FIREFLY"] = {
				name = "Firefly",
				id = 17,
			},
			["MANTIS"] = {
				name = "Mantis",
				id = 18,
			},
			["BADGER"] = {
				name = "Badger",
				id = 19,
			},
		},
	},
	moduleHandlers = {
		["onCreatePlayerHandler"] = true,
		["onDeletePlayerHandler"] = true,
		["onUpdatePlayerHandler"] = true,
		["onBaseCapturedPlayerHandler"] = true,
		["onCreateAIUnitHandler"] = true,
		["onDeleteAIUnitHandler"] = true,
		["onBaseCapturedAIUnitHandler"] = true,
		["onCreateGroupHandler"] = true,
		["onSpawnGroupHandler"] = true,
		["onDeleteGroupHandler"] = true,
		["onCreateStaticObjectHandler"] = true,
		["onDeleteStaticObjectHandler"] = true,
		["onCreateCargoHandler"] = true,
		["onDeleteCargoHandler"] = true,
		["onTimerHandler"] = true,
		["onStartMissionHandler"] = true,
		["onEndMissionHandler"] = true,
	},
	directionValue = {					-- Table de gestion des directions
		[-1] = "RIGHT_TO_LEFT",
		[1] = "LEFT_TO_RIGHT",
	},
	weaponFlags = {
		["NO_WEAPON"] = 0,
		["LGB"] = 2,
		["TVGB"] = 4,
		["SNSGB"] = 8,
		["HE_BOMB"] = 16,
		["PENETRATOR"] = 32,
		["NAPALM_BOMB"] = 64,
		["FAE_BOMB"] = 128,
		["CLUSTER_BOMB"] = 256,
		["DISPENCER"] = 512,
		["CANDLE_BOMB"] = 1024,
		["PARACHUTE_BOMB"] = 2147483648,
		["LIGHT_ROCKET"] = 2048,
		["MARKER_ROCKET"] = 4096,
		["CANDLE_ROCKET"] = 8192,
		["HEAVY_ROCKET"] = 16384,
		["ANTI_RADAR_MISSILE"] = 32768,
		["ANTI_SHIP_MISSILE"] = 65536,
		["ANTI_TANK_MISSILE"] = 131072,
		["FIRE_AND_FORGET_ASM"] = 262144,
		["LASER_ASM"] = 524288,
		["TELE_ASM"] = 1048576,
		["CRUISE_MISSILE"] = 2097152,
		["ANTI_RADAR_MISSILE_2"] = 1073741824,
		["LR_AAM"] = 16777216,
		["SR_AAM"] = 4194304,
		["MR_AAM"] = 8388608,
		["IR_AAM"] = 33554432,
		["SAR_AAM"] = 67108864,
		["AR_AAM"] = 134217728,
		["GUN_POD"] = 268435456,
		["BUILT_IN_CANNON"] = 536870912,
		["ANY_WEAPON"] = 3221225470,
		["MARKER_WEAPON"] = 13312,
		["ARM_WEAPON"] = 3221212158,
	},
	weaponExpend = {
		["QUARTER"] = AI.Task.WeaponExpend.QUARTER,
		["TWO"] = AI.Task.WeaponExpend.TWO,
		["ONE"] = AI.Task.WeaponExpend.ONE,
		["FOUR"] = AI.Task.WeaponExpend.FOUR,
		["HALF"] = AI.Task.WeaponExpend.HALF,
		["ALL"] = AI.Task.WeaponExpend.ALL,
	},
	orbitPattern = {
		["CIRCLE"] = AI.Task.OrbitPattern.CIRCLE,
		["RACE_TRACK"] = AI.Task.OrbitPattern.RACE_TRACK,
	},
	designationAI = {
		["NO"] = AI.Task.Designation.NO,
		["WP"] = AI.Task.Designation.WP,
		["IR_POINTER"] = AI.Task.Designation.IR_POINTER,
		["LASER"] = AI.Task.Designation.LASER,
		["AUTO"] = AI.Task.Designation.AUTO,
	},
	aircraftTypeJTAC = {
		["RQ-1A Predator"] = true,
		["MQ-9 Reaper"] = true,
	},
	targetsAttributes = {
		["PLANE_CARRIERS"] = "plane_carrier",
		["AA_FLAK"] = "AA_flak",
		["AA_MISSILES"] = "AA_missile",
		["CRUISE_MISSILES"] = "Cruise missiles",
		["ANTI_SHIP_MISSILES"] = "Anti-Ship missiles",
		["MISSILES"] = "Missiles",
		["AIR"] = "Air",
		["PLANES"] = "Planes",
		["UAVS"] = "UAVs",
		["HELICOPTERS"] = "Helicopters",
		["BATTLE_AIRPLANES"] = "Battle airplanes",
		["ATTACK_HELICOPTERS"] = "Attack helicopters",
		["FIGHTERS"] = "Fighters",
		["INTERCEPTORS"]= "Interceptors",
		["BOMBERS"] = "Bombers",
		["STRATEGIC_BOMBERS"] = "Strategic Bombers",
		["MULTIROLE_FIGHTERS"] = "Multirole fighters",
		["TRANSPORT_AIRPLANES"] = "Transports",
		["TRANSPORT_HELICOPTERS"] = "Transport helicopters",
		["AWACS"] = "AWACS",
		["TANKERS"] = "Tankers",
		["AUX"] = "Aux",
		["CARS"] = "Cars",
		["TRUCKS"] = "Trucks",
		["INFANTRY"] = "Infantry",
		["TANKS"] = "Tanks",
		["ARTILLERY"] = "Artillery",
		["MLRS"] = "MLRS",
		["IFV"] = "IFV",
		["APC"] = "APC",
		["FORTIFICATIONS"] = "Fortifications",
		["ARMED_VEHICLES"] = "Armed vehicles",
		["STATIC_AAA"] = "Static AAA",
		["MOBILE_AAA"]= "Mobile AAA",
		["SAM_SR"] = "SAM SR",
		["SAM_TR"] = "SAM TR",
		["SAM_LL"] = "SAM LL",
		["SAM_CC"] = "SAM CC",
		["SAM_AUX"] = "SAM AUX",
		["SR_SAM"] = "SR SAM",
		["MR_SAM"] = "MR SAM",
		["LR_SAM"] = "LR SAM",
		["SAM_ELEMENTS"] = "SAM Elements",
		["IR_GUIDED_SAM"] = "IR Guided SAM",
		["SAM"] = "SAM",
		["SAM_RELATED"] = "SAM related",
		["AAA"] = "AAA",
		["EWR"] = "EWR",
		["AIR_DEFENCE_VEHICLE"] = "Air Defence vehicle",
		["MANPADS"] = "MANPADS",
		["MANPADS_AUX"] = "MANPADS AUX",
		["UNARMED_VEHICLES"] = "Unarmed vehicles",
		["ARMED_GROUND_UNITS"] = "Armed ground units",
		["ARMED_AIR_DEFENCE"] = "Armed Air Defence",
		["AIR_DEFENCE"] = "Air defence",
		["AIRCRAFT_CARRIERS"] = "Aircraft Carriers",
		["CRUISERS"] = "Cruisers",
		["DESTROYERS"] = "Destroyers",
		["FRIGATES"] = "Frigates",
		["CORVETTES"] = "Corvettes",
		["HEAVY_ARMED_SHIPS"] = "Heavy armed ships",
		["LIGHT_ARMED_SHIPS"] = "Light armed ships",
		["ARMED_SHIPS"] = "Armed ships",
		["UNARMED_SHIPS"] = "Unarmed ships",
		["GROUND_VEHICLES"] = "Ground vehicles",
		["SHIPS"] = "Ships",
		["BUILDINGS"] = "Buildings",
		["HEAVY_ARMORED_UNITS"] = "HeavyArmoredUnits",
		["LIGHT_ARMORED_UNITS"] = "LightArmoredUnits",
		["ATGM"] = "ATGM",
		["OLD_TANKS"] = "Old Tanks",
		["MODERN_TANKS"] = "Modern Tanks",
		["ROCKET_ATTACK_VALID_AIR_DEFENCE"] = "Rocket Attack Valid AirDefence",
		["ALL"] = "All",
		["INFANTRY_CARRIERS"] = "Infantry carriers",
		["VEHICLES"] = "Vehicles",
		["GROUND_UNITS"] = "Ground Units",
		["GROUND_UNITS_NON_AIR_DEFENCE"] = "Ground Units Non Airdefence",
		["ARMORED_VEHICLES"] = " Armored vehicles",
		["ANTI_AIR_ARMED_VEHICLES"] = "AntiAir Armed Vehicles",
		["AIRBASES"] = "Airfields",
		["HELIPADS"] = "Helipad",
		["GRASS_AIRFIELDS"] = "Grass Airfields",
		["POINT"] = "Point",
		["NON_ARMORED_UNITS"] = "NonArmoredUnits",
		["NON_AND_LIGHT_ARMORED_UNITS"] = "NonAndLightArmoredUnits",
		["HUMAN_VEHICLE"] = "Human_vehicle",
		["RADAR_BAND1_FOR_ARM"] = "RADAR_BAND1_FOR_ARM",
		["RADAR_BAND2_FOR_ARM"] = "RADAR_BAND2_FOR_ARM",
		["PRONE"] = "Prone",
		["DETECTION_BY_AWACS"] = "DetectionByAWACS",
		["DATALINK"] = "Datalink",
		["CUSTOM_AIM_POINT"] = "CustomAimPoint",
		["INDIRECT_FIRE"] = "Indirect fire",
		["REFUELABLE"] = "Refuelable",
		["WEAPON"] = "Weapon",
	},
	backTargetsAttributes = {
		["plane_carrier"] = {
			value = "PLANE_CARRIERS",
			priority = 0,
		},
		["AA_flak"] = {
			value = "AA_FLAK",
			priority = 0,
		},
		["AA_missile"] = {
			value = "AA_MISSILES",
			priority = 0,
		},
		["Cruise missiles"] = {
			value = "CRUISE_MISSILES",
			priority = 0,
		},
		["Anti-Ship missiles"] = {
			value = "ANTI_SHIP_MISSILES",
			priority = 0,
		},
		["Missiles"] = {
			value = "MISSILES",
			priority = 0,
		},
		["Air"] = {
			value = "AIR",
			priority = 0,
		},
		["Planes"] = {
			value = "PLANES",
			priority = 0,
		},
		["UAVs"] = {
			value = "UAVS",
			priority = 0,
		},
		["Helicopters"] = {
			value = "HELICOPTERS",
			priority = 0,
		},
		["Battle airplanes"] = {
			value = "BATTLE_AIRPLANES",
			priority = 2,
		},
		["Attack helicopters"] = {
			value = "ATTACK_HELICOPTERS",
			priority = 2,
		},
		["Fighters"] = {
			value = "FIGHTERS",
			priority = 0,
		},
		["Interceptors"] = {
			value = "INTERCEPTORS",
			priority = 0,
		},
		["Bombers"] = {
			value = "BOMBERS",
			priority = 0,
		},
		["Strategic Bombers"] = {
			value = "STRATEGIC_BOMBERS",
			priority = 0,
		},
		["Multirole fighters"] = {
			value = "MULTIROLE_FIGHTERS",
			priority = 0,
		},
		["Transports"] = {
			value = "TRANSPORT_AIRPLANES",
			priority = 0,
		},
		["Transport helicopters"] = {
			value = "TRANSPORT_HELICOPTERS",
			priority = 0,
		},
		["AWACS"] = {
			value = "AWACS",
			priority = 0,
		},
		["Tankers"] = {
			value = "TANKERS",
			priority = 0,
		},
		["Aux"] = {
			value = "AUX",
			priority = 0,
		},
		["Cars"] = {
			value = "CARS",
			priority = 0,
		},
		["Trucks"] = {
			value = "TRUCKS",
			priority = 0,
		},
		["Infantry"] = {
			value = "INFANTRY",
			priority = 0,
		},
		["Tanks"] = {
			value = "TANKS",
			priority = 0,
		},
		["Artillery"] = {
			value = "ARTILLERY",
			priority = 0,
		},
		["MLRS"] = {
			value = "MLRS",
			priority = 0,
		},
		["IFV"] = {
			value = "IFV",
			priority = 0,
		},
		["APC"] = {
			value = "APC",
			priority = 0,
		},
		["Fortifications"] = {
			value = "FORTIFICATIONS",
			priority = 0,
		},
		["Armed vehicles"] = {
			value = "ARMED_VEHICLES",
			priority = 0,
		},
		["Static AAA"] = {
			value = "STATIC_AAA",
			priority = 0,
		},
		["Mobile AAA"] = {
			value = "MOBILE_AAA",
			priority = 0,
		},
		["SAM SR"] = {
			value = "SAM_SR",
			priority = 5,
		},
		["SAM TR"] = {
			value = "SAM_TR",
			priority = 9,
		},
		["SAM LL"] = {
			value = "SAM_LL",
			priority = 1,
		},
		["SAM CC"] = {
			value = "SAM_CC",
			priority = 0,
		},
		["SAM AUX"] = {
			value = "SAM_AUX",
			priority = 0,
		},
		["SR SAM"] = {
			value = "SR_SAM",
			priority = 6,
		},
		["MR SAM"] = {
			value = "MR_SAM",
			priority = 7,
		},
		["LR SAM"] = {
			value = "LR_SAM",
			priority = 8,
		},
		["SAM Elements"] = {
			value = "SAM_ELEMENTS",
			priority = 0,
		},
		["IR Guided SAM"] = {
			value = "IR_GUIDED_SAM",
			priority = 0,
		},
		["SAM"] = {
			value = "SAM",
			priority = 0,
		},
		["SAM related"] = {
			value = "SAM_RELATED",
			priority = 0,
		},
		["AAA"] = {
			value = "AAA",
			priority = 4,
		},
		["EWR"] = {
			value = "EWR",
			priority = 5,
		},
		["Air Defence vehicle"] = {
			value = "AIR_DEFENCE_VEHICLE",
			priority = 0,
		},
		["MANPADS"] = {
			value = "MANPADS",
			priority = 0,
		},
		["MANPADS AUX"] = {
			value = "MANPADS_AUX",
			priority = 0,
		},
		["Unarmed vehicles"] = {
			value = "UNARMED_VEHICLES",
			priority = 1,

		},
		["Armed ground units"] = {
			value = "ARMED_GROUND_UNITS",
			priority = 0,
		},
		["Armed Air Defence"] = {
			value = "ARMED_AIR_DEFENCE",
			priority = 0,
		},
		["Air defence"] = {
			value = "AIR_DEFENCE",
			priority = 0,
		},
		["Aircraft Carriers"] = {
			value = "AIRCRAFT_CARRIERS",
			priority = 0,
		},
		["Cruisers"] = {
			value = "CRUISERS",
			priority = 0,
		},
		["Destroyers"] = {
			value = "DESTROYERS",
			priority = 0,
		},
		["Frigates"] = {
			value = "FRIGATES",
			priority = 0,
		},
		["Corvettes"] = {
			value = "CORVETTES",
			priority = 0,
		},
		["Heavy armed ships"] = {
			value = "HEAVY_ARMED_SHIPS",
			priority = 0,
		},
		["Light armed ships"] = {
			value = "LIGHT_ARMED_SHIPS",
			priority = 0,
		},
		["Armed ships"] = {
			value = "ARMED_SHIPS",
			priority = 0,
		},
		["Unarmed ships"] = {
			value = "UNARMED_SHIPS",
			priority = 0,
		},
		["Ground vehicles"] = {
			value = "GROUND_VEHICLES",
			priority = 0,
		},
		["Ships"] = {
			value = "SHIPS",
			priority = 0,
		},
		["Buildings"] = {
			value = "BUILDINGS",
			priority = 0,
		},
		["HeavyArmoredUnits"] = {
			value = "HEAVY_ARMORED_UNITS",
			priority = 0,
		},
		["LightArmoredUnits"] = {
			value = "LIGHT_ARMORED_UNITS",
			priority = 0,
		},
		["ATGM"] = {
			value = "ATGM",
			priority = 0,
		},
		["Old Tanks"] = {
			value = "OLD_TANKS",
			priority = 0,
		},
		["Modern Tanks"] = {
			value = "MODERN_TANKS",
			priority = 0,
		},
		["Rocket Attack Valid AirDefence"] = {
			value = "ROCKET_ATTACK_VALID_AIR_DEFENCE",
			priority = 0,
		},
		["All"] = {
			value = "ALL",
			priority = 0,
		},
		["Infantry carriers"] = {
			value = "INFANTRY_CARRIERS",
			priority = 0,
		},
		["Vehicles"] = {
			value = "VEHICLES",
			priority = 0,
		},
		["Ground Units"] = {
			value = "GROUND_UNITS",
			priority = 0,
		},
		["Ground Units Non Airdefence"] = {
			value = "GROUND_UNITS_NON_AIR_DEFENCE",
			priority = 0,
		},
		[" Armored vehicles"] = {
			value = "ARMORED_VEHICLES",
			priority = 3,
		},
		["AntiAir Armed Vehicles"] = {
			value = "ANTI_AIR_ARMED_VEHICLES",
			priority = 0,
		},
		["Airfields"] = {
			value = "AIRBASES",
			priority = 0,
		},
		["Helipad"] = {
			value = "HELIPADS",
			priority = 0,
		},
		["Grass Airfields"] = {
			value = "GRASS_AIRFIELDS",
			priority = 0,
		},
		["Point"] = {
			value = "POINT",
			priority = 0,
		},
		["NonArmoredUnits"] = {
			value = "NON_ARMORED_UNITS",
			priority = 0,
		},
		["NonAndLightArmoredUnits"] = {
			value = "NON_AND_LIGHT_ARMORED_UNITS",
			priority = 0,
		},
		["Human_vehicle"] = {
			value = "HUMAN_VEHICLE",
			priority = 0,
		},
		["RADAR_BAND1_FOR_ARM"] = {
			value = "RADAR_BAND1_FOR_ARM",
			priority = 0,
		},
		["RADAR_BAND2_FOR_ARM"] = {
			value = "RADAR_BAND2_FOR_ARM",
			priority = 0,
		},
		["Prone"] = {
			value = "PRONE",
			priority = 0,
		},
		["DetectionByAWACS"] = {
			value = "DETECTION_BY_AWACS",
			priority = 0,
		},
		["Datalink"] = {
			value = "DATALINK",
			priority = 0,
		},
		["CustomAimPoint"] = {
			value = "CUSTOM_AIM_POINT",
			priority = 0,
		},
		["Indirect fire"] = {
			value = "INDIRECT_FIRE",
			priority = 0,
		},
		["Refuelable"] = {
			value = "REFUELABLE",
			priority = 0,
		},
		["Weapon"] = {
			value = "WEAPON",
			priority = 0,
		},
	},
	findCategories = {
		["AIRPLANE"] = true,
		["HELICOPTER"] = true,
		["GROUND"] = true,
		["FORTIFICATION"] = true,
		["WAREHOUSE"] = true,
		["CARGO"] = true,
		["SHIP"] = true,
		["HELIPAD"] = true,
		["AIRBASE"] = true,
	},
	WPActions = {
		["plane"] = {
			["TURNING_POINT"] = "Turning Point",
			["FLY_OVER_POINT"] = "Fly Over Point",
		},
		["helicopter"] = {
			["TURNING_POINT"] = "Turning Point",
			["FLY_OVER_POINT"] = "Fly Over Point",
		},
		["vehicle"] = {
			["VEE"] = "Vee",
			["ECHELON_RIGHT"] = "EchelonR",
			["OFF_ROAD"] = "Off Road",
			["RANK"] = "Rank",
			["ECHELON_LEFT"]  = "EchelonL",
			["ON_ROAD"] = "On Road",
			["CONE"] = "Cone",
			["DIAMOND"] = "Diamond",
		},
		["ship"] = {
			["TURNING_POINT"] = "Turning Point",
		},
	},
	taskAI = {
		["REFUELING"] = { -- Main task
			id = "Refueling",
			filters = {
				type = {
					["plane"] = true,
				},
			},
			args = {
			},
		},
		["FOLLOW"] = { -- Main task
			id = "Follow",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["group"] = {
					parameter = {
						name = "groupId",
						fct = "C_Group_getIDWithWP",
						linkedParams = {
							"lastWptIndexFlag",
							"lastWptIndex",
						},
					},
					multi = false,
					type = "C_Group",
					mandatory = true,
				},
				["pos"] = {
					parameter = {
						name = "pos",
						fct = "C_Vector_getPos",
					},
					multi = false,
					type = "C_Vector3D",
					mandatory = true,
				},
				["lastWaypointID"] = {
					parameter = {
						name = "lastWptIndex",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"lastWptIndexFlag",
						},
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = false,
				},
			},
		},
		["FOLLOW_BIG_FORMATION"] = { -- Main task
			id = "FollowBigFormation",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["group"] = {
					parameter = {
						name = "groupId",
						fct = "C_Group_getIDWithWP",
						linkedParams = {
							"lastWptIndexFlag",
							"lastWptIndex",
						},
					},
					multi = false,
					type = "C_Group",
					mandatory = true,
				},
				["pos"] = {
					parameter = {
						name = "pos",
						fct = "C_Vector_getPos",
					},
					multi = false,
					type = "C_Vector3D",
					mandatory = true,
				},
				["lastWaypointID"] = {
					parameter = {
						name = "lastWptIndex",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"lastWptIndexFlag",
						},
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = false,
				},
			},
		},
		["ESCORT"] = { -- Main task
			id = "Escort",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["group"] = {
					parameter = {
						name = "groupId",
						fct = "C_Group_getIDWithWP",
						linkedParams = {
							"lastWptIndexFlag",
							"lastWptIndex",
						},
					},
					multi = false,
					type = "C_Group",
					mandatory = true,
				},
				["pos"] = {
					parameter = {
						name = "pos",
						fct = "C_Vector_getPos",
					},
					multi = false,
					type = "C_Vector3D",
					mandatory = true,
				},
				["maxDistance"] = {
					parameter = {
						name = "engagementDistMax",
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = true,
				},
				["targetsType"] = {
					parameter = {
						name = "targetTypes",
						fct = "C_getTargetsType",
					},
					type = "string",
					multi = true,
					mandatory = true,
				},
				["lastWaypointID"] = {
					parameter = {
						name = "lastWptIndex",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"lastWptIndexFlag",
						},
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = false,
				},
			},
		},
		["LAND"] = { -- Main task
			id = "Land",
			filters = {
				type = {
					["helicopter"] = true,
				},
			},
			args = {
				["point"] = {
					parameter = {
						name = "point",
						fct = "C_getPoint2D",
					},
					multi = false,
					type = "table",
					mandatory = true,
				},
				["duration"] = {
					parameter = {
						name = "duration",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"durationFlag",
						},
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
			},
		},
		["FIRE_AT_POINT"] = { -- Main task
			id = "FireAtPoint",
			filters = {
				type = {
					["vehicle"] = true,
					["ship"] = true,
				},
			},
			const = {
				["radius"] = 100,
			},
			args = {
				["circle"] = {
					parameter = {
						name = "point",
						fct = "C_Circle_getCenterRadius",
						linkedParams = {
							"radius",
						},
					},
					multi = false,
					type = "C_Circle",
					mandatory = true,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["qty"] = {
					parameter = {
						name = "expendQty",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"expendQtyEnabled",
						},
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
			},
		},
		["BOMBING"] = { -- Main task
			id = "Bombing",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["point"] = {
					parameter = {
						name = "point",
						fct = "C_getPoint2D",
					},
					multi = false,
					type = "table",
					mandatory = true,
				},
				["qty"] = {
					parameter = {
						name = "attackQty",
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = true,
				},

				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["expend"] = {
					parameter = {
						name = "expend",
						fct = "C_getExpend",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["direction"] = {
					parameter = {
						name = "direction",
					},
					multi = false,
					type = "number",
					mandatory = false,
				},
				["altitude"] = {
					parameter = {
						name = "altitude",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"altitudeEnabled",
						},
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
				["groupAttack"] = {
					parameter = {
						name = "groupAttack",
					},
					multi = false,
					type = "boolean",
					mandatory = false,
				},
			},
		},
		["BOMBING_RUNWAY"] = { -- Main task
			id = "BombingRunway",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["airbase"] = {
					parameter = {
						name = "runwayId",
						fct = "C_getAirbase",
					},
					multi = false,
					type = "string",
					mandatory = true,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["expend"] = {
					parameter = {
						name = "expend",
						fct = "C_getExpend",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["qty"] = {
					parameter = {
						name = "attackQty",
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
				["direction"] = {
					parameter = {
						name = "direction",
					},
					multi = false,
					type = "number",
					mandatory = false,
				},
				["groupAttack"] = {
					parameter = {
						name = "groupAttack",
					},
					multi = false,
					type = "boolean",
					mandatory = false,
				},
			},
		},
		["CARPET_BOMBING"] = { -- Main task
			id = "CarpetBombing",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			const = {
				["attackType"] = "Carpet",
			},
			args = {
				["point"] = {
					parameter = {
						name = "point",
						fct = "C_getPoint2D",
					},
					multi = false,
					type = "table",
					mandatory = true,
				},
				["qty"] = {
					parameter = {
						name = "attackQty",
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = true,
				},
				["carpetLength"] = {
					parameter = {
						name = "carpetLength",
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = true,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["expend"] = {
					parameter = {
						name = "expend",
						fct = "C_getExpend",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["altitude"] = {
					parameter = {
						name = "altitude",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"altitudeEnabled",
						},
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
				["groupAttack"] = {
					parameter = {
						name = "groupAttack",
					},
					multi = false,
					type = "boolean",
					mandatory = false,
				},
			},
		},
		["ATTACK_MAP_OBJECT"] = { -- Main task
			id = "AttackMapObject",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["point"] = {
					parameter = {
						name = "point",
						fct = "C_getPoint2D",
					},
					multi = false,
					type = "table",
					mandatory = true,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["expend"] = {
					parameter = {
						name = "expend",
						fct = "C_getExpend",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["qty"] = {
					parameter = {
						name = "attackQty",
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
				["direction"] = {
					parameter = {
						name = "direction",
					},
					multi = false,
					type = "number",
					mandatory = false,
				},
				["groupAttack"] = {
					parameter = {
						name = "groupAttack",
					},
					multi = false,
					type = "boolean",
					mandatory = false,
				},
			},
		},
		["ATTACK_GROUP"] = { -- Main task
			id = "AttackGroup",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["group"] = {
					parameter = {
						name = "groupId",
						fct = "C_Group_getID",
					},
					multi = false,
					type = "C_Group",
					mandatory = true,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["expend"] = {
					parameter = {
						name = "expend",
						fct = "C_getExpend",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["qty"] = {
					parameter = {
						name = "attackQty",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"attackQtyLimit",
						},
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
				["direction"] = {
					parameter = {
						name = "direction",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"directionEnabled",
						},
					},
					multi = false,
					type = "number",
					mandatory = false,
				},
				["altitude"] = {
					parameter = {
						name = "altitude",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"altitudeEnabled",
						},
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
			},
		},
		["ATTACK_UNIT"] = { -- Main task
			id = "AttackUnit",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["unit"] = {
					parameter = {
						name = "unitId",
						fct = "C_Unit_getID",
					},
					multi = false,
					type = "C_Unit",
					mandatory = true,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["expend"] = {
					parameter = {
						name = "expend",
						fct = "C_getExpend",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["qty"] = {
					parameter = {
						name = "attackQty",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"attackQtyLimit",
						},
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
				["direction"] = {
					parameter = {
						name = "direction",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"directionEnabled",
						},
					},
					multi = false,
					type = "number",
					mandatory = false,
				},
				["groupAttack"] = {
					parameter = {
						name = "groupAttack",
					},
					multi = false,
					type = "boolean",
					mandatory = false,
				},
			},
		},
		["FAC_ATTACK_GROUP"] = { -- Main task
			id = "FAC_AttackGroup",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
					["vehicle"] = true,
				},
			},
			args = {
				["frequency"] = {
					parameter = {
						name = "frequency",
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = true,
				},
				["modulation"] = {
					parameter = {
						name = "modulation",
						fct = "C_getModulation",
					},
					multi = false,
					type = "string",
					mandatory = true,
				},
				["callsign"] = {
					parameter = {
						name = "callname",
						fct = "C_getCallname",
						linkedParams = {
							"number",
						},
					},
					multi = false,
					type = "string",
					mandatory = true,
				},
				["group"] = {
					parameter = {
						name = "groupId",
						fct = "C_Group_getID",
					},
					multi = false,
					type = "C_Group",
					mandatory = true,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["designation"] = {
					parameter = {
						name = "designation",
						fct = "C_getDesignation",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["datalink"] = {
					parameter = {
						name = "datalink",
					},
					multi = false,
					type = "boolean",
					mandatory = false,
				},
			},
		},
		["ORBIT"] = { -- Main task
			id = "Orbit",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["pattern"] = {
					parameter = {
						name = "pattern",
						fct = "C_getOrbitPattern",
					},
					multi = false,
					type = "string",
					mandatory = true,
				},
				["point"] = {
					parameter = {
						name = "point",
						fct = "C_getPoint2D",
					},
					multi = false,
					type = "table",
					mandatory = false,
				},
				["point2"] = {
					parameter = {
						name = "point2",
						fct = "C_getPoint2D",
					},
					multi = false,
					type = "table",
					mandatory = false,
				},
				["speed"] = {
					parameter = {
						name = "speed",
					},
					multi = false,
					min = 0,
					type = "number",
					mandatory = false,
				},
				["altitude"] = {
					parameter = {
						name = "altitude",
					},
					multi = false,
					min = 0,
					type = "number",
					mandatory = false,
				},
			},
		},
		["AWACS"] = { -- Enroute task
			id = "AWACS",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
			},
		},
		["TANKER"] = { -- Enroute task
			id = "Tanker",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
			},
		},
		["ENGAGE_GROUP"] = { -- Enroute task
			id = "EngageGroup",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["group"] = {
					parameter = {
						name = "groupId",
						fct = "C_Group_getID",
					},
					multi = false,
					type = "C_Group",
					mandatory = true,
				},
				["priority"] = {
					parameter = {
						name = "priority",
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = false,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["expend"] = {
					parameter = {
						name = "expend",
						fct = "C_getExpend",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["qty"] = {
					parameter = {
						name = "attackQty",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"attackQtyLimit",
						},
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
				["direction"] = {
					parameter = {
						name = "direction",
					},
					multi = false,
					type = "number",
					mandatory = false,
				},
			},
		},
		["ENGAGE_TARGETS"] = { -- Enroute task
			id = "EngageTargets",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["targetsType"] = {
					parameter = {
						name = "targetTypes",
						fct = "C_getTargetsType",
					},
					type = "string",
					multi = true,
					mandatory = true,
				},
				["priority"] = {
					parameter = {
						name = "priority",
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = false,
				},
				["maxDistance"] = {
					parameter = {
						name = "maxDist",
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
			},
		},
		["ENGAGE_TARGETS_IN_CIRCLE"] = { -- Enroute task
			id = "EngageTargetsInZone",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["circle"] = {
					parameter = {
						name = "point",
						fct = "C_Circle_getCenterRadius",
						linkedParams = {
							"zoneRadius",
						},
					},
					multi = false,
					type = "C_Circle",
					mandatory = true,
				},
				["targetsType"] = {
					parameter = {
						name = "targetTypes",
						fct = "C_getTargetsType",
					},
					type = "string",
					multi = true,
					mandatory = true,
				},
				["priority"] = {
					parameter = {
						name = "priority",
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = false,
				},
			},
		},
		["ENGAGE_UNIT"] = { -- Enroute task
			id = "EngageUnit",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
				},
			},
			args = {
				["unit"] = {
					parameter = {
						name = "unitId",
						fct = "C_Unit_getID",
					},
					multi = false,
					type = "C_Unit",
					mandatory = true,
				},
				["priority"] = {
					parameter = {
						name = "priority",
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = false,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["expend"] = {
					parameter = {
						name = "expend",
						fct = "C_getExpend",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["qty"] = {
					parameter = {
						name = "attackQty",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"attackQtyLimit",
						},
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = false,
				},
				["direction"] = {
					parameter = {
						name = "direction",
						fct = "C_getNumberWithFlag",
						linkedParams = {
							"directionEnabled",
						},
					},
					multi = false,
					type = "number",
					mandatory = false,
				},
				["groupAttack"] = {
					parameter = {
						name = "groupAttack",
					},
					multi = false,
					type = "boolean",
					mandatory = false,
				},
			},
		},
		["FAC_ENGAGE_GROUP"] = { -- Enroute task
			id = "FAC_EngageGroup",
			filters = {
				type = {
					["plane"] = true,
					["helicopter"] = true,
					["vehicle"] = true,
				},
			},
			args = {
				["frequency"] = {
					parameter = {
						name = "frequency",
					},
					multi = false,
					type = "number",
					min = 1,
					mandatory = true,
				},
				["modulation"] = {
					parameter = {
						name = "modulation",
						fct = "C_getModulation",
					},
					multi = false,
					type = "string",
					mandatory = true,
				},
				["callsign"] = {
					parameter = {
						name = "callname",
						fct = "C_getCallname",
						linkedParams = {
							"number",
						},
					},
					multi = false,
					type = "string",
					mandatory = true,
				},
				["group"] = {
					parameter = {
						name = "groupId",
						fct = "C_Group_getID",
					},
					multi = false,
					type = "C_Group",
					mandatory = true,
				},
				["priority"] = {
					parameter = {
						name = "priority",
					},
					multi = false,
					type = "number",
					min = 0,
					mandatory = false,
				},
				["weaponType"] = {
					parameter = {
						name = "weaponType",
						fct = "C_getWeaponFlag",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["designation"] = {
					parameter = {
						name = "designation",
						fct = "C_getDesignation",
					},
					multi = false,
					type = "string",
					mandatory = false,
				},
				["datalink"] = {
					parameter = {
						name = "datalink",
					},
					multi = false,
					type = "boolean",
					mandatory = false,
				},
			},
		},
	},
	badOptionAIFormationValues = {
		[9] = true,
		[10] = true,
		[655360] = true,
		[655616] = true,
		[720897] = true,
		[720898] = true,
		[851969] = true,
		[851970] = true,
		[917506] = true,
	},
	optionAI = {
		["plane"] = {
			["ROE"] = {
				id = AI.Option.Air.id.ROE,
				values = {
					args = {
						{
							type = "string",
							enum = {
								["WEAPON_FREE"] = AI.Option.Air.val.ROE.WEAPON_FREE,
								["OPEN_FIRE_WEAPON_FREE"] = AI.Option.Air.val.ROE.OPEN_FIRE_WEAPON_FREE,
								["OPEN_FIRE"] = AI.Option.Air.val.ROE.OPEN_FIRE,
								["RETURN_FIRE"] = AI.Option.Air.val.ROE.RETURN_FIRE,
								["WEAPON_HOLD"] = AI.Option.Air.val.ROE.WEAPON_HOLD,
							},
						},
					},
				},
			},
			["FORMATION"] = {
				id = AI.Option.Air.id.FORMATION,
				values = {
					validFct = "C_testFormation",
					args = {
						{
							type = "string",
							enum = {
								["LINE_ABREAST"] = 65536, 
								["TRAIL"] = 131072, 
								["WEDGE"] = 196608, 
								["ECHELON_RIGHT"] = 262144, 
								["ECHELON_LEFT"] = 327680, 
								["FINGER_FOUR"] = 393216, 
								["SPREAD"] = 458752, 
								["BOMBER_ELEMENT"] = 786432,
								["BOMBER_ELEMENT_HEIGHT"] = 851967,
								["FIGHTER_VIC"] = 917504,
							},
						},
						{
							type = "string",
							enum = {
								["CLOSE"] = 1, 
								["OPEN"] = 2,
								["GROUP_CLOSE"] = 3,
							},
						},
					},
				},
			},
			["REACTION_ON_THREAT"] = {
				id = AI.Option.Air.id.REACTION_ON_THREAT,
				values = {
					args = {
						{
							type = "string",
							enum = {
								["NO_REACTION"] = AI.Option.Air.val.REACTION_ON_THREAT.NO_REACTION,
								["PASSIVE_DEFENCE"] = AI.Option.Air.val.REACTION_ON_THREAT.PASSIVE_DEFENCE,
								["EVADE_FIRE"] = AI.Option.Air.val.REACTION_ON_THREAT.EVADE_FIRE,
								["BYPASS_AND_ESCAPE"] = AI.Option.Air.val.REACTION_ON_THREAT.BYPASS_AND_ESCAPE,
								["ALLOW_ABORT_MISSION"] = AI.Option.Air.val.REACTION_ON_THREAT.ALLOW_ABORT_MISSION,
							},
						},
					},
				},
			},
			["RADAR_USING"] = {
				id = AI.Option.Air.id.RADAR_USING,
				values = {
					args = {
						{
							type = "string",
							enum = {
								["NEVER"] = AI.Option.Air.val.RADAR_USING.NEVER,
								["FOR_ATTACK_ONLY"] = AI.Option.Air.val.RADAR_USING.FOR_ATTACK_ONLY,
								["FOR_SEARCH_IF_REQUIRED"] = AI.Option.Air.val.RADAR_USING.FOR_SEARCH_IF_REQUIRED,
								["FOR_CONTINUOUS_SEARCH"] = AI.Option.Air.val.RADAR_USING.FOR_CONTINUOUS_SEARCH,
							},
						},
					},
				},
			},
			["FLARE_USING"] = {
				id = AI.Option.Air.id.FLARE_USING,
				values = {
					args = {
						{
							type = "string",
							enum = {
								["NEVER"] = AI.Option.Air.val.FLARE_USING.NEVER,
								["AGAINST_FIRED_MISSILE"] = AI.Option.Air.val.FLARE_USING.AGAINST_FIRED_MISSILE,
								["FLYING_IN_SAM_WEZ"] = AI.Option.Air.val.FLARE_USING.WHEN_FLYING_IN_SAM_WEZ,
								["FLYING_NEAR_ENEMIES"] = AI.Option.Air.val.FLARE_USING.WHEN_FLYING_NEAR_ENEMIES,
							},
						},
					},
				},
			},
			["ECM_USING"] = {
				id = AI.Option.Air.id.ECM_USING,
				values = {
					args = {
						{
							type = "string",
							enum = {
								["NEVER"] = AI.Option.Air.val.ECM_USING.NEVER_USE,
								["ONLY_LOCKED_BY_RADAR"] = AI.Option.Air.val.ECM_USING.USE_IF_ONLY_LOCK_BY_RADAR,
								["DETECTED_LOCK_BY_RADAR"] = AI.Option.Air.val.ECM_USING.USE_IF_DETECTED_LOCK_BY_RADAR,
								["ALWAYS"] = AI.Option.Air.val.ECM_USING.ALWAYS_USE,
							},
						},
					},
				},
			},
			["MISSILE_ATTACK"] = {
				id = AI.Option.Air.id.MISSILE_ATTACK,
				values = {
					args = {
						{
							type = "string",
							enum = {
								["MAX_RANGE"] = AI.Option.Air.val.MISSILE_ATTACK.MAX_RANGE,
								["NEZ_RANGE"] = AI.Option.Air.val.MISSILE_ATTACK.NEZ_RANGE,
								["HALF_WAY_RMAX_NEZ"] = AI.Option.Air.val.MISSILE_ATTACK.HALF_WAY_RMAX_NEZ,
								["TARGET_THREATEST"] = AI.Option.Air.val.MISSILE_ATTACK.TARGET_THREAT_EST,
								["RANDOM_RANGE"] = AI.Option.Air.val.MISSILE_ATTACK.RANDOM_RANGE,
							},
						},
					},
				},
			},
			["SILENCE"] = {
				id = AI.Option.Air.id.SILENCE,
				values = {
					args = {
						{
							type = "boolean",
						},
					},
				},
			},
			["RTB_BINGO"] = {
				id = AI.Option.Air.id.RTB_ON_BINGO,
				values = {
					args = {
						{
							type = "boolean",
						},
					},
				},
			},
			["PROHIBIT_AA"] = {
				id = AI.Option.Air.id.PROHIBIT_AA,
				values = {
					args = {
						{
							type = "boolean",
						},
					},
				},
			},
			["PROHIBIT_AG"] = {
				id = AI.Option.Air.id.PROHIBIT_AG,
				values = {
					args = {
						{
							type = "boolean",
						},
					},
				},
			},
			["PROHIBIT_AFTER_BURNER"] = {
				id = AI.Option.Air.id.PROHIBIT_AB,
				values = {
					args = {
						{
							type = "boolean",
						},
					},
				},
			},
			["PROHIBIT_JETTISON"] = {
				id = AI.Option.Air.id.PROHIBIT_JETT,
				values = {
					args = {
						{
							type = "boolean",
						},
					},
				},
			},
			["PROHIBIT_WP_REPORT"] = {
				id = AI.Option.Air.id.PROHIBIT_WP_PASS_REPORT,
				values = {
					args = {
						{
							type = "boolean",
						},
					},
				},
			},
			["RTB_OUT_OF_AMMO"] = {
				id = AI.Option.Air.id.RTB_ON_OUT_OF_AMMO,
				values = {
					args = {
						{
							fct = "C_getWeaponFlag",
							type = "string",
						},
					},
				},
			},
		},
		["helicopter"] = {
			["ROE"] = {
				id = AI.Option.Air.id.ROE,
				values = {
					args = {
						{
							type = "string",
							enum = {
								["WEAPON_FREE"] = AI.Option.Air.val.ROE.WEAPON_FREE,
								["OPEN_FIRE_WEAPON_FREE"] = AI.Option.Air.val.ROE.OPEN_FIRE_WEAPON_FREE,
								["OPEN_FIRE"] = AI.Option.Air.val.ROE.OPEN_FIRE,
								["RETURN_FIRE"] = AI.Option.Air.val.ROE.RETURN_FIRE,
								["WEAPON_HOLD"] = AI.Option.Air.val.ROE.WEAPON_HOLD,
							},
						},
					},
				},
			},
			["FORMATION"] = {
				id = AI.Option.Air.id.FORMATION,
				values = {
					validFct = "C_testFormation",
					args = {
						{
							type = "string",
							enum = {
								["COLUMN"] = 720895, 
								["WEDGE"] = 7, 
								["FRONT_RIGHT"] = 655359,
								["FRONT_LEFT"] = 655615,
								["ECHELON_RIGHT"] = 589824, 
								["ECHELON_LEFT"] = 590080, 
							},
						},
						{
							type = "string",
							enum = {
								["70"] = 1, 
								["300"] = 2, 
								["600"] = 3, 
							},
						},
					}
				},
			},
			["REACTION_ON_THREAT"] = {
				id = AI.Option.Air.id.REACTION_ON_THREAT,
				values = {
					args = {
						{
							type = "string",
							enum = {
								["NO_REACTION"] = AI.Option.Air.val.REACTION_ON_THREAT.NO_REACTION,
								["PASSIVE_DEFENCE"] = AI.Option.Air.val.REACTION_ON_THREAT.PASSIVE_DEFENCE,
								["EVADE_FIRE"] = AI.Option.Air.val.REACTION_ON_THREAT.EVADE_FIRE,
								["BYPASS_AND_ESCAPE"] = AI.Option.Air.val.REACTION_ON_THREAT.BYPASS_AND_ESCAPE,
								["ALLOW_ABORT_MISSION"] = AI.Option.Air.val.REACTION_ON_THREAT.ALLOW_ABORT_MISSION,
							},
						},
					},
				},
			},
			["RADAR_USING"] = {
				id = AI.Option.Air.id.RADAR_USING,
				values = {
					args= {
						{
							type = "string",
							enum = {
								["NEVER"] = AI.Option.Air.val.RADAR_USING.NEVER,
								["FOR_ATTACK_ONLY"] = AI.Option.Air.val.RADAR_USING.FOR_ATTACK_ONLY,
								["FOR_SEARCH_IF_REQUIRED"] = AI.Option.Air.val.RADAR_USING.FOR_SEARCH_IF_REQUIRED,
								["FOR_CONTINUOUS_SEARCH"] = AI.Option.Air.val.RADAR_USING.FOR_CONTINUOUS_SEARCH,
							},
						},
					},
				},
			},
			["FLARE_USING"] = {
				id = AI.Option.Air.id.FLARE_USING,
				values = {
					args = {
						{
							type = "string",
							enum = {
								["NEVER"] = AI.Option.Air.val.FLARE_USING.NEVER,
								["AGAINST_FIRED_MISSILE"] = AI.Option.Air.val.FLARE_USING.AGAINST_FIRED_MISSILE,
								["FLYING_IN_SAM_WEZ"] = AI.Option.Air.val.FLARE_USING.WHEN_FLYING_IN_SAM_WEZ,
								["FLYING_NEAR_ENEMIES"] = AI.Option.Air.val.FLARE_USING.WHEN_FLYING_NEAR_ENEMIES,
							},
						},
					},
				},
			},
			["ECM_USING"] = {
				id = AI.Option.Air.id.ECM_USING,
				values = {
					args = {
						{
							type = "string",
							enum = {
								["NEVER"] = AI.Option.Air.val.ECM_USING.NEVER_USE,
								["ONLY_LOCKED_BY_RADAR"] = AI.Option.Air.val.ECM_USING.USE_IF_ONLY_LOCK_BY_RADAR,
								["DETECTED_LOCK_BY_RADAR"] = AI.Option.Air.val.ECM_USING.USE_IF_DETECTED_LOCK_BY_RADAR,
								["ALWAYS"] = AI.Option.Air.val.ECM_USING.ALWAYS_USE,
							},
						},
					},
				},
			},
			["MISSILE_ATTACK"] = {
				id = AI.Option.Air.id.MISSILE_ATTACK,
				values = {
					args = {
						{
							type = "string",
							enum = {
								["MAX_RANGE"] = AI.Option.Air.val.MISSILE_ATTACK.MAX_RANGE,
								["NEZ_RANGE"] = AI.Option.Air.val.MISSILE_ATTACK.NEZ_RANGE,
								["HALF_WAY_RMAX_NEZ"] = AI.Option.Air.val.MISSILE_ATTACK.HALF_WAY_RMAX_NEZ,
								["TARGET_THREATEST"] = AI.Option.Air.val.MISSILE_ATTACK.TARGET_THREAT_EST,
								["RANDOM_RANGE"] = AI.Option.Air.val.MISSILE_ATTACK.RANDOM_RANGE,
							},
						},
					},
				},
			},
			["SILENCE"] = {
				id = AI.Option.Air.id.SILENCE,
				values = {
					args = {
						{
							type = "boolean",
						},
					},
				},
			},
			["RTB_BINGO"] = {
				id = AI.Option.Air.id.RTB_ON_BINGO,
				values = {
					args = {
						{
							type = "boolean",
						},
					},
				},
			},
			["PROHIBIT_AA"] = {
				id = AI.Option.Air.id.PROHIBIT_AA,
				values = {
					args = {
						{
							type = "boolean",
						},
					},
				},
			},
			["PROHIBIT_AG"] = {
				id = AI.Option.Air.id.PROHIBIT_AG,
				values = {
					args = {
						{
							type = "boolean",
						},
					},
				},
			},
			["PROHIBIT_JETTISON"] = {
				id = AI.Option.Air.id.PROHIBIT_JETT,
				values = {
					args = {
						{
							type = "boolean",
						},
					},
				},
			},
			["PROHIBIT_WP_REPORT"] = {
				id = AI.Option.Air.id.PROHIBIT_WP_PASS_REPORT,
				values = {
					args = {
						{
							type = "boolean",
						},
					},
				},
			},
			["RTB_OUT_OF_AMMO"] = {
				id = AI.Option.Air.id.RTB_ON_OUT_OF_AMMO,
				values = {
					args = {
						{
							fct = "C_getWeaponFlag",
							type = "string",
						},
					}
				},
			},
		},
		["vehicle"] = {
			["ROE"] = {
				id = AI.Option.Ground.id.ROE,
				values = {
					args = {
						{
							type = "string",
							enum = {
								["OPEN_FIRE"] = AI.Option.Ground.val.ROE.OPEN_FIRE,
								["RETURN_FIRE"] = AI.Option.Ground.val.ROE.RETURN_FIRE,
								["WEAPON_HOLD"] = AI.Option.Ground.val.ROE.WEAPON_HOLD,
							},
						},
					},
				},
			},
			["ALARM_STATE"] = {
				id = AI.Option.Ground.id.ALARM_STATE,
				values = {
					args = {
						{
							type = "string",
							enum = {
								["AUTO"] = AI.Option.Ground.val.ALARM_STATE.AUTO,
								["RED"] = AI.Option.Ground.val.ALARM_STATE.RED,
								["GREEN"] = AI.Option.Ground.val.ALARM_STATE.GREEN,
							},
						},
					},
				},
			},
			["DISPERSE_ON_ATTACK"] = {
				id = AI.Option.Ground.id.DISPERSE_ON_ATTACK,
				values = {
					args = {
						{
							type = "number",
							min = 0,
							max = 120,
						},
					},
				},
			},
			["SILENCE"] = {
				id = AI.Option.Ground.id.SILENCE,
				values = {
					args = {
						{
							type = "boolean",
						},
					},
				},
			},
		},
		["ship"] = {
			["ROE"] = {
				id = AI.Option.Naval.id.ROE,
				values = {
					args = {
						{
							type = "string",
							enum = {
								["OPEN_FIRE"] = AI.Option.Naval.val.ROE.OPEN_FIRE,
								["RETURN_FIRE"] = AI.Option.Naval.val.ROE.RETURN_FIRE,
								["WEAPON_HOLD"] = AI.Option.Naval.val.ROE.WEAPON_HOLD,
							},
						},
					},
				},
			},
			["SILENCE"] = {
				id = AI.Option.Naval.id.SILENCE,
				values = {
					args = {
						{
							type = "boolean",
						},
					},
				},
			},
		},
	},
	typesOfSensor = {
		["VISUAL"] = Controller.Detection.VISUAL,
		["OPTIC"] = Controller.Detection.OPTIC,
		["RADAR"] = Controller.Detection.RADAR,
		["IRST"] = Controller.Detection.IRST,
		["RWR"] = Controller.Detection.RWR,
		["DLINK"] = Controller.Detection.DLINK,
		["ALL"] = true,
	},
	cardinal = {
		["N"] = "NORTH",
		["S"] = "SOUTH",
		["E"] = "EAST",
		["W"] = "WEST",
	},
	weaponAttributes = {
		--categories = {
			"SHELL",
			"MISSILE",
			"ROCKET",
			"BOMB",
		--guidanceType = {
			"INS",
			"IR",
			"RADAR_ACTIVE",
			"RADAR_SEMI_ACTIVE",
			"RADAR_PASSIVE",
			"TV",
			"LASER",
			"TELE",
		-- missileCategories = {
			"AAM",
			"SAM",
			"BM",
			"ANTI_SHIP",
			"CRUISE",
			"OTHER",    
		--warheadTypes = {
			"AP",
			"HE",
			"SHAPED_CHARGE",
	},
	stackTypeExtract = {
		["HEAD"] = {
			load = "first",
			direction = "next",
			raz = "previous",
		},
		["TAIL"] = {
			load = "last",
			direction = "previous",
			raz = "next",
		},
	},
	polygonMaxVertices = 40,
	debugSwitch = true,					-- Activation ou non du mode affichage debug pour le module Core
	maxTimeSmoke = 300,
	stackTimerMaxTotalCharge = 120,		-- Valeurs de tuning pour optimiser la gestion du timer sur les traitements sécables (stacktimer)
	modulationValues = {
		["AM"] = radio.modulation.AM,
		["FM"] = radio.modulation.FM, 
	},
	indZoneConfig ={
		["C_Circle"] = {
			fctIsInside = "C_Circle_isInside",
			fctGetRandomPosition = "C_Circle_getRandomPosition",
			fctEvalSearchVolume = "C_Circle_searchVolume",
			fctDuplicate = "C_Circle_duplicate",
			charge = 1,
		},
		["C_Rectangle"] = {
			fctIsInside = "C_Rectangle_isInside",
			fctGetRandomPosition = "C_Rectangle_getRandomPosition",
			fctEvalSearchVolume = "C_Rectangle_searchVolume",
			fctDuplicate = "C_Rectangle_duplicate",
			charge = 2,
		},
		["C_Sphere"] = {
			fctIsInside = "C_Sphere_isInside",
			fctGetRandomPosition = "C_Sphere_getRandomPosition",
			fctEvalSearchVolume = "C_Sphere_searchVolume",
			fctDuplicate = "C_Sphere_duplicate",
			charge = 1,
		},
		["C_Cylinder"] = {
			fctIsInside = "C_Cylinder_isInside",
			fctGetRandomPosition = "C_Cylinder_getRandomPosition",
			fctEvalSearchVolume = "C_Cylinder_searchVolume",
			fctDuplicate = "C_Cylinder_duplicate",
			charge = 1,
		},
		["C_Box"] = {
			fctIsInside = "C_Box_isInside",
			fctGetRandomPosition = "C_Box_getRandomPosition",
			fctEvalSearchVolume = "C_Box_searchVolume",
			fctDuplicate = "C_Box_duplicate",
			charge = 2,
		},
		["C_PickupZone"] = {
			fctIsInside = "C_Circle_isInside",
			fctGetRandomPosition = "C_Circle_getRandomPosition",
			fctEvalSearchVolume = "C_Circle_searchVolume",
			fctDuplicate = "C_PickupZone_duplicate",
			charge = 1,
		},
		["C_Polygon"] = {
			fctIsInside = "C_Polygon_isInside",
			fctGetRandomPosition = "C_Polygon_getRandomPosition",
			fctEvalSearchVolume = "C_Polygon_searchVolume",
			fctDuplicate = "C_Polygon_duplicate",
			charge = 1,
		},
	},
	randomSign = {
		[1] = -1,
		[2] = 1,
	},
	uncontrolledRouteDelay = 20,
	basicRouteDelay = 2,
	tokenTypes = {
		["RANDOM"] = {
			type = "TOKEN_TYPE_RANDOM",
		},
		["TRUE"] = {
			type = "TOKEN_INTEGER",
			value = 1,
		},
		["FALSE"] = {
			type = "TOKEN_INTEGER",
			value = 0,
		},
		["FLARE"] = {
			type = "TOKEN_SIGNAL_TYPE",
		},
		["SMOKE"] = {
			type = "TOKEN_SIGNAL_TYPE",
		},
		["WAYPOINT"] = {
			type = "TOKEN_TYPE_WAYPOINT",
		},
		["TIME"] = {
			type = "TOKEN_TYPE_TIME",
		},
		["MAX_TURNS"] = {
			type = "TOKEN_TYPE_MAX_TURNS",
		},
		["FLAG"] = {
			type = "TOKEN_TYPE_FLAG",
		},
		["FUEL_RATIO"] = {
			type = "TOKEN_TYPE_FUEL_RATIO",
		},
		["TACAN_AA"] = {
			type = "TOKEN_TYPE_TACAN_AA",
		},
		["TACAN"] = {
			type = "TOKEN_TYPE_TACAN",
		},
		["ICLS"] = {
			type = "TOKEN_TYPE_ICLS",
		}
	},
	relativeStatus = {
		"ABSOLUTE",
		"RELATIVE",
	},
	findClasses = { 
		["AIRBASES"] = { -- AIRBASES inopérant cause bug DCS pour recherche avec world.searchObject. Traitement spécifique
		},
		["GROUPS"] = {
			findSubCategories = true,
			category = {Object.Category.UNIT},
			search_fct = function(_object, _datas)
				ATMEFct.search_UNIT(_object, _datas)
			end,
			search_all_fct = function(_datas)
				for _, _unit in pairs(ATMEDatas.allAIUnitsAndPlayers) do
					ATMEFct.C_FindContext_search(_unit:getDCSUnit(), _datas)
				end
			end,
			fct = function(_datas, _group, _unitName)
				local groupName = _group:getName()
				if _datas.objects[groupName] == nil then -- Cause recherche multi-zones sur un ATME.C_ZonesMgr
					if _datas.nbUnitsGroup[groupName] == nil then
						_datas.allGroupIn[groupName] = {}
						_datas.nbUnitsGroup[groupName] = 0
					end

					table.insert(_datas.allGroupIn[groupName], _unitName)
					_datas.nbUnitsGroup[groupName] = _datas.nbUnitsGroup[groupName] + 1

					if _datas.nbUnitsGroup[groupName] == _group:getNbUnits() then
						local record
						local status
						if _datas.callbacks[ATMEDatas.currentModuleName] ~= nil then
							local datas = {
								zone = _datas.zone,
								object = _group,
								unitsName = _datas.allGroupIn[groupName],
							}
							status, record = xpcall(function() return _datas.callbacks[ATMEDatas.currentModuleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Find callback') end)
						end
						
						_datas.allGroupIn[groupName] = nil
						_datas.nbUnitsGroup[groupName] = nil
						if record == nil or (type(record) == "boolean" and record == true) then
							_datas.objects[groupName] = _group
							table.insert(_datas.objectsIndex, groupName)
						end
					end
				end
			end,
		},
		["UNITS"] = {
			findSubCategories = true,
			category = {Object.Category.UNIT},
			search_fct = function(_object, _datas)
				ATMEFct.search_UNIT(_object, _datas)
			end,
			search_all_fct = function(_datas)
				for _, _unit in pairs(ATMEDatas.allAIUnitsAndPlayers) do
					ATMEFct.C_FindContext_search(_unit:getDCSUnit(), _datas)
				end
			end,
			fct = function(_datas, _group, _unitName)
				if _datas.objects[_unitName] == nil then -- Cause recherche multi-zones sur un ATME.C_ZonesMgr
					if ATMEDatas.allAIUnitsAndPlayers[_unitName] ~= nil then
						local record
						local status
						if _datas.callbacks[ATMEDatas.currentModuleName] ~= nil then
							local datas = {
								zone = _datas.zone,
								object = ATMEDatas.allAIUnitsAndPlayers[_unitName],
							}
							status, record = xpcall(function() return _datas.callbacks[ATMEDatas.currentModuleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Find callback') end)
						end

						if record == nil or (type(record) == "boolean" and record == true) then

							_datas.objects[_unitName] = ATMEDatas.allAIUnitsAndPlayers[_unitName]
							table.insert(_datas.objectsIndex, _unitName)
						end
					end
				end
			end,
		},
		["AI_UNITS"] = {
			findSubCategories = true,
			category = {Object.Category.UNIT},
			search_fct = function(_object, _datas)
				ATMEFct.search_AIUNIT(_object, _datas)
			end,
			search_all_fct = function(_datas)
				for _, _unit in pairs(ATMEDatas.AIUnits) do
					ATMEFct.C_FindContext_search(_unit:getDCSUnit(), _datas)
				end
			end,
			fct = function(_datas, _group, _unitName)
				if _datas.objects[_unitName] == nil then -- Cause recherche multi-zones sur un ATME.C_ZonesMgr
					if ATMEDatas.AIUnits[_unitName] ~= nil then
						local record
						local status
						if _datas.callbacks[ATMEDatas.currentModuleName] ~= nil then
							local datas = {
								zone = _datas.zone,
								object = ATMEDatas.AIUnits[_unitName],
							}
							status, record = xpcall(function() return _datas.callbacks[ATMEDatas.currentModuleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Find callback') end)
						end
						
						if record == nil or (type(record) == "boolean" and record == true) then
							_datas.objects[_unitName] = ATMEDatas.AIUnits[_unitName]
							table.insert(_datas.objectsIndex, _unitName)
						end
					end
				end
			end,
		},
		["PLAYERS"] = {
			findSubCategories = true,
			category = {Object.Category.UNIT},
			search_fct = function(_object, _datas)
				ATMEFct.search_PLAYER(_object, _datas)
			end,
			search_all_fct = function(_datas)
				for _, _unit in pairs(ATMEDatas.players) do
					ATMEFct.C_FindContext_search(_unit:getDCSUnit(), _datas)
				end
			end,
			fct = function(_datas, _group, _unitName)
				if _datas.objects[_unitName] == nil then -- Cause recherche multi-zones sur un ATME.C_ZonesMgr
					
					if ATMEDatas.players[_unitName] ~= nil then
						local record
						local status
						if _datas.callbacks[ATMEDatas.currentModuleName] ~= nil then
							local datas = {
								zone = _datas.zone,
								object = ATMEDatas.players[_unitName],
							}
							status, record = xpcall(function() return _datas.callbacks[ATMEDatas.currentModuleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Find callback') end)
						end
						
						if record == nil or (type(record) == "boolean" and record == true) then
							_datas.objects[_unitName] = ATMEDatas.players[_unitName]
							table.insert(_datas.objectsIndex, _unitName)
						end
					end
				end
			end,
		},
		["WEAPONS"] = {
			findSubCategories = true,
			category = {
				Object.Category.WEAPON,
			},
			search_fct = function(_object, _datas)
				ATMEFct.search_WEAPON(_object, _datas)
			end,
			search_all_fct = function(_datas)
				for _, _object in pairs(ATMEDatas.weapons) do
					ATMEFct.C_FindContext_search(_object:getDCSWeapon(), _datas)
				end
			end,
			fct = function(_datas, _group, _objectName)
				if _datas.objects[_objectName] == nil then -- Cause recherche multi-zones sur un ATME.C_ZonesMgr
					local record
					local status
					if _datas.callbacks[ATMEDatas.currentModuleName] ~= nil then
						local datas = {
							zone = _datas.zone,
							object = ATMEDatas.weapons[_objectName],
						}
						status, record = xpcall(function() return _datas.callbacks[ATMEDatas.currentModuleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Find callback') end)
					end
				
					if record == nil or (type(record) == "boolean" and record == true) then
						_datas.objects[_objectName] = ATMEDatas.weapons[_objectName]
						
						table.insert(_datas.objectsIndex, _objectName)
					end
				end
			end,
		},
		["STATIC_OBJECTS"] = {
			findSubCategories = true,
			category = {
				Object.Category.STATIC,
			},
			search_fct = function(_object, _datas)
				ATMEFct.search_STATIC(_object, _datas)
			end,
			search_all_fct = function(_datas)
				for _, _object in pairs(ATMEDatas.staticObjects) do
					ATMEFct.C_FindContext_search(_object:getDCSStaticObject(), _datas)
				end
			end,
			fct = function(_datas, _group, _objectName)
				if _datas.objects[_objectName] == nil then -- Cause recherche multi-zones sur un ATME.C_ZonesMgr
					if ATMEDatas.staticObjects[_objectName] ~= nil then
						local record
						local status
						if _datas.callbacks[ATMEDatas.currentModuleName] ~= nil then
							local datas = {
								zone = _datas.zone,
								object = ATMEDatas.staticObjects[_objectName],
							}
							status, record = xpcall(function() return _datas.callbacks[ATMEDatas.currentModuleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Find callback') end)
						end
						
						
						if record == nil or (type(record) == "boolean" and record == true) then
							_datas.objects[_objectName] = ATMEDatas.staticObjects[_objectName]
							table.insert(_datas.objectsIndex, _objectName)
						end
					end
				end
			end,
		},
		["CARGOS"] = {
			findSubCategories = false,
			category = {
				Object.Category.CARGO,
			},
			search_fct = function(_object, _datas)
				ATMEFct.search_CARGO(_object, _datas)
			end,
			search_all_fct = function(_datas)
				for _, _object in pairs(ATMEDatas.cargos) do
					ATMEFct.C_FindContext_search(_object:getDCSCargo(), _datas)
				end
			end,
			fct = function(_datas, _group, _objectName)
				if _datas.objects[_objectName] == nil then -- Cause recherche multi-zones sur un ATME.C_ZonesMgr
					if ATMEDatas.cargos[_objectName] ~= nil then
						local record
						local status
						if _datas.callbacks[ATMEDatas.currentModuleName] ~= nil then
							local datas = {
								zone = _datas.zone,
								object = ATMEDatas.cargos[_objectName],
							}
							status, record = xpcall(function() return _datas.callbacks[ATMEDatas.currentModuleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Find callback') end)
						end
						
						if record == nil or (type(record) == "boolean" and record == true) then
							_datas.objects[_objectName] = ATMEDatas.cargos[_objectName]
							table.insert(_datas.objectsIndex, _objectName)
						end
					end
				end
			end,
		},
		["SCENERY_OBJECTS"] = {
			findSubCategories = false,
			category = {
				Object.Category.SCENERY,
			},
			search_fct = function(_object, _datas)
				ATMEFct.search_SCENERY(_object, _datas)
			end,
			search_all_fct = function(_datas)
				for _, _object in pairs(ATMEDatas.sceneryObjects) do
					ATMEFct.C_FindContext_search(_object:getDCSStaticObject(), _datas)
				end
			end,
			fct = function(_datas, _group, _objectName)
				if _datas.objects[_objectName] == nil then -- Cause recherche multi-zones sur un ATME.C_ZonesMgr
					local record
					local status
					if _datas.callbacks[ATMEDatas.currentModuleName] ~= nil then
						local datas = {
							zone = _datas.zone,
							object = ATMEDatas.sceneryObjects[_objectName],
						}
						status, record = xpcall(function() return _datas.callbacks[ATMEDatas.currentModuleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Find callback') end)
					end
					
					if record == nil or (type(record) == "boolean" and record == true) then
						_datas.objects[_objectName] = ATMEDatas.sceneryObjects[_objectName]
						
						table.insert(_datas.objectsIndex, _objectName)
					end
				end
			end,
		},
	},
	referenceClasses = {
			["C_AIUnit"] = true,
			["C_Player"] = true,
			["C_StaticObject"] = true,
			["C_Cargo"] = true,
			["C_Airbase"] = true,
			["C_SceneryObject"] = true,
			["C_Weapon"] = true,
	},
	tacanMode = {
		["X"] = 0,
		["Y"] = 1,
	},
	tacanFrequencyOffset = {
		["Xmit"] = {
			["X"] = 1024,
			["Y"] = 1024,
		},
		["Recv"] = {
			["X"] = 961,
			["Y"] = 1087,
		},
	},
	beaconTypes = {
		["TOKEN_TYPE_TACAN_AA"] = "TACAN",
		["TOKEN_TYPE_TACAN"] = "TACAN",
		["TOKEN_TYPE_ICLS"] = "ICLS",
	},
	raceTokenTypes = {
		["RACE_ON"] = true,
		["RACE_OFF"] = false,
	},
	coreEventTypes = {
		["TAKE_OFF"] = function(_localMembers)
			return {}
		end,
		["LAND"] = function(_localMembers)
			return {}
		end,
		["ENGINE_START"] = function(_localMembers)
			return {}
		end,
		["ENGINE_SHUTDOWN"] = function(_localMembers)
			return {}
		end,
		["RACE_PLAYER_TOO_HIGH"] = function(_localMembers)
			return {
				race = _localMembers.datas.race,
				penalty = _localMembers.datas.penalty,
			}
		end,
		["RACE_START"] = function(_localMembers)
			return {
				race = _localMembers.datas.race,
			}
		end,
		["RACE_TURN_TIME"] = function(_localMembers)
			return {
				race = _localMembers.datas.race,
				playTime = _localMembers.datas.playTime,
				totalPlayTime = _localMembers.datas.totalPlayTime,
				penalty = _localMembers.datas.penalty,
				turnNumber = _localMembers.datas.turnNumber,
			}	
		end,
		["RACE_PLAYER_BEST_TURN_TIME"] = function(_localMembers)
			return {
				race = _localMembers.datas.race,
				playTime = _localMembers.datas.playTime,
				turnNumber = _localMembers.datas.turnNumber,
			}
		end,
		["RACE_BEST_TURN_TIME"] = function(_localMembers)
			return {
				race = _localMembers.datas.race,
				playTime = _localMembers.datas.playTime,
				turnNumber = _localMembers.datas.turnNumber,
			}
		end,
		["RACE_LAP_TIME"] = function(_localMembers)
			return {
				race = _localMembers.datas.race,
				playTime = _localMembers.datas.playTime,
				totalPlayTime = _localMembers.datas.totalPlayTime,
				penalty = _localMembers.datas.penalty,
				turnNumber = _localMembers.datas.turnNumber,
				doorNumber = _localMembers.datas.doorNumber,
			}
		end,
		["RACE_BEST_LAP_TIME"] = function(_localMembers)
			return {
				race = _localMembers.datas.race,
				playTime = _localMembers.datas.playTime,
				turnNumber = _localMembers.datas.turnNumber,
				doorNumber = _localMembers.datas.doorNumber,
			}
		end,
		["RACE_PLAYER_BEST_FINAL_TIME"] = function(_localMembers)
			return {
				race = _localMembers.datas.race,
				totalPlayTime = _localMembers.datas.totalPlayTime,
			}
		end,
		["RACE_BEST_FINAL_TIME"] = function(_localMembers)
			return {
				race = _localMembers.datas.race,
				totalPlayTime = _localMembers.datas.totalPlayTime,
			}
		end,
		["RACE_FINAL_TIME"] = function(_localMembers)
			return {
				race = _localMembers.datas.race,
				totalPlayTime = _localMembers.datas.totalPlayTime,
				penalty = _localMembers.datas.penalty,
			}
		end,
		["TRANSPORT_END_OF_BOARDING"] = function(_localMembers)
			return {
				index = _localMembers.datas.index,
				onBoardGroupName = _localMembers.datas.onBoardGroupName,
				nbUnitsOnBoard = _localMembers.datas.nbUnitsOnBoard,
				nbDeadUnits = _localMembers.datas.nbDeadUnits,
			}
		end,
		["TRANSPORT_END_OF_DISEMBARK"] = function(_localMembers)
			return {
				index = _localMembers.datas.index,
				linkedGroup = _localMembers.datas.linkedGroup,
				nbUnitsOnBoard = _localMembers.datas.nbUnitsOnBoard,
				nbDeadUnits = _localMembers.datas.nbDeadUnits,
			}
		end,
		["SPOT_TARGET_OUT_OF_ZONE"] = function(_localMembers)
			local datas = {}
			datas.target = _localMembers.datas.target
				
			return datas
		end,
		["SPOT_TARGET_NO_LOS"] = function(_localMembers)
			return {
				target = _localMembers.datas.target,
			}
		end,
		["SPOT_NEW_TARGET"] = function(_localMembers)
			return {
				target = _localMembers.datas.target,
			}
		end,
		["SPOT_STOP_UNIT_MOVE"] = function(_localMembers)
			return {}
		end,
		["SPOT_STOP"] = function(_localMembers)
			return {}
		end,
		["TARGET_NEW_TRACK"] = function(_localMembers)
			return {
				target = _localMembers.datas.target,
			}
		end,
		["TARGET_END_TRACK"] = function(_localMembers)
			return {
				target = _localMembers.datas.target,
			}
		end,
		["SHOOTING_START"] = function(_localMembers)
			return {
				target = _localMembers.datas.target,
			}
		end,
		["WEAPON_SHOT"] = function(_localMembers)
			return {
				target = _localMembers.datas.target,
				weapon = _localMembers.datas.weapon,
			}
		end,
		["HIT_TARGET"] = function(_localMembers)
			return {
				target = _localMembers.datas.target,
				weapon = _localMembers.datas.weapon,
			}
		end,		
		["SHOOTING_END"] = function(_localMembers)
			return {}
		end,
		["REFUELING"] = function(_localMembers)
			return {}
		end,
		["REFUELING_STOP"] = function(_localMembers)
			return {}
		end,
		["MARK_ADDED"] = function(_localMembers)
			return {
				id = _localMembers.datas.id,
				group = _localMembers.datas.group,
				coalitionName = _localMembers.datas.coalitionName,
				pos = _localMembers.datas.pos,
				text = _localMembers.datas.text,
			}
		end,
		["MARK_REMOVED"] = function(_localMembers)
			return {
				id = _localMembers.datas.id,
				group = _localMembers.datas.group,
				coalitionName = _localMembers.datas.coalitionName,
				pos = _localMembers.datas.pos,
				text = _localMembers.datas.text,
			}
		end,
	},
}

-- Ajout des données nécessaires aux ATMEConst déjà définies

do
	ATMEConst.errorLabelReference = " or table (x,y)/(x,y,z)"
	local first = true
	for _id, _ in pairs(ATMEConst.referenceClasses) do
		if first == true then
			ATMEConst.errorLabelReference = "ATME." .. _id .. ATMEConst.errorLabelReference
		else
			ATMEConst.errorLabelReference = "ATME." .. _id .. ", " .. ATMEConst.errorLabelReference
		end
		
		first = false
	end
	
	for _id, _ in pairs(ATMEConst.supportedLanguages) do
		ATMEConst.tokenTypes[_id] = {
			type = "TOKEN_LANGUAGE",
		}
	end
	
	for _id, _ in pairs(ATMEConst.raceTokenTypes) do
		ATMEConst.tokenTypes[_id] = {
			type = "TOKEN_RACE",
		}
	end
	
	for _id, _ in pairs(ATMEConst.smokeColors) do
		if _id ~= "RANDOM" then
			ATMEConst.tokenTypes[_id] = {
				type = "TOKEN_COLOR",
			}
		end
	end
	
	for _id, _ in pairs(ATMEConst.flareColors) do
		if _id ~= "RANDOM" then
			ATMEConst.tokenTypes[_id] = {
				type = "TOKEN_COLOR",
			}
		end
	end
	
	for _id, _ in pairs(ATMEConst.blueCallsigns.aircrafts) do
		ATMEConst.tokenTypes[_id] = {
			type = "TOKEN_AIRCRAFT_CALLSIGN_NAME",
		}
	end
	
	for _id, _ in pairs(ATMEConst.blueCallsigns.jtacs) do
		ATMEConst.tokenTypes[_id] = {
			type = "TOKEN_JTAC_CALLSIGN_NAME",
		}
	end
end

-- Définitions complémentaires

local POINT_BAD = 0
local POINT_2D = 1
local POINT_3D = 2

local customEventHandler = {}			-- Déclaration du handler d'événements DCS


----------------------------------------------------------------------------------------------------------
-- DEFINITION DES FONCTIONS UTILES DIRECTEMENT A ATME
----------------------------------------------------------------------------------------------------------


ATMEFct.errorOutputCore = function(_module, _fctDefinition, _text, _needObjectPrecision)
	-- errorOutputCore : fonction de base de gestion des erreurs, engendre un arrêt de ATME (plus de réarmement des timers
	
	if ATMEDatas.exitATME == false then
		ATMEFct.setError("FATAL_ERROR")
	
		local err = debug.traceback()
		local id, text, stack = ATMEFct.analyzeTraceback(err)
	
		local label = "[ATME"
		if ATMEDatas.refModules[id] ~= nil then
			label = label .. "-" .. ATMEDatas.refModules[id]
		end
		
		label = label .. "] Error in "
		
		if ATMEDatas.refModules[id] ~= nil then
			label = label .. "module lua file" 
		else
			label = label .. "script lua (out of a module lua file)"
		end
		
		if text ~= nil then
			label = label .. "\n" .. text .. _fctDefinition .. "\n   -> " .. _text
			
			if _needObjectPrecision == true then
				label = label .. " (or missing ':' before function name)"
			end
		end
		
		label = label .. "\n\nATME STOPPED"
		env.error(label .. "\n\nstack traceback:\n" .. stack, true)
		
		ATMEDatas.exitATME = true
	end
end

ATMEFct.logDisplayText = function(_debugMode, _label, _level)
	if _debugMode == true and _level == 1 and ATMEDatas.logDisplay == true then
		trigger.action.outText(_label, 5)
	elseif _debugMode == false and _level > 2 then
		return
	end
	
	env.info(_label, false)
end

ATMEFct.outputCore = function(_module, _text, _level)
	-- outputCore : fonction de base des messages de debug en fonction du niveau et de l'activation ou non du mode debug
	if ATMEDatas.exitATME == false then
		local label = "[ATME"
		local debugMode = ATMEConst.debugSwitch
		
		if _module ~= nil then
			label = label .. "-" .. _module:getName()
			debugMode = _module:isDebugOn()
		end
		
		label = label .. "] " .. _text
		
		ATMEFct.logDisplayText(debugMode, label, _level)
	end
end

ATMEFct.outputVarCore = function(_module, _title, _var, _level)
	if ATMEDatas.exitATME == false then
		local debugMode = ATMEConst.debugSwitch
		local startLabel = "[ATME"
		
		if _module ~= nil then
			startLabel = startLabel .. "-" .. _module:getName()
			debugMode = _module:isDebugOn()
		end
		
		startLabel = startLabel .. "] "
		
		local function logVar(_v, _title)
			local label = _title
			
			if type(_v) == "boolean" then
				local v = "True"
				if _v == false then
					v = "False"
				end
				label = label .. " : " .. v
			elseif _v ~= nil and type(_v) ~= "function" then
				label = label ..  " : " .. _v
			end
			
			ATMEFct.logDisplayText(debugMode, label, _level)
		end
		
		local function logObject(_v, _title)
			ATMEFct.logDisplayText(debugMode, _title .. _v:getClassName()  .. " class", _level)
		end
		
		local function logTable(_t, _indent)
			if _indent == nil then
				_indent = "  "
			else
				_indent = _indent .. "  "
			end
			
			local found = false
			for _k, _v in pairs(_t) do
				found = true
				
				if type(_v) == "table" then
					if ATMEFct.isATMEObject(_v) == false then
						ATMEFct.logDisplayText(debugMode, _indent .. "table '" .. _k .. "' listing\n", _level)
						if logTable(_v, _indent) == false then
							ATMEFct.logDisplayText(debugMode, _indent .. "  Empty table", _level)
						elseif type(_v.x) == "number" and type(_v.y) == "number" and (_v.z == nil or type(_v.z) == "number") then
							local lat, lon, alt = coord.LOtoLL(ATMEFct.convertToPoint3D(_v))
					
							local dirlon, dlon, mlon, slon = ATMEFct.convertToDMS(lon)
							local dirlat, dlat, mlat, slat = ATMEFct.convertToDMS(lat)
							
							local label 
							if  _v.z == nil then
								label = _indent .. "  Found keys x,y point 2D at lat : "
							else
								label = _indent .. "  Found keys x,y,z point 3D at lat : "
							end
							
							if dirlat < 0 then
								label = label .. "S"
							else
								label = label .. "N"
							end
							label = label .. string.format("%02d°%02d'%02d\"", dlat, mlat, slat) .. " - lon : "
							
							if dirlon < 0 then
								label = label .. "W"
							else
								label = label .. "E"
							end
							
							ATMEFct.logDisplayText(debugMode, label .. string.format("%03d°%02d'%02d\"", dlon, mlon, slon) .. "\n -> alt : " .. alt .. "m (" .. math.floor(ATME.convertMtoFts(alt)) .. " feet)\n", _level)
						end
					else
						logObject(_v, _indent .. "key '" .. _k .. "' - Object instance of ATME.")
					end
				else
					logVar(_v, _indent .. "key '" .. _k .. "' - Type " .. type(_v))
				end
			end
			
			return found
		end
		
		if type(_var) == "table" then
			if ATMEFct.isATMEObject(_var) == false then
				startLabel = startLabel .. "Start Table '" .. _title .. "' :\n"
				ATMEFct.logDisplayText(debugMode, startLabel, _level)
				if logTable(_var) == false then
					ATMEFct.logDisplayText(debugMode, "  Empty table", _level)
				end
			else
				logObject(_var, startLabel .. "Object '" .. _title .. "' - Instance of ATME.")
			end
		else
			logVar(_var, startLabel .. "Var '" .. _title .. "' - Type " .. type(_var))
		end
	end
end

ATMEFct.setError = function(_code)
	ATMEDatas.error = {
		code = _code,
	}
end

ATMEFct.copyTable = function(_datas)
	local err = false 
	local function _copy(datas)
		if err == false then
			if type(datas) ~= "table" or ATMEFct.isATMEObject(datas) == true then
				return datas
			end
			
			if getmetatable(_datas) == nil then
				local newTable = {}
				for _index, _value in pairs(datas) do
					newTable[_index] = _copy(_value)
				end
					
				return newTable
			else
				err = true
			end
		end
	end
	
	local t = _copy(_datas)
	if err == false then
		return t
	end
end

ATMEFct.convertToDMS = function(_degrees)
	local dir = 1
	if _degrees < 0 then
		dir = -1
	end
	
	local d = math.abs(_degrees)
	local dfloor = math.floor(d)
	local m = (d - dfloor) * 60
	local mfloor = math.floor(m)
	local s = (m - mfloor) * 60
	local sfloor = math.floor(s)
	
	return dir, dfloor, mfloor, sfloor, math.floor((s - sfloor) * 100)
end

ATMEFct.convertToDMd = function(_degrees, _highPrecision)
	local dir = 1
	if _degrees < 0 then
		dir = -1
	end
	
	local factor = 10000
	if _highPrecision == false then
		factor = 100
	end

	local dabs = math.abs(_degrees)
	local dfloor = math.floor(dabs)
	local m = (dabs - dfloor) * 60
	local mfloor = math.floor(m)
	
	return dir, dfloor, mfloor, math.floor((m - mfloor) * factor)
end

ATMEFct.search_UNIT = function(_unit, _datas)
	local unitName = _unit:getName()

	if ATMEDatas.allAIUnitsAndPlayers[unitName] ~= nil then
		local group = ATMEDatas.allAIUnitsAndPlayers[unitName]:getGroup()
		local findOk = ATMEFct.testFindCategory(ATMEDatas.allAIUnitsAndPlayers[unitName], _datas.findCategories)
		
		if findOk == true then
			if _datas.coalitionName == nil or ATMEDatas.allAIUnitsAndPlayers[unitName]:getCoalitionName() == _datas.coalitionName then
				if ATMEFct.C_FindContext_Object_isInZone(ATMEDatas.allAIUnitsAndPlayers[unitName], _datas) == true then
					ATMEFct.C_FindContext_setToList(_datas, group, unitName)
				end
			end
		end
	end	
end

ATMEFct.search_AIUNIT = function(_unit, _datas)
	local unitName = _unit:getName()
	
	if ATMEDatas.AIUnits[unitName] ~= nil then
		local group = ATMEDatas.AIUnits[unitName]:getGroup()
		local findOk = ATMEFct.testFindCategory(ATMEDatas.AIUnits[unitName], _datas.findCategories)
		
		if findOk == true then
			if _datas.coalitionName == nil or ATMEDatas.AIUnits[unitName]:getCoalitionName() == _datas.coalitionName then
				if ATMEFct.C_FindContext_Object_isInZone(ATMEDatas.AIUnits[unitName], _datas) == true then
						
					ATMEFct.C_FindContext_setToList(_datas, group, unitName)
				end
			end
		end
	end
end

ATMEFct.search_PLAYER = function(_unit, _datas)
	local unitName = _unit:getName()
	
	if ATMEDatas.players[unitName] ~= nil then
		local group = ATMEDatas.players[unitName]:getGroup()
		local findOk = ATMEFct.testFindCategory(ATMEDatas.players[unitName], _datas.findCategories)
		
		if findOk == true then
			if _datas.coalitionName == nil or ATMEDatas.players[unitName]:getCoalitionName() == _datas.coalitionName then
				if ATMEFct.C_FindContext_Object_isInZone(ATMEDatas.players[unitName], _datas) == true then
						
					ATMEFct.C_FindContext_setToList(_datas, group, unitName)
				end
			end
		end
	end
end

ATMEFct.search_STATIC = function(_object, _datas)
	local objectName = _object:getName()
	if ATMEDatas.staticObjects[objectName] ~= nil then
		local findOk = ATMEFct.testFindCategory(ATMEDatas.staticObjects[objectName], _datas.findCategories)

		if findOk == true then
			if _datas.coalitionName == nil or ATMEDatas.staticObjects[objectName]:getCoalitionName() == _datas.coalitionName then
				if ATMEFct.C_FindContext_Object_isInZone(ATMEDatas.staticObjects[objectName], _datas) == true then
					ATMEFct.C_FindContext_setToList(_datas, nil, objectName)
				end
			end
		end
	end
end

ATMEFct.search_CARGO = function(_object, _datas)
	local objectName = _object:getName()
	if ATMEDatas.cargos[objectName] ~= nil then
		local findOk = ATMEFct.testFindCategory(ATMEDatas.cargos[objectName], _datas.findCategories)

		if findOk == true then
			if _datas.coalitionName == nil or ATMEDatas.cargos[objectName]:getCoalitionName() == _datas.coalitionName then
				if ATMEFct.C_FindContext_Object_isInZone(ATMEDatas.cargos[objectName], _datas) == true then
					ATMEFct.C_FindContext_setToList(_datas, nil, objectName)
				end
			end
		end
	end
end

ATMEFct.search_SCENERY = function(_object, _datas)
	local objectName = _object:getName()

	if ATMEDatas.sceneryObjects[objectName] == nil then
		ATMEFct.C_SceneryObject_create(_object)
	end
							
	if ATMEFct.C_FindContext_Object_isInZone(ATMEDatas.sceneryObjects[objectName], _datas) == true then
		ATMEFct.C_FindContext_setToList(_datas, nil, objectName)
	end
end

ATMEFct.search_WEAPON = function(_object, _datas)
	local objectName = _object:getName()
	
	if ATMEFct.C_FindContext_Object_isInZone(ATMEDatas.weapons[objectName], _datas) == true then
		ATMEFct.C_FindContext_setToList(_datas, nil, objectName)
	end
end

ATMEFct.fctZoneTask = {
	[1] = function(_param, _item)
		_param.groupId = ATMEAIIndFct.C_Group_getID(ATMEDatas.groups[_item.targetGroupName])[1]
	end,
}

ATMEFct.onZoneTasks = {	-- Tâche principale pilotable par ATME pour traiter des zones
						-- La fonction d'initialisation des paramètres est indiquée
	["FAC_AttackGroup"] = ATMEFct.fctZoneTask[1],
	["AttackGroup"] = ATMEFct.fctZoneTask[1],
}

ATMEFct.whichSide = function(_A, _B, _C)
	local vectAB = ATMEFct.C_Vector3D_create(nil, nil, _A, _B)
	local vectAC = ATMEFct.C_Vector3D_create(nil, nil, _A, _C)
	
	local det = vectAB:getX() * vectAC:getZ() - vectAB:getZ() * vectAC:getX()
	
	-- det < 0, C est à gauche de la droite AB
	-- det > 0,  C est à droite de la droite AB
	-- det = 0, C est dans sur la droite AB
	
	if det > 0 then
		return 1
	elseif det < 0 then
		return -1
	else
		return 0
	end
end	

ATMEFct.whichSideLine2D = function(_line2D, _C)
	local vector = _line2D:getVector3D()
	local origin = _line2D:getOrigin()
	
	local vectAC = ATMEFct.C_Vector3D_create(nil, nil, origin, _C)
	
	local det = vector:getX() * vectAC:getZ() - vector:getZ() * vectAC:getX()
	
	-- det < 0, C est à gauche de la droite orientée line2D
	-- det > 0,  C est à droite de la droite orientée line2D
	-- det = 0, C est dans sur la droite orientée line2D
	
	if det > 0 then
		return 1
	elseif det < 0 then
		return -1
	else
		return 0
	end
end		


ATMEFct.isAzimuthInLimits = function(_referencePoint, _azimuth, _targetPoint)
	local vector = ATMEFct.C_Vector3D_create(nil, nil, _referencePoint, _targetPoint)
	local radius = 1000
	
	if vector:getHModule() > 10000 then
		radius = 0.1 * vector:getHModule()
	end
	
	local h = math.sqrt(vector:getHModule()^2 + radius^2)
	local deltaAzimuth = math.deg(math.asin(radius/h))
	local vectorHeading = vector:getAzimuth()

	local minAzimuth = vectorHeading - deltaAzimuth
	local maxAzimuth = vectorHeading + deltaAzimuth

	if maxAzimuth >= 360 then
		maxAzimuth = maxAzimuth - 360
	end
	
	if minAzimuth < 0 then
		minAzimuth = minAzimuth + 360
	end
	
	if _azimuth >= minAzimuth and _azimuth <= maxAzimuth then
		return true
	end
	
	return false
end

ATMEFct.decompString = function(_sequence)
	local tokens = {}
	
	for _word in _sequence:gmatch("[^%s]+") do
		table.insert(tokens, _word)
    end 
   
	if #tokens > 0 then
		return tokens
	end
end

ATMEFct.joinWordsInString = function(_sequence)
	local noSpaceString = ""
	
	for _word in _sequence:gmatch("[^%s]+") do
		noSpaceString = noSpaceString .. _word
    end 
	
	return noSpaceString
end

ATMEFct.compareTags = function(_tags, _tagsToCompare)
	local function findTags(_sequence)
		local index = 1
		local cursor = 1
		local tags = {}
		local err = false
		local lg = _sequence:len()

		while cursor <= lg and err == false do
			local d,f = _sequence:find("[|]", cursor)
			if d ~= nil then
				local s = _sequence:sub(cursor, d - 1)
				local tagsTemp = ATMEFct.decompString(s)

				if tagsTemp ~= nil and #tagsTemp <= 1 then
					if #tagsTemp == 1 then
						tags[index] = tagsTemp[1]
					end
				end
				
				index = index + 1
				cursor = d + 1 
			else
				local s = _sequence:sub(cursor, lg) 
				local tagsTemp = ATMEFct.decompString(s)

				if tagsTemp ~= nil and #tagsTemp == 1 then
					tags[index] = tagsTemp[1]
					index = index + 1
					cursor = lg + 1 
				else
					err = true
				end  
			end
		end

		return err, tags
	end
	
	local err, tags = findTags(_tagsToCompare)

	if err == false then
		local ok = true
		for _i, _tag in ipairs(_tags) do
			if tags[_i] ~= nil and _tag ~= tags[_i] then
				ok = false
			end
		end

		return ok
	end
end

ATMEFct.analyzeString = function(_sequence)
	local pBegin = {}
	local pEnd = {}
	local pString = {}
	local index = 0
	
	local cursor = 1
	local lg =  _sequence:len()
      

	local function defineToken(_token)
		local sign = 1
		
		local function returnRelativeStatus(_token)
			local c = string.sub(_token, 1, 1)
			local relative = 1

			if c == "+" then
				relative = 2
			elseif c == "-" then
				relative = 2
				sign = -1
			end

			return relative
		end
		
		local function tokenType(_token, _item)
			local token = string.upper(_token)
			
			if ATMEConst.tokenTypes[token] ~= nil then
				_item.type = ATMEConst.tokenTypes[token].type
				_item.status = "ABSOLUTE"
        
				if ATMEConst.tokenTypes[token].value ~= nil then
				  _item.value = ATMEConst.tokenTypes[token].value
				else
				  _item.value = token
				end
        
				return true
			end
			
			return false
		end    

		local function tokenEval(_token, _item)
			local str
			local percent = 1
			
			local pos = returnRelativeStatus(_token)
			_item.status = ATMEConst.relativeStatus[pos]

			str = string.match(_token,"^%d%d:%d%d:%d%d$",pos)

			if str ~= nil then
				local hd = tonumber(string.sub(str,1,2))
				local md = tonumber(string.sub(str,4,5))
				local sd = tonumber(string.sub(str,7,8))

				if hd >= 0 and hd <= 23 and md >= 0 and md <= 59 and sd >= 0 and sd <= 59 then
					_item.type = "TOKEN_INTEGER"
					_item.value = (hd * 3600 + md * 60 + sd) * sign
					return true
				end
			end

			local lastChar = string.sub(_token, -1)

			local tacanChannel = false
			if lastChar == "%" then
				str = string.sub(_token, 1, _token:len() - 1)
				percent = 0.01
			elseif ATMEConst.tacanMode[lastChar] ~= nil then
				tacanChannel = true
				str = string.sub(_token, 1, _token:len() - 1)
			else
				str = _token
			end

			local n = tonumber(str)

			if n ~= nil then
				if tacanChannel == false then
					if n == math.floor(n) then
						_item.type = "TOKEN_INTEGER"
					else
						_item.type = "TOKEN_FLOAT"
					end
			
					_item.value = n * percent
					return true
				else
					local channel = math.floor(n)
					
					if channel > 0 and channel < 64 then
						_item.value = channel .. lastChar
						_item.type = "TOKEN_TACAN_CHANNEL"
						return true
					end
				end
			end

			return false
		end        

		local fcts = {
			[1] = tokenType,
			[2] = tokenEval,
		}

		local item = {}
		local indexfct = 1
		while indexfct <= #fcts and fcts[indexfct](_token, item) == false do
			indexfct = indexfct + 1
		end

		if indexfct > #fcts then
			item.value = _token
			item.type = "TOKEN_ERROR"
			item.status = "ABSOLUTE"
		end

		return item
	end
	
	local function setValue(_start, _end)
		local separator
		if _sequence:sub(_start,_start) == "'" then
			separator = "'"
		elseif _sequence:sub(_start,_start) == '"' then
			separator = '"'
		end
		
		if separator ~= nil then
			local ds,fs = string.find(_sequence, separator, _start + 1)
			if ds ~= nil then
				local d,f = string.find(_sequence, "%s+", fs)
				if fs == lg or d ~= nil then
					pBegin[index] = _start
					pEnd[index] = ds
					pString[index] = true

					if d ~= nil then
						return f
					else
						return fs
					end
				end
			end
		end
		
		pBegin[index] = _start
		pEnd[index] = _end
		pString[index] = false   
	end

	
	while cursor <= lg do
		local d,f = string.find(_sequence, "%s+", cursor)

		if d ~= nil then
			if d == cursor then
				cursor = f + 1
			else
				index = index + 1
				local fs = setValue(cursor, d - 1)
				
				if fs ~= nil then
					f = fs
				end
					
				cursor = f + 1
			end
		else
			index = index + 1
			setValue(cursor, lg)
			cursor = lg + 1
		end
	end

	local tokens = {}
	
	for i = 1,index do
		if pString[i] == false then
			tokens[i] = defineToken(_sequence:sub(pBegin[i], pEnd[i]))
		else
			local value = ""
			
			if pEnd[i] - pBegin[i] > 1 then
				value = _sequence:sub(pBegin[i] + 1, pEnd[i] - 1)
			end
			tokens[i] = {
				value = value,
				type = "TOKEN_STRING",
				status = "ABSOLUTE",
			}
		end
		tokens[i].pStart = pBegin[i]
		tokens[i].pEnd = pEnd[i]
	end

	return tokens
end

ATMEFct.getTimeValue = function(_rules)
	local tokenError = false
	local timeStart
	local isRelative = false
	
	if type(_rules) == "string" then
		local tokens = ATMEFct.analyzeString(_rules)
		
		if #tokens >= 1 and #tokens <= 3 then
			if #tokens >= 2 and tokens[1].type == "TOKEN_TYPE_RANDOM" then
				if tokens[2].type == "TOKEN_INTEGER" then
					local ok = false
					local minRandom = 0
					local maxRandom = 0
					
					if tokens[2].value >= 0 then
						maxRandom = tokens[2].value
						
						if #tokens == 3 then
							minRandom = tokens[2].value
							
							if tokens[3].type == "TOKEN_INTEGER" then
								if tokens[3].value >= 0 then
									maxRandom = tokens[3].value
									if tokens[3].status == tokens[2].status then
										ok = true
									else
										tokenError = true
										timeStart = "Parameters #2 & #3 must have same status (+ for all or nothing for all) in string format rules"
									end
								else
									tokenError = true
									timeStart = "Parameter #3 must be >= 0 in string format rules"
								end
							else
								tokenError = true
								timeStart = "Parameter #3 must be number or time value in string format rules"
							end
						else
							ok = true
						end
					else
						tokenError = true
						timeStart = "Parameter #2 must be >= 0 in string format rules"
					end
					
					if ok == true then
						if minRandom >= maxRandom  then
							tokenError = true
							timeStart = "Parameter #2 must be <= Parameter #3 in string format rules"
						else
							if tokens[2].status == "RELATIVE" then
								isRelative = true
							end
						
							timeStart = math.random(minRandom, maxRandom)
						end
					end
				else
					tokenError = true
					timeStart = "Parameter #2 must be number or time value in string format rules"
				end
			elseif tokens[1].type == "TOKEN_INTEGER" then
				if tokens[1].value >= 0 then
					if tokens[1].status == "RELATIVE" then
						isRelative = true
					end
					
					timeStart = tokens[1].value
				else
					tokenError = true
					timeStart = "Parameter #1 must be >= 0 in string format rules"
				end
			else
				tokenError = true
				timeStart = "Parameter #1 must be number or time value or 'RANDOM' in string format rules"
			end
		else
			tokenError = true
			timeStart = "Bad string format rules"
		end
	else
		isRelative = true
		if _rules >= 0 then
			timeStart = math.floor(_rules)
		else
			tokenError = true
			timeStart = "rules must be >= 0"
		end
	end

	return tokenError, timeStart, isRelative
end

ATMEFct.getRandomDatas = function(_datas)
	local randomSet = false
	local minValue = 0
	local maxValue
	
	if type(_datas) == "string" then
		local tokens = ATMEFct.analyzeString(_datas)
		local ok = false

		if #tokens >= 1 and #tokens <= 3 and (tokens[1].type == "TOKEN_TYPE_RANDOM" or tokens[1].type == "TOKEN_INTEGER") then
			if tokens[1].type == "TOKEN_TYPE_RANDOM" then
				randomSet = true
				if #tokens >= 2 and tokens[2].type == "TOKEN_INTEGER" then
					minValue = tokens[2].value
					
					if #tokens == 3 and tokens[3].type == "TOKEN_INTEGER" then
						maxValue = tokens[3].value
						ok = true
					elseif #tokens == 2 then
						maxValue = minValue
						minValue = 1
						ok = true
					else
						return "Parameter #3 must be number in string format "
					end
				else
					return "Parameter #2 must be number in string format "
				end
				
				if ok == true then
					if minValue >= maxValue then
						return "Parameter #2 must be < Parameter #3 in string format "
					end
				end
			else
				minValue = tokens[1].value
				ok = true
			end
		else
			return "Parameter #1 must be number or 'RANDOM' in string format "
		end
		
		if ok == false then
			return "Bad string format "
		end
	else
		minValue = math.floor(_datas)
	end
	
	return nil, randomSet, minValue, maxValue
end

ATMEFct.searchMainChunk = function(_err)
	local line = string.match(_err,'[^\n]* in main chunk')

	return string.sub(string.match(line,'[%w%s]*%.lua"%]'), 1, -7)
end

ATMEFct.analyzeTraceback = function (_err)
	local line = _err
	local id
	local startBack = 1
	local found = false
	local refName
	local last = false
	local numLine

	local _, pos = string.find(_err, "\n") -- La première ligne est ignorée
	pos = pos + 1

	while last == false and found == false do
		local _, f = string.find(_err, "\n", pos)

		if f ~= nil then
			line = string.sub(_err, pos, f - 1)
			id = string.sub(string.match(line,'[%w%s]*%.lua"%]'), 1, -7)
			startBack = pos
			pos = f + 1
		else -- Dernière ligne
			local d, f = string.find(_err,'[^\n]*$')
			line = string.sub(_err, d, f)
			id = string.sub(string.match(line,'[%w%s]*%.lua"%]'), 1, -7)

			startBack = d
			last = true
		end
		
		if refName == nil then
			refName = id
		elseif refName ~= id then 
			numLine = string.match(line, ":%d+:")
			if numLine ~= nil then
				found = true
			end	
		end
	end
	
	local text
	if numLine ~= nil then
		numLine = string.sub(numLine, 2, numLine:len() - 1)
		text = "At line " .. numLine .. " - "
	end
	
	return id, text, string.sub(_err, startBack, _err:len())
end

ATMEFct.getErrorPrefix = function(_object)
	ATMEDatas.error = nil
	if _object ~= nil then
		return "Class ATME." .. _object:getClassName().. "\nFunction :\n"
	else
		return "Function ATME."
	end
end

ATMEFct.errUserHandler = function(_handlerName, _isMenuHandler)
	if _isMenuHandler == nil or _isMenuHandler == false then
		ATMEFct.errorOutputCore(nil, "User function assigned to '" .. _handlerName .. "'", "fatal error or try to access a nil value", false)
	else
		ATMEFct.errorOutputCore(nil, "User function assigned to menu '" .. _handlerName .. "'", "fatal error or try to access a nil value", false)
	end
end

ATMEFct.addMark = function(_id, _coalition, _groupId, _name, _text, _point, _readOnly, _message)
	local id = _id
	
	if _id == nil then
		ATMEDatas.markId = ATMEDatas.markId + 1
		id = ATMEDatas.markId
	end
	
	if _coalition ~= nil then
		ATMEDatas.marks[id] = {
			coalitionName = _name,
		}
		
		trigger.action.markToCoalition(id, _text, _point, _coalition, _readOnly, _message)
	elseif _groupId ~= nil then
		ATMEDatas.marks[id] = {
			groupName = _name,
		}
		
		trigger.action.markToGroup(id, _text, _point, _groupId, _readOnly, _message)
	else
		ATMEDatas.marks[id] = {}
		trigger.action.markToAll(id, _text, _point, _readOnly, _message)
	end
	
	return id
end

ATMEFct.getRandomPlayerFromCoalition = function(_coalitionName)
	local playerUnits = coalition.getPlayers(ATMEConst.coalitionIds[_coalitionName])
	
	return ATMEDatas.players[playerUnits[math.random(1,#playerUnits)]:getName()]
end

ATMEFct.getDCSZone = function(_name)
	return trigger.misc.getZone(_name)
end

-- isATMEAObject : fonction interne ATME pour vérifier si une table est une instance
-- de classe ATME.

ATMEFct.isATMEObject = function(_object)
	local ok = false
	
	if type(_object) == "table" then
		xpcall(
			function() 
				local value = _object:getClassName()
		
				if type(value) == "string" then
					if type(ATMEDatas[value .. "_MT"].mt) == "table" then
						ok = true
					end
				end
			end, 
			function() 
				return false
			end)
	end
	
	return ok
end

-- isObjectClass : fonction interne ATME pour vérifier si une table est une instance
-- de classe ATME particulière.

ATMEFct.isObjectClass = function(_object, _class)
	local function testObjectClass(_object, _class)
		if _object:getClassName() == _class:getClassName() then
			return true
		else
			local object = _object:getParentClass()
			if object ~= nil then
				return testObjectClass(object, _class)
			end
		end
		
		return false
	end
	
	local function testObjectClassFromName(_object, _name)
		if _object:getClassName() == _name then
			return true
		else
			local object = _object:getParentClass()
			if object ~= nil then
				return testObjectClassFromName(object, _class)
			end
		end
		
		return false
	end

	if ATMEFct.isATMEObject(_object) == true then
		if ATMEFct.isATMEObject(_class) == true then
			return testObjectClass(_object, _class)
		elseif type(_class) == "string" then
			return testObjectClassFromName(_object, _class)
		end
	end
	
	return false
end



ATMEFct.isReference = function(_reference)
	if ATME.isPoint(_reference) ~= POINT_BAD then
		return true
	elseif ATMEFct.isATMEObject(_reference) == true then
		return ATMEConst.referenceClasses[_reference:getClassName()] ~= nil
	else
		return false
	end
end

ATMEFct.convertToPoint3D = function(_point)
	local point = {}
			
	point.x = _point.x
	if ATME.isPoint(_point) == POINT_2D then
		point.z = _point.y
		point.y = ATMEFct.getSurfaceAltitude(_point)
	else
		point.y = _point.y
		point.z = _point.z
	end
	
	return point
end

ATMEFct.getPointFromReference = function(_reference)
	if ATME.isPoint(_reference) ~= POINT_BAD then
		return ATMEFct.convertToPoint3D(_reference)
	else
		return _reference:getPosition()
	end
end

ATMEFct.coordLocalToDCS = function(_point, _originReference, _azimuth)
		-- coordLocalToDCS : Transforme les coordonnées locales d'un point relatives au repère défini par sa référence d'origine et son azimuth
		
	local angle = math.rad(_azimuth)
	local origin = ATMEFct.getPointFromReference(_originReference)
	local point
	
	if _point.z == nil then
		point = {
			x = _point.x,
			z = _point.y,
		}
	else
		point = _point
	end
	
	local pointDCS =  ATMEFct.convertToPoint3D({
										x = origin.x + point.x * math.cos(angle) - point.z * math.sin(angle),
										y = origin.z + point.x * math.sin(angle) + point.z * math.cos(angle),
									})

	local altitude = point.y
	if altitude ~= nil and altitude > pointDCS.y then
			pointDCS.y = altitude
	end

	return pointDCS
end

ATMEFct.hasAttributes = function(_object, _attributes)
	local found = false

	for _, _attr in ipairs(_attributes) do
		if _object:hasAttribute(ATMEConst.targetsAttributes[_attr]) == true or _attr == "ALL" then
			found = true
		end
	end

	return found
end

ATMEFct.testFindCategory = function(_objectATME, _findCategories)
	local found = false

	for _, _category in ipairs(_findCategories) do
		if _objectATME:getCategoryName() == _category or _category == "ALL" then
			found = true
		end
	end

	return found
end

ATMEFct.extractDatasFromStringOrTable = function(_attributes, _testFindCategorie)
	local datas = {}
	
	if type(_attributes) == "string" then
		local id = 1
		for _item in _attributes:gmatch("([^%s]+)") do
			if _testFindCategorie == true and ATMEConst.findCategories[_item] == nil then
				return true, "item #" .. id .. " must be 'AIRPLANE', 'HELICOPTER', 'GROUND', 'SHIP', 'CARGO', 'FORTIFICATION', 'WAREHOUSE', 'HELIPAD' or 'AIRBASE'"
			end
			id = id + 1
			table.insert(datas, string.upper(_item))
		end
	elseif type(_attributes) == "table" then
		for _id, _item in ipairs(_attributes) do
			if type(_item) == "string" then
				if _testFindCategorie == true and ATMEConst.findCategories[_item] == nil then
					return true, "item #" .. _id .. " must be 'AIRPLANE', 'HELICOPTER', 'GROUND', 'SHIP', 'CARGO', 'FORTIFICATION', 'WAREHOUSE', 'HELIPAD' or 'AIRBASE'"
				end
				
				table.insert(datas, string.upper(_item))
			else
				return true, "item #" .. _id .. " type must be string"
			end
		end
	else
		if _testFindCategorie == true then
			return true, "findCategories type must be nil, string or table"
		else
			return true, "attributes type must be string or table"
		end
	end
	
	return false, datas
end

ATMEFct.getSurfaceAltitude = function(_point)
	-- getSurfaceAltitude : Donne l'altitude du terrain à la verticale du point
	if  _point.z == nil then
		return land.getHeight({x = _point.x, y = _point.y})
	else
		return land.getHeight({x = _point.x, y = _point.z})
	end
end

ATMEFct.getAzimuthMod = function(_azimuth, _mod)
	local azimuth = math.fmod(_azimuth, 360)

	if _mod == 360 then	
		if azimuth < 0 then 
			azimuth = azimuth + 360
		end
	else
		if azimuth < -180 then
			azimuth = azimuth + 360
		elseif azimuth > 180 then
			azimuth = azimuth - 360
		end
	end
	
	return azimuth
end

ATMEFct.isTakeOffWaypoint = function(_type)
	return _type == "TakeOffParking" or _type == "TakeOffParkingHot" or _type == "TakeOff"
end

ATMEFct.isLandingWaypoint = function(_type)
	return _type == "Land"
end

----------------------------------------------------------------------------------------------------------
-- DEFINITION DE LA TABLE ATME ET DES FONCTIONS GENERALES
--
-- Bien que reprenant la structure des classes définies plus loin, ATME n'est pas une classe en tant que telle.
-- ATME définit des fonctions de base et la liste des classes existantes dans ATME_Core. A ce titre, elle n'est
-- pas préfixée par C_
--
-- ATME_MT définit une métatable pour ATME (mt) et pour les extensions de modules afin d'avoir une variable globale
-- ATME.modules["nom du module"] pour les modules lancés. L'entrée "keys" interdit la création de clés nouvelles
-- dans la table ATME.
----------------------------------------------------------------------------------------------------------

ATMEDatas.ATME_MT = {
	mt = {},
	mtModules = {},

	keys = {
		["C_Vector3D"] = true,
		["C_Line2D"] = true,
		["C_IndexList"] = true,
		["C_Stack"] = true,
		["C_Airbase"] = true,
		["C_ZonesMgr"] = true,
		["C_F10Menu"] = true,
		["C_Alarm"] = true,
		["C_Module"] = true, 
		["C_Smoke"] = true,
		["C_Flare"] = true,
		["C_Event"] = true,
		["C_CloningContext"] = true,
		["C_StaticCloningContext"] = true,
		["C_CargoCloningContext"] = true,
		["C_Group"] = true,
		["C_Player"] = true,
		["C_AIUnit"] = true,
		["C_StaticObject"] = true,
		["C_Cargo"] = true,
		["C_Race"] = true,
		["C_Task"] = true,
		["C_Route"] = true,
		["C_Sound"] = true,
		["C_Circle"] = true,
		["C_Rectangle"] = true,
		["C_Polygon"] = true,
		["C_Zone"] = true,
		["C_Zone2D"] = true,
		["C_Zone3D"] = true,
		["C_MovableZone2D"] = true,
		["C_MovableZone3D"] = true,
		["C_DynamicObject"] = true,
		["C_FindContext"] = true,
		["C_Unit"] = true,
		["C_Segment2D"] = true,
		["C_PickupZone"] = true,
		["C_Sphere"] = true,
		["C_Cylinder"] = true,
		["C_Box"] = true,
		["C_Weapon"] = true,
		["C_SceneryObject"] = true,
		["C_RadioChannel"] = true,
		["run"] = false,
		["modules"] = false,
		["onTrackWaypoints"] = false,
	},
}

ATME = {
	getLanguage = function()
		return ATMEDatas.language
	end,
	
	getNbPlayers = function()
		return ATMEDatas.nbPlayers
	end,
	
	isPoint = function(_point)
		-- isPoint : Vérifie si la table passée en paramètre est un point et retourne POINT_BAD (0) si ce n'est pas un point :
		--           Si point 2D si présence de deux nombres _point.x et _point.y -> Retourne POINT_2D (1)
		--			 Si point 3D si présence de trois nombres _point.x, _point.y et _point.z/gModule -> Retourne POINT_3D (2)
		if type(_point) == "table" then
			if type(_point.x) == "number" and type(_point.y) == "number" then
				if type(_point.z) == "number" then
					return POINT_3D
				else
					return POINT_2D
				end
			end
		end
		
		return POINT_BAD
	end,
	
	isPoint2D = function(_point)
		-- isPoint2D : Vérifie si la table passée en paramètre est un point 2D, avec x et y renseignés
		--             Ces nombres sont relatifs au plan horizontal XZ de DCS, y étant en 2D la coordonnée sur Z.
		if type(_point) == "table" then
			if type(_point.x) == "number" and type(_point.y) == "number" then
				return true
			end
		end
		
		return false
	end,
	
	isPoint3D = function(_point)
		-- isPoint3D : Vérifie si la table passée en paramètre est un point 3D, avec x, y et z renseignés
		--             Ces nombres sont relatifs au repère XYZ de DCS
		if type(_point) == "table" then
			if type(_point.x) == "number" and type(_point.y) == "number" then
				if type(_point.z) == "number" then
					return true
				end
			end
		end
		
		return false
	end,
	
	getError = function()
		if ATMEDatas.error ~= nil then
			local errorCode = ATMEDatas.error.code
			ATMEDatas.error = nil
			
			return errorCode
		end
	end,
	
	isObject = function(_object, _class)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "isObject(object, [class])"
		
		if _class ~= nil then
			if ATMEFct.isATMEObject(_class) == true then
				return ATMEFct.isObjectClass(_object, _class)
			elseif type(_class) == "string" then
				if ATMEDatas.ATME_MT.keys[_class] ~= nil and ATMEDatas.ATME_MT.keys[_class] == true then
					return ATMEFct.isObjectClass(_object, _class)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "class must be an ATME class name", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "class type must be string or ATME class", false)
			end
		else
			return ATMEFct.isATMEObject(_object)
		end
	end,
	
	getTheatre = function()
		-- getTheatre : Retourne la carte (map) associée à la mission
		-- 			    actuellement "NEVADA", "CAUCASUS", "NORMANDY" ou ""
		
		return ATMEDatas.theatreName
	end,
	
	getFlag = function(_id)
		-- getFlag : Retourne la valeur d'un flag. 
		-- 			 Pour être compatible avec le flags de l'éditeur de mission, il faut un id entier positif.
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getFlag(id)"
		
		if type(_id) == "number" or type(_id) == "string" then
			local id
			if type(_id) == "number" then
				if _id > 0 then
					id = tostring(math.floor(_id))
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "id must be > 0 if number type", false)
					return
				end
			else
				id = _id
			end
			
			return trigger.misc.getUserFlag(id)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "id type must be number", false)
		end
	end,
	
	setFlag = function(_id, _value)
		-- setFlag : Modifie la valeur l'état d'un flag.
		--			 Pour être compatible avec le flags de l'éditeur de mission, il faut un id entier positif. Si le flag
		--			 est modifié ici et qu'il existe un trigger dans l'éditeur sur ce même flag pour la valeur renseignée,
		--           le trigger défini dans l'éditeur de mission sera déclenché
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "setFlag(id, value)"
		
		if (type(_id) == "number" or type(_id) == "string") and (type(_value) == "number" or type(_value) == "boolean") then
			local id
			if type(_id) == "number" then
				if _id > 0 then
					id = tostring(math.floor(_id))
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "id must be > 0 if number type", false)
					return
				end
			else
				id = _id
			end
			
			if (type(_value) == "number" and _value >= 0) or type(_value) == "boolean" then
				if type(_value) == "number" then
					_value = math.floor(_value)
				end
				
				trigger.action.setUserFlag(id, _value)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "value must be > 0 or true/false", false)
			end
		else
			if type(_id) ~= "number" and type(_id) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "id type must be number or string", false)
			elseif type(_value) ~= "number" and type(_value) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "value type must be number or boolean", false)
			end
		end
	end,
	
	getDCSZone = function(_name)
		-- getDCSZone : Retourne une zone définie dans l'éditeur de mission, nil si inexistante.
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getDCSZone(name)"
		
		if type(_name) == "string" then
			return ATMEFct.getDCSZone(_name) 
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
		end
	end,
	
	getTime = function()
		-- getTime : Retourne le temps de la mission.
		
		return timer.getTime()
	end,
	
	getAzimuth360 = function(_azimuth)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getAzimuth360(azimuth)"
		
		if type(_azimuth) == "number" then
			return ATMEFct.getAzimuthMod(_azimuth, 360)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
		end
	end,
	
	getAzimuth180 = function(_azimuth)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getAzimuth180(azimuth)"
		
		if type(_azimuth) == "number" then
			return ATMEFct.getAzimuthMod(_azimuth, 180)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
		end
	end,
	
	copyTable = function(_tableToCopy)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "copyTable(tableToCopy)"
		
		if type(_tableToCopy) == "table" then
			return ATMEFct.copyTable(_tableToCopy)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tableToCopy type must be table", false)
		end
	end,
	
	markToAll = function(_text, _reference, _readOnly, _message)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "markToAll(text, reference, readOnly, [message])"

		if type(_text) == "string" and ATMEFct.isReference(_reference) == true and type(_readOnly) == "boolean" and 
					(_message == nil or type(_message) == "string") then
			local point = ATMEFct.getPointFromReference(_reference)

			return ATMEFct.addMark(nil, nil, nil, nil, _text, point, _readOnly, _message)
		else
			if type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", false)
			elseif ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_readOnly) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "readOnly type must be boolean", false)
			elseif _message ~= nil and type(_message) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "message type must be nil or string", false)
			end
		end
	end,
	
	markToCoalition = function(_coalitionName, _text, _reference, _readOnly, _message)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "markToCoalition(coalitionName, text, reference, readOnly, [message])"
		
		if type(_coalitionName) == "string" and type(_text) == "string" and ATMEFct.isReference(_reference) == true and
					type(_readOnly) == "boolean" and (_message == nil or type(_message) == "string") then
					
			local point = ATMEFct.getPointFromReference(_reference)
			local coalitionName = string.upper(_coalitionName)
			
			if ATMEConst.coalitionIdsBlueRed[coalitionName] ~= nil then
				return ATMEFct.addMark(nil, ATMEConst.coalitionIds[coalitionName], nil, coalitionName, _text, point, _readOnly, _message)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'BLUE' or 'RED'", false)
			end
		else
			if type(_coalitionName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", false)
			elseif type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", false)
			elseif ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_readOnly) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "readOnly type must be boolean", false)
			elseif _message ~= nil and type(_message) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "message type must be nil or string", false)
			end
		end
	end,
	
	markRemove = function(_id)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "markRemove(id)"
		
		if type(_id) == "number" then
			local id = math.floor(_id)
			
			if ATMEDatas.marks[id] ~= nil then
				trigger.action.removeMark(id)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "id type must be number", false)
		end
	end,
	
	soundForAll = function(_text, _readMode)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "soundForAll(text, [readMode])"
		
		if type(_text) == "string" and (_readMode == nil or type(_readMode) == "boolean") then
			local randomRead = false
			local readOnce = true
			
			if type(_readMode) == "boolean" then
				randomRead = _readMode
				readOnce = false
			end
			
			local err, playlist = ATMEFct.C_Sound_getPlaylist(_text, randomRead)

			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, playlist, false)
				return
			end

			if playlist == nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text string has no sound token", false)
				return
			end
			
			ATMEFct.C_Sound_initPlaylist({
				randomRead = randomRead, 
				readOnce = readOnce, 
				playlist = playlist,
				timeSound = timer.getTime(),
			})
		else
			if type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", false)
			elseif _readMode ~= nil and type(_readMode) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "readMode type must be nil or boolean", false)
			end
		end
	end,
	
	soundForCoalition = function(_coalitionName, _text, _readMode)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "soundForCoalition(coalitionName, text, [readMode])"
		
		if type(_coalitionName) == "string" and type(_text) == "string" and (_readMode == nil or type(_readMode) == "boolean") then
			local coalitionName = string.upper(_coalitionName)
			if ATMEConst.coalitionIdsBlueRed[coalitionName] ~= nil then
				local randomRead = false
				local readOnce = true
				
				if type(_readMode) == "boolean" then
					randomRead = _readMode
					readOnce = false
				end
				
				local err, playlist = ATMEFct.C_Sound_getPlaylist(_text, randomRead)

				if err == true then
					ATMEFct.errorOutputCore(nil, fctDefinition, playlist, false)
					return
				end

				if playlist == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "text string has no sound token", false)
					return
				end
			
				ATMEFct.C_Sound_initPlaylist({
					coalition = ATMEConst.coalitionIds[coalitionName],
					randomRead = randomRead, 
					readOnce = readOnce, 
					playlist = playlist,
					timeSound = timer.getTime(),
				})
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'BLUE' or 'RED'", false)
			end
		else
			if type(_coalitionName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", false)
			elseif type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", false)
			elseif _readMode ~= nil and type(_readMode) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "readMode type must be nil or boolean", false)
			end
		end
	end,
	
	soundForCountry = function(_countryName, _text, _readMode)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "soundForCountry(countryName, text, [readMode])"
		
		if type(_countryName) == "string" and type(_text) == "string" and (_readMode == nil or type(_readMode) == "boolean") then
			local countryName = string.upper(_countryName)
			if ATMEConst.countries[countryName] ~= nil then
				local randomRead = false
				local readOnce = true
				
				if type(_readMode) == "boolean" then
					randomRead = _readMode
					readOnce = false
				end
				
				local err, playlist = ATMEFct.C_Sound_getPlaylist(_text, randomRead)

				if err == true then
					ATMEFct.errorOutputCore(nil, fctDefinition, playlist, false)
					return
				end

				if playlist == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "text string has no sound token", false)
					return
				end
			
				ATMEFct.C_Sound_initPlaylist({
					country = ATMEConst.countries[countryName],
					randomRead = randomRead, 
					readOnce = readOnce, 
					playlist = playlist,
					timeSound = timer.getTime(),
				})
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Country " .. _countryName .. " doesn't exist", false)
			end
		else
			if type(_countryName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "countryName type must be string", false)
			elseif type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", false)
			elseif _readMode ~= nil and type(_readMode) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "readMode type must be nil or boolean", false)
			end
		end
	end,
	
	explode = function(_reference, _power)
		-- explode : Génère une explosion au point indiqué.
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "explode(reference, power)"

		if ATMEFct.isReference(_reference) == true and type(_power) == "number" then
			local power = math.floor(_power)
			if power > 0 then
				trigger.action.explosion(ATMEFct.getPointFromReference(_reference), power)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power must be > 0", false)
			end
		else
			if ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_power) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", false)
			end
		end
	end,
	
	displayForAll = function(_text, _duration, _clear)
		-- displayForAll : Affiche un message à tous les joueurs actifs avec un durée donnée
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "displayForAll(text, duration, [clear])"
		
		if type(_text) == "string" and type(_duration) == "number" and
				(_clear == nil or type(_clear) == "boolean") then
				
			local clear = false
			
			if type(_clear) == "boolean" then
				clear = _clear
			end
			
			if _duration >= 0 then
				trigger.action.outText(_text, _duration, clear)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be >= 0", false)
			end
		else
			if type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", false)
			elseif type(_duration) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be string", false)
			elseif _clear ~= nil and type(_clear) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "clear type must be nil or string", false)
			end
		end
	end,
	
	displayForCoalition = function(_coalitionName, _text, _duration, _clear)
		-- displayForCoalition : Affiche un message à tous les joueurs actifs d'une coalition avec une durée donnée
		--                       La coalition doit être "RED" ou "BLUE"
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "displayForCoalition(coalitionName, text, duration, [clear])"
		
		if type(_coalitionName) == "string" and type(_text) == "string" and type(_duration) == "number" and
				(_clear == nil or type(_clear) == "boolean") then
				
			local clear = false
			
			if type(_clear) == "boolean" then
				clear = _clear
			end
			
			if _duration >= 0 then
				local coalitionName = string.upper(_coalitionName)
				if ATMEConst.coalitionIdsBlueRed[coalitionName] ~= nil then
					trigger.action.outTextForCoalition(ATMEConst.coalitionIds[coalitionName], _text, _duration, clear)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'BLUE' or 'RED'", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be >= 0", false)
			end
		else
			if type(_coalitionName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", false)
			elseif type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", false)
			elseif type(_duration) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be string", false)
			elseif _clear ~= nil and type(_clear) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "clear type must be nil or string", false)
			end
		end
	end,
	
	displayForCountry = function(_countryName, _text, _duration, _clear)
		-- displayForCoalition : Affiche un message à tous les joueurs actifs d'une coalition avec une durée donnée
		--                       La nation est celle définie par DCS.
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "displayForCountry(countryName, text, duration, [clear])"
		
		if type(_countryName) == "string" and type(_text) == "string" and type(_duration) == "number" and
				(_clear == nil or type(_clear) == "boolean") then
			
			local clear = false
			
			if type(_clear) == "boolean" then
				clear = _clear
			end
			
			if _duration >= 0 then
				_countryName = string.upper(_countryName)
				if country.id[_countryName] ~= nil then
					trigger.action.outTextForCountry(country.id[_countryName], _text, _duration, clear)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Country " .. _countryName .. " doesn't exist", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be >= 0", false)
			end
		else
			if type(_countryName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "countryName type must be string", false)
			elseif type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", false)
			elseif type(_duration) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be string", false)
			elseif _clear ~= nil and type(_clear) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "clear type must be nil or string", false)
			end
		end
	end,
	
	logDisplay = function()
		-- setDebugLevel : Définit le niveau de débug. Tout module en mode debug est concerné. 
		--				   Les modules en mode normal ne sont pas concernés
		--                 Ce niveau permet de filtre l'affichage des messages d'erreur pour les niveaux supérieurs ou égaux
		--                (niveau 1, le plus haut à niveau 9 le plus bas).
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "logDisplay()"
		
		ATMEDatas.logDisplay = true
	end,
	
	isSurfaceWater = function(_reference)
		-- isSurfaceWater : Vérifie si un point est au dessus de l'eau
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "isSurfaceWater(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATMEFct.getPointFromReference(_reference)
			
			local landType = land.getSurfaceType({
													x = point.x,
													y = point.z,
												})
			
			return landType == land.SurfaceType.WATER or landType == land.SurfaceType.SHALLOW_WATER
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		end
	end,
	
	isSurfaceRunway = function(_reference)
		-- isSurfaceRunway : Vérifie si un point est au dessus d'une piste ou d'un taxiway
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "isSurfaceRunway(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATMEFct.getPointFromReference(_reference)
			
			local landType = land.getSurfaceType({
													x = point.x,
													y = point.z,
												})
			
			return landType == land.SurfaceType.RUNWAY
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		end
	end,
	
	isSurfaceLand = function(_reference)
		-- isSurfaceRunway : Vérifie si un point est au dessus de la terre hors piste ou taxiway
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "isSurfaceLand(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATMEFct.getPointFromReference(_reference)
			
			local landType = land.getSurfaceType({
													x = point.x,
													y = point.z,
												})
			
			return landType == land.SurfaceType.LAND
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		end
	end,
	
	isSurfaceRoad = function(_reference)
		-- isSurfaceRoad : Vérifie si un point est au dessus de la terren hors piste ou taxiway
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "isSurfaceRoad(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATMEFct.getPointFromReference(_reference)
			
			local landType = land.getSurfaceType({
													x = point.x,
													y = point.z,
												})
			
			return landType == land.SurfaceType.ROAD
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		end
	end,

	getSurfaceAltitude = function(_reference)
		-- getSurfaceAltitude : Donne l'altitude du terrain à la verticale du point
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getSurfaceAltitude(reference)"
		
		if  ATMEFct.isReference(_reference) == true then
			return ATMEFct.getSurfaceAltitude(ATMEFct.getPointFromReference(_reference))
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		end
	end,
	
	getPointAltitude = function(_point)
		-- getPointAltitude : Donne l'altitude du point. Si le point est 2D, ce sera
		--					  l'altitude du terrain
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getPointAltitude(point)"
		
		local isPoint = ATME.isPoint(_point)
		if  isPoint ~= POINT_BAD then
			if isPoint == POINT_3D then
				return _point.y
			else
				return ATMEFct.getSurfaceAltitude(_point)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "point type must be table (x,y)/(x,y,z)", false)
		end
	end,
	
	getOnRoadPoint = function(_reference)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getOnRoadPoint(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATMEFct.getPointFromReference(_reference)
			
			local xTemp, yTemp = land.getClosestPointOnRoads("roads", point.x, point.z)
			
			return ATMEFct.convertToPoint3D({x = xTemp, y = yTemp})
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		end
	end,
	
	convertToPoint3D = function(_point)
		-- Permet de convertir un point en point 3D
		-- afin de s'assurer que le point retourné est bien 3D.
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertToPoint3D(point)"
		
		if ATME.isPoint(_point) ~= POINT_BAD then
			return ATMEFct.convertToPoint3D(_point)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "point type must be table (x,y)/(x,y,z)", false)
		end
	end,
	
	areVisible = function(_refA, _refB)
		-- areVisible : Vérifie si deux références sont mutuellement visibles (ie non masqués) par le terrain
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "areVisible(refA, refB)"
		
		if ATMEFct.isReference(_refA) == true and ATMEFct.isReference(_refB) == true then
			local pA = ATMEFct.getPointFromReference(_refA)
			local pB = ATMEFct.getPointFromReference(_refB)
			
			-- Ajout d'un offset sur l'altitude pour éviter un pb si l'objet est sur le 
			-- sol avec une altitude dans le sol (cf doc DCS pour land.isVisible)
			pA.y = pA.y + 3
			pB.y = pB.y + 3
			return land.isVisible(pA, pB)
		else
			if ATMEFct.isReference(_refA) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "refA type must be" .. ATMEConst.errorLabelReference, false)
			elseif ATMEFct.isReference(_refB) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "refB type must be" .. ATMEConst.errorLabelReference, false)
			end
		end
	end,
	
	getTemperatureAndPressure = function(_reference)
		-- getWindVector : Retourne un instance de ATME.C_Vector3D correspond au vecteur vent
		--                 avec ou sans turbulences
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getTemperatureAndPressure(reference)"
		
		if  ATMEFct.isReference(_reference) == true then
			local point = ATMEFct.getPointFromReference(_reference)
			
			return atmosphere.getTemperatureAndPressure(point) 
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		end
	end,
	
	getWindVector = function(_reference, _withTurbulence)
		-- getWindVector : Retourne un instance de ATME.C_Vector3D correspond au vecteur vent
		--                 avec ou sans turbulences
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getWindVector(reference, withTurbulence)"
		
		if  ATMEFct.isReference(_reference) == true and type(_withTurbulence) == "boolean" then
			local point = ATMEFct.getPointFromReference(_reference)
			
			local alt = ATMEFct.getSurfaceAltitude(point)
			if point.y - alt < 20 then 
				point.y = alt + 20
			end
			
			if _withTurbulence == true then
				local wind = atmosphere.getWindWithTurbulence(point)
				return ATMEFct.C_Vector3D_create(nil, nil, wind) 
			else
				local wind = atmosphere.getWind(point)
				return ATMEFct.C_Vector3D_create(nil, nil, wind) 
			end
		else
			if ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_withTurbulence) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "withTurbulence type must be boolean", false)
			end
		end
	end,
	
	getWindSpeed = function(_reference, _withTurbulence)
		-- getWindSpeed : Retourne la vitesse du vent calculée sur les 3 dimensions
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getWindSpeed(reference, withTurbulence)"
		
		if ATMEFct.isReference(_reference) == true and type(_withTurbulence) == "boolean" then
			local wind = ATME.getWindVector(_reference, _withTurbulence)

			return wind:getModule()
		else
			if ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_withTurbulence) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "withTurbulence type must be boolean", false)
			end
		end
	end,
	
	getWindHSpeed = function(_reference, _withTurbulence)
		-- getWindSpeed : Retourne la vitesse horizontale du vent calculée sur les 2 dimensions x et z
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getWindHSpeed(reference, withTurbulence)"
		
		if ATMEFct.isReference(_reference) == true and type(_withTurbulence) == "boolean" then
			local wind = ATME.getWindVector(_reference, _withTurbulence)

			return wind:getHModule()
		else
			if ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_withTurbulence) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "withTurbulence type must be boolean", false)
			end
		end
	end,
	
	getWindAzimuth = function(_reference, _withTurbulence)
		-- getWindAzimuth : Retourne l'azimuth du vent (attention pas son origine mais sa direction)
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getWindAzimuth(reference, withTurbulence)"
		
		if ATMEFct.isReference(_reference) == true and type(_withTurbulence) == "boolean" then
			local wind = ATME.getWindVector(_reference, _withTurbulence)

			return wind:getAzimuth()
		else
			if ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_withTurbulence) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "withTurbulence type must be boolean", false)
			end
		end
	end,
	
	getTags = function(_text)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getTags(text)"
		
		if type(_text) == "string" then
			return ATMEFct.decompString(_text)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", false)
		end
	end,
		
	convertMtoNM = function(_distance)
		-- convertMtoNM : Convertit des mètres en miles nautiques
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertMtoNM(distance)"
		
		if type(_distance) == "number" then
			return _distance * 0.000539957
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "distance type must be number", false)
		end
	end,

	convertNMtoM = function(_distance)
		-- convertNMtoM : Convertit des miles nautiques en mètres
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertNMtoM(distance)"
		
		if type(_distance) == "number" then
			return _distance / 0.000539957
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "distance type must be number", false)
		end
	end,
	
	convertMtoFts = function(_distance)
		-- convertMtoFts : Convertit des mètres en pieds
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertMtoFts(distance)"
		
		if type(_distance) == "number" then
			return _distance * 3.28084
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "distance type must be number", false)
		end	
	end,
	
	convertFtstoM =  function(_distance)
		-- convertFtstoM : Convertit des pieds en mètres
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertFtstoM(distance)"
		
		if type(_distance) == "number" then
			return _distance / 3.28084
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "distance type must be number", false)
		end
	end,
	
	convertMpstoNds = function(_speed)
		-- convertMpstoNds : Convertit des m/s en noeuds
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertMpstoNds(speed)"
		
		if type(_speed) == "number" then
			return _speed * 1.9438
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "speed type must be number", false)
		end
	end,
	
	convertNdstoMps = function(_speed)
		-- convertNdstoMps : Convertit des noeuds en m/s
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertNdstoMps(speed)"
		
		if type(_speed) == "number" then
			return _speed / 1.9438
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "speed type must be number", false)
		end
	end,

	convertMpstoKph = function(_speed)
		-- convertMpstoKph : Convertit des m/s en km/h
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertMpstoKph(speed)"
		
		if type(_speed) == "number" then
			return _speed * 3.6
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "speed type must be number", false)
		end
	end,

	convertKphtoMps = function(_speed)
		-- convertKphtoMps : Convertit des km/h en m/s
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertKphtoMps(speed)"
		
		if type(_speed) == "number" then
			return _speed / 3.6
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "speed type must be number", false)
		end
	end,
	
	convertToDMS = function(_degrees)
		-- convertToDMS : Convertit des angles en degrés décimaux en degrés/minutes/secondes/centièmes
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertToDMS(degrees)"
		
		if type(_degrees) == "number" then
			return ATMEFct.convertToDMS(_degrees)
		else
			if type(_degrees) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "degrees type must be number", false)
			end
		end
	end,
	
	convertToDMd = function(_degrees)
		-- convertToDMd : Convertit des angles en degrés décimaux en degrés/minutes/décimales haute précision
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertToDMd(degrees)"
		
		if type(_degrees) == "number" then
			return ATMEFct.convertToDMd(_degrees, true)
		else
			if type(_degrees) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "degrees type must be number", false)
			end
		end
	end,
	
	LLToDMSString = function(_latitude, _longitude, _highPrecision)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "LLToDMSString(latitude, longitude, [highPrecision])"
		
		if type(_latitude) == "number" and type(_longitude) == "number" and (_highPrecision == nil or type(_highPrecision) == "boolean") then
			local highPrecision = false
			if type(_highPrecision) == "boolean" then
				highPrecision = _highPrecision
			end
			
			local dirlat, dlat, mlat, slat, ddlat  = ATMEFct.convertToDMS(_latitude)
			local dirlon, dlon, mlon, slon, ddlon  = ATMEFct.convertToDMS(_longitude)
			
			local westEast = "N"
			if dirlat == -1 then
				westEast = "S"
			end
			
			local northSouth = "E"
			if dirlon == -1 then
				northSouth = "W"
			end
			
			local textlat
			local textlon

			if highPrecision == true then
				textlat = string.format("%s%02d°%02d'%02d\"%02d", westEast, dlat, mlat, slat, ddlat)
				textlon = string.format("%s%03d°%02d'%02d\"%02d", northSouth, dlon, mlon, slon, ddlon)
			else
				textlat = string.format("%s%02d°%02d'%02d\"", westEast, dlat, mlat, slat)
				textlon = string.format("%s%03d°%02d'%02d\"", northSouth, dlon, mlon, slon)
			end
			
			return textlat, textlon
		else
			if type(_latitude) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "latitude type must be number", false)
			elseif type(longitude) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "longitude type must be number", false)
			elseif _highPrecision ~= nil and type(_highPrecision) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "highPrecision type must be nil or boolean", false)
			end
		end
	end,
	
	LLToDMdString = function(_latitude, _longitude, _highPrecision)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "LLToDMdString(latitude, longitude, [highPrecision])"
		
		if type(_latitude) == "number" and type(_longitude) == "number" and (_highPrecision == nil or type(_highPrecision) == "boolean") then
			local highPrecision = false
			if type(_highPrecision) == "boolean" then
				highPrecision = _highPrecision
			end

			local dirlat, dlat, mlat, slat  = ATMEFct.convertToDMd(_latitude, highPrecision)
			local dirlon, dlon, mlon, slon  = ATMEFct.convertToDMd(_longitude, highPrecision)

			local westEast = "N"
			if dirlat == -1 then
				westEast = "S"
			end
			
			local northSouth = "E"
			if dirlon == -1 then
				northSouth = "W"
			end
		
			local textlat
			local textlon
			if highPrecision == true then
				textlat = string.format("%s%02d°%02d.%04d", westEast, dlat, mlat, slat)
				textlon = string.format("%s%03d°%02d.%04d", northSouth, dlon, mlon, slon)
			else
				textlat = string.format("%s%02d°%02d.%02d", westEast, dlat, mlat, slat)
				textlon = string.format("%s%03d°%02d.%02d", northSouth, dlon, mlon, slon)	
			end
			
			return textlat, textlon
		else
			if type(_latitude) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "latitude type must be number", false)
			elseif type(longitude) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "longitude type must be number", false)
			elseif _highPrecision ~= nil and type(_highPrecision) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "highPrecision type must be nil or boolean", false)
			end
		end
	end,
	
	convertToHMS = function (_seconds)
		-- convertToHMS : Convertit un temps en secondes en heures/minutes/secondes/100ème
		-- En cas de temps négatif, retourne une valeur négative sur la première valeur significative,
		-- h, m ou s -> exemple : -60s donnera m = -1
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertToHMS(seconds)"
		
		if type(_seconds) == "number" then
			local seconds = _seconds
			local sign = 1
			if _seconds < 0 then
				seconds = -_seconds
				sign = -1
			end
	
			local h = math.floor(seconds/3600)
			local m =  math.floor((seconds - h * 3600)/60)
			local s = math.floor(seconds - h * 3600 - m * 60)
			local d = (seconds - math.floor(seconds)) * 100
			
			if h > 0 then
				h = h * sign
			elseif m > 0 then
				m = m * sign
			elseif s > 0 then
				s = s * sign
			else
				d = d * sign
			end
			
			return h, m, s, d
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "seconds type must be number", false)
		end
	end,

	convertPointToLL = function(_point)
		-- convertPointToLL : Convertit un point en coordonnées latitude/longitude/altitude
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertPointToLL(point)"
		
		if ATME.isPoint(_point) ~= POINT_BAD then
			return coord.LOtoLL(ATMEFct.convertToPoint3D(_point))
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "point type must be table (x,y)/(x,y,z)", false)
		end
	end,
	
	convertLLToPoint = function(_latitude, _longitude, _altitude)
		-- convertLLToPoint : Convertit un triplet latitude/longitude/altitude en un point 3D
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertLLToPoint(latitude, longitude, altitude)"
		
		if type(_latitude) == "number" and type(_longitude) == "number" and type(_altitude) == "number" then
			return coord.LLtoLO(_latitude, _longitude, _altitude)
		else
			if type(_latitude) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "latitude type must be number", false)
			elseif type(_longitude) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "longitude type must be number", false)
			elseif type(_altitude) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "altitude type must be number", false)
			end
		end
	end,
	
	convertLLToMGRS = function(_latitude, _longitude)
		-- convertLLToMGRS : Convertit une latitude/longitude en référence MGRS
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertLLToMGRS(latitude, longitude)"
		
		if type(_latitude) == "number" and type(_longitude) == "number" then
			return coord.LLtoMGRS(_latitude, _longitude)
		else
			if type(_latitude) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "latitude type must be number", false)
			elseif type(_longitude) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "longitude type must be number", false)
			end
		end
	end,
	
	convertMGRSToLL = function(_tableMGRS)
		-- convertMGRSToLL : Convertit une référence MGRS en latitude/longitude
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "convertMGRSToLL(tableMGRS)"
		
		if type(_tableMGRS) == "table" and type(table.UTMZone) == "string" and type(table.MGRSDigraph) == "string" and type(table.Easting) == "number" and type(table.Northing ) == "number" then
			return coord.MGRStoLL(_tableMGRS)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tableMGRS must have MGRS format", false)
		end
	end,
	
	getHScalarProductFromPoints = function(_A, _B, _C)
		-- HScalarProductFromPoints : Calcule le produit scalaire des vecteurs AB et AC, sur leurs coordonnées x,z
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "getHScalarProductFromPoints(A, B, C)"
		
		if ATME.isPoint(_A) ~= POINT_BAD and ATME.isPoint(_B) ~= POINT_BAD and ATME.isPoint(_C) ~= POINT_BAD then
			local vectAB = ATMEFct.C_Vector3D_create(nil, nil, _A, _B)
			local vectAC = ATMEFct.C_Vector3D_create(nil, nil, _A, _C)
			
			return vectAB:getX() * vectAC:getX() + vectAB:getZ() * vectAC:getZ()
		else
			if ATME.isPoint(_A) == POINT_BAD then
				ATMEFct.errorOutputCore(nil, fctDefinition, "A type must be table (x,y)/(x,y,z)", false)
			elseif ATME.isPoint(_B) == POINT_BAD then
				ATMEFct.errorOutputCore(nil, fctDefinition, "B type must be table (x,y)/(x,y,z)", false)
			elseif ATME.isPoint(_C) == POINT_BAD then
				ATMEFct.errorOutputCore(nil, fctDefinition, "C type must be table (x,y)/(x,y,z)", false)
			end
		end
	end,
	
	whichSide = function(_referenceA, _referenceB, _referenceC)
		-- whichSide : Vérifie si un point C est à droite ou à gauche d'un axe orienté AB :
		--               - 0 si le point est sur la droite
		--				 - 1 si le point est à droite
		--               - -1 si le point est à gauche
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "whichSide(referenceA, referenceB, referenceC)"
		
		if ATMEFct.isReference(_referenceA) == true and ATMEFct.isReference(_referenceB) == true and ATMEFct.isReference(_referenceC) == true then
			return ATMEFct.whichSide(ATMEFct.getPointFromReference(_referenceA), ATMEFct.getPointFromReference(_referenceB), ATMEFct.getPointFromReference(_referenceC))
		else
			if ATMEFct.isReference(_referenceA) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "referenceA type must be" .. ATMEConst.errorLabelReference, false)
			elseif ATMEFct.isReference(_referenceB) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "referenceB type must be" .. ATMEConst.errorLabelReference, false)
			elseif ATMEFct.isReference(_referenceC) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "referenceC type must be".. ATMEConst.errorLabelReference, false)
			end
		end
	end,
	
	coordLocalToDCS = function(_point, _originReference, _azimuth)
		-- coordLocalToDCS : Transforme les coordonnées locales d'un point relatives au repère défini par sa référence d'origine et son azimuth
		-- L'azimuth est en degrés. A noter : le repère est inversé x en vertical et repère xOy indirect (norme DCS)
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "coordLocalToDCS(point, originReference, azimuth)"

		if ATME.isPoint(_point) ~= POINT_BAD and ATMEFct.isReference(_originReference) == true and type(_azimuth) == "number" then
			return ATMEFct.coordLocalToDCS(_point, _originReference, _azimuth)
		else
			if ATMEFct.isPoint(_point) == POINT_BAD then
				ATMEFct.errorOutputCore(nil, fctDefinition, "point type must be table (x,y)/(x,y,z)", false)
			elseif ATMEFct.isReference(_originReference) == false  then
				ATMEFct.errorOutputCore(nil, fctDefinition, "centerReference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_azimuth) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
			end
		end
	end,
	
	coordDCSToLocal = function(_reference, _originReference, _azimuth)		
		-- localCoord : Transforme les coordonnées DCS d'une référence en coordonnées locales relatives au repère défini par sa référence 
		-- d'origine en coordonnées DCS et son azimuth
		-- L'azimuth est en degrés. A noter : le repère est inversé x en vertical et repère xOy indirect (norme DCS)
		
		local fctDefinition = ATMEFct.getErrorPrefix() .. "coordDCSToLocal(reference, originReference, azimuth)"
		
		if ATMEFct.isReference(_reference) == true and ATMEFct.isReference(_originReference) == true and type(_azimuth) == "number" then
			local angle = math.rad(_azimuth)
			local point = ATMEFct.getPointFromReference(_reference)
			local origin = ATMEFct.getPointFromReference(_originReference)
			
			return {
				x = (point.x - origin.x) * math.cos(angle) + (point.z - origin.z) * math.sin(angle),
				y = point.y,
				z = -(point.x - origin.x) * math.sin(angle) + (point.z - origin.z) * math.cos(angle),
			}
		else
			if ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif ATMEFct.isReference(_originReference) == false  then
				ATMEFct.errorOutputCore(nil, fctDefinition, "centerReference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_azimuth) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
			end
		end
	end,
	effectBigSmoke = function(_reference, _typeSmoke, _density)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "effectBigSmoke(reference, typeSmoke, density)"
		
		if ATMEFct.isReference(_reference) == true and type(_typeSmoke) == "string" and type(_density) == "number" then
			if _density > 0 then
				local tokens = ATMEFct.decompString(_typeSmoke)
				local value = (1 - #tokens) * 4
				
				if #tokens >= 1 and #tokens <= 2 then
					if #tokens == 1 or tokens[2] == "FIRE" then
						if ATMEConst.bigSmokeType[tokens[1]] ~= nil then
							local point = ATMEFct.getPointFromReference(_reference)
							trigger.action.effectSmokeBig(point, ATMEConst.bigSmokeType[tokens[1]] + value, _density)
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "typeSmoke string bad format", false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "typeSmoke string bad format", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "typeSmoke string bad format", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "density must be > 0", false)
			end
		else
			if ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_typeSmoke) ~= "string"  then
				ATMEFct.errorOutputCore(nil, fctDefinition, "typeSmoke type must be string", false)
			elseif type(_azimuth) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "density type must be number", false)
			end
		end
	end,
}

setmetatable(ATME, ATMEDatas.ATME_MT.mt)

ATMEDatas.ATME_MT.mt.__metatable = "Private"
ATMEDatas.ATME_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.ATME_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "ATME : Bad key " .. k, "table locked", false)
	end
end


ATME.modules = {}  -- Déclaration de ATME.modules pour les variables globales des modules

setmetatable(ATME.modules, ATMEDatas.ATME_MT.mtModules)
ATMEDatas.ATME_MT.mtModules.__metatable = "Private"
ATMEDatas.ATME_MT.mtModules.__newindex = function(t, k, v)
	if ATMEDatas.modules[k] ~= nil then
		rawset(t, k, v)
	else  
		ATMEFct.errorOutputCore(nil,"ATME : ATME.modules Bad key " .. k, "table locked", false)
	end
end



----------------------------------------------------------------------------------------------------------
-- DEFINITION DES CLASSES D'OBJETS DEFINIS DANS ATME
--
-- Quelques points essentiels :
-- 			- Une classe d'objet est une table lua particulière, préfixée de "C_"
--			- Il existe une notion de classes virtuelles qui sont des regroupements de classes
--			- Une instance d'objet est un objet créé à partir d'une classe, c'est donc aussi une table qui peut
--			  être cependant étendue par le création de méthodes dans le constructeur local C_Class_create.
--			- Les fonctions liées à une classe et non à une instance sont définies dans la classe.
--			- Une classe d'objet a également une classe spéciale suffixée par _MT et locale à ATME_Core. Cette
--            classe définit : 
--				-> la metatable des instances (entrée "mt")
--				-> l'entrée "keys" des instances pour en interdire la création, lors de la création de l'instance. En effet,
--                 le constructeur local C_Class_create crée l'instance et lui associe des méthodes spécifique pour  
--				   les variables locales (closure). Ce sont ces méthodes, non gérées par la metaclasse qui sont listées
--				   dans l'entrée "keys" pour le contrôle des entrées de la table représentant l'instance.
--				-> le nom de la classe entrée "className"
--				-> la création du constructeur "public" de la classe (__call). Par défaut seule la fonction locale
--                 C_Class_create est créée ce qui interdit toute création d'instance dans les modules externes. C'est
--				   le cas pour les classes C_Group (Groupes au format ATME), C_Player (unités joueurs au format ATME),
--                 C_AIUnit (Unités AI au format ATME) et C_StaticObject (objets statiques au format ATME sauf FARP et
--				   Airfields.
--				-> Certaines propriétés dont l'accès ne doit être que local et dont l'accès à la modification pourrait
--  			   engendrer des incohérences. Ces propriétés sont regroupées dans l'entrée "localMembers" et accessibles
--				   par les fonctions appropriées à partir du module ATME_Core uniquement (en lecture comme en écriture).
--				   L'instance contient un index spécifique accessible en lecture seule y compris à partir de modules
--				   externes, mais pour ces derniers cet index ne sert à rien.
----------------------------------------------------------------------------------------------------------


----------------------------------------------------------------------------------------------------------
-- DEFINITION DES CLASSES VIRTUELLES
----------------------------------------------------------------------------------------------------------

----------------------------------------------------------------------------------------------------------
--	ATME.Zone : 
--  Classe virtuelle pour
-- 		                   ATME.C_Zone2D
--		                   ATME.C_Zone3D
----------------------------------------------------------------------------------------------------------

ATMEFct.C_Zone_createNamed = function(_namedTable, _name, _fct, ...)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return false
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			local args = {...}
			return false, _fct(_name, _namedTable, unpack(args))
		else
			return true, "name can't contain '|' character"
		end
	end
end

ATMEDatas.C_Zone_MT = {
	className = "C_Zone",
	
	mt = {},
}

ATME.C_Zone = {
	getClassName = function(self)
		return ATMEDatas.C_Zone_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Zone_MT.parentClass
	end,
	
	getByName = function(_name)	
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_PickupZone.getByName(name)", "name type must be string", false)
		end
	end,
}

setmetatable(ATME.C_Zone, ATMEDatas.C_Zone_MT)

ATMEDatas.C_Zone_MT.mt.__metatable = "Private"
ATMEDatas.C_Zone_MT.mt.__index = ATME.C_Zone

ATMEDatas.C_Zone_MT.mt.__newindex = function(t, k, v)
	ATMEFct.errorOutputCore(nil, "Function ATME.C_Zone : Bad key " .. k, "table locked", false)
end

----------------------------------------------------------------------------------------------------------
--	ATME.Zone2D : 
--  Classe virtuelle pour
-- 		                   ATME.C_Polygon
--		                   ATME.C_MovableZone2D
--						   ATME.C_PickupZone
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Zone2D_MT = {
	className = "C_Zone2D",
	parentClass = ATME.C_Zone,
	
	mt = {},
}

ATME.C_Zone2D = {
	getClassName = function(self)
		return ATMEDatas.C_Zone2D_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Zone2D_MT.parentClass
	end,
	
	getByName = function(_name)	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_Zone2D) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_PickupZone.getByName(name)", "name type must be string", false)
		end
	end,
}

setmetatable(ATME.C_Zone2D, ATMEDatas.C_Zone2D_MT)

ATMEDatas.C_Zone2D_MT.mt.__metatable = "Private"
ATMEDatas.C_Zone2D_MT.mt.__index = ATME.C_Zone2D

ATMEDatas.C_Zone2D_MT.mt.__newindex = function(t, k, v)
	ATMEFct.errorOutputCore(nil, "Function ATME.C_Zone2D : Bad key " .. k, "table locked", false)
end

----------------------------------------------------------------------------------------------------------
--	ATME.Zone3D : 
--  Classe virtuelle pour
-- 		                   ATME.C_MovableZone3D
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Zone3D_MT = {
	className = "C_Zone3D",
	parentClass = ATME.C_Zone,
	
	mt = {},
}

ATME.C_Zone3D = {
	getClassName = function(self)
		return ATMEDatas.C_Zone3D_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Zone3D_MT.parentClass
	end,
	
	getByName = function(_name)	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_Zone3D) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_PickupZone.getByName(name)", "name type must be string", false)
		end
	end,
}

setmetatable(ATME.C_Zone3D, ATMEDatas.C_Zone3D_MT)

ATMEDatas.C_Zone3D_MT.mt.__metatable = "Private"
ATMEDatas.C_Zone3D_MT.mt.__index = ATME.C_Zone3D

ATMEDatas.C_Zone3D_MT.mt.__newindex = function(t, k, v)
	ATMEFct.errorOutputCore(nil, "Function ATME.C_Zone3D : Bad key " .. k, "table locked", false)
end

----------------------------------------------------------------------------------------------------------
--	ATME.MovableZone2D : Définit les zones 2D pouvant être rattachées à une unité si associée à un zonesMgr
--  Classe virtuelle pour
--		                   ATME.C_Circle
--						   ATME.C_Rectangle
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_MovableZone2D_MT = {
	className = "C_MovableZone2D",
	parentClass = ATME.C_Zone2D,
	
	mt = {},
}

ATME.C_MovableZone2D = {
	getClassName = function(self)
		return ATMEDatas.C_MovableZone2D_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_MovableZone2D_MT.parentClass
	end,
	
	getByName = function(_name)	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_MovableZone2D) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_PickupZone.getByName(name)", "name type must be string", false)
		end
	end,
}

setmetatable(ATME.C_MovableZone2D, ATMEDatas.C_MovableZone2D_MT)

ATMEDatas.C_MovableZone2D_MT.mt.__metatable = "Private"
ATMEDatas.C_MovableZone2D_MT.mt.__index = ATME.C_MovableZone2D

ATMEDatas.C_MovableZone2D_MT.mt.__newindex = function(t, k, v)
	ATMEFct.errorOutputCore(nil, "Function ATME.C_MovableZone2D : Bad key " .. k, "table locked", false)
end

----------------------------------------------------------------------------------------------------------
--	ATME.MovableZone3D : Définit les zones 3D pouvant être rattachées à une unité si associée à un zonesMgr
--  Classe virtuelle pour
--		                   ATME.C_Sphere
--						   ATME.C_Cylinder
--                         ATME.C_Box
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_MovableZone3D_MT = {
	className = "C_MovableZone3D",
	parentClass = ATME.C_Zone3D,
	
	mt = {},
}

ATME.C_MovableZone3D = {
	getClassName = function(self)
		return ATMEDatas.C_MovableZone3D_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_MovableZone3D_MT.parentClass
	end,
	
	getByName = function(_name)	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_MovableZone3D) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_PickupZone.getByName(name)", "name type must be string", false)
		end
	end,
}

setmetatable(ATME.C_MovableZone3D, ATMEDatas.C_MovableZone3D_MT)

ATMEDatas.C_MovableZone3D_MT.mt.__metatable = "Private"
ATMEDatas.C_MovableZone3D_MT.mt.__index = ATME.C_MovableZone3D

ATMEDatas.C_MovableZone3D_MT.mt.__newindex = function(t, k, v)
	ATMEFct.errorOutputCore(nil, "Function ATME.C_MovableZone3D : Bad key " .. k, "table locked", false)
end

----------------------------------------------------------------------------------------------------------
--	ATME.C_DynamicObject : 
--  Classe virtuelle pour
-- 		                   ATME.C_Unit
--		                   ATME.C_Group
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_DynamicObject_MT = {
	className = "C_DynamicObject",
	
	mt = {},
}

ATME.C_DynamicObject = {
	getClassName = function(self)
		return ATMEDatas.C_DynamicObject_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_DynamicObject_MT.parentClass
	end,
}

setmetatable(ATME.C_DynamicObject, ATMEDatas.C_DynamicObject_MT)

ATMEDatas.C_DynamicObject_MT.mt.__metatable = "Private"
ATMEDatas.C_DynamicObject_MT.mt.__index = ATME.C_DynamicObject

ATMEDatas.C_DynamicObject_MT.mt.__newindex = function(t, k, v)
	ATMEFct.errorOutputCore(nil, "Function ATME.C_DynamicObject : Bad key " .. k, "table locked", false)
end

----------------------------------------------------------------------------------------------------------
--	ATME.C_Unit : 
--  Classe virtuelle pour
-- 		                   ATME.C_AIUnit
--		                   ATME.C_Player
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Unit_MT = {
	className = "C_Unit",
	parentClass = ATME.C_DynamicObject,
	
	mt = {},
}

ATME.C_Unit = {
	getClassName = function(self)
		return ATMEDatas.C_Unit_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Unit_MT.parentClass
	end,
	
	getByName = function(_name)
		if type(_name) == "string" then
			return ATMEDatas.allAIUnitsAndPlayers[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Unit.getByName(name)", "name type must be string", false)
		end
	end,
}

setmetatable(ATME.C_Unit, ATMEDatas.C_Unit_MT)

ATMEDatas.C_Unit_MT.mt.__metatable = "Private"
ATMEDatas.C_Unit_MT.mt.__index = ATME.C_Unit

ATMEDatas.C_Unit_MT.mt.__newindex = function(t, k, v)
	ATMEFct.errorOutputCore(nil, "Function ATME.C_Unit : Bad key " .. k, "table locked", false)
end
----------------------------------------------------------------------------------------------------------
-- DEFINITION DES CLASSES STANDARD
----------------------------------------------------------------------------------------------------------

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_IndexList : gère une liste indexée par des entiers. Constructeur de classe accessible aux modules 
-- externes. 
-- Elle optimise la place utilisée sans recourir aux fonctions table.insert etc...
-- Le premier indice commence à 1. En cas de suppression, l'entrée indexée est mise à nil et sera réutilisée
-- au prochain ajout dans la liste. Si toutes les entrées sont utilisées, la liste est étendue d'un élément.
-- De plus en cas de suppression, l'index des suivants ne change pas.
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_IndexList_MT = {
	className = "C_IndexList",
	mt = {},
	
	keys = {
		["add"] = true,
		["remove"] = true,
		["get"] = true,
		["getFirst"] = true,
		["getNext"] = true,
		["getPrevious"] = true,
		["getRandom"] = true,
		["getLast"] = true,
		["pairs"] = true,
		["getCount"] = true,
		["removeAll"] = true,
	},
}

ATME.C_IndexList = {
	getClassName = function(self)
		return ATMEDatas.C_IndexList_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_IndexList_MT.parentClass
	end,
}

ATMEFct.C_IndexList_create = function()
	local maxIndex = 1
	local nbItems = 0
	local idsRef = {}
	local list = {}
	local newList = {}
	local firstValidId
	local lastValidId
	local scanUp, scanDown
		
	setmetatable(newList, ATMEDatas.C_IndexList_MT.mt)
	
	function newList.add(self, _item)
		local fctDefinition = ATMEFct.getErrorPrefix(newList) .. "add(item)"
		
		if _item ~= nil then
			local index
			local previousId
			
			index = maxIndex
			
			if nbItems == 0 then
				firstValidId = index
			else
				previousId = index - 1
				list[previousId].next = index
			end
			
			lastValidId = index
			maxIndex = maxIndex + 1
		
			list[index] = {
				item = _item,
				previous = previousId,
			}
			
			nbItems = nbItems + 1
			
			return index
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "item type can't be nil", true)
		end
	end
	
	function newList.remove(self, _id)
		local fctDefinition = ATMEFct.getErrorPrefix(newList) .. "remove([id])"
		if _id == nil or type(_id) == "number" then
			local id = firstValidId
			
			if _id ~= nil then
				id = math.floor(_id)
			end
			
			if id > 0 then
				if list[id] ~= nil then
					if nbItems == 1 then
						firstValidId = nil
						lastValidId = nil
						maxIndex = 1
					else
						local previousId = list[id].previous
						local nextId = list[id].next
						
						if id == scanDown then
							scanDown = list[id].previous
						elseif id == scanUp then
							scanUp = list[id].next
						end
						
						if  previousId ~= nil then
							list[previousId].next = nextId
						end
						if nextId ~= nil then
							list[nextId].previous = previousId
						end
						
						if id == firstValidId then
							firstValidId = nextId
						elseif id == lastValidId then
							lastValidId = previousId
						end
					end
					
					nbItems = nbItems - 1
					
					local item = list[id].item
					list[id] = nil
					
					return item
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "id must be >0", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "id type must be nil or number", true)
		end
	end
	
	function newList.get(self,_id)
		local fctDefinition = ATMEFct.getErrorPrefix(newList) .. "get(id)"
		if type(_id) == "number" then
			local id = math.floor(_id)
			if id > 0 then
				if list[id] ~= nil then
					scanDown = list[id].previous
					scanUp = list[id].next
			
					return list[id].item, id
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "id must be >0", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "id type must be number", true)
		end
	end
	
	function newList.getFirst(self)
		if firstValidId ~= nil then
			scanDown = nil
			scanUp = list[firstValidId].next
			return list[firstValidId].item, firstValidId
		end
	end
	
	function newList.getLast(self)
		if lastValidId ~= nil then
			scanDown = list[lastValidId].previous
			scanUp = nil
			return list[lastValidId].item, lastValidId
		end
	end
	
	function newList.getNext(self)
		if scanUp ~= nil and list[scanUp] ~= nil then
			local id = scanUp
			scanDown = list[id].previous
			scanUp = list[id].next
			
			return list[id].item, id
		else
			scanUp = nil
		end
	end
	
	function newList.getPrevious(self)
		if scanDown ~= nil and list[scanDown] ~= nil then
			local id = scanDown
			scanDown = list[id].previous
			scanUp = list[id].next
			
			return list[id].item, id
		else
			scanOn = nil
		end
	end
	
	function newList.getCount(self)
		return nbItems
	end
	
	function newList.getRandom(self)
		if nbItems > 0 then
			local rand = math.random(1, nbItems) - 1
			local id = firstValidId
			local scanUpRandom = list[id].next
			
			while rand > 0 do
				id = scanUpRandom
				scanUpRandom = list[id].next
				rand = rand - 1
			end
				
			return list[id].item, id
		end
	end
	
    function newList.pairs(self)
		local i = firstValidId
		return function()
			
			if i ~= nil then
				local index = i
				i = list[index].next

				return index, list[index].item
			end
		end
    end
	
	function newList.removeAll(self)
		list = {}
		nbItems = 0
		maxIndex = 1
		firstValidId = nil
		lastValidId = nil
	end
	
	return newList
end

ATMEDatas.C_IndexList_MT.__call = function(t)
	return ATMEFct.C_IndexList_create()
end

setmetatable(ATME.C_IndexList,ATMEDatas.C_IndexList_MT)

ATMEDatas.C_IndexList_MT.mt.__metatable = "Private"
ATMEDatas.C_IndexList_MT.mt.__index = ATME.C_IndexList
ATMEDatas.C_IndexList_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_IndexList_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else  
		ATMEFct.errorOutputCore(nil, "Function ATME.C_IndexList : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Stack : gère une pile. Constructeur de classe accessible aux modules 
-- externes. 
-- Elle permet une gestion des piles FIFO et LIFO
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Stack_MT = {
	className = "C_Stack",
	mt = {},
	
	keys = {
		["push"] = true,
		["extract"] = true,
		["pairs"] = true,
		["getCount"] = true,
		["removeAll"] = true,
		["getIndex"] = true,
	},
	
	add = function(self)
		local item = {
			nbItems= 0,
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
	end,
}

ATME.C_Stack = {
	getClassName = function(self)
		return ATMEDatas.C_Stack_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Stack_MT.parentClass
	end,
}

ATMEFct.C_Stack_extract = function(_localMembers, _typeExtract)
	if _localMembers[ATMEConst.stackTypeExtract[_typeExtract].load] ~= nil then
		local item = _localMembers[ATMEConst.stackTypeExtract[_typeExtract].load].item
		if _localMembers.nbItems ~= 1 then
			_localMembers[ATMEConst.stackTypeExtract[_typeExtract].load] = _localMembers[ATMEConst.stackTypeExtract[_typeExtract].load][ATMEConst.stackTypeExtract[_typeExtract].direction]
			_localMembers[ATMEConst.stackTypeExtract[_typeExtract].load][ATMEConst.stackTypeExtract[_typeExtract].raz] = nil
		else
			_localMembers.first = nil
			_localMembers.last = nil
		end
		
		_localMembers.nbItems = _localMembers.nbItems - 1
		
		return item
	end
end

ATMEFct.C_Stack_create = function()
	local newStack = {}

	setmetatable(newStack, ATMEDatas.C_Stack_MT.mt)
	local index = ATMEDatas.C_Stack_MT:add()
	local localMembers = ATMEDatas.localMembers[index]

	function newStack.getIndex(self)
		return index
	end
	
	function newStack.push(self, _item)
		local data = {
			item = _item,
		}

		if localMembers.first == nil then
			localMembers.first = data
			localMembers.last = data
		else
			localMembers.first.previous = data
			data.next = localMembers.first
			localMembers.first = data
		end

		localMembers.nbItems = localMembers.nbItems + 1
	end
	
	function newStack.extract(self, _typeExtract)
		local fctDefinition = ATMEFct.getErrorPrefix(newStack) .. "extract([type])"
		
		if _typeExtract == nil or type(_typeExtract) == "string" then
			local typeExtract
			if _typeExtract == nil then
				typeExtract = "TAIL"
			else
				typeExtract = string.upper(_typeExtract)
			end
			
			if ATMEConst.stackTypeExtract[typeExtract] ~= nil then		
				return ATMEFct.C_Stack_extract(localMembers, typeExtract)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "type must be 'HEAD' or 'TAIL'", false)
			end
		else
			if _typeExtract ~= nil and type(_typeExtract) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "type type must be nil or string", true)
			end
		end
	end
	
	function newStack.getCount(self)
		return localMembers.nbItems
	end
	
    function newStack.pairs(self, _typeRead)
		local data
		local i = 0
		
		local typeRead
		if _typeRead == nil or type(_typeRead) ~= "string" then
			typeRead = "TAIL"
		else
			typeRead = _typeRead:upper()
			if ATMEConst.stackTypeExtract[typeRead] == nil then
				typeRead = "TAIL"
			end
		end
		
		return function()
			if data == nil then
				data = localMembers[ATMEConst.stackTypeExtract[typeRead].load]
			else
				data = data[ATMEConst.stackTypeExtract[typeRead].direction]
			end
			
			i = i + 1
			
			if data ~= nil then
				return i, data.item
			end
		end
    end
	
	function newStack.removeAll(self)
		localMembers.first = nil
		localMembers.last = nil
		localMembers.nbItems = 0
	end

	return newStack
end

ATMEDatas.C_Stack_MT.__call = function(t)
	return ATMEFct.C_Stack_create()
end

setmetatable(ATME.C_Stack,ATMEDatas.C_Stack_MT)

ATMEDatas.C_Stack_MT.mt.__metatable = "Private"
ATMEDatas.C_Stack_MT.mt.__index = ATME.C_Stack
ATMEDatas.C_Stack_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Stack_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else  
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Stack : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Variables globales utilisant les C_Index_List & C_Stack
----------------------------------------------------------------------------------------------------------
ATMEDatas.signalCapabilityUnits = ATMEFct.C_IndexList_create()

----------------------------------------------------------------------------------------------------------
-- Fonctions générales
----------------------------------------------------------------------------------------------------------		

ATMEFct.searchNearestWaypoint = function(_unit, _datas)
	local localMembersGroup = ATMEDatas.localMembers[_datas.group:getIndex()]

	if localMembersGroup.ATMEUnits[_unit:getName()] ~= nil then
		_datas.ok = true
	end
end

-- Fonction de disable pour les AIUnits et players
ATMEFct.C_Player_C_AIUnit_disable = function(_unit)
	local unitDCS = _unit:getDCSUnit()
	local localMembers = ATMEDatas.localMembers[_unit:getIndex()]
	
	localMembers.isEnabled = false
	
	unitDCS:destroy()
	ATMEFct.C_Airbase_delete(unitDCS)
	ATMEFct[_unit:getClassName() .. "_delete"](_unit)
end			

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_RadioChannel : gère les fréquences radios pour les transmissions.
----------------------------------------------------------------------------------------------------------

ATMEFct.C_RadioChannel_getIdFrequency = function(_modulation, _frequency)
	return "FR_" .. string.format("%09d", _frequency) .. _modulation
end

ATMETimers.C_RadioChannel_transmissionTimer = function(_localMembersRadioChannel)
	if ATMEDatas.exitATME == false then
		local currentTransmission = _localMembersRadioChannel.current
		local toRemove = false

		if currentTransmission.randomRead == true then 
			currentTransmission.currentReadIndex = math.random(1, #currentTransmission.playlist)
		end
		
		if currentTransmission.unitName == nil then
			ATMEFct.C_RadioChannel_stopTransmission(currentTransmission)
			
			if _localMembersRadioChannel.stopCurrent == nil or (currentTransmission.currentReadIndex == 1 and currentTransmission.readOnce == true) then
				ATMEFct.C_RadioChannel_startTransmission(currentTransmission)
				
				currentTransmission.tipsBase = currentTransmission.playlist[currentTransmission.currentReadIndex].tipsBase
				
				if currentTransmission.tipsBase < 1.5 then
					currentTransmission.tipsBase = 1.5
				end

				if currentTransmission.autoRestartDCS == true then
					_localMembersRadioChannel.autoRestartDCS = true
				end
			else
				toRemove = true
			end
		else
			if ATMEDatas.AIUnits[currentTransmission.unitName] ~= nil then
				ATMEFct.C_AIUnit_stopTransmission(currentTransmission)
				if _localMembersRadioChannel.stopCurrent == nil or (currentTransmission.currentReadIndex == 1 and currentTransmission.readOnce == true) then 
					ATMEFct.C_AIUnit_startTransmission(currentTransmission)
					currentTransmission.tipsBase = currentTransmission.playlist[currentTransmission.currentReadIndex].tipsBase
					
					if currentTransmission.autoRestartDCS == true then
						_localMembersRadioChannel.autoRestartDCS = true
					end
				else
					toRemove = true
				end
			else
				toRemove = true
			end
		end
		
		_localMembersRadioChannel.stopCurrent = nil
		
		if currentTransmission.readOnce == true then
			currentTransmission.currentReadIndex = currentTransmission.currentReadIndex + 1
			if currentTransmission.currentReadIndex > #currentTransmission.playlist then
				toRemove = true
			end
		else
			if currentTransmission.randomRead == false then
				currentTransmission.currentReadIndex = currentTransmission.currentReadIndex + 1
				if currentTransmission.currentReadIndex > #currentTransmission.playlist then
					currentTransmission.currentReadIndex = 1
				end
			end
		end
		
		if toRemove == true then
			if currentTransmission.autoRestartDCS == false then
				if _localMembersRadioChannel.serialized == true then
					local stackLocalMembers = ATMEDatas.localMembers[_localMembersRadioChannel.stack:getIndex()]
					_localMembersRadioChannel.current = ATMEFct.C_Stack_extract(stackLocalMembers, "TAIL")
					
					if _localMembersRadioChannel.current ~= nil then 
						return timer.getTime() + math.random(1,5)
					end
				end
			end
			
			_localMembersRadioChannel.timerId = nil
		else
			-- On réarme pour la durée du son
			return timer.getTime() + currentTransmission.tipsBase
		end
	end
end

ATMEFct.C_RadioChannel_initTransmission = function(_extraDatas)
	-- C_RadioChannel_initTransmission : fonction d'initialisation des paramètres pour les transmissions radio
	-- 						             pour une mise en forme au format DCS

	local newRadioTransmission = {
		playlist = _extraDatas.playlist,
		unitName = _extraDatas.unitName,
		autoRestartDCS = _extraDatas.autoRestartDCS,
		tipsBase = 0,
		readOnce = _extraDatas.readOnce,
		randomRead = _extraDatas.randomRead,
		replay = true,
		currentReadIndex = 1,
		started = false,
		voice = _extraDatas.voice,
		localMembersRadioChannel = _extraDatas.localMembersRadioChannel,
	}
	
	if _extraDatas.unitName == nil then 			-- Pour la transmission à partir d'un point
		newRadioTransmission.point = _extraDatas.point
		newRadioTransmission.power = _extraDatas.power
		newRadioTransmission.textId = "Z" .. ATMEDatas.idDCSRadioTransmissions
		ATMEDatas.idDCSRadioTransmissions = ATMEDatas.idDCSRadioTransmissions + 1
	end
	
	if _extraDatas.localMembersRadioChannel.timerId == nil then
		if _extraDatas.localMembersRadioChannel.autoRestartDCS == false then
			_extraDatas.localMembersRadioChannel.current = newRadioTransmission
			_extraDatas.localMembersRadioChannel.timerId = timer.scheduleFunction(ATMETimers.C_RadioChannel_transmissionTimer, _extraDatas.localMembersRadioChannel, timer.getTime() + 1)
		else
			if _extraDatas.localMembersRadioChannel.serialized == true then
				_extraDatas.localMembersRadioChannel.stack:push(newRadioTransmission)
			end
		end
	else
		if _extraDatas.localMembersRadioChannel.serialized == true then
			_extraDatas.localMembersRadioChannel.stack:push(newRadioTransmission)
		else
			_extraDatas.localMembersRadioChannel.current = newRadioTransmission
		end
	end
	
	return newRadioTransmission
end

ATMEFct.C_RadioChannel_startTransmission = function(_radioTransmission)
	-- Pour "trigger.action.radioTransmission", si 4è paramètre à true, la boucle est gérée par DCS directement.

	if _radioTransmission.started == false and _radioTransmission.playlist[_radioTransmission.currentReadIndex].pause == false then
		trigger.action.radioTransmission("l10n/DEFAULT/" .. _radioTransmission.voice .. _radioTransmission.playlist[_radioTransmission.currentReadIndex].fileName,
											_radioTransmission.point , ATMEConst.modulationValues[_radioTransmission.localMembersRadioChannel.modulation],
											_radioTransmission.autoRestartDCS, _radioTransmission.localMembersRadioChannel.frequency,
											_radioTransmission.power, _radioTransmission.textId)
		_radioTransmission.started = true
	end
end

ATMEFct.C_RadioChannel_stopTransmission = function(_radioTransmission)
	-- Pour "trigger.action.radioTransmission", si 4è paramètre à true, la boucle est gérée par DCS directement.
	
	if _radioTransmission.started == true then
		trigger.action.stopRadioTransmission(_radioTransmission.textId)
		_radioTransmission.started = false
	end
end

ATMEFct.C_RadioChannel_createNamed = function(_namedTable, _name, _modulation, _frequency, _serialized)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return false
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			return ATMEFct.C_RadioChannel_create(_name, _namedTable, _modulation, _frequency, _serialized)
		else
			return true, "name can't contain '|' character"
		end
	end
end

ATMEDatas.C_RadioChannel_MT = {
	className = "C_RadioChannel",
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["setVoice"] = true,
		["getVoice"] = true,
		["sendFromUnit"] = true,
		["sendFromReference"] = true,
		["loopFromUnit"] = true,
		["loopFromReference"] = true,
		["stop"] = true,
		["stopAll"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
	},
	
	add = function(self, _modulation, _frequency, _serialized)
		local item = {
			frequency = _frequency,
			modulation = _modulation,
			idFrequency = ATMEFct.C_RadioChannel_getIdFrequency(_modulation, _frequency),
			userPrivateDatas = {},
			currentVoice = "",
			serialized = _serialized,
			stack = ATMEFct.C_Stack_create(),
			autoRestartDCS = false,
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_RadioChannel = {
	getClassName = function(self)
		return ATMEDatas.C_RadioChannel_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_RadioChannel_MT.parentClass
	end,
	
	createNamed = function(_name, _modulation, _frequency, _serialized)
		local fctDefinition = "Function ATME.C_RadioChannel.createNamed(name, modulation, frequency, [serialized])"
		
		if type(_name) == "string" and type(_modulation) == "string" and type(_frequency) == "number" and (_serialized == nil or type(_serialized) == "boolean") then
			if ATMEConst.modulationValues[_modulation] ~= nil then
				local frequencyValue = math.floor(_frequency)
				if frequencyValue > 0 then
					local serialized = true
					
					if _serialized ~= nil then
						serialized = _serialized
					end
					local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
					local err, newRadioChannel = ATMEFct.C_RadioChannel_createNamed(localMembersModule.radioChannels, _name, _modulation, frequencyValue, serialized)

					if err == true then
						ATMEFct.errorOutputCore(nil, fctDefinition, newRadioChannel, false)
						return
					end

					return newRadioChannel
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "frequency must be > 0", false)
				end	
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "modulation must be 'AM' or 'FM'", false)
			end
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif type(_modulation) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "modulation type must be string", false)
			elseif type(_frequency) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "frequency type must be number", false)
			elseif _serialized ~= nil and type(_serialized) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "serialized type must be nil or boolean", false)
			end
		end
	end,
	
	getByName = function(_name)
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].radioChannels[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_RadioChannel.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_RadioChannel_create = function(_name, _radioChannels, _modulation, _frequency, _serialized)
	local newRadioChannel = {}
	local name
	local tagsName

	setmetatable(newRadioChannel, ATMEDatas.C_RadioChannel_MT.mt)
	
	local index = ATMEDatas.C_RadioChannel_MT:add(_modulation, _frequency, _serialized)
	local localMembers = ATMEDatas.localMembers[index]
	
	if ATMEDatas.radioFrequencies[localMembers.idFrequency]~= nil then
		return true, "Radio channel '" .. _modulation .. "' " .. _frequency/1000000.0 .. "MHz already exists"
	end
	
	if _name ~= nil and _radioChannels ~= nil then
		name = _name
		_radioChannels[name] = newRadioChannel
		tagsName = ATMEFct.decompString(_name)
	end
	
	ATMEDatas.radioFrequencies[localMembers.idFrequency] = true
	
	function newRadioChannel.getIndex(self)
		return index
	end
	
	function newRadioChannel.getName(self)
		return name
	end
	
	function newRadioChannel.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newRadioChannel) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newRadioChannel.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newRadioChannel.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newRadioChannel) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newRadioChannel.setVoice(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newRadioChannel) .. "setVoice([name])"
		
		if _name == nil or type(_name) == "string" then
			if _name ~= nil then
				local d, f, temp = _name:find("^(%a[%a%d_]*)")
				
				if temp == _name then
					localMembers.currentVoice = _name
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "name bad format", false)
				end
			else
				localMembers.currentVoice = ""
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be nil or string", true)
		end
	end
	
	function newRadioChannel.getVoice(self)
		if localMembers.currentVoice ~= "" then
			return localMembers.currentVoice
		end
	end
	
	function newRadioChannel.sendFromUnit(self, _text, _unit, _readMode)
		local fctDefinition = ATMEFct.getErrorPrefix(newRadioChannel) .. "sendFromUnit(text, unit, [readMode])"
		
		if type(_text) == "string" and ATMEFct.isObjectClass(_unit, ATME.C_AIUnit) == true and 
				(_readMode == nil or type(_readMode) == "boolean") then
			if _unit:getGroup():getCategoryName() ~= "SHIP" then -- Bug DCS avec les navires -> Plantage
				local randomRead = false
				local readOnce = true
				
				if type(_readMode) == "boolean" then
					randomRead = _readMode
					readOnce = false
				end
				
				local err, playlistNew = ATMEFct.C_Sound_getPlaylist(_text, randomRead)
				
				if err == true then
					ATMEFct.errorOutputCore(nil, fctDefinition, playlistNew, false)
					return
				end

				if playlistNew == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "text string has no sound token", false)
					return
				end
				
				local localMembersUnit = ATMEDatas.localMembers[_unit:getIndex()]
				
				ATMEFct.C_RadioChannel_initTransmission({
					unitName = _unit:getName(),
					playlist = playlistNew,
					autoRestartDCS = false,
					readOnce = readOnce,
					randomRead = randomRead,
					voice = localMembersUnit.currentVoice,
					localMembersRadioChannel = localMembers,
				})
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "unit can't be a ship", false)
			end
		else
			if type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", true)
			elseif ATMEFct.isObjectClass(_unit, ATME.C_AIUnit) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "unit type must be ATME.C_AIUnit", false)
			elseif _readMode ~= nil and type(_readMode) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "readMode type must be nil or boolean", false)
			end
		end
	end
	
	function newRadioChannel.sendFromReference(self, _text, _reference, _power, _readMode)
		local fctDefinition = ATMEFct.getErrorPrefix(newRadioChannel) .. "sendFromReference(text, reference, power, readMode)"
		
		if type(_text) == "string" and ATMEFct.isReference(_reference) == true and type(_power) == "number" and 
												(_readMode == nil or type(_readMode) == "boolean") then
			local powerValue = math.floor(_power)
			
			if powerValue > 0 then
				local randomRead = false
				local readOnce = true
				
				if type(_readMode) == "boolean" then
					randomRead = _readMode
					readOnce = false
				end
				
				local err, playlistNew = ATMEFct.C_Sound_getPlaylist(_text, false)

				if err == true then
					ATMEFct.errorOutputCore(nil, fctDefinition, playlistNew, false)
					return
				end

				if playlistNew == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "text string has no sound token", false)
					return
				end
				
				ATMEFct.C_RadioChannel_initTransmission({
					playlist = playlistNew,
					autoRestartDCS = false,
					readOnce = readOnce,
					randomRead = randomRead,
					point = ATMEFct.getPointFromReference(_reference),
					power = powerValue,
					voice = localMembers.currentVoice,
					localMembersRadioChannel = localMembers,
				})
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power must be > 0", false)
			end
		else
			if type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", true)
			elseif ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_power) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", false)
			elseif _readMode ~= nil and type(_readMode) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "readMode type must be nil or boolean", false)
			end
		end
	end
	
	function newRadioChannel.loopFromUnit(self, _text, _unit)
		local fctDefinition = ATMEFct.getErrorPrefix(newRadioChannel) .. "loopFromUnit(text, unit)"

		if type(_text) == "string" and ATMEFct.isObjectClass(_unit, ATME.C_AIUnit) == true then
		
			local err, playlistNew = ATMEFct.C_Sound_getPlaylist(_text, false, true)

			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, playlistNew, false)
				return
			end

			if playlistNew == nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text string has no sound token", false)
				return
			end
			
			local localMembersUnit = ATMEDatas.localMembers[_unit:getIndex()]
			
			ATMEFct.C_RadioChannel_initTransmission({
				unitName = _unit:getName(),
				playlist = playlistNew,
				autoRestartDCS = true,
				readOnce = true,
				randomRead = false,
				voice = localMembersUnit.currentVoice,
				localMembersRadioChannel = localMembers,
			})
		else
			if type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", true)
			elseif ATMEFct.isObjectClass(_unit, ATME.C_AIUnit) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "unit type must be ATME.C_AIUnit", false)
			end
		end		
	end
	
	function newRadioChannel.loopFromReference(self, _text, _reference, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newRadioChannel) .. "loopFromReference(text, reference, power)"
		
		if type(_text) == "string" and ATMEFct.isReference(_reference) == true and type(_power) == "number" then
			local powerValue = math.floor(_power)
			
			if powerValue > 0 then
				local err, playlistNew = ATMEFct.C_Sound_getPlaylist(_text, false, true)

				if err == true then
					ATMEFct.errorOutputCore(nil, fctDefinition, playlistNew, false)
					return
				end

				if playlistNew == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "text string has no sound token", false)
					return
				end
				
				ATMEFct.C_RadioChannel_initTransmission({
					playlist = playlistNew,
					autoRestartDCS = true,
					readOnce = true,
					randomRead = false,
					point = ATMEFct.getPointFromReference(_reference),
					power = powerValue,
					voice = localMembers.currentVoice,
					localMembersRadioChannel = localMembers,
				})
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power must be > 0", false)
			end
		else
			if type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", true)
			elseif ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_power) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", false)
			end
		end		
	end

	function newRadioChannel.stop(self)
		if localMembers.autoRestartDCS == true then
			
			if localMembers.current.unitName ~= nil then
				ATMEFct.C_AIUnit_stopTransmission(localMembers.current)
			else
				ATMEFct.C_RadioChannel_stopTransmission(localMembers.current)
			end
			
			localMembers.autoRestartDCS = false
			
			if localMembers.serialized == true then	
				local stackLocalMembers = ATMEDatas.localMembers[localMembers.stack:getIndex()]
				localMembers.current = ATMEFct.C_Stack_extract(stackLocalMembers, "TAIL")
				
				if localMembers.current ~= nil then
					localMembers.timerId = timer.scheduleFunction(ATMETimers.C_RadioChannel_transmissionTimer, localMembers, timer.getTime() + 1)
					return
				end
			end
		else
			localMembers.stopCurrent = true
		end
	end
	
	function newRadioChannel.stopAll(self)
		if localMembers.autoRestartDCS == true then
			if localMembers.current.unitName ~= nil then
				ATMEFct.C_AIUnit_stopTransmission(localMembers.current)
			else
				ATMEFct.C_RadioChannel_stopTransmission(localMembers.current)
			end

			localMembers.autoRestartDCS = false
		else
			localMembers.stopCurrent = true
		end
		
		localMembers.stack:removeAll()
	end
	
	function newRadioChannel.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return false, newRadioChannel
end

ATMEDatas.C_RadioChannel_MT.__call = function(t, _modulation, _frequency, _serialized)
	local fctDefinition = "Constructor :\nATME.C_RadioChannel(modulation, frequency, [serialized])"
	
	if type(_modulation) == "string" and type(_frequency) == "number" and (_serialized == nil or type(_serialized) == "boolean") then
		if ATMEConst.modulationValues[_modulation] ~= nil then
			local frequencyValue = math.floor(_frequency)
			if frequencyValue > 0 then
				local serialized = true
					
				if _serialized ~= nil then
					serialized = _serialized
				end
				
				local err, newRadioChannel =  ATMEFct.C_RadioChannel_create(nil, nil, _modulation, frequencyValue, serialized)
				
				if err == true then
					ATMEFct.errorOutputCore(nil, fctDefinition, newRadioChannel, false)
					return
				end
					
				return newRadioChannel
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "frequency must be > 0", false)
			end	
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "modulation must be 'AM' or 'FM'", false)
		end
	else
		if type(_modulation) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "modulation type must be string", false)
		elseif type(_frequency) ~= "number" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "frequency type must be number", false)
		elseif _serialized ~= nil and type(_serialized) ~= "boolean" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "serialized type must be nil or boolean", false)
		end
	end
end

setmetatable(ATME.C_RadioChannel,ATMEDatas.C_RadioChannel_MT)

ATMEDatas.C_RadioChannel_MT.mt.__metatable = "Private"
ATMEDatas.C_RadioChannel_MT.mt.__index = ATME.C_RadioChannel

ATMEDatas.C_RadioChannel_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_RadioChannel_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_RadioChannel : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Vector3D : gère les vecteurs 3D. Constructeur de classe accessible aux modules externes.
-- Elle donne accès aux calculs de module (module du vecteur AB = distance AB), à l'angle formé avec le nord
-- par AB (Azimuth), à la transformation en vecteur unitaire (vecteur identique de module 1) et au produit
-- scalaire calculé pour les coordonnées horizontales (x et z).
----------------------------------------------------------------------------------------------------------

ATMEFct.C_Vector3D_createNamed = function(_namedTable, _name, ...)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return false
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			local args = {...}
			
			if  #args == 0 or (#args == 3 and type(args[1]) == "number" and type(args[2]) == "number" and type(args[3]) == "number") or
					(#args == 1 and ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == true) or
					(#args == 2 and ATMEFct.isReference(args[1]) == true and ATMEFct.isReference(args[2]) == true) then

				return false, ATMEFct.C_Vector3D_create(_name, _namedTable, args)
			else
				if #args == 3 then
					if type(args[1]) ~= "number" then
						return true, "x type must be number"
					elseif type(args[2]) ~= "number" then
						return true, "y type must be number"
					elseif type(args[3]) ~= "number" then
						return true, "z type must be number"
					end
				elseif #args == 2 then
					if ATMEFct.isReference(args[1]) == false then
						return true, "refA type must be string," .. ATMEConst.errorLabelReference
					elseif ATMEFct.isReference(args[2]) == false then
						return true, "refB type must be string," .. ATMEConst.errorLabelReference
					end
				elseif #args == 1 and ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == false then
					return true, "vector type must be ATME.C_Vector3D"
				else
					return true, "must have 0 to 3 parameters"
				end
			end
		else
			return true, "name can't contain '|' character"
		end
	end
end

ATMEDatas.C_Vector3D_MT = {
	className = "C_Vector3D",
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["getAzimuth"] = true, 
		["getHModule"] = true,
		["getModule"] = true,
		["get"] = true, 
		["getX"] = true,
		["getY"] = true,
		["getZ"] = true,
		["toUnitVector"] = true,
		["getHScalarProduct"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
	},
	
	add = function(self)
		local item = {
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Vector3D = {
	getClassName = function(self)
		return ATMEDatas.C_Vector3D_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Vector3D_MT.parentClass
	end,
	
	createNamed = function(_name, ...)
		local fctDefinition = "Function ATME.C_Vector3D.createNamed(name)\nATME.C_Vector3D.createNamed(name, vector)\nATME.C_Vector3D.createNamed(name, refA, refB)\nATME.C_Vector3D.createNamed(name, x, y, z)"
		
		if type(_name) == "string" then
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newVector3D = ATMEFct.C_Vector3D_createNamed(localMembersModule.vectors3D, _name, ...)				
			
			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, newVector3D, false)
				return
			end
			
			return newVector3D
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
		end
	end,
	
	getByName = function(_name)
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].vectors3D[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Vector3D.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Vector3D_create = function(_name, _vectors3D, ...)
	local args = {...}
	local x = 0
	local y = 0
	local z = 0
	
	local newVector = {}
	local localMembers
	local index
	
	local name
	local tagsName
	
	setmetatable(newVector, ATMEDatas.C_Vector3D_MT.mt)
	index = ATMEDatas.C_Vector3D_MT:add()
	localMembers = ATMEDatas.localMembers[index]
		
	if _name ~= nil and _vectors3D ~= nil then
		name = _name
		_vectors3D[name] = newVector
		tagsName = ATMEFct.decompString(_name)
	end

	if #args == 3 and type(args[1]) == "number" and type(args[2]) == "number" and type(args[3]) == "number" then
		x = args[1]
		y = args[2]
		z = args[3]
	elseif #args == 1 and ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == true then  
		x = args[1]:getX()
		y = args[1]:getY()
		z = args[1]:getZ()
	elseif #args == 1 and type(args[1]) == "table" and type(args[1].x) == "number" and type(args[1].y) == "number" and type(args[1].z) == "number" then -- Uniquement interne ATME
		x = args[1].x
		y = args[1].y
		z = args[1].z
	elseif #args == 2 and ATMEFct.isReference(args[1]) == true and ATMEFct.isReference(args[2]) == true then
		local p1 = ATMEFct.getPointFromReference(args[1])
		local p2 = ATMEFct.getPointFromReference(args[2])

		x = p2.x - p1.x
		y = p2.y - p1.y
		z = p2.z - p1.z
	end
	
	local hModule = math.sqrt(x^2 + z^2)
	local gModule = math.sqrt(x^2 + y^2 + z^2)

	local angle = 0
	if x ~= 0 and z~= 0 then
		angle = math.deg(math.atan2(z, x))
	end
	
	
	if angle < 0 then
		angle = angle + 360
	end
	
	function newVector.getIndex(self)
		return index
	end
	
	function newVector.getName(self)
		return name
	end
	
	function newVector.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newVector) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newVector.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newVector.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newVector) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newVector.getAzimuth(self)
		return angle
	end
	
	function newVector.getHModule(self)
		return hModule
	end
	
	function newVector.getModule(self)
		return gModule
	end
	
	function newVector.get(self)
		return x, y, z
	end
	
	function newVector.getX(self)
		return x
	end

	function newVector.getY(self)
		return y
	end

	function newVector.getZ(self)
		return z
	end
	
	function newVector.toUnitVector(self)
		local unitVector
		if gModule ~= 0 then
			unitVector = ATMEFct.C_Vector3D_create(nil, nil, x/gModule, y/gModule, z/gModule)
		else
			unitVector = ATMEFct.C_Vector3D_create(nil, nil)
		end
		
		return unitVector
	end
	
	function newVector.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	function newVector.getHScalarProduct(self, v1)
		return v1:getX() * x + v1:getZ() * z
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newVector
end

ATMEDatas.C_Vector3D_MT.__call = function(t, ...)
	local args = {...}
	
	local fctDefinition = "Constructor :\nATME.C_Vector3D()\nATME.C_Vector3D(vector)\nATME.C_Vector3D(refA, refB)\nATME.C_Vector3D(x, y, z)"
	
	if  #args == 0 or (#args == 3 and type(args[1]) == "number" and type(args[2]) == "number" and type(args[3]) == "number") or
		(#args == 1 and ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == true) or
		(#args == 2 and ATMEFct.isReference(args[1]) == true and ATMEFct.isReference(args[2]) == true) then
		
		return ATMEFct.C_Vector3D_create(nil, nil, ...)
	else
		if #args == 3 then
			if type(args[1]) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "x type must be number", false)
			elseif type(args[2]) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "y type must be number", false)
			elseif type(args[3]) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "z type must be number", false)
			end
		elseif #args == 2 then
			if ATMEFct.isReference(args[1]) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "refA type must be".. ATMEConst.errorLabelReference, false)
			elseif ATME.isPoint(args[2]) == POINT_BAD then
				ATMEFct.errorOutputCore(nil, fctDefinition, "refB type must be".. ATMEConst.errorLabelReference, false)
			end
		elseif #args == 1 and ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == false then
			ATMEFct.errorOutputCore(nil, fctDefinition, "vector type must be ATME.C_Vector3D", false)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "must have 0 to 3 parameters", false)
		end
	end
end

setmetatable(ATME.C_Vector3D,ATMEDatas.C_Vector3D_MT)

ATMEDatas.C_Vector3D_MT.mt.__metatable = "Private"
ATMEDatas.C_Vector3D_MT.mt.__index = ATME.C_Vector3D

ATMEDatas.C_Vector3D_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Vector3D_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Vector3D : Bad key " .. k, "table locked", false)
	end
end

ATMEDatas.C_Vector3D_MT.mt.__add = function(v1, v2)
	v1x, v1y, v1z = v1:get()
	v2x, v2y, v2z = v2:get()

	return C_Vector3D(v1x + v2x, v1y + v2y, v1z + v2z)
end

ATMEDatas.C_Vector3D_MT.mt.__sub = function(v1, v2)
	v1x, v1y, v1z = v1:get()
	v2x, v2y, v2z = v2:get()
	
	return C_Vector3D(v1x - v2x, v1y - v2y, v1z - v2z)
end



----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Line2D : gère les droites horizontales (2D). Format DCS repère z0x
--
-- Une droite est formée d'un vecteur et d'un point d'origine, référence 0 de la droite pour cette classe.
--
-- Forme Az + Bx + C = 0
-- Vecteur directeur v(xv, yv, zv) => A = xv et B = -zv
-- Point d'origine (xref, yref, zref) => C = -A * zref - B * xref
--
-- Droite perpendiculaire déduite et passant au point d'origine :
-- Vecteur directeur vp(xvp, yvp, zvp) => xvp = B et zvp = A
-- Forme : -Bz + Ax + C = 0 et C = -A * zref + B * xref
--
-- NOTE : l'axe y est non significatif
----------------------------------------------------------------------------------------------------------

ATMEFct.C_Line2D_createNamed = function(_namedTable, _name, ...)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return false
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			local args = {...}
			
			if (#args == 1 and ATMEFct.isObjectClass(args[1], ATME.C_Line2D) == true)  or
					(#args == 2 and ATMEFct.isReference(args[1]) == true and ATMEFct.isReference(args[2]) == true) or
					(#args == 2 and ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == true and ATMEFct.isReference(args[2]) == true) then
				return false, ATMEFct.C_Line2D_create(_name, _namedTable, ...)
			else
				if #args == 2 then
					if ATMEFct.isReference(args[1]) == false and ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == false then
						return true, "vector or refA type must be ATME.C_Vector3D, string," .. ATMEConst.errorLabelReference
					elseif ATMEFct.isReference(args[2]) == false then
						return true, "origin or refB type must be string,".. ATMEConst.errorLabelReference
					end
				elseif #args == 1 and ATMEFct.isObjectClass(args[1], ATME.C_Line2D) == false then
					return true, "line type must be ATME.C_Line2D"
				else
					return true, "must have 1 or 2 parameters"
				end
			end
		else
			return true, "name can't contain '|' character"
		end
	end
end

ATMEDatas.C_Line2D_MT = {
	className = "C_Line2D",
	mt = {},
	
	keys = {
		["get"] = true,
		["getA"] = true,
		["getB"] = true,
		["getC"] = true,
		["getX"] = true,
		["getZ"] = true,
		["isNSAxis"] = true,
		["isWEAxis"] = true,
		["getOrigin"] = true,
		["getPerpendicular"] = true,
		["getSegment2D"] = true,
		["getBissectrix"] = true,
		["getPointFromOrigin"] = true,
		["getIntersection"] = true,
		["getPointHDistance"] = true,
		["getPointDistance"] = true,
		["getVector3D"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
	},
	
	add = function(self)
		local item = {
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Line2D = {
	getClassName = function(self)
		return ATMEDatas.C_Line2D_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Line2D_MT.parentClass
	end,
	
	createNamed = function(_name, ...)
		local fctDefinition = "Function ATME.C_Line2D.createNamed(name, line)\nATME.C_Line2D.createNamed(name, vector, origin)\nATME.C_Line2D.createNamed(name, refA, refB)"
		
		if type(_name) == "string" then
			local args = {...}
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newLine2D = ATMEFct.C_Line2D_createNamed(localMembersModule.lines2D, _name, ...)				
			
			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, newLine2D, false)
				return
			end
			
			return newLine2D
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
		end
	end,
	
	getByName = function(_name)
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].lines2D[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Line2D.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Line2D_special = function(_line2D, _line2D1, _intersect, _parallel)
	local a, b, c = _line2D:get()
	local a1, b1, c1 = _line2D1:get()
	local result
	
	local delta = a1 * b - b1 * a
			
	if delta ~= 0 then
		if a1 ~= 0 then
			if a ~= 0 then
				local xTemp = (a * c1 - a1 * c) / delta
				result = ATMEFct.convertToPoint3D({
									x = xTemp,
									y = (-b * xTemp - c) / a,
								})
			else
				local xTemp = -c / b
				result = ATMEFct.convertToPoint3D({
									x = xTemp,
									y = (-b1 * xTemp - c1) / a1,
								})
			end
		else
			local xTemp = -c1 / b1
			result = ATMEFct.convertToPoint3D({
								x = xTemp,
								y = (-b * xTemp - c) / a,
							})
		end
		
		if _intersect == true then
			-- Intersection
			return result
		end
		
		-- Bissectrices
		
		local d = math.sqrt(a * a + b * b)
		local d1 = math.sqrt(a1 * a1 + b1 * b1)
		
		local v1 = ATMEFct.C_Vector3D_create(nil, nil, d1 * a + d * a1, 0, -d1 * b - d * b1)
		local bis1 = ATMEFct.C_Line2D_create(nil, nil, v1, result)
		
		local v2 = ATMEFct.C_Vector3D_create(nil, nil,  d1 * a - d * a1, 0, -d1 * b + d * b1)
		local bis2 = ATMEFct.C_Line2D_create(nil, nil, v2, result)
		
		return bis1, bis2
	else
		if _parallel == true then	
			
			local deltac = a1 * c - c1 * a
			if math.abs(deltac) < 1e-3 then
				return _line2D
			end
		else
			-- Test si parallèle, pour éviter les erreurs d'arrondis sur tracking WP 
			-- => _parallel = false
			
			return _line2D
		end
	end
end

ATMEFct.C_Line2D_getPointFromOrigin = function(_a, _b, _origin, _offset)
	local n = math.sqrt(_offset^2/(_a^2 + _b^2))
	local point = {}
	
	if _offset < 0 then
		n = -n
	end

	point.x = _origin.x + n * _a
	point.z = _origin.z - n * _b
	point.y = ATMEFct.getSurfaceAltitude(point)
	
	return point
end

ATMEFct.C_Line2D_getSegment2D = function(_origin, _a, _b, _offset1, _offset2)
	local n1 = math.sqrt(_offset1^2/(_a^2 + _b^2))
	local points = {}
	
	if _offset1 < 0 then
		n1 = -n1
	end

	points[1] = {}
	points[1].x = _origin.x + n1 * _a
	points[1].z = _origin.z - n1 * _b
	points[1].y = ATMEFct.getSurfaceAltitude(points[1])
	
	local n2
	if _offset2 == nil then
		n2 = -n1
	else
		n2 = math.sqrt(_offset2^2/(_a^2 + _b^2))
		
		if _offset2 < 0 then
			n2 = -n2
		end
	end
	
	points[2] = {}
	points[2].x = _origin.x + n2 * _a
	points[2].z = _origin.z - n2 * _b
	points[2].y = ATMEFct.getSurfaceAltitude(points[2])
	
	return ATMEFct.C_Segment2D_create(nil, nil, points)
end

ATMEFct.C_Line2D_create = function(_name, _lines2D, ...)
	local args = {...}
	local a = 0
	local b = 0
	local c = 0
	local origin
	local newLine = {}
	
	setmetatable(newLine, ATMEDatas.C_Line2D_MT.mt)
	
	local index = ATMEDatas.C_Line2D_MT:add()
	local localMembers = ATMEDatas.localMembers[index]
	
	local name
	local tagsName
	
	if _name ~= nil and _lines2D ~= nil then
		name = _name
		_lines2D[name] = newLine

		tagsName = ATMEFct.decompString(_name)
	end
	
	if #args == 1 then
		a = args[1]:getA()
		b = args[1]:getB()
		c = args[1]:getC()
		origin = args[1]:getOrigin()
	elseif #args == 2 then
		if ATMEFct.isReference(args[1]) == true and ATMEFct.isReference(args[2]) == true then
			origin = ATMEFct.getPointFromReference(args[1])
			local p2 = ATMEFct.getPointFromReference(args[2])

			if p2.x ~= origin.x or p2.z ~= origin.z then
				a = p2.x - origin.x
				b = origin.z - p2.z
				c = - b * origin.x - a * origin.z
			else
				return
			end
		else
			origin = ATMEFct.getPointFromReference(args[2])

			a = args[1]:getX()
			b = -args[1]:getZ()
			c = - b * origin.x - a * origin.z
		end
	end
	
	local vector = ATMEFct.C_Vector3D_create(nil, nil, a, 0, -b)
	
	function newLine.getIndex(self)
		return index
	end
	
	function newLine.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	function newLine.getName(self)
		return name
	end
	
	function newLine.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newLine.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newLine.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newLine.get(self)
		return a, b, c
	end
	
	function newLine.getA(self)
		return a
	end

	function newLine.getB(self)
		return b
	end

	function newLine.getC(self)
		return c
	end
	
	function newLine.getZ(self, _x)
		if a ~= 0 then
			return (-b * _x - c) / a
		end
	end
	
	function newLine.getX(self, _z)
		if b ~= 0 then
			return (-a * _z - c) / b
		end
	end
	
	function newLine.isNSAxis(self)
		return b == 0
	end
	
	function newLine.isWEAxis(self)
		return a == 0
	end
	
	function newLine.getOrigin(self)
		return {
			x = origin.x,
			y = origin.y,
			z = origin.z,
		}
	end
	
	function newLine.getVector3D(self)
		return vector
	end
	
	function newLine.getPointFromOrigin(self, _offset)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "getPointFromOrigin(offset)"
		
		if type(_offset) == "number" then
			return ATMEFct.C_Line2D_getPointFromOrigin(a, b, origin, _offset)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "offset type must be number", true)
		end
	end
	
	function newLine.getSegment2D(self, _offset1, _offset2)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "getSegment2D(offset1, [offset2])"
		
		if type(_offset1) == "number" and (_offset2 == nil or type(_offset2) == "number") then
			if (_offset1 ~= 0 and _offset2 == nil) or _offset1 ~= _offset2 then
				return ATMEFct.C_Line2D_getSegment2D(origin, a, b, _offset1, _offset2)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "offset1 must be ~= 0 or ~= offset2", false)
			end
		else
			if type(_offset1) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "offset1 type must be number", true)
			elseif _offset2 ~= nil and  type(_offset2) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "offset2 type must be nil or number", false)
			end
		end
	end
		
	function newLine.getPerpendicular(self, _offset)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "getPerpendicular([offset])"
		
		if _offset == nil or type(_offset) == "number" then
			local v = ATMEFct.C_Vector3D_create(nil, nil, b, 0, a)
			
			local perpendicularLine
			
			if _offset == nil or _offset == 0 then
				perpendicularLine = ATMEFct.C_Line2D_create(nil, nil, v, self:getOrigin())
			else
				perpendicularLine = ATMEFct.C_Line2D_create(nil, nil, v, ATMEFct.C_Line2D_getPointFromOrigin(a, b, origin, _offset))
			end
			
			return perpendicularLine
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "offset type must be nil or number", true)
		end
	end
	
	function newLine.getBissectrix(self, _line2D)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "getBissectrix(line2D)"
		
		if ATMEFct.isObjectClass(_line2D, ATME.C_Line2D) == true then
			return ATMEFct.C_Line2D_special(self, _line2D, false, true)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "line2D type must be ATME.C_Line2D", true)
		end
	end
	
	function newLine.getPointHDistance(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "getPointHDistance(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATME.getPointFromReference(_reference)
			
			return math.abs(a * point.z + b * point.x + c) / math.sqrt(a * a + b * b)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
	end
	
	function newLine.getPointDistance(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "getPointDistance(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATME.getPointFromReference(_reference)
			
			local hdist = math.abs(a * point.z + b * point.x + c) / math.sqrt(a * a + b * b)
			
			return math.sqrt(hdist * hdist + point.y * point.y)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
	end
	
	function newLine.getIntersection(self, _line2D)
		local fctDefinition = ATMEFct.getErrorPrefix(newLine) .. "getIntersection(line2D)"
		
		if ATMEFct.isObjectClass(_line2D, ATME.C_Line2D) == true then
			return ATMEFct.C_Line2D_special(self, _line2D, true, true)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "line2D type must be ATME.C_Line2D", true)
		end
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newLine
end

ATMEDatas.C_Line2D_MT.__call = function(t, ...)
	local args = {...}
	local fctDefinition = "Constructor :\nATME.C_Line2D(line)\nATME.C_Line2D(vector, origin)\nATME.C_Line(refA, refB)"
	
	if (#args == 1 and ATMEFct.isObjectClass(args[1], ATME.C_Line2D) == true)  or
			(#args == 2 and ATMEFct.isReference(args[1]) == true and ATMEFct.isReference(args[2]) == true) or
			(#args == 2 and ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == true and ATMEFct.isReference(args[2]) == true) then
			
		local line2D = ATMEFct.C_Line2D_create(nil, nil, ...)
		
		if line2D == nil then
			ATMEFct.errorOutputCore(nil, fctDefinition, "refA and refB must be differents", false)
		else
			return line2D
		end
	else
		if #args == 2 then
			if ATMEFct.isReference(args[1]) == false and ATMEFct.isObjectClass(args[1], ATME.C_Vector3D) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "vector or refA type must be ATME.C_Vector3D," .. ATMEConst.errorLabelReference, false)
			elseif ATMEFct.isReference(args[2]) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "origin or refB type must be".. ATMEConst.errorLabelReference, false)
			end
		elseif #args == 1 and ATMEFct.isObjectClass(args[1], ATME.C_Line2D) == false then
			ATMEFct.errorOutputCore(nil, fctDefinition, "line type must be ATME.C_Line2D", false)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "must have 1 or 2 parameters", false)
		end
	end
end

setmetatable(ATME.C_Line2D,ATMEDatas.C_Line2D_MT)

ATMEDatas.C_Line2D_MT.mt.__metatable = "Private"
ATMEDatas.C_Line2D_MT.mt.__index = ATME.C_Line2D

ATMEDatas.C_Line2D_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Line2D_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Line2D : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Segment2D : gère des segments 2D ayant ou non des points communs. 
-- Format DCS repère z0x.
--
-- Un ensemble de segments se caractérise par un ensemble de points consécutifs.
--
-- NOTE : l'axe y est non significatif
----------------------------------------------------------------------------------------------------------

ATMEFct.C_Segment2D_createNamed = function(_namedTable, _name, _points)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return false
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			return false, ATMEFct.C_Segment2D_create(_name, _namedTable, _points)
		else
			return true, "name can't contain '|' character"
		end
	end
end

ATMEFct.C_Segment2D_intersect = function(_datas, _pA, _pB)
	local p1 = _datas.p1
	local p2 = _datas.p2
	local vector1 = _datas.vector
	
	local vector2 = {
		x = _pB.x - _pA.x,
		y = 0,
		z = _pB.z - _pA.z,
	}

	local det = vector1:getZ() * vector2.x - vector1:getX() * vector2.z
	
	if det == 0 then
		return false
	end
	
	local t = ((p1.x - _pA.x) * vector2.z - (p1.z - _pA.z) * vector2.x) / det
	if t < 0 or t > 1 then
		return false
	end
	
	local u = ((p1.x - _pA.x) * vector1:getZ() - (p1.z - _pA.z) * vector1:getX()) / det
	if u < 0 or u > 1 then
		return false
	end
	
	local side = ATMEFct.whichSide(p1, p2, _pB)
	if side < 0 then
		side = -1
	else
		side = 1
	end
	
	return true, side
end

ATMETimers.C_Segment2D_timer = function(_datas)
	if _datas.tracking.localMembersObject.objectDCS:isExist() == true then
		if _datas.tracking.object:isActivated() == true then
			local currentPos = _datas.tracking.object:getPosition()
			local intersect, side = ATMEFct.C_Segment2D_intersect(_datas.localMembers.datas, _datas.tracking.pos, currentPos)
			
			_datas.tracking.pos = currentPos

			if intersect == true then		
				local datas = {
					direction = ATMEConst.directionValue[side],
					object = _datas.tracking.object,
				}

				for _moduleName, _callback in pairs(_datas.tracking.callbacks) do
					ATMEDatas.currentModuleName = _moduleName
					
					ATMEFct.outputCore(nil,"Appel callback - Franchissement de segment par l'unité " .. _datas.tracking.name .. " pour le module " .. _moduleName, 2)
					xpcall(function() _callback(_datas.segment2D, datas) end, function() ATMEFct.errUserHandler('Tracking callback') end)
				end
			end
		end
		
		return timer.getTime() + 2
	else
		_datas.localMembers.tracking[_datas.tracking.name] = nil
	end
end

ATMEDatas.C_Segment2D_MT = {
	className = "C_Segment2D",
	mt = {},
	
	keys = {
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["getDatas"] = true,
		["addTrackingCallback"] = true,
		["removeTrackingCallback"] = true,
	},
	
	add = function(self)
		local item = {
			datas = {},
			tracking = {},
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].tracking = nil
	end,
}

ATME.C_Segment2D = {
	getClassName = function(self)
		return ATMEDatas.C_Segment2D_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Segment2D_MT.parentClass
	end,
	
	createNamed = function(_name, _refA, _refB)
		local fctDefinition = "Function ATME.C_Segment2D.createNamed(name, refA, refB)"
		
		if type(_name) == "string" and ATMEFct.isReference(_refA) == true and ATMEFct.isReference(_refB) == true then
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local points = {
				[1] = ATMEFct.getPointFromReference(_refA),
				[2] = ATMEFct.getPointFromReference(_refB),
			}

			local err, newSegment2D = ATMEFct.C_Segment2D_createNamed(localMembersModule.segments2D, _name, points)				
			
			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, newSegment2D, false)
				return
			end

			return newSegment2D
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif ATMEFct.isReference(_refA) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "refA type must be".. ATMEConst.errorLabelReference, false)
			elseif ATMEFct.isReference(_refB) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "refB type must be".. ATMEConst.errorLabelReference, false)
			end
		end
	end,
	
	getByName = function(_name)	
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].segments2D[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Segment2D.getByName(name)", "name type must be string", false)
		end
	end,
	
	createSegments = function(_points)
		local fctDefinition = "Function : ATME.C_Segment2D.createSegments(points)\nATME.C_Segment2D.createSegments(dcsZoneName)"
		
		if type(_points) == "table" or type(_points) == "string" then
			local pErr
			local points = {}
			if type(_points) == "table" then
				for _key, _value in ipairs(_points) do
					if pErr == nil then
						local point
						
						if ATME.isPoint(_value) ~= POINT_BAD then
							point = ATME.convertToPoint3D(_value)
						elseif type(_value) == "string" then
							local zone = ATME.getDCSZone(_value)
							
							if zone ~= nil then
								point = zone.point
							else
								pErr = "Bad zone in item key " .. _key .. " : DCS Zone " .. _value .. " doesn't exist"
							end
						else
							pErr = "Bad type for item key " .. _key
						end
						
						if pErr == nil then
							table.insert(points, point)
						end
					end
				end
			else
				for _i = 1, 40 do
					local currentName = _points .. string.format("#%02d", _i)
					local zone = ATME.getDCSZone(currentName)
					
					if zone ~= nil then
						table.insert(points, zone.point)
					end
				end
			end
			
			if pErr ~= nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, pErr, false)
			else
				if #points >= 2 then
					local segments = {}
					
					for _i = 2, #points do
						table.insert(segments, ATMEFct.C_Segment2D_create(nil, nil, {points[_i - 1], points[_i]}))
					end
					
					return segments
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "#points must be >= 2 and <= 40", false)
				end
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "points type must be table or string", false)
		end
	end,
}

ATMEFct.C_Segment2D_create = function(_name, _segments2D, _points)
	local newSegment = {}
	
	setmetatable(newSegment, ATMEDatas.C_Segment2D_MT.mt)
	
	local index = ATMEDatas.C_Segment2D_MT:add()
	local localMembers = ATMEDatas.localMembers[index]
	
	local name
	local tagsName
	
	if _name ~= nil and _segments2D ~= nil then
		name = _name
		_segments2D[name] = newSegment
		tagsName = ATMEFct.decompString(_name)
	end
	
	localMembers.datas.p1 = {
		x = _points[1].x,
		y = _points[1].y,
		z = _points[1].z,
	}
	localMembers.datas.p2 = {
		x = _points[2].x,
		y = _points[2].y,
		z = _points[2].z,
	}
	
	localMembers.datas.vector = ATMEFct.C_Vector3D_create(nil, nil, _points[2].x - _points[1].x, 0, _points[2].z - _points[1].z)
	
	function newSegment.getIndex(self)
		return index
	end
	
	function newSegment.getName(self)
		return name
	end
	
	function newSegment.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newSegment) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newSegment.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newSegment.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newSegment) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newSegment.getDatas(self)
		local fctDefinition = ATMEFct.getErrorPrefix(newSegment) .. "getDatas()"
		
		return localMembers.datas.p1, localMembers.datas.p2, localMembers.datas.vector
	end
	
	function newSegment.addTrackingCallback(self, _fct, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newSegment) .. "addTrackingCallback(fct, object)"

		if type(_fct) == "function" and ATMEFct.isObjectClass(_object, ATME.C_Unit) == true then
			local className = _object:getClassName()
			local name = _object:getName()
			if localMembers.tracking[name] == nil then
				localMembers.tracking[name] = {
					className = className,
					name = name,
					object = _object,
					localMembersObject = ATMEDatas.localMembers[_object:getIndex()],
					callbacks = {},
				}
				
				localMembers.tracking[name].pos = _object:getPosition()
				
				local datas = {
					tracking = localMembers.tracking[name],
					localMembers = localMembers,
					segment2D = self,
				}
				
				timer.scheduleFunction(ATMETimers.C_Segment2D_timer, datas, timer.getTime() + 1)
			end
			
			if localMembers.tracking[name].callbacks[ATMEDatas.currentModuleName] == nil then
				localMembers.tracking[name].callbacks[ATMEDatas.currentModuleName] = _fct
			end
		else
			if type(_fct) ~= "function" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
			elseif ATMEFct.isObjectClass(_object, ATME.C_Unit) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Player or ATME.C_AIUnit", false)
			end
		end
	end
	
	function newSegment.removeTrackingCallback(self, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newSegment) .. "removeTrackingCallback(object)"
		
		if ATMEFct.isObjectClass(_object, ATME.C_Unit) == true then
			if localMembers.tracking[_object:getName()] ~= nil then
				localMembers.tracking[_object:getName()] = nil
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Player or ATME.C_AIUnit", true)
		end
	end
	
	function newSegment.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newSegment
end

ATMEDatas.C_Segment2D_MT.__call = function(t, _refA, _refB)
	local fctDefinition = "Constructor :\nATME.C_Segment2D(refA, refB)"
	
	if ATMEFct.isReference(_refA) == true and ATMEFct.isReference(_refB) == true then	
		local points = {
			[1] = ATMEFct.getPointFromReference(_refA),
			[2] = ATMEFct.getPointFromReference(_refB),
		}

		return ATMEFct.C_Segment2D_create(nil, nil, points)
	else
		if ATMEFct.isReference(_refA) == false then
			ATMEFct.errorOutputCore(nil, fctDefinition, "refA type must be" .. ATMEConst.errorLabelReference, false)
		elseif ATMEFct.isReference(_refB) == false then
			ATMEFct.errorOutputCore(nil, fctDefinition, "refB type must be" .. ATMEConst.errorLabelReference, false)
		end
	end
end

setmetatable(ATME.C_Segment2D,ATMEDatas.C_Segment2D_MT)

ATMEDatas.C_Segment2D_MT.mt.__metatable = "Private"
ATMEDatas.C_Segment2D_MT.mt.__index = ATME.C_Segment2D

ATMEDatas.C_Segment2D_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Segment2D_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Segment2D : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classes Zones 2D et 3D
--      ATME.C_Circle : gère les cercles sur le plan zOx.
--		ATME.C_Polygon : gère les polygones non complexes dans le plan zOx
-- 		ATME.C_Rectangle : gère les rectangles dans le plan zOx
-- 		ATME.C_PickupZone : gère les zones de chargement d'unité qui sont des zones DCS particulières
--                          donc aussi des cercles 2D.
--		ATME.C_Sphere : gère les sphères en 3D
--		ATME.C_Cylinder : gère les cylindres verticaux en 3D (rotation autour de l'axe y de DCS)
--		ATME.C_Box : gère les boites en 3D
----------------------------------------------------------------------------------------------------------

ATMEFct.getActualCenter = function( _center, _unitDatas)
	if _unitDatas ~= nil then
		if ATMEDatas.allAIUnitsAndPlayers[_unitDatas.unitName] ~= nil then
			local origin = ATMEDatas.allAIUnitsAndPlayers[_unitDatas.unitName]:getPosition()				

			local point = {
				x = _unitDatas.distance * math.cos(_unitDatas.azimuthRad),
				y = _unitDatas.distance * math.sin(_unitDatas.azimuthRad),
			}
			
			local unitAngle = math.rad(ATMEDatas.allAIUnitsAndPlayers[_unitDatas.unitName]:getAzimuth())

			return {
				x = origin.x + point.x * math.cos(unitAngle) - point.y * math.sin(unitAngle),
				y = origin.y + _unitDatas.deltaAltitude,
				z = origin.z + point.x * math.sin(unitAngle) + point.y * math.cos(unitAngle),
			}
		else 
			return
		end
	elseif _center ~= nil then
		return {
			x = _center.x,
			y = _center.y,
			z = _center.z,
		}
	end
end

ATMEFct.getActualAzimuth = function(_azimuth, _unitDatas)
	if _unitDatas ~= nil and _azimuth ~= nil then
		if ATMEDatas.allAIUnitsAndPlayers[_unitDatas.unitName] ~= nil then
			return ATMEFct.getAzimuthMod(_azimuth + ATMEDatas.allAIUnitsAndPlayers[_unitDatas.unitName]:getAzimuth(), 360)
		else 
			return
		end
	else
		return _azimuth
	end
end

ATMEFct.C_Circle_isInside = function(_point, _datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	
	local dx2 = (_point.x - _datas.actualCenter.x) * (_point.x - _datas.actualCenter.x)
	local dz2 = (_point.z - _datas.actualCenter.z) * (_point.z - _datas.actualCenter.z)
	local d = math.sqrt(dx2 + dz2) 
	
	return d >= localMembers.radiusMin and d <= localMembers.radiusMax
end

ATMEFct.C_Circle_getRandomPosition = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	local angle = 2 * math.pi * math.random()
	local radius = (localMembers.radiusMax - localMembers.radiusMin) * math.random() + localMembers.radiusMin

	
	local point = {
		x = _datas.actualCenter.x + radius * math.sin(angle),
		y = _datas.actualCenter.z + radius * math.cos(angle),
	}
	
	local altitude = ATMEFct.getSurfaceAltitude(point)
	
	return {
		x = point.x,
		y = altitude,
		z = point.y,
	}
end

ATMEFct.C_Circle_duplicate = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	
	local newCircle = ATMEFct.C_Circle_create(nil, nil, _datas.actualCenter, localMembers.radiusMax)
	
	local localMembersNew = ATMEDatas.localMembers[newCircle:getIndex()]
	localMembersNew.radiusMin = localMembers.radiusMin
	localMembersNew.thickness = localMembers.thickness
	
	return newCircle
end

ATMEFct.C_Circle_searchVolume = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]

	return 	world.VolumeType.BOX,
			"C_FindContext_search",
			{
				min = {
					x = _datas.actualCenter.x - localMembers.radiusMax,
					y = 0,
					z = _datas.actualCenter.z - localMembers.radiusMax,
				},
				max = {
					x = _datas.actualCenter.x + localMembers.radiusMax,
					y = 200000,
					z = _datas.actualCenter.z + localMembers.radiusMax,
				},
			}
end

ATMEDatas.C_Circle_MT = {
	className = "C_Circle",
	parentClass = ATME.C_MovableZone2D,
	
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getUserPrivateDatas"] = true,
		["isInside"] = true,
		["getRandomPosition"] = true,
		["getCenter"] = true,
		["getRadius"] = true,
		["changeThickness"] = true,
		["getThickness"] = true,
	},
	
	add = function(self, _center, _radiusMax)
		local item = {
			center = _center,
			radiusMax = _radiusMax,
			radiusMin = 0,
			thickness = 0,
			box = {
				min = {
					x = _center.x - _radiusMax,
					y = 0,
					z = _center.z - _radiusMax,
				},
				max = {
					x = _center.x + _radiusMax,
					y = 200000,
					z = _center.z + _radiusMax,
				},
			},
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].box = nil
	end,
}

ATME.C_Circle = {
	getClassName = function(self)
		return ATMEDatas.C_Circle_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Circle_MT.parentClass
	end,
	
	createNamed = function(_name, ...)
		local args = {...}
		local fctDefinition = "Function :\nATME.C_Circle.createNamed(name, DCSZoneName)\nATME.C_Circle.createNamed(name, reference, radius)"
		local point, radius
		
		if type(_name) == "string" then
			if type(args[1]) == "string" then
				local zone = ATMEFct.getDCSZone(args[1])
				
				if zone ~=  nil then
					point = zone.point
					radius = zone.radius
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. args[1] .. "' unknown", false)
					return
				end
			elseif ATMEFct.isReference(args[1]) == true then
				if type(args[2]) == "number" then
					radius = args[2]
					if radius > 0 then
						point = ATMEFct.getPointFromReference(args[1])
						point.y = 0
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", false)
					return
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 type must be string," .. ATMEConst.errorLabelReference, false)
				return
			end
			
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newCircle = ATMEFct.C_Zone_createNamed(localMembersModule.zones, _name, ATMEFct.C_Circle_create, point, radius)				
		
			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, newCircle, false)
				return
			end

			return newCircle
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
		end
	end,
	
	getByName = function(_name)	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_Circle) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_PickupZone.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Circle_create = function(_name, _zones, _center, _radiusMax)
	local newCircle = {}
	
	setmetatable(newCircle, ATMEDatas.C_Circle_MT.mt)
	local index = ATMEDatas.C_Circle_MT:add(_center, _radiusMax)
	local localMembers = ATMEDatas.localMembers[index]

	local name
	local tagsName

	if _name ~= nil and _zones ~= nil then
		name = _name
		_zones[name] = newCircle
		tagsName = ATMEFct.decompString(_name)
	end
	
	function newCircle.getIndex(self)
		return index
	end
	
	function newCircle.getName(self)
		return name
	end
	
	function newCircle.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newCircle) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newCircle.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newCircle.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newCircle) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newCircle.changeThickness(self, _thickness)
		local fctDefinition = ATMEFct.getErrorPrefix(newCircle) .. "changeThickness([thickness])"
		
		if _thickness == nil or type(_thickness) == "number" then
			local thickness = 0
			if type(_thickness) == "number" then
				thickness = _thickness
			end
			
			if thickness >= localMembers.radiusMax then
				thickness = 0
			elseif thickness < 0 then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness must be >= 0", false)
				return
			end
			
			localMembers.thickness = thickness
			
			if thickness == 0 then
				localMembers.radiusMin = 0
			else
				localMembers.radiusMin = localMembers.radiusMax - thickness
			end
		else
			if _thickness ~= nil and type(_thickness) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness type must be nil or number", true)
			end
		end
	end
	
	function newCircle.getThickness(self)	
		if localMembers.thickness ~= 0 then
			return localMembers.thickness
		end
	end
	
	function newCircle.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	function newCircle.isInside(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newCircle) .. "isInside(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATMEFct.getPointFromReference(_reference)
			
			local datas = {
				zone = self,
				actualCenter = localMembers.center,
			}
			
			return ATMEFct.C_Circle_isInside(point, datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
		
		return false
	end
	
	function newCircle.getRandomPosition(self)
		local datas = {
			zone = self,
			actualCenter = localMembers.center,
		}
			
		return ATMEFct.C_Circle_getRandomPosition(datas)
	end
	
	function newCircle.getRadius(self)
		return localMembers.radiusMax
	end
	
	function newCircle.getCenter(self)
		return {
			x = localMembers.center.x,
			y = localMembers.center.y,
			z = localMembers.center.z,
		}
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newCircle
end


ATMEDatas.C_Circle_MT.__call = function(t, ...)
	local args = {...}
	local fctDefinition = "Constructor :\nATME.C_Circle(DCSZoneName)\nATME.C_Circle(reference, radius)"
	local point, radius
	
	if type(args[1]) == "string" then
		local zone = ATMEFct.getDCSZone(args[1])
		
		if zone ~=  nil then	
			point = zone.point
			radius = zone.radius
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. args[1] .. "' unknown", false)
			return
		end
	elseif ATMEFct.isReference(args[1]) == true then
		if type(args[2]) == "number" then
			radius = args[2]
			if radius > 0 then
				point = ATMEFct.getPointFromReference(args[1])
				point.y = 0
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
				return
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", false)
			return
		end
	else
		ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #1 type must be string," .. ATMEConst.errorLabelReference, false)
		return
	end

	return ATMEFct.C_Circle_create(nil, nil, point, radius)
end

setmetatable(ATME.C_Circle, ATMEDatas.C_Circle_MT)

ATMEDatas.C_Circle_MT.mt.__metatable = "Private"
ATMEDatas.C_Circle_MT.mt.__index = ATME.C_Circle

ATMEDatas.C_Circle_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Circle_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Circle : Bad key " .. k, "table locked", false)
	end
end

ATMEFct.C_PickupZone_duplicate = function(_datas)
	return _datas.zone
end

ATMEFct.generateSignal = function(_datas)
	if _datas.timeStart == nil then
		_datas.timeStart = timer:getTime()
		
		if _datas.type == "SMOKE" then
			local color
			if ATMEConst.smokeColors[_datas.color] == -1 then
				color = ATMEConst.smokeColors[ATMEConst.smokeColorsIndex[math.random(1,#ATMEConst.smokeColorsIndex)]]
			else
				color = ATMEConst.smokeColors[_datas.color]
			end
			
			trigger.action.smoke(_datas.point, color)
			_datas.offsetTime = math.random(5,10) + ATMEConst.maxTimeSmoke
		else
			local color
			if ATMEConst.flareColors[_datas.color] == -1 then
				color = ATMEConst.flareColors[ATMEConst.flareColorsIndex[math.random(1, #ATMEConst.flareColorsIndex)]]
			else
				color = ATMEConst.flareColors[_datas.color]
			end
			
			trigger.action.signalFlare(_datas.point, color, math.random() * 2 * math.pi)
			_datas.offsetTime = math.random(5,20)
		end
	else
		if _datas.timeStart + _datas.offsetTime < timer:getTime() then
			if _datas.type == "SMOKE" then
				local color
				if ATMEConst.smokeColors[_datas.color] == -1 then
					color = ATMEConst.smokeColors[ATMEConst.smokeColorsIndex[math.random(1,#ATMEConst.smokeColorsIndex)]]
				else
					color = ATMEConst.smokeColors[_datas.color]
				end
				
				ATMEFct.outputCore(nil, "Je passe fumée activée" , 1)
				trigger.action.smoke(_datas.point, color)
				_datas.offsetTime = math.random(5,10) + ATMEConst.maxTimeSmoke
			else
				local color
				if ATMEConst.flareColors[_datas.color] == -1 then
					color = ATMEConst.flareColors[ATMEConst.flareColorsIndex[math.random(1, #ATMEConst.flareColorsIndex)]]
				else
					color = ATMEConst.flareColors[_datas.color]
				end
				
				trigger.action.signalFlare(_datas.point, color, math.random() * 2 * math.pi)
				ATMEFct.outputCore(nil, "Je passe flare activé" , 1)
				_datas.offsetTime = math.random(5,20)
			end
			
			_datas.timeStart = timer.getTime()
		end
	end
end

ATMETimers.C_PickupZone_signalTimer = function(_localMembersPickupZone)

	ATMEFct.outputCore(nil, "C_PickupZone_signalTimer" , 2)

	if _localMembersPickupZone.datasSignal.currentDatas == nil then
		_localMembersPickupZone.datasSignal.currentDatas = ATMEDatas.signalCapabilityUnits:getFirst()
	else
		_localMembersPickupZone.datasSignal.currentDatas = ATMEDatas.signalCapabilityUnits:getNext()
		
		if _localMembersPickupZone.datasSignal.currentDatas == nil then
			_localMembersPickupZone.datasSignal.currentDatas = ATMEDatas.signalCapabilityUnits:getFirst()
		end
	end
	
	if _localMembersPickupZone.datasSignal.currentDatas ~= nil then
		local name = _localMembersPickupZone.datasSignal.currentDatas.localMembers.name
		if _localMembersPickupZone.datasSignal.currentDatas.localMembers.objectDCS:isExist() == true then
			if _localMembersPickupZone.datasSignal.currentDatas.object:getCoalitionName() == _localMembersPickupZone.coalitionName then
				if _localMembersPickupZone.datasSignal.currentDatas.object:inAir() == true then
					local v = ATMEFct.C_Vector3D_create(nil, nil, _localMembersPickupZone.datasSignal.signal.point, _localMembersPickupZone.datasSignal.currentDatas.object:getPosition())
					if _localMembersPickupZone.datasSignal.coalitionUnitsInZone[_localMembersPickupZone.datasSignal.currentDatas.localMembers.name] == nil and v:getHModule() <= _localMembersPickupZone.datasSignal.signal.radius then
						_localMembersPickupZone.datasSignal.coalitionUnitsInZone[_localMembersPickupZone.datasSignal.currentDatas.localMembers.name] = true
						ATMEFct.outputCore(nil, "Je passe Inzone" , 1)
						local datas = {
							unit = _localMembersPickupZone.datasSignal.currentDatas.object,
							isInside = true,
						}
						
						for _moduleName, _callback in pairs(_localMembersPickupZone.callbacks) do
							ATMEDatas.currentModuleName = _moduleName

							ATMEFct.outputCore(nil,"Appel handler d'entrée de pickupzone signalée pour le module " .. _moduleName, 2)
							xpcall(function() _callback(_localMembersPickupZone.datasSignal.pickupZone, datas) end, function() ATMEFct.errUserHandler('pickupZoneSignalCallbackHandler') end)
						end
									
						_localMembersPickupZone.datasSignal.nbUnits = _localMembersPickupZone.datasSignal.nbUnits + 1
					elseif _localMembersPickupZone.datasSignal.coalitionUnitsInZone[_localMembersPickupZone.datasSignal.currentDatas.localMembers.name] ~= nil and v:getHModule() > _localMembersPickupZone.datasSignal.signal.radius then
						_localMembersPickupZone.datasSignal.coalitionUnitsInZone[_localMembersPickupZone.datasSignal.currentDatas.localMembers.name] = nil
						ATMEFct.outputCore(nil, "Je passe outZone" , 1)
						local datas = {
							unit = _localMembersPickupZone.datasSignal.currentDatas.object,
							isInside = false,
						}
						
						for _moduleName, _callback in pairs(_localMembersPickupZone.callbacks) do
							ATMEDatas.currentModuleName = _moduleName

							ATMEFct.outputCore(nil,"Appel handler de sortie de pickupzone signalée pour le module " .. _moduleName, 2)
							xpcall(function() _callback(_localMembersPickupZone.datasSignal.pickupZone, datas) end, function() ATMEFct.errUserHandler('pickupZoneSignalCallbackHandler') end)
						end
						
						_localMembersPickupZone.datasSignal.nbUnits = _localMembersPickupZone.datasSignal.nbUnits - 1
					end
				end
			end
		else
			if _localMembersPickupZone.datasSignal.coalitionUnitsInZone[_localMembersPickupZone.datasSignal.currentDatas.localMembers.name] ~= nil then
				_localMembersPickupZone.datasSignal.nbUnits = _localMembersPickupZone.datasSignal.nbUnits - 1
				_localMembersPickupZone.datasSignal.coalitionUnitsInZone[_localMembersPickupZone.datasSignal.currentDatas.localMembers.name] = nil
			end
		end
	end

	if _localMembersPickupZone.datasSignal.signal ~= nil then
		if _localMembersPickupZone.datasSignal.nbUnits > 0 then 
			ATMEFct.generateSignal(_localMembersPickupZone.datasSignal.signal)
		end
	end
	
	
	if _localMembersPickupZone.nbReadyToBoard == 0 then
		_localMembersPickupZone.idTimer = nil
		return
	end
		
	return timer.getTime() + 0.5
end

ATMETimers.C_PickupZone_readyToBoardTimer = function(_datas)
	if _datas.localMembersObject.destroyed == false then
		ATMEFct.outputCore(nil, "Tracking pickupZone sur group ".. _datas.name , 2)
		
		if _datas.object:isActivated() == true then
			if _datas.object:isReadyToBoard() == false then
				local distance = ATMEFct.C_Group_evalDistances(_datas.object, _datas.localMembersPickupZone.center, "MIN_H")

				if distance <= _datas.localMembersPickupZone.radiusMax then
					_datas.isInZone = true
				end
				
				if _datas.isInZone == true then
					local firstUnit = ATMEFct.getFirstUnit(_datas.object)
					
					if firstUnit ~= nil then
						if _datas.nbTries == 5 or math.random() < _datas.randomMax or firstUnit:getSpeed() == 0 then
							ATMEFct.C_Group_changeReadyToBoard(_datas.object, true, _datas.nextWaypoint)
							ATMEFct.outputCore(nil, "C_Group_changeReadyToBoard ok random " .. _datas.nbTries, 2)
							_datas.localMembersPickupZone.nbReadyToBoard = _datas.localMembersPickupZone.nbReadyToBoard + 1
							_datas.localMembersObject.idTimerPickupZone = nil
							
							if _datas.localMembersPickupZone.datasSignal ~= nil and _datas.localMembersPickupZone.idTimer == nil then
								_datas.localMembersPickupZone.idTimer = timer.scheduleFunction(ATMETimers.C_PickupZone_signalTimer, _datas.localMembersPickupZone, timer.getTime() + 1)
							end
							
							return
						else
							_datas.nbTries = _datas.nbTries + 1
						end
					end
				end
			end
		end
		
		return timer:getTime() + 2
	else
		_datas.localMembersObject.idTimerPickupZone = nil
		_datas.localMembersObject.pickupZone = nil
	end
end

ATMEDatas.C_PickupZone_MT = {
	className = "C_PickupZone",
	parentClass = ATME.C_Zone2D,
	
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getUserPrivateDatas"] = true,
		["getRandomPosition"] = true,
		["isInside"] = true,
		["addSignalCallback"] = true,
		["removeSignalCallback"] = true,
	},
	
	add = function(self, _coalitionName, _center, _radiusMax)
		local item = {
			coalitionName = _coalitionName,
			center = _center,
			radiusMax = _radiusMax,
			radiusMin= 0,
			box = {
				min = {
					x = _center.x - _radiusMax,
					y = 0,
					z = _center.z - _radiusMax,
				},
				max = {
					x = _center.x + _radiusMax,
					y = 200000,
					z = _center.z + _radiusMax,
				},
			},
			nbReadyToBoard = 0,
			userPrivateDatas = {},
			callbacks = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].box = nil
		ATMEDatas.localMembers[_id].callbacks = nil
	end,
}

ATME.C_PickupZone = {
	getClassName = function(self)
		return ATMEDatas.C_PickupZone_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_PickupZone_MT.parentClass
	end,
	
	createNamed = function(_name, ...)
		local args = {...}
		local fctDefinition = "Function :\nATME.C_PickupZone.createNamed(name, DCSZoneName, coalitionName)\nATME.C_PickupZone.createNamed(name, reference, radius, coalitionName)\nATME.C_PickupZone.createNamed(name, DCSZoneName, coalitionName, signalRadius, [signalType])\nATME.C_PickupZone.createNamed(name, reference, radius, coalitionName, signalRadius, [signalType])"
		local point, radius
		local argId = 2
		
		if type(_name) == "string" then
			if type(args[1]) == "string" then
				local zone = ATMEFct.getDCSZone(args[1])
				
				if zone ~= nil then
					point = zone.point
					radius = zone.radius
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. args[1] .. "' unknown", false)
					return
				end
			elseif ATMEFct.isReference(args[1]) == true then
				argId = 3
				if type(args[2]) == "number" then
					radius = args[2]
					if radius > 0 then
						point = ATMEFct.getPointFromReference(args[1])
						point.y = 0
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", false)
					return
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 type must be string," .. ATMEConst.errorLabelReference, false)
				return
			end
			
			if type(args[argId]) == "string" then
				local coalitionName = string.upper(args[argId])
				if ATMEConst.coalitionIdsBlueRed[coalitionName] == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'BLUE' or 'RED'", false)
					return
				end
				
				if args[argId + 1] == nil or type(args[argId + 1]) == "number" then
					local signalRadius
					if args[argId + 1] ~= nil then
						signalRadius = math.floor(args[argId + 1])
						if signalRadius < 2000 then
							signalRadius = 2000
						end
					else
						signalRadius = 2000
					end
					
					local datasSignal = {
						radius = signalRadius,
					}

					if args[argId + 2] == nil or type(args[argId + 2]) == "string" then
						if args[argId + 2] ~= nil then
							ATMEFct.outputCore(nil, args[argId + 2], 1)
							local tokens = ATMEFct.analyzeString(args[argId + 2])
							if #tokens == 2 and tokens[1].type == "TOKEN_SIGNAL_TYPE" and (tokens[2].type == "TOKEN_COLOR" or tokens[2].type == "TOKEN_TYPE_RANDOM") then
								if tokens[1].value == "SMOKE" and ATMEConst.smokeColors[tokens[2].value] ~= nil then
									datasSignal.type = tokens[1].value
									datasSignal.color = tokens[2].value
								elseif tokens[1].value == "FLARE" and ATMEConst.flareColors[tokens[2].value] ~= nil then
									datasSignal.type = tokens[1].value
									datasSignal.color = tokens[2].value
								else
									ATMEFct.errorOutputCore(nil, fctDefinition, "signalType string bad format", false)
									return
								end
							end
						end

						local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
						local err, newPickupZone = ATMEFct.C_Zone_createNamed(localMembersModule.zones, _name, ATMEFct.C_PickupZone_create, coalitionName, point, radius, datasSignal)				

						if err == true then
							ATMEFct.errorOutputCore(nil, fctDefinition, newPickupZone, false)
							return
						end

						return newPickupZone
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "signalType type must be nil or string", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "signalRadius type must be nil or number", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
		end
	end,
	
	getByName = function(_name)	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_PickupZone) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_PickupZone.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_PickupZone_create = function(_name, _zones, _coalitionName, _center, _radiusMax, _datasSignal)
	local newPickupZone = {}

	setmetatable(newPickupZone, ATMEDatas.C_PickupZone_MT.mt)
	local index = ATMEDatas.C_PickupZone_MT:add(_coalitionName, _center, _radiusMax)
	local localMembers = ATMEDatas.localMembers[index]
							
	local name
	local tagsName

	if _name ~= nil and _zones ~= nil then
		name = _name
		_zones[name] = newPickupZone
		tagsName = ATMEFct.decompString(_name)
	end

	function newPickupZone.getIndex(self)
		return index
	end
	
	function newPickupZone.getName(self)
		return name
	end
	
	function newPickupZone.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newPickupZone) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newPickupZone.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newPickupZone.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newPickupZone) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newPickupZone.isInside(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newPickupZone) .. "isInside(reference)"
		
		if ATMEFct.isReference(_reference) == true then		
			local point = ATMEFct.getPointFromReference(_reference)
			
			local datas = {
				zone = self,
				actualCenter = localMembers.center,
			}
			
			return ATMEFct.C_Circle_isInside(point, datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
		
		return false
	end
	
	function newPickupZone.getRandomPosition(self)
		local datas = {
			zone = self,
			actualCenter = localMembers.center,
		}
			
		return ATMEFct.C_Circle_getRandomPosition(datas)
	end
	
	function newPickupZone.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	function newPickupZone.addSignalCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newPickupZone) .. "addSignalCallback(fct)"
		
		if type(_fct) == "function" then
			localMembers.callbacks[ATMEDatas.currentModuleName] = _fct
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newPickupZone.removeSignalCallback(self)
		localMembers.callbacks[ATMEDatas.currentModuleName] = nil
	end
	
	if _datasSignal ~= nil then
		localMembers.datasSignal = {
			coalitionUnitsInZone = {},
			signal = {
				type = _datasSignal.type,
				color = _datasSignal.color,
				radius = _datasSignal.radius,
			},
			nbUnits = 0,
			delay = 0,
			pickupZone = self,
		}
		
		if localMembers.datasSignal.signal.type == "SMOKE" then
			localMembers.datasSignal.signal.point = {
				x = _center.x + math.random(-20,20),
				y = _center.y,
				z = _center.z + math.random(-20,20),
			}
		elseif _datasSignal.type == "FLARE" then
			localMembers.datasSignal.signal.point = _center
		end
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end

	return newPickupZone
end


ATMEDatas.C_PickupZone_MT.__call = function(t, ...)
	local args = {...}
	local fctDefinition = "Constructor :\nATME.C_PickupZone(DCSZoneName, coalitionName)\nATME.C_PickupZone(reference, radius, coalitionName)\nATME.C_PickupZone(DCSZoneName, coalitionName, signalRadius, [signalType])\nATME.C_PickupZone(reference, radius, coalitionName, signalRadius, [signalType])"
	local point, radius
	local argId = 2
	
	if type(args[1]) == "string" then
		local zone = ATMEFct.getDCSZone(args[1])
		
		if zone ~= nil then
			point = zone.point
			radius = zone.radius
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. args[1] .. "' unknown", false)
			return
		end
	elseif ATMEFct.isReference(args[1]) == true then
		argId = 3
		if type(args[2]) == "number" then
			radius = args[2]
			if radius > 0 then
				point = ATMEFct.getPointFromReference(args[1])
				point.y = 0
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
				return
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", false)
			return
		end
	else
		ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #1 type must be string," .. ATMEConst.errorLabelReference, false)
		return
	end
	
	if type(args[argId]) == "string" then
		local coalitionName = string.upper(args[argId])
		if ATMEConst.coalitionIdsBlueRed[coalitionName] == nil then
			ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'BLUE' or 'RED'", false)
			return
		end
		
		if args[argId + 1] == nil or type(args[argId + 1]) == "number" then
			local signalRadius
			if args[argId + 1] ~= nil then
				signalRadius = math.floor(args[argId + 1])
				if signalRadius < 2000 then
					signalRadius = 2000
				end
			else
				signalRadius = 2000
			end
			
			local datasSignal = {
				radius = signalRadius,
			}

			if args[argId + 2] == nil or type(args[argId + 2]) == "string" then
				if args[argId + 2] ~= nil then
					local tokens = ATMEFct.analyzeString(args[argId + 2])
					if #tokens == 2 and tokens[1].type == "TOKEN_SIGNAL_TYPE" and (tokens[2].type == "TOKEN_COLOR" or tokens[2].type == "TOKEN_TYPE_RANDOM") then
						if tokens[1].value == "SMOKE" and ATMEConst.smokeColors[tokens[2].value] ~= nil then
							datasSignal.type = tokens[1].value
							datasSignal.color = tokens[2].value
						elseif tokens[1].value == "FLARE" and ATMEConst.flareColors[tokens[2].value] ~= nil then
							datasSignal.type = tokens[1].value
							datasSignal.color = tokens[2].value
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "signalType string bad format", false)
							return
						end
					end
				end
				
				return ATMEFct.C_PickupZone_create(nil, nil, coalitionName, point, radius, datasSignal)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "signalType type must be nil or string", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "signalRadius type must be nil or number", false)
		end
	else
		ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", false)
	end
end

setmetatable(ATME.C_PickupZone, ATMEDatas.C_PickupZone_MT)

ATMEDatas.C_PickupZone_MT.mt.__metatable = "Private"
ATMEDatas.C_PickupZone_MT.mt.__index = ATME.C_PickupZone

ATMEDatas.C_PickupZone_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_PickupZone_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_PickupZone : Bad key " .. k, "table locked", false)
	end
end

ATMEFct.C_Polygon_isInTriangle = function(_points, _point)
	local dRef
	local isInside = true

	for _i = 1, 3 do
		local j = _i + 1
		if j == 4 then
			j = 1
		end

		local d = ATMEFct.whichSide(_points[_i], _points[j], _point)
		if d ~= 0 then
			if dRef == nil then
				dRef = d
			elseif dRef ~= d then
				isInside = false
			end
		end
	end

	return isInside
end

ATMEFct.C_Polygon_isInside = function(_point, _datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	local isInside = true
	local index = 1
	
	repeat
		isInside = ATMEFct.C_Polygon_isInTriangle(localMembers.triangles[index].points, _point)
		index = index + 1
	until isInside == true or index == #localMembers.triangles + 1

	return isInside
end

ATMEFct.C_Polygon_getRandomPosition = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	
	local idTriangle = math.random(1, #localMembers.triangles)
	
	local sectorMin = math.random(1, #localMembers.triangles[idTriangle].points)
	local sectorMax = sectorMin + 1
	
	if sectorMin == #localMembers.triangles[idTriangle].points then
		sectorMax = 1
	end
	
	local angleMin
	local angleMax
	
	if localMembers.triangles[idTriangle].direction == 1 then
		angleMin = localMembers.triangles[idTriangle].angleRad[sectorMin].angle
		angleMax = localMembers.triangles[idTriangle].angleRad[sectorMax].angle
	else
		angleMin = localMembers.triangles[idTriangle].angleRad[sectorMax].angle
		angleMax = localMembers.triangles[idTriangle].angleRad[sectorMin].angle
	end
	
	local angleRad = angleMin + math.random() * (angleMax - angleMin)
	
	local pt = {
		x = localMembers.triangles[idTriangle].barycentre.x + 10 * math.cos(angleRad),
		y = localMembers.triangles[idTriangle].barycentre.z + 10 * math.sin(angleRad),
	}
	
	local line = ATMEFct.C_Line2D_create(nil, nil, localMembers.triangles[idTriangle].barycentre, pt)
	
	local intersectPoint = line:getIntersection(localMembers.triangles[idTriangle].sides[sectorMin])
	
	local dx2 = (intersectPoint.x - localMembers.triangles[idTriangle].barycentre.x) * (intersectPoint.x - localMembers.triangles[idTriangle].barycentre.x)
	local dy2 = (intersectPoint.z - localMembers.triangles[idTriangle].barycentre.z) * (intersectPoint.z - localMembers.triangles[idTriangle].barycentre.z)
	local distance = math.sqrt(dx2 + dy2)
	
	local randomDistance = distance * math.random()
	
	return ATMEFct.C_Line2D_getPointFromOrigin(line:getA(), line:getB(), line:getOrigin(), randomDistance)
end

ATMEFct.C_Polygon_duplicate = function(_datas)
	return _datas.zone
end

ATMEFct.C_Polygon_searchVolume = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	
	return world.VolumeType.BOX,
			"C_FindContext_search",
			localMembers.box
end

ATMEDatas.C_Polygon_MT = {
	className = "C_Polygon",
	parentClass = ATME.C_Zone2D,
	
	mt = {},
	
	keys = {
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["isInside"] = true,
		["getRandomPosition"] = true,
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["getCenter"] = true,
	},
	
	add = function(self, _points, _triangles)
		local point = {
			x = 0,
			y = 0,
		}
		
		local minPoint = {
			x = _points[1].x,
			y = 0,
			z = _points[1].z,
		}
			
		local maxPoint = {
			x = _points[1].x,
			y = 200000,
			z = _points[1].z,
		}
		
		for _, _point in ipairs(_points) do
			point.x = point.x + _point.x
			point.y = point.y + _point.z
			
			if minPoint.x > _point.x then
				minPoint.x = _point.x
			end
			if minPoint.z > _point.z then
				minPoint.z = _point.z
			end
			
			if maxPoint.x < _point.x then
				maxPoint.x = _point.x
			end
			if maxPoint.z < _point.z then
				maxPoint.z = _point.z
			end
		end
		
		point.x = point.x / #_points
		point.y = point.y / #_points
		
		local item = {
			points = _points,
			center = ATMEFct.convertToPoint3D(point),
			triangles = _triangles,
			box = {
				min = minPoint,
				max = maxPoint,
			},
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].box = nil
	end,
}

ATME.C_Polygon = {
	getClassName = function(self)
		return ATMEDatas.C_Polygon_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Polygon_MT.parentClass
	end,
	
	createNamed = function(_name, _vertices)
		local fctDefinition = "Function :\nATME.C_Polygon.createNamed(name, vertices)\nATME.C_Polygon.createNamed(name, DCSZoneName)"
		
		if type(_name) == "string" and type(_vertices) == "table" or type(_vertices) == "string" then
			local pErr
			local points = {}
			if type(_vertices) == "table" then
				for _key, _value in ipairs(_vertices) do
					if pErr == nil then
						local point
						
						if ATME.isPoint(_value) ~= POINT_BAD then
							point = ATMEFct.convertToPoint3D(_value)
						elseif type(_value) == "string" then
							local zone = ATMEFct.getDCSZone(_value)
							
							if zone ~= nil then
								point = zone.point
							else
								pErr = "Bad zone in item key " .. _key .. " : DCS Zone " .. _value .. " doesn't exist"
							end
						else
							pErr = "Bad type for item key " .. _key
						end
						
						if pErr == nil then
							table.insert(points, point)
						end
					end
				end
			else
				for _i = 1, ATMEConst.polygonMaxVertices do
					local currentName = _vertices .. string.format("#%03d", _i)
					local zone = ATMEFct.getDCSZone(currentName)
					
					if zone ~= nil then
						table.insert(points, zone.point)
					end
				end
			end
			
			if pErr ~= nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, pErr, false)
			else
				if #points > 2 and #points <= ATMEConst.polygonMaxVertices then
				
					-- Vérification de polygone croisé
					local intersect = false
					
					for _i = 1, #points - 1 do
						for _j = _i + 1, #points - 1 do
							local nr = (points[_i].z - points[_j].z) * (points[_j + 1].x - points[_j].x) - (points[_i].x - points[_j].x) * (points[_j + 1].z - points[_j].z)
							local ns = (points[_i].z - points[_j].z) * (points[_i + 1].x - points[_i].x) - (points[_i].x - points[_j].x) * (points[_i + 1].z - points[_i].z)
							local d = (points[_i + 1].x - points[_i].x) * (points[_j + 1].z - points[_j].z) - (points[_i + 1].z - points[_i].z) * (points[_j + 1].x - points[_j].x)

							if d ~= 0 then
								local r = nr / d
								local s = ns / d
								if r > 0 and r < 1 and s > 0 and s < 1 then
									intersect = true
								end
							end
						end
					end


					if intersect == false then
						local pErr
						-- Vérification des distances entre points consécutifs
						for _i = 1, #points - 1 do
							for _j = _i + 1, #points do
								local dX2 = (points[_i].x - points[_j].x) * (points[_i].x - points[_j].x)
								local dY2 = (points[_i].z - points[_j].z) * (points[_i].z - points[_j].z)
								local dist2 = dX2 + dY2

								if dist2 < 1 then
									-- Points trop proches distance < 1
									
									pErr = {
										first = _i,
										last = _j,
									}
								end
							end
						end

						if pErr == nil then
							local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
							local err, newPolygon = ATMEFct.C_Zone_createNamed(localMembersModule.zones, _name, ATMEFct.C_Polygon_create, points)				
			
							if err == true then
								ATMEFct.errorOutputCore(nil, fctDefinition, newPolygon, false)
								return
							end

							return newPolygon
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Bad polygon : Vertex " .. pErr.first .. " and " .. pErr.last .. " too near (distance < 1m)", false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Bad polygon : crossed vertices", false)
					end
				else
					if #points < 3 then
						ATMEFct.errorOutputCore(nil, fctDefinition, "Bad polygon : Less than 3 vertices", false)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Bad polygon : More than limit of " .. ATMEConst.polygonMaxVertices .. " vertices", false)
					end
				end
			end
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "vertices type must be table or DCSZoneName type must be string", false)
			end
		end
	end,
	
	getByName = function(_name)	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_Polygon) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_PickupZone.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Polygon_create = function(_name, _zones, _points)
	local triangles = {}
	
	local function zMin(_points)
		local zMin = _points[1].z
		local index = 1
		for _i = 2, #_points do
			if _points[_i].z < zMin then
				index = _i
				zMin = _points[_i].z
			end
		end

		return index
	end
  
	local function nextVertexIndex(_points, _index)
		local newIndex = _index + 1
		local nbPoints = #_points

		if newIndex > nbPoints then
			newIndex = 1
		end

		return newIndex
	end

	local function previousVertexIndex(_points, _index)
		local newIndex = _index - 1
		local nbPoints = #_points

		if newIndex == 0 then
			newIndex = nbPoints
		end

		return newIndex
	end

	local function newPolygon(_polygon, _start, _end)
		local points = {}
		local i = _start

		while i ~= _end do
			table.insert(points, {
									x = _polygon[i].x,
									y = _polygon[i].y,
									z = _polygon[i].z,
									})

			i = nextVertexIndex(_polygon, i)
		end
		
		table.insert(points, {
								x = _polygon[i].x,
								y = _polygon[i].y,
								z = _polygon[i].z,
								})

		return points
	end

	local function inTriangle(_points, _M)
		local det = -1
		
		if ATMEFct.C_Polygon_isInTriangle(_points, _M) == true then
			det = math.abs((_points[3].z - _points[2].z) * (_M.x - _points[2].x) - 
								(_points[3].x - _points[2].x) * (_M.z - _points[2].z))
		end

		return det
	end
	
	local function findVertexInTriangle(_points, _iP1, _iP2, _iP3)
		local nbPoints = #_points
		local j
		local maxDet = 0

		for _i = 1, nbPoints do
			if _i ~= _iP1 and _i ~= _iP2 and _i ~= _iP3 then
				local pts = {
					[1] = _points[_iP1], 
					[2] = _points[_iP2], 
					[3] = _points[_iP3],
				}
				local det = inTriangle(pts, _points[_i])

				if det > maxDet then
					maxDet = det
					j = _i  
				end
			end
		end

		return j
	end
  
	local function analyze(_points)
		-- Fonction récursive de triangulation d'un polygone
		-- Pour un polygone en cours défini par ses point, cette triangulation 
		-- est basée sur :
		--    - La recherche du z minimal dans le plan zOx appelé Start
		--    - Si aucun autre point du polygon coupe le segment défini par précédent/suivant, 
		--      le premier triangle est crée avec ces 3 points. Le point Start est supprimé et 
		--      une nouvelle analyse est faite.
		--    - sinon le point le plus loin (pour optimiser utilisation du produit vectoriel)
		--      du segment défini par précédent/suivant est mémorisé (ici M) et le nouveau segment
		--      défini par Start et M induisent la création de deux polygones à partir du premier,
		--      une nouvelle analyse est alors lancée pour chaque polygone.
		--
		-- Dans tous les cas, l'analyse s'arrête s'il ne reste que 3 points et ce tringle est alors
		-- mémorisé. Il faut noter que les triangles plats ne sont pas mémorisés. Le triangle ABC est 
		-- un triangle plat si le produit vectoriel vaut 0 pour les vecteurs AB et AC.
		
		local startIndex = zMin(_points)
		local nextIndex = nextVertexIndex(_points, startIndex)
		local prevIndex = previousVertexIndex(_points, startIndex)

		local j = findVertexInTriangle(_points, startIndex, nextIndex, prevIndex)
		
		if j == nil then
			if ATMEFct.whichSide(_points[startIndex], _points[nextIndex], _points[prevIndex]) ~= 0 then
				table.insert(triangles, {
											points = {
												_points[startIndex], 
												_points[nextIndex], 
												_points[prevIndex],
											}
										})
			end
			
			local poly = newPolygon(_points, nextIndex, prevIndex)
			if #poly == 3 then
				if ATMEFct.whichSide(poly[1], poly[2], poly[3]) ~= 0 then
					table.insert(triangles, {
												points = poly,
											})
				end
			else
				analyze(poly)
			end
		else
			local poly1 = newPolygon(_points, startIndex, j)
			local poly2 = newPolygon(_points, j, startIndex)
			if #poly1 == 3 then
				if ATMEFct.whichSide(poly1[1], poly1[2], poly1[3]) ~= 0 then
					table.insert(triangles, {
												points = poly1,
											})
				end
			else
				analyze(poly1)
			end
			if #poly2 == 3 then
				if ATMEFct.whichSide(poly2[1], poly2[2], poly2[3]) ~= 0 then
					table.insert(triangles, {
												points = poly2,
											})
				end
			else
				analyze(poly2)
			end
		end
	end
  
 
	analyze(_points)
	
	
	-- Calcul des barycentres et des droites rayon qui relient le barycentre à chaque sommet.
	-- Mémorisation des côtés représentés par des droites 2D. La droite 2D i relie le sommet i au 
	-- sommet i + 1. Un triangle = 3 côtés et 3 rayons, donc pas besoin de recherche de direction.
	-- L'angle de chaque rayon est également mémorisé. Toutes ces valeurs seront utilisées pour déterminer
	-- de manière déterministe un point aléatoire dans un polygone.
	
	-- Calcul du barycentre et des côtés de chaque triangle.
	
	for _, _triangle in ipairs(triangles) do
		_triangle.sides = {}
		
		local sx = 0
		local sy = 0
		for _i = 1, #_triangle.points - 1 do
			sx = sx + _triangle.points[_i].x
			sy = sy + _triangle.points[_i].z
			
			_triangle.sides[_i] = ATMEFct.C_Line2D_create(nil, nil, _triangle.points[_i], _triangle.points[_i + 1])
		end
		sx = sx + _triangle.points[#_triangle.points].x
		sy = sy + _triangle.points[#_triangle.points].z
			
		_triangle.sides[#_triangle.points] = ATMEFct.C_Line2D_create(nil, nil, _triangle.points[#_triangle.points], _triangle.points[1])
		
		_triangle.barycentre = ATMEFct.convertToPoint3D({
								x = sx / #_triangle.points,
								y = sy / #_triangle.points,
							})
		
		-- Mémorisation des droites rayons : lient le barycentre de chaque triangle à ses sommets. 
		-- Evaluation de l'angle
		_triangle.angleRad = {}
		for _i = 1, #_triangle.points do
			local line = ATMEFct.C_Line2D_create(nil, nil, _triangle.barycentre,  _triangle.points[_i])
			_triangle.angleRad[_i] = {
				angle = math.atan2(-line:getB(), line:getA()),
				index = _i,
			}
		end
	end


	local newPolygon = {}
	
	setmetatable(newPolygon, ATMEDatas.C_Polygon_MT.mt)
	local index = ATMEDatas.C_Polygon_MT:add(_points, triangles)
	local localMembers = ATMEDatas.localMembers[index]
	
	local name
	local tagsName

	if _name ~= nil and _zones ~= nil then
		name = _name
		_zones[name] = newPolygon
		tagsName = ATMEFct.decompString(_name)
	end

	function newPolygon.getIndex(self)
		return index
	end
	
	function newPolygon.getName(self)
		return name
	end
	
	function newPolygon.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newPolygon) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newPolygon.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newPolygon.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newPolygon) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end

	function newPolygon.getCenter(self)
		return {
				x = localMembers.center.x,
				y = localMembers.center.y,
				z = localMembers.center.z,
			}
	end
	
	function newPolygon.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	function newPolygon.isInside(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newPolygon) .. "isInside(reference)"
		local isInside = false
		
		if ATMEFct.isReference(_reference) == true then		
			local point = ATMEFct.getPointFromReference(_reference)
			
			local datas = {
				zone = self,
			}
				
			isInside = ATMEFct.C_Polygon_isInside(point, datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
		
		return isInside
	end
	
	function newPolygon.getRandomPosition(self)
		local datas = {
			zone = self,
		}
		
		return ATMEFct.C_Polygon_getRandomPosition(datas)
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newPolygon
end


ATMEDatas.C_Polygon_MT.__call = function(t, _vertices)
	local fctDefinition = "Constructor :\nATME.C_Polygon(vertices)\nATME.C_Polygon(DCSZoneName)"
	
	if type(_vertices) == "table" or type(_vertices) == "string" then
		local pErr
		local points = {}
		if type(_vertices) == "table" then
			for _key, _value in ipairs(_vertices) do
				if pErr == nil then
					local point
					
					if ATME.isPoint(_value) ~= POINT_BAD then
						point = ATMEFct.convertToPoint3D(_value)
					elseif type(_value) == "string" then
						local zone = ATMEFct.getDCSZone(_value)
						
						if zone ~= nil then
							point = zone.point
						else
							pErr = "Bad zone in item key " .. _key .. " : DCS Zone " .. _value .. " doesn't exist"
						end
					else
						pErr = "Bad type for item key " .. _key
					end
					
					if pErr == nil then
						table.insert(points, point)
					end
				end
			end
		else
			for _i = 1, ATMEConst.polygonMaxVertices do
				local currentName = _vertices .. string.format("#%03d", _i)
				local zone = ATMEFct.getDCSZone(currentName)
				
				if zone ~= nil then
					table.insert(points, zone.point)
				end
			end
		end
		
		if pErr ~= nil then
			ATMEFct.errorOutputCore(nil, fctDefinition, pErr, false)
		else
			if #points > 2 and #points <= ATMEConst.polygonMaxVertices then
			
				-- Vérification de polygone croisé
				local intersect = false
				
				for _i = 1, #points - 1 do
					for _j = _i + 1, #points - 1 do
						local nr = (points[_i].z - points[_j].z) * (points[_j + 1].x - points[_j].x) - (points[_i].x - points[_j].x) * (points[_j + 1].z - points[_j].z)
						local ns = (points[_i].z - points[_j].z) * (points[_i + 1].x - points[_i].x) - (points[_i].x - points[_j].x) * (points[_i + 1].z - points[_i].z)
						local d = (points[_i + 1].x - points[_i].x) * (points[_j + 1].z - points[_j].z) - (points[_i + 1].z - points[_i].z) * (points[_j + 1].x - points[_j].x)

						if d ~= 0 then
							local r = nr / d
							local s = ns / d
							if r > 0 and r < 1 and s > 0 and s < 1 then
								intersect = true
							end
						end
					end
				end


				if intersect == false then
					local pErr
					-- Vérification des distances entre points consécutifs
					for _i = 1, #points - 1 do
						for _j = _i + 1, #points do
							local dX2 = (points[_i].x - points[_j].x) * (points[_i].x - points[_j].x)
							local dY2 = (points[_i].z - points[_j].z) * (points[_i].z - points[_j].z)
							local dist2 = dX2 + dY2

							if dist2 < 1 then
								-- Points trop proches distance < 1
								
								pErr = {
									first = _i,
									last = _j,
								}
							end
						end
					end

					if pErr == nil then
						return ATMEFct.C_Polygon_create(nil, nil, points)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Bad polygon : Vertex " .. pErr.first .. " and " .. pErr.last .. " too near (distance < 1m)", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Bad polygon : crossed vertices", false)
				end
			else
				if #points < 3 then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Bad polygon : Less than 3 vertices", false)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Bad polygon : More than limit of " .. ATMEConst.polygonMaxVertices .. " vertices", false)
				end
			end
		end
	else
		ATMEFct.errorOutputCore(nil, fctDefinition, "vertices type must be table or DCSZoneName type must be string", false)
	end
end

setmetatable(ATME.C_Polygon, ATMEDatas.C_Polygon_MT)

ATMEDatas.C_Polygon_MT.mt.__metatable = "Private"
ATMEDatas.C_Polygon_MT.mt.__index = ATME.C_Polygon

ATMEDatas.C_Polygon_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Polygon_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Polygon : Bad key " .. k, "table locked", false)
	end
end

ATMEFct.C_Rectangle_isInside = function(_point, _datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	local dx = _point.x - _datas.actualCenter.x
	local dz = _point.z - _datas.actualCenter.z
	
	local angleRad = math.rad(_datas.actualAzimuth)
	
	local lPoint = {
		x = dx * math.cos(angleRad) + dz * math.sin(angleRad),
		y = -dx * math.sin(angleRad) + dz * math.cos(angleRad),
	}
	
	local ok = (math.abs(lPoint.x) >= localMembers.halfLengthMin and math.abs(lPoint.x) <= localMembers.halfLength and 
					math.abs(lPoint.y) <= localMembers.halfWidth and localMembers.isRounded == false) or
			 (math.abs(lPoint.y) >= localMembers.halfWidthMin and math.abs(lPoint.y) <= localMembers.halfWidth and 
					math.abs(lPoint.x) <= localMembers.halfLength)
					
	if ok == false and localMembers.isRounded == true then	
		local dxPoint = lPoint.x - localMembers.halfLength
		local dyPoint = lPoint.y
		local d = math.sqrt(dxPoint * dxPoint + dyPoint * dyPoint)
		local angle = math.atan2(dyPoint, dxPoint)
	
		ok = angle > -math.pi / 2 and angle < math.pi / 2 and d >= localMembers.halfWidthMin and d <= localMembers.halfWidth
		
		if ok == false then
			dxPoint = lPoint.x + localMembers.halfLength
			local d = math.sqrt(dxPoint * dxPoint + dyPoint * dyPoint)
			local angle = math.atan2(dyPoint, dxPoint)
		
			ok = (angle < -math.pi / 2 or angle > math.pi / 2) and d >= localMembers.halfWidthMin and d <= localMembers.halfWidth
		end
	end

	return ok   
end

ATMEFct.C_Rectangle_getRandomPosition = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	local angleRad = math.rad(_datas.actualAzimuth)
	local point
	
	local rand = math.random()
	if rand < localMembers.probLength then
		-- Soit la surface correspondant au grand coté avec l'épaisseur
		
		local dx = ATMEConst.randomSign[math.random(1,2)] * localMembers.halfLength * math.random()
		local dy = ATMEConst.randomSign[math.random(1,2)] * (localMembers.deltaWidth * math.random() + localMembers.halfWidthMin)
		
		
		point = {
				x = _datas.actualCenter.x + dx * math.cos(angleRad) - dy * math.sin(angleRad),
				y = _datas.actualCenter.z + dx * math.sin(angleRad) + dy * math.cos(angleRad),
		}
	else 
		-- Soit la surface correspondant au petit coté moins l'épaisseur
		-- Soit sur 2x le demi-cercle de part et d'autre du centre sur l'axe x.
		
		if localMembers.isRounded == true then
			local angle = 2 * math.pi * math.random()
			local radius = localMembers.deltaWidth * math.random() + localMembers.halfWidthMin
			local i = 1
			
			if angle >= math.pi then
				i = -1
			end
			
			local localPoint = {
				x = i * localMembers.halfLength + radius * math.sin(angle),
				y = radius * math.cos(angle),
			}
			
			point = {
				x = _datas.actualCenter.x + localPoint.x * math.cos(angleRad) - localPoint.y * math.sin(angleRad),
				y = _datas.actualCenter.z + localPoint.x * math.sin(angleRad) + localPoint.y * math.cos(angleRad),
			}
		else
			local dx = ATMEConst.randomSign[math.random(1,2)] * (localMembers.deltaLength * math.random() + localMembers.halfLengthMin)
			local dy = ATMEConst.randomSign[math.random(1,2)] * localMembers.halfWidthMin * math.random()
			
			point = {
				x = _datas.actualCenter.x + dx * math.cos(angleRad) - dy * math.sin(angleRad),
				y = _datas.actualCenter.z + dx * math.sin(angleRad) + dy * math.cos(angleRad),
			}
		end
	end
	
	return ATMEFct.convertToPoint3D(point)
end

ATMEFct.C_Rectangle_duplicate = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	
	local newRectangle = ATMEFct.C_Rectangle_create(nil, nil, _datas.actualCenter, _datas.actualAzimuth, localMembers.length, 
													localMembers.width, localMembers.isRounded)
	
	local localMembersNew = ATMEDatas.localMembers[newRectangle:getIndex()]
	localMembersNew.halfLengthMin = localMembers.halfLengthMin
	localMembersNew.halfWidthMin = localMembers.halfWidthMin
	localMembersNew.thickness = localMembers.thickness
	localMembersNew.deltaLength = localMembers.deltaLength
	localMembersNew.deltaWidth = localMembers.deltaWidth
	localMembersNew.thicknessMax = localMembers.thicknessMax
	localMembersNew.probLength = localMembers.probLength
	
	return newRectangle
end

ATMEFct.C_Rectangle_searchVolume = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	local lPoints = {}
	local points = {}
	
	local angleRad = math.rad(_datas.actualAzimuth)
	
	local length = localMembers.halfLength
	if localMembers.isRounded == true then
		length = length + localMembers.halfWidth
	end

	table.insert(lPoints, {x = length, y = localMembers.halfWidth})
	table.insert(lPoints, {x = -length, y = localMembers.halfWidth})
	table.insert(lPoints, {x = -length, y = -localMembers.halfWidth})
	table.insert(lPoints, {x = length, y = -localMembers.halfWidth})

	for _index, _point in ipairs(lPoints) do
		points[_index] =  {
			x = _datas.actualCenter.x + _point.x * math.cos(angleRad) - _point.y * math.sin(angleRad),
			y = _datas.actualCenter.z + _point.x * math.sin(angleRad) + _point.y * math.cos(angleRad),
		}
	end
	
	local minPoint = {
		x = points[1].x,
		y = 0,
		z = points[1].y,
	}
		
	local maxPoint = {
		x = points[1].x,
		y = 200000,
		z = points[1].y,
	}
	
	for _, _point in ipairs(points) do
		if minPoint.x > _point.x then
			minPoint.x = _point.x
		end
		
		if minPoint.z > _point.y then
			minPoint.z = _point.y
		end
		
		if maxPoint.x < _point.x then
			maxPoint.x = _point.x
		end
		
		if maxPoint.z < _point.y then
			maxPoint.z = _point.y
		end
	end
	
	return 	world.VolumeType.BOX,
			"C_FindContext_search",
			{
				min = minPoint,
				max = maxPoint,
			}
end

ATMEDatas.C_Rectangle_MT = {
	className = "C_Rectangle",
	parentClass = ATME.C_MovableZone2D,
	
	mt = {},
	
	keys = {
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["isInside"] = true,
		["getRandomPosition"] = true,
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["getLength"] = true,
		["getWidth"] = true,
		["getAzimuth"] = true,
		["getCenter"] = true,
		["changeThickness"] = true,
		["getThickness"] = true,
		["isRounded"] = true,
	},
	
	add = function(self, _center, _azimuth, _length, _width, _isRounded)
		local item = {
			center = _center,
			azimuth = _azimuth,
			length = _length,
			width = _width,
			halfLength = _length / 2,
			halfWidth = _width / 2,
			halfLengthMin = 0,
			halfWidthMin = 0,
			thickness = 0,
			isRounded = _isRounded,
			userPrivateDatas = {},
		}
		
		-- item.box est initialisé dans C_Rectangle_create
		
		item.thicknessMax = math.min(item.halfLength, item.halfWidth)
		item.deltaLength = item.halfLength
		item.deltaWidth = item.halfWidth
		
		
		local sCircle = 0
		if _isRounded == true then
			sCircle = math.pi * (item.halfWidth * item.halfWidth)
		end
		
		local sLength = item.deltaWidth * item.length * 2
		local sWidth = item.deltaLength * (item.width - 2 * item.deltaWidth) * 2
		
		local sTotal = sCircle + sLength + sWidth
		
		item.probLength = sLength/sTotal
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].box = nil
	end,
}

ATME.C_Rectangle = {
	getClassName = function(self)
		return ATMEDatas.C_Rectangle_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Rectangle_MT.parentClass
	end,
	
	createNamed = function(_name, _reference, _azimuth, _length, _width, _isRounded)
		local fctDefinition = "Function :\nATME.C_Rectangle.createNamed(name, reference, azimuth, length, width, [isRounded])"

		if type(_name) == "string" and ATMEFct.isReference(_reference) == true and type(_azimuth) == "number" and type(_length) == "number" and type(_width) == "number" and
						(_isRounded == nil or type(_isRounded) == "boolean") then
			local center = ATMEFct.getPointFromReference(_reference)
			if _length > 0 then
				if _width > 0 then
					local isRounded = false
					if type(_isRounded) == "boolean" then
						isRounded = _isRounded
					end
					
					local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
					local err, newRectangle = ATMEFct.C_Zone_createNamed(localMembersModule.zones, _name, ATMEFct.C_Rectangle_create, center, ATMEFct.getAzimuthMod(_azimuth, 360), _length, _width, isRounded)				
			
					if err == true then
						ATMEFct.errorOutputCore(nil, fctDefinition, newRectangle, false)
						return
					end

					return newRectangle
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "width must be > 0", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "length must be > 0", false)
			end
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_azimuth) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
			elseif type(_length) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "length type must be number", false)
			elseif type(_width) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "width type must be number", false)
			elseif _isRounded ~= nil and type(_isRounded) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "isRounded type must be nil or boolean", false)
			end
		end
	end,
	
	getByName = function(_name)	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_Rectangle) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Rectangle.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Rectangle_create = function(_name, _zones, _center, _heading, _length, _width, _isRounded)
	local newRectangle = {}

	setmetatable(newRectangle, ATMEDatas.C_Rectangle_MT.mt)
	local index = ATMEDatas.C_Rectangle_MT:add(_center, _heading, _length, _width, _isRounded)
	local localMembers = ATMEDatas.localMembers[index]

	local name
	local tagsName

	if _name ~= nil and _zones ~= nil then
		name = _name
		_zones[name] = newRectangle
		tagsName = ATMEFct.decompString(_name)
	end

	function newRectangle.getIndex(self)
		return index
	end
	
	function newRectangle.getName(self)
		return name
	end
	
	function newRectangle.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newRectangle) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newRectangle.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newRectangle.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newRectangle) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newRectangle.getLength(self)
		return localMembers.length
	end
	
	function newRectangle.getWidth(self)
		return localMembers.width
	end
	
	function newRectangle.getAzimuth(self)
		return localMembers.azimuth
	end

	function newRectangle.getCenter(self)
		return {
			x = localMembers.center.x,
			y = localMembers.center.y,
			z = localMembers.center.z,
		}
	end
	
	function newRectangle.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
		
	function newRectangle.changeThickness(self, _thickness)
		local fctDefinition = ATMEFct.getErrorPrefix(newRectangle) .. "changeThickness([thickness])"
				
		if _thickness == nil or type(_thickness) == "number" then
			local thickness = 0
			if type(_thickness) == "number" then
				thickness = _thickness
			end
			
			if thickness >= localMembers.thicknessMax then
				thickness = 0
			elseif thickness < 0 then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness must be >= 0", false)
				return
			end
			
			if thickness > 0 then
				localMembers.halfLengthMin = localMembers.halfLength - thickness
				localMembers.halfWidthMin = localMembers.halfWidth - thickness
			else
				localMembers.halfLengthMin = 0
				localMembers.halfWidthMin = 0
			end
			
			localMembers.thickness = thickness
			localMembers.deltaLength = localMembers.halfLength - localMembers.halfLengthMin
			localMembers.deltaWidth = localMembers.halfWidth - localMembers.halfWidthMin
			
			local sLength = localMembers.deltaWidth * localMembers.length * 2
			local sCircle = 0
			local sWidth = 0
			if localMembers.isRounded == true then
				sCircle = math.pi * (localMembers.halfWidth * localMembers.halfWidth - localMembers.halfWidthMin * localMembers.halfWidthMin)
				sWidth = 0
			else
				sCircle = 0
				sWidth = localMembers.deltaLength * (localMembers.width - 2 * localMembers.deltaWidth) * 2
			end
			
			local sTotal = sCircle + sLength + sWidth
			
			localMembers.probLength = sLength/sTotal
		else
			if _thickness ~= nil and type(_thickness) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness type must be nil or number", true)
			end
		end
	end
	
	function newRectangle.getThickness(self)	
		if localMembers.thickness ~= 0 then
			return localMembers.thickness
		end
	end
	
	function newRectangle.isRounded(self)
		return localMembers.isRounded
	end
	
	function newRectangle.isInside(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newRectangle) .. "isInside(reference)"
		local isInside = false
		
		if ATMEFct.isReference(_reference) == true then		
			local point = ATMEFct.getPointFromReference(_reference)
			
			local datas = {
				zone = self,
				actualCenter = localMembers.center,
				actualAzimuth = localMembers.azimuth,
			}
			
			ATMEFct.C_Rectangle_isInside(point, datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
		
		return isInside
	end
	
	function newRectangle.getRandomPosition(self)
		local datas = {
			zone = self,
			actualCenter = localMembers.center,
			actualAzimuth = localMembers.azimuth,
		}

		return ATMEFct.C_Rectangle_getRandomPosition(datas)
	end
	
	local datas = {
		zone = newRectangle,
		actualCenter = _center,
		actualAzimuth = _heading,
	}
	

	local _, _, cbox = ATMEFct.C_Rectangle_searchVolume(datas)
	localMembers.box = cbox
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newRectangle
end

ATMEDatas.C_Rectangle_MT.__call = function(t, _reference, _azimuth, _length, _width, _isRounded)
	local fctDefinition = "Constructor :\nATME.C_Rectangle(reference, azimuth, length, width, [isRounded])"

	if ATMEFct.isReference(_reference) == true and type(_azimuth) == "number" and type(_length) == "number" and type(_width) == "number" and
					(_isRounded == nil or type(_isRounded) == "boolean") then
		local center = ATMEFct.getPointFromReference(_reference)
		if _length > 0 then
			if _width > 0 then
				local isRounded = false
				if type(_isRounded) == "boolean" then
					isRounded = _isRounded
				end

				return ATMEFct.C_Rectangle_create(nil, nil, center, ATMEFct.getAzimuthMod(_azimuth, 360), _length, _width, isRounded)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "width must be > 0", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "length must be > 0", false)
		end
	else
		if ATMEFct.isReference(_reference) == false then
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		elseif type(_azimuth) ~= "number" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
		elseif type(_length) ~= "number" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "length type must be number", false)
		elseif type(_width) ~= "number" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "width type must be number", false)
		elseif _isRounded ~= nil and type(_isRounded) ~= "boolean" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "isRounded type must be nil or boolean", false)
		end
	end
end

setmetatable(ATME.C_Rectangle, ATMEDatas.C_Rectangle_MT)

ATMEDatas.C_Rectangle_MT.mt.__metatable = "Private"
ATMEDatas.C_Rectangle_MT.mt.__index = ATME.C_Rectangle

ATMEDatas.C_Rectangle_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Rectangle_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Rectangle : Bad key " .. k, "table locked", false)
	end
end

ATMEFct.C_Sphere_isInside = function(_point, _datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	
	local dx2 = (_point.x - _datas.actualCenter.x) * (_point.x - _datas.actualCenter.x)
	local dy2 = (_point.y - _datas.actualCenter.y) * (_point.y - _datas.actualCenter.y)
	local dz2 = (_point.z - _datas.actualCenter.z) * (_point.z - _datas.actualCenter.z)
	local d = math.sqrt(dx2 + dy2 + dz2) 
	
	return d >= localMembers.radiusMin and d <= localMembers.radiusMax
end

ATMEFct.C_Sphere_getRandomPosition = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	local angle1 = math.pi - 2 * math.pi * math.random()
	local angle2 = math.pi / 2 - math.pi * math.random()
	local radius = (localMembers.radiusMax - localMembers.radiusMin) * math.random() + localMembers.radiusMin

	return {
		x = _datas.actualCenter.x + radius * math.cos(angle1) * cos(angle2),
		y = _datas.actualCenter.y + radius * math.sin(angle2), 
		z = _datas.actualCenter.z + radius * math.sin(angle1) * cos(angle2),
	}
end

ATMEFct.C_Sphere_duplicate = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	
	local newSphere = ATMEFct.C_Sphere_create(nil, nil, _datas.actualCenter, localMembers.radiusMax)
	
	local localMembersNew = ATMEDatas.localMembers[newSphere:getIndex()]
	localMembersNew.radiusMin = localMembers.radiusMin
	localMembersNew.thickness = localMembers.thickness
	
	return newSphere
end

ATMEFct.C_Sphere_searchVolume = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	
	return 	world.VolumeType.SPHERE, 
			"C_FindContext_search",
			{
				point = {
					x = _datas.actualCenter.x,
					y = _datas.actualCenter.y,
					z = _datas.actualCenter.z,
				},
				radius = localMembers.radiusMax,
			}
end

ATMEDatas.C_Sphere_MT = {
	className = "C_Sphere",
	parentClass = ATME.C_MovableZone3D,
	
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getUserPrivateDatas"] = true,
		["isInside"] = true,
		["getRandomPosition"] = true,
		["getCenter"] = true,
		["getRadius"] = true,
		["changeThickness"] = true,
		["getThickness"] = true,
	},
	
	add = function(self, _center, _radiusMax)
		local item = {
			center = _center,
			radiusMax = _radiusMax,
			radiusMin = 0,
			thickness = 0,
			sphere = {
				point = _center,
				radius = _radiusMax,
			},
			box = {
				min = {
					x = _center.x - _radiusMax,
					y = _center.y - _radiusMax,
					z = _center.z - _radiusMax,
				},
				max = {
					x = _center.x + _radiusMax,
					y = _center.y + _radiusMax,
					z = _center.z + _radiusMax,
				},
			},
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].box = nil
	end,
}

ATME.C_Sphere = {
	getClassName = function(self)
		return ATMEDatas.C_Sphere_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Sphere_MT.parentClass
	end,
	
	createNamed = function(_name, ...)
		local args = {...}
		local fctDefinition = "Function :\nATME.C_Sphere.createNamed(name, DCSZoneName, [altitudeMSL])\nATME.C_Sphere.createNamed(name, reference, radius)"
		local point, radius
		
		if type(name) == "string" then
			if type(args[1]) == "string" then
				local zone = ATMEFct.getDCSZone(args[1])
				
				if zone ~=  nil then
					local altitude = zone.point.y
					
					if #args > 1 then
						if type(args[2]) == "number" then
							altitude = args[2]
							
							if altitude < 0 then
								ATMEFct.errorOutputCore(nil, fctDefinition, "altitude must be >= 0", false)
								return
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "altitude type must be nil or number", false)
							return
						end
					end
					
					point = {
						x = zone.point.x,
						y = altitude,
						z = zone.point.z,
					}

					radius = zone.radius
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. args[1] .. "' unknown", false)
					return
				end
			elseif ATMEFct.isReference(args[1]) == true then
				if type(args[2]) == "number" then
					radius = args[2]
					if radius > 0 then
						point = ATMEFct.getPointFromReference(args[1])
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", false)
					return
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 type must be string," .. ATMEConst.errorLabelReference, false)
				return
			end
			
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newSphere = ATMEFct.C_Zone_createNamed(localMembersModule.zones, _name, ATMEFct.C_Sphere_create, point, radius)				
		
			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, newSphere, false)
				return
			end

			return newSphere
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
		end
	end,
	
	getByName = function(_name)	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_Sphere) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_PickupZone.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Sphere_create = function(_name, _zones, _center, _radiusMax)
	local newSphere = {}

	setmetatable(newSphere, ATMEDatas.C_Sphere_MT.mt)
	local index = ATMEDatas.C_Sphere_MT:add(_center, _radiusMax)
	local localMembers = ATMEDatas.localMembers[index]

	local name
	local tagsName

	if _name ~= nil and _zones ~= nil then
		name = _name
		_zones[name] = newSphere
		tagsName = ATMEFct.decompString(_name)
	end
	
	function newSphere.getIndex(self)
		return index
	end
	
	function newSphere.getName(self)
		return name
	end
	
	function newSphere.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newSphere) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newSphere.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newSphere.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newSphere) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end

	function newSphere.changeThickness(self, _thickness)
		local fctDefinition = ATMEFct.getErrorPrefix(newSphere) .. "changeThickness([thickness])"
				
		if _thickness == nil or type(_thickness) == "number" then
			local thickness = 0
			if type(_thickness) == "number" then
				thickness = _thickness
			end
			
			if thickness >= localMembers.radiusMax then
				thickness = 0
			elseif thickness < 0 then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness must be >= 0", false)
				return
			end
			
			localMembers.thickness = thickness
			
			if thickness == 0 then
				localMembers.radiusMin = 0
			else
				localMembers.radiusMin = localMembers.radiusMax - thickness
			end
		else
			if _thickness ~= nil and type(_thickness) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness type must be nil or number", true)
			end
		end
	end
	
	function newSphere.getThickness(self)	
		if localMembers.thickness ~= 0 then
			return localMembers.thickness
		end
	end
	
	function newSphere.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	function newSphere.isInside(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newSphere) .. "isInside(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATMEFct.getPointFromReference(_reference)
			
			local datas = {
				zone = self,
				actualCenter = localMembers.center,
			}
			
			return ATMEFct.C_Sphere_isInside(point, datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
		
		return false
	end
	
	function newSphere.getRandomPosition(self)
		local datas = {
			zone = self,
			actualCenter = localMembers.center,
		}
		
		return ATMEFct.C_Sphere_getRandomPosition(datas)
	end
	
	function newSphere.getRadius(self)
		return localMembers.radiusMax
	end
	
	function newSphere.getCenter(self)
		return {
			x = localMembers.center.x,
			y = localMembers.center.y,
			z = localMembers.center.z,
		}
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newSphere
end


ATMEDatas.C_Sphere_MT.__call = function(t, ...)
	local args = {...}
	local fctDefinition = "Constructor :\nATME.C_Sphere(DCSZoneName, [altitudeMSL])\nATME.C_Sphere(reference, radius)"
	local point, radius
		
	if type(args[1]) == "string" then
		local zone = ATMEFct.getDCSZone(args[1])
		
		if zone ~=  nil then
			local altitude = zone.point.y
			
			if #args > 1 then
				if type(args[2]) == "number" then
					altitude = args[2]
					
					if altitude < 0 then
						ATMEFct.errorOutputCore(nil, fctDefinition, "altitude must be >= 0", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "altitude type must be nil or number", false)
					return
				end
			end
			
			point = {
				x = zone.point.x,
				y = altitude,
				z = zone.point.z,
			}

			radius = zone.radius
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. args[1] .. "' unknown", false)
			return
		end
	elseif ATMEFct.isReference(args[1]) == true then
		if type(args[2]) == "number" then
			radius = args[2]
			if radius > 0 then
				point = ATMEFct.getPointFromReference(args[1])
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
				return
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", false)
			return
		end
	else
		ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #1 type must be string," .. ATMEConst.errorLabelReference, false)
		return
	end
	
	return ATMEFct.C_Sphere_create(nil, nil, point, radius)
end

setmetatable(ATME.C_Sphere, ATMEDatas.C_Sphere_MT)

ATMEDatas.C_Sphere_MT.mt.__metatable = "Private"
ATMEDatas.C_Sphere_MT.mt.__index = ATME.C_Sphere

ATMEDatas.C_Sphere_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Sphere_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Sphere : Bad key " .. k, "table locked", false)
	end
end


ATMEFct.C_Cylinder_isInside = function(_point, _datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]

	local dx2 = (_point.x - _datas.actualCenter.x) * (_point.x - _datas.actualCenter.x)
	local dy = _point.y - _datas.actualCenter.y
	local dz2 = (_point.z - _datas.actualCenter.z) * (_point.z - _datas.actualCenter.z)
	local d = math.sqrt(dx2 + dz2)
	
	return d >= localMembers.radiusMin and d <= localMembers.radiusMax and math.abs(dy) < localMembers.deltaAltitude
end

ATMEFct.C_Cylinder_getRandomPosition = function(_datas)
	local angle = 2 * math.pi * math.random()
	
	local radius = (localMembers.radiusMax - localMembers.radiusMin) * math.random() + localMembers.radiusMin
	local dAlt = ATMEConst.randomSign[math.random(1,2)] * localMembers.deltaAltitude * math.random()
	
	return {
		x = _datas.actualCenter.x + radius * math.sin(angle),
		y = _datas.actualCenter.y + dAlt,
		z = _datas.actualCenter.z + radius * math.cos(angle),
	}
end

ATMEFct.C_Cylinder_duplicate = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	
	local newCylinder = ATMEFct.C_Cylinder_create(nil, nil, _datas.actualCenter, localMembers.radiusMax, localMembers.deltaAltitude)
	
	local localMembersNew = ATMEDatas.localMembers[newCylinder:getIndex()]
	localMembersNew.radiusMin = localMembers.radiusMin
	localMembersNew.thickness = localMembers.thickness
	localMembersNew.deltaAltitudeMin = localMembers.deltaAltitudeMin
	
	return newCylinder
end

ATMEFct.C_Cylinder_searchVolume = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	
	return world.VolumeType.BOX, 
			"C_FindContext_search",
			{
				min = {
					x = _datas.actualCenter.x - localMembers.radiusMax,
					y = _datas.actualCenter.y - localMembers.deltaAltitude,
					z = _datas.actualCenter.z - localMembers.radiusMax,
				},
				max = {
					x = _datas.actualCenter.x + localMembers.radiusMax,
					y = _datas.actualCenter.y + localMembers.deltaAltitude,
					z = _datas.actualCenter.z + localMembers.radiusMax,
				},
			}
end

ATMEDatas.C_Cylinder_MT = {
	className = "C_Cylinder",
	parentClass = ATME.C_MovableZone3D,
	
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getUserPrivateDatas"] = true,
		["isInside"] = true,
		["getRandomPosition"] = true,
		["getCenter"] = true,
		["getRadius"] = true,
		["getHeight"] = true,
		["changeThickness"] = true,
		["getThickness"] = true,
	},
	
	add = function(self, _center, _radiusMax, _deltaAltitude)
		local item = {
			center = _center,
			radiusMax = _radiusMax,
			radiusMin = 0,
			thickness = 0,
			deltaAltitude = _deltaAltitude,
			deltaAltitudeMin = 0,
			box = {
				min = {
					x = _center.x - _radiusMax,
					y = _center.y - _deltaAltitude,
					z = _center.z - _radiusMax,
				},
				max = {
					x = _center.x + _radiusMax,
					y = _center.y + _deltaAltitude,
					z = _center.z + _radiusMax,
				},
			},
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].box = nil
	end,
}

ATME.C_Cylinder = {
	getClassName = function(self)
		return ATMEDatas.C_Cylinder_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Cylinder_MT.parentClass
	end,
	
	createNamed = function(_name, ...)
		local args = {...}
		local fctDefinition = "Function :\nATME.C_Cylinder.createNamed(name, DCSZoneName, height, [altitudeMSL])\nATME.C_Cylinder.createNamed(name, reference, radius, height)"
		local point, radius, deltaAltitude

		if type(_name) == "string" then
			if type(args[1]) == "string" then
				local zone = ATMEFct.getDCSZone(args[1])
				if zone ~=  nil then
					local altitude = zone.point.y
					
					if type(args[2]) ~= "number" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "height type must be number", false)
						return
					end

					deltaAltitude = args[2] / 2
					if deltaAltitude <= 0 then
						ATMEFct.errorOutputCore(nil, fctDefinition, "height must be > 0", false)
						return
					end
					
					if #args > 2 then
						if type(args[3]) == "number" then
							if args[3] >= 0 then
								altitude = args[3]
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "altitude must be >= 0", false)
								return
							end
						else	
							ATMEFct.errorOutputCore(nil, fctDefinition, "altitude type must be nil or number", false)
							return
						end
					end
					
					point = {
						x = zone.point.x,
						y = altitude,
						z = zone.point.z,
					}

					radius = zone.radius
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. args[1] .. "' unknown", false)
					return
				end
			elseif ATMEFct.isReference(args[1]) == true then
				if type(args[2]) == "number" then
					radius = args[2]
					if radius > 0 then
						point = ATMEFct.getPointFromReference(args[1])
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
						return
					end
					
					if type(args[3]) ~= "number" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "deltaAltitude type must be number", false)
						return
					end
					
					deltaAltitude = args[3] / 2
					if deltaAltitude <= 0 then
						ATMEFct.errorOutputCore(nil, fctDefinition, "deltaAltitude must be > 0", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", false)
					return
				end
			else

				ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 type must be string," .. ATMEConst.errorLabelReference, false)
				return
			end
			
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newCylinder = ATMEFct.C_Zone_createNamed(localMembersModule.zones, _name, ATMEFct.C_Cylinder_create, point, radius, deltaAltitude)				
		
			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, newCylinder, false)
				return
			end

			return newCylinder
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
		end
	end,
	
	getByName = function(_name)	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_Cylinder) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Cylinder.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Cylinder_create = function(_name, _zones, _center, _radiusMax, _deltaAltitude)
	local newCylinder = {}
	
	setmetatable(newCylinder, ATMEDatas.C_Cylinder_MT.mt)
	local index = ATMEDatas.C_Cylinder_MT:add(_center, _radiusMax, _deltaAltitude)
	local localMembers = ATMEDatas.localMembers[index]

	local name
	local tagsName

	if _name ~= nil and _zones ~= nil then
		name = _name
		_zones[name] = newCylinder
		tagsName = ATMEFct.decompString(_name)
	end
	
	function newCylinder.getIndex(self)
		return index
	end
	
	function newCylinder.getName(self)
		return name
	end
	
	function newCylinder.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newCylinder) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newCylinder.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newCylinder.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newCylinder) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newCylinder.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	function newCylinder.changeThickness(self, _thickness)
		local fctDefinition = ATMEFct.getErrorPrefix(newCylinder) .. "changeThickness([thickness])"
				
		if _thickness == nil or type(_thickness) == "number" then
			local thickness = 0
			if type(_thickness) == "number" then
				thickness = _thickness
			end
			
			
			if thickness >= localMembers.radiusMax then
				thickness = 0
			elseif thickness < 0 then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness must be >= 0", false)
				return
			end
			
			localMembers.thickness = thickness
			
			if thickness == 0 then
				localMembers.radiusMin = 0
				localMembers.deltaAltitudeMin = 0
			else
				localMembers.radiusMin = localMembers.radiusMax - thickness
				localMembers.deltaAltitudeMin = localMembers.deltaAltitude - thickness
			end
		else
			if _thickness ~= nil and type(_thickness) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness type must be nil or number", true)
			end
		end
	end
	
	function newCylinder.getThickness(self)	
		if localMembers.thickness ~= 0 then
			return localMembers.thickness
		end
	end
	
	function newCylinder.isInside(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newCylinder) .. "isInside(reference)"
		
		if ATMEFct.isReference(_reference) == true then
			local point = ATMEFct.getPointFromReference(_reference)
			
			local datas = {
				zone = self,
				actualCenter = localMembers.center,
			}
			
			return ATMEFct.C_Cylinder_isInside(point, datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
		
		return false
	end
	
	function newCylinder.getRandomPosition(self)
		local datas = {
			zone = self,
			actualCenter = localMembers.center,
		}
			
		return ATMEFct.C_Cylinder_getRandomPosition(datas)
	end
	
	function newCylinder.getRadius(self)
		return localMembers.radiusMax
	end
	
	function newCylinder.getCenter(self)
		return {
			x = localMembers.center.x,
			y = localMembers.center.y,
			z = localMembers.center.z,
		}
	end
	
	function newCylinder.getHeight(self)
		return localMembers.deltaAltitude * 2
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end

	return newCylinder
end


ATMEDatas.C_Cylinder_MT.__call = function(t, ...)
	local args = {...}
	local fctDefinition = "Constructor :\nATME.C_Cylinder(DCSZoneName, height, [altitudeMSL])\nATME.C_Cylinder(reference, radius, height)"
	local point, radius, deltaAltitude

	if type(args[1]) == "string" then
		local zone = ATMEFct.getDCSZone(args[1])
		if zone ~=  nil then
			local altitude = zone.point.y
			
			if type(args[2]) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "height type must be number", false)
				return
			end

			deltaAltitude = args[2] / 2
			if deltaAltitude <= 0 then
				ATMEFct.errorOutputCore(nil, fctDefinition, "height must be > 0", false)
				return
			end
			
			if #args > 2 then
				if type(args[3]) == "number" then
					if args[3] >= 0 then
						altitude = args[3]
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "altitude must be >= 0", false)
						return
					end
				else	
					ATMEFct.errorOutputCore(nil, fctDefinition, "altitude type must be nil or number", false)
					return
				end
			end
			
			point = {
				x = zone.point.x,
				y = altitude,
				z = zone.point.z,
			}

			radius = zone.radius
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. args[1] .. "' unknown", false)
			return
		end
	elseif ATMEFct.isReference(args[1]) == true then
		if type(args[2]) == "number" then
			radius = args[2]
			if radius > 0 then
				point = ATMEFct.getPointFromReference(args[1])
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
				return
			end
			
			if type(args[3]) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "deltaAltitude type must be number", false)
				return
			end
			
			deltaAltitude = args[3] / 2
			if deltaAltitude <= 0 then
				ATMEFct.errorOutputCore(nil, fctDefinition, "deltaAltitude must be > 0", false)
				return
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", false)
			return
		end
	else

		ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #1 type must be string," .. ATMEConst.errorLabelReference, false)
		return
	end
	
	return ATMEFct.C_Cylinder_create(nil, nil, point, radius, deltaAltitude)
end

setmetatable(ATME.C_Cylinder, ATMEDatas.C_Cylinder_MT)

ATMEDatas.C_Cylinder_MT.mt.__metatable = "Private"
ATMEDatas.C_Cylinder_MT.mt.__index = ATME.C_Cylinder

ATMEDatas.C_Cylinder_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Cylinder_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Cylinder : Bad key " .. k, "table locked", false)
	end
end

ATMEFct.C_Box_isInside = function(_point, _datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	local dx = _point.x - _datas.actualCenter.x
	local dy = _point.y - _datas.actualCenter.y
	local dz = _point.z - _datas.actualCenter.z
	
	local angleRad = math.rad(_datas.actualAzimuth)
	
	local lPoint = {
		x = dx * math.cos(angleRad) + dz * math.sin(angleRad),
		y = -dx * math.sin(angleRad) + dz * math.cos(angleRad),
	}

	local ok = (math.abs(lPoint.x) >= localMembers.halfLengthMin and math.abs(lPoint.x) <= localMembers.halfLength and 
					math.abs(lPoint.y) <= localMembers.halfWidth and localMembers.isRounded == false) or
			 (math.abs(lPoint.y) >= localMembers.halfWidthMin and math.abs(lPoint.y) <= localMembers.halfWidth and 
					math.abs(lPoint.x) <= localMembers.halfLength)
					
	if ok == false and localMembers.isRounded == true then	
		local dxPoint = lPoint.x - localMembers.halfLength
		local dyPoint = lPoint.y
		local d = math.sqrt(dxPoint * dxPoint + dyPoint * dyPoint)
		local angle = math.atan2(dyPoint, dxPoint)
	
		ok = angle > -math.pi / 2 and angle < math.pi / 2 and d >= localMembers.halfWidthMin and d <= localMembers.halfWidth
		
		if ok == false then
			dxPoint = lPoint.x + localMembers.halfLength
			local d = math.sqrt(dxPoint * dxPoint + dyPoint * dyPoint)
			local angle = math.atan2(dyPoint, dxPoint)
		
			ok = (angle < -math.pi / 2 or angle > math.pi / 2) and d >= localMembers.halfWidthMin and d <= localMembers.halfWidth
		end
	end
	
	dy = lPoint.x * localMembers.a - dy
	
	return ok == true and math.abs(dy) <= localMembers.deltaAltitude
end

ATMEFct.C_Box_getRandomPosition = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	local angleRad = math.rad(_datas.actualAzimuth)
	local point
	
	local rand = math.random()
	if rand < localMembers.probLength then
		-- Soit la surface correspondant au grand coté avec l'épaisseur
		
		local dx = ATMEConst.randomSign[math.random(1,2)] * localMembers.halfLength * math.random()
		local dy = ATMEConst.randomSign[math.random(1,2)] * (localMembers.deltaWidth * math.random() + localMembers.halfWidthMin)
		local dAlt = ATMEConst.randomSign[math.random(1,2)] * localMembers.deltaAltitude * math.random()
		
		return {
			x = _datas.actualCenter.x + dx * math.cos(angleRad) - dy * math.sin(angleRad),
			y = _datas.actualCenter.y + dAlt - dx * localMembers.a,
			z = _datas.actualCenter.z + dx * math.sin(angleRad) + dy * math.cos(angleRad),
		}
	else
		-- Soit la surface correspondant au petit coté moins l'épaisseur
		-- Soit sur 2x le demi-cercle de part et d'autre du centre sur l'axe x
		
		local dAlt = ATMEConst.randomSign[math.random(1,2)] * localMembers.deltaAltitude * math.random()
		
		if localMembers.isRounded == true then
			local angle = 2 * math.pi * math.random()
			local radius = localMembers.deltaWidth * math.random() + localMembers.halfWidthMin
			local i = 1
			
			if angle >= math.pi then
				i = -1
			end
			
			local localPoint = {
				x = i * localMembers.halfLength + radius * math.sin(angle),
				y = radius * math.cos(angle),
			}
			
			return {
				x = _datas.actualCenter.x + localPoint.x * math.cos(angleRad) - localPoint.y * math.sin(angleRad),
				y = _datas.actualCenter.y + dAlt - localPoint.x * localMembers.a,
				z = _datas.actualCenter.z + localPoint.x * math.sin(angleRad) + localPoint.y * math.cos(angleRad),
			}
		else
			local dx = ATMEConst.randomSign[math.random(1,2)] * (localMembers.deltaLength * math.random() + localMembers.halfLengthMin)
			local dy = ATMEConst.randomSign[math.random(1,2)] * localMembers.halfWidthMin * math.random()
			
			return {
				x = _datas.actualCenter.x + dx * math.cos(angleRad) - dy * math.sin(angleRad),
				y = _datas.actualCenter.y + dAlt - dx * localMembers.a,
				z = _datas.actualCenter.z + dx * math.sin(angleRad) + dy * math.cos(angleRad),
			}
		end
	end
end

ATMEFct.C_Box_duplicate = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	
	local newBox = ATMEFct.C_Box_create(nil, nil, _datas.actualCenter, _datas.actualAzimuth, localMembers.length, 
										localMembers.width, localMembers.deltaAltitude, localMembers.glide,
										localMembers.isRounded)
	
	local localMembersNew = ATMEDatas.localMembers[newBox:getIndex()]
	localMembersNew.halfLengthMin = localMembers.halfLengthMin
	localMembersNew.halfWidthMin = localMembers.halfWidthMin
	localMembersNew.deltaAltitudeMin = localMembers.deltaAltitudeMin
	localMembersNew.thickness = localMembers.thickness
	localMembersNew.deltaLength = localMembers.deltaLength
	localMembersNew.deltaWidth = localMembers.deltaWidth
	localMembersNew.thicknessMax = localMembers.thicknessMax
	localMembersNew.probLength = localMembers.probLength
	
	return newBox
end

ATMEFct.C_Box_searchVolume = function(_datas)
	local localMembers = ATMEDatas.localMembers[_datas.zone:getIndex()]
	local lPoints = {}
	local points = {}
	
	local angleRad = math.rad(_datas.actualAzimuth)

	local length = localMembers.halfLength
	if localMembers.isRounded == true then
		length = length + localMembers.halfWidth
	end

	table.insert(lPoints, {x = length, y = localMembers.halfWidth})
	table.insert(lPoints, {x = -length, y = localMembers.halfWidth})
	table.insert(lPoints, {x = -length, y = -localMembers.halfWidth})
	table.insert(lPoints, {x = length, y = -localMembers.halfWidth})

	for _index, _point in ipairs(lPoints) do
		points[_index] =  {
			x = _datas.actualCenter.x + _point.x * math.cos(angleRad) - _point.y * math.sin(angleRad),
			y = _datas.actualCenter.z + _point.x * math.sin(angleRad) + _point.y * math.cos(angleRad),
		}
	end
	
	local dy = math.abs(localMembers.halfLength * localMembers.a)

	local minPoint = {
		x = points[1].x,
		y = _datas.actualCenter.y - localMembers.deltaAltitude - dy,
		z = points[1].y,
	}
		
	local maxPoint = {
		x = points[1].x,
		y = _datas.actualCenter.y + localMembers.deltaAltitude + dy,
		z = points[1].y,
	}

	for _, _point in ipairs(points) do
		if minPoint.x > _point.x then
			minPoint.x = _point.x
		end
		
		if minPoint.z > _point.y then
			minPoint.z = _point.y
		end
		
		if maxPoint.x < _point.x then
			maxPoint.x = _point.x
		end
		
		if maxPoint.z < _point.y then
			maxPoint.z = _point.y
		end
	end

	return  world.VolumeType.BOX,
			"C_FindContext_search",
			{
				min = minPoint,
				max = maxPoint,
			}
end

ATMEDatas.C_Box_MT = {
	className = "C_Box",
	parentClass = ATME.C_MovableZone3D,
	
	mt = {},
	
	keys = {
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["isInside"] = true,
		["getRandomPosition"] = true,
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["getLength"] = true,
		["getWidth"] = true,
		["getAzimuth"] = true,
		["getCenter"] = true,
		["getGlide"] = true,
		["getHeight"] = true,
		["changeThickness"] = true,
		["getThickness"] = true,
		["isRounded"] = true,
	},
	
	add = function(self, _center, _azimuth, _length, _width, _deltaAltitude, _glide, _isRounded)
		local lPoints = {}
		local points = {}
		local glideRadValue = math.rad(_glide)
		
		local item = {
			points = _points,
			localPoints = lPoints,
			a = math.sin(glideRadValue),
			center = _center,
			azimuth = _azimuth,
			glideRad = glideRadValue,
			glide = _glide,
			length = _length,
			width = _width,
			halfLength = _length / 2,
			halfWidth = _width / 2,
			halfLengthMin = 0,
			halfWidthMin = 0,
			deltaAltitude = _deltaAltitude,
			deltaAltitudeMin = 0,
			thickness = 0,
			isRounded = _isRounded,
			userPrivateDatas = {},
		}
		
		-- item.box est initialisé dans C_Box_create
		
		item.thicknessMax = math.min(item.halfLength, item.halfWidth)
		item.deltaLength = item.halfLength - item.halfLengthMin
		item.deltaWidth = item.halfWidth - item.halfWidthMin
		
		local vCircle = 0
		local vWidth = 0
		if _isRounded == true then
			vCircle = math.pi * (item.halfWidth * item.halfWidth)
		else
			vWidth = item.deltaLength * (item.width - 2 * item.deltaWidth) * 2
		end
		
		local vLength = item.deltaWidth * item.length * 2
		
		local vTotal = vCircle + vLength + vWidth
		
		item.probLength = vLength/vTotal
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].box = nil
	end,
}

ATME.C_Box = {
	getClassName = function(self)
		return ATMEDatas.C_Box_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Box_MT.parentClass
	end,
	
	createNamed = function(_name, _reference, _azimuth, _length, _width, _height, ...)
		local fctDefinition = "Function :\nATME.C_Box.createNamed(name, reference, azimuth, length, width, height)\nATME.C_Box.createNamed(name, reference, azimuth, length, width, height, glide)\nATME.C_Box.createNamed(name, reference, azimuth, length, width, height, isRounded)ATME.C_Box(reference, azimuth, length, width, height, glide, isRounded)"
		local args = {...}
		
		if type(_name) == "string" and ATMEFct.isReference(_reference) == true and type(_azimuth) == "number" and  
									type(_length) == "number" and type(_width) == "number" and 
									type(_height) == "number" then
			local center = ATMEFct.getPointFromReference(_reference)

			if _length > 0 then
				if _width > 0 then
					if _height > 0 then
						local glide = 0
						local isRounded = false

						if #args == 1 then
							if type(args[1]) == "number" then
								glide = ATMEFct.getAzimuthMod(args[1], 180)
								if glide <= -90 or glide >= 90 then
									ATMEFct.errorOutputCore(nil, fctDefinition, "glide must be > -90 and < 90", false)
									return
								end
							elseif type(args[1]) == "boolean" then
								isRounded = args[1]
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #6 type must be nil, number or boolean", false)
								return
							end
						elseif #args > 1 then
							if type(args[1]) == "number" then
								glide = ATMEFct.getAzimuthMod(args[1], 180)
								
								if glide <= -90 or glide >= 90 then
									ATMEFct.errorOutputCore(nil, fctDefinition, "glide must be > -90 and < 90", false)
									return
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "glide type must be nil or number", false)
								return
							end
							
							if type(args[2]) == "boolean" then
								isRounded = args[2]
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "isRounded type must be nil or boolean", false)
								return
							end

						end
						
						local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
						local err, newBox = ATMEFct.C_Zone_createNamed(localMembersModule.zones, _name, ATMEFct.C_Box_create, center, ATMEFct.getAzimuthMod(_azimuth, 360), _length, _width, _height / 2, glide, isRounded)				
			
						if err == true then
							ATMEFct.errorOutputCore(nil, fctDefinition, newBox, false)
							return
						end

						return newBox
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "height must be > 0", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "width must be > 0", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "length must be > 0", false)
			end
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif ATMEFct.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
			elseif type(_azimuth) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
			elseif type(_length) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "length type must be number", false)
			elseif type(_width) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "width type must be number", false)
			elseif type(_height) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "height type must be number", false)
			end
		end
	end,
	
	getByName = function(_name)	
		if type(_name) == "string" then
			if ATMEFct.isObjectClass(ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name], ATME.C_Box) == true then
				return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zones[_name]
			end
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_PickupZone.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Box_create = function(_name, _zones, _center, _heading, _length, _width, _deltaAltitude, _glide, _isRounded)
	local newBox = {}
	
	setmetatable(newBox, ATMEDatas.C_Box_MT.mt)
	local index = ATMEDatas.C_Box_MT:add(_center, _heading, _length, _width, _deltaAltitude, _glide, _isRounded)
	local localMembers = ATMEDatas.localMembers[index]
			
	local name
	local tagsName

	if _name ~= nil and _zones ~= nil then
		name = _name
		_zones[name] = newBox
		tagsName = ATMEFct.decompString(_name)
	end
	
	function newBox.getIndex(self)
		return index
	end
	
	function newBox.getName(self)
		return name
	end
	
	function newBox.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newBox) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newBox.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newBox.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newBox) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newBox.getLength(self)
		return localMembers.length
	end
	
	function newBox.getHeight(self)
		return localMembers.deltaAltitude * 2
	end
	
	function newBox.getWidth(self)
		return localMembers.width
	end
	
	function newBox.getAzimuth(self)
		return localMembers.azimuth
	end

	function newBox.getGlide(self)
		return localMembers.glide
	end
	
	function newBox.getCenter(self)
		return {
			x = localMembers.center.x,
			y = localMembers.center.y,
			z = localMembers.center.z,
		}
	end
	
	function newBox.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	function newBox.changeThickness(self, _thickness)
		local fctDefinition = ATMEFct.getErrorPrefix(newBox) .. "changeThickness([thickness])"
				
		if _thickness == nil or type(_thickness) == "number" then
			local thickness = 0
			if type(_thickness) == "number" then
				thickness = _thickness
			end
			
			
			if thickness >= localMembers.thicknessMax then
				thickness = 0
			elseif thickness < 0 then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness must be >= 0", false)
				return
			end
			
			localMembers.thickness = thickness
			
			if thickness > 0 then
				localMembers.halfLengthMin = localMembers.halfLength - thickness
				localMembers.halfWidthMin = localMembers.halfWidth - thickness
				localMembers.deltaAltitudeMin = localMembers.deltaAltitude - thickness
			else
				localMembers.halfLengthMin = 0
				localMembers.halfWidthMin = 0
				localMembers.deltaAltitudeMin = 0
			end
			
			localMembers.deltaLength = localMembers.halfLength - localMembers.halfLengthMin
			localMembers.deltaWidth = localMembers.halfWidth - localMembers.halfWidthMin
			
			local vLength = localMembers.deltaWidth * localMembers.length * 2 * localMembers.deltaAltitude
			local vCircle = 0
			local vWidth = 0
			if localMembers.isRounded == true then
				vCircle = math.pi * (localMembers.halfWidth * localMembers.halfWidth - localMembers.halfWidthMin * localMembers.halfWidthMin) * localMembers.deltaAltitude
			else
				vWidth = localMembers.deltaLength * (localMembers.width - 2 * localMembers.deltaWidth) * 2 * localMembers.deltaAltitude
			end
			
			local vTotal = vCircle + vLength + vWidth
			
			localMembers.probLength = vLength/vTotal
		else
			if _thickness ~= nil and type(_thickness) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "thickness type must be nil or number", true)
			end
		end
	end
	
	function newBox.getThickness(self)	
		if localMembers.thickness ~= 0 then
			return localMembers.thickness
		end
	end
	
	function newBox.isRounded(self)
		return localMembers.isRounded
	end
	
	function newBox.isInside(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newBox) .. "isInside(reference)"
		local isInside = false
		
		if ATMEFct.isReference(_reference) == true then		
			local point = ATMEFct.getPointFromReference(_reference)
			
			local datas = {
				zone = self,
				actualCenter = localMembers.center,
				actualAzimuth = localMembers.azimuth,
			}
			
			ATMEFct.C_Box_isInside(point, datas)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
		
		return isInside
	end
	
	function newBox.getRandomPosition(self)
		local datas = {
			zone = self,
			actualCenter = localMembers.center,
			actualAzimuth = localMembers.azimuth,
		}
		
		return ATMEFct.C_Box_getRandomPosition(datas)
	end
	
	local datas = {
			zone = newBox,
			actualCenter = _center,
			actualAzimuth = _heading,
		}
	
	local _, _, cbox = ATMEFct.C_Box_searchVolume(datas)
	localMembers.box = cbox
		
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newBox
end

ATMEDatas.C_Box_MT.__call = function(t, _reference, _azimuth, _length, _width, _height, ...)
	local fctDefinition = "Constructor :\nATME.C_Box(reference, azimuth, length, width, height)\nATME.C_Box(reference, azimuth, length, width, height, glide)\nATME.C_Box(reference, azimuth, length, width, height, isRounded)ATME.C_Box(reference, azimuth, length, width, height, glide, isRounded)"
	local args = {...}
	
	if ATMEFct.isReference(_reference) == true and type(_azimuth) == "number" and  
								type(_length) == "number" and type(_width) == "number" and 
								type(_height) == "number" then
		local center = ATMEFct.getPointFromReference(_reference)

		if _length > 0 then
			if _width > 0 then
				if _height > 0 then
					local glide = 0
					local isRounded = false

					if #args == 1 then
						if type(args[1]) == "number" then
							glide = ATMEFct.getAzimuthMod(args[1], 180)
							if glide <= -90 or glide >= 90 then
								ATMEFct.errorOutputCore(nil, fctDefinition, "glide must be > -90 and < 90", false)
								return
							end
						elseif type(args[1]) == "boolean" then
							isRounded = args[1]
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #6 type must be nil, number or boolean", false)
							return
						end
					elseif #args > 1 then
						if type(args[1]) == "number" then
							glide = ATMEFct.getAzimuthMod(args[1], 180)
							
							if glide <= -90 or glide >= 90 then
								ATMEFct.errorOutputCore(nil, fctDefinition, "glide must be > -90 and < 90", false)
								return
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "glide type must be nil or number", false)
							return
						end
						
						if type(args[2]) == "boolean" then
							isRounded = args[2]
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "isRounded type must be nil or boolean", false)
							return
						end

					end
					
					return ATMEFct.C_Box_create(nil, nil, center, ATMEFct.getAzimuthMod(_azimuth, 360), _length, _width, _height / 2, glide, isRounded)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "height must be > 0", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "width must be > 0", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "length must be > 0", false)
		end
	else
		if ATMEFct.isReference(_reference) == false then
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		elseif type(_azimuth) ~= "number" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
		elseif type(_length) ~= "number" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "length type must be number", false)
		elseif type(_width) ~= "number" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "width type must be number", false)
		elseif type(_height) ~= "number" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "height type must be number", false)
		end
	end
end

setmetatable(ATME.C_Box, ATMEDatas.C_Box_MT)

ATMEDatas.C_Box_MT.mt.__metatable = "Private"
ATMEDatas.C_Box_MT.mt.__index = ATME.C_Box

ATMEDatas.C_Box_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Box_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Box : Bad key " .. k, "table locked", false)
	end
end


----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Airbase : gère les bases aériennes (aéroports, bateaux ou FARP). 
--                         Pas de contructeur accessible aux modules externes.
-- 
-- Deux listes globales sont créés et initialisées :
--			-> airbasesById pour le référencement par ID (number pour les bases de la carte, string sinon)
--			-> airbases pour le référencement par nom
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Airbase_MT = {
	className = "C_Airbase",
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["getName"] = true,
		["getCategoryName"] = true,
		["getCoalitionName"] = true,
		["getPosition"] = true,
		["getDCSHelipadId"] = true,
		["getDCSAirdromeId"] = true,
		["getDCSAirbase"] = true,
		["getDCSID"] = true,
		["getAttributes"] = true,
		["hasAttributes"] = true,
		["isHelipad"] = true,
		["isInZones"] = true,
		["isInDCSZone"] = true,
		["isInZone2D"] = true,
		["isInZone3D"] = true,
		["getParking"] = true,
	},
	
	add = function(self, _airbase, _categoryName, _id)
		-- Bug DCS V2.5.6
		
		local id = _id
		if _id == nil then
			id = _airbase:getID()
			
			if type(id) == "string" then
				id = tonumber(id)
			end
		end
		
		local item = {
			objectDCS = _airbase,
			objectDCSId = id,
			categoryName = _categoryName,
			userPrivateDatas = {},
			destroyed = false,
		}

		if _id == nil then
			item.helipadId = id
		else
			item.airdromeId = id
		end
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Airbase = {
	getClassName = function(self)
		return ATMEDatas.C_Airbase_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Airbase_MT.parentClass
	end,
	
	getByName = function(_name, _categoryName)
		if type(_name) == "string" and (_categoryName == nil or type(_categoryName) == "string") then
			local categoryName = _categoryName
			
			if categoryName == nil then
				return ATMEDatas.airbases["AIRBASE"][_name]
			elseif type(categoryName) == "string" then
				if categoryName ~= "SHIP" and categoryName ~= "HELIPAD" and categoryName ~= "AIRBASE" then
					ATMEFct.errorOutputCore(nil, "Function ATME.C_Airbase.getByName(name, [categoryName])", "categoryName must be 'SHIP', 'HELIPAD' or 'AIRBASE'", false)
				else
					return ATMEDatas.airbases[categoryName][_name]
				end
			else
				ATMEFct.errorOutputCore(nil, "Function ATME.C_Airbase.getByName(name, [categoryName])", "categoryName type must be string", false)
			end
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, "Function ATME.C_Airbase.getByName(name, [categoryName])", "name type must be string", false)
			elseif _categoryName ~= nil and type(_categoryName) ~= "string" then
				ATMEFct.errorOutputCore(nil, "Function ATME.C_Airbase.getByName(name, [categoryName])", "categoryName type must be nil or string", false)
			end
		end
	end,
	
	getByID = function(_id)
		if type(_id) == "string" or type(_id) == "number" then
			local id = _id
			if type(_id) == "number" then
				id = math.floor(_id)
			end
			
			return ATMEDatas.airbasesById[id]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Airbase.getByID(id)", "id type must be string or number", false)
		end
	end,
}

ATMEFct.C_Airbase_create_basic = function(_categoryName, _helicoptersOnly, _airbase, _id)
	local name = _airbase:getName()
	local newAirbase = {}
	local position = _airbase:getPosition().p

	local helicoptersOnly = _helicoptersOnly

	setmetatable(newAirbase, ATMEDatas.C_Airbase_MT.mt)
	local index = ATMEDatas.C_Airbase_MT:add(_airbase, _categoryName, _id)
	local localMembers = ATMEDatas.localMembers[index]

	ATMEDatas.airbasesById[localMembers.objectDCSId] = newAirbase
	ATMEDatas.airbases[localMembers.categoryName][name] = newAirbase
	
	function newAirbase.getIndex(self)
		return index
	end
	
	function newAirbase.getName(self)
		if localMembers.destroyed == false then
			return name
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newAirbase.getDCSHelipadId(self)
		if localMembers.destroyed == false then
			return localMembers.helipadId
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newAirbase.getDCSAirdromeId(self)
		if localMembers.destroyed == false then
			return localMembers.airdromeId
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newAirbase.getDCSAirbase(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newAirbase.getDCSID(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCSId
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newAirbase.getParking(self, _available)
		local fctDefinition = ATMEFct.getErrorPrefix(newAirbase) .. "getParking(available)"
		
		if localMembers.destroyed == false then
			if type(_available) == nil or type(_available) == "string" then
				local available = true
				
				if type(_available) == "string" then
					available = _available
				end
				
				return localMembers.objectDCS:getParking(available)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zonesMgr type must be ATME.C_ZonesMgr", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newAirbase.getUserPrivateDatas(self)
		if localMembers.destroyed == false then
			return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
		
	function newAirbase.isHelipad(self)
		if localMembers.destroyed == false then
			return helicoptersOnly
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newAirbase.getCoalitionName(self)
		if localMembers.destroyed == false then
			return ATMEConst.coalitionNames[localMembers.objectDCS:getCoalition()]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newAirbase.getCategoryName(self)
		if localMembers.destroyed == false then
			return localMembers.categoryName
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newAirbase.getPosition(self)
		if localMembers.destroyed == false then
			if localMembers.categoryName == "SHIP" then
				if ATMEDatas.AIUnits[name] ~= nil then
					return ATMEDatas.AIUnits[name]:getPosition()
				end
			else
				return position
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newAirbase.isInZones(self, _zonesMgr)
		if localMembers.destroyed == false then
			local fctDefinition = ATMEFct.getErrorPrefix(newAirbase) .. "isInZones(zonesMgr)"
			
			if ATMEFct.isObjectClass(_zonesMgr, ATME.C_ZonesMgr) == true then
				return _zonesMgr:isInside(self) 
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zonesMgr type must be ATME.C_ZonesMgr", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end				
	
	function newAirbase.isInDCSZone(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newAirbase) .. "isInDCSZone(name)"
		
		if localMembers.destroyed == false then
			if type(_name) == "string" then
				local zone = ATMEFct.getDCSZone(_name)
				
				if zone ~= nil then
					local vector = ATMEFct.C_Vector3D_create(nil, nil, self:getPosition(), zone.point)
					if vector:getHModule() <= zone.radius then
						return true
					else
						return false
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _name .. " unknown", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newAirbase.isInZone2D(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newAirbase) .. "isInZone2D(zone)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zone, ATME.C_Zone2D) == true then
				local point = self:getPosition()
				local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
				
				local datas = {
					zone = _zone,
					actualCenter = localMembersZone.center,
					azimuth = localMembersZone.azimuth,
				}
			
				return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](point, datas)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Circle, ATME.C_PickupZone, ATME.C_Rectangle or ATME.C_Polygon", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newAirbase.isInZone3D(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newAirbase) .. "isInZone3D(zone)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zone, ATME.C_Zone3D) == true then
				local point = self:getPosition()
				local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
		
				local datas = {
					zone = _zone,
					actualCenter = localMembersZone.center,
					azimuth = localMembersZone.azimuth,
				}
				
				return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](point, datas)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Sphere, ATME.C_Cylinder or ATME.C_Box", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newAirbase.getAttributes(self)
		if localMembers.destroyed == false then
			local datas = Airbase.getDesc(localMembers.objectDCS)
			local attributes = {}
			
			for _id, _ in pairs(datas.attributes) do
				if ATMEConst.backTargetsAttributes[_id] ~= nil then
					table.insert(attributes, ATMEConst.backTargetsAttributes[_id].value)
				end
			end
			
			return attributes
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newAirbase.hasAttributes(self, _attributes)
		if localMembers.destroyed == false then
			local fctDefinition = ATMEFct.getErrorPrefix(newAirbase) .. "hasAttributes(attributes)"
			
			if type(_attributes) == "string" or type(_attributes) == "table" then
				local err, datas = ATMEFct.extractDatasFromStringOrTable(_attributes, false)
		
				if err == false then
					return ATMEFct.hasAttributes(localMembers.objectDCS, datas)
				end
				
				ATMEFct.errorOutputCore(nil, fctDefinition, datas, false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "attributes type must be string or table", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end

	ATMEFct.outputCore(nil, "Création entrée airbasesById .. " .. name .. " " .. localMembers.categoryName .. " -> " .. localMembers.objectDCSId .. "(type " .. type(localMembers.objectDCSId) .. ") " .. newAirbase:getCoalitionName(),2)
	
	return newAirbase
end

ATMEFct.C_Airbase_delete = function(_unit)
	if ATMEDatas.airbasesById[_unit:getID()] ~= nil then
		-- L'unité détruite est un FARP ou un bateau avec FARP ou porte-avion
		localMembers.destroyed = true
		ATMEDatas.airbasesById[_unit:getID()] = nil
		
		ATMEFct.outputCore(nil, "Suppression entrée airbasesById .. " .. _unit:getName(), 2)
	end
end

setmetatable(ATME.C_Airbase,ATMEDatas.C_Airbase_MT)

ATMEDatas.C_Airbase_MT.mt.__metatable = "Private"
ATMEDatas.C_Airbase_MT.mt.__index = ATME.C_Airbase
ATMEDatas.C_Airbase_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Airbase_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Airbase : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_ZonesMgr : Gère les ensembles de zones DCS ou zones 2D/3D. 
-- Contructeur accessible aux modules externes.
-- Cette classe permet d'étendre la notion de zones par l'union de zones multiples.
-- Une zone peut être liée à un objet pour les classes dérivées de ATME.C_MovableZone2D 
-- et ATME.C_MovableZone3D.
----------------------------------------------------------------------------------------------------------

ATMEFct.C_ZonesMgr_isInside = function(_zonesMgr, _point)
	local localMembersZonesMgr = ATMEDatas.localMembers[_zonesMgr:getIndex()]
	local isInside = false
	local zones = {}
	
	-- La table ATMEConst.indZoneConfig optimise les tests, les appels de fonctions
	-- et le passage de paramètres.
	
	if #localMembersZonesMgr.zonesList > 0 then
		for _id, _item in ipairs(localMembersZonesMgr.zonesList) do
			local localMembers = ATMEDatas.localMembers[_item.zone:getIndex()]
			local datas = {
				zone = _item.zone,
				actualCenter = ATMEFct.getActualCenter(localMembers.center, _item.linkedUnitDatas),
				actualAzimuth = ATMEFct.getActualAzimuth(localMembers.azimuth, _item.linkedUnitDatas),
			}
			
			if ATMEFct[ATMEConst.indZoneConfig[_item.zone:getClassName()].fctIsInside](_point, datas) == true then
				isInside = true
				table.insert(zones, _id)
			end
		end
	end
	
	return isInside, zones
end

ATMEFct.C_ZonesMgr_createNamed = function(_namedTable, _name)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return false
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			return false, ATMEFct.C_ZonesMgr_create(_name, _namedTable)
		else
			return true, "name can't contain '|' character"
		end
	end
end

ATMETimers.C_ZonesMgr_timer = function(_datas)
	if _datas.tracking.localMembersObject.objectDCS:isExist() == true then
		ATMEFct.outputCore(nil, "Tracking sur gestionnaire de zone ".. _datas.tracking.name .. " - classe " .. _datas.tracking.className , 2)
		
		if _datas.tracking.object:isActivated() == true then
			local isInside
			local zonesId
			isInside, zonesId = ATMEFct.C_ZonesMgr_isInside(_datas.zonesMgr, _datas.tracking.object:getPosition())
			
			if isInside == true then
				if _datas.tracking.flag == nil or _datas.tracking.flag  == false then
					_datas.tracking.flag = true
					local datas = {
						isInside = true,
						object = _datas.tracking.object,
					}
					
					for _moduleName, _callback in pairs(_datas.tracking.callbacks) do
						ATMEDatas.currentModuleName = _moduleName
						
						ATMEFct.outputCore(nil,"Appel callback - Entrée zonesMgr pour unité " .. _datas.tracking.name .. " pour le module " .. _moduleName, 2)
						xpcall(function() _callback(_datas.zonesMgr, datas) end, function() ATMEFct.errUserHandler('Tracking callback') end)
					end
				end
			else
				if _datas.tracking.flag  == true then
					_datas.tracking.flag = false
					local datas = {
						isInside = false,
						object = _datas.tracking.object,
					}
					
					for _moduleName, _callback in pairs(_datas.tracking.callbacks) do
						ATMEDatas.currentModuleName = _moduleName
						
						ATMEFct.outputCore(nil,"Appel callback - Sortie zonesMgr pour unité " .. _datas.tracking.name .. " pour le module " .. _moduleName, 2)
						xpcall(function() _callback(_datas.zonesMgr, datas) end, function() ATMEFct.errUserHandler('Tracking callback') end)
					end
				end
			end
		end
		
		return timer.getTime() + 1
	else
		_datas.localMembers.tracking[_datas.tracking.className] = nil
	end
end

ATMEDatas.C_ZonesMgr_MT = {
	className = "C_ZonesMgr",
	mt = {},
	
	keys = {
		["add"] = true,
		["addUnitRelative"] = true,
		["isInside"] = true,
		["getIndex"] = true,
		["getRandomPosition"] = true,
		["getCount"] = true,
		["addTrackingCallback"] = true,
		["removeTrackingCallback"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getInitialZone"] = true,
		["getZone"] = true,
		["getZoneLinkedUnitName"] = true,
		["getInsideZonesIndex"] = true,
		["getUserPrivateDatas"] = true,
	},
	
	add = function(self)
		local item = {
			zonesList = {},
			tracking = {},
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].tracking = nil
		ATMEDatas.localMembers[_id].zonesList = nil
	end,
}

ATME.C_ZonesMgr = {
	getClassName = function(self)
		return ATMEDatas.C_ZonesMgr_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_ZonesMgr_MT.parentClass
	end,
	
	createNamed = function(_name)
		local fctDefinition = "Function ATME.C_ZonesMgr.createNamed(name)"
		
		if type(_name) == "string" then
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newZonesMgr = ATMEFct.C_ZonesMgr_createNamed(localMembersModule.zonesMgr, _name)				
			
			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, newZonesMgr, false)
				return
			end
			
			return newZonesMgr
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
		end
	end,
	
	getByName = function(_name)	
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].zonesMgr[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_ZonesMgr.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_ZonesMgr_create = function(_name, _zonesMgr)
	local newZonesMgr = {}

	setmetatable(newZonesMgr, ATMEDatas.C_ZonesMgr_MT.mt)
	
	local name
	local tagsName
	
	if _name ~= nil and _zonesMgr ~= nil then
		name = _name
		_zonesMgr[name] = newZonesMgr
		tagsName = ATMEFct.decompString(_name)
	end
	
	local index = ATMEDatas.C_ZonesMgr_MT:add()
	local localMembers = ATMEDatas.localMembers[index]
	
	local function updateZone(_index)
		local datas = {
			zone = localMembers.zonesList[_index].zone,
			actualCenter = ATMEFct.getActualCenter(localMembers.center, localMembers.zonesList[_index].linkedUnitDatas),
			actualAzimuth = ATMEFct.getActualAzimuth(localMembers.azimuth, localMembers.zonesList[_index].linkedUnitDatas),
		}
			
		return ATMEFct[ATMEConst.indZoneConfig[localMembers.zonesList[_index].zone:getClassName()].fctDuplicate](datas)
	end
	
	function newZonesMgr.getName(self)
		return name
	end
	
	function newZonesMgr.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newZonesMgr) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newZonesMgr.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newZonesMgr.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newZonesMgr) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newZonesMgr.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	function newZonesMgr.add(self, _zone)
		local item = {}
		local fctDefinition = ATMEFct.getErrorPrefix(newZonesMgr) .. "add(zone)"
	
		if type(_zone) == "string" then
			local zone = ATMEFct.getDCSZone(_zone)
			if zone ~= nil then
				item.zone = ATMEFct.C_Circle_create(nil, nil, zone.point, zone.radius)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. _zone .. "' unknown", false)
				return
			end
		elseif ATMEFct.isObjectClass(_zone, ATME.C_Zone) == true then	
			item.zone = _zone
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be string, ATME.C_Circle, ATME.C_PickupZone, ATME.C_Rectangle or ATME.C_Polygon", false)
			return
		end
		
		return table.insert(localMembers.zonesList, item)
	end
	
	function newZonesMgr.addUnitRelative(self, _zone, _linkedUnit, _azimuth, _distance, _deltaAltitude)
		local fctDefinition = ATMEFct.getErrorPrefix(newZonesMgr) .. "addUnitRelative(zone, linkedUnit, [azimuth, distance, [deltaAltitude]])"

		local item = {}
		
		if type(_zone) == "string" then
			local zone = ATMEFct.getDCSZone(_zone)
			if zone ~= nil then
				item.zone = ATMEFct.C_Circle_create(nil, nil, zone.point, zone.radius)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone '" .. _zone .. "' unknown", false)
				return
			end
		elseif ATMEFct.isObjectClass(_zone, ATME.C_MovableZone2D) == true or ATMEFct.isObjectClass(_zone, ATME.C_MovableZone3D) == true then
			item.zone = _zone
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be derived from ATME.C_MovableZone2D and ATME.C_MovableZone3D", true)
			return
		end

		if ATMEFct.isObjectClass(_linkedUnit, ATME.C_Unit) == true then
			item.linkedUnitDatas = {}
			item.linkedUnitDatas.unitName = _linkedUnit:getName()
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "linkedUnit type must be ATME.C_AIUnit or ATME.C_Player", false)
			return
		end
		
		if _azimuth == nil and _distance == nil and _deltaAltitude == nil then
			item.linkedUnitDatas.azimuthRad = 0
			item.linkedUnitDatas.distance = 0
			item.linkedUnitDatas.deltaAltitude = 0
		elseif type(_azimuth) == "number" and type(_distance) == "number" then
			if _deltaAltitude == nil or type(_deltaAltitude) == "number" then
				local azimuth = ATMEFct.getAzimuthMod(_azimuth, 360)
				local deltaAltitude = 0
				
				if type(_deltaAltitude) == "number" then
					deltaAltitude = _deltaAltitude
				end
				
				item.linkedUnitDatas.azimuthRad = math.rad(azimuth)
				item.linkedUnitDatas.distance = _distance
				item.linkedUnitDatas.deltaAltitude = deltaAltitude
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "deltaAltitude type must be nil or number", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth and distance type must be both nil or number", false)
		end
		
		return table.insert(localMembers.zonesList, item)
	end
	
	function newZonesMgr.getInitialZone(self, _index)
		local fctDefinition = ATMEFct.getErrorPrefix(newZonesMgr) .. "getInitialZone(index)"
		
		if type(_index) == "number" then
			if index > 0 then
				if localMembers.zonesList[index] ~= nil then
					return localMembers.zonesList[index].zone
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "index must be > 0", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "index type must be number", true)
		end
	end
	
	function newZonesMgr.getZone(self, _index)
		local fctDefinition = ATMEFct.getErrorPrefix(newZonesMgr) .. "getZone(index)"
		
		if type(_index) == "number" then
			local index = math.floor(_index)
			
			if index > 0 then
				if localMembers.zonesList[index] ~= nil then
					return updateZone(index)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "index must be > 0", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "index type must be number", true)
		end
	end
	
	function newZonesMgr.getZoneLinkedUnitName(self, _index)
		local fctDefinition = ATMEFct.getErrorPrefix(newZonesMgr) .. "getZoneLinkedUnitName(index)"
		
		if type(_index) == "number" then
			local index = math.floor(_index)
			
			if index > 0 then
				if localMembers.zonesList[index] ~= nil then
					if localMembers.zonesList[index].linkedUnitDatas ~= nil then
						return localMembers.zonesList[index].linkedUnitDatas.unitName
					end
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "index must be > 0", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "index type must be number", true)
		end
	end

	function newZonesMgr.isInside(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newZonesMgr) .. "isInside(reference)"
		local isInside = false
		
		
		if ATMEFct.isReference(_reference) == true then		

			isInside, _ = ATMEFct.C_ZonesMgr_isInside(self, ATMEFct.getPointFromReference(_reference))

			return isInside
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
	end
	
	function newZonesMgr.getInsideZonesIndex(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newZonesMgr) .. "getInsideZones(reference)"
		local zones
		
		if ATMEFct.isReference(_reference) == true then
			_, zones = ATMEFct.C_ZonesMgr_isInside(self, ATMEFct.getPointFromReference(_reference))
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
		end
		
		return zones
	end
	
	function newZonesMgr.getCount(self)
		return #localMembers.zonesList
	end
	
	function newZonesMgr.getIndex(self)
		return index
	end
	
	function newZonesMgr.getRandomPosition(self, _surfaceFilter, _isFlat, _altMin)
		local fctDefinition = ATMEFct.getErrorPrefix(newZonesMgr) .. "getRandomPosition(surfaceFilter, [flat], [altMin])"
	
		if type(_surfaceFilter) == "string" and (_surfaceFilter == "ALL" or _surfaceFilter == "WATER" or _surfaceFilter == "LAND") and
					(_isFlat == nil or type(_isFlat) =="boolean") and (_altMin == nil or type(_altMin) =="number") then
			if self:getCount() > 0 then
				local nbTries = 0
				local point
				local isInWater
				
				local isFlat = _isFlat
				
				if _isFlat == nil then
					isFlat = false
				end
				
				local altMin = 0
				
				if type(_altMin) == "number" then
					altMin = _altMin
				end
							
				
				local found = false

				repeat
					local randomIndex = math.random(1, #localMembers.zonesList)
					local currentZone = localMembers.zonesList[randomIndex].zone

					local localMembersZone = ATMEDatas.localMembers[currentZone:getIndex()]

					local datas = {
						zone = currentZone,
						actualCenter = ATMEFct.getActualCenter(localMembersZone.center, currentZone.linkedUnitDatas),
						actualAzimuth = ATMEFct.getActualAzimuth(localMembersZone.azimuth, currentZone.linkedUnitDatas),
					}

					point = ATMEFct[ATMEConst.indZoneConfig[currentZone:getClassName()].fctGetRandomPosition](datas)

					nbTries = nbTries + 1
					isInWater = ATME.isSurfaceWater(point)
					
					if _surfaceFilter == "ALL" or (isInWater == true and _surfaceFilter == "WATER") or 
							(isInWater == false and _surfaceFilter == "LAND") then
						if point.y > altMin then
							if isFlat == true and _surfaceFilter == "LAND" then
								local ok = true
								for _i = -1, 1 do
									for _j = -1, 1 do
										if _i ~= 0 or _j ~= 0 then
											local deltaX = _i * 10
											local deltaZ = _j * 10
											local d2 = deltaX * deltaX + deltaZ * deltaZ
											local h = ATMEFct.getSurfaceAltitude({x = point.x + deltaX, y = point.z + deltaZ}) - point.y

											-- Pente de 10% max soit h > d*10/100 -> h > d/10
											if h * h > d2 / 100 then
												ok = false
											end
										end
									end
								end
								
								found = ok
							else
								found = true
							end
						else
							found = false
						end
					end
				until found == true or nbTries >= 100
				
				if found == true then
					return point
				end
			end
		else
			if type(_surfaceFilter) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "surfaceFilter type must be string ('ALL', 'WATER' or 'LAND')", true)
			elseif _isFlat ~= nil and type(_isFlat) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "flat type must be nil or boolean", false)
			elseif _altMin ~= nil and type(_altMin) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "altMin type must be nil or number", false)
			end
		end
	end
	
	function newZonesMgr.addTrackingCallback(self, _fct, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newZonesMgr) .. "addTrackingCallback(fct, object)"

		if type(_fct) == "function" and ATMEFct.isObjectClass(_object, ATME.C_Unit) == true then
			local className = _object:getClassName()
			local name = _object:getName()
			if localMembers.tracking[name] == nil then
				localMembers.tracking[name] = {
					className = className,
					name = name,
					object = _object,
					localMembersObject = ATMEDatas.localMembers[_object:getIndex()],
					callbacks = {},
				}
			end
				
			localMembers.tracking[name].pos = _object:getPosition()
			
			localMembers.tracking[name].callbacks[ATMEDatas.currentModuleName] = _fct
			
			local datas = {
				tracking = localMembers.tracking[name],
				zonesMgr = self,
			}
			
			timer.scheduleFunction(ATMETimers.C_ZonesMgr_timer, datas, timer.getTime() + 1)
		else
			if type(_fct) ~= "function" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
			elseif ATMEFct.isObjectClass(_object, ATME.C_Unit) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Group, ATME.C_AIUnit or ATME.C_Player", false)
			end
		end
	end
	
	function newZonesMgr.removeTrackingCallback(self, _object)
		local fctDefinition = ATMEFct.getErrorPrefix(newZonesMgr) .. "removeTrackingCallback(object)"
		
		if ATMEFct.isObjectClass(_object, ATME.C_Unit) == true then
			if localMembers.tracking[_object:getName()] ~= nil then
				localMembers.tracking[_object:getName()].callbacks[ATMEDatas.currentModuleName] = nil
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "object type must be ATME.C_Group, ATME.C_AIUnit or ATME.C_Player", true)
		end
	end

	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end

	return newZonesMgr
end

setmetatable(ATME.C_ZonesMgr,ATMEDatas.C_ZonesMgr_MT)

ATMEDatas.C_ZonesMgr_MT.__call = function(t)
	return ATMEFct.C_ZonesMgr_create(nil, nil)
end

ATMEDatas.C_ZonesMgr_MT.mt.__metatable = "Private"
ATMEDatas.C_ZonesMgr_MT.mt.__index = ATME.C_ZonesMgr
ATMEDatas.C_ZonesMgr_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_ZonesMgr_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_ZonesMgr : Bad key " .. k, "table locked", false)
	end
end


-----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Sound : gère les fichiers son définis dans la mission
--
-----------------------------------------------------------------------------------------------------------

ATMEFct.C_Sound_initPlaylist = function(_datas)
	local datas = {
		randomRead = _datas.randomRead,
		readOnce = _datas.readOnce,
		playlist = _datas.playlist,
		currentReadIndex = 1,
		toRemove = false,
	}
	
	if _datas.player ~= nil then
		local localMembers = ATMEDatas.localMembers[_datas.player:getIndex()]
		datas.playerName = _datas.player:getName()
		if localMembers.soundDatas ~= nil then
			localMembers.soundDatas.toRemove = true
		end
		
		localMembers.soundDatas = datas
		
		timer.scheduleFunction(ATMETimers.C_Sound_timer, datas, timer.getTime() + 1)
	else
		datas.players = {}
		for i = 1, 2 do
			datas.players[i] = coalition.getPlayers(i)
		end
		
		datas.coalition = _datas.coalition
		datas.country = _datas.country
		datas.indexCoalition = 1

		timer.scheduleFunction(ATMETimers.C_Sound_playersTimer, datas, timer.getTime() + 1)
	end
end

ATMETimers.C_Sound_playersTimer = function(_datas)
	if ATMEDatas.exitATME == false then
		if #_datas.players[_datas.indexCoalition] > 0 then
			if _datas.players[_datas.indexCoalition][#_datas.players[_datas.indexCoalition]]:isExist() == true then
				local datas = {
					randomRead = _datas.randomRead,
					readOnce = _datas.readOnce,
					playlist = _datas.playlist,
					currentReadIndex = 1,
					toRemove = false,
					coalition = _datas.coalition,
					country = _datas.country,
					playerName = _datas.players[_datas.indexCoalition][#_datas.players[_datas.indexCoalition]]:getName()
				}
							
				local localMembers = ATMEDatas.localMembers[ATMEDatas.players[datas.playerName]:getIndex()]
				if localMembers.soundDatas ~= nil then
					localMembers.soundDatas.toRemove = true
				end
				
				localMembers.soundDatas = datas
				
				if _datas.country ~= nil then
					if _datas.players[_datas.indexCoalition][#_datas.players[_datas.indexCoalition]]:getCountry() == _datas.country then
						timer.scheduleFunction(ATMETimers.C_Sound_timer, datas, timer.getTime() + 1)
					end
				elseif _datas.coalition ~= nil then
					if _datas.coalition == _datas.indexCoalition then
						timer.scheduleFunction(ATMETimers.C_Sound_timer, datas, timer.getTime() + 1)
					end
				else
					timer.scheduleFunction(ATMETimers.C_Sound_timer, datas, timer.getTime() + 1)
				end
			end
			
			table.remove(_datas.players[_datas.indexCoalition],#_datas.players[_datas.indexCoalition])
		else
			if _datas.indexCoalition == 1 then
				_datas.indexCoalition = 2
			else
				return
			end
		end
		
		return timer.getTime() + 0.1
	end
end

ATMETimers.C_Sound_timer = function(_soundDatas)
	if ATMEDatas.exitATME == false then
		if _soundDatas.playerName ~= nil then
			if ATMEDatas.players[_soundDatas.playerName] ~= nil then
				local localMembersPlayer = ATMEDatas.localMembers[ATMEDatas.players[_soundDatas.playerName]:getIndex()]
				
				if _soundDatas.toRemove == false then
					if _soundDatas.randomRead == true then 
						_soundDatas.currentReadIndex = math.random(1, #_soundDatas.playlist)
					end
					
					local tips = _soundDatas.playlist[_soundDatas.currentReadIndex].tipsBase
					trigger.action.outSoundForGroup(localMembersPlayer.groupId, localMembersPlayer.currentVoice .. _soundDatas.playlist[_soundDatas.currentReadIndex].fileName)

					if _soundDatas.playlist[_soundDatas.currentReadIndex].subtitle.duration > 0 then
						trigger.action.outTextForGroup(localMembersPlayer.groupId, _soundDatas.playlist[_soundDatas.currentReadIndex].subtitle.text, _soundDatas.playlist[_soundDatas.currentReadIndex].subtitle.duration)
					end

					_soundDatas.currentReadIndex = _soundDatas.currentReadIndex + 1
					if _soundDatas.currentReadIndex > #_soundDatas.playlist then
						if _soundDatas.readOnce == true then
							return
						else
							_soundDatas.currentReadIndex = 1
						end
					end
				
					return timer.getTime() + tips
				end
			else
				return
			end
		end
	end
end

ATMEFct.C_Sound_getPlaylist = function(_text, _random, _special)
	local tokens = ATMEFct.decompString(_text)

	if tokens ~= nil then
		local playlist = {}

		for _, _token in ipairs(tokens) do
			local _, _, temp = _token:find("([%d%a_]+)")
			
			if temp == _token then
				local idToken = _token:upper()
				if ATMEDatas.sounds[idToken] ~= nil then
					local localMembers = ATMEDatas.localMembers[ATMEDatas.sounds[idToken]:getIndex()]
					table.insert(playlist, {
												fileName = ATMEDatas.sounds[idToken]:getFileName(),
												tipsBase =  localMembers.tipsBase,
												tips = localMembers.tips,
												pause = false,
												spell = false,
												subtitle = {
													text = localMembers.subtitle.text,
													duration = localMembers.subtitle.duration,
												},
											})
				else
					return true, "Token " .. _token .. " - Sound doesn't exist"
				end
			elseif _special == nil then
				_, _, temp = _token:find("(%([%d]+.?[%d]*%))")

				if temp == _token then  -- Cas de la pause
					table.insert(playlist, {
												fileName = "",
												pause = true,
												spell = false,
												tipsBase =  tonumber(_token:sub(2, -2)),
												tips =  math.ceil(tonumber(_token:sub(2, -2))),
												subtitle = {
													text = "",
													duration = 0,
												},
											})
				else
					_, _, temp = _token:find("({[%d]+.?[%d]+})")
					local ok = false
					
					if temp == _token then -- Cas nombre à épeler
						ok = true
					else
						_, _, temp = _token:find("({[%a%d]+})")
						
						if temp == _token then -- Cas lettres et chiffres à épeler 
							temp = _token:upper()
							ok = true
						end
					end
					
					if ok == true then
						local lg = temp:len() - 1
						for i = 2, lg do
							local id = temp:sub(i,i)
							
							if id == "." then
								id = "DECIMAL"
							end
							
							if ATMEDatas.sounds[id] ~= nil then
								local localMembers = ATMEDatas.localMembers[ATMEDatas.sounds[id]:getIndex()]
								table.insert(playlist, {
															fileName = ATMEDatas.sounds[id]:getFileName(),
															tipsBase =  localMembers.tipsBase,
															tips = localMembers.tips,
															pause = false,
															spell = true,
															subtitle = {
																text = localMembers.subtitle.text,
																duration = localMembers.subtitle.duration,
															},
														})
							else
								return true, "Token " .. _token .. " spelling '" .. temp:sub(i,i) .. "' - Sound " .. id .. " is missing"
							end
						end
					else
						local _, _, cardinal, d, m, cent = _token:find("%[(%a)(%d%d%d?).?(%d%d).?(%d%d)%]")
						
						if cardinal ~= nil then
							local cardinalUpper = cardinal:upper()
							if ATMEConst.cardinal[cardinalUpper] ~= nil then
								local values = {
									tonumber(d),
									tonumber(m),
									tonumber(cent),
								}
								
								local maxdeg = 90
								
								if cardinalUpper == "E" or cardinalUpper == "W" then
									maxdeg = 180
									d = string.format("%03d", values[1])
								end
								
								if values[1] <= maxdeg and values[2] < 60 then
									if ATMEDatas.sounds[ATMEConst.cardinal[cardinalUpper]] ~= nil then
										local localMembers = ATMEDatas.localMembers[ATMEDatas.sounds[ATMEConst.cardinal[cardinalUpper]]:getIndex()]
										table.insert(playlist, {
																	fileName = ATMEDatas.sounds[ATMEConst.cardinal[cardinalUpper]]:getFileName(),
																	tipsBase =  localMembers.tipsBase,
																	tips = localMembers.tips,
																	pause = false,
																	spell = true,
																	subtitle = {
																		text = localMembers.subtitle.text,
																		duration = localMembers.subtitle.duration,
																	},
																})
										
										local coord = d .. "." .. m .. "." .. cent									
										local lg = coord:len()
										local extCar = ""
										
										for i = 1, lg do
											local id = coord:sub(i,i)
											
											if id == "." and ATMEDatas.sounds["POINT"] ~= nil then
												id = "POINT"
											end
											
											if id ~= "." then
												if ATMEDatas.sounds[id] ~= nil then
													local localMembers = ATMEDatas.localMembers[ATMEDatas.sounds[id]:getIndex()]
													table.insert(playlist, {
																				fileName = ATMEDatas.sounds[id]:getFileName(),
																				tipsBase =  localMembers.tipsBase,
																				tips = localMembers.tips,
																				pause = false,
																				spell = true,
																				subtitle = {
																					text = extCar .. localMembers.subtitle.text,
																					duration = localMembers.subtitle.duration,
																				},
																			})
													extCar = ""
												else
													return true, "Token '" .. _token .. "' spelling '" .. _token:sub(i,i) .. "' - Sound " .. id .. " is missing"
												end
											else
												extCar = "."
											end
										end
									else
										return true, "Token '" .. _token .. "' spelling '" .. cardinal .. "' - Sound " .. ATMEConst.cardinal[cardinalUpper] .. " is missing"
									end
								else
									return true, "Token '" .. _token .. "' - bad coordinates"
								end
							else
								return true, "Token '" .. _token .. "' - cardinal value must be N, S, E or W"
							end
						else
							return true, "Token '" .. _token .. "' - bad format"
						end
					end
				end
			end
			
			if _special == true then
				break
			end
		end
						
		if #playlist > 1 and _random == false then
			-- Si playlist > 1 alors les sous titres sont regroupés dans le premier item
			local spelling = playlist[1].spell
			for i = 2, #playlist do
				if (spelling == true and playlist[i].spell == true) or playlist[1].subtitle.text:sub(-1, -1) == "\n" or playlist[i].subtitle.text:sub(1, 1) == "\n" or
					playlist[1].subtitle.text:sub(-1, -1) == " " or playlist[i].subtitle.text:sub(1, 1) == " " then
					playlist[1].subtitle.text = playlist[1].subtitle.text .. playlist[i].subtitle.text
				elseif playlist[i].subtitle.text ~= "" then
					playlist[1].subtitle.text = playlist[1].subtitle.text .. " " .. playlist[i].subtitle.text
				end
				
				playlist[i].subtitle.text = ""
				playlist[1].subtitle.duration = playlist[1].subtitle.duration + playlist[i].subtitle.duration
				playlist[i].subtitle.duration = 0
				
				spelling = playlist[i].spell
			end
		elseif #playlist == 0 then
			playList = nil
		end
		
		return false, playlist
	end
	
	return false
end

ATMEDatas.C_Sound_MT = {
	className = "C_Sound",
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getID"] = true,
		["getDuration"] = true,
		["getRoundDuration"] = true,
		["getFileName"] = true,
		["getSubtitle"] = true,
		["getUserPrivateDatas"] = true,
	},
	
	add = function(self, _tips, _tipsBase)
		local item = {
			tipsBase = _tipsBase,
			tips = _tips,
			subtitle = {
				text = "",
				duration = 0,
			},
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Sound = {
	getClassName = function(self)
		return ATMEDatas.C_Sound_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Sound_MT.parentClass
	end,
	
	getByID = function(_id)
		if type(_id) == "string" then
			return ATMEDatas.sounds[string.upper(_id)]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Sound.getByID(id)", "id type must be string", false)
		end
	end,
}

ATMEFct.C_Sound_create = function(_id, _tips, _tipsBase, _fileName, _subtitle)	
	local newSound = {}
	local id = _id
	local fileName = _fileName
	local idFrequency
	
	local index = ATMEDatas.C_Sound_MT:add(_tips, _tipsBase)
	local localMembers = ATMEDatas.localMembers[index]

	local durationSec = _tips
	local roundDurationSec = math.ceil(durationSec)
	
	setmetatable(newSound, ATMEDatas.C_Sound_MT.mt)
	
	function newSound.getIndex(self)
		return index
	end
	
	function newSound.getID(self)
		return id
	end
	
	function newSound.getDuration(self)
		return durationSec
	end
	
	function newSound.getRoundDuration(self)
		return roundDurationSec
	end
	
	function newSound.getFileName(self)
		return fileName
	end
	
	function newSound.getSubtitle(self)
		return localMembers.subtitle.text, localMembers.subtitle.duration
	end
	
	function newSound.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	localMembers.subtitle.text = _subtitle.text
	localMembers.subtitle.duration = _subtitle.duration
	
	ATMEDatas.sounds[id] = newSound
	
	return newSound
end

setmetatable(ATME.C_Sound,ATMEDatas.C_Sound_MT)

ATMEDatas.C_Sound_MT.__call = function(t, _id, ...)
	local fctDefinition = "Constructor ATME.C_Sound(ID, fileName, duration)\nATME.C_Sound(ID, fileName, duration, subtitle, subtitleDuration)"
	local args = {...}
	
	if type(_id) == "string" then
		local tokens = ATMEFct.decompString(_id)
			
		if tokens ~= nil then
			local d, f, temp = tokens[1]:find("([%d%a_]+)")
			
			if temp == tokens[1] then
				if #tokens > 1 then
					ATMEFct.errorOutputCore(nil, fctDefinition, "ID consist of letters, numbers and '_' char", false)
				else
					local tokenName = string.upper(tokens[1])
					if ATMEDatas.sounds[tokenName] ~= nil then
						ATMEFct.errorOutputCore(nil, fctDefinition, "ID " .. tokens[1] .. " already exists", false)
					else
						local tips
						local fileName
						local subtitle
						
						if #args >= 2 then
							if type(args[1]) == "string" and type(args[2]) == "number" then
								fileName = args[1]
								tips = math.floor(args[2] + 1)
								
								if #args == 4 then
									if type(args[3]) == "string" and type(args[4]) == "number" then
										local subtitleDuration = math.floor(args[4])
										
										if subtitleDuration >= 1 then
											subtitle = {
												text = args[3],
												duration = args[4],
											}
										else
											ATMEFct.errorOutputCore(nil, fctDefinition, "durationSubtitle must be >= 1", false)
											return
										end
									else
										if type(args[3]) ~= "string" then
											ATMEFct.errorOutputCore(nil, fctDefinition, "subtitle type must be string", false)
											return
										elseif type(args[4]) ~= "number" then
											ATMEFct.errorOutputCore(nil, fctDefinition, "durationSubtitle type must be number", false)
											return
										end
									end
								elseif #args == 2 then
									subtitle = {
										text = "",
										duration = 0,
									}
								else
									ATMEFct.errorOutputCore(nil, fctDefinition, "bad parameters", false)
									return
								end
							else
								if type(args[1]) ~= "string" then
									ATMEFct.errorOutputCore(nil, fctDefinition, "fileName type must be string", false)
									return
								elseif type(args[2]) ~= "number" then
									ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be number", false)
									return
								end
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "bad parameters", false)
							return
						end

						if args[2] > 0 then
							return ATMEFct.C_Sound_create(tokenName, tips, args[2], fileName, subtitle)
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "duration must be >= 0.1", false)
						end
					end
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "ID " .. token[1] .. " - bad format", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "no ID defined", false)
		end
	else
		if type(_id) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "ID type must be string", false)
		elseif type(_duration) ~= "number" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be number", false)
		elseif _fileName ~= nil and type(_fileName) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "fileName type must be nil or string", false)
		end
	end
end

ATMEDatas.C_Sound_MT.mt.__metatable = "Private"
ATMEDatas.C_Sound_MT.mt.__index = ATME.C_Sound
ATMEDatas.C_Sound_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Sound_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Sound : Bad key " .. k, "table locked", false)
	end
end



----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_F10Menu : gère les menus accessibles par F10. Constructeur accessible aux modules externes 
-- Cette classe est complexe et gère :
--			- Les entrées d'un menu (menu de base root ou sous menu). Un entrée engendre une action par l'appel
--			  d'une fonction.
--			- Les sous-menus rattachés :
--					-> au menu de base root (accessible directement par F10)
--					-> à un menu, lui même sous-menu du menu de base rout ou d'un autre menu.
--			- L'affichage dans l'ordre choisi (alpha par défaut) des sous-menus du menu courant et de ses entrées
--					-> sous menus d'abord
--					-> entrées ensuite
--			- La gestion des entrées par une identification de type groupId pour une gestion en groupe :
--					-> lors de l'affichage dans l'ordre par groupId puis alpha si groupId identique
--					-> pour la suppression de plusieurs entrées  simultanément.
--					-> le groupId d'une entrée s'applique à toutes les entrées d'un menu donné, y compris 
--					   pour les entrées de ses sous-menus
--			- Un menu vide n'est plus affiché. Un menu est vide si lui-même et ses sous-menus ne contiennent
--			  aucune entrée
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_F10Menu_MT = {
	className = "C_F10Menu",
	mt = {},
	
	keys = {
		["getParent"] = true,
		["isRoot"] = true,
		["getLabel"] = true,
		["getIndex"] = true,
		["getPlayer"] = true,
		["append"] = true,
		["removeAll"] = true,
		["remove"] = true,
		["getNbEntries"] = true,
		["getUserPrivateDatas"] = true,
	},
	
	add = function(self, _player)
		local item = {
			itemCommandsList = {},
			itemSubMenusList = {},
			isActive = false,
			path = nil,
			nbSubmenus = 0,
			erased = false,									-- Utilisé pour les données privées. Passe à true si le menu n'est pas affiché
			handlers = ATMEFct.C_IndexList_create(),		-- Table des handlers définis (proxyHandler d'appel)
			userPrivateDatas = {},
			player = _player,
			localMembersPlayer = ATMEDatas.localMembers[_player:getIndex()],
		}

		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].player = nil
		ATMEDatas.localMembers[_id].localMembersPlayer = nil
	end,
}


ATME.C_F10Menu = {
	getClassName = function(self)
		return ATMEDatas.C_F10Menu_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_F10Menu_MT.parentClass
	end,
}

ATMEFct.C_F10Menu_handler = function(_args)
	local localMembers = ATMEDatas.localMembers[_args.indexMenu] 
	local tempModuleName = ATMEDatas.currentModuleName
	
	ATMEDatas.currentModuleName = _args.moduleName
	ATMEFct.outputCore(nil,"Appel handler du menu '" .. _args.labelMenu .. "' du module " .. ATMEDatas.currentModuleName, 2)
	xpcall(function() localMembers.handlers:get(_args.idHandler)(_args.menu, _args.args) end, function() ATMEFct.errUserHandler(_args.labelMenu, true) end)
	
	ATMEDatas.currentModuleName = tempModuleName
end

ATMEFct.C_F10Menu_create = function(_player, _label, _parent, _moduleName)
	local found = false
	local nbEntries = 0

	if _parent ~= nil then
		local localMembersParent = ATMEDatas.localMembers[_parent:getIndex()]
		if _parent:isRoot() == true then -- Root menu direct avec _moduleName défini
			for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
				for _k, _itemSubMenu in pairs(localMembersParent.itemSubMenusList) do
					if _itemSubMenu:getLabel() == _label then
						found = true
					end
				end
			end
		else -- Sous menu
			for _k, _itemSubMenu in pairs(localMembersParent.itemSubMenusList) do
				if _itemSubMenu:getLabel() == _label then
					found = true
				end
			end
		end
		nbEntries = _parent:getNbEntries()
	end

	if nbEntries < 10 and found == false then
		local parent = _parent
		local isRoot = false
		local label = _label
		
		local newF10SubMenu = {}
		
		setmetatable(newF10SubMenu, ATMEDatas.C_F10Menu_MT.mt)
			
		local index = ATMEDatas.C_F10Menu_MT:add(_player)
		local localMembers = ATMEDatas.localMembers[index]
		
		-- Si _moduleName ~= nil -> Menu root, à mémoriser pour le joueur
		-- Plusieurs arborescences root peuvent être créées pour un joueur donné
		if _parent == nil then
			isRoot = true
			label = "Root"
		end
		
		function newF10SubMenu.getParent(self)
			return parent
		end
		
		function newF10SubMenu.isRoot(self)
			return isRoot
		end
		
		function newF10SubMenu.getLabel(self)
			return label
		end
		
		function newF10SubMenu.getIndex(self)
			return index
		end
		
		function newF10SubMenu.getPlayer(self)
			return localMembers.player
		end
		
		function newF10SubMenu.getNbEntries(self)
			local nbEntries = 0
			
			if self:isRoot() == true then
				for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
					local menuRoot = localMembers.localMembersPlayer.itemsF10MenuCommandRoot[_localModuleMembers.name]
					local localMembersRoot = ATMEDatas.localMembers[menuRoot:getIndex()]
					nbEntries = nbEntries + localMembersRoot.nbSubmenus + #localMembersRoot.itemCommandsList
				end
			else
				nbEntries = localMembers.nbSubmenus + #localMembers.itemCommandsList
			end
			
			return nbEntries
		end
		
		local function C_F10Menu_deleteAllCommands(_itemF10Menu, _deleteSubMenu)
			ATMEFct.outputCore(nil, "Start Suppression -> " .. _itemF10Menu:getLabel(), 3)
			
			local localMembersSub = ATMEDatas.localMembers[_itemF10Menu:getIndex()]
			while #localMembersSub.itemCommandsList > 0 do
				local toRemove= table.remove(localMembersSub.itemCommandsList)
				missionCommands.removeItemForGroup(localMembersSub.localMembersPlayer.groupId, toRemove.itemF10MenuCommandId)
				localMembersSub.handlers:remove(toRemove.idHandler)
			end
			
			for _k, _itemSubMenu in pairs(localMembersSub.itemSubMenusList) do
				C_F10Menu_deleteAllCommands(_itemSubMenu, true)
				localMembersSub.itemSubMenusList[_k] = nil
				localMembersSub.nbSubmenus = localMembersSub.nbSubmenus - 1
			end
		  
			if _deleteSubMenu == true and _itemF10Menu:isRoot() == false then
				missionCommands.removeItemForGroup(localMembersSub.localMembersPlayer.groupId, localMembersSub.path)
				localMembersSub.erased = true
			end

			ATMEFct.outputCore(nil, "End Suppression -> " .. _itemF10Menu:getLabel(), 3)
		end

		local function C_F10Menu_deleteCommands(_itemF10Menu, _groupId, _deleteSubMenu)
			local function deleteCommands(_itemF10Menu, _groupId, _deleteSubMenu)
				local erase = false

				ATMEFct.outputCore(nil, "Start Suppression -> " .. _itemF10Menu:getLabel(), 3)
			  
				local localMembersSub = ATMEDatas.localMembers[_itemF10Menu:getIndex()]
				local i = 1
				while i <= #localMembersSub.itemCommandsList do
					if localMembersSub.itemCommandsList[i].groupId == _groupId then    
						missionCommands.removeItemForGroup(localMembersSub.localMembersPlayer.groupId, localMembersSub.itemCommandsList[i].itemF10MenuCommandId)
						localMembersSub.handlers:remove(localMembersSub.itemCommandsList[i].idHandler)
						table.remove(localMembersSub.itemCommandsList, i)
					else
						i = i + 1
					end
				end
				
				for _k, _itemSubMenu in pairs(localMembersSub.itemSubMenusList) do
					if deleteCommands(_itemSubMenu, _groupId, true) == true then
						localMembersSub.itemSubMenusList[_k] = nil
						localMembersSub.nbSubmenus = localMembersSub.nbSubmenus - 1
					end
				end
				
				if #localMembersSub.itemCommandsList == 0 then
					if _deleteSubMenu == true and _itemF10Menu:isRoot() == false then
						missionCommands.removeItemForGroup(localMembersSub.localMembersPlayer.groupId, localMembersSub.path)
						localMembersSub.erased = true
					end
					
					erase = true
				end
				
				ATMEFct.outputCore(nil, "End Suppression -> " .. _itemF10Menu:getLabel(), 3)
					
				return erase
			end
			
			deleteCommands(_itemF10Menu, _groupId, _deleteSubMenu)
		end
		
		function newF10SubMenu.append(self, _groupId, _label, _handler, _args)
			local fctDefinition = ATMEFct.getErrorPrefix(newF10SubMenu) .. "append(groupId, label, handler, [args])"
			if type(_groupId) == "number" and type(_label) == "string" and type(_handler) == "function" then
				if self:getNbEntries() < 10 then
					if _groupId >= 1 then
						ATMEFct.outputCore(nil, "Start append -> " .. _label, 3)
						
						local id = localMembers.handlers:add(_handler)
						
						local args = {
							idHandler = id,
							indexMenu = index,
							labelMenu = _label,
							menu = self,
							moduleName = ATMEDatas.currentModuleName,
							args = _args,
						}

						local itemRadio = {
								groupId = math.floor(_groupId),
								label = _label,
								idHandler = id,
								itemF10MenuCommandId = missionCommands.addCommandForGroup(localMembers.localMembersPlayer.groupId, _label, localMembers.path, ATMEFct.C_F10Menu_handler , args)
						}

						table.insert(localMembers.itemCommandsList,itemRadio)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "groupId must be >= 1", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Max of entries per menu is 10. Can't add new menu item", false)
				end
			else
				if type(_groupId) ~= "number" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "groupId type must be number", true)
				elseif type(_label) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "label type must be string", false)
				elseif type(_handler) ~= "function" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "handler type must be function", false)
				end
			end
		end
		
		function newF10SubMenu.removeAll(self, _deleteSubMenu)
			local fctDefinition = ATMEFct.getErrorPrefix(newF10SubMenu) .. "removeAll([deleteSubMenu])"
			
			if  _deleteSubMenu == nil or type(_deleteSubMenu) == "boolean" then
				local flag = false
				if _deleteSubMenu ~= nil then
					flag = _deleteSubMenu
				end

				C_F10Menu_deleteAllCommands(self, flag)
				if flag == true and parent ~= nil then
					local localParentMembers = ATMEDatas.localMembers[parent:getIndex()]
					
					localParentMembers.itemSubMenusList[label] = nil
					localParentMembers.nbSubmenus = localParentMembers.nbSubmenus - 1
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "deleteSubMenu type must be nil or boolean", true)
			end
		end
		
		function newF10SubMenu.remove(self, _groupId, _deleteSubMenu)
			local fctDefinition = ATMEFct.getErrorPrefix(newF10SubMenu) .. "remove(groupId, [deleteSubMenu])"
			
			if type(_groupId) == "number" and (_deleteSubMenu == nil or type(_deleteSubMenu) == "boolean") then
				local flag = false
				if _deleteSubMenu ~= nil then
					flag = _deleteSubMenu
				end
				
				if _groupId >= 1 then
					C_F10Menu_deleteCommands(self, math.floor(_groupId), flag) 
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "groupId must be >= 1", false)
				end
			else
				if type(_groupId) ~= "number" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "groupId type must be number", true)
				elseif _deleteSubMenu ~= nil and type(_deleteSubMenu) ~= "boolean" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "deleteSubMenu type must be nil or boolean", false)
				end
			end
		end
		
		function newF10SubMenu.getUserPrivateDatas(self)
			return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
		end

		for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
			localMembers.userPrivateDatas[_localModuleMembers.name] = {}
		end

		if isRoot == false then
			local localMembersParent = ATMEDatas.localMembers[parent:getIndex()]
			localMembersParent.itemSubMenusList[label] = newF10SubMenu
			
			localMembersParent.nbSubmenus = localMembersParent.nbSubmenus + 1
			
			ATMEFct.outputCore(nil, "Création subMenu " .. label .. " à partir de " .. parent:getLabel(), 3)
		end
		
		
		
		return nbEntries, newF10SubMenu
	end
	
	return nbEntries
end


ATMEDatas.C_F10Menu_MT.__call = function(t, _label, _parent)
	local fctDefinition = "Constructor ATME.C_F10Menu(label, parent)"
	
	if type(_label) == "string" and ATMEFct.isObjectClass(_parent, ATME.C_F10Menu) == true then
		local nbEntries, menu = ATMEFct.C_F10Menu_create(_parent:getPlayer(), _label, _parent, nil)
		
		if menu ~= nil then
			local parentPath = nil
			if _parent:isRoot() == false then
				local localParentMembers = ATMEDatas.localMembers[_parent:getIndex()]
				parentPath = localParentMembers.path
			end
			
			local localMembers = ATMEDatas.localMembers[menu:getIndex()]
			localMembers.path = missionCommands.addSubMenuForGroup(localMembers.localMembersPlayer.groupId, menu:getLabel(), parentPath)
			
			return menu
		else
			if nbEntries == 10 then
				ATMEFct.errorOutputCore(nil, fctDefinition, "Max of entries per menu is 10. Can't add new submenu", false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Label '" .. _label .. "' already exist in parent or root menu", false)
			end
		end
	else
		if type(_label) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "label type must be string", false)
		elseif ATMEFct.isObjectClass(_parent, ATME.C_F10Menu) == false then
			ATMEFct.errorOutputCore(nil, fctDefinition, "parent type must be nil or ATME.C_F10Menu", false)
		end
	end
end

setmetatable(ATME.C_F10Menu,ATMEDatas.C_F10Menu_MT)

ATMEDatas.C_F10Menu_MT.mt.__metatable = "Private"
ATMEDatas.C_F10Menu_MT.mt.__index = ATME.C_F10Menu
ATMEDatas.C_F10Menu_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_F10Menu_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_F10Menu : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Alarm : gère les alarmes définies par l'utilisateur.
-- Pas de contructeur accessible aux modules externes.
-- En effet, la création d'une alarme passe par deux méthodes de la classe ATME.C_Module.
-- Il existe deux types d'alarme :
--		- Les alarmes absolues se déclenchant au moment programmé.
--		- Les alarmes relative à un flag et donc associée à l'activation de ce flag pour se déclencher.
-- La condition de temps est donné par un string au format particulier :
--		- "X..X" pour un temps en secondes depuis le début de mission (ex : "10" ou "100" pour 10s ou 100s)
--		- "+X..X" pour un temps en seconde depuis la création de l'alarme (ex : "+10" pour 10s après la création)
--		- "HH:MM" ou "HH:MM:SS" pour remplacer une expression en secondes par une expression en m/s ou h/m/s
----------------------------------------------------------------------------------------------------------

ATMEFct.C_Alarm_createNamed = function(_namedTable, _name, _flagID, _rules)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return false
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			return ATMEFct.C_Alarm_create(_namedTable, _name, _flagID, _rules)
		else
			return true, "name can't contain '|' character"
		end
	end
end

ATMEDatas.C_Alarm_MT = {
	className = "C_Alarm",
	mt = {},
	
	keys = {
		["getFlag"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getIndex"] = true,
		["getUserPrivateDatas"] = true, 
		["reschedule"] = true,
		["addTriggeringsLimit"] = true,
		["removeTriggeringsLimit"] = true,
		["addCallback"] = true,
		["removeCallback"] = true,
		["remove"] = true,
	},
	
	add = function(self)
		local item = {
			triggered = false,
			restarted = false,
			eventCallback = {},
			maxTriggerings = -1,
			maxTriggeringsCallback = {},
			userPrivateDatas = {},
			destroyed = false,
			moduleName = ATMEDatas.currentModuleName,
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].eventCallback = nil
		ATMEDatas.localMembers[_id].maxTriggeringsCallback = nil
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

----------------------------------------------------------------------------------------------------------
-- Fonctions timers appelées pour les alarmes
----------------------------------------------------------------------------------------------------------

ATMETimers.C_Alarm_timer = function(_alarm)
	if ATMEDatas.exitATME == false then
		local localMembers = ATMEDatas.localMembers[_alarm:getIndex()]
		local localMembersModule = ATMEDatas.modulesLocalMembers[localMembers.moduleName]
		
		localMembers.triggered = true
		if localMembers.flag ~= nil then
			if trigger.misc.getUserFlag(localMembers.flag) == 0 then
				ATMEFct.C_Alarm_delete(localMembersModule.alarms, _alarm)
				ATMEFct.outputCore(nil,"Alarme annulée -> flag non défini" , 2)
				return
			end
		end
		
		ATMEFct.outputCore(nil,"Alarme déclenchée" , 2)
		
		for _moduleName, _callback in pairs(localMembers.eventCallback) do
			ATMEDatas.currentModuleName = _moduleName

			ATMEFct.outputCore(nil,"Appel handler d'alarme pour le module " .. _moduleName, 2)
			xpcall(function() _callback(_alarm) end, function() ATMEFct.errUserHandler('alarmEventCallbackHandler') end)
		end
			
		if localMembers.maxTriggerings > 0 then
			localMembers.maxTriggerings = localMembers.maxTriggerings - 1
			
			if localMembers.maxTriggerings == 0 then
				for _moduleName, _callback in pairs(localMembers.maxTriggeringsCallback) do
					ATMEDatas.currentModuleName = _moduleName
						
					ATMEFct.outputCore(nil,"Appel handler maxTriggerings d'alarme pour le module " .. _moduleName, 2)
					xpcall(function() _callback(_alarm) end, function() ATMEFct.errUserHandler('maxTriggeringsCallbackHandler') end)
				end
				
				if localMembers.maxTriggerings ~= - 1 then  -- removeTriggeringLimit appelé dans les callbacks
					localMembers.restarted = false
				end
			end
		end
		
		if localMembers.restarted == false then
			ATMEFct.C_Alarm_delete(localMembersModule.alarms, _alarm)
		else
			localMembers.restarted = false
		end
	end
end	
				
ATMETimers.C_Alarm_timerSchedule = function(_args)
	if ATMEDatas.exitATME == false then
		ATMEDatas.currentModuleName = _args[3]
		xpcall(function() _args[1](unpack(_args[2])) end, function() ATMEFct.errUserHandler('Scheduled function') end)
	end
end

ATME.C_Alarm = {
	getClassName = function(self)
		return ATMEDatas.C_Alarm_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Alarm_MT.parentClass
	end,
	
	schedule = function(_rules, _fct, ...)
		local fctDefinition = "Function ATME.C_Alarm.schedule(rules, fct, fctArgs)"
		local args = {...}
		
		if (type(_rules) == "number" or type(_rules) == "string") and type(_fct) == "function" then
			local err, timeStart, isRelative = ATMEFct.getTimeValue(_rules)
	
			if err == false then
				local currentTime = timer.getTime()
				
				if isRelative == true then
					timeStart = timeStart + currentTime
				end
				
				if timeStart > currentTime then
					timer.scheduleFunction(ATMETimers.C_Alarm_timerSchedule, {_fct, args, ATMEDatas.currentModuleName}, timeStart)
					return true
				else
					ATMEFct.outputCore(nil, "Schedule refusé -> startTime < current time", 2)
					return false
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, timeStart, false)
			end
		else
			if type(args[1]) ~= "number" and type(args[1]) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "flagID type must be number or string", false)
			elseif type(_fct) ~= "function" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", false)
			end
		end
	end,
	
	createNamed = function(_name, ...)
		local fctDefinition = "Function ATME.C_Alarm.createNamed(name, rules)\nATME.C_Alarm.createNamed(name, flagID, rules)"
		local args = {...}
		
		if type(_name) == "string" then
			local d,f = string.find(_name, "|", 1)
			if d == nil then
				if #args == 1 then
					if type(args[1]) == "number" or type(args[1]) == "string" then
						local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
						local err, newAlarm = ATMEFct.C_Alarm_createNamed(localMembersModule.alarms, _name, nil, args[1])
							
						if err == true then
							ATMEFct.errorOutputCore(nil, fctDefinition, newAlarm, false)
							return
						end
						
						return newAlarm
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "rules type must be number or string", false)
					end
				elseif #args >= 2 then
					if (type(args[1]) == "number" or type(args[1]) == "string") and (type(args[2]) == "number" or type(args[2]) == "string") then
						local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
						local flag
						
						if type(args[1]) == "number" then
							flag = math.floor(args[1])
							
							if flag > 0 then
								flag = tostring(flag)
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "flagID must be > 0", false)
								return
							end
						else
							flag = args[1]
						end
						
						local err, newAlarm = ATMEFct.C_Alarm_createNamed(localMembersModule.alarms, _name, flag, args[2])
						
						if err == true then
							ATMEFct.errorOutputCore(nil, fctDefinition, newAlarm, false)
							return
						end
						
						return newAlarm
					else
						if type(args[1]) ~= "number" and type(args[1]) ~= "string" then
							ATMEFct.errorOutputCore(nil, fctDefinition, "flagID type must be number or string", false)
						elseif type(args[2]) ~= "number" and type(args[2]) ~= "string" then
							ATMEFct.errorOutputCore(nil, fctDefinition, "rules type must be number or string", false)
						end
					end
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "name can't contain '|' character", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", true)
		end
	end,
	
	getByName = function(_name)
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].alarms[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Alarm.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Alarm_create = function(_alarms, _name, _flagID, _rules)
	local name
	local tagsName
	local index
	local localMembers
	local err, timeStart, isRelative = ATMEFct.getTimeValue(_rules)
	
	if err == false then
		local currentTime = timer.getTime()

		if isRelative == true then
			timeStart = timeStart + currentTime
		end
		
		if timeStart > currentTime then
			local newAlarm = {}
			
			if _name ~= nil then
				name = _name
				_alarms[name] = newAlarm
				tagsName = ATMEFct.decompString(_name)
			end
		
			setmetatable(newAlarm, ATMEDatas.C_Alarm_MT.mt)
	
			index = ATMEDatas.C_Alarm_MT:add()
			localMembers = ATMEDatas.localMembers[index]
			localMembers.flag = _flagID

			function newAlarm.getIndex(self)
				return index
			end
			
			function newAlarm.getFlag(self)
				if localMembers.destroyed == false then
					if localMembers.flag ~= nil then
						return trigger.misc.getUserFlag(localMembers.flag)
					end
				else
					ATMEFct.setError("ALREADY_DESTROYED")
				end
			end
			
			function newAlarm.getName(self)
				if localMembers.destroyed == false then
					return name
				else
					ATMEFct.setError("ALREADY_DESTROYED")
				end
			end
			
			function newAlarm.tagGetByID(self, _position)
				local fctDefinition = ATMEFct.getErrorPrefix(newAlarm) .. "tagGetByID(position)"
				
				if localMembers.destroyed == false then
					if type(_position) == "number" then
						local pos = math.floor(_position)
						if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
							return tagsName[pos]
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
					end
				else
					ATMEFct.setError("ALREADY_DESTROYED")
				end
			end
			
			function newAlarm.tagGetCount(self)
				if localMembers.destroyed == false then
					if tagsName ~= nil then
						return #tagsName
					else
						return 0
					end
				else
					ATMEFct.setError("ALREADY_DESTROYED")
				end
			end
			
			function newAlarm.tagCompare(self, _tagsToCompare)
				local fctDefinition = ATMEFct.getErrorPrefix(newAlarm) .. "tagCompare(tagsToCompare)"
				
				if localMembers.destroyed == false then
					if type(_tagsToCompare) == "string" then
						if tagsName ~= nil then
							local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
							
							if ok ~= nil then
								return ok
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
							end
						else
							return false
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
					end
				else
					ATMEFct.setError("ALREADY_DESTROYED")
				end
			end
			
			function newAlarm.getUserPrivateDatas(self)
				if localMembers.destroyed == false then
					return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
				else
					ATMEFct.setError("ALREADY_DESTROYED")
				end
			end
			
			function newAlarm.addCallback(self, _fct)
				local fctDefinition = ATMEFct.getErrorPrefix(newAlarm) .. "addCallback(fct)"
				
				if localMembers.destroyed == false then
					if type(_fct) == "function" then
						localMembers.eventCallback[ATMEDatas.currentModuleName] = _fct
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
					end
				else
					ATMEFct.setError("ALREADY_DESTROYED")
				end
			end
			
			function newAlarm.removeCallback(self)
				if localMembers.destroyed == false then
					localMembers.eventCallback[ATMEDatas.currentModuleName] = nil
				else
					ATMEFct.setError("ALREADY_DESTROYED")
				end
			end
			
			function newAlarm.addTriggeringsLimit(self, _value, _callback)
				local fctDefinition = ATMEFct.getErrorPrefix(newAlarm) .. "addTriggeringsLimit(value, [callback])"
				
				if localMembers.destroyed == false then
					if type(_value) == "number" and (_callback == nil or type(_callback) == "function") then
						local value = math.floor(_value)
						
						if value > 0 then
							localMembers.maxTriggerings = value
							localMembers.maxTriggeringsCallback[ATMEDatas.currentModuleName] = _callback
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "value must be > 0", true)
						end
					else
						if type(_value) ~= "number" then
							ATMEFct.errorOutputCore(nil, fctDefinition, "value type must be number", true)
						elseif _callback ~= nil and type(_callback) ~= "function" then
							ATMEFct.errorOutputCore(nil, fctDefinition, "callback type must be nil or function", false)
						end
					end
				else
					ATMEFct.setError("ALREADY_DESTROYED")
				end
			end
			
			function newAlarm.removeTriggeringsLimit(self)
				if localMembers.destroyed == false then
					localMembers.maxTriggerings = -1
					localMembers.maxTriggeringsCallback[ATMEDatas.currentModuleName] = nil
				else
					ATMEFct.setError("ALREADY_DESTROYED")
				end
			end
			
			function newAlarm.reschedule(self, _rules)
				local fctDefinition = ATMEFct.getErrorPrefix(newAlarm) .. "reschedule(rules)"
				
				if localMembers.destroyed == false then
					if type(_rules) == "number" or type(_rules) == "string" then
						local err, timeStart, isRelative = ATMEFct.getTimeValue(_rules)
						
						if err == false then
							local currentTime = timer.getTime()
				
							if isRelative == true then
								timeStart = timeStart + currentTime
							end
							
							if timeStart > currentTime then
								localMembers.restarted = true
								
								if localMembers.triggered == false then
									timer.setFunctionTime(localMembers.idTimer, timeStart)
								else
									localMembers.idTimer = timer.scheduleFunction(ATMETimers.C_Alarm_timer, self, timeStart)
								end
								
								localMembers.triggered = false
								

								ATMEFct.outputCore(nil, "Reschedule alarme -> alarme reprogrammée (timeStart = ".. timeStart .. ")" , 2)
							else
								ATMEFct.outputCore(nil, "Reschedule alarme -> startTime < current time : alarme non prise en compte", 2)
								return false
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, timeStart, false)
						end
							
						return localMembers.restarted
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "rules type must be number or string", true)
					end
				else
					ATMEFct.setError("ALREADY_DESTROYED")
				end
			end
			
			function newAlarm.remove(self)
				if localMembers.destroyed == false then
					local localMembersModule = ATMEDatas.modulesLocalMembers[localMembers.moduleName]
					ATMEFct.C_Alarm_delete(localMembersModule.alarms, self)
				else
					ATMEFct.setError("ALREADY_DESTROYED")
				end
			end

			for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
				localMembers.userPrivateDatas[_localModuleMembers.name] = {}
			end

			localMembers.idTimer = timer.scheduleFunction(ATMETimers.C_Alarm_timer, newAlarm, timeStart)
			ATMEFct.outputCore(nil, "Alarme créée" , 2)

			return false, newAlarm
		else
			ATMEFct.outputCore(nil, "Alarme startTime < current time : Alarme non crée", 2)
			return false
		end
	else
		return true, timeStart
	end
end

ATMEFct.C_Alarm_delete = function(_alarms, _alarm)
	local name = _alarm:getName()
	
	if name ~= nil then
		_alarms[name] = nil
	end
	
	local localMembers = ATMEDatas.localMembers[_alarm:getIndex()]
	localMembers.destroyed = true
	
	if localMembers.triggered == false then
		timer.removeFunction(localMembers.idTimer)
	end
	
	localMembers.triggered = false
	ATMEDatas.C_Alarm_MT:remove(_alarm:getIndex())
	
	ATMEFct.outputCore(nil, "Alarme détruite", 2)
end


setmetatable(ATME.C_Alarm,ATMEDatas.C_Alarm_MT)

ATMEDatas.C_Alarm_MT.__call = function(t, ...)
	local fctDefinition = "Constructor ATME.C_Alarm(rules)\nATME.C_Alarm(flagID, rules)"
	local args = {...}
	
	if #args == 1 then
		if type(args[1]) == "number" or type(args[1]) == "string" then
			local err, newAlarm = ATMEFct.C_Alarm_create(nil, nil, nil, args[1])
			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, newAlarm, false)
				return
			end
			return newAlarm
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "rules type must be number or string", false)
		end
	elseif #args >= 2 then
		if (type(args[1]) == "number" or type(args[1]) == "string") and (type(args[2]) == "number" or type(args[2]) == "string") then
			local flag
			
			if type(args[1]) == "number" then
				flag = math.floor(args[1])
				
				if flag > 0 then
					flag = tostring(flag)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "flagID must be > 0", false)
					return
				end
			else
				flag = args[1]
			end
			
			local err, newAlarm = ATMEFct.C_Alarm_create(nil, nil, flag, args[2])
		
			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, newAlarm, false)
				return
			end
			
			return newAlarm
		else
			if type(args[1]) ~= "number" and type(args[1]) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "flagID type must be number or string", false)
			elseif type(args[2]) ~= "number" and type(args[2]) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "rules type must be number or string", false)
			end
		end
	end
end

ATMEDatas.C_Alarm_MT.mt.__metatable = "Private"
ATMEDatas.C_Alarm_MT.mt.__index = ATME.C_Alarm
ATMEDatas.C_Alarm_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Alarm_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Alarm : Bad key " .. k, "table locked", false)
	end
end


----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Event : La classe ATME.C_Event permet la création d'instance events Core.
-- Pas de contructeur accessible aux modules externes. instances entièrement gérées par ATME.
---------------------------------------------------------------------------------------------------------- 

ATMEDatas.C_Event_MT = {
	className = "C_Event",
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getType"] = true,
		["getDatas"] = true,
	},
	
	add = function(self, _item)
		local item = {
			type = _item.type,
			datas = _item.datas,
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
	end,
}

ATME.C_Event = {
	getClassName = function(self)
		return ATMEDatas.C_Event_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Event_MT.parentClass
	end,
}

setmetatable(ATME.C_Event, ATMEDatas.C_Event_MT)

ATMEDatas.C_Event_MT.mt.__metatable = "Private"
ATMEDatas.C_Event_MT.mt.__index = ATME.C_Event
ATMEDatas.C_Event_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Event_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Event : Bad key " .. k, "table locked", false)
	end
end

ATMEFct.C_Event_create = function(_item)
	local newEvent = {}
	
	setmetatable(newEvent, ATMEDatas.C_Event_MT.mt)
	local index = ATMEDatas.C_Event_MT:add(_item)
	local localMembers = ATMEDatas.localMembers[index]
	
	function newEvent.getIndex()
		return index
	end
	
	function newEvent.getType(self)
		return localMembers.type
	end
	
	function newEvent.getDatas(self)
		return ATMEConst.coreEventTypes[localMembers.type](localMembers)
	end
	
	return newEvent
end

-- Définition du deux events mgr : la bascule se fait dans la fonction main timer
-- Ceci permet si des événements sont créés dans cette fonction, de les traiter
-- au coup suivant. Dans main timer, activeEvents passe de 1 à 0 et vice versa.

local events = {
	[0] = ATMEFct.C_IndexList_create(),
	[1] = ATMEFct.C_IndexList_create(),
}

local activeEvents = 0

ATMEFct.C_Event_add = function(_item)
	local newEvent = ATMEFct.C_Event_create(_item)
	events[activeEvents]:add(newEvent)
end

ATMEFct.C_Event_removeAll = function(_id)
	events[_id]:removeAll()
end

ATMEFct.C_Event_fire = function(_event)
	local localEventMembers = ATMEDatas.localMembers[_event:getIndex()]
			
	local name = localEventMembers.datas.unitName
	
	if ATMEDatas.allAIUnitsAndPlayers[name] ~= nil then
		local localMembersUnit = ATMEDatas.localMembers[ATMEDatas.allAIUnitsAndPlayers[name]:getIndex()]
		
		if localEventMembers.type == "TRANSPORT_END_OF_DISEMBARK" then
			local localMembersGroup = ATMEDatas.localMembers[localEventMembers.datas.linkedGroup:getIndex()]
			
			localMembersGroup.disembarking = false
		end
		
		local moduleName = ATMEDatas.currentModuleName
		for _moduleName, _callbacks in pairs(localMembersUnit.eventCallbacks) do
			ATMEDatas.currentModuleName = _moduleName

			if _callbacks[localEventMembers.type] ~= nil then
				ATMEFct.outputCore(nil,"Appel event (type " .. localEventMembers.type .. ") callback de l'unité " .. name .. " pour le module " .. _moduleName, 2)
				xpcall(function() _callbacks[localEventMembers.type](ATMEDatas.allAIUnitsAndPlayers[name], _event) end, function() ATMEFct.errUserHandler('eventCallbackHandler') end)
			end
		end
		
		ATMEDatas.currentModuleName = moduleName
	end
end

ATMEFct.C_Event_target_class = {
	["C_AIUnit"] = function(_targetName)
		return ATMEDatas.allAIUnitsAndPlayers[_targetName]
	end,
	["C_Player"] = function(_targetName)
		return ATMEDatas.allAIUnitsAndPlayers[_targetName]
	end,
	["C_StaticObject"] = function(_targetName)
		return ATMEDatas.staticObjects[_targetName]
	end,
	["C_Cargo"] = function(_targetName)
		return ATMEDatas.cargos[_targetName]
	end,
	["C_SceneryObject"] = function(_targetName)
		return ATMEDatas.sceneryObjects[_targetName]
	end,
}

ATMEFct.C_Event_update = function(_event, _remove)
	local localEventMembers = ATMEDatas.localMembers[_event:getIndex()]

	local name = localEventMembers.datas.unitName
	
	if name ~= nil then
		if ATMEDatas.allAIUnitsAndPlayers[name] == nil then
			if _remove == true then
				events[activeEvents]:remove(_id)
				return
			end
		end
	end
	
	if localEventMembers.datas.linkedGroupName ~= nil then  -- Type string
		if ATMEDatas.groups[localEventMembers.datas.linkedGroupName] ~= nil then
			-- Traitement des cas type TRANSPORT_END_OF_DISEMBARK
			-- Associe le groupe à partir de son nom si il existe toujours au moment du traitement.
		
			localEventMembers.datas.linkedGroup = ATMEDatas.groups[localEventMembers.datas.linkedGroupName]
			
		elseif _remove == true then
			events[activeEvents]:remove(_id)
			return
		end
	end

	if localEventMembers.datas.targetName ~= nil then -- Type string ou number suivant les cas
		local target = ATMEFct.C_Event_target_class[localEventMembers.datas.targetClassName](localEventMembers.datas.targetName)
		if target ~= nil then
			-- Traitement des cas SPOT_NEW_TARGET, SPOT_TARGET_OUT_OF_ZONE, SPOT_TARGET_NO_LOS,
			-- TARGET_NEW_TRACK, TARGET_END_TRACK, HIT_TARGET, SHOOTING_START, SHOT
			
			localEventMembers.datas.target = target
			
			if ATMEDatas.weapons[localEventMembers.datas.weaponName] ~= nil then
				localEventMembers.datas.weapon = ATMEDatas.weapons[localEventMembers.datas.weaponName]
			end
		elseif _remove == true then
			events[activeEvents]:remove(_id)
			return
		end
	end
	
	if localEventMembers.datas.weaponName ~= nil then -- Type number
		if ATMEDatas.weapons[localEventMembers.datas.weaponName] ~= nil then
			localEventMembers.datas.weapon = ATMEDatas.weapons[localEventMembers.datas.weaponName]
		elseif _remove == true then
			events[activeEvents]:remove(_id)
		end
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Module : gère les modules externes. Contructeur accessible aux modules externes.
-- Tout module externe doit créer une instance de cette classe afin de pouvoir activer les handlers définis
-- dans le module externe. Un handler à nil, donc non défini, ne sera pas activé par ATME_Core.
-- Cette classe offre des fonction liées au event ATME et à leurs données, permet la création des triggers
-- utilisateurs dont les identifiants sont spécifiques à un module et ajoute des fonction de debug.
--
-- Il est à noter que tous les appels xpcall au travers de localMembers (Instance de ATME.C_IndexList) se font
-- toujours dans l'ordre de création des instance de cette classe. Ainsi, un module ayant besoin d'un autre
-- module devra être défini dans les triggers après le module qu'il utilise pour que le bon ordre d'appel
-- soit respecté.
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Module_MT = {
	className = "C_Module",
	mt = {},
	
	keys = {
		["error"] = true,
		["output"] = true,
		["outputVar"] = true,
		["getName"] = true,
		["getIndex"] = true,
		["isDebugOn"] = true,
	},
	
	add = function(self, _name, _newHandlers)
		local item = {
			alarms = {},
			zonesMgr = {},
			tasks = {},
			lines2D = {},
			vectors3D = {},
			segments2D = {},
			zones = {},
			cloningContexts = {},
			staticCloningContexts = {},
			cargoCloningContexts = {},
			findContexts = {},
			radioChannels = {},
			name = _name,
			id = #ATMEDatas.modulesNames + 1,
			handlers = _newHandlers,
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].alarms = nil
		ATMEDatas.localMembers[_id].namedAlarms = nil
		ATMEDatas.localMembers[_id].handlers = nil
		ATMEDatas.localMembers[_id].zonesMgr = nil
		ATMEDatas.localMembers[_id].tasks = nil
		ATMEDatas.localMembers[_id].lines2D = nil
		ATMEDatas.localMembers[_id].vectors3D = nil
		ATMEDatas.localMembers[_id].segments2D = nil
		ATMEDatas.localMembers[_id].cloningContexts = nil
		ATMEDatas.localMembers[_id].zones = nil
		ATMEDatas.localMembers[_id].staticCloningContexts = nil
		ATMEDatas.localMembers[_id].cargoCloningContexts = nil
		ATMEDatas.localMembers[_id].findContexts = nil
		ATMEDatas.localMembers[_id].radioChannels = nil
	end,
}

ATME.C_Module = {
	getClassName = function(self)
		return ATMEDatas.C_Module_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Module_MT.parentClass
	end,
	
	isLoaded = function(_name)
		local fctDefinition = "ATME.C_Module.isLoaded(name)"
		
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[_name] ~= nil
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", true)
		end
	end,
	
	getNames = function()
		local names = {}
		for _i, _name in ipairs(ATMEDatas.modulesNames) do
			names[_i] = _name
		end
		
		return names
	end
}

ATMEFct.C_Module_create = function(_name, _handlers, _debugSwitch)		   
    local newHandlers = {}
   
    for _id, _ in pairs(_handlers) do
		newHandlers[_id] = _handlers[_id]
	end
	
	local newModule = {}
	setmetatable(newModule, ATMEDatas.C_Module_MT.mt)
	
	local debugSwitch = _debugSwitch
	local index = ATMEDatas.C_Module_MT:add(_name, newHandlers)
	local localMembers = ATMEDatas.localMembers[index]
	
	local err = debug.traceback()

	local referenceFile = ATMEFct.searchMainChunk(err)
	
	ATMEDatas.refModules[referenceFile] = _name
	ATMEDatas.modules[_name] = newModule
	ATMEDatas.modulesLocalMembers[_name] = localMembers
	table.insert(ATMEDatas.modulesNames, _name)
	
	if ATME.modules[_name] == nil then
		ATME.modules[_name] = {}
	end
	
	function newModule.getName(self)	
		return localMembers.name
	end
	
	function newModule.getIndex(self)
		return index
	end
	
	function newModule.isDebugOn(self)
		return debugSwitch
	end
	
	function newModule.output(self, _text, _level)
		local fctDefinition = ATMEFct.getErrorPrefix(newModule) .. "output(text, [level])"
		
		if type(_text) == "string" and (_level == nil or type(_level) == "number") then
			local level = 1
			if type(_level) == "number" then
				level = math.floor(_level)
			end
			
			if level >= 0 then
				ATMEFct.outputCore(self, _text, level)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "level must be >= 0", false)
			end
		else
			if type(_text) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", true)
			elseif _level ~= nil and type(_level) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "level type must be nil or number", false)
			end
		end
	end
	
	function newModule.outputVar(self, _title, _var, _level)
		local fctDefinition = ATMEFct.getErrorPrefix(newModule) .. "outputVar(text, var, [level])"
		
		if type(_title) == "string" and (_level == nil or type(_level) == "number") then
			local level = 1
			if type(_level) == "number" then
				level = math.floor(_level)
			end
			
			if level >= 0 then
				ATMEFct.outputVarCore(self, _title, _var, level)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "level must be >= 0", false)
			end
		else
			if type(_title) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "_title type must be string", true)
			elseif _vel ~= nil and type(_level) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "level type must be nil or number", false)
			end
		end
	end
	
	function newModule.error(self, _text)
		local fctDefinition = ATMEFct.getErrorPrefix(newModule) .. "error(text)"
		
		if type(_text) == "string" then
			ATMEFct.errorOutputCore(self, "test", _text, false)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", true)
		end
	end
	
	ATMEFct.outputCore(nil, "Module " .. localMembers.name .." -> loaded", 2)
		
	return newModule			
end

ATMEDatas.C_Module_MT.__call = function(t, _name, _handlers, _debugSwitch)
	local fctDefinition = "Constructor ATME.C_Module(name, handlers, debugSwitch)"
	
	if type(_name) == "string" and type(_handlers) == "table" and type(_debugSwitch) == "boolean" then
		if ATMEDatas.isRunning == false then
			if ATMEDatas.modulesLocalMembers[_name] == nil then
				for _id, _ in pairs(_handlers) do
					if ATMEConst.moduleHandlers[_id] ~= nil then
						if type(_handlers[_id]) ~= "function" then
							ATMEFct.errorOutputCore(nil, fctDefinition, "Module " .. _name .. " - Handler '" .. _id .. " must be nil or function", false)
							return
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Module " .. _name .. " - Handler bad id : '" .. _id, false)
						return
					end
				end
					
				return ATMEFct.C_Module_create(_name, _handlers, _debugSwitch)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Module " .. _name .. " already loaded", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "Module " .. _name .. " is not created - ATME is running", false)
		end
	else
		if type(_name) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
		elseif type(_handlers) ~= "table" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "handlers type must be table of functions", false)
		elseif type(_debugSwitch) ~= "boolean" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "debugSwitch type must be boolean", false)
		end
	end
end

setmetatable(ATME.C_Module,ATMEDatas.C_Module_MT)

ATMEDatas.C_Module_MT.mt.__metatable = "Private"
ATMEDatas.C_Module_MT.mt.__index = ATME.C_Module
ATMEDatas.C_Module_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Module_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Module : Bad key " .. k, "table locked", false)
	end
end


----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Smoke : gère les fumées. Contructeur accessible aux modules externes.
-- Une fumée peut être déclenché simplement. DCS l'arrêtera au bout de 300s (5 minutes). Elle peut être également
-- redéclenchée au bout de cette durée jusqu'à un arrêt volontaire. 
-- La couleur est définie par une string dans ATME.
-- On notera qu'il est aussi possible de déclencher une fumée à partir d'un unité joueur, AI, ou d'un objet statique.
----------------------------------------------------------------------------------------------------------

ATMETimers.C_Smoke_timer = function(_localMembers)
	if _localMembers.activated == true then
		if _localMembers.always == true then
			trigger.action.smoke(_localMembers.point, _localMembers.color)
			
			return timer.getTime() + ATMEConst.maxTimeSmoke
		else
			_localMembers.activated = false
		end
	end
end

ATMEDatas.C_Smoke_MT = {
	className = "C_Smoke",
	mt = {},
	
	keys = {
		["activate"] = true,
		["stop"] = true,
		["getIndex"] = true,
		["getUserPrivateDatas"]= true,
	},
	
	add = function(self, _point, _color)
		local item = {
			point = _point,
			color = _color,
			always = false,
			activated = false,
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Smoke = {
	getClassName = function(self)
		return ATMEDatas.C_Smoke_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Smoke_MT.parentClass
	end,
}

ATMEFct.C_Smoke_create = function(_color, _point)
	local newSmoke = {}
	local color

	setmetatable(newSmoke, ATMEDatas.C_Smoke_MT.mt)

	if ATMEConst.smokeColors[_color] == -1 then
		color = ATMEConst.smokeColors[ATMEConst.smokeColorsIndex[math.random(1,#ATMEConst.smokeColorsIndex)]]
	else
		color = ATMEConst.smokeColors[_color]
	end
	
	local index = ATMEDatas.C_Smoke_MT:add(_point, color)
	local localMembers = ATMEDatas.localMembers[index]
	
	function newSmoke.activate(self, _always)
		local fctDefinition = ATMEFct.getErrorPrefix(newSmoke) .. "activate([always])"
		
		if _always == nil or type(_always) == "boolean" then
			localMembers.always = _always
			if _always == nil then
				localMembers.always = false
			end
			
			if localMembers.activated == false then
				localMembers.activated = true
				trigger.action.smoke(localMembers.point, localMembers.color)
				timer.scheduleFunction(ATMETimers.C_Smoke_timer, localMembers, timer.getTime() + ATMEConst.maxTimeSmoke)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "always type must be nil or boolean", true)
		end
	end

	function newSmoke.stop(self)
		localMembers.activated = false
		localMembers.always = false
	end
	
	function newSmoke.getIndex(self)
		return index
	end
	
	function newSmoke.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newSmoke
end

setmetatable(ATME.C_Smoke,ATMEDatas.C_Smoke_MT)

ATMEDatas.C_Smoke_MT.__call = function(t, _color, _reference)
	local fctDefinition = "Constructor ATME.C_Smoke(color, reference)"
	
	if type(_color) == "string" and ATMEFct.isReference(_reference) == true then
		local color = string.upper(_color)
		if ATMEConst.smokeColors[color]~= nil then
			return ATMEFct.C_Smoke_create(color, ATMEFct.getPointFromReference(_reference))
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a smoke color or 'RANDOM'", false)
		end
	else
		if type(_color) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", false)
		elseif ATMEFct.isReference(_reference) == false then
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		end
	end
end

ATMEDatas.C_Smoke_MT.mt.__metatable = "Private"
ATMEDatas.C_Smoke_MT.mt.__index = ATME.C_Smoke
ATMEDatas.C_Smoke_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Smoke_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Smoke : Bad key " .. k, "table locked", false)
	end
end


----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Flare : gère les flares de couleur. Contructeur accessible aux modules externes.
-- Un tir de flare peut être déclenché unitairement. Il peut être aussi redéclenché automatiquement et aléatoirement
-- sur une durée souhaitée ou jusqu'à un arrêt volontaire.
-- La couleur est définie par une string dans ATME. Il existe une couleur "RANDOM" pour un choix aléatoire.
-- On notera qu'il est aussi possible de déclencher un flare à partir d'un unité joueur, AI, ou d'un objet statique.
----------------------------------------------------------------------------------------------------------

ATMETimers.C_Flare_timer = function(_localMembers)
	if _localMembers.activated == true then
		local color
			
		if ATMEConst.flareColors[_localMembers.colorName] == -1 then
			color = ATMEConst.flareColors[ATMEConst.flareColorsIndex[math.random(1, #ATMEConst.flareColorsIndex)]]
		else
			color = ATMEConst.flareColors[_localMembers.colorName]
		end
		
		trigger.action.signalFlare(_localMembers.point, color, math.random() * 2 * math.pi)
		
		if _localMembers.always == true then
			return timer.getTime() + math.random(_localMembers.minTimer, _localMembers.maxTimer)
		else
			_localMembers.activated = false
		end
	end
end

ATMEDatas.C_Flare_MT = {
	className = "C_Flare",
	mt = {},
	
	keys = {
		["activate"] = true,
		["stop"] = true,
		["getIndex"] = true,
		["getUserPrivateDatas"]= true,
	},
	
	add = function(self, _point, _colorName)
		local item = {
			point = _point,
			colorName = _colorName,
			activated = false,
			always = false,
			minTimer = 1,
			maxTimer = 4,
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Flare = {
	getClassName = function(self)
		return ATMEDatas.C_Flare_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Flare_MT.parentClass
	end,
}

ATMEFct.C_Flare_create = function(_color, _point)	
	local newFlare = {}
	
	setmetatable(newFlare, ATMEDatas.C_Flare_MT.mt)
	local index = ATMEDatas.C_Flare_MT:add(_point, string.upper(_color))
	local localMembers = ATMEDatas.localMembers[index]
	
	function newFlare.activate(self, _always)
		local fctDefinition = ATMEFct.getErrorPrefix(newSmoke) .. "activate([always])"
		
		if _always == nil or type(_always) == "boolean" then
			localMembers.always = _always
			if _always == nil then
				localMembers.always = false
			end
			
			if localMembers.activated == false then
				localMembers.activated = true
				timer.scheduleFunction(ATMETimers.C_Flare_timer, localMembers, timer.getTime() + 1)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "always type must be nil or boolean", true)
		end
	end
	
	function newFlare.stop(self)
		localMembers.activated = false
		localMembers.always = false
	end
	
	function newFlare.getIndex(self)
		return index
	end
	
	function newFlare.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newFlare
end

setmetatable(ATME.C_Flare,ATMEDatas.C_Flare_MT)

ATMEDatas.C_Flare_MT.__call = function(t, _color, _reference)
	local fctDefinition = "Constructor ATME.C_Flare(color, reference)"
	
	if type(_color) == "string" and ATMEFct.isReference(_reference) == true then
		if ATMEConst.flareColors[string.upper(_color)] ~= nil then
			return ATMEFct.C_Flare_create(_color, ATMEFct.getPointFromReference(_reference))
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a flare color or 'RANDOM'", false)
		end
	else
		if type(_color) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", false)
		elseif ATMEFct.isReference(_reference) == false then
			ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
		end
	end
end


ATMEDatas.C_Flare_MT.mt.__metatable = "Private"
ATMEDatas.C_Flare_MT.mt.__index = ATME.C_Flare
ATMEDatas.C_Flare_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Flare_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Flare : Bad key " .. k, "table locked", false)
	end
end


-----------------------------------------------------------------------------------------------------------
-- Définition des fonctions indirectes ATMEAIIndFct liées aux tâches et définies dans la table 
-- ATMEConst.taskAI. Ces fonctions retournent toujours une table, qui comporte un ou plusieurs éléments 
-- indexés. Si plusieurs éléments sont retournées, les paramètres associés à linkedParams, définis dans
-- la table ATMEConst.taskAI seront initialisés.
-----------------------------------------------------------------------------------------------------------

ATMEAIIndFct.C_testFormation = function(_arg)	
	if ATMEConst.badOptionAIFormationValues[_arg] == nil then
		return {args}, 0
	end
end

ATMEAIIndFct.C_getNumberWithFlag = function(_arg)
	return {_arg, true}, 0
end

ATMEAIIndFct.C_Circle_getCenterRadius = function(_arg)
	local point = _arg:getCenter()
	local rMax = _arg:getRadius()
	
	return {{x = point.x, y = point.z}, rMax}, 0
end

ATMEAIIndFct.C_Group_getID = function(_arg)
	localMembers = ATMEDatas.localMembers[_arg:getIndex()]
	return {localMembers.spawnDatas.datas.groupId}, 0
end

ATMEAIIndFct.C_Group_getIDWithWP = function(_arg)
	localMembers = ATMEDatas.localMembers[_arg:getIndex()]

	return {localMembers.spawnDatas.datas.groupId, true, #localMembers.waypoints}, 0
end

ATMEAIIndFct.C_Unit_getID = function(_arg)
	return {_arg:getDCSUnit():getID()}, 0
end

ATMEAIIndFct.C_getWeaponFlag = function(_arg)
	local value = 0
	
	for _item in _arg:gmatch("([^%s]+)") do
		if ATMEConst.weaponFlags[string.upper(_item)] ~= nil then
			value = value + ATMEConst.weaponFlags[string.upper(_item)]
		else
			return {}, 0
		end
	end

	return {value}, 0
end

ATMEAIIndFct.C_getExpend = function(_arg)	
	return {ATMEConst.weaponExpend[string.upper(_arg)]}, 0
end

ATMEAIIndFct.C_getDesignation = function(_arg)	
	return {ATMEConst.designationAI[string.upper(_arg)]}, 0
end

ATMEAIIndFct.C_getTargetsType = function(_arg)
	return {ATMEConst.targetsAttributes[string.upper(_arg)]}, 0
end

ATMEAIIndFct.C_getModulation = function(_arg)
	local value = string.upper(_arg)
	
	if value == "FM" then
		return {radio.modulation.FM}, 0
	elseif value == "AM" then
		return {radio.modulation.AM}, 0
	end
end

ATMEAIIndFct.C_getCallname = function(_arg)
	local tokens = ATMEFct.analyzeString(string.upper(_arg))
	
	if #tokens == 2 and tokens[1].type == "TOKEN_JTAC_CALLSIGN_NAME" and tokens[2].type == "TOKEN_INTEGER" then
		return {ATMEConst.blueCallsigns.jtacs[tokens[1].value].id, tokens[2].value}, 1
	elseif #tokens == 2 and tokens[1].type == "TOKEN_AIRCRAFT_CALLSIGN_NAME" and tokens[2].type == "TOKEN_INTEGER" then
		return {ATMEConst.blueCallsigns.aircrafts[tokens[1].value].id, tokens[2].value}, 2
	end
end

ATMEAIIndFct.C_getPoint2D = function(_arg)
	if ATMEFct.isObjectClass(_arg, ATME.C_Unit) or ATMEFct.isObjectClass(_arg, ATME.C_Group) then
		local pos = _arg:getPosition()
		return {{x = pos.x, y = pos.z}}, 0
	elseif _arg.x ~= nil and _arg.y ~= nil then
		local point = ATMEFct.convertToPoint3D(_arg)
		return {{x = point.x, y = point.z}}, 0
	end
end

ATMEAIIndFct.C_getOrbitPattern = function(_arg)
	return {ATMEConst.orbitPattern[string.upper(_arg)]}, 0
end	

ATMEAIIndFct.C_getAirbase = function(_arg)
	if ATMEDatas.airbases[_arg] ~= nil then
		return {ATMEDatas.airbases[_arg]:getID()}, 0
	end
end

ATMEAIIndFct.C_Vector_getPos = function(_arg)
	return {{x = _arg:getX(), y = _arg:getY(), z = _arg:getZ()}}, 0
end

ATMEFct.testDatas = function(_data, _param)
	if _data ~= nil then
		if _param.enum ~= nil then
			local data = _data
			if type(data) == "string" then
				data = string.upper(_data)
			end
			if _param.enum[data] ~= nil then
				return _param.enum[data]
			end
		elseif type(_data) == "number" then
			if _param.max ~= nil and type(_data) == "number" then
				if _data > _param.max then
					return
				end
			end
			
			if _param.min ~= nil and type(_data) == "number" then
				if _data < _param.min then
					return
				end
			end
			
			return _data
		else
			return _data
		end
	end
end

-----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Task : gère la création dynamique des taches associées ensuite à des WP ou à des groupes
-- d'unités.
-----------------------------------------------------------------------------------------------------------

-- Analyse des paramètres pour une tache :
-- _id = 1 pour les paramètres obligatoires (voir déclaration de taskAI)
-- _id = 2 pour les paramètres optionnels (voir déclaration de taskAI)
--
-- Cette analyse se fait argument par argument lié à {...}
-- Le type des arguments est testé.
-- Un argument peut représenter plusieurs paramètres de la tâche (exemple argument cercle qui renseigne
-- un centre et un rayon. Chaque paramètre peut faire appel à une fonction indirecte spécifique ou pas.
-- Un argument peut également lié plusieurs paramètres de la tâche, dans ce cas, la fonction indirecte
-- doit exister et retourner une table avec plusieurs valeurs, une par paramètre.
-- Enfin, chaque paramètre est testé en terme de valeur sur une plage min/max ou énumération.

ATMEFct.C_Task_createNamed = function(_namedTable, _name, _task, _datas)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return false
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			local taskDatas = {}
			local err, usedOnZoneOnly, filters, taskType = ATMEFct.C_Task_setParams(_task, taskDatas, _datas)
			
			if err == nil then
				return false, ATMEFct.C_Task_create(_name, _namedTable, _task, usedOnZoneOnly, taskDatas, filters, taskType)
			else
				return true, err
			end
		else
			return true, "name can't contain '|' character"
		end
	end
end

ATMEFct.C_Task_setArg = function(_name, _arg, _item, _datas)
	if _arg.type == type(_item) or ATMEFct.isObjectClass(_item, _arg.type) == true then
		if _arg.parameter.fct ~= nil then
			local values, taskT = ATMEAIIndFct[_arg.parameter.fct](_item)	-- taskT : Uniquement utilisé pour le callsign aircraft/jtacs sol
																			-- 		   taches liées au JTAC (1 sol, 2 air), sinon 0
			
			if #values == 0 then
				return true, _name .. " - Bad token '" .. _item .. "'"
			end

			local value = ATMEFct.testDatas(values[1], _arg)
			if value == nil then
				return true, _name .. " - Bad value"
			end
			
			table.insert(_datas, value)
			
			return false, taskT
		else
			local value = ATMEFct.testDatas(_item, _arg)
			if  value == nil then
				return true, _name .. " - Bad value"
			end
			
			table.insert(_datas, value)
			
			return false, 0
		end
	else
		return true, _name .. " - Bad type"
	end
end

ATMEFct.C_Task_setParams = function(_task, _taskDatas, _datas)
	local taskID = string.upper(_task)
	local taskType = 0
	local usedOnZoneOnly = false

	if ATMEConst.taskAI[taskID] ~= nil then
		local index = 1
		
		_taskDatas["id"] = ATMEConst.taskAI[taskID].id
		_taskDatas["params"] = {}
		
		if ATMEConst.taskAI[taskID].const ~= nil then
			for _name, _value in pairs(ATMEConst.taskAI[taskID].const) do
				_taskDatas["params"][_name] = _value
			end
		end
			
		for _argName, _arg in pairs(ATMEConst.taskAI[taskID].args) do
			if _datas ~= nil and _datas[_argName] ~= nil then
				if _arg.multi == true then
					local datas = {}
					if type(_datas[_argName]) == "string" then
						for _item in _datas[_argName]:gmatch("([^%s]+)") do
							local err, value = ATMEFct.C_Task_setArg(_argName, _arg, _item, datas)
							
							if err == true then
								return value
							else
								taskType = value
							end
						end
					elseif type(_datas[_argName]) == "table" then
						for _, _item in ipairs(_datas[_argName]) do
							local err, value = ATMEFct.C_Task_setArg(_argName, _arg, _item, datas)
							
							if err == true then
								return value
							else
								taskType = value
							end
						end
					else
						return _argName .. " - Bad type"
					end

					_taskDatas["params"][_arg.parameter.name] = datas
				elseif type(_datas[_argName]) == _arg.type or ATMEFct.isObjectClass(_datas[_argName], _arg.type) == true then
					if _arg.parameter.fct ~= nil then
						local values, taskT = ATMEAIIndFct[_arg.parameter.fct](_datas[_argName])	-- taskT : Uniquement utilisé pour le callsign aircraft/jtacs sol
																									-- 		   taches liées au JTAC (1 sol, 2 air), sinon 0
						
						taskType = taskT
						
						if #values == 0 then
							return _argName .. " - Bad value"
						end
						
						for _valueId, _value in ipairs(values) do
							local value = ATMEFct.testDatas(_value, _arg)
							if value ~= nil then
								if _valueId == 1 then
									_taskDatas["params"][_arg.parameter.name] = value
								else
									_taskDatas["params"][_arg.parameter.linkedParams[_valueId - 1]] = value
								end
							else
								return _argName .. " - Bad value"
							end
						end
					elseif ATMEFct.isObjectClass(_datas[_argName], _arg.type) == false then
						local value = ATMEFct.testDatas(_datas[_argName], _arg)
						if value ~= nil then
							_taskDatas["params"][_arg.parameter.name] = value
						else
							return _argName .. " - Bad value"
						end
					else
						return _argName .. " - Bad value"
					end
				end
			elseif _arg.mandatory == true then
				return _argName .. " - Missing mandatory data"
			end
		end
		
		return nil, usedOnZoneOnly, ATMEConst.taskAI[taskID].filters, taskType
	else
		return "Task '" .. _task .. "' doesn't exist"
	end
end

ATMEFct.C_Task_getTaskDatas = function(_task, _type, _stopConditionOk)
	local localMembers = ATMEDatas.localMembers[_task:getIndex()]
	

	if localMembers.taskType == 0 or ATMEConst.JTACStatus[_type] == localMembers.taskType then
		local ok = false
		local taskDatas = ATMEFct.copyTable(localMembers.datas)
	
		for _id, _ in pairs(localMembers.filters.type) do
			if _id == _type then
				ok = true
			end
		end
		
		if ok == true then
			if (localMembers.stopCondition ~= nil or localMembers.startCondition ~= nil) and _stopConditionOk == true then
				local temp = taskDatas
				taskDatas = { 
					["id"] = "ControlledTask", 
					["params"] = { 
						task = temp,
					} 
				}
				
				if localMembers.stopCondition ~= nil then
					taskDatas.params.stopCondition = ATMEFct.copyTable(localMembers.stopCondition) 
				end
				
				if localMembers.startCondition ~= nil then
					taskDatas.params.condition = ATMEFct.copyTable(localMembers.startCondition)
				end
			end
			
			return true, taskDatas
		else
			return false, "'" .. taskDatas.id .. "' is not allowed"
		end
	end
	
	return false, "'" .. taskDatas.id .. "' bad JTAC callsign"
end

ATMEDatas.C_Task_MT = {
	className = "C_Task",
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["addStopCondition"] = true,
		["removeStopCondition"] = true,
		["addStartCondition"] = true,
		["removeStartCondition"] = true,
		["getUserPrivateDatas"] = true,
	},
	
	add = function(self, _taskID, _usedOnZoneOnly, _taskDatas, _filters, _taskType)
		local item = {
			taskID = _taskID,
			datas = _taskDatas,
			filters = _filters,
			taskType = _taskType,
			usedOnZoneOnly = _usedOnZoneOnly,
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Task = {
	getClassName = function(self)
		return ATMEDatas.C_Task_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Task_MT.parentClass
	end,
	
	createNamed = function(_name, _task, _datas)
		local fctDefinition = "Function ATME.C_Task.createNamed(name, taskID, datas)"
		
		if type(_name) == "string" and type(_task) == "string" then
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newTask = ATMEFct.C_Task_createNamed(localMembersModule.tasks, _name, _task, _datas)				
			
			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, newTask, false)
				return
			end
			
			return newTask
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif type(_task) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "taskID type must be string", false)
			end
		end
	end,
	
	getByName = function(_name)
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].tasks[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Task.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_Task_create = function(_name, _tasks, _taskID, _usedOnZoneOnly, _taskDatas, _filters, _taskType)
	local newTask = {}

	setmetatable(newTask, ATMEDatas.C_Task_MT.mt)
	
	local name
	local tagsName
	
	if _name ~= nil and _tasks ~= nil then
		name = _name
		_tasks[name] = newTask
		tagsName = ATMEFct.decompString(_name)
	end
	
	local index = ATMEDatas.C_Task_MT:add(_taskID, _usedOnZoneOnly, _taskDatas, _filters, _taskType)
	local localMembers = ATMEDatas.localMembers[index]
	
	function newTask.getName(self)
		return name
	end
	
	function newTask.getIndex(self)
		return index
	end
	
	function newTask.addStopCondition(self, _condition)
		local fctDefinition = ATMEFct.getErrorPrefix(newTask) .. "addStopCondition(condition)"
		
		if type(_condition) == "string" then
			local stopCondition = {}
			local tokens = ATMEFct.analyzeString(_condition)

			if #tokens == 2 then
				if tokens[1].type == "TOKEN_TYPE_TIME" then
					if tokens[2].type == "TOKEN_INTEGER" then
						if  tokens[2].value > 0 then
							if tokens[2].status == "RELATIVE" then
								stopCondition.duration = tokens[2].value
							else
								stopCondition.time = tokens[2].value
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be > 0 in string format condition", false)
							return
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be integer or time value in string format condition", false)
						return
					end
				elseif tokens[1].type == "TOKEN_TYPE_WAYPOINT" then
					if tokens[2].type == "TOKEN_INTEGER" then
						if tokens[2].value >= 0 then
							stopCondition.lastWaypoint = tokens[2].value
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be >= 0 in string format condition", false)
							return
						end
					elseif tokens[2].type == "TOKEN_STRING" then
						stopCondition.lastWaypoint = tokens[2].value
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be integer or string in string format condition", false)
						return
					end
				elseif tokens[1].type == "TOKEN_TYPE_FLAG" then
					stopCondition.userFlagValue = true
			
					if tokens[2].type == "TOKEN_INTEGER" then
						if tokens[2].value > 0 then
							stopCondition.userFlag = string.format("%d", tokens[2].value)
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be > 0 in string format condition", false)
							return
						end
					elseif tokens[2].type == "TOKEN_STRING" then
						stopCondition.userFlag = tokens[2].value
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be integer or string in string format condition", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Bad string format condition", false)
					return
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Bad string format condition", false)
				return
			end
			
			if localMembers.stopCondition == nil then
				localMembers.stopCondition = stopCondition
			else
				for _id, _item in pairs(stopCondition) do
					localMembers.stopCondition[_id] = _item
				end
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "condition type must be string", true)
		end
	end
	
	function newTask.removeStopCondition(self)
		localMembers.stopCondition = nil
	end
	
	function newTask.addStartCondition(self,_condition)
		local fctDefinition = ATMEFct.getErrorPrefix(newTask) .. "addStartCondition(condition)"
		
		if type(_condition) == "string" then
			local startCondition = {}
			local tokens = ATMEFct.analyzeString(_condition)

			if #tokens == 2 then
				if tokens[1].type == "TOKEN_TYPE_TIME" then
					if tokens[2].type == "TOKEN_INTEGER" and tokens[2].value > 0 then
						if tokens[2].status == "RELATIVE" then
							ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be an absolute time value", false)
							return
						else
							startCondition.time = tokens[2].value
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be integer or time value in string format condition", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Bad string format condition", false)
					return
				end
			elseif #tokens == 3 then
				if tokens[1].type == "TOKEN_TYPE_FLAG" then
					startCondition.userFlagValue = false
					if tokens[3].type == "TOKEN_INTEGER" and tokens[3].value >= 0 then
						if tokens[3].value > 0 then
							startCondition.userFlagValue = true
						end
						
						if tokens[2].type == "TOKEN_INTEGER" and tokens[2].value > 0 then
							startCondition.userFlag = string.format("%d", tokens[2].value)
						elseif tokens[2].type == "TOKEN_STRING" then
							startCondition.userFlag = tokens[2].value
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be integer > 0 in string format condition", false)
							return
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #3 must be integer > 0 or boolean in string format condition", false)
						return
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Bad string format condition", false)
					return
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Bad string format condition", false)
				return
			end
			
			localMembers.startCondition = startCondition
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "condition type must be string", true)
		end
	end
	
	function newTask.removeStartCondition(self)
		localMembers.startCondition = nil
	end
	
	function newTask.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newTask) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newTask.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newTask.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newTask) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newTask.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newTask
end

setmetatable(ATME.C_Task,ATMEDatas.C_Task_MT)

ATMEDatas.C_Task_MT.__call = function(t, _task, _datas)
	local fctDefinition = "Constructor ATME.C_Task(taskID, datas)"
	
	if type(_task) == "string" then
		local taskDatas = {}
		
		local err, usedOnZoneOnly, filters, taskType = ATMEFct.C_Task_setParams(_task, taskDatas, _datas)
		
		if err == nil then
			return ATMEFct.C_Task_create(nil, nil, _task, usedOnZoneOnly, taskDatas, filters, taskType)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, err, false)
		end
	else
		if type(_task) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "taskID type must be string", false)
		end
	end
end

ATMEDatas.C_Task_MT.mt.__metatable = "Private"
ATMEDatas.C_Task_MT.mt.__index = ATME.C_Task
ATMEDatas.C_Task_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Task_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Task : Bad key " .. k, "table locked", false)
	end
end

-----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Route : gère la création dynamique de route composée de WP
--
-- ATTENTION : WP 0 dans l'éditeur de mission est waypoints[1] de cette classe avec les coordonnées de la
-- première unité du groupe.
-----------------------------------------------------------------------------------------------------------
-- Set option des AI
ATMEFct.C_setOption = function(_type, _option, _param)
	local optionName = string.upper(_option)

	if ATMEConst.optionAI[_type][optionName] ~= nil then
		local optionValue 
		local tokens
		local values = {}
		
		if type(_param) == "string" then
			tokens = ATMEFct.decompString(_param)
			optionValue = 0
			
			if #tokens ~= #ATMEConst.optionAI[_type][optionName].values.args then
				if #ATMEConst.optionAI[_type][optionName].values > 1 then
					return true, "Option '" .. _option .. "' : string value must have " .. #ATMEConst.optionAI[_type][optionName].values .. " tokens"
				else
					return true, "Option '" .. _option .. "' : string value must have 1 token"
				end
			end
		end
		
		for _i, _item in ipairs(ATMEConst.optionAI[_type][optionName].values.args) do
			if tokens ~= nil then -- Si ok string
				if _item.fct ~= nil then
					local values = ATMEAIIndFct[_item.fct](tokens[_i])
					local value = ATMEFct.testDatas(values[1], _item)
					
					if value ~= nil then
						optionValue = optionValue + value
					else
						return true, "Option '" .. _option .. " : string value token '" .. tokens[_i] .. "' - Bad value"
					end
				else
					local value = ATMEFct.testDatas(tokens[_i], _item)
					if value ~= nil then
						optionValue = optionValue + value
					else
						return true, "Option '" .. _option .. " : string value token '" .. tokens[_i] .. "' - Bad value"
					end
				end
			elseif type(_param) == _item.type then  -- Number ou Boolean uniquement
				if _item.fct ~= nil then
					local values = ATMEAIIndFct[_item.fct](_param)
					local value = ATMEFct.testDatas(values[1], _item)
					
					if value ~= nil then
						optionValue =  value
					else
						if type(_param) == "number" then
							return true, "Option '" .. _option .. " : number value " .. _param .. " - Bad value"
						end
					end
				else
					local value = ATMEFct.testDatas(_param, _item)
					if value ~= nil then
						optionValue = value
					else
						if type(_param) == "number" then
							return true, "Option '" .. _option .. " : number value " .. _param .. " - Bad value"
						end
					end
				end
			else
				return true, "Value : Bad type"
			end
		end
		
		if ATMEConst.optionAI[_type][optionName].values.validFct ~= nil then
			local values = ATMEAIIndFct[ATMEConst.optionAI[_type][optionName].values.validFct](optionValue)
			
			if values[1] ~= nil then
				return true, "Option '" .. _option .. " : string value '" .. _param .. "' not allowed"
			end
		end	
		
		return false, optionName, optionValue
	else
		return true, "Option '" .. _option .. "' doesn't exist"
	end
end

ATMEFct.C_Route_getWaypointPosition = function(_route, _position)
	local localMembers = ATMEDatas.localMembers[_route:getIndex()]
	local surfaceFilter = "LAND"
	local inWater = false
	local position
	
	if localMembers.type == "ship" then
		surfaceFilter = "WATER"
		inWater = true
	elseif localMembers.type == "plane" or  localMembers.type == "helicopter" then
		surfaceFilter = "ALL"
	end
	
	if ATME.isPoint(_position) ~= POINT_BAD then
		position = ATMEFct.convertToPoint3D(_position)
		
		if surfaceFilter ~= "ALL" and ATME.isSurfaceWater(position) ~= inWater then
			return true, "Bad surface for position"
		end
	elseif type(_position) == "string" then
		local zone = ATMEFct.getDCSZone(_position)
		
		if zone ~= nil then
			position = ATMEFct.C_Circle_create(nil, nil, zone.point, zone.radius)
		else
			return true, "DCS zone " .. _position .. " unknown"
		end
	else
		if _position:getCount() > 0 then
			position = _position
		else
			return true, "ZonesMgr without zones"
		end
	end
	
	return false, position, surfaceFilter
end								

ATMEFct.C_Route_createWaypoints = function(_route)
	local localMembers = ATMEDatas.localMembers[_route:getIndex()]

	if localMembers.waypoints == nil or localMembers.forced == true then
		localMembers.waypoints = {}
		if #localMembers.initWP > 0 then
			
			for _index, _initWP in ipairs(localMembers.initWP) do
				local absPoint
				local altitude = _initWP.alt
				
				if _initWP.type ~= "WAYPOINT" then
					if _initWP.position ~= nil then
						local position = _initWP.position:getRandomPosition(_initWP.surfaceFilter)
							
						absPoint = {
							x = position.x,
							y = _initWP.alt,
							z = position.z,
						}
					else
						local point = {
							x = _initWP.x,
							y = _initWP.alt,
							z = _initWP.y,
						}
						
						if _initWP.isRelative == true then
							if _initWP.unitName ~= nil and ATMEDatas.allAIUnitsAndPlayers[_initWP.unitName] ~= nil then
								-- Evaluation des positions relatives
								absPoint = ATMEFct.coordLocalToDCS(point, ATMEDatas.allAIUnitsAndPlayers[_initWP.unitName], ATMEDatas.allAIUnitsAndPlayers[_initWP.unitName]:getAzimuth())
							elseif _initWP.reference ~= nil then
								if type(_initWP.reference) == "string" and ATMEDatas.allAIUnitsAndPlayers[_initWP.reference] ~= nil then
									local position = ATMEDatas.allAIUnitsAndPlayers[_initWP.reference]:getPosition()
									
									absPoint = {
										x = _initWP.x + position.x,
										y = _initWP.alt,
										z = _initWP.y + position.z,
									}
								else
									absPoint = {
										x = _initWP.x + _initWP.reference.x,
										y = _initWP.alt,
										z = _initWP.y + _initWP.reference.z,
									}
								end
							end
						else
							absPoint = point
						end
					end
					
					if _initWP.action == "On Road" then
						absPoint.x, absPoint.z = land.getClosestPointOnRoads("roads", absPoint.x, absPoint.z)
					end
					
					if  _initWP.alt == nil or ATMEConst.altRouteSet[localMembers.type] == true then
						absPoint.y = ATMEFct.getSurfaceAltitude(absPoint)
					end

					local speedLocked = true

					table.insert(localMembers.waypoints, {
						x = absPoint.x,
						y = absPoint.z,
						alt = absPoint.y,
						name = _initWP.name,
						speed = _initWP.speed,
						speed_locked = speedLocked,
						alt_type = "BARO",
						ETA = nil,
						ETA_locked = false,
						type = _initWP.typeWP,
						action = _initWP.action,
						helipadId = _initWP.helipadId,
						airdromeId = _initWP.airdromeId,
						linkUnit = _initWP.linkUnit,
						task = {
							["id"] = "ComboTask",
							["params"] = 
							{
								["tasks"] = 
								{
								},
							},
						},
	
					})
				else
					table.insert(localMembers.waypoints, ATMEFct.copyTable(_initWP.datas))
				end
				
				for _, _task in ipairs(_initWP.tasks) do
					table.insert(localMembers.waypoints[#localMembers.waypoints].task.params.tasks, _task)
				end
			end
			
			if #localMembers.waypoints == 0 then
				return true
			end
		else
			return true
		end
	end

	local waypoints = {}
	
	for _i, _waypoint in ipairs(localMembers.waypoints) do
		table.insert(waypoints, _waypoint)
	end
	
	return false, waypoints
end

ATMEFct.C_Route_addCurrentPositionWaypoint = function(_waypoints, _ATMEGroup)
	local waypoint = ATMEFct.copyTable(_waypoints[1])
	local localMembers = ATMEDatas.localMembers[_ATMEGroup:getIndex()]
	
	if ATMEConst.canFly[localMembers.category] == true and ATMEFct.isTakeOffWaypoint(_waypoints[1].type) == true then
		table.remove(_waypoints, 1)
	end

	if #_waypoints > 0 then
		local pos = ATMEFct.getFirstUnit(_ATMEGroup):getPosition()
		local line = ATMEFct.C_Line2D_create(nil, nil, pos, _waypoints[1])

		if line ~= nil then
			local pt = ATMEFct.C_Line2D_getPointFromOrigin(line:getA(), line:getB(), line:getOrigin(), 100)
			
			waypoint.x = pt.x
			waypoint.y = pt.z
		end
		
		waypoint.alt = pos.y
		
		table.insert(_waypoints, 1, waypoint)
		
		return pos
	end
end

ATMEFct.addInitWP = function(_route, _fctDefinition, _unit, _reference, _datas, _copy)
	local newPoint
	local localMembers = ATMEDatas.localMembers[_route:getIndex()]
	
	if _copy == false then
		if _unit == nil and _reference == nil then
			newPoint = {
				type = "POINT",
				isRelative = false,
				x = _datas.x,
				y = _datas.y,
				speed = _datas.speed,
				alt = _datas.altitude,
				typeWP = _datas.typeWP,
				position = _datas.position,
				surfaceFilter = _datas.surfaceFilter,
			}
		elseif _unit ~= nil then		
			newPoint = {
				type = "POINT",
				isRelative = true,
				x = _datas.x,
				y = _datas.y,
				alt = _datas.altitude,
				speed = _datas.speed,
				typeWP = _datas.typeWP,
				unitName = _unit:getName(),
			}
		else
			local data
			
			if ATMEFct.isObjectClass(_reference, C_AIUnit) == true then
				data = _reference:getName()
			else
				data = ATMEFct.getPointFromReference(_reference)
			end
			
			newPoint = {
				type = "POINT",
				isRelative = true,
				x = _datas.x,
				y = _datas.y,
				alt = _datas.altitude,
				speed = _datas.speed,
				typeWP = _datas.typeWP,
				reference = data,
			}
		end

		if _datas.typeWP == "Turning Point" then
			if _datas.action == nil then
				newPoint.action = "Turning Point"
				if localMembers.type == "vehicle" then
					newPoint.action = "Off Road"
				elseif localMembers.type == "ship" then
					newPoint.action = "Turning Point"
				end
			else
				newPoint.action = _datas.action
			end
		elseif _datas.typeWP == "Land" or _datas.typeWP == "TakeOffParkingHot" or _datas.typeWP == "TakeOffParking" then
			if _datas.typeWP == "Land" then
				newPoint.action = "Landing"
			else
				newPoint.action = "From Parking Area"
			end
			
			local localMembersAirbase = ATMEDatas.localMembers[_datas.airbase:getIndex()]
			
			newPoint.airdromeId = localMembersAirbase.airdromeId
			newPoint.helipadId = localMembersAirbase.helipadId
			newPoint.linkUnit = localMembersAirbase.helipadId
			newPoint.airbaseName = _datas.airbase:getName()
			newPoint.airbaseCategoryName = _datas.airbase:getCategoryName()
		end
		
		newPoint.name = ""
		newPoint.tasks = {}
	else
		newPoint = {
			type = "WAYPOINT",
			datas = _datas,
			tasks = {},
		}
	end
	
	
	-- Si le dernier WP est landing, il est supprimé.
	if #localMembers.initWP > 0 then
		if localMembers.initWP[#localMembers.initWP].typeWP == "Land" then
			table.remove(localMembers.initWP)
		end
	end
	
	table.insert(localMembers.initWP, newPoint)
	
	-- On retourne le numéro du waypoint qui débute à 0
	return #localMembers.initWP - 1
end		

ATMEFct.addWPFrom = function(_route, _groupName, _originalWP, _withTask, _fromWP, _nbWP)
	local waypoints, typeRoute = ATMEFct.C_CloningContext_getInitialWaypointsAndType(_groupName, _originalWP)
	local localMembers = ATMEDatas.localMembers[_route:getIndex()]
	
	if waypoints ~= nil then		
		if localMembers.type == typeRoute then

			ATMEFct.C_CloningContext_startRouteAt(waypoints, _fromWP, 0)

			if #waypoints > 0 then
				if ATMEFct.isTakeOffWaypoint(waypoints[1].type) == true then
					if #localMembers.initWP ~= 0 then
						return true, "can't add a takeoff waypoint after existing waypoints"
					end
				end
				
				local first = #localMembers.initWP
				local last
				local nb = #waypoints
				
				if _nbWP > 0 and _nbWP <= #waypoints then
					nb = _nbWP
				end

				for _index = 1, nb do
					if _withTask == false then
						waypoints[_index].task = {
							["id"] = "ComboTask",
							["params"] = {
								["tasks"] = {
								},
							},
						}
					end
						
					last = ATMEFct.addInitWP(_route, fctDefinition, nil, nil, waypoints[_index], true)
				end

				return false, first, last
			else
				ATMEFct.setError("NO_COPY")
				return false
			end
		else
			return true, _groupName .. " type and route type are different"
		end
	else
		return true, _groupName .. " doesn't exist"
	end
end

ATMEDatas.C_Route_MT = {
	className = "C_Route",
	mt = {},
	
	keys = {
		["addRampHotWaypoint"] = true,
		["addRampWaypoint"] = true,
		["addUnitRelativePolarWaypoint"] = true,
		["addReferencePolarWaypoint"] = true,
		["addAbsoluteWaypoint"] = true,
		["addLandingWaypoint"] = true,
		["addTaskToWaypoint"] = true,
		["addOptionToWaypoint"] = true,
		["setNameToWaypoint"] = true,
		["getIndex"] = true,
		["addWaypointsFrom"] = true,
		["addInitialWaypointsFrom"] = true,
		["removeAllWaypoints"] = true,
		["getLandingAirbase"] = true,
		["getNbWaypoints"] = true,
		["addFromRoute"] = true,
		["getUserPrivateDatas"] = true,
	},
	
	add = function(self, _type, _forced)
		local item = {
			initWP = {},
			waypoints = nil,
			type = _type,
			forced = _forced,
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].initWP = nil
		ATMEDatas.localMembers[_id].waypoints = nil
	end,
}

ATME.C_Route = {
	getClassName = function(self)
		return ATMEDatas.C_Route_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Route_MT.parentClass
	end,
}

ATMEFct.C_Route_create = function(_type, _forced)	
	local newRoute = {}
	
	setmetatable(newRoute, ATMEDatas.C_Route_MT.mt)
	local index = ATMEDatas.C_Route_MT:add(_type, _forced)
	local localMembers = ATMEDatas.localMembers[index]
	local landingWaypointExist = false
	
	local function analyzeAltitudeAction(_typeRoute, ...)
		local args = {...}
		local altitude
		local action
		local err
		
		if type(args[1]) == "number" then
			altitude = args[1]
			if type(args[2]) == "string" then
				if ATMEConst.WPActions[_typeRoute][args[2]] ~= nil then
					action = ATMEConst.WPActions[_typeRoute][args[2]]
				else
					err = "action must be a correct value. See doc for further informations"
				end
			elseif args[2] ~= nil then
				err = "action type must be nil or string"
			end
		elseif type(args[1]) == "string" then
			if ATMEConst.WPActions[_typeRoute][args[1]] ~= nil then
				action = ATMEConst.WPActions[_typeRoute][args[1]]
			else
				err = "action must be a correct value. See doc for further informations"
			end
		elseif args[1] ~= nil then
			err = "altitude type must be nil or number"
		end

		return err, altitude, action
	end
	
	local function getNumWaypointFromName(_name)
		local index = #localMembers.initWP
		local ok = false
		while index > 0 and ok == false do
			if localMembers.initWP[index].name ~= nil and localMembers.initWP[index].name == _name then
				ok = true
			else
				index = index - 1
			end
		end
		
		return index
	end
	
	function newRoute.getIndex(self)
		return index
	end
	
	function newRoute.removeAllWaypoints(self)
		localMembers.initWP = {}
		localMembers.waypoints = nil
		landingWaypointExist = false
	end
	
	function newRoute.addUnitRelativePolarWaypoint(self, _unit, _azimuth, _distance, _speed, ...)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addUnitRelativePolarWaypoint(unit, azimuth, distance, speed)\naddUnitRelativePolarWaypoint(unit, azimuth, distance, speed, altitude)\naddUnitRelativePolarWaypoint(unit, azimuth, distance, speed, action)\naddUnitRelativePolarWaypoint(unit, azimuth, distance, speed, altitude, action)"
		
		if ATMEFct.isObjectClass(_unit, ATME.C_Unit) == true and type(_azimuth) == "number" and type(_distance) == "number" and 
				type(_speed) == "number" then
			
			local err, altitude, action = analyzeAltitudeAction(localMembers.type, ...)

			if err == nil then
				if _speed > 0 then
					if altitude == nil or altitude >= 0 then
						local azimuthRad =  math.rad(ATMEFct.getAzimuthMod(_azimuth, 360))
						
						local point = {
							x = _distance * math.cos(azimuthRad),
							y = _distance * math.sin(azimuthRad),
						}

						local datas = {
							x = point.x,
							y = point.y,
							speed = _speed,
							altitude = altitude,
							typeWP = "Turning Point",
							action = action,
						}
						
						return ATMEFct.addInitWP(self, fctDefinition, _unit, nil, datas, false)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "altitude must be >= 0", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "speed must be > 0", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, err, false)
			end
		else
			if ATMEFct.isObjectClass(_unit, ATME.C_Unit) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "unit type must be ATME.C_AIUnit or ATME.C_Player", true)
			elseif type(_azimuth) ~= "number" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
			elseif type(_distance) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "distance type must be number", false)
			elseif type(_speed) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "speed type must be number", false)
			end
		end
	end
	
	function newRoute.addReferencePolarWaypoint(self, _reference, _azimuth, _distance, _speed, ...)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addReferencePolarWaypoint(reference, azimuth, distance, speed)\naddReferencePolarWaypoint(reference, azimuth, distance, speed, altitude)\naddReferencePolarWaypoint(reference, azimuth, distance, speed, action)\naddReferencePolarWaypoint(reference, azimuth, distance, speed, altitude, action)"
		
		if ATMEFct.isReference(_reference) == true and type(_azimuth) == "number" and type(_distance) == "number" and 
				type(_speed) == "number" then
			
			if landingWaypointExist == false then
				local err, altitude, action = analyzeAltitudeAction(localMembers.type, ...)

				if err == nil then
					if _speed > 0 then
						if altitude == nil or altitude >= 0 then
							local azimuthRad =  math.rad(ATMEFct.getAzimuthMod(_azimuth, 360))

							local point = {
								x = _distance * math.cos(azimuthRad),
								y = _distance * math.sin(azimuthRad),
							}
							
							local datas = {
								x = point.x,
								y = point.y,
								speed = _speed,
								altitude = altitude,
								typeWP = "Turning Point",
								action = action,
							}
							
							return ATMEFct.addInitWP(self, fctDefinition, nil, _reference, datas, false)
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "altitude must be >= 0", false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "speed must be > 0", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, err, false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "can't a new waypoint after a landing waypoint", false)
			end
		else
			if ATME.isReference(_reference) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
			elseif type(_azimuth) ~= "number" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "azimuth type must be number", false)
			elseif type(_distance) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "distance type must be number", false)
			elseif type(_speed) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "speed type must be number", false)
			end
		end
	end
	
	function newRoute.addAbsoluteWaypoint(self, _position, _speed, ...)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addAbsoluteWaypoint(position, speed)\naddAbsoluteWaypoint(position, speed, altitude)\naddAbsoluteWaypoint(position, speed, action)\naddAbsoluteWaypoint(position, speed, altitude, action)"
		
		if (ATME.isPoint(_position) ~= POINT_BAD or type(_position) == "string" or ATMEFct.isObjectClass(_position, ATME.C_ZonesMgr) == true) and 
				type(_speed) == "number" then
			
			if landingWaypointExist == false then
				local err, altitude, action = analyzeAltitudeAction(localMembers.type, ...)
				
				if err == nil then
					local err1, position, surfaceFilter = ATMEFct.C_Route_getWaypointPosition(self, _position)
					
					if err1 == false then
						if _speed >= 0 then
							if altitude == nil or altitude >= 0 then
								local datas
								
								if ATME.isPoint(position) ~= POINT_BAD then
									datas = {
										x = position.x,
										y = position.z,
										speed = _speed,
										altitude = altitude,
										typeWP = "Turning Point",
										action = action,
									}
								else
									datas = {
										speed = _speed,
										altitude = altitude,
										typeWP = "Turning Point",
										position = position,
										surfaceFilter = surfaceFilter,
										action = action,
									}
								end

								return ATMEFct.addInitWP(self, fctDefinition, nil, nil, datas, false)
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "altitude must be >= 0", false)
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "speed must be >= 0", false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, position, false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, err, false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "can't a new waypoint after a landing waypoint", false)
			end
		else
			if ATME.isPoint(_position) == POINT_BAD and type(_position) ~= "string" and ATMEFct.isObjectClass(_position, ATME.C_ZonesMgr) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be string, ATME.C_ZonesMgr or table (x,y)/(x,y,z) ", true)
			elseif type(_speed) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "speed type must be number", false)
			end
		end
	end
	
	function newRoute.addRampWaypoint(self, _airbase)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addRampWaypoint(airbase)"
		
		if #localMembers.initWP == 0 then
			if localMembers.type == "plane" or localMembers.type == "helicopter" then
				if ATMEFct.isObjectClass(_airbase, ATME.C_Airbase) == true then
					local point = _airbase:getPosition()
					
					local datas = {
						x = point.x,
						y = point.z,
						altitude = point.y,
						typeWP = "TakeOffParking",
						airbase = _airbase,
					}

					return ATMEFct.addInitWP(self, fctDefinition, nil, nil, datas, false)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "airbase type must be ATME.C_Airbase", true)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "reserved to 'AIRPLANE' or 'HELICOPTER' route", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "can't add a takeoff waypoint after existing waypoints", false)
		end
	end
	
	function newRoute.addRampHotWaypoint(self, _airbase)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addRampHotWaypoint(airbase)"
		
		if #localMembers.initWP == 0 then
			if localMembers.type == "plane" or localMembers.type == "helicopter" then
				if ATMEFct.isObjectClass(_airbase, ATME.C_Airbase) == true then
					local point = _airbase:getPosition()
					
					local datas = {
						x = point.x,
						y = point.z,
						altitude = point.y,
						typeWP = "TakeOffParkingHot",
						airbase = _airbase,
					}

					return ATMEFct.addInitWP(self, fctDefinition, nil, nil, datas, false)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "airbase type must be ATME.C_Airbase", true)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "reserved to 'AIRPLANE' or 'HELICOPTER' route", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "can't add a takeoff waypoint after existing waypoints", false)
		end
	end
	
	function newRoute.addLandingWaypoint(self, _airbase)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addLandingWaypoint(airbase)"
		
		if localMembers.type == "plane" or localMembers.type == "helicopter" then
			if ATMEFct.isObjectClass(_airbase, ATME.C_Airbase) == true then
				if _airbase:isHelipad() == true and localMembers.type == "plane" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "Only helicopters can land on Helipad", false)
					return
				end
				
				local point = _airbase:getPosition()
				
				local datas = {
					x = point.x,
					y = point.z,
					altitude = point.y,
					typeWP = "Land",
					airbase = _airbase,
				}
				
				landingWaypointExist = true
				
				return ATMEFct.addInitWP(self, fctDefinition, nil, nil, datas, false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "airbase type must be ATME.C_Airbase", true)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "reserved to 'AIRPLANE' or 'HELICOPTER' route", false)
		end
	end
	
	function newRoute.addWaypointsFrom(self, _group, ...)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addWaypointsFrom(group)\naddWaypointsFrom(group, fromWP, [nbWP])\naddWaypointsFrom(group, withTask)\naddWaypointsFrom(group, withTask, fromWP, [nbWP])"
		local args = {...}

		if type(_group) == "string" or ATMEFct.isObjectClass(_group, ATME.C_Group) == true then
			local withTask = false
			local fromWP = 0
			local nbWP = 0
			local groupName
			
			if type(_group) == "string" then
				groupName = _group
			else
				groupName = _group:getName()
			end
			
			if #args > 0 then
				local index = 1
				if type(args[1]) == "boolean" then
					withTask = args[1]
					index = index + 1
				end
				
				if args[index] == nil or ATMEFct.isReference(args[index]) == true or 
							type(args[index]) == "string" or type(args[index]) == "number" then
				
					if args[index] ~= nil then
						if type(args[index]) == "number" then
							fromWP = math.floor(args[index])
							
							if fromWP < 0 then
								ATMEFct.errorOutputCore(nil, fctDefinition, "fromWP must be >= 0 if number", false)
								return
							end
						end
						
						index = index + 1
						if args[index] == nil or type(args[index]) == "number" then
							if type(args[index]) == "number" then
								nbWP = math.floor(args[index])
								
								if nbWP < 0 then
									ATMEFct.errorOutputCore(nil, fctDefinition, "nbWP must be > 0", false)
									return
								end
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "nbWP type must be nil or number", false)
							return
						end
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "fromWP type must be nil, string, number," .. ATMEConst.errorLabelReference, false)
					return
				end
			end

			local err, first, last = ATMEFct.addWPFrom(self, groupName, false, withTask, fromWP, nbWP)

			if err == false then
				return first, last
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, first, false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "group type must be string or ATME.C_Group", true)
		end
	end
	
	function newRoute.addInitialWaypointsFrom(self, _group, ...)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addInitialWaypointsFrom(group)\naddInitialWaypointsFrom(group, fromWP, [nbWP])\naddInitialWaypointsFrom(group, withTask)\naddInitialWaypointsFrom(group, withTask, fromWP, [nbWP])"
		local args = {...}

		if type(_group) == "string" or ATMEFct.isObjectClass(_group, ATME.C_Group) == true then
			local withTask = false
			local fromWP = 0
			local nbWP = 0
			local groupName
			
			if type(_group) == "string" then
				groupName = _group
			else
				groupName = _group:getName()
			end
			
			
			if #args > 0 then
				local index = 1
				if type(args[1]) == "boolean" then
					withTask = args[1]
					index = index + 1
				end
				
				if args[index] == nil or ATMEFct.isReference(args[index]) == true or 
						type(args[index]) == "string" or type(args[index]) == "number" then
				
					if args[index] ~= nil then
						fromWP = args[index]
						if type(args[index]) == "number" then
							fromWP = math.floor(args[index])
							
							if fromWP < 0 then
								ATMEFct.errorOutputCore(nil, fctDefinition, "fromWP must be >= 0 if number", false)
								return
							end
						end
						
						index = index + 1
						if args[index] == nil or type(args[index]) == "number" then
							if type(args[index]) == "number" then
								nbWP = math.floor(args[index])
								
								if nbWP < 0 then
									ATMEFct.errorOutputCore(nil, fctDefinition, "nbWP must be > 0", false)
									return
								end
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "nbWP type must be nil or number", false)
							return
						end
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "fromWP type must be nil, string, number," .. ATMEConst.errorLabelReference, false)
					return
				end
			end

			local err, first, last = ATMEFct.addWPFrom(self, groupName, true, withTask, fromWP, nbWP)
			
			if err == false then
				return first, last
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, first, false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "group type must be string or ATME.C_Group", true)
		end
	end
	
	function newRoute.addFromRoute(self, _route, _fromWP, _nbWP)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addFromRoute(route, [fromWP, [nbWP]])"
		
		if ATMEFct.isObjectClass(_route, ATME.C_Route) == true and (_fromWP == nil or type(_fromWP) == "number" or type(_fromWP) == "string") and (_nbWP == nil or type(_nbWP) == "number") then
			local localMembersRouteToCopy = ATMEDatas.localMembers[_route:getIndex()]
			local nbWP = #localMembersRouteToCopy.initWP
			
			if _fromWP == nil and type(_nbWP) == "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "nbWP type must be nil if fromWP type is nil", false)
				return
			end
			
			if type(_nbWP) == "number" then
				nbWP = math.floor(_nbWP)
				if nbWP <= 0 then
					ATMEFct.errorOutputCore(nil, fctDefinition, "nbWP must be > 0", false)
					return
				end
			end
			
			if localMembers.type ~= localMembersRouteToCopy.type then
				ATMEFct.errorOutputCore(nil, fctDefinition,  "Route types are different", false)
				return
			end

			if ATMEFct.isTakeOffWaypoint(localMembersRouteToCopy.initWP[1].typeWP) == true then
				if #localMembers.initWP ~= 0 then
					ATMEFct.errorOutputCore(nil, fctDefinition, "can't add a takeoff waypoint after existing waypoints", false)
					return
				end
			end
			
			local startIndex = 1

			if type(_fromWP) == "string" then -- Relatif au nom du waypoint
				-- Si le nom n'est pas trouvé, il s'agira du premier waypoint (0 dans l'éditeur)
				local name = string.upper(_fromWP)
				local index = #localMembersRouteToCopy.initWP
				while index > 0 do
					if name == string.upper(localMembersRouteToCopy.initWP[index].name) then
						startIndex  = index
						break
					end
					
					index = index - 1
				end
			elseif type(_fromWP) == "number" then -- relatif au numéro de waypoint (début à 0 pour rappel)
				if _fromWP >= 0 and _fromWP < #localMembersRouteToCopy.initWP then
					startIndex = _fromWP + 1
				end
			end
			
			-- Si le numéro est plus grand que le dernier waypoint, il s'agira du dernier waypoint.
			local maxWP = startIndex + nbWP - 1
			if maxWP > #localMembersRouteToCopy.initWP then
				maxWP = #localMembersRouteToCopy.initWP
			end

			for _i = startIndex, maxWP do
				local currentWP = ATMEFct.copyTable(localMembersRouteToCopy.initWP[_i])

				table.insert(localMembers.initWP, currentWP)
			end
			
			return startIndex - 1, #localMembers.initWP - 1
		else
			if ATMEFct.isObjectClass(_route, ATME.C_Route) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "route type must be ATME.C_Route", true)
			elseif _fromWP ~= nil and type(_fromWP) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "fromWP type must be nil or number", false)
			elseif _nbWP ~= nil and type(_nbWP) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "nbWP type must be nil or number", false)
			end
		end
	end
	
	function newRoute.getLandingAirbase(self)
		if #localMembers.initWP > 0 then
			if localMembers.initWP[#localMembers.initWP].typeWP == "Land" then
				return localMembers.initWP[#localMembers.initWP].airbaseName, localMembers.initWP[#localMembers.initWP].categoryName
			end
		end
	end
	
	function newRoute.setNameToWaypoint(self, _numWaypoint, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "setNameToWaypoint(numWaypoint, name)"
		
		if type(_numWaypoint) == "number" and type(_name) == "string" then
			local numWaypoint = math.floor(_numWaypoint) + 1
			if numWaypoint > 0 and numWaypoint <= #localMembers.initWP then
				localMembers.initWP[numWaypoint].name = _name
			end
		else
			if type(_numWaypoint) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "numWaypoint type must be number", true)
			elseif type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			end
		end
	end
	
	function newRoute.getNbWaypoints(self)
		return #localMembers.initWP
	end
	
	function newRoute.addTaskToWaypoint(self, _waypointId, _task)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addTaskToWaypoint(waypointId, task)"
		
		if (type(_waypointId) == "number" or type(_waypointId) == "string") and ATMEFct.isObjectClass(_task, ATME.C_Task) == true then
			local numWaypoint
			if type(_waypointId) == "number" then
				numWaypoint = math.floor(_waypointId) + 1
			else
				numWaypoint = getNumWaypointFromName(_waypointId)
			end
			
			if numWaypoint >= 0 and numWaypoint <= #localMembers.initWP then
				local ok, taskDatas = ATMEFct.C_Task_getTaskDatas(_task, localMembers.type, true)
				if ok == true then
					taskDatas.enabled = true
					
					if taskDatas.id == "Orbit" then
						taskDatas.auto = false
					else
						taskDatas.auto = true
					end
					taskDatas.number = 1
					
					table.insert(localMembers.initWP[numWaypoint].tasks, taskDatas)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, taskDatas, false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Bad waypoint ID", false)
			end
		else
			if type(_waypointId) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "waypointId type must be number or string", true)
			elseif ATMEFct.isObjectClass(_task, ATME.C_Task) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "task type must be ATME.C_Task", false)
			end
		end
	end
	
	function newRoute.addOptionToWaypoint(self, _waypointId, _option, _value)
		local fctDefinition = ATMEFct.getErrorPrefix(newRoute) .. "addOptionToWaypoint(waypointId, option, value)"
		
		if (type(_waypointId) == "number" or type(_waypointId) == "string") and type(_option) == "string" and _value ~= nil then
			
			local numWaypoint
			if type(_waypointId) == "number" then
				numWaypoint = math.floor(_waypointId) + 1
			else
				numWaypoint = getNumWaypointFromName(_waypointId)
				if numWaypoint == 0 then
					ATMEFct.setError("NO_WAYPOINT_FOUND")
					return
				end
			end
			
			if numWaypoint > 0 and numWaypoint <= #localMembers.initWP then
				local err, option, optionValue = ATMEFct.C_setOption(localMembers.type, _option, _value) 

				if err == false then
					local taskDatas = {
						["id"] = "WrappedAction",		   
						["params"] = {
							["action"] = 
							{
								["id"] = "Option",
								["params"] = 
								{
									["value"] = optionValue,
									["name"] = option,
								},
							},
						},
					}
					table.insert(localMembers.initWP[numWaypoint].tasks, taskDatas)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, option, false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Bad waypoint ID", false)
			end
		else
			if type(_waypointId) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "waypointId type must be number or string", true)
			elseif type(_option) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "option type must be string", false)
			elseif _value == nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, "value type can't be nil", false)
			end
		end
	end
	
	function newRoute.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newRoute
end

setmetatable(ATME.C_Route,ATMEDatas.C_Route_MT)

ATMEDatas.C_Route_MT.__call = function(t, _typeRoute, _rebuild)
	local fctDefinition = "Constructor ATME.C_Route(typeRoute, [rebuild])"
		
	if type(_typeRoute) == "string" and (_rebuild == nil or type(_rebuild) == "boolean") then
		if ATMEConst.categoryNameFilters[_typeRoute] ~= nil then
			local rebuild = _rebuild
			if rebuild == nil then
				rebuild = false
			end
			return ATMEFct.C_Route_create(ATMEConst.categoryNameFilters[_typeRoute], _rebuild)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "type must be 'AIRPLANE', 'HELICOPTER', 'GROUND' or 'SHIP'", false)
		end
	else
		if type(_typeRoute) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "typeRoute type must be string", false)
		elseif _rebuild ~= nil and type(_rebuild) ~= "boolean" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "rebuild type must be il or boolean", false)
		end
	end
end


ATMEDatas.C_Route_MT.mt.__metatable = "Private"
ATMEDatas.C_Route_MT.mt.__index = ATME.C_Route
ATMEDatas.C_Route_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Route_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Route : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_CloningContext : gère les données de clonage nécessaires à la duplication de groupes
--
-- Cette classe permet de gérer les chargements/déchargements d'unités d'un véhicule (voir load/unload)
-- des classes C_Player et C_AIUnit. Cette fonctionnalité n'est pas accessible aux autres modules
--
-- Cette classe offre également le spawn d'unités définie à partir d'un groupe existant.
-- Le spawn sera en échec si le groupe existe. 
-- Pour l'attribution des noms :
-- 		* Pour le nom des groupes : Si le nom n'existe pas dans la mission de base (donnée initiale) ni 
--		  dans les groupes existants (actifs), le nom est celui donnée à la construction du contexte de 
--		  de clonage. Sinon, le num sera postfixé avec " - Spawned #XXX". La valeur XXX est incrémentée 
--        automatiquement un nom de groupe particulier. Elle est comprise en 001 et 999. Si le groupe
--		  est détruit, cette valeur sera réutilisée par le prochain spawn.
--		* Pour les noms d'unités : Le nom d'unité dépend du groupe initial (défini dans la mission ou créé
--	      directement "from scratch"). En conséquence, le nom de l'unité existe dès le clonage, aussi tous
--        tous les noms d'unités clonées sont postfixée avec " - Spawned #XXX". La valeur XXX est incrémentée 
--        automatiquement un nom de l'unité particulière. Elle est comprise en 001 et 999. Si l'unité
--		  est détruite, cette valeur sera réutilisée par le prochain spawn.
--
-- Il est possible de construire un contexte de clonage à partir d'un groupe existants (actifs) ou défini initialement
-- dans la mission. Le nom du groupe à cloner sera alors une entrée pour le contexte de clonage.  Il est également
-- possible de construire un groupe "from scratch", l'entrée du contexte de clonage sera alors la table datas de 
-- DCS correctement renseignée.
--
-- Pour les waypoints : La modification d'une route initiale par la fonction setRoute de ATME.C_Group n'a aucun incidence
-- sur la route du nouveau groupe cloné. Le groupe cloné peut avoir :
--			- Par défaut : les WP définis initialement pour le groupe source du clonage.
--			- Une route spécifique définie par une instance de ATME.C_Route
----------------------------------------------------------------------------------------------------------

ATMEFct.C_CloningContext_createNamed = function(_namedTable, _name, _origin, _newGroupName)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return false
	else
		if type(_origin) == "table" then
			if ATMEFct.C_CloningContext_controlTable(_origin, _newGroupName) == false then
				return true, "Bad cloning table format"
			end
			
			if _newGroupName ~= nil then
				_origin.datas.name = _newGroupName
			end
		end
		
		local newGroupName = _newGroupName
		if _newGroupName == nil then
			if type(_origin) == "string" then
				newGroupName = _origin
			else
				newGroupName = _origin.datas.name 
			end
		else
			local d,f = string.find(newGroupName, "|", 1)
			if d ~= nil then
				return true, "Name of group " .. newGroupName .. " can't contain '|' character"
			end
			
			d,f = string.find(newGroupName:upper(), "%s+SPAWNED%s+", 1)
			if d ~= nil then
				return true, "Name of group " .. newGroupName .. " can't contain 'Spawned' word"
			end
		end
		
		local err, initialNames = ATMEFct.C_CloningContext_getInitialNames(_origin)
		
		if err == nil then
			local foundPlayer = false
			for _, _unitDatas in pairs(ATMEDatas.allGroupsInMission[initialNames.groupName].units) do
				if _unitDatas.skill == "Player" or _unitDatas.skill == "Client" then
					foundPlayer = true
				end
			end
			
			if foundPlayer == false then
				-- Le nom du nouveau groupe (_newGroupName) ne doit pas correspondre à un groupe existant,
				-- à un groupe embarqué ou à un groupe en cours de spawn.

				return false, ATMEFct.C_CloningContext_create(_name, _namedTable, initialNames, newGroupName)
			else
				return true, "name can't contain '|' character"
			end
		end
	end
end

ATMEFct.C_CloningContext_controlTable = function(_WPGroup, _newGroupName)
	if ATMEConst.countries[_WPGroup.countryName] == nil or
		ATMEConst.categoryNameFilters[_WPGroup.typeName] == nil or
		_WPGroup.datas == nil or
		_WPGroup.datas.task == nil or _WPGroup.datas.tasks == nil then
		
		return false
	end
	
	if _WPGroup.datas.units == nil or type(_WPGroup.datas.units) ~= "table"  then
		_WPGroup.datas.units = {}
	end
	
	if (type(_WPGroup.datas.name) ~= "string" and type(_newGroupName) ~= "string") or type(_WPGroup.datas.tasks) ~= "table" then
		return false
	end
	
	for _, _unit in ipairs(_WPGroup.datas.units) do
		if type(_unit.name) ~= "string" then
			return false
		end
	end
	
	return true
end

ATMEFct.C_CloningContext_addUnit = function(_allGroupsInMissionItem, _unit, _newName, _initialNames, _initialType)
	local newName = _newName
	
	if newName == nil then
		newName = _unit.name
		
		if newName == nil then
			return true, "Group " .. _allGroupsInMissionItem.name .. " - New unit name is nil"
		end
	end
	
	if ATMEDatas.allAIUnitsInMission[newName] ~= nil then
		return true, "Unit " .. newName .. " exists in mission"
	end
	
	if _unit.skill ~= "Client" and _unit.skill ~= "Player" then
		local payload = _unit.payload
		
		if payload == nil and (_initialType == "plane" or _initialType == "helicopter") then
			payload = {
				["pylons"] = {},
			}
		end
		
		_allGroupsInMissionItem.units[newName] = {
			name = newName,
			row = #_initialNames.unitsName + 1,
			type = _unit.type,
			livery_id = _unit.livery_id,
			speed = _unit.speed,
			psi = _unit.psi,
			payload = payload,
			skill = _unit.skill,
			x = _unit.x,
			y = _unit.y,
			playerCanDrive = _unit.playerCanDrive,
			heading = _unit.heading,
			alt = _unit.alt,
			alt_type = _unit.alt_type,
			callsign = _unit.callsign,
			onboard_num = _unit.onboard_num,
			parking = _unit.parking,
			parking_id = _unit.parking_id,
		}
		
		table.insert(_initialNames.unitsName, newName)
		
		ATMEDatas.allAIUnitsInMission[newName] = {
			group = _allGroupsInMissionItem,
			unit = _allGroupsInMissionItem.units[newName],
		}
		
		return false
	else
		return true, "Group " .. _allGroupsInMissionItem.name .. " has one or more player units"
	end
end

ATMEFct.C_CloningContext_getInitialNames = function(_WPGroup)
	if type(_WPGroup) == "string" then
		if ATMEDatas.allGroupsInMission[_WPGroup] ~= nil then
			return nil, ATMEDatas.allInitialGroupNames[_WPGroup]
		elseif ATMEDatas.groups[_WPGroup] ~= nil then
			local localMembersGroup = ATMEDatas.localMembers[ATMEDatas.groups[_WPGroup]:getIndex()]
			
			return nil, localMembersGroup.spawnDatas.initialNames
		elseif ATMEDatas.groupsOnBoard[_WPGroup] ~= nil then
			return nil, ATMEDatas.groupsOnBoard[_WPGroup].spawnDatas.initialNames
		elseif ATMEDatas.groupsReadyToBirth[_WPGroup] ~= nil then
			return nil, ATMEDatas.groupsReadyToBirth[_WPGroup].item.spawnDatas.initialNames
		elseif ATMEDatas.newCreatedGroups[_WPGroup] ~= nil then
			local localMembersGroup = ATMEDatas.localMembers[ATMEDatas.newCreatedGroups[_WPGroup].group:getIndex()]
			
			return nil, localMembersGroup.spawnDatas.initialNames
		end
		
		return "Group " .. _WPGroup .. " not in mission"
	else
		if ATMEDatas.allGroupsInMission[_WPGroup.datas.name] == nil then
			if ATMEConst.categoryNameFilters[_WPGroup.typeName] ~= nil then
				local initialNames = {
					groupName = _WPGroup.datas.name,
					unitsName = {},
				}
				
				-- Pour ces groupes qui servent au spawn, le group ID n'est pas initialisé.
				-- Il le sera lors du spawn par la fonction addGroup.
				local allGroupsInMissionItem = {
					name = _WPGroup.datas.name,
					country = ATMEConst.countries[_WPGroup.countryName],
					coalition = ATMEConst.coalitionNames[coalition.getCountryCoalition(ATMEConst.countries[_WPGroup.countryName])],
					category = ATMEConst.groupCategories[ATMEConst.categoryNameFilters[_WPGroup.typeName]],
					type = ATMEConst.categoryNameFilters[_WPGroup.typeName],
					units = {},
					taskSelected = true,
					hidden = _WPGroup.datas.hidden,
					start_time = 0,
					task = _WPGroup.datas.task,
					tasks = _WPGroup.datas.tasks,
					radioSet = _WPGroup.datas.radioSet,
					modulation = _WPGroup.datas.modulation,
					frequency = _WPGroup.datas.frequency,
					communication = _WPGroup.datas.communication,
					waypoints = {},
				}

				for _, _unit in ipairs(_WPGroup.datas.units) do
					local err, value = ATMEFct.C_CloningContext_addUnit(allGroupsInMissionItem, _unit, nil, initialNames, ATMEConst.categoryNameFilters[_WPGroup.typeName])

					if err == true then
						return err, value
					end
				end
				
				if _WPGroup.datas.route ~= nil and _WPGroup.datas.route.points ~= nil and #_WPGroup.datas.route.points > 0 then
					for _pointNum, _point in ipairs(_WPGroup.datas.route.points) do
						allGroupsInMissionItem.waypoints[#allGroupsInMissionItem.waypoints + 1] = {
							name = _point.name,
							x = _point.x,
							y = _point.y,
							speed = _point.speed,
							speed_locked = _point.speed_locked,
							alt = _point.alt,
							alt_type = _point.alt_type,
							ETA = _point.ETA,
							ETA_locked = _point.ETA_locked,
							airdromeId = _point.airdromeId,
							helipadId = _point.helipadId,
							type = _point.type,
							action = _point.action,
							properties = _point.properties,
							formation_template = _point.formation_template,
						}
						
						if _point.task == nil then
							ATMEFct.C_CloningContext_setWaypointComboTask(allGroupsInMissionItem.waypoints[#allGroupsInMissionItem.waypoints])
						else
							allGroupsInMissionItem.waypoints[#allGroupsInMissionItem.waypoints].task = ATMEFct.copyTable(_point.task)
						end
					end
				end

				-- Préparation des données globales si tout est ok
				
				ATMEDatas.allGroupsInMission[_WPGroup.datas.name] = allGroupsInMissionItem
				ATMEDatas.allInitialGroupNames[_WPGroup.datas.name] = initialNames
				
				return nil, initialNames
			else
				return "Group " .. _WPGroup.datas.name .. " bad type name"
			end
		else
			return "Group " .. _WPGroup.datas.name .. " exists in mission"
		end
	end
end

ATMEFct.C_CloningContext_getInitialWaypointsAndType = function(_WPGroupName, _onlyInMission)
	local initialWaypoints
	local initialType
	
	if ATMEDatas.groups[_WPGroupName] ~= nil and _onlyInMission == false then
		local localMembersGroup = ATMEDatas.localMembers[ATMEDatas.groups[_WPGroupName]:getIndex()]
		initialWaypoints = localMembersGroup.spawnDatas.datas.route.points
		initialType = localMembersGroup.spawnDatas.type
	elseif ATMEDatas.allGroupsInMission[_WPGroupName] ~= nil then
		initialWaypoints = ATMEDatas.allGroupsInMission[_WPGroupName].waypoints
		initialType = ATMEDatas.allGroupsInMission[_WPGroupName].type
	end
	
	if initialWaypoints ~= nil then
		local waypoints = ATMEFct.C_CloningContext_copyAllWaypoints(initialWaypoints)

		return waypoints, initialType
	end
end

ATMEFct.C_CloningContext_getInitialWaypointFromID = function(_WPGroupName, _id)
	local initialWaypoints
	
	if ATMEDatas.groups[_WPGroupName] ~= nil then
		local localMembersGroup = ATMEDatas.localMembers[ATMEDatas.groups[_WPGroupName]:getIndex()]
		initialWaypoints = localMembersGroup.spawnDatas.datas.route.points
	elseif ATMEDatas.allGroupsInMission[_WPGroupName] ~= nil then
		initialWaypoints = ATMEDatas.allGroupsInMission[_WPGroupName].waypoints
	end
	
	if initialWaypoints ~= nil then
		local waypoint = initialWaypoints[_id]

		return waypoint
	end
end

ATMEFct.C_CloningContext_copyAllWaypoints = function(_waypoints)
	local newWaypoints = {}

	for _, _waypoint in ipairs(_waypoints) do
		table.insert(newWaypoints, ATMEFct.copyTable(_waypoint))
	end

	return newWaypoints
end

ATMEFct.C_CloningContext_setWaypointComboTask = function(_waypoint)
	if _waypoint.task == nil then
		_waypoint.task = {
			["id"] = "ComboTask",
			["params"] = {
				["tasks"] = {
				},
			},
		}
	end
end

ATMEFct.C_CloningContext_setMission = function(_ATMEGroup, _waypoints, _delayed)
	local localMembers = ATMEDatas.localMembers[_ATMEGroup:getIndex()]
	
	--TODO _delayed toujours à false ;) à voir
	if _delayed == false and localMembers.routeUncontrolled == false then
		local offsetTaskWaypointTracking

		local task = { 
			id = 'Mission', 
			params = { 
				route = {
					points = _waypoints,
				}
			} 
		}
		
	
		local controller = _ATMEGroup:getDCSGroup():getController()

		controller:setTask(task)
	else
		local name = _ATMEGroup:getName()
		local item = {
			name = _ATMEGroup:getName(),
			waypoints = _waypoints,
			delay = localMembers.routeDelay,
		}
		
		ATMEDatas.missionsDelayed[name] = item
	end
end

ATMEFct.C_CloningContext_setWaypoints = function(_ATMEGroup, _waypoints, _isOnRoad)
	local firstUnit = ATMEFct.getFirstUnit(_ATMEGroup)
	
	if firstUnit ~= nil then
		-- On ajoute la position courante à la nouvelle route

		local newWP = ATMEFct.copyTable(_waypoints[1])
		newWP.x = firstUnit:getPosition().x
		newWP.y = firstUnit:getPosition().z
		
		if _waypoints[1].action == "On Road" and _isOnRoad == false then
			newWP.action = "Off Road"
		end
		
		table.insert(_waypoints, 1, newWP)
		
		ATMEFct.C_CloningContext_setMission(_ATMEGroup, _waypoints, false)
	end
end

ATMEFct.C_CloningContext_getWaypointFromName = function(_waypoints, _name)
	local index = 0
	local name = _name
	
	for _i, _wp in ipairs(_waypoints) do
		if name == _wp.name then
			index = _i
		end
	end
	
	if index > 0 then
		return index - 1
	else
		return nil
	end
end

ATMEFct.C_CloningContext_startRouteAt = function(_waypoints, _reference, _radius)
	-- _radius défini la zone d'exclusion par rapport à la référence de type objet. Non pris en compte sinon.
	-- Dans ce cas, la route débutera au waypoint le plus proche de la référence, mais en excluant un cercle de rayon 
	-- _radius autour de la référence.
	-- Pour les autres cas, l'offset ou le nom du waypoint est recherché. S'il n'est pas trouvé, la route reste la 
	-- route par défaut

	if type(_reference) == "string" then -- Relatif au nom du waypoint
		-- Si le nom n'est pas trouvé, il s'agira du premier waypoint (0 dans l'éditeur)
		
		local index = ATMEFct.C_CloningContext_getWaypointFromName(_waypoints, _reference)
		
		if index ~= nil then
			for _i = 1, index do
				table.remove(_waypoints, 1)
			end
			
			return index
		else
			return #_waypoints - 1
		end
	elseif type(_reference) == "number" then -- relatif au numéro de waypoint (début à 0 pour rappel)
		-- Si le numéro est plus grand  ou egal au dernier waypoint, il s'agira du dernier waypoint.
		
		local nb = _reference - _radius
		if nb < #_waypoints then
			while nb > 0 do 
				table.remove(_waypoints, 1)
				nb = nb - 1
			end

			return _reference - _radius -- radius contient alors waypointId (si débarquement) ou 0
		end
		
		return #_waypoints - _radius	-- radius contient alors waypointId (si débarquement) ou 0
	else -- Relatif à un objet
		local point = ATMEFct.getPointFromReference(_reference)

		local index = #_waypoints
		local v = ATMEFct.C_Vector3D_create(nil, nil, point, _waypoints[index])
		local minDistance = v:getHModule()
		
		local found = index

		while index > 1 do
			index = index - 1
			
			local v = ATMEFct.C_Vector3D_create(nil, nil, point, _waypoints[index])
			
			if v:getHModule() >= _radius then
				if minDistance > v:getHModule() then
					minDistance = v:getHModule()
					found = index
				end
			end
		end
		
		for _i = 1, found - 1 do
			table.remove(_waypoints, 1)
		end

		return found - 1
	end
end
								

-- Fonction spawn des objets DCS avec gestion ATME
ATMEFct.C_CloningContext_spawn = function(_spawnDatas)
	-- ATTENTION deux cas :
	--	*	Pour les groupes d'unités au sol ou les bateaux, les unités sont créées (event S_EVENT_BIRTH) 
	--		pendant le addgroup mais parfois leur groupe associé ne l'est pas encore ... 
	--		Donc on mémorise lors de l'événement DCS et on traite ci après
	--	*	Pour les groupes aircraft, l'event S_EVENT_BIRTH est appelé bien après. On mémorise ici pour un
	--		traitement lors de l'événement DCS.

	ATMEDatas.spawnedGroupsWaypoints[_spawnDatas.datas.name] = {
		waypoints = ATMEFct.C_CloningContext_copyAllWaypoints(_spawnDatas.datas.route.points),
		initialWaypoints = ATMEFct.C_CloningContext_copyAllWaypoints(_spawnDatas.datas.route.points),
	}
	
	coalition.addGroup(_spawnDatas.country, _spawnDatas.category, _spawnDatas.datas)

	if #ATMEDatas.unitsReadyToCreate > 0 then
		while #ATMEDatas.unitsReadyToCreate > 0 do
			local newUnit = table.remove(ATMEDatas.unitsReadyToCreate)
			
			-- L'unité à créer est un bateau avec FARP ou porte-avion
			
			if newUnit:hasAttribute("AircraftCarrier") == true then
				ATMEFct.C_Airbase_create_basic("SHIP", false, newUnit)
			elseif newUnit:hasAttribute("HelicopterCarrier") == true then
				ATMEFct.C_Airbase_create_basic("SHIP", true, newUnit)
			end
			
			local datas = {
				uncontrolled = false,
				visible = true,
				spawnDatas = _spawnDatas,
			}
			
			ATMEFct.C_AIUnit_create(newUnit, datas, true)
		end
		
		ATMEDatas.newCreatedGroups[_spawnDatas.datas.name].isSpawned = true
		ATMEDatas.newCreatedGroups[_spawnDatas.datas.name].spawnCallback = _spawnDatas.spawnCallback
		ATMEDatas.newCreatedGroups[_spawnDatas.datas.name].cloningContext = _spawnDatas.cloningContext
	else
		local itemDatas = {
			name = _spawnDatas.datas.name,
			units = {},
			nbUnits = 0,
			uncontrolled = false,
			visible = true,
			spawnDatas = _spawnDatas,
		}
		
		for _i, _unit in ipairs(_spawnDatas.datas.units) do
			itemDatas.units[_unit.name] = true
			itemDatas.nbUnits = itemDatas.nbUnits + 1
			
			ATMEDatas.unitsReadyToBirth[_unit.name] = true
		end
		
		ATMEDatas.groupsReadyToBirth[_spawnDatas.datas.name] = {
			item = itemDatas,
			spawnCallback = _spawnDatas.spawnCallback,
			cloningContext = _spawnDatas.cloningContext,
		}
	end
end

-- Spawn sur débarquement des troupes embarquées dans un transport
ATMEFct.C_CloningContext_spawnFromTransportUnit = function(_transportUnit, _currentOnBoardGroupDatas)
	if ATMEDatas.exitATME == false then -- Test nécessaire car sinon problème sur l'événement du spawn (S_EVENT_BIRTH) désactivé par ATMEDatas.exitATME à true
		local spawnDatas = _currentOnBoardGroupDatas.spawnDatas
		local unitsOnBoard = _currentOnBoardGroupDatas.unitsOnBoard
		local initialWaypoints = ATMEFct.C_CloningContext_copyAllWaypoints(spawnDatas.datas.route.points)
		
		local nb = _transportUnit:getNbUnitsOnBoard()

		-- Min une unité par 3° soit 118 unités au maximum
		-- Calcul : pour 180° -> 180/3 - 1 unités, soit pour 360°, 2 x (180/3 - 1) = 118
		-- Max units par groupe terrestre = 99 < 118 donc pas de soucis actuellement
		-- sachant que la charge d'un hélico est bien inférieure à 99 unités.

		local point = _transportUnit:getPosition()
		local azimuth = _transportUnit:getAzimuth()

		-- Vérification d'absence d'eau côté droit/gauche 
		
		local sides = {}
		local startAngle = 0
		local nbPointsWater = {0,0}
		local index = 1
		local distance = ATMEConst.personnelCarrierAttributes[_transportUnit:getTypeName()].minDistance
		local deltaRad = math.pi / 4
		for _side = 1,-1,-2 do
			for _angle = deltaRad / 4 * _side, 3 * deltaRad * _side, deltaRad * _side do
				local result = ATMEFct.coordLocalToDCS({x = distance * math.cos(_angle), y = distance * math.sin(_angle)}, point, azimuth)
				if ATME.isSurfaceWater(result) == true then
					nbPointsWater[index] = nbPointsWater[index] + 1
				end
			end
			
			if nbPointsWater[index] < 2 then
				table.insert(sides, startAngle)
			end
			startAngle = startAngle + 180
			index = index + 1
		end
		
		if #sides > 0 then
			if #sides == 2 then
				if nbPointsWater[1] > nbPointsWater[2] then -- On permute les côtés pour privilégier le côté sans eau (1 unité de plus si impair)
					sides[1] = 180
					sides[2] = 0
				end
			end
			-- Ajustements en fonction du nombre d'unités et du nombre
			-- de côtés possibles.
			
			-- Si moins de 2 unités par côté pas de symétrie
			if nb/#sides <= 2 and #sides > 1 then
				table.remove(sides)
			end
			
			-- Si symétrie (2 côtés) et nombre impair d'unités -> On ajuste pour rendre pairs
			if nb/#sides ~= math.floor(nb/#sides) then
				-- Si le nombre d'unités est impair quand 2 côtés ok, on ajoute 1
				nb = nb + 1
			end

			-- Calcul de l'angle entre chaque unité
			
			local deltaAngle = math.floor(180 / (nb + 1))
			local unitAngle = startAngle
			local id = 1
			local currentSide = 1

			while id <= #spawnDatas.datas.units do
				if unitsOnBoard[spawnDatas.datas.units[id].name] ~= nil then
					local unit = spawnDatas.datas.units[id]
					unitAngle = unitAngle + deltaAngle


					local unitAngleRad = math.rad(unitAngle + sides[currentSide])
					currentSide = currentSide + 1
					if currentSide > #sides then
						currentSide = 1
					end
					
					local result = ATMEFct.coordLocalToDCS(
												{
													x = distance * math.cos(unitAngleRad), 
													y = distance  * math.sin(unitAngleRad),
												}, 
												point, 
												azimuth)
					
					-- Si l'unité est bien sur terre, alors ok
					if ATME.isSurfaceWater(result) == false then
					
						unit.x = result.x
						unit.y = result.z

						unit.heading = unitAngleRad + math.rad(azimuth)
						id = id + 1
					else -- si elle est dans l'eau alors elle meure (elle ne sera pas réactivée).
						table.remove(spawnDatas.datas.units, id)
						_currentOnBoardGroupDatas.nbDeadUnits = _currentOnBoardGroupDatas.nbDeadUnits + 1
					end
				else
					table.remove(spawnDatas.datas.units, id)
				end
			end

			if #spawnDatas.datas.units > 0 then
				local startWP = ATMEFct.copyTable(_currentOnBoardGroupDatas.waypoints[1])
				local nextWaypoint
				local waypointId = _currentOnBoardGroupDatas.waypointId + _currentOnBoardGroupDatas.deltaWaypointId

				if _currentOnBoardGroupDatas.nextWaypoint ~= nil then
					if type(_currentOnBoardGroupDatas.nextWaypoint) == "number" then
						if _currentOnBoardGroupDatas.nextWaypoint > _currentOnBoardGroupDatas.waypointId and 
									_currentOnBoardGroupDatas.nextWaypoint < #_currentOnBoardGroupDatas.waypoints then
							nextWaypoint = _currentOnBoardGroupDatas.nextWaypoint
						end
					else
						for _, _waypoint in ipairs(_currentOnBoardGroupDatas.waypoints) do
							if _waypoint.name == _currentOnBoardGroupDatas.nextWaypoint then
								nextWaypoint = _currentOnBoardGroupDatas.nextWaypoint
							end
						end
					end
				end

				local newWaypoints = {}
				for i = _currentOnBoardGroupDatas.waypointId, #_currentOnBoardGroupDatas.waypoints do
					table.insert(newWaypoints, _currentOnBoardGroupDatas.waypoints[i])
				end
				
				_currentOnBoardGroupDatas.waypoints = newWaypoints			

				local indexWP
				if nextWaypoint == nil then
					-- Dans ce cas, les unités débarquées suivent une route qui débute
					-- au waypoint le plus proche du transport (exclusion de 200m)
					-- sans prendre en compte les waypoints déjà passés.

					indexWP = ATMEFct.C_CloningContext_startRouteAt(newWaypoints, _transportUnit, 200)
				else
					-- Dans ce cas soit le nom soit le numéro du waypoint est connu.
					-- Dans ce cas, les unités débarquées suivent une route qui débute au waypoint désigné.

					indexWP = ATMEFct.C_CloningContext_startRouteAt(newWaypoints, nextWaypoint, _currentOnBoardGroupDatas.waypointId)
				end

				_currentOnBoardGroupDatas.delta = _currentOnBoardGroupDatas.waypointId + indexWP - 3	-- offset ajusté cause 2 WP de débarquement -> 1er WP ok à 3.
				
				_currentOnBoardGroupDatas.nbWaypoints = #initialWaypoints
				
				table.insert(_currentOnBoardGroupDatas.waypoints, 1, startWP)

				-- Mémorisation de la route initiale du spawnDatas pour après le spawn
				local initialRoute = spawnDatas.datas.route
				
				-- Pour le spawn, création de la bonne route
				spawnDatas.datas.route = {}
				spawnDatas.datas.route.points = _currentOnBoardGroupDatas.waypoints

				-- Mise à jour de la position du premier WP (WP0 dans l'éditeur) avec
				-- les nouvelles coordonnées de la première unité du groupe qui débarque.
				-- Si le groupe est en maintenir position (task = Hold), cette tache est supprimée
				
				spawnDatas.datas.route.points[1].x = spawnDatas.datas.units[1].x
				spawnDatas.datas.route.points[1].y = spawnDatas.datas.units[1].y
				spawnDatas.datas.route.points[1].action = "Off Road"

				local indexTask = 0
				for _i, _task in ipairs(spawnDatas.datas.route.points[1].task.params.tasks) do
					if _task.id == "Hold" then -- Maintenir position activé
						indexTask = _i
					end
				end

				if indexTask ~= 0 then
					table.remove(spawnDatas.datas.route.points[1].task.params.tasks, indexTask)
				end
				
				local rollAxis = ATMEFct.C_Line2D_create(nil, nil, _transportUnit:getRollAxisVector(), _transportUnit:getPosition())
				
				local pAxis
				local newWP = ATMEFct.copyTable(spawnDatas.datas.route.points[1])

				if #spawnDatas.datas.route.points > 1 then
					-- Les unités traversent le transport si elles le percutent (bug DCS pas de contournement)
					-- Aussi, un WP est ajouté dans l'axe du transport à 100m, au plus près du
					-- deuxième WP (WP1 dans l'éditeur de mission ou créé s'il n'existe pas) si
					-- le groupe dispose d'une route de plus d'un WP.

					pAxis = {}
					pAxis[1] = ATMEFct.C_Line2D_getPointFromOrigin(rollAxis:getA(), rollAxis:getB(), rollAxis:getOrigin(), 100)
					pAxis[2] = ATMEFct.C_Line2D_getPointFromOrigin(rollAxis:getA(), rollAxis:getB(), rollAxis:getOrigin(), -100)
					
					local indexPAxis = 1
					local vAxis1 = ATMEFct.C_Vector3D_create(nil, nil, pAxis[1], spawnDatas.datas.route.points[2])
					local vAxis2 = ATMEFct.C_Vector3D_create(nil, nil, pAxis[2], spawnDatas.datas.route.points[2])
					
					if vAxis2:getHModule() < vAxis1:getHModule() then
						indexPAxis = 2
					end
					
					newWP.x = pAxis[indexPAxis].x
					newWP.y = pAxis[indexPAxis].z
					table.insert(spawnDatas.datas.route.points, 2, newWP)
				else
					-- Si le groupe ne dispose pas de route de plus d'un WP, on en crée un à 300m devant
					-- l'unité de transport. Ainsi le déplacement du groupe est maitrisé par l'azimuth du 
					-- transport.
					
					pAxis = ATMEFct.C_Line2D_getPointFromOrigin(rollAxis:getA(), rollAxis:getB(), rollAxis:getOrigin(), 300)
					newWP.x = pAxis.x
					newWP.y = pAxis.z
					table.insert(spawnDatas.datas.route.points, newWP)
				end
				
				-- Mémorisation du groupe spawné pour traitement à la création
				-- du groupe ATME.C_Group
	
				ATMEDatas.spawnedGroupsWaypoints[spawnDatas.datas.name] = {
					waypoints = spawnDatas.datas.route.points,
					initialWaypoints = initialWaypoints,
				}

				coalition.addGroup(spawnDatas.country, spawnDatas.category, spawnDatas.datas)
				
				
				-- Les unités sont créées (event S_EVENT_BIRTH) pendant le addgroup mais parfois
				-- leur groupe associé ne l'est pas encore ... Dans ce cas on mémorise et on traite ci après.
				
				-- On remet le spawnDatas.datas.route sur la route initiale
				spawnDatas.datas.route = initialRoute
				
				while #ATMEDatas.unitsReadyToCreate > 0 do
					local newUnit = table.remove(ATMEDatas.unitsReadyToCreate)
					
					local datas = {
						uncontrolled = false,
						visible = true,
						spawnDatas = spawnDatas,
					}

					ATMEFct.C_AIUnit_create(newUnit, datas, false) -- False car pas de mémo sur débarquement
				end
				
				return false
			else
				ATMEFct.setError("ALL_GROUP_DEAD", 0)
				return true
			end
		else
			ATMEFct.setError("ONLY_WATER", 0)
			return true
		end
	end
end

-- Fonction de création des données de base nécessaire au spawn simple ou multiple.
--   Le groupe _groupeName s'il existe peut lui même être un groupe spawné.
--   Dans ce cas, il faudra utiliser les noms originaux du groupe et unités associés 
--   tels que définis dans la mission. D'où la variable originalMissionNames
--   Si ce groupe n'existe pas alors on se base sur les données de missions.

ATMEFct.C_CloningContext_create = function(_name, _cloningContexts, _initialNames, _newGroupName)
	local groupNameToCopy = _initialNames.groupName
	local newGroupName = _newGroupName
	local initialCoalition
	local initialType
	local initialNames = _initialNames
	
	initialType = ATMEDatas.allGroupsInMission[groupNameToCopy].type
	initialCoalition = ATMEDatas.allGroupsInMission[groupNameToCopy].coalition
	
	local newClonigContext = {}
	
	local name
	local tagsName
	local transportUnitName
	
	if _name ~= nil and _cloningContexts ~= nil then
		name = _name
		_cloningContexts[name] = newClonigContext
		tagsName = ATMEFct.decompString(_name)
	end
	
	setmetatable(newClonigContext, ATMEDatas.C_CloningContext_MT.mt)
	
	local index = ATMEDatas.C_CloningContext_MT:add()
	local localMembers = ATMEDatas.localMembers[index]
	
	function newClonigContext.getIndex(self)
		return index
	end
	
	function newClonigContext.setAircraftsCallsign(self, _callsign)
		local fctDefinition = ATMEFct.getErrorPrefix(newClonigContext) .. "setAircraftsCallsign(callsign)"
		
		if type(_callsign) == "string" or type(_callsign) == "number" then
			if initialType == "plane" or initialType == "helicopter" then
				if #initialNames.unitsName > 0 then
					if initialCoalition == "BLUE" then -- Coalition BLUE
						local aircraftType = ATMEDatas.allGroupsInMission[groupNameToCopy].units[initialNames.unitsName[1]].type
						
						if type(_callsign) == "string" then
							local tokens = ATMEFct.analyzeString(_callsign)
							if #tokens >= 1 and tokens[1].type == "TOKEN_AIRCRAFT_CALLSIGN_NAME" then
								local value
								if #tokens == 2 and tokens[2].type == "TOKEN_INTEGER" and tokens[2].status == "ABSOLUTE" and tokens[2].value <= 9 then
									value = tokens[2].value
								else
									value = 1
								end
								
								if (ATMEConst.specialAircraftTypes[aircraftType] == nil and 
											ATMEConst.blueCallsigns.aircrafts[tokens[1].value].type == "standard") or
									(ATMEConst.specialAircraftTypes[aircraftType] ~= nil and 
											ATMEConst.blueCallsigns.aircrafts[tokens[1].value].type == ATMEConst.specialAircraftTypes[aircraftType]) then
						
									localMembers.callsign = {
										[1] = ATMEConst.blueCallsigns.aircrafts[tokens[1].value].id,
										[2] = value,
										name = ATMEConst.blueCallsigns.aircrafts[tokens[1].value].name .. tokens[2].value,
									}
								else
									ATMEFct.errorOutputCore(nil, fctDefinition, "For BLUE coalition, callsign depends on aircraft type (tanker, awacs...)", false)
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "For BLUE coalition, callsign format must be 'Callsign' or 'Callsign X' x < 9", false)
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "For BLUE coalition, callsign type must be string only", false)
						end
					else --Coalition RED
						local callsignId = 0
						if type(_callsign) == "string" then
							local tokens = ATMEFct.analyzeString(_callsign)
							if #tokens == 1 and tokens[1].type == "TOKEN_INTEGER" and tokens[1].status == "ABSOLUTE" then
								callsignId = math.floor(tonumber(tokens[1].value))
							end
						else
							callsignId = math.floor(_callsign)
						end
						
						if callsignId > 0 and callsignId <= 999 then
							localMembers.callsign[i] = callsignId
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "For RED coalition, callsign format must be 'XXX' or XXX (XXX > 0 and <= 999)", false)
						end
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "can't set callsign - #units = 0 in group", false)
				end
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "callsign type must be string or number ('CALLSIGN X' or 'XXX'/XXX)", true)
		end
	end
	
	function newClonigContext.setAutoRepeatSpawn(self, _quantity, _repeatPeriod)
		local fctDefinition = ATMEFct.getErrorPrefix(newClonigContext) .. "setAutoRepeatSpawn(quantity, repeatPeriod)"
		
		if (type(_quantity) == "number" or type(_quantity) == "string") and (type(_repeatPeriod) == "number" or type(_repeatPeriod) == "string") then
			local maxQuantity = 0

			local err, qtyRandom, qtyMin, qtyMax = ATMEFct.getRandomDatas(_quantity)
			if err ~= nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, err .. "quantity", false)
				return
			else
				if qtyRandom == true then
					if qtyMin >= 1 then
						if qtyMax ~= nil then 
							if qtyMax <= 20 then
								maxQuantity = math.random(qtyMin, qtyMax)
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #3 must be <= 20 in string format quantity", false)
								return
							end
						else
							maxQuantity = qtyMin
							if maxQuantity < 1 then
								ATMEFct.errorOutputCore(nil, fctDefinition, "quantity must be >= 1", false)
								return
							end
						end
						
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be >= 1 in string format quantity", false)
						return
					end
				else
					maxQuantity = qtyMin
					if maxQuantity < 1 then
						ATMEFct.errorOutputCore(nil, fctDefinition, "quantity must be >= 1", false)
						return
					end
				end
			end
			
			local err, repeatRandom, repeatPeriodMin, repeatPeriodMax = ATMEFct.getRandomDatas(_repeatPeriod)
			if err ~= nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, err .. "repeatPeriod", false)
				return
			else
				if repeatRandom == true then
					if repeatPeriodMin < 1 then
						ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be >= 1 in string format repeatPeriod", false)
						return
					end
				else
					if repeatPeriodMin < 1 then
						ATMEFct.errorOutputCore(nil, fctDefinition, "repeatPeriod must be >= 1", false)
						return
					end
				end
			end

			ATMEDatas.spawnGroupDatasRepeat[groupNameToCopy] = {
				groupDatas = self,
				repeatPeriodMin = repeatPeriodMin,
				repeatPeriodMax = repeatPeriodMax,
				repeatRandom = repeatRandom,
				maxQuantity = maxQuantity,
				nbSpawn = 1,
			}
				
		else
			if type(_maxQuantity) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "quantity type must be number or string", true)
			elseif type(_repeatPeriod) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "repeatPeriod type must be number or string", false)
			end
		end
	end
	
	function newClonigContext.resetAutoRepeatSpawn(self)
		ATMEDatas.spawnGroupDatasRepeat[spawnDatas.datas.name] = nil
	end
	
	function newClonigContext.addUnit(self, _unitDatas, _newUnitName)
		local fctDefinition = ATMEFct.getErrorPrefix(newClonigContext) .. "addUnit(unitDatas, [newUnitName])"
		
		if type(_unitDatas) == "table" and (_newUnitName == nil or type(_newUnitName) == "string") then
			local err, value = ATMEFct.C_CloningContext_addUnit(ATMEDatas.allGroupsInMission[groupNameToCopy], _unitDatas, _newUnitName, initialNames, initialType)
			
			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, value, false)
			end
		else
			if type(_unitDatas) ~= "table" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "unitDatas type must be table", true)
			elseif _newUnitName ~= nil and type(_newUnitName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "newUnitName type must be nil or string", true)
			end
		end
	end
	
	function newClonigContext.getName(self)
		return name
	end
	
	function newClonigContext.setWeaponToUnit(self, _unitName, _pylonID, _weapon)
		local fctDefinition = ATMEFct.getErrorPrefix(newClonigContext) .. "addWeaponsToUnit(unitName, pylonID, weapon)"
		
		if type(_unitName) == "string" and type(_pylonID) == "number" and (_weapon == nil or type(_weapon) == "table") then
			if initialType == "plane" or initialType == "helicopter" then
				local unitDatas = ATMEDatas.allGroupsInMission[groupNameToCopy].units[_unitName]
				
				if unitDatas ~= nil then
					if _pylonID > 0 then
						if type(_weapon) == "table" then
							unitDatas.payload.pylons[_pylonID] = ATMEFct.copyTable(_weapon)
						else
							unitDatas.payload.pylons[_pylonID] = nil
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "pylonID must be > 0", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "unit " .. _unitName .. " doesn't exist", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "unit " .. _unitName .. " must be an aircraft unit", false)
			end
		else
			if type(_unitName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "unitName type must be string", true)
			elseif type(_pylonID) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "pylonID type must be number", false)
			elseif _weapon ~= nil and type(_weapon) ~= "table" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "weapon type must be nil or table", true)
			end
		end
	end
	
	function newClonigContext.getGroupName(self)
		return groupNameToCopy
	end
	
	function newClonigContext.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newClonigContext) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newClonigContext.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newClonigContext.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newClonigContext) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newClonigContext.addSpawnCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newClonigContext) .. "addSpawnCallback(fct)"
		
		if type(_fct) == "function" then
			localMembers.spawnCallback[ATMEDatas.currentModuleName] = _fct
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newClonigContext.removeSpawnCallback(self)
		localMembers.spawnCallback[ATMEDatas.currentModuleName] = nil
	end
	
	function newClonigContext.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	function newClonigContext.spawn(self, _route)
		if ATMEDatas.exitATME == false then -- Test nécessaire car sinon problème sur l'événement du spawn (S_EVENT_BIRTH) 
											-- désactivé par ATMEDatas.exitATME à true
			local route
			local fctDefinition = ATMEFct.getErrorPrefix(newClonigContext) .. "spawn([route])"

			if _route ~= nil then
				if ATMEFct.isObjectClass(_route, ATME.C_Route) == true then
					route = _route
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "route type must be nil or ATME.C_Route", false)
					return
				end
			else
				route = ATMEFct.C_Route_create(initialType)
				ATMEFct.addWPFrom(route, groupNameToCopy, false, true, 0, 0)
			end

			local localMembersRoute = ATMEDatas.localMembers[route:getIndex()]

			if initialType == localMembersRoute.type then
				-- Mémorisation du groupe spawné pour traitement à la création
				-- du groupe (C_Group)

				local err, waypoints = ATMEFct.C_Route_createWaypoints(route)

				if err == false then
					local spawnDatas
					local newInitialNames

					
					-- Si le groupe à créer existe, il faudra toujours postfixer son nom d'origine
					-- avec " - Spawned #XXX" où XXX est une valeur incrémentale. Néanmoins,
					-- si un groupe avec le même nom ainsi posfixé a déjà été créé puis détruit,
					-- la valeur XXX associée pourra être réutilisée (entrée freeIds).
					-- Idem pour les unités. Attention, même si le nouveau groupe a un nom différent,
					-- les unités elles conservent leur nom d'origine donc l'incrément s'applique 
					-- plus souvent.
					
					local idSpawnedList = {
						units = {},
					}


					if ATMEDatas.groups[newGroupName] ~= nil or ATMEDatas.groupsOnBoard[newGroupName] ~= nil or ATMEDatas.groupsToRespawn[newGroupName] ~= nil or
																	ATMEDatas.groupsReadyToBirth[newGroupName] ~= nil or ATMEDatas.newCreatedGroups[newGroupName] ~= nil then

						_, newInitialNames = ATMEFct.C_CloningContext_getInitialNames(newGroupName)

						local newId
						if ATMEDatas.spawnedGroups[newInitialNames.groupName] == nil then
							ATMEDatas.spawnedGroups[newInitialNames.groupName] = {
								max = 1,
								freeIds = {},
								cloningContext = self,
							}
							
							newId = ATMEDatas.spawnedGroups[newInitialNames.groupName].max
						else
							if ATMEDatas.spawnedGroups[newInitialNames.groupName].cloningContext == self then
								if #ATMEDatas.spawnedGroups[newInitialNames.groupName].freeIds == 0 then
									ATMEDatas.spawnedGroups[newInitialNames.groupName].max = ATMEDatas.spawnedGroups[newInitialNames.groupName].max + 1
									newId = ATMEDatas.spawnedGroups[newInitialNames.groupName].max
									
									if newId > 999 then
										ATMEFct.errorOutputCore(nil, fctDefinition, "Too much group with the same name (#XXX > 999)", false)
										return
									end
								else
									newId = table.remove(ATMEDatas.spawnedGroups[newInitialNames.groupName].freeIds)
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "New clonning groups " .. newInitialNames.groupName .. " already used", false)
								return
							end
						end

						newGroupName = string.format("%s - Spawned #%03d", newInitialNames.groupName, newId)
						idSpawnedList.groupSpawnedId = newId
					else
						-- Le nom du groupe est nouveau mais le nom des unités sont bien la duplication d'unités existantes
						newInitialNames = {
							groupName = newGroupName,
							unitsName = ATMEDatas.allInitialGroupNames[groupNameToCopy].unitsName,
						}
					end

					spawnDatas = {
						initialNames = newInitialNames,
						idSpawnedList = idSpawnedList,
						country = ATMEDatas.allGroupsInMission[groupNameToCopy].country,
						category = ATMEDatas.allGroupsInMission[groupNameToCopy].category,
						type = initialType,
						datas = {
							name = newGroupName,
							hidden = ATMEDatas.allGroupsInMission[groupNameToCopy].hidden,
							start_time = 0,
							task = ATMEDatas.allGroupsInMission[groupNameToCopy].task,
							visible = false,
							taskSelected = true,
							lateActivation = false,
							tasks = {},
							units = {},
							route = {},
						},
						
						spawnCallback = localMembers.spawnCallback,
						cloningContext = self,
					}
					
					local first = true
					local position0 = {}

					for _i, _unitName in ipairs(newInitialNames.unitsName) do
						local unitName = _unitName

						if ATMEDatas.AIUnits[_unitName] ~= nil or ATMEDatas.unitsReadyToBirth[_unitName] ~= nil then
							local newId
							if ATMEDatas.spawnedUnits[_unitName] == nil then
								ATMEDatas.spawnedUnits[_unitName] = {
									max = 1,
									freeIds = {},
								}
								newId = ATMEDatas.spawnedUnits[_unitName].max
							else
								if #ATMEDatas.spawnedUnits[_unitName].freeIds == 0 then
									ATMEDatas.spawnedUnits[_unitName].max = ATMEDatas.spawnedUnits[_unitName].max + 1
									newId = ATMEDatas.spawnedUnits[_unitName].max
									if newId > 999 then
										ATMEFct.errorOutputCore(nil, fctDefinition, "Too much units with the same name (#XXX > 999)", false)
										return
									end
								else
									newId = table.remove(ATMEDatas.spawnedUnits[_unitName].freeIds)
								end
							end

							unitName = string.format("%s - Spawned #%03d", _unitName, newId)
							idSpawnedList.units[_unitName] = newId
						end
						
						local initialUnit = ATMEDatas.allAIUnitsInMission[_unitName].unit
						local x = 0
						local y = 0
						local heading = 0
						
						if initialUnit.x ~= nil then
							x = initialUnit.x
						end
						
						if initialUnit.y ~= nil then
							y = initialUnit.y
						end
						
						if first == true then
							position0.x = x
							position0.y = y
							first = false
						end
						
						x = waypoints[1].x + x - position0.x
						y = waypoints[1].y + y - position0.y
						
						if #waypoints > 1 then
							local dx = waypoints[2].x - x
							local dy = waypoints[2].y - y
							
							if dx == 0 and dy == 0 then
								if initialUnit.heading == nil then
									heading = math.rad(math.random(-179, 180))
								else
									heading = initialUnit.heading
								end
							else
								heading = math.atan2(dy, dx)
							end
						else
							if initialUnit.heading == nil then
								heading = math.rad(math.random(-179, 180))
							else
								heading = initialUnit.heading
							end
						end
						
						spawnDatas.datas.units[_i] = {
							type = initialUnit.type,
							name = unitName,
							skill = initialUnit.skill,
							alt = waypoints[1].alt,
							x = x,
							y = y,
							alt_type = initialUnit.alt_type,
							playerCanDrive = initialUnit.playerCanDrive,
							heading = heading,
							payload = initialUnit.payload,
							speed = initialUnit.speed,
							callsign = initialUnit.callsign,
							onboard_num = initialUnit.onboard_num,
							livery_id = initialUnit.livery_id,
						}
					end

					if initialType == "plane" or initialType == "helicopter" then
						if localMembers.callsign ~= nil then
							if initialCoalition == "BLUE" then
								for _i, _unit in ipairs(spawnDatas.datas.units) do
									_unit.callsign = {
										[1] = localMembers.callsign[1],
										[2] = localMembers.callsign[2],
										[3] = _i,
										name = localMembers.callsign.name .. _i,
									}
								end
								
								localMembers.callsign[2] = localMembers.callsign[2] + 1
								if localMembers.callsign[2] > 9 then
									localMembers.callsign[2] = 1
								end
							else -- RED
								for _i, _unit in ipairs(spawnDatas.datas.units) do
									_unit.callsign = string.format("%03d", localMembers.callsign + _i - 1)
								end
								
								localMembers.callsign = localMembers.callsign + 10
							end
						end
					end
					
					spawnDatas.datas.route.points = waypoints
					
					-- Mémorisation des GroupDatas pour un éventuel spawn multiple
					if ATMEDatas.groups[spawnDatas.datas.name] == nil and ATMEDatas.groupsOnBoard[spawnDatas.datas.name] == nil then
						if ATMEDatas.spawnGroupDatasRepeat[groupNameToCopy] ~= nil then
							ATMEDatas.spawnGroupDatasRepeat[groupNameToCopy].arg = route
							if ATMEDatas.spawnGroupDatasRepeat[groupNameToCopy].repeatRandom == true then
								ATMEDatas.spawnGroupDatasRepeat[groupNameToCopy].currentTime = math.random(ATMEDatas.spawnGroupDatasRepeat[groupNameToCopy].repeatPeriodMax, ATMEDatas.spawnGroupDatasRepeat[groupNameToCopy].repeatPeriodMax)
							else
								ATMEDatas.spawnGroupDatasRepeat[groupNameToCopy].currentTime = ATMEDatas.spawnGroupDatasRepeat[groupNameToCopy].repeatPeriodMin
							end
						end

						ATMEFct.C_CloningContext_spawn(spawnDatas)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Group " .. localMembers.datas.name .. " already exists or on board", false)
						return
					end
				else
					ATMEFct.setError("SPAWN_NO_WAYPOINTS")
					return
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Route must have same type as " .. localMembers.datas.name, false)
				return
			end
		end
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newClonigContext
end

ATMEDatas.C_CloningContext_MT = {
	className = "C_CloningContext",
	
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["spawn"] = true,
		["setAutoRepeatSpawn"] = true,
		["resetAutoRepeatSpawn"] = true,
		["setAircraftsCallsign"] = true,
		["addUnit"] = true,
		["setWeaponToUnit"] = true,
		["getName"] = true,
		["getGroupName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["addSpawnCallback"] = true,
		["removeSpawnCallback"] = true,
		["getUserPrivateDatas"] = true,
	},
	
	add = function(self)
		local item = {
			callsigns = nil,
			userPrivateDatas = {},						-- Point d'entrée des données privées : un par module
			spawnCallback = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].spawnCallback = nil
	end,
}

ATME.C_CloningContext = {
	getClassName = function(self)
		return ATMEDatas.C_CloningContext_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_CloningContext_MT.parentClass
	end,
	
	createNamed = function(_name, _origin, _newGroupName)
		local fctDefinition = "Function ATME.CloningContext.createNamed(name, origin, [newGroupName])"
		
		if type(_name) == "string" and 
			(type(_origin) == "string" or type(_origin) == "table") and (type(_newGroupName) == "string" or _newGroupName == nil) then
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newContext = ATMEFct.C_CloningContext_createNamed(localMembersModule.cloningContexts, _name, _origin, _newGroupName)				
			
			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, newContext, false)
				return
			end
			
			return newContext
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif _newGroupName ~= nil and type(_newGroupName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "newGroupName type must be nil or string", false)
			end
		end
	end,
	
	getByName = function(_name)
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].cloningContexts[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_CloningContext.getByName(name)", "name type must be string", false)
		end
	end,
}

setmetatable(ATME.C_CloningContext,ATMEDatas.C_CloningContext_MT)

ATMEDatas.C_CloningContext_MT.__call = function(t, _origin, _newGroupName)
	local fctDefinition = "Constructor ATME.C_CloningContext(origin, [newGroupName])"

	if (type(_origin) == "string" or type(_origin) == "table") and (type(_newGroupName) == "string" or _newGroupName == nil) then
		if type(_origin) == "table" then
			if ATMEFct.C_CloningContext_controlTable(_origin, _newGroupName) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "Bad cloning table format", false)
				return
			end
			
			if _newGroupName ~= nil then
				_origin.datas.name = _newGroupName
			end
		end
		
		local newGroupName = _newGroupName
		if _newGroupName == nil then
			if type(_origin) == "string" then
				newGroupName = _origin
			else
				newGroupName = _origin.datas.name 
			end
		else
			local d,f = string.find(newGroupName, "|", 1)
			if d ~= nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, "Name of group " .. newGroupName .. " can't contain '|' character", false)
				return
			end
			
			d,f = string.find(newGroupName:upper(), "%s+SPAWNED%s+", 1)
			if d ~= nil then
				ATMEFct.errorOutputCore(nil, fctDefinition, "Name of group " .. newGroupName .. " can't contain 'Spawned' word", false)
				return
			end
		end
		
		local err, initialNames = ATMEFct.C_CloningContext_getInitialNames(_origin)
		
		if err == nil then
			local foundPlayer = false
			for _, _unitDatas in pairs(ATMEDatas.allGroupsInMission[initialNames.groupName].units) do
				if _unitDatas.skill == "Player" or _unitDatas.skill == "Client" then
					foundPlayer = true
				end
			end
			
			if foundPlayer == false then
				-- Le nom du nouveau groupe (_newGroupName) ne doit pas correspondre à un groupe existant,
				-- à un groupe embarqué ou à un groupe en cours de spawn.

				return ATMEFct.C_CloningContext_create(nil, nil, initialNames, newGroupName)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Group " .. _origin .. " has one or more player units", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, err, false)
		end
	else
		if type(_origin) ~= "string" and type(_origin) ~= "table" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "origin type must be string or table (DCS fromat)", false)
		elseif _newGroupName ~= nil and type(_newGroupName) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "newGroupName type must be nil or string", false)
		end
	end
end

ATMEDatas.C_CloningContext_MT.mt.__metatable = "Private"
ATMEDatas.C_CloningContext_MT.mt.__index = ATME.C_CloningContext
ATMEDatas.C_CloningContext_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_CloningContext_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_CloningContext : Bad key " .. k, "table locked", false)
	end
end


----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_StaticCloningContext : gère les données de clonage nécessaires à la duplication des objets 
-- statiques
----------------------------------------------------------------------------------------------------------

ATMEFct.C_StaticCloningContext_createNamed = function(_namedTable, _name, _countryName, _staticObjectName, _staticObjectDatas)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return false
	else
		local countryName = string.upper(_countryName)
		if ATMEConst.countries[countryName] ~= nil then
			if ATMEDatas.staticObjects[_staticObjectName] == nil then
				local d,f = string.find(_staticObjectName, "|", 1)
				if d == nil then
					if type(_staticObjectDatas.category) == "string" and type(_staticObjectDatas.type) == "string" then
						if ATMEConst.staticObjectCategories[string.upper(_staticObjectDatas.category)] ~= nil then
							return false, ATMEFct.C_StaticCloningContext_create(_name, _namedTable, countryName, _staticObjectName, _staticObjectDatas)	
						else
							return true, "Category " .. _staticObjectDatas.category .. " unknown"
						end
					else
						return true, "Bad type for staticObjectDatas entries"
					end
				else
					return true, "Name " .. _staticObjectName .. " can't contain '|' character"
				end
			else
				return true, "Cargo " .. _staticObjectName .. " already exists"
			end
		else
			return true, "Country " .. _countryName .. " unknown"
		end
	end
end

ATMEFct.C_StaticCloningContext_create = function(_name, _staticCloningContexts, _countryName, _staticObjectName, _staticObjectDatas)
	local staticObjectInitialName = _staticObjectName
	local countryName = _countryName
	local staticObjectDatas
	
	local newStaticCloningContext = {}
	
	local name
	local tagsName
	
	staticObjectDatas = {
		type = _staticObjectDatas.type,
		canCargo = false,
		category = ATMEConst.staticObjectCategories[string.upper(_staticObjectDatas.category)].category,
		shape_name = _staticObjectDatas.shape_name,
		livery_id = _staticObjectDatas.livery_id,
	}
	
	if _name ~= nil and _staticCloningContexts ~= nil then
		name = _name
		_staticCloningContexts[name] = newStaticCloningContext
		tagsName = ATMEFct.decompString(_name)
	end

	
	setmetatable(newStaticCloningContext, ATMEDatas.C_StaticCloningContext_MT.mt)
	
	local index = ATMEDatas.C_StaticCloningContext_MT:add()
	local localMembers = ATMEDatas.localMembers[index]
						
	function newStaticCloningContext.getIndex(self)
		return index
	end
	
	function newStaticCloningContext.getName(self)
		return name
	end
	
	function newStaticCloningContext.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newStaticCloningContext) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newStaticCloningContext.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newStaticCloningContext.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newStaticCloningContext) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newStaticCloningContext.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	function newStaticCloningContext.addSpawnCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newStaticCloningContext) .. "addSpawnCallback(fct)"
		
		if type(_fct) == "function" then
			localMembers.spawnCallback[ATMEDatas.currentModuleName] = _fct
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newStaticCloningContext.removeSpawnCallback(self)
		localMembers.spawnCallback[ATMEDatas.currentModuleName] = nil
	end
	
	function newStaticCloningContext.spawn(self, _position, _heading, _dead)
		local fctDefinition = ATMEFct.getErrorPrefix(newStaticCloningContext) .. "spawn(position, heading, [dead])"
		
		if ATMEDatas.exitATME == false then -- Test nécessaire car sinon problème sur l'événement du spawn (S_EVENT_BIRTH) 
											-- désactivé par ATMEDatas.exitATME à true
											
			if (type(_position) == "string" or ATME.isPoint(_position) ~= POINT_BAD) and 
				type(_heading) == "number" and (_dead == nil or type(_dead) == "boolean") then		
				
				local point
				local headingValue = _heading
				local newId
											
				if ATMEDatas.spawnedStaticObjects[staticObjectInitialName] == nil then
					ATMEDatas.spawnedStaticObjects[staticObjectInitialName] = {
						max = 1,
						freeIds = {},
					}
					newId = ATMEDatas.spawnedStaticObjects[staticObjectInitialName].max
				else
					if #ATMEDatas.spawnedStaticObjects[staticObjectInitialName].freeIds == 0 then
						ATMEDatas.spawnedStaticObjects[staticObjectInitialName].max = ATMEDatas.spawnedStaticObjects[staticObjectInitialName].max + 1
						newId = ATMEDatas.spawnedStaticObjects[staticObjectInitialName].max
						
						if newId > 999 then
							ATMEFct.errorOutputCore(nil, fctDefinition, "Too much static objects with the same name (#XXX > 999)", false)
							return
						end
					else
						newId = table.remove(ATMEDatas.spawnedStaticObjects[staticObjectInitialName].freeIds)
					end
				end
				
				local newStaticObjectName = string.format("%s - Spawned #%03d", staticObjectInitialName, newId)
				
				if type(_position) == "string" then
					local zone = ATMEFct.getDCSZone(_position)
					if zone ==  nil then
						ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _position .. " unknown", false)
						return
					end

					point = {}
					local angle = 2 * math.pi * math.random()
					local radius = math.random(0, zone.radius)
					point.x = zone.point.x + radius * math.sin(angle)
					point.z = zone.point.z + radius * math.cos(angle)
				else
					point = ATMEFct.convertToPoint3D(_position) 
				end
					
				if headingValue >= 0 and headingValue <= 359 then
					staticObjectDatas.name = newStaticObjectName
					staticObjectDatas.x = point.x
					staticObjectDatas.y = point.z
					staticObjectDatas.heading = math.rad(headingValue)
					staticObjectDatas.dead = _dead
					
					-- Permet de transmettre la catégorie lors de la création de l'objet. A ce moment
					-- là sera remis à nil
			
			
					ATMEDatas.staticObjectsToBirth[newStaticObjectName] = {
						isSpawned = true,
						spawnCallback = localMembers.spawnCallback,
						staticCloningContext = self,
						category = staticObjectDatas.category,
					}
					
					coalition.addStaticObject(ATMEConst.countries[countryName], staticObjectDatas)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "heading must be >= 0 and <= 359", false)
				end
			end
		else
			if ATME.isPoint(_position) == POINT_BAD and type(_position) ~= string then
				ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be string or table (x,y)/(x,y,z)", true)
			elseif type(_heading) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "heading type must be number", false)
			elseif _dead ~= nil and type(_dead) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "dead type must be boolean or nil", false)
			end
		end
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end

	return newStaticCloningContext
end

ATMEDatas.C_StaticCloningContext_MT = {
	className = "C_StaticCloningContext",
	
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["spawn"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getUserPrivateDatas"] = true,
		["addSpawnCallback"] = true,
		["removeSpawnCallback"] = true,
	},

	add = function(self)
		local item = {
			userPrivateDatas = {},						-- Point d'entrée des données privées : un par module
			spawnCallback = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].spawnCallback = nil
	end,
}

ATME.C_StaticCloningContext = {
	getClassName = function(self)
		return ATMEDatas.C_StaticCloningContext_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_StaticCloningContext_MT.parentClass
	end,
	
	createNamed = function(_name, _countryName, _staticObjectName, _staticObjectDatas)
		local fctDefinition = "Function ATME.C_StaticCloningContext.createNamed(name, countryName, staticObjectName, staticObjectDatas)"
		
		if type(_name) == "string" and type(_countryName) == "string" and type(_staticObjectDatas) == "table" and type(_staticObjectName) == "string" then
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newContext = ATMEFct.C_StaticCloningContext_createNamed(localMembersModule.staticCloningContexts, _name, _countryName, _staticObjectName, _staticObjectDatas)				
			
			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, newContext, false)
				return
			end
			
			return newContext
		else
			if type(_name) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif type(_countryName) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "countryName type must be string", false)
			elseif type(_staticObjectName) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "staticObjectName type must be string", false)
			elseif type(_staticObjectData) ~= "table" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "staticObjectDatas type must be table", false)
			
			end
		end
	end,
	
	getByName = function(_name)
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].staticCloningContexts[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_StaticCloningContext.getByName(name)", "name type must be string", false)
		end
	end,
}

setmetatable(ATME.C_StaticCloningContext,ATMEDatas.C_StaticCloningContext_MT)

ATMEDatas.C_StaticCloningContext_MT.__call = function(t, _countryName, _staticObjectName, _staticObjectDatas)
	local fctDefinition = "Constructor ATME.C_StaticCloningContext(countryName, staticObjectName, staticObjectDatas)"

	if type(_countryName) == "string" and type(_staticObjectDatas) == "table" and type(_staticObjectName) == "string" then
		local countryName = string.upper(_countryName)
		if ATMEConst.countries[countryName] ~= nil then
			if ATMEDatas.staticObjects[_staticObjectName] == nil then
				local d,f = string.find(_staticObjectName, "|", 1)
				if d == nil then
					if type(_staticObjectDatas.category) == "string" and type(_staticObjectDatas.type) == "string" then
						if ATMEConst.staticObjectCategories[string.upper(_staticObjectDatas.category)] ~= nil then
							return ATMEFct.C_StaticCloningContext_create(nil, nil, countryName, _staticObjectName, _staticObjectDatas)	
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Category " .. _staticObjectDatas.category .. " unknown", false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Bad type for staticObjectDatas entries", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "name " .. _staticObjectName .. " can't contain '|' character", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Static object " .. _staticObjectName .. " already exists", false)
			end
			
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "Country " .. _countryName .. " unknown", false)
		end
	else
		if type(_countryName) ~= "string" then 
			ATMEFct.errorOutputCore(nil, fctDefinition, "countryName type must be string", false)
		elseif type(_staticObjectName) ~= "string" then 
			ATMEFct.errorOutputCore(nil, fctDefinition, "staticObjectName type must be string", false)
		elseif type(_staticObjectData) ~= "table" then 
			ATMEFct.errorOutputCore(nil, fctDefinition, "staticObjectDatas type must be table", false)
		end
	end
end

ATMEDatas.C_StaticCloningContext_MT.mt.__metatable = "Private"
ATMEDatas.C_StaticCloningContext_MT.mt.__index = ATME.C_StaticCloningContext
ATMEDatas.C_StaticCloningContext_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_StaticCloningContext_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_StaticCloningContext : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_CargoCloningContext : gère les données de clonage nécessaires à la duplication 
-- des cargos
----------------------------------------------------------------------------------------------------------

ATMEFct.C_CargoCloningContext_createNamed = function(_namedTable, _name, _countryName, _cargoName, _cargoType)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return false
	else
		local countryName = string.upper(_countryName)
		if ATMEConst.countries[countryName] ~= nil then
			if ATMEDatas.cargos[_cargoName] == nil then
				local d,f = string.find(_cargoName, "|", 1)
				if d == nil then
					local typeCargo = string.upper(_cargoType)
					if ATMEConst.cargos[typeCargo] ~= nil then
						return false, ATMEFct.C_CargoCloningContext_create(_name, _namedTable, countryName, _cargoName, typeCargo)	
					else
						return true, "Type name " .. _cargoType .. " unknown"
					end
				else
					return true, "name " .. _cargoName .. " can't contain '|' character"
				end
			else
				return true, "Cargo " .. _cargoName .. " already exists"
			end
		else
			return true, "Country " .. _countryName .. " unknown"
		end
	end
end

ATMEFct.C_CargoCloningContext_create = function(_name, _staticCloningContexts, _countryName, _cargoName, _typeCargo)
	local staticObjectInitialName = _cargoName
	local countryName = _countryName
	local cargoDatas
	
	local newCargoCloningContext = {}
	
	local name
	local tagsName
	local mass = 100
	
	local rateSpecial = ATMEConst.cargos[_typeCargo].rate
	if rateSpecial == nil then
		rateSpecial = 100
	end
		
	cargoDatas = {
		type = ATMEConst.cargos[_typeCargo].type,
		canCargo = true,
		shape_name = ATMEConst.cargos[_typeCargo].shape,
		rate = rateSpecial,
	}
	
	if _name ~= nil and _staticCloningContexts ~= nil then
		name = _name
		_staticCloningContexts[name] = newCargoCloningContext
		tagsName = ATMEFct.decompString(_name)
	end
	
	setmetatable(newCargoCloningContext, ATMEDatas.C_CargoCloningContext_MT.mt)
	
	local index = ATMEDatas.C_CargoCloningContext_MT:add()
	local localMembers = ATMEDatas.localMembers[index]
						
	function newCargoCloningContext.getIndex(self)
		return index
	end
	
	function newCargoCloningContext.getName(self)
		return name
	end
	
	function newCargoCloningContext.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargoCloningContext) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newCargoCloningContext.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newCargoCloningContext.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargoCloningContext) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newCargoCloningContext.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	function newCargoCloningContext.addSpawnCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargoCloningContext) .. "addSpawnCallback(fct)"
		
		if type(_fct) == "function" then
			localMembers.spawnCallback[ATMEDatas.currentModuleName] = _fct
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newCargoCloningContext.removeSpawnCallback(self)
		localMembers.spawnCallback[ATMEDatas.currentModuleName] = nil
	end
	
	function newCargoCloningContext.setWeight(self, _mass)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargoCloningContext) .. "setWeight(mass)"
		
		if type(_mass) == "number" then
			if math.floor(_mass) >= 0 then
				mass = math.floor(_mass)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "mass must be >= 0", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "mass type must be number", true)
		end
	end
	
	function newCargoCloningContext.spawn(self, _position, _heading, _dead)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargoCloningContext) .. "spawn(position, heading, [dead])"
		
		if ATMEDatas.exitATME == false then -- Test nécessaire car sinon problème sur l'événement du spawn (S_EVENT_BIRTH) 
											-- désactivé par ATMEDatas.exitATME à true
											
			if (type(_position) == "string" or ATME.isPoint(_position) ~= POINT_BAD) and 
				type(_heading) == "number" and (_dead == nil or type(_dead) == "boolean") then		
				
				local point
				local headingValue = _heading
				local newId
											
				if ATMEDatas.spawnedCargos[staticObjectInitialName] == nil then
					ATMEDatas.spawnedCargos[staticObjectInitialName] = {
						max = 1,
						freeIds = {},
					}
					newId = ATMEDatas.spawnedCargos[staticObjectInitialName].max
				else
					if #ATMEDatas.spawnedCargos[staticObjectInitialName].freeIds == 0 then
						ATMEDatas.spawnedCargos[staticObjectInitialName].max = ATMEDatas.spawnedCargos[staticObjectInitialName].max + 1
						newId = ATMEDatas.spawnedCargos[staticObjectInitialName].max
						
						if newId > 999 then
							ATMEFct.errorOutputCore(nil, fctDefinition, "Too much static objects with the same name (#XXX > 999)", false)
							return
						end
					else
						newId = table.remove(ATMEDatas.spawnedCargos[staticObjectInitialName].freeIds)
					end
				end
				
				local newStaticObjectName = string.format("%s - Spawned #%03d", staticObjectInitialName, newId)
				
				if type(_position) == "string" then
					local zone = ATMEFct.getDCSZone(_position)
					if zone ==  nil then
						ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _position .. " unknown", false)
						return
					end

					point = {}
					local angle = 2 * math.pi * math.random()
					local radius = math.random(0, zone.radius)
					point.x = zone.point.x + radius * math.sin(angle)
					point.z = zone.point.z + radius * math.cos(angle)
				else
					point = ATMEFct.convertToPoint3D(_position) 
				end
					
				if headingValue >= 0 and headingValue <= 359 then
					cargoDatas.name = newStaticObjectName
					cargoDatas.x = point.x
					cargoDatas.y = point.z
					cargoDatas.heading = math.rad(headingValue)
					cargoDatas.dead = _dead
					cargoDatas.mass = mass
					
					ATMEDatas.cargosToBirth[newStaticObjectName] = {
						isSpawned = true,
						spawnCallback = localMembers.spawnCallback,
						cargoCloningContext = self,
					}

					coalition.addStaticObject(ATMEConst.countries[countryName], cargoDatas)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "heading must be >= 0 and <= 359", false)
				end
			end
		else
			if ATME.isPoint(_position) == POINT_BAD and type(_position) ~= string then
				ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be string or table (x,y)/(x,y,z)", true)
			elseif type(_heading) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "heading type must be number", false)
			elseif _dead ~= nil and type(_dead) ~= "boolean" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "dead type must be boolean or nil", false)
			end
		end
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newCargoCloningContext
end

ATMEDatas.C_CargoCloningContext_MT = {
	className = "C_CargoCloningContext",
	
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["setWeight"] = true,
		["spawn"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getUserPrivateDatas"] = true,
		["addSpawnCallback"] = true,
		["removeSpawnCallback"] = true,
	},
	
	add = function(self)
		local item = {
			userPrivateDatas = {},						-- Point d'entrée des données privées : un par module
			spawnCallback = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].spawnCallback = nil
	end,
}

ATME.C_CargoCloningContext = {
	getClassName = function(self)
		return ATMEDatas.C_CargoCloningContext_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_CargoCloningContext_MT.parentClass
	end,
	
	createNamed = function(_name, _countryName, _cargoName, _cargoType)
		local fctDefinition = "Function ATME.C_CargoCloningContext.createNamed(name, countryName, cargoName, typeName)"
		
		if type(_name) == "string" and type(_countryName) == "string" and type(_cargoType) == "string" and type(_cargoName) == "string" then
			local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
			local err, newContext = ATMEFct.C_CargoCloningContext_createNamed(localMembersModule.cargoCloningContexts, _name, _countryName, _cargoName, _cargoType)				
			
			if err == true then
				ATMEFct.errorOutputCore(nil, fctDefinition, newContext, false)
				return
			end
			
			return newContext
		else
			if type(_name) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif type(_countryName) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "countryName type must be string", false)
			elseif type(_cargoName) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "cargoName type must be string", false)
			elseif type(_cargoType) ~= "string" then 
				ATMEFct.errorOutputCore(nil, fctDefinition, "typeName type must be string", false)
			
			end
		end
	end,
	
	getByName = function(_name)
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].cargoCloningContexts[_name]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_CargoCloningContext.getByName(name)", "name type must be string", false)
		end
	end,
}

setmetatable(ATME.C_CargoCloningContext,ATMEDatas.C_CargoCloningContext_MT)

ATMEDatas.C_CargoCloningContext_MT.__call = function(t, _countryName, _cargoName, _cargoType)
	local fctDefinition = "Constructor ATME.C_CargoCloningContext(countryName, staticObjectName, typeName)"

	if type(_countryName) == "string" and type(_cargoType) == "string" and type(_cargoName) == "string" then
		local countryName = string.upper(_countryName)
		if ATMEConst.countries[countryName] ~= nil then
			if ATMEDatas.cargos[_cargoName] == nil then
				local d,f = string.find(_cargoName, "|", 1)
				if d == nil then
					local typeCargo = string.upper(_cargoType)
					if ATMEConst.cargos[typeCargo] ~= nil then
						return ATMEFct.C_CargoCloningContext_create(nil, nil, countryName, _cargoName, typeCargo)	
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Type name " .. _cargoType .. " unknown", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "name " .. _cargoName .. " can't contain '|' character", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Cargo " .. _cargoName .. " already exists", false)
			end
			
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "Country " .. _countryName .. " unknown", false)
		end
	else
		if type(_countryName) ~= "string" then 
			ATMEFct.errorOutputCore(nil, fctDefinition, "countryName type must be string", false)
		elseif type(_cargoName) ~= "string" then 
			ATMEFct.errorOutputCore(nil, fctDefinition, "cargoName type must be string", false)
		elseif type(_cargoType) ~= "string" then 
			ATMEFct.errorOutputCore(nil, fctDefinition, "typeName type must be string", false)
		end
	end
end

ATMEDatas.C_CargoCloningContext_MT.mt.__metatable = "Private"
ATMEDatas.C_CargoCloningContext_MT.mt.__index = ATME.C_CargoCloningContext
ATMEDatas.C_CargoCloningContext_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_CargoCloningContext_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_CargoCloningContext : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_FindContext : Gestion des contextes de recherche 
--                   .
-- Un context de recherche permet de fixer les critères, le type d'unités et les filtres. Une callback peut
-- être associée
----------------------------------------------------------------------------------------------------------

ATMEFct.C_FindContext_createNamed = function(_namedTable, _name, _findType, _findCategories)
	if _namedTable[_name] ~= nil then
		ATMEFct.setError("ALREADY_CREATED")
		return false
	else
		local d,f = string.find(_name, "|", 1)
		if d == nil then
			return false, ATMEFct.C_FindContext_create(_name, _namedTable, _findType, _findCategories)
		else
			return true, "name can't contain '|' character"
		end
	end
end

ATMEFct.C_FindContext_Object_isInZone = function(_object, _datas)
	if _datas.zone ~= nil then
		local point = _object:getPosition()
		return ATMEFct[ATMEConst.indZoneConfig[_datas.zone:getClassName()].fctIsInside](point, _datas)
	end
	
	return true
end

ATMEFct.C_FindContext_setToList = function(_datas, _group, _id)
	ATMEConst.findClasses[_datas.type].fct(_datas, _group, _id)
end

ATMEFct.C_FindContext_search = function(_object, _datas)
	ATMEConst.findClasses[_datas.type].search_fct(_object, _datas)
end

ATMEFct.C_FindContext_find = function(_type, _zone, _datas)
	if _type ~= "AIRBASES" then -- world.searchObjects ne fonctionne pas avec BASE - Bug DCS pas d'objet retourné.
		if ATMEFct.isObjectClass(_zone, ATME.C_ZonesMgr) == true then
			local localMembersZonesMgr = ATMEDatas.localMembers[_zone:getIndex()]

			for _, _zoneItem in ipairs(localMembersZonesMgr.zonesList) do
				local localMembersZone = ATMEDatas.localMembers[_zoneItem.zone:getIndex()]

				_datas.zone = _zoneItem.zone
				_datas.actualCenter = ATMEFct.getActualCenter(localMembersZone.center, _zoneItem.linkedUnitDatas)
				_datas.actualAzimuth = ATMEFct.getActualAzimuth(localMembersZone.azimuth, _zoneItem.linkedUnitDatas)
					
				local idVolume, searchFct, paramsVolume = ATMEFct[ATMEConst.indZoneConfig[_zoneItem.zone:getClassName()].fctEvalSearchVolume](_datas)
				
				local volume = {
					id = idVolume,
					params = paramsVolume,
				}

				world.searchObjects(ATMEConst.findClasses[_datas.type].category, volume, ATMEFct[searchFct], _datas)
			end
		elseif _zone ~= nil then
			local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]

			_datas.zone = _zone
			_datas.actualCenter = localMembersZone.center
			_datas.actualAzimuth = localMembersZone.azimuth
			
			local idVolume, searchFct, paramsVolume = ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctEvalSearchVolume](_datas)
					
			local volume = {
				id = idVolume,
				params = paramsVolume,
			}

			world.searchObjects(ATMEConst.findClasses[_datas.type].category, volume, ATMEFct[searchFct], _datas)
		else
			ATMEConst.findClasses[_datas.type].search_all_fct(_datas)
		end
		
		_datas.nbUnitsGroup = nil
	else -- Recherche AIRBASES
		_datas.objects = {}
		for _, _airbase in pairs(ATMEDatas.airbasesById) do
			local findOk = ATMEFct.testFindCategory(_airbase, _datas.findCategories)
			if findOk == true then
				if _datas.coalitionName == nil or _airbase:getCoalitionName() == _datas.coalitionName then
					local point = _airbase:getPosition()
					
					if ATMEFct.isObjectClass(_zone, ATME.C_ZonesMgr) == true then
						local localMembersZonesMgr = ATMEDatas.localMembers[_zone:getIndex()]

						for _, _zoneItem in ipairs(localMembersZonesMgr.zonesList) do
							local localMembersZone = ATMEDatas.localMembers[_zoneItem.zone:getIndex()]

							_datas.zone = _zoneItem.zone
							_datas.actualCenter = ATMEFct.getActualCenter(localMembersZone.center, _zoneItem.linkedUnitDatas)
							_datas.actualAzimuth = ATMEFct.getActualAzimuth(localMembersZone.azimuth, _zoneItem.linkedUnitDatas)
								
							if ATMEFct[ATMEConst.indZoneConfig[_datas.zone:getClassName()].fctIsInside](point, _datas) then
								if _datas.objects[_airbase:getName()] == nil then
									local record
									local status
									if _datas.callbacks[ATMEDatas.currentModuleName] ~= nil then
										local datas = {
											zone = _zone,
											object = _airbase,
										}
										status, record = xpcall(function() return _datas.callbacks[ATMEDatas.currentModuleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Find callback') end)
									end
									
									if record == nil or (type(record) == "boolean" and record == true) then
										table.insert(_datas.objectsIndex, _airbase:getName())
										_datas.objects[_airbase:getName()] = _airbase
									end
								end
							end
						end
					elseif _zone ~= nil then
						local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]

						_datas.zone = _zone
						_datas.actualCenter = localMembersZone.center
						_datas.actualAzimuth = localMembersZone.azimuth
						
						if ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](point, _datas) then
							
							if _datas.objects[_airbase:getName()] == nil then
								local record
								local status
								if _datas.callbacks[ATMEDatas.currentModuleName] ~= nil then
									local datas = {
										zone = _zone,
										object = _airbase,
									}
									status, record = xpcall(function() return _datas.callbacks[ATMEDatas.currentModuleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Find callback') end)
								end
								
								if record == nil or (type(record) == "boolean" and record == true) then
									table.insert(_datas.objectsIndex, _airbase:getName())
									_datas.objects[_airbase:getName()] = _airbase
								end
							end
						end
					else
						if _datas.objects[_airbase:getName()] == nil then
							local record
							local status
							if _datas.callbacks[ATMEDatas.currentModuleName] ~= nil then
								local datas = {
									object = _airbase,
								}
								status, record = xpcall(function() return _datas.callbacks[ATMEDatas.currentModuleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Find callback') end)
							end
							
							if record == nil or (type(record) == "boolean" and record == true) then
								table.insert(_datas.objectsIndex, _airbase:getName())
								_datas.objects[_airbase:getName()] = _airbase
							end
						end
					end
				end
			end
		end
	end
	
	for _name, _ in pairs(_datas.allGroupIn) do -- allGroupIn n'est rempli que pour GROUPS
		local record
		local status
		if _datas.callbacks[ATMEDatas.currentModuleName] ~= nil then
			local datas = {
				zone = _zone,
				object = ATMEDatas.groups[_name],
				unitsName = _datas.allGroupIn[_name],
			}
			status, record = xpcall(function() return _datas.callbacks[ATMEDatas.currentModuleName](_datas.findContext, datas) end, function() ATMEFct.errUserHandler('Find callback') end)
		end
		
		if type(record) ~= "boolean" or record == true then
			_datas.objects[_name] = _group
			table.insert(_datas.objectsIndex, _name)
		end
		
		_datas.allGroupIn[_name] = nil
	end
end

ATMEDatas.C_FindContext_MT = {
	className = "C_FindContext",
	mt = {},
	
	keys = {
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["addCallback"] = true,
		["removeCallback"] = true,
		["findForAll"] = true,
		["findForCoalition"] = true,
	},
	
	add = function(self, _type, _findCategories)
		local item = {
			findCategories = _findCategories,
			type = _type,
			userPrivateDatas = {},
			callbacks = {}
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].callbacks = nil
	end,
}

ATME.C_FindContext = {
	getClassName = function(self)
		return ATMEDatas.C_FindContext_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_FindContext_MT.parentClass
	end,
	
	createNamed = function(_name, _findType, _findCategories)
		local fctDefinition = "Function ATME.C_FindContext.createNamed(name, findType, [findCategories])"
		if type(_name) == "string" and type(_findType) == "string" and 
				(_findCategories == nil or type(_findCategories) == "string" or type(_findCategories) == "table") then
				
			local findType = string.upper(_findType) 
			if ATMEConst.findClasses[findType] ~= nil then
				local findCategories
				
				if _findCategories == nil then
					if findType == "AIRBASES" then
						findCategories = {"AIRBASE"}
					else
						findCategories = {"ALL"}
					end
				else
					if ATMEConst.findClasses[findType].findSubCategories == true then
						local err
						
						err, findCategories = ATMEFct.extractDatasFromStringOrTable(_findCategories, true)
						
						if #findCategories > 1 and findType == "AIRBASES" then
							ATMEFct.errorOutputCore(nil, fctDefinition, "Limited to one category for 'AIRBASES' type", false)
							return
						end
						
						if err == true then
							ATMEFct.errorOutputCore(nil, fctDefinition, findCategories, false)
							return
						end
					else
						findCategories = {"ALL"}
					end
				end


				local localMembersModule = ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName]
				local err1, newFindContext = ATMEFct.C_FindContext_createNamed(localMembersModule.findContexts, _name, findType, findCategories)				

				if err1 == true then
					ATMEFct.errorOutputCore(nil, fctDefinition, newFindContext, false)
					return
				end
				
				return newFindContext
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "findType must be 'GROUPS', 'UNITS', 'AI_UNITS', 'PLAYERS', 'WEAPONS', 'STATIC_OBJECTS', 'CARGOS', 'SCENERY_OBJECTS' or 'AIRBASES'", false)
			end
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif type(_findType) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "findType type must be string", false)
			elseif type(_findCategories) ~= "string" and type(_findCategories) ~= "table" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "findCategories type must be string or table", false)
			end
		end
	end,
	
	getByName = function(_name)
		if type(_name) == "string" then
			return ATMEDatas.modulesLocalMembers[ATMEDatas.currentModuleName].findContexts[_name]
		else
			ATMEFct.errorOutputCore(nil, "ATME.C_FindContext.getByName(name)", "name type must be string", false)
		end
	end,
}

ATMEFct.C_FindContext_create = function(_name, _findContexts, _type, _findCategories)
	local newFindContext = {}

	setmetatable(newFindContext, ATMEDatas.C_FindContext_MT.mt)
	
	local name
	local tagsName
	
	if _name ~= nil and _findContexts ~= nil then
		name = _name
		_findContexts[name] = newFindContext
		tagsName = ATMEFct.decompString(_name)
	end

	local index = ATMEDatas.C_FindContext_MT:add(_type, _findCategories)
	local localMembers = ATMEDatas.localMembers[index]
	
	function newFindContext.getName(self)
		return name
	end
	
	function newFindContext.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newFindContext) .. "tagGetByID(position)"
		
		if type(_position) == "number" then
			local pos = math.floor(_position)
			if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
				return tagsName[pos]
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
		end
	end
	
	function newFindContext.tagGetCount(self)
		if tagsName ~= nil then
			return #tagsName
		else
			return 0
		end
	end
	
	function newFindContext.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newFindContext) .. "tagCompare(tagsToCompare)"
		
		if type(_tagsToCompare) == "string" then
			if tagsName ~= nil then
				local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
				
				if ok ~= nil then
					return ok
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
				end
			else
				return false
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
		end
	end
	
	function newFindContext.getIndex(self)
		return index
	end
	
	function newFindContext.getUserPrivateDatas(self)
		return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
	end
	
	function newFindContext.addCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newFindContext) .. "addCallback(fct)"
		if type(_fct) == "function" then
			localMembers.callbacks[ATMEDatas.currentModuleName] = _fct
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
		end
	end
	
	function newFindContext.removeCallback(self)
		localMembers.callbacks[ATMEDatas.currentModuleName] = nil
	end
	
	function newFindContext.findForAll(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newFindContext) .. "findForAll([zone])"
		
		if _zone == nil or type(_zone) == "string" or ATMEFct.isObjectClass(_zone, ATME.C_ZonesMgr) == true or ATMEFct.isObjectClass(_zone, ATME.C_Zone) == true then
			local datas = {
				objects = {},
				objectsIndex = {},
				nbUnitsGroup = {},
				allGroupIn = {},
				findCategories = localMembers.findCategories,
				callbacks = localMembers.callbacks,
				findContext = self,
				type = localMembers.type,
			}
			
			local zone = _zone
			if type(zone) == "string" then
				local zoneDCS = ATMEFct.getDCSZone(zone)
		
				if zoneDCS ~= nil then
					zone = ATMEFct.C_Circle_create(nil, nil, zoneDCS.point, zoneDCS.radius)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _zone .. " unknown", false)
					return
				end
			end
			
			ATMEFct.C_FindContext_find(localMembers.type, zone, datas)
			
			return datas.objectsIndex, datas.objects
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be nil, string, ATME.C_ZonesMgr or ATME.C_Zone", true)
		end
	end
	
	function newFindContext.findForCoalition(self, _coalitionName, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newFindContext) .. "findForCoalition(coalitionName, [zone])"
		
		if type(_coalitionName) == "string" and 
				(_zone == nil or type(_zone) == "string" or ATMEFct.isObjectClass(_zone, ATME.C_ZonesMgr) == true or ATMEFct.isObjectClass(_zone, ATME.C_Zone) == true) then
			
			local coalitionName = string.upper(_coalitionName)
			if ATMEConst.coalitionIds[coalitionName] ~= nil then
				local datas = {
					objects = {},
					objectsIndex = {},
					nbUnitsGroup = {},
					allGroupIn = {},
					findCategories = localMembers.findCategories,
					coalitionName = coalitionName,
					callbacks = localMembers.callbacks,
					findContext = self,
					type = localMembers.type,
				}
				
				if localMembers.type == "SCENERY_OBJECTS" then
					datas.coalitionName = nil
				end
				
				local zone = _zone
				if type(zone) == "string" then
					local zoneDCS = ATMEFct.getDCSZone(zone)
			
					if zoneDCS ~= nil then
						zone = ATMEFct.C_Circle_create(nil, nil, zoneDCS.point, zoneDCS.radius)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _zone .. " unknown", false)
						return
					end
				end
				
				ATMEFct.C_FindContext_find(localMembers.type, zone, datas)
				
				return datas.objectsIndex, datas.objects
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'RED', 'BLUE' or 'NEUTRAL'", false)
			end
		else
			if type(_coalitionName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", true)
			elseif _zone ~= nil and type(_zone) ~= "string" and ATMEFct.isObjectClass(_zone, ATME.C_ZonesMgr) == false and ATMEFct.isObjectClass(_zone, ATME.C_Zone) == false then
				ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be nil, string, ATME.C_ZonesMgr or ATME.C_Zone", false)
			end
		end
	end
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
	end
	
	return newFindContext
end

setmetatable(ATME.C_FindContext,ATMEDatas.C_FindContext_MT)

ATMEDatas.C_FindContext_MT.__call = function(t, _findType, _findCategories)
	local fctDefinition = "Constructor ATME.C_FindContext(findType, [findCategories])"
	
	if type(_findType) == "string" and
			(_findCategories == nil or type(_findCategories) == "string" or type(_findCategories) == "table") then 
		local findType = string.upper(_findType)
		if ATMEConst.findClasses[findType] ~= nil then
			local findCategories
			if _findCategories == nil then
				if findType == "AIRBASES" then
					findCategories = {"AIRBASE"}
				else
					findCategories = {"ALL"}
				end
			else
				if ATMEConst.findClasses[findType].findSubCategories == true then
					local err
					
					err, findCategories = ATMEFct.extractDatasFromStringOrTable(_findCategories, true)
					
					if #findCategories > 1 and findType == "AIRBASES" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "Limited to one category for 'AIRBASES' type", false)
						return
					end
					
					if err == true then
						ATMEFct.errorOutputCore(nil, fctDefinition, findCategories, false)
						return
					end
				else
					findCategories = {"ALL"}
				end
			end
			
			return ATMEFct.C_FindContext_create(nil, nil, findType, findCategories)
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "findType must be 'GROUPS', 'UNITS', 'AI_UNITS', 'PLAYERS', 'WEAPONS', 'STATIC_OBJECTS', 'CARGOS', 'SCENERY_OBJECTS' or 'AIRBASES'", false)
		end
	else
		if type(_findType) ~= "string" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "findType type must be string", false)
		elseif type(_findCategories) ~= "string" and type(_findCategories) ~= "table" then
			ATMEFct.errorOutputCore(nil, fctDefinition, "findCategories type must be string or table", false)
		end
	end
end

ATMEDatas.C_FindContext_MT.mt.__metatable = "Private"
ATMEDatas.C_FindContext_MT.mt.__index = ATME.C_FindContext
ATMEDatas.C_FindContext_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_FindContext_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_FindContext : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Group : gère les groupes d'unités (AI ou joueur) au format ATME. Pas de contructeur accessible
-- pour les modules externes.
-- Les instances de cette classe sont intégralement gérées par le module ATME_Core. La création et la destruction
-- dépendent du contexte de début de mission et des événements qui se produisent dans DCS en cours de mission.
-- Les modules externes accèdent à cette gestion indirectement par les handlers définis dans ceux-ci et les méthodes
-- accessibles de cette classe. Ceci permet de garantir une cohérence sur les listes internes ATME.
-- ATME crée des instances de cette classe à partir de la création d'instance d'unité joueur ou AI.
-- Chaque unité AI ou joueur dispose d'une variable représentant le groupe ATME qui lui est associé.
-- En cas de destruction (suppression) d'une unité (AI ou joueur), le groupe ne sera détruit qu'après la destruction
-- de la dernière unité (AI ou joueur) du groupe.
----------------------------------------------------------------------------------------------------------

ATMEFct.C_Group_CreateTrackingWaypoints = function(_localMembers, _flagInit)
		
	for _i = 1, #_localMembers.waypoints do
		_localMembers.trackingDatas[_i] = {
			pos = {
				x = _localMembers.waypoints[_i].x,
				y = _localMembers.waypoints[_i].alt,
				z = _localMembers.waypoints[_i].y,
			},
		}
	end

	-- Du WP0 au WP1, départ segment non renseigné car lastWaypoint à nil (cf fonction stackTimerBeforeL2Fcts)
	if _flagInit == false then
		_localMembers.waypointId = 2
	end

	for _i = 2, #_localMembers.waypoints - 1 do
		if ATMEConst.canFly[_localMembers.category] == false then
			_localMembers.trackingDatas[_i].fctRadius = ATMEConst.fctRadius[_localMembers.category]
			_localMembers.trackingDatas[_i].action = _localMembers.waypoints[_i].action
		else
			local line1 = ATMEFct.C_Line2D_create(nil, nil, _localMembers.trackingDatas[_i - 1].pos, _localMembers.trackingDatas[_i].pos)
			local line2 = ATMEFct.C_Line2D_create(nil, nil, _localMembers.trackingDatas[_i].pos, _localMembers.trackingDatas[_i + 1].pos)
			
			local a,b,c = line1:get()

			local bis1, bis2 = ATMEFct.C_Line2D_special(line1, line2, false, false)

			-- Choix de la bonne bissectrice, les points doivent être de part et d'autre
			
			local a1, b1, c1 = bis1:get()
			
			local t1 = a1 * _localMembers.trackingDatas[_i - 1].pos.z + b1 * _localMembers.trackingDatas[_i - 1].pos.x + c1
			local t2 = a1 * _localMembers.trackingDatas[_i + 1].pos.z + b1 * _localMembers.trackingDatas[_i + 1].pos.x + c1

			local distance = ATMEConst.fctRadius[_localMembers.category](_localMembers.waypoints[_i].action) * 2

			if t1 * t2 <= 0 then
				_localMembers.trackingDatas[_i].segment = ATMEFct.C_Line2D_getSegment2D(bis1:getOrigin(), bis1:getA(), bis1:getB(), distance)
			else
				_localMembers.trackingDatas[_i].segment = ATMEFct.C_Line2D_getSegment2D(bis2:getOrigin(), bis2:getA(), bis2:getB(), distance)
			end
		end
	end

	if #_localMembers.waypoints > 1 then
		-- Arrivée prise un cercle centré sur le WPn

		_localMembers.trackingDatas[#_localMembers.waypoints].fctRadius = ATMEConst.fctRadius[_localMembers.category]
		_localMembers.trackingDatas[#_localMembers.waypoints].action = _localMembers.waypoints[#_localMembers.waypoints].action
	end
end

ATMETimers.C_Group_timerTrackingWaypoint = function(_group)
	local localMembersGroup = ATMEDatas.localMembers[_group:getIndex()]
	--ATMEFct.outputCore(nil, "Group " .. localMembersGroup.name .." -> Entrée C_Group_timerTrackingWaypoint " .. localMembersGroup.waypointId, 1)
	
	if localMembersGroup.disembarking == false then
		local unitsDCS = localMembersGroup.objectDCS:getUnits()
		local unit = ATMEFct.getFirstUnit(_group)
		local localMembersUnit = ATMEDatas.localMembers[unit:getIndex()]
		local pos = unit:getPosition()
		local intersect = unit:getDCSUnit():isActive()
		-- TODO local speed = ATMEFct.C_Vector3D_create(nil, nil, localMembersUnit.objectDCS:getVelocity()):getHModule()
		--local refresh
		--if localMembersGroup.category == "AIRPLANE" then
		--	refresh = 100/speed
		--else
		--	refresh = 10/speed
		--end
		
		--ATMEFct.outputCore(nil, "Speed first unit " .. speed .. "-> Refresh à " .. refresh, 1)
		if intersect == true then
			if localMembersGroup.trackingDatas[localMembersGroup.waypointId].segment ~= nil then
				local localMembersSegment = ATMEDatas.localMembers[localMembersGroup.trackingDatas[localMembersGroup.waypointId].segment:getIndex()]
				intersect, _ = ATMEFct.C_Segment2D_intersect(localMembersSegment.datas, localMembersGroup.lastWaypoint, pos)
			elseif localMembersGroup.trackingDatas[localMembersGroup.waypointId].fctRadius ~= nil then
				local datas = {
					group = _group,
					ok = false,
				}

				local volume = {
					id = world.VolumeType.SPHERE,
					params = {
						point = localMembersGroup.trackingDatas[localMembersGroup.waypointId].pos,
						radius = localMembersGroup.trackingDatas[localMembersGroup.waypointId].fctRadius(localMembersGroup.trackingDatas[localMembersGroup.waypointId].action),
					},
				}

				world.searchObjects(Object.Category.UNIT, volume, ATMEFct.searchNearestWaypoint, datas)
				
				intersect = datas.ok
			end

			localMembersGroup.lastWaypoint = pos
		end

		if intersect == true then
			ATMEFct.outputCore(nil, "Group " .. localMembersGroup.name .." => Route courante - Passage WP " .. localMembersGroup.waypointId - 1 .. "/" .. #localMembersGroup.waypoints - 1, 2)
			if localMembersGroup.waypointId > localMembersGroup.offsetWaypoint then
				-- Passage au waypoint
				-- Préparation de l'event core et WP suivante	
				local id = localMembersGroup.waypointId
				local datas = {
					waypointID = id + localMembersGroup.deltaWaypointId - 1,
					lastWaypointID = localMembersGroup.nbInitialWaypoints - 1,
					point = {
						x = localMembersGroup.trackingDatas[id].pos.x,
						y = localMembersGroup.trackingDatas[id].pos.y,
						z = localMembersGroup.trackingDatas[id].pos.z,
					},
					name = localMembersGroup.waypoints[id].name,
				}
				
				localMembersGroup.waypointId = localMembersGroup.waypointId + 1
				
				if localMembersGroup.offsetWaypoint > 0 then
					ATMEFct.outputCore(nil, "              -> Groupe sortant d'un véhicule, Route originale " .. datas.waypointID .. "/" .. localMembersGroup.nbInitialWaypoints - 1, 2)
				end
				
				for _moduleName, _callbacks in pairs(localMembersGroup.trackingCallbacks) do
					ATMEDatas.currentModuleName = _moduleName
					
					if _callbacks[id] ~= nil then
						ATMEFct.outputCore(nil, "Appel callback Tracking waypoint du module " .. _moduleName .. " --> Group ".. localMembersGroup.name, 2)
						xpcall(function() _callbacks[id](_group, datas) end, function() ATMEFct.errUserHandler('Tracking callback') end)
					end
				end
			else
				localMembersGroup.waypointId = localMembersGroup.waypointId + 1
				ATMEFct.outputCore(nil, "Group " .. localMembersGroup.name .." => Pas de traitement sur ce WP", 2)
			end
			
			--ATMEFct.outputCore(nil, "Group " .. localMembersGroup.name .." -> Fin C_Group_timerTrackingWaypoint -> franchissement", 1)
			if localMembersGroup.waypointId > #localMembersGroup.trackingDatas then
				localMembersGroup.endOfRoute = true
				localMembersGroup.idTimer = nil
			else
				return timer.getTime() + 2
			end
		else
			return timer.getTime() + 2
		end
	else
		ATMEFct.outputCore(nil, "Group ".. localMembersGroup.name .. " débarquement en cours", 1)
		return timer.getTime() + 2
	end
end

ATMEFct.C_Group_disable = function(_group)
	local localMembers = ATMEDatas.localMembers[_group:getIndex()]
	localMembers.isEnabled = false
			
	local units = _group:getUnits()
	for _, _unit in pairs(units) do
		ATMEFct.C_Player_C_AIUnit_disable(_unit)
	end
end

ATMEDatas.C_Group_MT = {
	className = "C_Group",
	parentClass = ATME.C_DynamicObject,
	
	mtModules = {},
	mt = {},
	
	keys = {
		["disable"] = true,
		["getFirstUnit"] = true,
		["getNextUnit"] = true,
		["getPreviousUnit"] = true,
		["getLastUnit"] = true,
		["getUnitByName"] = true,
		["getUnits"] = true,
		["getDCSID"] = true,
		["getNbUnits"] = true,
		["hasAttributes"] = true,
		["getCoalitionName"] = true,
		["getCountryName"] = true,
		["isOnlyInfantry"] = true,
		["isOnlyAI"] = true,
		["isInZones"] = true,
		["isInDCSZone"] = true,
		["isInZone2D"] = true,
		["isInZone3D"] = true,
		["getCategoryName"] = true,
		["isReadyToBoard"] = true,
		["setReadyToBoard"] = true,
		["resetReadyToBoard"]= true,
		["isBoarding"] = true,
		["move"] = true,
		["stop"] = true,
		["freeze"] = true,
		["getDCSGroup"] = true,
		["hasAutoRespawn"] = true,
		["setAutoRespawn"] = true,
		["resetAutoRespawn"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getIndex"] = true,
		["activate"] = true,
		["isActivated"] = true,
		["isAlwaysVisible"] = true,
		["setRoute"] = true,
		["setTask"] = true,
		["resetTask"] = true,
		["pushTask"] = true,
		["popTask"] = true,
		["getMinDistance"] = true,
		["getMaxDistance"] = true, 
		["getMinHDistance"] = true,
		["getMaxHDistance"] = true,
		["getPosition"] = true,
		["modules"] = true,
		["setPatrol"] = true,
		["resetPatrol"] = true,
		["start"] = true,
		["isControlled"] = true,
		["canFly"] = true,
		["getUserPrivateDatas"] = true,
		["setInvisible"] = true,
		["setImmortal"] = true,
		["setOption"] = true,
		["addTrackingCallback"] = true,
		["removeTrackingCallback"] = true,
		["addNewRouteCallback"] = true,
		["removeNewRouteCallback"] = true,
		["addReadyToBoardCallback"] = true,
		["removeReadyToBoardCallback"] = true,
	},
	
	add = function(self, _name, _group, _datas)
		local delay = ATMEConst.basicRouteDelay
		if _datas.uncontrolled == true then
			delay = ATMEConst.uncontrolledRouteDelay
		end

		local item = {
			name = _name,
			objectDCS = _group,
			objectDCSId = _group:getID(),
			trackingDatas = {},							-- Données pour le tracking waypoints
			trackingCallbacks = {}, 					-- Liste des callbacks pour les waypoints trackés									
			maxWaypoints = 0, 							-- Nombre de waypoints tenant compte de l'ajustement si le dernier WP
														-- n'est pas de type "Land" pour les aéronefs.
			waypointId = 1,
			offsetWaypoint = 0,							-- Offset de 2 si débarquement, 0 sinon
			nbInitialWaypoints = 0,						-- Indique le nombre de WP initiaux (utile pour le load/unload de troupes)
			deltaWaypointId = 0, 						-- Indique l'écart d'ID entre la première route et la route actuelle pour troupes load/unload
			endOfRoute = false,							-- Route en cours ... Utilisé pour le suivi des waypoint sur modif de route
			disembarking = false,						-- True si débarquement en cours, permet de synchroniser les événements
			groupId = _group:getID(),
			hasPlayers = false,
			onBoard = false,
			autoRespawn = {
				activated = false,
				time = 0,
				random = false,
				minTime = 0,
				maxTime = 0,
				restore = false,
			},
			coalition = ATMEConst.coalitionNames[_group:getCoalition()],
			category = ATMEConst.categoryNames[_group:getCategory()],
			routeDelay = delay,
			routeUncontrolled = _datas.uncontrolled,
			onlyInfantry = true,
			boardingStarted = false,
			readyToBoardSet = false,
			isEnabled = true,
			nbUnitsToEnable = 0,
			isStopped = false,
			isFreezed = false,
			engineStopped = false,						-- Utilisé si le dernier waypoint d'un aéronef n'est pas franchi. Sera fait au premier arret moteur
			ATMEUnits = {},
			ATMEUnitNames = ATME.C_IndexList(),
			userPrivateDatas = {},						-- Point d'entrée des données privées : un par module
			newRouteCallback = {},
			tagsName = ATMEFct.decompString(_name),
			spawnDatas = _datas.spawnDatas,
			readyToBoardCallback = {},
			zoneCleanedCallback = {},
			destroyed = false,							-- Utile lors du delete (car isExist est déjà false)
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].ATMEUnits = nil
		ATMEDatas.localMembers[_id].ATMEUnitNames = nil
		ATMEDatas.localMembers[_id].trackingDatas = nil
		ATMEDatas.localMembers[_id].trackingCallbacks = nil
		ATMEDatas.localMembers[_id].newRouteCallback = nil
		ATMEDatas.localMembers[_id].readyToBoardCallback = nil
		ATMEDatas.localMembers[_id].zoneCleanedCallback = nil
	end,
}

ATME.C_Group = {
	getClassName = function(self)
		return ATMEDatas.C_Group_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Group_MT.parentClass
	end,
	
	getByName = function(_groupName)
		if type(_groupName) == "string" then
			return ATMEDatas.groups[_groupName]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Group.getByName(name)", "name type must be string", false)
		end
	end,
	
	findReadyToBoardGroupsForAll = function()
		local groupsReady = {}
		local nbGroups = 0
		for _, _group in pairs(ATMEDatas.readyToBoardGroups) do
			groupsReady[_group:getName()] = _group
			nbGroups = nbGroups + 1
		end
		
		return groupsReady, nbGroups
	end,
	
	findReadyToBoardGroupsForCoalition = function(_coalitionName)
		local fctDefinition = "Function ATME.C_Group.findReadyToBoardGroupsForCoalition(coalitionName)"
		
		if type(_coalitionName) == "string" then
			local coalitionName = string.upper(_coalitionName)
			if ATMEConst.coalitionIdsBlueRed[coalitionName] ~= nil then
				local groupsReady = {}
				local nbGroups = 0
				for _, _group in pairs(ATMEDatas.readyToBoardGroups) do
					if _group:getCoalitionName() == coalitionName then
						groupsReady[_group:getName()] = _group
						nbGroups = nbGroups + 1
					end
				end
				
				return groupsReady, nbGroups
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'BLUE' or 'RED'", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", false)
		end
	end,
}


ATMEFct.C_Group_addUnit = function(_ATMEGroup, _unit)
	local unitName = _unit:getName()

	local localMembers = ATMEDatas.localMembers[_ATMEGroup:getIndex()]
	
	if ATMEFct.isObjectClass(_unit, ATME.C_Player) == true then
		localMembers.hasPlayers = true
	end

	local localMembersUnit = ATMEDatas.localMembers[_unit:getIndex()]
	localMembersUnit.num = localMembers.ATMEUnitNames:add(unitName)

	localMembers.ATMEUnits[unitName] = _unit

	ATMEFct.outputCore(nil,"Function C_Group_addUnit : ajout unité " .. unitName, 2)
	
	if ATMEFct.isObjectClass(_unit, ATME.C_Player) == true or 
		(ATMEFct.isObjectClass(_unit, ATME.C_AIUnit) == true and localMembers.onlyInfantry == true and _unit:isInfantry() == false) then
		localMembers.onlyInfantry = false
	end
end


ATMEFct.C_Group_deleteUnit = function(_ATMEGroup, _unitName)
	local localMembers = ATMEDatas.localMembers[_ATMEGroup:getIndex()]
	local localMembersUnit = ATMEDatas.localMembers[localMembers.ATMEUnits[_unitName]:getIndex()]
	
	localMembers.ATMEUnitNames:remove(localMembersUnit.num)
	localMembers.ATMEUnits[_unitName] = nil
	
	if localMembers.ATMEUnitNames:getCount() == 0 then
		ATMEFct.C_Group_delete(_ATMEGroup)
	end
end

ATMEFct.C_Group_evalTypeDistance = {
	["MIN_H"] = function(_pA, _pB, _value)
		local d2 = (_pA.x - _pB.x) * (_pA.x - _pB.x) + (_pA.z - _pB.z) * (_pA.z - _pB.z)
		if _value == nil or _value > d2 then
			return d2
		else
			return _value
		end
	end,
	["MAX_H"] = function(_pA, _pB, _value)
		local d2 = (_pA.x - _pB.x) * (_pA.x - _pB.x) + (_pA.z - _pB.z) * (_pA.z - _pB.z)
		if _value == nil or _value < d2 then
			return d2
		else
			return _value
		end
	end,
	["MIN"] = function(_pA, _pB, _value)
		local d2 = (_pA.x - _pB.x) * (_pA.x - _pB.x) + (_pA.z - _pB.z) * (_pA.z - _pB.z) + (_pA.y - _pB.y) * (_pA.y - _pB.y)
		if _value == nil or _value > d2 then
			return d2
		else
			return _value
		end
	end,
	["MAX"] = function(_pA, _pB, _value)
		local d2 = (_pA.x - _pB.x) * (_pA.x - _pB.x) + (_pA.z - _pB.z) * (_pA.z - _pB.z) + (_pA.y - _pB.y) * (_pA.y - _pB.y)
		if _value == nil or _value < d2 then
			return d2
		else
			return _value
		end
	end,
}

ATMEFct.C_Group_evalDistances = function(_ATMEGroup, _point, _typeTest)
	local value
	local units = _ATMEGroup:getUnits()
	
	for _, _unit in pairs(units) do
		value = ATMEFct.C_Group_evalTypeDistance[_typeTest](_point, _unit:getPosition(), value)
	end
	
	return math.sqrt(value)
end


ATMEFct.C_Group_changeReadyToBoard = function(_group, _value, _nextWaypoint)
	local localMembers = ATMEDatas.localMembers[_group:getIndex()]
	
	if _value == true and ATMEDatas.readyToBoardGroups[_group:getName()] == nil then
		if type(_nextWaypoint) == "number" then
			localMembers.nextWaypoint = _nextWaypoint - localMembers.deltaWaypointId
		else
			localMembers.nextWaypoint = _nextWaypoint
		end
		
		ATMEDatas.readyToBoardGroups[_group:getName()] = _group
		_group:resetPatrol()
		_group:stop()

		for _moduleName, _callback in pairs(localMembers.readyToBoardCallback) do
			ATMEDatas.currentModuleName = _moduleName
				
			ATMEFct.outputCore(nil,"Appel de readyToBoardCallback du groupe " .. _group:getName() .. " pour le module " .. _moduleName, 2)
			xpcall(function() _callback(_group) end, function() ATMEFct.errUserHandler('readyToBoardCallback du groupe ' .. _group:getName()) end)
		end
	elseif _value == false and _group:isBoarding() == false and ATMEDatas.readyToBoardGroups[_group:getName()] ~= nil then
		_group:move()
		ATMEDatas.readyToBoardGroups[_group:getName()] = nil
		localMembers.nextWaypoint = nil
		localMembers.pickupZone = nil
	end	
end

ATMEFct.C_Group_finalizeBoarding = function(_localMembersTransportUnit, _groupToBoard, _transportUnit)
	_localMembersTransportUnit.nbUnitsOnBoard = _localMembersTransportUnit.nbUnitsOnBoard + _groupToBoard.nbUnitsOnBoard
	
	local index = 0
	for _i, _groupItem in ipairs(_localMembersTransportUnit.groupsBoarding) do
		if _groupToBoard.activeGroupName == _groupItem.groupName then
			index = _i
		end
	end
	
	if index > 0 then
		table.remove(_localMembersTransportUnit.groupsBoarding, index)
	end
	
	-- TODO -> Ajuster les valeurs passées
	local idGroupOnBoard = ATMEFct.C_Player_C_AIUnit_addOnBoardGroup(_transportUnit,  
																	{
																		name = _groupToBoard.activeGroupName,
																		spawnDatas = _groupToBoard.spawnDatas,
																		unitsOnBoard = _groupToBoard.unitsOnBoard,
																		nbUnitsOnBoard = _groupToBoard.nbUnitsOnBoard,
																		waypoints = _groupToBoard.waypoints,
																		waypointId = _groupToBoard.waypointId,
																		nbInitialWaypoints = _groupToBoard.nbInitialWaypoints,
																		trackingCallbacks = _groupToBoard.trackingCallbacks,
																		userPrivateDatas = _groupToBoard.userPrivateDatas,
																		autoRespawn = _groupToBoard.autoRespawn,
																		nextWaypoint = _groupToBoard.nextWaypoint,
																		deltaWaypointId = _groupToBoard.deltaWaypointId,
																		nbDeadUnits = 0,
																		newRouteCallback = _groupToBoard.newRouteCallback,
																		readyToBoardCallback = _groupToBoard.readyToBoardCallback,
																		zoneCleanedCallback = _groupToBoard.zoneCleanedCallback,
																	})

	-- Evénement de fin d'embarquement
	-- On ne stocke que le nom. La vérification de l'existance de l'unité ne se fera que lors
	-- de l'activation de l'événement. 
	-- ATTENTION : Ici name est le nom du group qui va être disable... Donc pas de réactivation
	-- à l'activation de l'événement.
	
	
	local event = {
		type = "TRANSPORT_END_OF_BOARDING",
		datas = {
			unitName = _localMembersTransportUnit.name,
			index = idGroupOnBoard,
			onBoardGroupName = _groupToBoard.activeGroupName,
			nbOnBoardUnits = _groupToBoard.unitsOnBoard,
			nbDeadUnits = #_groupToBoard.unitsLost,
		},
	}
	
	ATMEFct.C_Event_add(event)
end

ATMETimers.C_Group_startLoad_timer = function(_groupToBoard)
	if ATMEDatas.allAIUnitsAndPlayers[_groupToBoard.transportUnitName] ~= nil then
		local transportUnit = ATMEDatas.allAIUnitsAndPlayers[_groupToBoard.transportUnitName]

		local transportUnitPosition = transportUnit:getPosition()
		local toBoardGroup = ATME.C_Group.getByName(_groupToBoard.activeGroupName)
		local localMembersTransportUnit = ATMEDatas.localMembers[transportUnit:getIndex()]
		
		if toBoardGroup ~= nil then
			local localMembersGroup = ATMEDatas.localMembers[toBoardGroup:getIndex()]
			if _groupToBoard.boardingStopped == false then
				if (transportUnit:isGroundVehicle() == true or (transportUnit:canFly() == true and transportUnit:inAir() == true)) and
								transportUnit:getSpeed() > 3 then
								
					_groupToBoard.delayBeforeStopped = _groupToBoard.delayBeforeStopped + 1
				
					if _groupToBoard.delayBeforeStopped == 40 then -- 40 cycles soit 2s
						_groupToBoard.boardingStopped = true
						_groupToBoard.currentUnitName = localMembersGroup.ATMEUnitNames:getFirst()
					end
				else
					_groupToBoard.delayBeforeStopped = 0
				end
			end
			
			-- Vitesse des infanteries à 5 m/s soit un temps de parcours = dist/5
			-- Marge de sécurité à 10s en plus. Au delà, tout le groupe est chargé directement
			-- Ceci permet de contourner le problème d'une unité capricieuse qui louperait le transport et le rayon de chargement.
				
			if _groupToBoard.boardingStopped == false then
				if ATMEDatas.AIUnits[_groupToBoard.currentUnitName] ~= nil then
					local unit = ATMEDatas.AIUnits[_groupToBoard.currentUnitName]
					local disabled = false
					
					if _groupToBoard.boardingStart == false then
						local item = {
							groupName = _groupToBoard.activeGroupName,
							nbUnits = localMembersGroup.ATMEUnitNames:getCount(),
						}
						
						table.insert(localMembersTransportUnit.groupsBoarding, item)
						_groupToBoard.boardingStart = true
					end
					
					if timer.getTime() < _groupToBoard.maxDelay then
						local v = ATMEFct.C_Vector3D_create(nil, nil, unit:getPosition(), transportUnitPosition)
						local ok = ATMEFct.whichSide(_groupToBoard.A, _groupToBoard.B, unit:getPosition()) == _groupToBoard.sideOfWP2

						if v:getHModule() < ATMEConst.personnelCarrierAttributes[transportUnit:getTypeName()].minDistance or ok == true then
							-- tant que l'unité de transport est au sol, les unités à transporter
							-- sont désactivées à l'approche de l'unité de transport. Distance paramétrée dans ATMEConst.personnelCarrierAttributes
							_groupToBoard.unitsOnBoard[unit:getName()] = true
							_groupToBoard.nbUnitsOnBoard = _groupToBoard.nbUnitsOnBoard + 1
							
							ATMEFct.C_Player_C_AIUnit_disable(unit)
							disabled = true
						end
					else
						ATMEFct.C_Player_C_AIUnit_disable(unit)
						disabled = true
					end
					
					if localMembersGroup.ATMEUnitNames == nil then
						localMembersGroup.onBoard = true
						ATMEFct.C_Group_finalizeBoarding(localMembersTransportUnit, _groupToBoard, transportUnit)
						
						_groupToBoard.unitsOnBoard = nil
						_groupToBoard.unitsLost = nil
					else
						if disabled == false then
							_groupToBoard.currentUnitName = localMembersGroup.ATMEUnitNames:getNext()
							
							if _groupToBoard.currentUnitName == nil then
								_groupToBoard.currentUnitName = localMembersGroup.ATMEUnitNames:getFirst()
							end
						else
							_groupToBoard.currentUnitName = localMembersGroup.ATMEUnitNames:getFirst()
						end
						
						return timer.getTime() + 0.1
					end	
				end
			else
				if _groupToBoard.nbUnitsOnBoard > 0 then
					local disabled = false
					
					-- Si l'unité de transport décolle ou se déplace plus vite que 3m/s, tout le groupe est disable
					-- Embarquement en urgence ! des pertes sont possibles
				
					if ATMEDatas.AIUnits[_groupToBoard.currentUnitName] ~= nil then
						local unit = ATMEDatas.AIUnits[_groupToBoard.currentUnitName]
						if math.random(1,10) <= 7 - _groupToBoard.alea then
							_groupToBoard.unitsOnBoard[unit:getName()] = true
							_groupToBoard.nbUnitsOnBoard = _groupToBoard.nbUnitsOnBoard + 1

						else 
							table.insert(_groupToBoard.unitsLost, unit:getName())
						end
						ATMEFct.C_Player_C_AIUnit_disable(unit)
						disabled = true
						
						if _groupToBoard.alea < 7 then
							_groupToBoard.alea = _groupToBoard.alea + 1
						end
					end
					
					if localMembersGroup.ATMEUnitNames == nil then
						localMembersGroup.onBoard = true
						ATMEFct.C_Group_finalizeBoarding( localMembersTransportUnit, _groupToBoard, transportUnit)
						
						_groupToBoard.unitsOnBoard = nil
						_groupToBoard.unitsLost = nil
					else
						if disabled == false then
							_groupToBoard.currentUnitName = localMembersGroup.ATMEUnitNames:getNext()
							
							if _groupToBoard.currentUnitName == nil then
								_groupToBoard.currentUnitName = localMembersGroup.ATMEUnitNames:getFirst()
							end
						else
							_groupToBoard.currentUnitName = localMembersGroup.ATMEUnitNames:getFirst()
						end
						
						return timer.getTime() + 0.1
					end	
				else
					toBoardGroup:stop()
					localMembersGroup.boardingStarted = false
				end
			end
		else
			if _groupToBoard.nbUnitsOnBoard > 0 then
				ATMEFct.C_Group_finalizeBoarding(localMembersTransportUnit, _groupToBoard, transportUnit)
			end
			
			_groupToBoard.unitsOnBoard = nil
			_groupToBoard.unitsLost = nil
		end
	else
		-- L'unité de transport est inexistante (détruit ou changement de slot)
		-- Tout est perdu. Seules restent les unités du groupe extérieures à l'unité de transport.
		
		_groupToBoard.unitsOnBoard = nil
		_groupToBoard.unitsLost = nil
		local toBoardGroup = ATME.C_Group.getByName(_groupToBoard.activeGroupName)
		if toBoardGroup ~= nil then
			local localMembers = ATMEDatas.localMembers[toBoardGroup:getIndex()]
			localMembers.boardingStarted = false
		end
	end
end

ATMEFct.getFirstUnit = function(_group)
	local localMembers = ATMEDatas.localMembers[_group:getIndex()]
	return ATMEDatas.allAIUnitsAndPlayers[localMembers.objectDCS:getUnits()[1]:getName()]
end

ATMEFct.C_Group_create = function(_ATMEUnit, _datas, _newObject)
	local uncontrolled = _datas.uncontrolled
	local alwaysVisible = _datas.visible
	
	local newGroup = {}
	setmetatable(newGroup, ATMEDatas.C_Group_MT.mt)

	local index = ATMEDatas.C_Group_MT:add(_ATMEUnit:getDCSUnit():getGroup():getName(), _ATMEUnit:getDCSUnit():getGroup(), _datas)
	local localMembers = ATMEDatas.localMembers[index]
	
	
	function newGroup.getIndex(self)
		return index
	end
		
	function newGroup.getName(self)
		if localMembers.destroyed == false then
			return localMembers.name
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.getDCSID(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCSId
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "tagGetByID(position)"
		
		if localMembers.destroyed == false then
			if type(_position) == "number" then
				local pos = math.floor(_position)
				if localMembers.tagsName ~= nil and pos >= 1 and pos <= #localMembers.tagsName then
					return localMembers.tagsName[pos]
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.tagGetCount(self)
		if localMembers.destroyed == false then
			if tagsName ~= nil then
				return #tagsName
			else
				return 0
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "tagCompare(tagsToCompare)"
		
		if localMembers.destroyed == false then
			if type(_tagsToCompare) == "string" then
				if localMembers.tagsName ~= nil then
					local ok = ATMEFct.compareTags(localMembers.tagsName, _tagsToCompare)
					
					if ok ~= nil then
						return ok
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
					end
				else
					return false
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.isAlwaysVisible(self)
		if localMembers.destroyed == false then
			return alwaysVisible
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.getCategoryName(self)
		if localMembers.destroyed == false then
			return localMembers.category
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.getCoalitionName(self)
		if localMembers.destroyed == false then
			return localMembers.coalition
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.getDCSGroup(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.activate(self)
		if localMembers.destroyed == false then
			localMembers.objectDCS:activate()
			localMembers.justActivated = true
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.isActivated(self)
		-- ATTENTION fonction DCS sur l'unité et non sur le groupe
		-- Risque de BUG DCS car pas homogène
		
		if localMembers.destroyed == false then
			local unit = ATMEFct.getFirstUnit(self)

			if unit ~= nil then	
				return unit:getDCSUnit():isActive()
			else
				return false
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.hasAttributes(self, _attributes)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "hasAttributes(attributes)"
		
		if localMembers.destroyed == false then
			if type(_attributes) == "string" or type(_attributes) == "table" then
				local unit = ATMEFct.getFirstUnit(self)
				local unitsATME = {}
				
				while unit ~= nil do
					local err, datas = ATMEFct.extractDatasFromStringOrTable(_attributes, false)
		
					if err == false then
						if ATMEFct.hasAttributes(unit, datas) == true then
							table.insert(unitsATME, unit)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, datas, false)
						return
					end
					
					unit = self:getNextUnit()
				end
				
				return unitsATME
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "attributes type must be string or table", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.getUserPrivateDatas(self)
		if localMembers.destroyed == false then
			return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.setInvisible(self, _invisible)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "setInvisible(invisible)"
		
		if localMembers.destroyed == false then
			if type(_invisible) == "boolean" then
				local controller = self:getDCSGroup():getController()
				local command = {
					id = "SetInvisible", 
					params = { 
						value = _invisible, 
					} 
				}
				
				controller:setCommand(command)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "invisible type must be boolean", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.setImmortal(self, _immortal)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "setImmortal(immortal)"
		
		if localMembers.destroyed == false then
			if type(_immortal) == "boolean" then
				local controller = self:getDCSGroup():getController()
				local command = {
					id = "SetImmortal", 
					params = { 
						value = _immortal, 
					} 
				}
				
				controller:setCommand(command)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "immortal type must be boolean", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.setRoute(self, _route)
		local localMembersRoute
		
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "setRoute(route)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_route, ATME.C_Route) == true then
				localMembersRoute = ATMEDatas.localMembers[_route:getIndex()]

				if localMembers.spawnDatas.type == localMembersRoute.type then
					if uncontrolled == false then
						self:resetPatrol()
						
						local err, waypoints = ATMEFct.C_Route_createWaypoints(_route)

						if err == false then
							localMembers.lastWaypoint =  ATMEFct.C_Route_addCurrentPositionWaypoint(waypoints, self)
							localMembers.waypoints = waypoints
							localMembers.initialWaypoints = waypoints
							localMembers.waypointId = 1
							
							ATMEFct.C_Group_CreateTrackingWaypoints(localMembers, false)

							for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
								localMembers.trackingCallbacks[_localModuleMembers.name] = {}
							end
		
							localMembers.spawnDatas.datas.route.points = waypoints
							localMembers.offsetWaypoint = 0
							localMembers.deltaWaypointId = 0

							ATMEFct.C_CloningContext_setMission(self, localMembers.waypoints, false)
							
							-- Ajustement du nombre de waypoints si pour un aéronef le dernier waypoint
							-- n'est pas un waypoint d'atterrissage.
							
							localMembers.maxWaypoints = #localMembers.waypoints
							
							if ATMEConst.canFly[localMembers.category] == true then
								if 	localMembers.waypoints[#localMembers.waypoints].type ~= "Land" then
									localMembers.maxWaypoints = localMembers.maxWaypoints + 1
								end
							end
							
							localMembers.nbInitialWaypoints = localMembers.maxWaypoints
							
							ATMEDatas.groupsWithNewRouteSet[localMembers.name] = true
						else
							ATMEFct.setError("SPAWN_NO_WAYPOINTS")
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, localMembers.name .. " is an uncontrolled group.", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "route must have same type as " .. localMembers.name, false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "route type must be ATME.C_Route", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.setTask(self, _task, _zone2D)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "setTask(task)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_task, ATME.C_Task) == true then
				
				local ok, taskDatas = ATMEFct.C_Task_getTaskDatas(_task, localMembers.spawnDatas.type, false)
				
				if ok == true then
				
					local controller = self:getDCSGroup():getController()

					controller:setTask(taskDatas)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, taskDatas .. " for group " .. localMembers.name, false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "task type must be ATME.C_Task", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.resetTask(self)
		if localMembers.destroyed == false then
			local controller = self:getDCSGroup():getController()

			controller:resetTask()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.pushTask(self, _task)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "pushTask(task)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_task, ATME.C_Task) == true then
				local ok, taskDatas = ATMEFct.C_Task_getTaskDatas(_task, localMembers.spawnDatas.type, false)
				
				if ok == true then
					local controller = self:getDCSGroup():getController()

					controller:pushTask(taskDatas)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, taskDatas .. " for group " .. localMembers.name, false)
				end	
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "task type must be ATME.C_Task", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.popTask(self)
		if localMembers.destroyed == false then
			local controller = self:getDCSGroup():getController()

			controller:popTask()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.canFly(self)
		if localMembers.destroyed == false then
			return ATMEConst.canFly[localMembers.category]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.getMinDistance(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "getMinDistance(reference)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isReference(_reference) == true then
				return ATMEFct.C_Group_evalDistances(self, ATMEFct.getPointFromReference(_reference), "MIN")
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.getMaxDistance(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "getMaxDistance(reference)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isReference(_reference) == true then
				return ATMEFct.C_Group_evalDistances(self, ATMEFct.getPointFromReference(_reference), "MAX")
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.getMinHDistance(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "getMinHDistance(reference)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isReference(_reference) == true then
				return ATMEFct.C_Group_evalDistances(self, ATMEFct.getPointFromReference(_reference), "MIN_H")
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
		
	function newGroup.getMaxHDistance(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "getMaxHDistance(reference)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isReference(_reference) == true then
				return ATMEFct.C_Group_evalDistances(self, ATMEFct.getPointFromReference(_reference), "MAX_H")
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.getPosition(self)
		if localMembers.destroyed == false then
			local unit = ATMEFct.getFirstUnit(self)	
			
			if unit ~= nil then
				return unit:getPosition()
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.disable(self)
		if localMembers.destroyed == false then
			if localMembers.hasPlayers == false then
				ATMEDatas.groupsToDisable[localMembers.name] = self
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.addNewRouteCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "addNewRouteCallback(fct)"
		
		if localMembers.destroyed == false then
			if type(_fct) == "function" then
				localMembers.newRouteCallback[ATMEDatas.currentModuleName] = _fct
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.removeNewRouteCallback(self)
		if localMembers.destroyed == false then
			localMembers.newRouteCallback[ATMEDatas.currentModuleName] = nil
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.disable(self)
		if localMembers.destroyed == false then
			if localMembers.hasPlayers == false then
				ATMEDatas.groupsToDisable[localMembers.name] = self
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.addReadyToBoardCallback(self, _fct)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "addReadyToBoardCallback(fct)"
		
		if localMembers.destroyed == false then
			if type(_fct) == "function" then
				localMembers.readyToBoardCallback[ATMEDatas.currentModuleName] = _fct
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.removeReadyToBoardCallback(self)
		if localMembers.destroyed == false then
			localMembers.readyToBoardCallback[ATMEDatas.currentModuleName] = nil
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	
	function newGroup.addTrackingCallback(self, _fct, _waypointId)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "addTrackingCallback(fct, [waypointId])"
		
		if localMembers.destroyed == false then
			local index
							
			if type(_fct) == "function" then
				if type(_waypointId) == "number" then
					index = math.floor(_waypointId) + 1
					
					if index > 0 and index <= localMembers.nbInitialWaypoints then
						localMembers.trackingCallbacks[ATMEDatas.currentModuleName][index - localMembers.deltaWaypointId] = _fct
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "waypointId must be >=0 and <= max waypoints of current route", false)
					end
				elseif type(_waypointId) == "string" then
					index = #localMembers.waypoints
					local ok = false
					while index > localMembers.offsetWaypoint and ok == false do
						if localMembers.waypoints[index].name ~= nil and localMembers.waypoints[index].name == _waypointId then
							ok = true
						else
							index = index - 1
						end
					end
					
					if ok == true then
						localMembers.trackingCallbacks[ATMEDatas.currentModuleName][index] = _fct
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "waypointId name '" .. _waypointId .. "' unknown in current route", false)
					end
				elseif _waypointId == nil then
					for _i = localMembers.offsetWaypoint + 1, #localMembers.waypoints do
						localMembers.trackingCallbacks[ATMEDatas.currentModuleName][_i] = _fct
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "waypointId type must be nil, number or string", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.removeTrackingCallback(self, _waypointId)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "removeTrackingCallback([waypointId])"
		
		if localMembers.destroyed == false then
			local index
			
			if type(_waypointId) == "number" then
				index = math.floor(_waypointId) + 1
				
				if index > 0 and index <= localMembers.nbInitialWaypoints then
					localMembers.trackingCallbacks[ATMEDatas.currentModuleName][index + localMembers.offsetWaypoint] = nil
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "waypointId must be >=0 and <= max waypoints of current route", false)
				end
			elseif type(_waypointId) == "string" then
				index = #localMembers.waypoints
				local ok = false
				while index > localMembers.offsetWaypoint and ok == false do
					if localMembers.waypoints[index].name ~= nil and localMembers.waypoints[index].name == _waypointId then
						ok = true
					else
						index = index - 1
					end
				end
				
				if ok == true then
					localMembers.trackingCallbacks[ATMEDatas.currentModuleName][index] = nil
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "waypointId name '" .. _waypointId .. "' unknown in current route", false)
				end
			elseif _waypointId == nil then
				for _i = localMembers.offsetWaypoint + 1, #localMembers.waypoints do
					localMembers.trackingCallbacks[ATMEDatas.currentModuleName][_i] = nil
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "waypointId type must be nil, number or string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.setAutoRespawn(self, _rules)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "setAutoRespawn(rules)"
		
		if localMembers.destroyed == false then
			if (type(_rules) == "string" or type(_rules) == "number") then
				if localMembers.hasPlayers == false then
					local restore = true
					if _restore ~= nil then
						restore = _restore
					end
					
					if _rules == nil then
						localMembers.autoRespawn = {
							activated = true,
							time = 0,
							random = false,
							minTime = 0,
							maxTime = 0,
							restore = restore,
						}
					elseif type(_rules) == "number" then
						if _rules >= 0 then
							local value = math.floor(_rules)
							localMembers.autoRespawn = {
								activated = true,
								time = value,
								random = false,
								minTime = 0,
								maxTime = 0,
								restore = restore,
							}
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "rules must be >= 0 when number", false)
						end
					else
						local tokens = ATMEFct.analyzeString(_rules)
						
						if #tokens >= 1 and #tokens <= 3 then
							local ok = false
							local minRandom = 0
							local maxRandom = 0
							local value = 0
							local randomFlag = false
							
							if #tokens >= 2 and tokens[1].type == "TOKEN_TYPE_RANDOM" then
								randomFlag = true
								
								if tokens[2].type == "TOKEN_INTEGER" then
									if tokens[2].value >= 0 then
										maxRandom = tokens[2].value
										
										if #tokens == 3 then
											minRandom = tokens[2].value
											
											if tokens[3].type == "TOKEN_INTEGER" then
												if tokens[3].value >= 0 then
													maxRandom = tokens[3].value
													ok = true
												else
													ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #3 must be >= 0 in string format rules", false)
												end
											else
												ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #3 must be number or time value in string format rules", false)
											end
										else
											maxRandom = tokens[2].value
											ok = true
										end
									else
										ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be >= 0 in string format rules", false)
									end
								else
									ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be number or time value in string format rules", false)
								end
							else
								if tokens[1].type == "TOKEN_INTEGER" then
									if tokens[1].value >= 0 then
										value = tokens[1].value
										ok = true
									else
										ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #1 must be >= 0 in string format rules", false)
									end
								else
									ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #1 must be number or time value or 'RANDOM' in string format rules", false)
								end
							end
							
							if ok == true then
								if randomFlag == true then
									if maxRandom > minRandom then
										localMembers.autoRespawn = {
											activated = true,
											time = 0,
											random = true,
											minTime = minRandom,
											maxTime = maxRandom,
											restore = restore,
										}
									else
										ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be < parameter #3 in string format rules", false)
									end
								else
									localMembers.autoRespawn = {
										activated = true,
										time = value,
										random = false,
										minTime = 0,
										maxTime = 0,
										restore = restore,
									}
								end
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Bad string format rules", false)
						end
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Group " .. localMembers.name .. " has player units - Do not use setAutoRespawn", false)
				end
			else
				if type(_rules) ~= "string" and type(_rules) ~= "number" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "rules type must be string or number", true)
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.resetAutoRespawn(self)
		if localMembers.destroyed == false then
			localMembers.autoRespawn = {
				activated = false,
				time = 0,
				random = false,
				minTime = 0,
				maxTime = 0,
				restore = false,
			}
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.hasAutoRespawn(self)
		if localMembers.destroyed == false then
			return localMembers.autoRespawn.activated
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.setPatrol(self, _lastWaypoint, _restartWaypoint, _stopCondition)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "setPatrol(lastWaypoint, restartWaypoint, [stopCondition])"
		
		if localMembers.destroyed == false then
			if type(_lastWaypoint) == "number" or type(_lastWaypoint) == "string" or 
					type(_restartWaypoint) == "number" or type(_restartWaypoint) == "string" or
					_stopCondition == nil or type(_stopCondition) == "string" then
					
				if localMembers.hasPlayers == false then
					local lastWaypoint
					local restartWaypoint
					local randomValue = -1
					local maxTurnsValue = -1
					local maxTimeValue = -1
					local flag
					
					if type(_lastWaypoint) == "number" then
						lastWaypoint = _lastWaypoint
						if lastWaypoint + 1 >= #localMembers.waypoints then
							ATMEFct.errorOutputCore(nil, fctDefinition, "lastWaypoint bad number " .. _lastWaypoint .. " - Group " .. localMembers.name, false)
							return
						end
					else
						lastWaypoint = ATMEFct.C_CloningContext_getWaypointFromName(localMembers.waypoints, _lastWaypoint)
						if lastWaypoint == nil then
							ATMEFct.errorOutputCore(nil, fctDefinition, "Waypoint " .. _lastWaypoint .. " not found in current route - Group " .. localMembers.name, false)
							return
						end
					end
					
					if type(_restartWaypoint) == "number" then
						restartWaypoint = _restartWaypoint
						if restartWaypoint + 1 >= #localMembers.waypoints then
							ATMEFct.errorOutputCore(nil, fctDefinition, "restartWaypoint bad number " .. restartWaypoint .. " - Group " .. localMembers.name, false)
							return
						end
					else
						restartWaypoint = ATMEFct.C_CloningContext_getWaypointFromName(localMembers.waypoints, _restartWaypoint)
						if restartWaypoint == nil then
							ATMEFct.errorOutputCore(nil, fctDefinition, "Waypoint " .. _restartWaypoint .. " not found in current route - Group " .. localMembers.name, false)
							return
						end
					end
					
					if lastWaypoint <= restartWaypoint then
						ATMEFct.errorOutputCore(nil, fctDefinition, "lastWaypoint must be > at restartWaypoint - Group " .. localMembers.name, false)
						return
					end
					
					local stopConditionType = "NO_COND"
					if _stopCondition ~= nil then
						local tokens = ATMEFct.analyzeString(_stopCondition)
							
						if #tokens == 2 then
							if tokens[1].type == "TOKEN_TYPE_RANDOM" then
								if tokens[2].type == "TOKEN_INTEGER" and tokens[2].value > 0 and tokens[2].value < 1 then
									randomValue = tokens[2].value
									stopConditionType = "RANDOM_COND"
								else
									ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be integer > 0 and < 1 in string format stopCondition", false)
								end
							elseif tokens[1].type == "TOKEN_TYPE_TIME" then
								if tokens[2].type == "TOKEN_INTEGER" and tokens[2].value > 0 then
									maxTimeValue = tokens[2].value + timer.getTime()
									stopConditionType = "MAX_TIMER_COND"
								else
									ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be integer or time value in string format stopCondition", false)
								end
							elseif tokens[1].type == "TOKEN_TYPE_MAX_TURNS" then
								if tokens[2].type == "TOKEN_INTEGER" and tokens[2].value > 0 then
									maxTurnsValue = tokens[2].value
									stopConditionType = "MAX_TURN_COND"
								else
									ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be integer > 0 in string format stopCondition", false)
								end
							elseif tokens[1].type == "TOKEN_TYPE_FLAG" then
								stopConditionType = "FLAG_COND"
								if tokens[2].type == "TOKEN_INTEGER" and tokens[2].value > 0 then
									flag = tokens[2].value
								elseif tokens[2].type == "TOKEN_STRING" then
									flag = tokens[2].value
								else
									ATMEFct.errorOutputCore(nil, fctDefinition, "Parameter #2 must be integer or string > 0 in string format condition", false)
									return
								end
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Bad string format stopCondition", false)
						end
					end
					
					-- La patrouille ne peut se faire que du deuxième jusqu'à l'avant dernier waypoint de la route pour un aéronef
					if localMembers.waypoints[lastWaypoint + 1].type == "Turning Point" and 
								localMembers.waypoints[restartWaypoint + 1].type == "Turning Point"  then

						ATMEDatas.patrolsToUpdate[localMembers.name] = {
								type = stopConditionType,
								last = lastWaypoint + 1,
								restart = restartWaypoint + 1,
								nbTurns = 0,
								maxTurns = maxTurnsValue,
								maxTime = maxTimeValue,
								maxRandom = randomValue,
								flag = flag,
								isArmed = false,
						}
					else
						if localMembers.waypoints[lastWaypoint + 1].type ~= "Turning Point" then
							ATMEFct.errorOutputCore(nil, fctDefinition, "lastWaypoint must be 'Turning Point' type - Group " .. localMembers.name, false)
							return
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "restartWaypoint must be 'Turning Point' type - Group " .. localMembers.name, false)
							return
						end
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Group " .. localMembers.name .. " has player units - Do not use setPatrol", false)
				end
			else
				if type(_lastWaypoint) ~= "string" and type(_lastWaypoint) ~= "number" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "lastWaypoint type must be string or number", true)
				elseif type(_restartWaypoint) ~= "string" and type(_restartWaypoint) ~= "number" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "restartWaypoint type must be string or number", true)
				elseif _stopCondition ~= nil and type(_stopCondition) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "stopCondition type must be nil or string", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.resetPatrol(self)
		if localMembers.destroyed == false then
			if ATMEDatas.missionsDelayed[localMembers.name] == nil then
				ATMEDatas.patrolsToUpdate[localMembers.name] = nil
			else
				ATMEDatas.missionsDelayed[localMembers.name].patrolToUpdate = nil
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.start(self)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "start()"
		
		if localMembers.destroyed == false then
			if localMembers.hasPlayers == false then
				if ATMEConst.canFly[localMembers.category] == true then
					if uncontrolled == true then
						uncontrolled = false
						local controller = self:getDCSGroup():getController()

						controller:setCommand({ 
												id = 'Start', 
												params = {},
											})
					end
				else
					self:freeze(false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Group " .. localMembers.name .. " can't have player units", false)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.isControlled(self)
		if localMembers.destroyed == false then
			return uncontrolled == false
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.getFirstUnit(self)
		if localMembers.destroyed == false then
			local unitName = localMembers.ATMEUnitNames:getFirst()
			if unitName ~= nil then
				return ATMEDatas.allAIUnitsAndPlayers[unitName]
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newGroup.getNextUnit(self)
		if localMembers.destroyed == false then
			local unitName = localMembers.ATMEUnitNames:getNext()
			
			if unitName ~= nil then
				return ATMEDatas.allAIUnitsAndPlayers[unitName]
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newGroup.getPreviousUnit(self)
		if localMembers.destroyed == false then
			local unitName = localMembers.ATMEUnitNames:getPrevious()
			
			if unitName ~= nil then
				return ATMEDatas.allAIUnitsAndPlayers[unitName]
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newGroup.getLastUnit(self)
		if localMembers.destroyed == false then
			local unitName = localMembers.ATMEUnitNames:getLast()
			
			if unitName ~= nil then
				return ATMEDatas.allAIUnitsAndPlayers[unitName]
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end


	function newGroup.getUnitByName(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "getUnitByName(name)"
		
		if localMembers.destroyed == false then
			if type(_name) == "string" then
				return localMembers.ATMEUnits[_name]
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", true)
			end	
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newGroup.getUnits(self)
		if localMembers.destroyed == false then
			return localMembers.ATMEUnits
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newGroup.getNbUnits(self)
		if localMembers.destroyed == false then
			return localMembers.ATMEUnitNames:getCount()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newGroup.isOnlyInfantry(self)
		if localMembers.destroyed == false then
			return localMembers.onlyInfantry
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newGroup.isOnlyAI(self)
		if localMembers.destroyed == false then
			return localMembers.hasPlayers == false
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newGroup.isInZones(self, _zonesMgr, _allGroup)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "isInZones(zonesMgr, allGroup)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zonesMgr, ATME.C_ZonesMgr) == true and type(_allGroup) == "boolean" then
				local allUnitInZonesMgr = true
				local found = false
				local units = self:getUnits()
				for _, _unit in pairs(units) do
					local point = _unit:getPosition()
					local isInside, _ = ATMEFct.C_ZonesMgr_isInside(_zonesMgr, point)
					if  isInside == false then
						allUnitInZonesMgr = false
					else 
						found = true
					end
				end
				
				return(_allGroup == false and found == true) or (_allGroup == true and allUnitInZonesMgr == true)
			else
				if ATMEFct.isObjectClass(_zonesMgr, ATME.C_ZonesMgr) == false then
					ATMEFct.errorOutputCore(nil, fctDefinition, "zonesMgr type must be ATME.C_ZonesMgr", true)
				elseif type(_allGroup) ~= "boolean" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "allGroup type must be boolean", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end				
	
	function newGroup.isInDCSZone(self, _zoneName, _allGroup)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "isInDCSZone(zoneName, allGroup)"
		
		if localMembers.destroyed == false then
			if type(_zoneName) == "string" and type(_allGroup) == "boolean" then
				local zone = ATMEFct.getDCSZone(_zoneName)
				if zone ~= nil then
					if _allGroup == false then
						return ATMEFct.C_Group_evalDistances(self, zone.point, "MIN_H") <= zone.radius
					else
						return ATMEFct.C_Group_evalDistances(self, zone.point, "MAX_H") <= zone.radius
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _zoneName .. " unknown", false)
				end
			else
				if type(_zoneName) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "zoneName type must be string", true)
				elseif type(_allGroup) ~= "boolean" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "allGroup type must be boolean", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newGroup.isInZone2D(self, _zone, _allGroup)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "isInZone2D(zone, allGroup)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zone, ATME.C_Zone2D) == true and type(_allGroup) == "boolean" then
				local units = self:getUnits()
				local allGroupInZone = true
				local found = false
				
				for _, _unit in pairs(units) do
					local point = _unit:getPosition()
					local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
					
					local datas = {
						zone = _zone,
						actualCenter = localMembersZone.center,
						azimuth = localMembersZone.azimuth,
					}

					if ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](point, datas) == true then
						found = true
					end
				end
				
				return (_allGroup == true and allGroupInZone == true) or (_allGroup == false and found == true) 
			else
				if ATMEFct.isObjectClass(_zone, ATME.C_Zone2D) == false then
					ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Circle, ATME.C_PickupZone, ATME.C_Rectangle or ATME.C_Polygon", true)
				elseif type(_radius) ~= "number" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", false)
				elseif type(_allGroup) ~= "boolean" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "allGroup type must be boolean", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end	
			
	function newGroup.isInZone3D(self, _zone, _allGroup)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "isInZone3D(zone, allGroup)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zone, ATME.C_Zone3D) == true and type(_allGroup) == "boolean" then
				local units = self:getUnits()
				local allGroupInZone = true
				local found = false
				
				for _, _unit in pairs(units) do
					local point = _unit:getPosition()
					local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
					
					local datas = {
						zone = _zone,
						actualCenter = localMembersZone.center,
						azimuth = localMembersZone.azimuth,
					}
					
					if ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](point, datas) == true then
						found = true
					end
				end
				
				return (_allGroup == true and allGroupInZone == true) or (_allGroup == false and found == true) 
			else
				if ATMEFct.isObjectClass(_zone, ATME.C_Zone2D) == false then
					ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Sphere, ATME.C_Cylinder or ATME.C_Box", true)
				elseif type(_allGroup) ~= "boolean" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "allGroup type must be boolean", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end	

	function newGroup.isReadyToBoard(self)
		if localMembers.destroyed == false then
			return ATMEDatas.readyToBoardGroups[self:getName()] ~= nil
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newGroup.setReadyToBoard(self, ...)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "setReadyToBoard([nextWaypoint])\nsetReadyToBoard(pickupZone, [nextWaypoint])"
		
		if localMembers.destroyed == false then
			local args = {...}
			local argId = 1
			local pickupZone 
			
			if #args > 0 then
				if ATMEFct.isObjectClass(args[1], ATME.C_PickupZone) == true then
					pickupZone = args[1]
					
					if args[2] == nil or type(args[2]) == "number" or type(args[2]) == "string" then
						argId = 2
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "nextWaypoint type must be nil, number or string", true)
						return
					end
				elseif args[1] ~=  nil and type(args[1]) ~= "number" and type(args[1]) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "nextWaypoint type must be nil, number or string", true)
					return
				end
			end
			
			if self:isOnlyInfantry() == true then
				if localMembers.readyToBoardSet == false then
					if localMembers.boardingStarted == false then
						local nextWaypoint
						if type(args[argId]) == "number" then
							if args[argId] >= 0 then
								nextWaypoint = math.floor(args[argId])
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "nextWaypoint must be >= 0", false)
								return
							end
						elseif type(args[argId]) == "string" then
							local index = ATMEFct.C_CloningContext_getWaypointFromName(localMembers.waypoints, args[argId])
							if index == nil then
								ATMEFct.errorOutputCore(nil, fctDefinition, "nextWaypoint name '" .. args[argId] .. "' unknown in current route", false)
								return
							end
							
							nextWaypoint = args[argId]
						end
						
						if pickupZone == nil then
							ATMEFct.C_Group_changeReadyToBoard(self, true, nextWaypoint)
							localMembers.readyToBoardSet = true
						else
							local localMembersPickupZone = ATMEDatas.localMembers[pickupZone:getIndex()]
						
							if localMembersPickupZone.coalitionName == self:getCoalitionName() then
								local datas = {
									randomMax = 0.25 + math.random() / 2,
									name = localMembers.name,
									object = self,
									localMembersObject = localMembers,
									localMembersPickupZone = localMembersPickupZone,
									start = timer.getTime(),
									isInZone = false,
									nextWaypoint = nextWaypoint,
									nbTries = 0,
								}
								
								localMembers.pickupZone = pickupZone
								
								localMembers.idTimerPickupZone = timer.scheduleFunction(ATMETimers.C_PickupZone_readyToBoardTimer, datas, timer.getTime() + 1)
								localMembers.readyToBoardSet = true
							else
								ATMEFct.setError("BAD_COALITION")
							end
						end
					else
						ATMEFct.setError("BOARDING_STARTED")
					end
				else
					ATMEFct.setError("ALREADY_SET_READYTOBOARD")
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, localMembers.name .. " must be infantry group", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.resetReadyToBoard(self)
		if localMembers.destroyed == false then
			if localMembers.pickupZone ~= nil then
				if localMembers.idTimerPickupZone ~= nil then
					timer.removeFunction(localMember.idTimerPickupZone)
					localMembers.idTimerPickupZone = nil
				end
				
				local localMembersPickupZone = ATMEDatas.localMembers[localMembers.pickupZone:getIndex()]
				if ATMEDatas.readyToBoardGroups[self:getName()] ~= nil then
					localMembersPickupZone.nbReadyToBoard = localMembersPickupZone.nbReadyToBoard - 1
				end
				
				localMembers.pickupZone = nil
			end
			
			localMembers.readyToBoardSet = false
			ATMEFct.C_Group_changeReadyToBoard(self, false, nil)
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newGroup.isBoarding(self)
		if localMembers.destroyed == false then
			return localMembers.boardingStarted
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.move(self)
		if localMembers.destroyed == false then
			if ATMEConst.canFly[localMembers.category] == false then
				if localMembers.isStopped == true then
					trigger.action.groupContinueMoving(localMembers.objectDCS)
					localMembers.isStopped = false
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "only ground groups can be moved", false)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newGroup.stop(self)
		if localMembers.destroyed == false then
			if ATMEConst.canFly[localMembers.category] == false then
				if localMembers.isStopped == false then
					trigger.action.groupStopMoving(localMembers.objectDCS)
					localMembers.isStopped = true
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "only ground groups can be stopped", false)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newGroup.freeze(self, _freezeOn)
		if localMembers.destroyed == false then
			if ATMEConst.canFly[localMembers.category] == false then
				local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "freeze(freezeOn)"
				
				if type(_freezeOn) == "boolean" then
					if _freezeOn == true then
						trigger.action.setGroupAIOff(localMembers.objectDCS)
						localMembers.isFreezed = true
					else
						trigger.action.setGroupAIOn(localMembers.objectDCS)
						localMembers.isFreezed = false
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "freezeOn type must be boolean", true)
				end	
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "only ground groups can be freezed", false)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newGroup.setOption(self, _option, _value)
		local fctDefinition = ATMEFct.getErrorPrefix(newGroup) .. "setOption(option, value)"

		if localMembers.destroyed == false then
			if type(_option) == "string" and _value ~= nil then
				local err, option, optionValue = ATMEFct.C_setOption(localMembers.spawnDatas.type, _option, _value) 

				if err == false then
					local controller = self:getDCSGroup():getController()
					controller:setOption(ATMEConst.optionAI[localMembers.spawnDatas.type][option].id, optionValue)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, option, false)
				end
			else
				if type(_option) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "option type must be string", true)
				elseif _value == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "value type can't be nil", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	ATMEFct.C_Group_addUnit(newGroup, _ATMEUnit)

	if ATMEDatas.spawnedGroupsWaypoints[localMembers.name] == nil then  -- Groupe non spawné
		-- Les waypoints de la route initiale sont recopiés
		-- pour avoir la route courante.
		localMembers.waypoints = ATMEFct.C_CloningContext_copyAllWaypoints(localMembers.spawnDatas.datas.route.points)
		localMembers.initialWaypoints = localMembers.waypoints
		ATMEFct.C_CloningContext_setMission(newGroup, localMembers.waypoints, false)
	else -- Groupe spawné, pas de setMission car inclus dans addgroup.
		localMembers.waypoints = ATMEDatas.spawnedGroupsWaypoints[localMembers.name].waypoints
		localMembers.initialWaypoints = ATMEDatas.spawnedGroupsWaypoints[localMembers.name].initialWaypoints
		ATMEDatas.spawnedGroupsWaypoints[localMembers.name] = nil
	end
	
	-- Ajustement du nombre de waypoints si pour un aéronef le dernier waypoint
	-- n'est pas un waypoint d'atterrissage.
	localMembers.maxWaypoints = #localMembers.waypoints
	localMembers.nbInitialWaypoints = localMembers.maxWaypoints

	if ATMEConst.canFly[localMembers.category] == true then
		if localMembers.maxWaypoints > 0 then
			if 	localMembers.waypoints[#localMembers.waypoints].type ~= "Land" then
				localMembers.maxWaypoints = localMembers.maxWaypoints + 1
			end
		end
	end
	
	newGroup.modules = {}
	setmetatable(newGroup.modules, ATMEDatas.C_Group_MT.mtModules)
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
		localMembers.trackingCallbacks[_localModuleMembers.name] = {}
		localMembers.readyToBoardCallback[_localModuleMembers.name] = {}
		localMembers.newRouteCallback[_localModuleMembers.name] = {}
		localMembers.zoneCleanedCallback[_localModuleMembers.name] = {}
		if newGroup.modules[_localModuleMembers.name] == nil then
			newGroup.modules[_localModuleMembers.name] = {}
		end
	end
	
	ATMEDatas.newCreatedGroups[localMembers.name] = {
		timerOffset = 1,
		isSpawned = false,
		disembark = ATMEDatas.groupsOnBoard[localMembers.name] ~= nil,
		group = newGroup,
	}

	ATMEFct.outputCore(nil,"Prepare group = " .. localMembers.name .. " -> ok", 2)
end

ATMEFct.C_Group_delete = function(_ATMEGroup)
	local localMembers = ATMEDatas.localMembers[_ATMEGroup:getIndex()]

	local flag = "DESTROYED"
	if localMembers.isEnabled == false then
		flag = "DISABLED"
	end
	
	if localMembers.idTimer ~= nil then
		timer.removeFunction(localMembers.idTimer)
		localMembers.idTimer = nil
	end
	
	if localMembers.pickupZone ~= nil then
		if ATMEDatas.readyToBoardGroups[localMembers.name] ~= nil then
			local localMembersPickupZone = ATMEDatas.localMembers[localMembers.pickupZone:getIndex()]
			localMembersPickupZone.nbReadyToBoard = localMembersPickupZone.nbReadyToBoard - 1
		end
	end

	for _i = #ATMEDatas.modulesNames, 1, -1 do
		local localModuleMembers = ATMEDatas.modulesLocalMembers[ATMEDatas.modulesNames[_i]]
		if localMembers.onBoard == false then
			-- Les groupes chargés dans un autre véhicule ne génèrent pas d'appel 
			-- à onDeleteGroupHandler
			
			if localModuleMembers.handlers.onDeleteGroupHandler ~= nil then
				ATMEDatas.currentModuleName = localModuleMembers.name
				
				-- Appel du module
				ATMEFct.outputCore(nil,"Appel onDeleteGroupHandler du module " .. localModuleMembers.name, 2)
				xpcall(function() localModuleMembers.handlers.onDeleteGroupHandler(_ATMEGroup, flag) end, function() ATMEFct.errUserHandler('onDeleteGroupHandler') end)
			end
		end
	end
	
	-- Si le groupe est détruit normalement (pas à bord)
	if localMembers.onBoard == false then
		
		-- Si le groupe a un respawn automatique il est préparé
		if localMembers.autoRespawn.activated == true then
			local respawnTime = localMembers.autoRespawn.time
			if localMembers.autoRespawn.random == true then
				respawnTime = math.random(localMembers.autoRespawn.minTime, localMembers.autoRespawn.maxTime)
			end
			ATMEDatas.groupsToRespawn[localMembers.name]  = {
				time = respawnTime + timer.getTime(),
				spawnDatas = localMembers.spawnDatas,
			}
			ATMEDatas.respawnedGroups[localMembers.name] = true
		elseif localMembers.spawnDatas.idSpawnedList ~= nil then
			if localMembers.spawnDatas.idSpawnedList.groupSpawnedId ~= nil then
				table.insert(ATMEDatas.spawnedGroups[localMembers.spawnDatas.initialNames.groupName].freeIds, localMembers.spawnDatas.idSpawnedList.groupSpawnedId)
			end
			
			for _name, _id in pairs(localMembers.spawnDatas.idSpawnedList.units) do
				table.insert(ATMEDatas.spawnedUnits[_name].freeIds, _id)
			end
		end
	
		_ATMEGroup.modules = nil
	end
			
	ATMEDatas.C_Group_MT:remove(_ATMEGroup:getIndex())
	
	ATMEDatas.groups[localMembers.name] = nil
	ATMEDatas.readyToBoardGroups[localMembers.name] = nil
	ATMEDatas.newCreatedGroups[localMembers.name] = nil
	ATMEDatas.patrolsToUpdate[localMembers.name] = nil
	localMembers.destroyed = true
	
	ATMEFct.outputCore(nil,"Delete group = " .. localMembers.name, 2)
end

setmetatable(ATME.C_Group,ATMEDatas.C_Group_MT)

ATMEDatas.C_Group_MT.mt.__metatable = "Private"
ATMEDatas.C_Group_MT.mt.__index = ATME.C_Group
ATMEDatas.C_Group_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Group_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Group : Bad key " .. k, "table locked", false)
	end
end

ATMEDatas.C_Group_MT.mtModules.__metatable = "Private"
ATMEDatas.C_Group_MT.mtModules.__newindex = function(t, k, v)
	if ATMEDatas.modules[k] ~= nil then
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil,"ATME : ATME.C_Group.modules Bad key " .. k, "table locked", false)
	end
end


-- Fonctions communes à ATME.C_Player et ATME.C_AIUnit

ATMEFct.C_Player_C_AIUnit_unload_now = function(_transportUnit, _id, _currentOnBoardGroupDatas)
	local localMembers = ATMEDatas.localMembers[_transportUnit:getIndex()]
	local err = ATMEFct.C_CloningContext_spawnFromTransportUnit(_transportUnit, _currentOnBoardGroupDatas)

	if err == true then
		return false
	end
	
	-- ATTENTION : Lors du spawn, les unités sont crées après la fonction addGroup car unités au sol.
	-- Donc, à ce stade, les infos ATME pour le groupe et les unités sont à jour.
	-- La mise à jour de ATMEDatas.groupsOnBoard se fait ici.
	
	ATMEDatas.groupsOnBoard[_currentOnBoardGroupDatas.spawnDatas.datas.name] = nil
	
	local localSpawnGroupMembers = ATMEDatas.localMembers[ATMEDatas.newCreatedGroups[_currentOnBoardGroupDatas.spawnDatas.datas.name].group:getIndex()]
	localSpawnGroupMembers.autoRespawn = _currentOnBoardGroupDatas.autoRespawn
	localSpawnGroupMembers.disembarking = true
	localSpawnGroupMembers.offsetWaypoint = 2
	localSpawnGroupMembers.nbInitialWaypoints = _currentOnBoardGroupDatas.nbInitialWaypoints
	localSpawnGroupMembers.deltaWaypointId = _currentOnBoardGroupDatas.deltaWaypointId + _currentOnBoardGroupDatas.delta
	localSpawnGroupMembers.userPrivateDatas = _currentOnBoardGroupDatas.userPrivateDatas
	localSpawnGroupMembers.newRouteCallback = _currentOnBoardGroupDatas.newRouteCallback
	localSpawnGroupMembers.readyToBoardCallback = _currentOnBoardGroupDatas.readyToBoardCallback
	localSpawnGroupMembers.zoneCleanedCallback = _currentOnBoardGroupDatas.zoneCleanedCallback
	
	-- Permet de savoir si le groupe a déjà débarqué, pour rappel 2 WP sont ajoutés pour sortir du véhicule
	-- (position initiale et position pour s'en aller)

	localSpawnGroupMembers.trackingCallbacks = {}
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localSpawnGroupMembers.trackingCallbacks[_localModuleMembers.name] = {}
	end

	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		for _i = 3, #_currentOnBoardGroupDatas.waypoints do
			localSpawnGroupMembers.trackingCallbacks[_localModuleMembers.name][_i] = _currentOnBoardGroupDatas.trackingCallbacks[_localModuleMembers.name][_i + _currentOnBoardGroupDatas.delta]
		end
	end
	
	_currentOnBoardGroupDatas.trackingCallbacks = nil
	
	
	localMembers.nbUnitsOnBoard = localMembers.nbUnitsOnBoard - _currentOnBoardGroupDatas.nbUnitsOnBoard
	localMembers.onBoardGroupDatas:remove(_id)
	
	ATMEFct.outputCore(nil, "Fin de débarquement",1)
	ATMEDatas.newCreatedGroups[_currentOnBoardGroupDatas.spawnDatas.datas.name].event = {
		type = "TRANSPORT_END_OF_DISEMBARK",
		datas = {
			unitName = _transportUnit:getName(),
			index = _id,
			linkedGroupName = _currentOnBoardGroupDatas.spawnDatas.datas.name,
			nbUnitsOnBoard = localMembers.nbUnitsOnBoard,
			nbDeadUnits = _currentOnBoardGroupDatas.nbDeadUnits,
		},
	}
end

ATMETimers.C_Player_C_AIUnit_autoUnload = function(_localMembers)
	ATMEFct.outputCore(nil, "C_Player_C_AIUnit_autoUnload " .. _localMembers.name, 2)
	if _localMembers.objectDCS:isExist() == true then
		local unit = ATMEDatas.allAIUnitsAndPlayers[_localMembers.name]
		if _localMembers.autoUnloadDatas.all == false then
			local ok = false
			for _id, _item in pairs(_localMembers.autoUnloadDatas.list) do
				ok = true
				if _item.armed == false then
					if (unit:isGroundVehicle() == true and unit:getHSpeed() > 3) or 
							((unit:isHelicopter() or unit:isPlane()) and unit:inAir() == true) then
						_item.armed = true
						_item.wait = 10
					end
				end
				
				if _item.armed == true then
					if unit:getHSpeed() < 1 and (unit:isGroundVehicle() == true or 
							((unit:isHelicopter() or unit:isPlane()) and unit:inAir() == false)) then
					
						_item.wait = _item.wait - 1
						
						if _item.wait == 0 then
							local currentOnBoardGroupDatas = _localMembers.onBoardGroupDatas:get(_id)
							if currentOnBoardGroupDatas ~= nil then
								ATMEFct.C_Player_C_AIUnit_unload_now(unit, _id, currentOnBoardGroupDatas)
							end
							
							_localMembers.autoUnloadDatas.list[_id] = nil
						end
					else
						_item.wait = 10
					end
				end
			end

			if ok == false then
				_localMembers.autoUnloadDatas = nil
				ATMEFct.outputCore(nil, "C_Player_C_AIUnit_autoUnload stoppé" .. _localMembers.name, 2)
			else
				return timer.getTime() + 3
			end
		else
			if _localMembers.autoUnloadDatas.armed == false then
				if (unit:isGroundVehicle() == true and unit:getHSpeed() > 3) or 
						((unit:isHelicopter() or unit:isPlane()) and unit:inAir() == true) then
					_localMembers.autoUnloadDatas.armed = true
					_localMembers.autoUnloadDatas.wait = 10
				else
					return timer.getTime() + 13
				end
			end
			
			if _localMembers.autoUnloadDatas.armed == true then
				if unit:getHSpeed() < 1 and (unit:isGroundVehicle() == true or 
						((unit:isHelicopter() or unit:isPlane()) and unit:inAir() == false)) then
				
					_localMembers.autoUnloadDatas.wait = _localMembers.autoUnloadDatas.wait - 1
					
					if _localMembers.autoUnloadDatas.wait == 0 then
						for _id, _current in _localMembers.onBoardGroupDatas:pairs() do
							ATMEFct.C_Player_C_AIUnit_unload_now(unit, _id, _current)
						end
						
						_localMembers.autoUnloadDatas = nil
						ATMEFct.outputCore(nil, "C_Player_C_AIUnit_autoUnload stoppé" .. _localMembers.name, 2)
					else
						return timer.getTime() + 3
					end
				else
					_localMembers.autoUnloadDatas.wait = 10
					return timer.getTime() + 3
				end
			end
		end
	end
end

ATMEFct.C_Player_C_AIUnit_unload = function(_localMembers, _id, _whenStopped)
	local first = false
	if _localMembers.autoUnloadDatas == nil then
		_localMembers.autoUnloadDatas = {}
		first = true
	end
	
	if _id == nil then
		_localMembers.autoUnloadDatas.all = true
		_localMembers.autoUnloadDatas.armed = not _whenStopped
		_localMembers.autoUnloadDatas.wait = 1
		_localMembers.autoUnloadDatas.list = nil
		ATMEFct.outputCore(nil, "Je passe all",1)
	else
		ATMEFct.outputCore(nil, "Je passe id = " .. _id,1)
		_localMembers.autoUnloadDatas.all = false
		if _localMembers.autoUnloadDatas.list == nil then
			_localMembers.autoUnloadDatas.list = {}
		end
		if _localMembers.autoUnloadDatas.list[_id] == nil then
			_localMembers.autoUnloadDatas.list[_id] =  {
				armed = not _whenStopped,
				wait = 1,
			}
		end
	end
	
	if first == true then
		ATMEFct.outputCore(nil, "Start timer pour " .. _localMembers.name,1)
		timer.scheduleFunction(ATMETimers.C_Player_C_AIUnit_autoUnload, _localMembers, timer.getTime() + 1)
	end
end

ATMEFct.C_Player_C_AIUnit_whichSide = function(_transportUnit, _point)
	local rollAxis = ATMEFct.C_Line2D_create(nil, nil, _transportUnit:getRollAxisVector(), _transportUnit:getPosition())

	return ATMEFct.whichSideLine2D(rollAxis, _point)
end

ATMEFct.C_Player_C_AIUnit_load = function(_transportUnit, _groupToBoard, _radius)
	local maxUnits = ATMEConst.personnelCarrierAttributes[_transportUnit:getTypeName()].maxUnits
	local nbUnits
	local nbUnitsBoarding = 0

	local localMembers = ATMEDatas.localMembers[_transportUnit:getIndex()]

	nbUnits = localMembers.nbUnitsOnBoard
								
	for _, _groupItem in ipairs(localMembers.groupsBoarding) do
		nbUnitsBoarding = nbUnitsBoarding + _groupItem.nbUnits
	end
	
	if _groupToBoard:getNbUnits() + nbUnits + nbUnitsBoarding <= maxUnits then
		if (_transportUnit:isGroundVehicle() == true and _transportUnit:getHSpeed() < 2) or 
				((_transportUnit:isHelicopter() or _transportUnit:isPlane()) and _transportUnit:inAir() == false) then
				
			local transportUnitPosition = _transportUnit:getPosition()

			if ATMEFct.C_Group_evalDistances(_groupToBoard, transportUnitPosition, "MIN_H") <= _radius then	
				-- On mémorise le nb waypoints avant le changement de route pour monter
				-- dans le transport.
	
				local localMembersGroup = ATMEDatas.localMembers[_groupToBoard:getIndex()]
				local initialWaypoints = localMembersGroup.initialWaypoints
				
				local pos = _groupToBoard:getPosition()
				
				local v = ATMEFct.C_Vector3D_create(nil, nil, pos, transportUnitPosition)
				local v1 = v:toUnitVector()

				local waypoints = {}
				waypoints[1] = {
					action = "Off Road",
					x = transportUnitPosition.x,
					y = transportUnitPosition.z,
					type = "Turning Point",
					speed = 5,
				}
				
				waypoints[2] = {
					action = "Off Road",
					x = transportUnitPosition.x + v1:getX() * 500,
					y = transportUnitPosition.z + v1:getZ() * 500,
					type = "Turning Point",
					speed = 5,
				}
				
				local line = ATMEFct.C_Line2D_create(nil, nil, transportUnitPosition, pos)
				local axis = line:getPerpendicular(0)
				
				ATMEFct.C_CloningContext_setWaypoints(_groupToBoard, waypoints, false)
				
				localMembersGroup.boardingStarted = true
				
				if localMembersGroup.pickupZone ~= nil then
					local localMembersPickupZone = ATMEDatas.localMembers[localMembersGroup.pickupZone:getIndex()]
					localMembersPickupZone.nbReadyToBoard = localMembersPickupZone.nbReadyToBoard - 1
					localMembersGroup.pickupZone = nil
				end
					
				_groupToBoard:move()
							
				local pA = axis:getOrigin()
				local pB = ATMEFct.C_Line2D_getPointFromOrigin(axis:getA(), axis:getB(), pA, 10)
				
-- TODO --> Ajuster les valeurs passées
				local itemGroupToBoard = {
					activeGroupName = _groupToBoard:getName(),
					spawnDatas = localMembersGroup.spawnDatas,
					nbInitialWaypoints = localMembersGroup.nbInitialWaypoints,
					trackingCallbacks = localMembersGroup.trackingCallbacks,
					waypoints = localMembersGroup.waypoints,
					waypointId = localMembersGroup.waypointId,
					deltaWaypointId = localMembersGroup.deltaWaypointId,
					nextWaypoint = localMembersGroup.nextWaypoint,
					userPrivateDatas = localMembersGroup.userPrivateDatas,
					transportUnitName = _transportUnit:getName(),
					unitsOnBoard = {},
					nbUnitsOnBoard = 0,
					unitsLost = {},
					maxDelay = timer.getTime() + ATMEFct.C_Group_evalDistances(_groupToBoard, transportUnitPosition, "MAX_H") / 5 + 10,
					A = pA,
					B = pB,
					sideOfWP2 = ATMEFct.C_Player_C_AIUnit_whichSide(_transportUnit, waypoints[2]),
					autoRespawn = localMembersGroup.autoRespawn,
					newRouteCallback = localMembersGroup.newRouteCallback,
					readyToBoardCallback = localMembersGroup.readyToBoardCallback,
					zoneCleanedCallback = localMembersGroup.zoneCleanedCallback,
					currentUnitName = localMembersGroup.ATMEUnitNames:getFirst(),
					boardingStart = false,
					boardingStopped = false,
					delayBeforeStopped = 0,
					alea = 0,
				}

				timer.scheduleFunction(ATMETimers.C_Group_startLoad_timer, itemGroupToBoard, timer.getTime() + 1)
				
				return _groupToBoard
			else
				ATMEFct.setError("TOO_FAR")
				return
			end
		else
			if _transportUnit:isGroundVehicle() == true then
				ATMEFct.setError("BAD_SPEED")
				return
			else
				ATMEFct.setError("IN_AIR")
				return
			end
		end
	else
		ATMEFct.setError("TOO_MUCH_UNITS")
		return
	end
end

ATMEFct.C_Player_C_AIUnit_searchNearestReadyToBoardGroup = function(_unit, _datas)
	local ATMEUnit = ATMEDatas.AIUnits[_unit:getName()]
	local group = ATMEUnit:getGroup()
	
	if ATMEDatas.readyToBoardGroups[group:getName()] ~= nil and group:getCoalitionName() == _datas.transportUnit:getCoalitionName() and 
				group:isBoarding() == false and group:isActivated() == true and 
				ATMEConst.personnelCarrierAttributes[_datas.transportUnit:getTypeName()].maxUnits >= group:getNbUnits() then

		local posUnit = ATMEUnit:getPosition()
		local pos = _datas.transportUnit:getPosition()
		local d2 = (pos.x - posUnit.x) * (pos.x - posUnit.x) + (pos.z - posUnit.z) * (pos.z - posUnit.z)
		
		if _datas.minD2 == nil or _datas.minD2 > d2 then
			_datas.group = group
			_datas.minD2 = d2
		end
	end
end

ATMEFct.C_Player_C_AIUnit_getNearestReadyToBoardGroup = function(_transportUnit, _radius)
	local datas = {
		transportUnit = _transportUnit,
	}

	local volume = {
		id = world.VolumeType.SPHERE,
		params = {
			point = _transportUnit:getPosition(),
			radius = _radius,
		},
	}

	world.searchObjects(Object.Category.UNIT, volume, ATMEFct.C_Player_C_AIUnit_searchNearestReadyToBoardGroup, datas)
	return datas.group
end

ATMEFct.C_Player_C_AIUnit_addOnBoardGroup = function(_transportUnit, _datas)
	local localMembers = ATMEDatas.localMembers[_transportUnit:getIndex()]
	local id = localMembers.onBoardGroupDatas:add(_datas)

	ATMEDatas.groupsOnBoard[_datas.name] = {
		transportUnit = _transportUnit,
		spawnDatas = _datas.spawnDatas,
	}
	
	return id
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Player : gère les unités joueurs au format ATME. Pas de contructeur accessible
-- pour les modules externes.
-- Les instances de cette classe sont intégralement gérées par le module ATME_Core. La création et la destruction
-- dépendent du contexte de début de mission et des événements qui se produisent dans DCS en cours de mission.
-- Les modules externes accèdent à cette gestion indirectement par les handlers définis dans ceux-ci et les méthodes
-- accessibles de cette classe. Ceci permet de garantir une cohérence sur les listes internes ATME.
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Player_MT = {
	className = "C_Player",
	parentClass = ATME.C_Unit,
	
	mtModules = {},
	mt = {},
	
	keys = {
		["isEngineStarted"] = true,
		["getIndex"] = true,
		["getDCSUnit"] = true,
		["getDCSID"] = true,
		["getName"] = true,
		["isActivated"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getTypeName"] = true,
		["getNbUnitsOnBoard"] = true,
		["getOnBoardGroupNames"] = true,
		["getF10MenuRoot"] = true,
		["getPseudo"] = true,
		["getGroup"] = true,
		["getAttributes"] = true,
		["hasAttributes"] = true,
		["display"] = true,
		["disable"] = true,
		["mark"] = true,
		["setVoice"] = true,
		["getVoice"] = true,
		["soundPlay"] = true,
		["load"] = true,
		["unload"] = true,
		["soundStop"] = true,
		["getFuelRatio"] = true,
		["getLife"] = true,
		["getLifeRatio"] = true,
		["getCallsign"] = true,
		["getNearestReadyToBoardGroup"] = true,
		["getPosition"] = true,
		["getVelocityVector"] = true,
		["getSpeed"] = true,
		["getHSpeed"] = true,
		["getVSpeed"] = true,
		["getRollAxisVector"] = true,
		["getSpeedAzimuth"] = true,
		["getAzimuth"] = true, 
		["getCountryName"] = true,
		["getCoalitionName"] = true,
		["getCategoryName"] = true,
		["isHelicopter"] = true,
		["isPersonnelCarrier"] = true,
		["isPlane"] = true,
		["inAir"] = true,
		["isGroundVehicle"] = true,
		["getMSLAltitude"] = true,
		["getAGLAltitude"] = true, 
		["explode"] = true,
		["isInZones"] = true,
		["isInDCSZone"] = true,
		["isInZone2D"] = true,
		["isInZone3D"] = true,
		["isPilotDead"] = true,
		["isPilotEjected"] = true,
		["whichSide"] = true,
		["isInDirection"] = true,
		["modules"] = true,
		["canFly"] = true,
		["getUserPrivateDatas"] = true,
		["addEventCallback"] = true,
		["removeEventCallback"] = true,
	},
	
	add = function(self, _name, _unit)
		local item = {
			name = _name,
			objectDCS = _unit,
			objectDCSId = _unit:getID(),
			groupId = _unit:getGroup():getID(),
			engineOn = false,
			nbUnitsOnBoard = 0,
			groupsBoarding = {},
			onBoardGroupDatas = ATMEFct.C_IndexList_create(),
			itemsF10MenuCommandRoot = {},				-- Table des menus root : une par module.
			userPrivateDatas = {},						-- Point d'entrée des données privées : un par module
			eventCallbacks = {},						-- Callbacks d'événement
			tagsName = ATMEFct.decompString(_name),
			currentVoice = "",
			pilotEjected = false,
			pilotDead = false,
			isEnabled = true,
			destroyed = false,							-- Utile lors du delete (car isExist est déjà false)
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		for _k, _itemF10MenuCommandRoot in pairs(ATMEDatas.localMembers[_id].itemsF10MenuCommandRoot) do
			_itemF10MenuCommandRoot:removeAll()
			ATMEDatas.C_F10Menu_MT:remove(_itemF10MenuCommandRoot:getIndex())
			ATMEDatas.localMembers[_id].itemsF10MenuCommandRoot[_k] = nil
		end
		ATMEDatas.localMembers[_id].eventCallbacks = nil
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Player = {
	getClassName = function(self)
		return ATMEDatas.C_Player_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Player_MT.parentClass
	end,
	
	getByName = function(_playerName)
		if type(_playerName) == "string" then
			return ATMEDatas.players[_playerName]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Player.getByName(name)", "name must be string", false)
		end
	end,
	
	getRandomFromCoalition = function(_coalitionName)
		local fctDefinition = "Function ATME.C_Player.getRandomFromCoalition(coalitionName)"
		
		if type(_coalitionName) == "string" then
			if ATMEConst.coalitionIdsBlueRed[_coalitionName] ~= nil then
				return ATMEFct.getRandomPlayerFromCoalition(_coalitionName)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName must be 'BLUE' or 'RED'", false)
			end
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "coalitionName type must be string", false)
		end
	end,
}

ATMEFct.C_Player_create = function(_unit, _datas, _newObject)
	local newPlayer = {}
	setmetatable(newPlayer, ATMEDatas.C_Player_MT.mt)
	
	local typeName = _unit:getTypeName()								
	local airbase
	local oldSide
	
	local groupName = _unit:getGroup():getName()
	
	local index = ATMEDatas.C_Player_MT:add(_unit:getName(), _unit)
	local localMembers = ATMEDatas.localMembers[index]
	
	function newPlayer.getIndex(self)
		return index
	end
	
	function newPlayer.getDCSUnit(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.getDCSID(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCSId
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.getName(self)
		if localMembers.destroyed == false then
			return localMembers.name
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.isActivated(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:isActive()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.disable(self)
		if localMembers.destroyed == false then
			ATMEDatas.unitsToDisable[localMembers.name] = self
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.getCategoryName(self)
		if localMembers.destroyed == false then
			return localMembers.category
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "tagGetByID(position)"
		
		if localMembers.destroyed == false then
			if type(_position) == "number" then
				local pos = math.floor(_position)
				if localMembers.tagsName ~= nil and pos >= 1 and pos <= #localMembers.tagsName then
					return localMembers.tagsName[pos]
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.tagGetCount(self)
		if localMembers.destroyed == false then
			if tagsName ~= nil then
				return #tagsName
			else
				return 0
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "tagCompare(tagsToCompare)"
		
		if localMembers.destroyed == false then
			if type(_tagsToCompare) == "string" then
				if localMembers.tagsName ~= nil then
					local ok = ATMEFct.compareTags(localMembers.tagsName, _tagsToCompare)

					if ok ~= nil then
						return ok
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
					end
				else
					return false
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.getPseudo(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getPlayerName()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.getTypeName(self)
		if localMembers.destroyed == false then
			return typeName
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.getGroup(self)
		if localMembers.destroyed == false then
			return localMembers.group
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.getAttributes(self)
		if localMembers.destroyed == false then
			local datas = Unit.getDesc(localMembers.objectDCS)
			local attributes = {}
			for _id, _ in pairs(datas.attributes) do
				if ATMEConst.backTargetsAttributes[_id] ~= nil then
					table.insert(attributes, ATMEConst.backTargetsAttributes[_id].value)
				end
			end
			
			return attributes
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.hasAttributes(self, _attributes)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "hasAttributes(attributes)"
		
		if localMembers.destroyed == false then
			if type(_attributes) == "string" or type(_attributes) == "table" then
				local err, datas = ATMEFct.extractDatasFromStringOrTable(_attributes, false)
		
				if err == false then
					return ATMEFct.hasAttributes(localMembers.objectDCS, datas)
				end
				
				ATMEFct.errorOutputCore(nil, fctDefinition, datas, false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "attributes type must be string or table", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.addEventCallback(self, _fct, _eventType)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "addEventCallback(fct, [eventType])"
		
		if localMembers.destroyed == false then
			if type(_fct) == "function" and (_eventType == nil or type(_eventType) == "string") then
				if _eventType == nil then
					for _typeName, _ in pairs(ATMEConst.coreEventTypes) do
						localMembers.eventCallbacks[ATMEDatas.currentModuleName][_typeName] = _fct
					end
				else
					local eventType = string.upper(_eventType)
					if ATMEConst.coreEventTypes[eventType] ~= nil then
						localMembers.eventCallbacks[ATMEDatas.currentModuleName][eventType] = _fct
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "eventType must be an existing event type (see manual)", false)
					end
				end
			else
				if type(_fct) ~= "function" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
				elseif type(_eventType) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "eventType type must be nil or string", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.removeEventCallback(self, _eventType)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "removeEventCallback([eventType])"
		
		if localMembers.destroyed == false then
			if _eventType == nil or type(_eventType) == "string" then
				
				if _eventType == nil then
					for _typeName, _ in pairs(ATMEConst.coreEventTypes) do
						localMembers.eventCallbacks[ATMEDatas.currentModuleName][_typeName] = nil
					end
				else
					local eventType = string.upper(_eventType)
					if ATMEConst.coreEventTypes[eventType] ~= nil then
						localMembers.eventCallbacks[ATMEDatas.currentModuleName][eventType] = nil
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "eventType must be an existing event type (see manual)", false)
					end
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "eventType type must be nil or string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.getUserPrivateDatas(self)
		if localMembers.destroyed == false then
			return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.display(self, _text, _duration, _clear)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "display(text, duration,[clear])"
		
		if localMembers.destroyed == false then
			if type(_text) == "string" and type(_duration) == "number" and
					(_clear == nil or type(_clear) == "boolean") then
				
				local clear = false
				
				if type(_clear) == "boolean" then
					clear = _clear
				end
				
				if _duration >= 0 then
					trigger.action.outTextForGroup(localMembers.groupId, _text, _duration, clear)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be >= 0", false)
				end
			else
				if type(_text) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", true)
				elseif type(_duration) ~= "number" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be number", false)
				elseif _clear ~= nil and type(_clear) ~= "boolean" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "clear type must be nil or string", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.mark(self, _text, _reference, _readOnly, _message)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "mark(text, reference, readOnly, [message])"
		
		if localMembers.destroyed == false then
			if type(_text) == "string" and ATMEFct.isReference(_reference) == true and type(_readOnly) == "boolean" and
							(_message == nil or type(_message) == "string") then
				local point = ATMEFct.getPointFromReference(_reference)

				return ATMEFct.addMark(nil, nil, localMembers.groupId, self:getGroup():getName(), _text, point, _readOnly, _message)
			else
				if type(_text) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", true)
				elseif ATMEFct.isReference(_reference) == false then
					ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, false)
				elseif type(_readOnly) ~= "boolean" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "readOnly type must be boolean", false)
				elseif _message ~= nil and type(_message) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "message type must be nil or string", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.setVoice(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "setVoice([name])"
		
		if localMembers.destroyed == false then
			if _name == nil or type(_name) == "string" then
				if _name ~= nil then
					local d, f, temp = _name:find("^(%a[%a%d_]*)")
					
					if temp == _name then
						localMembers.currentVoice = _name
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "name bad format", false)
					end
				else
					localMembers.currentVoice = ""
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be nil or string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.getVoice(self)
		if localMembers.destroyed == false then
			if localMembers.currentVoice ~= "" then
				return localMembers.currentVoice
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.soundPlay(self, _text, _readMode)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "soundOnce(_text, [readMode])"
		
		if localMembers.destroyed == false then
			if type(_text) == "string" and (_readMode == nil or type(_readMode) == "boolean") then
				local randomRead = false
				local readOnce = true
				
				if type(_readMode) == "boolean" then
					randomRead = _readMode
					readOnce = false
				end
				
				local err, playlist = ATMEFct.C_Sound_getPlaylist(_text, randomRead)

				if err == true then
					ATMEFct.errorOutputCore(nil, fctDefinition, playlist, false)
					return
				end

				if playlist == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "text string has no sound token", false)
					return
				end
			
				ATMEFct.C_Sound_initPlaylist({
					player = self, 
					randomRead = randomRead, 
					readOnce = readOnce, 
					playlist = playlist,
					timeSound = timer.getTime(),
				})
			else
				if type(_text) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", true)
				elseif _readMode ~= nil and type(_readMode) ~= "boolean" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "readMode type must be nil or boolean", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.soundStop(self)
		if localMembers.destroyed == false then
			if localMembers.soundDatas ~= nil then
				localMembers.soundDatas.toRemove = true
				localMembers.soundDatas = nil
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.getFuelRatio(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getFuel()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.getLife(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getLife()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.getLifeRatio(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getLife()/localMembers.objectDCS:getLife0()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.getCallsign(self)
		if localMembers.destroyed == false then
			local identification
			local id1
			local id2
			
			if localMembers.objectDCS:hasAttribute("Helicopters") or localMembers.objectDCS:hasAttribute("Planes") then
				local callsign = localMembers.objectDCS:getCallsign()
				if localMembers.objectDCS:getCoalition() == 2 then
					identification = string.match(callsign, "^%a*")
					id1 = string.match(callsign, "%d")
					id2 = string.match(callsign, "%d$")
				else
					identification = callsign
				end
			end
			
			return identification, id1, id2
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.getPosition(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getPoint()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.getVelocityVector(self)
		if localMembers.destroyed == false then
			local vector = ATMEFct.C_Vector3D_create(nil, nil, localMembers.objectDCS:getVelocity())
			return vector
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.getSpeed(self)
		if localMembers.destroyed == false then
			local vector = self:getVelocityVector()
			return vector:getModule(vector)
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.getHSpeed(self)
		if localMembers.destroyed == false then
			local vector = self:getVelocityVector()
			return vector:getHModule(vector)
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.getVSpeed(self)
		if localMembers.destroyed == false then
			local vector = localMembers.objectDCS:getVelocity()
			return vector.y
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.getSpeedAzimuth(self)
		if localMembers.destroyed == false then
			local vector = self:getVelocityVector()
			
			if vector:getModule() > 1 then 
				return vector:getAzimuth()
			else
				return 0
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.getRollAxisVector(self)
		if localMembers.destroyed == false then
			return ATMEFct.C_Vector3D_create(nil, nil, localMembers.objectDCS:getPosition().x)
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.getAzimuth(self)
		if localMembers.destroyed == false then
			local vector = self:getRollAxisVector()
			return vector:getAzimuth()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.getCountryName(self)
		if localMembers.destroyed == false then
			local id = localMembers.objectDCS:getCountry()
			
			return country.name[id]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.getCoalitionName(self)
		if localMembers.destroyed == false then
		local id = localMembers.objectDCS:getCoalition()
			return ATMEConst.coalitionNames[id]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.isHelicopter(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("Helicopters")
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.isPersonnelCarrier(self)
		if localMembers.destroyed == false then
			return ATMEConst.personnelCarrierAttributes[self:getTypeName()] ~= nil
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.isPlane(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("Planes")
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.canFly(self)
		if localMembers.destroyed == false then
			local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
			return ATMEConst.canFly[localMembersGroup.category]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.inAir(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:inAir()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.isGroundVehicle(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("Ground vehicles")
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.getMSLAltitude(self)
		if localMembers.destroyed == false then
			return ATME.getPointAltitude(self:getPosition())
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.getAGLAltitude(self)
		if localMembers.destroyed == false then
			local landHeight = ATMEFct.getSurfaceAltitude(self:getPosition())
			return ATME.getPointAltitude(self:getPosition()) - landHeight
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.explode(self, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "explode(power)"
		
		if localMembers.destroyed == false then
			if type(_power) == "number" then
				if _power >= 1 then
					ATME.explode(self:getPosition(), _power)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.isInZones(self, _zonesMgr)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "isInZones(zonesMgr)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zonesMgr, ATME.C_ZonesMgr) == true then
				return _zonesMgr:isInside(self) 
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zonesMgr type must be ATME.C_ZonesMgr", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end				
	
	function newPlayer.isInDCSZone(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "isInDCSZone(name)"
		
		if localMembers.destroyed == false then
			if type(_name) == "string" then
				local zone = ATMEFct.getDCSZone(_name)
				
				if zone ~= nil then
					local vector = ATMEFct.C_Vector3D_create(nil, nil, self:getPosition(), zone.point)
					if vector:getHModule() <= zone.radius then
						return true
					else
						return false
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _name .. " unknown", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.isInZone2D(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "isInZone2D(zone)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zone, ATME.C_Zone2D) == true then
				local point = self:getPosition()
				local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
				
				local datas = {
					zone = _zone,
					actualCenter = localMembersZone.center,
					azimuth = localMembersZone.azimuth,
				}
			
				return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](point, datas)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Circle, ATME.C_PickupZone, ATME.C_Rectangle or ATME.C_Polygon", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.isInZone3D(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "isInZone3D(zone)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zone, ATME.C_Zone3D) == true then
				local point = self:getPosition()
				local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
		
				local datas = {
					zone = _zone,
					actualCenter = localMembersZone.center,
					azimuth = localMembersZone.azimuth,
				}
				
				return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](point, datas)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Sphere, ATME.C_Cylinder or ATME.C_Box", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.whichSide(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "whichSide(reference)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isReference(_reference) == true  then
				local point = ATMEFct.getPointFromReference(_reference)

				return ATMEFct.C_Player_C_AIUnit_whichSide(self, point)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.isInDirection(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "isInDirection(reference)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isReference(_reference) == true then
				return ATMEFct.isAzimuthInLimits(self:getPosition(), self:getSpeedAzimuth(), ATMEFct.getPointFromReference(_reference))
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.isEngineStarted(self)
		if localMembers.destroyed == false then
			return localMembers.engineOn
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.getF10MenuRoot(self)
		if localMembers.destroyed == false then
			return localMembers.itemsF10MenuCommandRoot[ATMEDatas.currentModuleName]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.getNbUnitsOnBoard(self)
		if localMembers.destroyed == false then
			return localMembers.nbUnitsOnBoard
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.getOnBoardGroupNames(self)
		if localMembers.destroyed == false then
			local temp = {}
			
			for _id, _item in localMembers.onBoardGroupDatas:pairs() do
				temp[_item.name] = _id
			end
			
			return temp
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.unload(self, _param1, _param2)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "unload([whenStopped])\nunload(id, [whenStopped])"

		if localMembers.destroyed == false then	
			if (_param1 == nil and _param2 == nil) or type(_param1) == "boolean" then
				local whenStopped = false
				 if type(_param1) == "boolean" then
					whenStopped = _param1
				end
				
				ATMEFct.C_Player_C_AIUnit_unload(localMembers, nil, whenStopped)
			elseif type(_param1) == "number" and (_param2 == nil or type(_param2) == "boolean") then
				local id = math.floor(_param1)
				
				local whenStopped = _param2
				
				if _param2 == nil then
					whenStopped = false
				end
				
				ATMEFct.C_Player_C_AIUnit_unload(localMembers, id, whenStopped)
			else
				if type(_param1) ~= "boolean" and type(_param1) ~= "number" and _param2 == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "whenStopped must be nil or boolean\nid type must be number", true)
				else
					if type(_param1) ~= "number" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "id type must be number", true)
					elseif _param2 ~= nil and type(_param2) ~= "boolean" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "whenStopped type must be nil or boolean", false)
					end
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newPlayer.load(self, _radius, _groupToBoard)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "load(radius, [groupToBoard])"
		
		if localMembers.destroyed == false then
			if type(_radius) == "number" and (_groupToBoard == nil or ATMEFct.isObjectClass(_groupToBoard, ATME.C_Group) == true or type(_groupToBoard) == "string") then
				if self:isActivated() == true then
					local groupToBoard
					
					if _groupToBoard == nil then
						groupToBoard = ATMEFct.C_Player_C_AIUnit_getNearestReadyToBoardGroup(self, _radius)
					elseif type(_groupToBoard) == "string" then
						groupToBoard = ATME.C_Group.getByName(_groupToBoard)
					else
						groupToBoard = _groupToBoard
					end
					
					if groupToBoard ~= nil then
						if groupToBoard:isOnlyInfantry() == true then
							if _radius > 0 then
								if groupToBoard:isReadyToBoard() == true and groupToBoard:isBoarding() == false then
									if ATMEConst.personnelCarrierAttributes[self:getTypeName()] ~= nil then
										return ATMEFct.C_Player_C_AIUnit_load(self, groupToBoard, _radius)
									else
										ATMEFct.errorOutputCore(nil, fctDefinition, self:getName() .. " not a transport unit", false)
									end
								else
									ATMEFct.setError("INFANTRY_GROUP_NOT_READY")
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Not infantry group", false)
						end
					else
						ATMEFct.setError("NO_GROUP_FOUND")
					end
				else
					ATMEFct.setError("UNACTIVATED_UNIT")
				end
			else
				if type(_radius) ~= "number" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", true)
				elseif ATMEFct.isObjectClass(_groupToBoard, ATME.C_Group) == false and _groupToBoard ~= nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "groupToBoard type must be nil or ATME.C_Group or string", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.isPilotDead(self)
		if localMembers.destroyed == false then
			return localMembers.pilotDead
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.isPilotEjected(self)
		if localMembers.destroyed == false then
			return localMembers.pilotEjected
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newPlayer.getNearestReadyToBoardGroup(self, _radius)
		local fctDefinition = ATMEFct.getErrorPrefix(newPlayer) .. "getNearestReadyToBoardGroup(radius)"
		
		if localMembers.destroyed == false then
			if type(_radius) == "number" then
				if _radius >= 0 then
					return ATMEFct.C_Player_C_AIUnit_getNearestReadyToBoardGroup(self, _radius)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be number", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	newPlayer.modules = {}
	setmetatable(newPlayer.modules, ATMEDatas.C_Player_MT.mtModules)
	
	if ATMEConst.personnelCarrierAttributes[newPlayer:getTypeName()] ~= nil and newPlayer:isHelicopter() == true then
		localMembers.indexPickupZoneSignal = ATMEDatas.signalCapabilityUnits:add({ 
																					object = newPlayer,
																					localMembers = localMembers})
	end
	
	if ATMEDatas.newCreatedGroups[groupName] == nil then
		ATMEFct.C_Group_create(newPlayer, _datas, _newObject)
	else
		ATMEFct.C_Group_addUnit(ATMEDatas.newCreatedGroups[groupName].group, newPlayer)
	end

	localMembers.group = ATMEDatas.newCreatedGroups[groupName].group
	
	local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
	
	if localMembersGroup.waypoints[1].airdromeId ~= nil or localMembersGroup.waypoints[1].helipadId ~= nil then
		-- Bug DCS ou changement de gestion des airbases ID
		if localMembersGroup.waypoints[1].airdromeId ~= nil then
			airbase = ATMEDatas.airbasesById[localMembersGroup.waypoints[1].airdromeId]:getName()
		elseif localMembersGroup.waypoints[1].helipadId ~= nil then
			airbase = ATMEDatas.airbasesById[localMembersGroup.waypoints[1].helipadId]:getName()
		end
		
		if localMembersGroup.waypoints[1].type == "TakeOffParkingHot" then
			-- Au parking moteur allumé
			
			localMembers.engineOn = true
		elseif localMembersGroup.waypoints[1].type == "TakeOff" then
			-- Sur la piste, prêt à décoler
			
			localMembers.engineOn = true
		end
	else
		-- En vol....
		
		localMembers.engineOn = true
	end
	
	localMembers.category = localMembersGroup.category
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		local _, menu = ATMEFct.C_F10Menu_create(newPlayer, nil, nil, _localModuleMembers.name)

		localMembers.itemsF10MenuCommandRoot[_localModuleMembers.name] = menu
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
		localMembers.eventCallbacks[_localModuleMembers.name] = {}
		newPlayer.modules[_localModuleMembers.name] = {}
	end
	
	ATMEFct.outputCore(nil,"Prepare Player = " .. localMembers.name .. " -> ok", 2)
end

ATMEFct.C_Player_delete = function(_player)
	local localMembers = ATMEDatas.localMembers[_player:getIndex()]
	local flag = "DISABLED"
	if localMembers.isEnabled == true then
		flag = "EJECTION"
		if _player:isPilotDead() == true then
			flag = "DEAD"
		end
	end
	
	for _i = #ATMEDatas.modulesNames, 1, -1 do
		local localModuleMembers = ATMEDatas.modulesLocalMembers[ATMEDatas.modulesNames[_i]]
		if localModuleMembers.handlers.onDeletePlayerHandler ~= nil then
			ATMEDatas.currentModuleName = localModuleMembers.name
		
			-- Appel du module
			ATMEFct.outputCore(nil,"Appel onDeletePlayerHandler du module " .. localModuleMembers.name, 2)
			xpcall(function() localModuleMembers.handlers.onDeletePlayerHandler(_player, flag) end, function() ATMEFct.errUserHandler('onDeletePlayerHandler') end)
		end
	end
	
	for _, _race in pairs(ATMEDatas.races) do
		_race:removePlayer(_player)
	end
	
	if localMembers.indexPickupZoneSignal ~= nil then
		ATMEDatas.signalCapabilityUnits:remove(localMembers.indexPickupZoneSignal)
		localMembers.indexPickupZoneSignal = nil
	end
	
	_player.modules = nil
	
	for _, _item in localMembers.onBoardGroupDatas:pairs() do
		-- Si un des groupes embarqués a un respawn automatique il est préparé
		
		if _item.autoRespawn.activated == true then
			local respawnTime = _item.autoRespawn.time
			if _item.autoRespawn.random == true then
				respawnTime = math.random(_item.autoRespawn.minTime, _item.autoRespawn.maxTime)
			end
			
			ATMEDatas.groupsToRespawn[_item.name] = {
				time = respawnTime + timer.getTime(),
				spawnDatas = _item.spawnDatas,
			}
		end
		
		ATMEDatas.groupsOnBoard[_item.name] = nil
	end

	localMembers.onBoardGroupDatas:removeAll()
	
	ATMEFct.C_Group_deleteUnit(_player:getGroup(), localMembers.name)
	ATMEDatas.C_Player_MT:remove(_player:getIndex())
	
	ATMEDatas.players[localMembers.name] = nil
	ATMEDatas.allAIUnitsAndPlayers[localMembers.name] = nil
	
	ATMEDatas.nbPlayers = ATMEDatas.nbPlayers - 1
	localMembers.destroyed = true

	ATMEFct.outputCore(nil,"Delete Player : " .. localMembers.name , 2)
end

setmetatable(ATME.C_Player,ATMEDatas.C_Player_MT)

ATMEDatas.C_Player_MT.mt.__metatable = "Private"
ATMEDatas.C_Player_MT.mt.__index = ATME.C_Player
ATMEDatas.C_Player_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Player_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Player : Bad key " .. k, "table locked", false)
	end
end

ATMEDatas.C_Player_MT.mtModules.__metatable = "Private"
ATMEDatas.C_Player_MT.mtModules.__newindex = function(t, k, v)
	if ATMEDatas.modules[k] ~= nil then
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil,"ATME : ATME.C_Player.modules Bad key " .. k, "table locked", false)
	end
end


----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_AIUnit : gère les unités AI au format ATME. Pas de contructeur accessible
-- pour les modules externes.
-- Les instances de cette classe sont intégralement gérées par le module ATME_Core. La création et la destruction
-- dépendent du contexte de début de mission et des événements qui se produisent dans DCS en cours de mission.
-- Les modules externes accèdent à cette gestion indirectement par les handlers définis dans ceux-ci et les méthodes
-- accessibles de cette classe. Ceci permet de garantir une cohérence sur les listes internes ATME.
----------------------------------------------------------------------------------------------------------

ATMETimers.C_Unit_spotReassign_timer = function(_item)
	local targetName
	local targetClassName
	
	if _item.reassign == false then
		targetName = ATMEDatas.spots[_name].targetName
		targetClassName = ATMEDatas.spots[_name].targetClassName
	else
		if _item.groupName ~= nil and ATMEDatas.groups[_item.groupName] ~= nil then
			targetName, targetClassName = ATMEFct.C_AIUnit_findTargetInGroup(ATMEDatas.groups[_item.groupName], ATMEDatas.spots[_name].ignoredTargets)
		elseif _item.zone ~= nil then
			if ATMEDatas.allAIUnitsAndPlayers[ATMEDatas.spots[_name].targetName] ~= nil then
				if _item.outOfZone == true then
					ATMEFct.outputCore(nil, "Spot par unité " .. _name .. " : unité " .. ATMEDatas.spots[_name].targetName .. " -> Hors zone", 2)
					local event = {
						type = "SPOT_TARGET_OUT_OF_ZONE",
						datas = {
							unitName = _name,
							targetName = ATMEDatas.spots[_name].targetName,
							targetClassName = ATMEDatas.spots[_name].targetClassName,
						},
					}
					
					ATMEFct.C_Event_add(event)
				elseif _item.noLOS == true then
					if _item.neverHaveLOS == false then
						ATMEFct.outputCore(nil, "Spot par unité " .. _name .. " : unité " .. ATMEDatas.spots[_name].targetName .. " -> Plus de visée", 2)
						local event = {
							type = "SPOT_TARGET_NO_LOS",
							datas = {
								unitName = _name,
								targetName = ATMEDatas.spots[_name].targetName,
								targetClassName = ATMEDatas.spots[_name].targetClassName,
							},
						}
						
						ATMEFct.C_Event_add(event)
					end
				end
			end

			targetName, targetClassName = ATMEFct.C_AIUnit_findTargetInZone2D(_item.zone, ATMEConst.coalitionEnemy[ATMEDatas.AIUnits[_name]:getCoalitionName()], ATMEDatas.spots[_name].ignoredTargets)
		end

		ATMEDatas.spots[_name].targetName = targetName
	end
	
	if targetName ~= nil then
		local point = ATMEDatas.allAIUnitsAndPlayers[targetName]:getPosition()
		local posUnit = ATMEDatas.AIUnits[_name]:getPosition()
		
		ATMEDatas.spots[_name].vector =  ATMEFct.C_Vector3D_create(nil, nil, posUnit, point)
		ATMEDatas.spots[_name].noLOS = 0
		ATMEDatas.spots[_name].ignoredTargetOn = false
		
		local offset = {x = 0, y = -2, z = 0}
		if ATMEDatas.spots[_name].fromAir == false then
			offset = {x = 0, y = 2, z = 0}
		end
				
		if ATMEDatas.spots[_name].code ~= nil then
			ATMEDatas.spots[_name].spot = Spot.createLaser(ATMEDatas.AIUnits[_name]:getDCSUnit(),
															offset, 
															point, 
															ATMEDatas.spots[_name].code)
			ATMEFct.outputCore(nil, "Spot LASER par unité " .. _name .. " : ciblage unité " .. targetName, 2)
		else
			ATMEDatas.spots[_name].spot = Spot.createInfraRed(ATMEDatas.AIUnits[_name]:getDCSUnit(),
														offset, 
														point)
			ATMEFct.outputCore(nil, "Spot IR par unité " .. _name .. " : ciblage unité " .. targetName, 2)
		end
		
		local event = {
			type = "SPOT_NEW_TARGET",
			datas = {
				unitName = _name,
				targetName = targetName,
				targetClassName = ATMEDatas.spots[_name].targetClassName,
			},
		}
		
		ATMEFct.C_Event_add(event)
	else
		ATMEFct.outputCore(nil, "Spot par unité " .. _name .. " : plus d'unités à cibler. Arrêt", 2)
		ATMEDatas.spots[_name] = nil
	
		local event = {
			type = "SPOT_STOP",
			datas = {
				unitName = _name,
			},
		}

		ATMEFct.C_Event_add(event)
	end
end

ATMETimers.C_Unit_spot_timer = function(_localMembers)
	if ATMEDatas.exitATME == false then
		-- Suppression des spots actifs pour les unités AI au sol qui se mettent à bouger
		-- Tracking si pointeur sur objet dynamique
		if _localMembers.destroyed == false then		
			if _localMembers.spotDatas.fromAir == false and ATMEDatas.AIUnits[localMembers.name]:getSpeed() > 0 then
				-- L'unité au sol (JTAC) se déplace, arrêt du spot
				
				_localMembers.spotDatas.spot:destroy()
				_localMembers.spotDatas.spot = nil
				
				local event = {
					type = "SPOT_STOP_UNIT_MOVE",
					datas = {
						unitName = localMembers.name,
					},
				}
				ATMEFct.outputCore(nil, "Spot par unité " .. localMembers.name .. " : Arrêt cause déplacement", 2)		
				ATMEFct.C_Event_add(event)
			else
				if _localMembers.spotDatas.spot == nil then -- Pas de spot actif
					local item = {
						reassign = false,
					}
					
					timer.scheduleFunction(ATMETimers.C_Unit_spotReassign_timer, item, timer.getTime() + 1)
				elseif _localMembers.spotDatas.spot ~= nil then -- Un spot est actif
					if ATMEDatas.allAIUnitsAndPlayers[_localMembers.spotDatas.targetName] ~= nil and _localMembers.spotDatas.ignoredTargetOn == false then
						local posUnit = ATMEDatas.AIUnits[localMembers.name]:getPosition()
						local point = ATMEDatas.allAIUnitsAndPlayers[_localMembers.spotDatas.targetName]:getPosition()
						
						if (_localMembers.spotDatas.fromAir == false and land.isVisible({x = posUnit.x, y = posUnit.y + 2, z = posUnit.z} , {x = point.x, y = point.y + 2, z = point.z}) == false) or 
						   (_localMembers.spotDatas.fromAir == true and land.isVisible({x = posUnit.x, y = posUnit.y - 2, z = posUnit.z}  , {x = point.x, y = point.y + 2, z = point.z}) == false) then
							_localMembers.spotDatas.noLOS = _localMembers.spotDatas.noLOS + 1
						else
							_localMembers.spotDatas.neverHaveLOS = false
							_localMembers.spotDatas.noLOS = 0
							_localMembers.spotDatas.ignoredTargets[_localMembers.spotDatas.targetName] = nil
						end
						
						if _localMembers.spotDatas.noLOS == 50 then -- Pas de ligne de visée pendant 5 secondes. Abandon
							_localMembers.spotDatas.spot:destroy()
							_localMembers.spotDatas.spot = nil
							local item = {
								groupName = _localMembers.spotDatas.groupName,
								zone = _localMembers.spotDatas.zone,
								reassign = true,
								noLOS = true, 
								neverHaveLOS = _localMembers.spotDatas.neverHaveLOS,
							}
							_localMembers.spotDatas.ignoredTargets[_localMembers.spotDatas.targetName] = true
							timer.scheduleFunction(ATMETimers.C_Unit_spotReassign_timer, item, timer.getTime() + 1)
						elseif _localMembers.spotDatas.dynamicSpot == true and (ATMEDatas.allAIUnitsAndPlayers[_localMembers.spotDatas.targetName]:getSpeed() > 0 or _localMembers.spotDatas.fromAir == true) then
							local datas
							
							if _localMembers.spotDatas.zone ~= nil then
								local localMembers = ATMEDatas.localMembers[_localMembers.spotDatas.zone:getIndex()]
								datas = {
									zone = _localMembers.spotDatas.zone,
									actualCenter = localMembers.center,
									actualAzimuth = localMembers.azimuth,
								}
							end
							
							if _localMembers.spotDatas.zone == nil or 
								(datas ~= nil and ATMEFct.C_FindContext_Object_isInZone(ATMEDatas.allAIUnitsAndPlayers[_localMembers.spotDatas.targetName], datas) == true) then
								-- La cible existe et elle est dynamique... Le spot suit la cible.
								
								_localMembers.spotDatas.vector =  ATMEFct.C_Vector3D_create(nil, nil, posUnit, point)
								_localMembers.spotDatas.spot:setPoint(point)
								
								return timer:getTime() + 0.01
							else -- La cible est sortie de la zone de surveillance
								_localMembers.spotDatas.spot:destroy()
								_localMembers.spotDatas.spot = nil
								local item = {
									zone = _localMembers.spotDatas.zone,
									outOfZone = true,
									reassign = true,
								}
								
								timer.scheduleFunction(ATMETimers.C_Unit_spotReassign_timer, item, timer.getTime() + 1)
							end
						end
					else -- La cible n'existe plus
						_localMembers.spotDatas.spot:destroy()
						_localMembers.spotDatas.spot = nil
						
						ATMEDatas.spotsToReassign[localMembers.name] = {
							delay = 1,
							groupName = _localMembers.spotDatas.groupName,
							zone = _localMembers.spotDatas.zone,
							reassign = true,
						}
					end
				end
			end
		end
	end
end

ATMETimers.C_Unit_targetTracking = function(_localMembers)
	if ATMEDatas.exitATME == false then
		if _localMembers.objectDCS:isExist() == true then
			local targets = ATMEFct.C_AIUnit_getDetectedTargets(_localMembers, _localMembers.targetsTracking.typeOfSensorName)
			for _targetName, _ in pairs(targets) do
				if targets[_targetName].stillDetected == false then
					if targets[_targetName].confirmed > 0 then
						targets[_targetName].confirmed = targets[_targetName].confirmed - 1	
					elseif targets[_targetName].confirmed == 0 and targets[_targetName].trackingOn == true then
						targets[_targetName] = nil

						local event = {
							type = "TARGET_END_TRACK",
							datas = {
								unitName = _localMembers.name,
								targetName = _targetName,
							},
						}
						
						ATMEFct.C_Event_add(event)
					end
				else
					if targets[_targetName].confirmed < 5 then
						targets[_targetName].confirmed = targets[_targetName].confirmed + 1
					elseif targets[_targetName].confirmed == 5 then
						if targets[_targetName].trackingOn == false and (_localMembers.targetsTracking.flag == nil or ATME.getFlag(_localMembers.targetsTracking.flag) ~= 0) then	
							-- Event nouvelle target dans le tracking confirmée
							
							targets[_targetName].trackingOn = true
							local event = {
								type = "TARGET_NEW_TRACK",
								datas = {
									unitName = _localMembers.name,
									targetName = _targetName,
								},
							}
							
							ATMEFct.C_Event_add(event)
						elseif targets[_targetName].trackingOn == true and _localMembers.targetsTracking.flag ~= nil and ATME.getFlag(_localMembers.targetsTracking.flag) == 0 then
							targets[_targetName].trackingOn = false
							local event = {
								type = "TARGET_END_TRACK",
								datas = {
									unitName = _localMembers.name,
									targetName = _targetName,
								},
							}
							
							ATMEFct.C_Event_add(event)
						end
					end
				end
			end
			
			return timer.getTime() + 2
		else		
			_localMembers.targetsTracking.idTimer = nil 
		end
	end
end

ATMEFct.C_AIUnit_startTransmission = function(_radioTransmission)
	if _radioTransmission.started == false and _radioTransmission.playlist[_radioTransmission.currentReadIndex].pause == false then
		local controller = ATMEDatas.AIUnits[_radioTransmission.unitName]:getDCSUnit():getController()

		local commandFreq = { 
			id = "SetFrequency", 
			params = {
				frequency = _radioTransmission.localMembersRadioChannel.frequency, 
				modulation = ATMEConst.modulationValues[_radioTransmission.localMembersRadioChannel.modulation],
			},
		}
		
		controller:setCommand(commandFreq)
		
		local command = { 
			id = "TransmitMessage", 
			params = {
				file = "l10n/DEFAULT/" .. _radioTransmission.voice .. _radioTransmission.playlist[_radioTransmission.currentReadIndex].fileName,
				duration = _radioTransmission.playlist[_radioTransmission.currentReadIndex].subtitle.duration,
				subtitle = _radioTransmission.playlist[_radioTransmission.currentReadIndex].subtitle.text,
				loop = _radioTransmission.autoRestartDCS,
			} 
		}
		controller:setCommand(command)
		_radioTransmission.started = true
	end
end

ATMEFct.C_AIUnit_stopTransmission = function(_radioTransmission)
	if _radioTransmission.started == true then
		local controller = ATMEDatas.AIUnits[_radioTransmission.unitName]:getDCSUnit():getController()
		
		local command = { 
			id = "StopTransmission", 
			params = {},
				
		}
		
		controller:setCommand(command)
	
		_radioTransmission.started = false
	end
end

ATMEFct.C_AIUnit_loopAttributes = function(_unit, _datas)
	local datas = Unit.getDesc(_unit)

	for _id, _ in pairs(datas.attributes) do
		if ATMEConst.backTargetsAttributes[_id] ~= nil then
			if _datas.currentPriority < ATMEConst.backTargetsAttributes[_id].priority then
				_datas.currentPriority = ATMEConst.backTargetsAttributes[_id].priority
				_datas.targetName = _unit:getName()
				_datas.targetClassName = _unit:getClassName()
			end
		end
	end
end

ATMEFct.C_AIUnit_isTargetOk = function(_ATMEUnit, _ignoredTargets)
	local group = _ATMEUnit:getGroup()
	local active = _ATMEUnit:getDCSUnit():isActive()
	local ok = false
	
	if _ignoredTargets[_ATMEUnit:getName()] ~= nil then
		ok = true
	end
	
	return (_ATMEUnit:getCategoryName() == "GROUND" or 
			((_ATMEUnit:getCategoryName() == "AIRPLANE" or _ATMEUnit:getCategoryName() == "HELICOPTER") and _ATMEUnit:inAir() == false)) and 
			(active == true or (active == false and group:isAlwaysVisible() == true)) and ok == false
end

ATMEFct.isTargetOk = {
	["C_AIUnit"] = function(_ATMEUnit, _ignoredTargets)
		return ATMEFct.C_AIUnit_isTargetOk(_ATMEUnit, _ignoredTargets)
	end,
	["C_Player"] = function(_ATMEUnit, _ignoredTargets)
		return ATMEFct.C_AIUnit_isTargetOk(_ATMEUnit, _ignoredTargets)
	end,
	["C_StaticObject"] = function(_ATMEObject, _ignoredTargets)
		local ok = false
		
		if _ignoredTargets[_ATMEObject:getName()] ~= nil then
			ok = true
		end
		
		return (_ATMEUnit:getCategoryName() == "GROUND" or 
				_ATMEUnit:getCategoryName() == "AIRPLANE" or _ATMEUnit:getCategoryName() == "HELICOPTER") and 
				ok == false
	end,
	["C_SceneryObject"] = function(_ATMEObject, _ignoredTargets)
		-- TODO
	end,
}

ATMEFct.C_AIUnit_searchTargetsInBox = function(_unit, _datas)
	local ATMEUnit = ATMEDatas.allAIUnitsAndPlayers[_unit:getName()]
	
	if ATMEFct.isTargetOk[ATMEUnit:getClassName()](ATMEUnit, _datas.ignoredTargets) == true and ATMEUnit:getCoalitionName() == _datas.targetsCoalition then
		if ATMEFct.C_FindContext_Object_isInZone(ATMEUnit, _datas) == true then	
			ATMEFct.C_AIUnit_loopAttributes(_unit, _datas)
			
			if _datas.targetName == nil then
				_datas.targetName = _unit:getName()
				_datas.targetClassName = _unit:getClassName()
			end
		end
	end
end

ATMEFct.searchTargetsClassInBox = {
	["C_AIUnit"] = function(_unit, _datas)
		ATMEFct.C_AIUnit_searchTargetsInBox(_unit, _datas)
	end,
	["C_Player"] = function(_unit, _datas)
		ATMEFct.C_AIUnit_searchTargetsInBox(_unit, _datas)
	end,
	["C_StaticObject"] = function(_unit, _datas)
		local ATMEObject = ATMEDatas.staticObjects[_unit:getName()]
		
		if ATMEFct.isTargetOk[ATMEObject:getClassName()](ATMEObject, _datas.ignoredTargets) == true and ATMEObject:getCoalitionName() == _datas.targetsCoalition then
			if ATMEFct.C_FindContext_Object_isInZone(ATMEObject, _datas) == true then	
				ATMEFct.C_AIUnit_loopAttributes(_unit, _datas)
				
				if _datas.targetName == nil then
					_datas.targetName = _unit:getName()
					_datas.targetClassName = _unit:getClassName()
				end
			end
		end
	end,
	["C_Cargo"] = function(_unit, _datas)
		local ATMEObject = ATMEDatas.cargos[_unit:getName()]
		
		if ATMEFct.isTargetOk[ATMEObject:getClassName()](ATMEObject, _datas.ignoredTargets) == true and ATMEObject:getCoalitionName() == _datas.targetsCoalition then
			if ATMEFct.C_FindContext_Object_isInZone(ATMEObject, _datas) == true then	
				ATMEFct.C_AIUnit_loopAttributes(_unit, _datas)
				
				if _datas.targetName == nil then
					_datas.targetName = _unit:getName()
					_datas.targetClassName = _unit:getClassName()
				end
			end
		end
	end,
	["C_SceneryObject"] = function(_unit, _datas)
		local ATMEObject = ATMEDatas.sceneryObjects[_unit:getName()]
		
		if ATMEFct.isTargetOk[ATMEObject:getClassName()](ATMEObject, _datas.ignoredTargets) == true and ATMEObject:getCoalitionName() == _datas.targetsCoalition then
			if ATMEFct.C_FindContext_Object_isInZone(ATMEObject, _datas) == true then	
				ATMEFct.C_AIUnit_loopAttributes(_unit, _datas)
				
				if _datas.targetName == nil then
					_datas.targetName = _unit:getName()
					_datas.targetClassName = _unit:getClassName()
				end
			end
		end
	end,
}

ATMEFct.searchTargetsInBox = function(_unit, _datas)
	ATMEFct.searchTargetsClassInBox[_unit:getCategory()](_unit, _datas)
end

ATMEFct.C_AIUnit_findTargetInGroup = function(_group , _ignoredTargets)
	local unit = ATMEFct.getFirstUnit(_group)
	
	if unit ~= nil then
		local datas = {
			currentPriority = 0,
		}
		
		repeat
			if ATMEFct.isTargetOk[unit:getClassName()](unit, _ignoredTargets) == true then
				local found = false
				local unitName = unit:getName()

				ATMEFct.C_AIUnit_loopAttributes(unit:getDCSUnit(), datas)
				
				if datas.targetName == nil then
					datas.targetName = unitName
					datas.targetClassName = _unit:getClassName()
				end
			end
			
			unit = _group:getNextUnit()
		until unit == nil
		
		return datas.targetName, datas.targetClassName
	end
end

ATMEFct.C_AIUnit_findTargetInZone2D = function(_zone, _targetsCoalition, _ignoredTargets)
	local localMembers = ATMEDatas.localMembers[_zone:getIndex()]
	local volume = {
		id = world.VolumeType.BOX,
		params = localMembers.box,
	}
	
	local datas = {
		currentPriority = 0,
		zone = _zone,
		targetsCoalition = _targetsCoalition,
		ignoredTargets = _ignoredTargets,
		actualCenter = localMembers.center,
		actualAzimuth = localMembers.azimuth,
	}

	world.searchObjects(Object.Category.UNIT, volume, ATMEFct.searchTargetsInBox, datas)
		
	
	return datas.targetName, datas.targetClassName
end

ATMEFct.C_AIUnit_findTargetInZone2D = {
	["C_AIUnit"] = function(_zone, _targetsCoalition, _ignoredTargets)
		return ATMEFct.C_AIUnit_findTargetInZone2D(_zone, _targetsCoalition, _ignoredTargets)
	end,
	["C_Player"] = function(_zone, _targetsCoalition, _ignoredTargets)
		return ATMEFct.C_AIUnit_findTargetInZone2D(_zone, _targetsCoalition, _ignoredTargets)
	end,
	["C_StaticObject"] = function(_zone, _targetsCoalition, _ignoredTargets)
		local localMembers = ATMEDatas.localMembers[_zone:getIndex()]
		local volume = {
			id = world.VolumeType.BOX,
			params = localMembers.box,
		}
		
		local datas = {
			currentPriority = 0,
			zone = _zone,
			targetsCoalition = _targetsCoalition,
			ignoredTargets = _ignoredTargets,
			actualCenter = localMembers.center,
			actualAzimuth = localMembers.azimuth,
		}

		world.searchObjects({Object.Category.STATIC, Object.Category.CARGO} , volume, ATMEFct.C_AIUnit_searchTargetsInBox, datas)
	
		return datas.targetName, datas.targetClassName
	end,
	["C_SceneryObject"] = function(_zone, _targetsCoalition, _ignoredTargets)
		local localMembers = ATMEDatas.localMembers[_zone:getIndex()]
		local volume = {
			id = world.VolumeType.BOX,
			params = localMembers.box,
		}
		
		local datas = {
			currentPriority = 0,
			zone = _zone,
			targetsCoalition = _targetsCoalition,
			ignoredTargets = _ignoredTargets,
			actualCenter = localMembers.center,
			actualAzimuth = localMembers.azimuth,
		}

		world.searchObjects(Object.Category.SCENERY, volume, ATMEFct.C_AIUnit_searchTargetsInBox, datas)
	
		return datas.targetName, datas.targetClassName
	end,
}

ATMEFct.C_AIUnit_getDetectedTargets = function(_localMembers, _typeOfSensorName)
	local controller = _localMembers.objectDCS:getController()
	local detectedTargetsList
	
	if _typeOfSensorName ~= nil then
		detectedTargetsList = controller:getDetectedTargets(ATMEConst.typesOfSensor[_typeOfSensorName])
	else
		detectedTargetsList = controller:getDetectedTargets()
	end
	

	local nb = 0

	for _, _target in pairs(_localMembers.targets) do
		_target.stillDetected = false
	end

	for _, _target in ipairs(detectedTargetsList) do
		if _target.object ~= nil then
			local targetName = _target.object:getName()
			if ATMEDatas.allAIUnitsAndPlayers[targetName] ~= nil then
				-- Test si la target est bien une unité ou un joueur
				
				if _localMembers.targets[targetName] ~= nil then
					_localMembers.targets[targetName].stillDetected = true
				else
					_localMembers.targets[targetName] = {
						stillDetected = true,
						trackingOn = false,
						confirmed = 0,
					}
				end
	
				nb = nb + 1
			end
		end
	end
	
	return _localMembers.targets, nb
end

ATMEDatas.C_AIUnit_MT = {
	className = "C_AIUnit",
	parentClass = ATME.C_Unit,
	mtModules = {},
	mt = {},
	
	keys = {
		["setVoice"] = true,
		["getVoice"] = true,
		["getPosition"] = true,
		["getVelocityVector"] = true,
		["getRollAxisVector"] = true,
		["getIndex"] = true,
		["getDCSUnit"] = true,
		["getDCSID"] = true,
		["getName"] = true,
		["getAirbase"] = true,
		["getAttributes"] = true,
		["hasAttributes"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getTypeName"] = true,
		["getGroup"] = true,
		["disable"] = true,
		["getFuelRatio"] = true,
		["getLife"] = true,
		["getLifeRatio"] = true,
		["getCallsign"] = true,
		["getSpeed"] = true,
		["getHSpeed"] = true,
		["getVSpeed"] = true,
		["getSpeedAzimuth"] = true,
		["getAzimuth"] = true,
		["getCountryName"] = true,
		["getCoalitionName"] = true,
		["getCategoryName"] = true,
		["isActivated"] = true,
		["isHelicopter"] = true,
		["isPlane"] = true,
		["isPersonnelCarrier"] = true,
		["isInfantry"] = true,
		["isGroundVehicle"] = true,
		["isAirDefence"] = true, 
		["isManPad"] = true,
		["isSAMVehicle"] = true,
		["isSAMSiteLauncher"] = true,
		["isSAMSiteRadar"] = true,
		["isSAMSiteCommandCenter"] = true,
		["isAAA"] = true,
		["inAir"] = true,
		["getMSLAltitude"] = true,  
		["getAGLAltitude"] = true,
		["explode"] = true,
		["isInZones"] = true,
		["isInDCSZone"] = true,
		["isInZone2D"] = true,
		["isInZone3D"] = true,
		["isPilotDead"] = true,
		["isPilotEjected"] = true,
		["isEngineStarted"] = true,
		["whichSide"] = true,
		["getNbUnitsOnBoard"] = true,
		["getOnBoardGroupNames"] = true,
		["load"] = true,
		["unload"] = true,
		["getNearestReadyToBoardGroup"] = true,
		["fireFlare"] = true,
		["fireSmoke"] = true,
		["fireIlluminationBomb"] = true,
		["modules"] = true,
		["canFly"] = true,
		["getUserPrivateDatas"] = true,
		["spotActivate"] = true,
		["spotDesactivate"] = true,
		["spotWhichIsActivated"] = true,
		["spotGetDatas"] = true,
		["spotIgnoreTarget"] = true,
		["assignTarget"] = true,
		["sensorStart"] = true,
		["sensorStop"] = true,
		["beaconOn"] = true,
		["beaconOff"] = true,
		["setFrequency"] = true,
		["addEventCallback"] = true,
		["removeEventCallback"] = true,
	},

	add = function(self, _name, _unit)
		local item = {
			name = _name,
			objectDCS = _unit,
			objectDCSId = _unit:getID(),
			engineOn = false,
			nbUnitsOnBoard = 0,
			groupsBoarding = {},
			spotDatas = {},
			onBoardGroupDatas = ATMEFct.C_IndexList_create(),
			isEnabled = true,
			currentVoice = "",
			userPrivateDatas = {},						-- Point d'entrée des données privées : un par module
			eventCallbacks = {},						-- Callbacks d'événement
			targets = {},								-- Table des cibles trackées à un instant T
			tagsName = ATMEFct.decompString(_name),
			pilotDead = false,
			pilotEjected = false,
			destroyed = false,							-- Utile lors du delete (car isExist est déjà false)
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
		ATMEDatas.localMembers[_id].eventCallbacks = nil
		ATMEDatas.localMembers[_id].targets = nil
	end,
}

ATME.C_AIUnit = {
	getClassName = function(self)
		return ATMEDatas.C_AIUnit_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_AIUnit_MT.parentClass
	end,
	
	getByName = function(_unitName)
		if type(_unitName) == "string" then
			return ATMEDatas.AIUnits[_unitName]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_AIUnit.getByName(name)", "name type must be string", false)
		end	
	end,
}

ATMEFct.C_AIUnit_create = function(_unit, _datas, _newObject)
	local newUnit = {}
	
	setmetatable(newUnit, ATMEDatas.C_AIUnit_MT.mt)
	local index = ATMEDatas.C_AIUnit_MT:add(_unit:getName(), _unit)
	local localMembers = ATMEDatas.localMembers[index]
	
	local typeName = _unit:getTypeName()
	local oldSide
	
	local groupName = _unit:getGroup():getName()
	local linkedAirbase = ATME.C_Airbase.getByName(localMembers.name, "SHIP")
	
	function newUnit:getIndex(self)
		return index
	end
	
	function newUnit:getAirbase()
		if localMembers.destroyed == false then
			return linkedAirbase
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.getDCSUnit(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.getDCSID(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCSId
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.getName(self)
		if localMembers.destroyed == false then
			return localMembers.name
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.isActivated(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:isActive()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.getCategoryName(self)
		if localMembers.destroyed == false then
			return localMembers.category
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "tagGetByID(position)"
		
		if localMembers.destroyed == false then
			if type(_position) == "number" then
				local pos = math.floor(_position)
				if localMembers.tagsName ~= nil and pos >= 1 and pos <= #localMembers.tagsName then
					return localMembers.tagsName[pos]
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.tagGetCount(self)
		if localMembers.destroyed == false then
			if tagsName ~= nil then
				return #tagsName
			else
				return 0
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "tagCompare(tagsToCompare)"
		
		if localMembers.destroyed == false then
			if type(_tagsToCompare) == "string" then
				if localMembers.tagsName ~= nil then
					local ok = ATMEFct.compareTags(localMembers.tagsName, _tagsToCompare)
					
					if ok ~= nil then
						return ok
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
					end
				else
					return false
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.getTypeName(self)
		if localMembers.destroyed == false then
			return typeName
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.getGroup(self)
		if localMembers.destroyed == false then
			return localMembers.group
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.getUserPrivateDatas(self)
		if localMembers.destroyed == false then
			return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.getAttributes(self)
		if localMembers.destroyed == false then
			local datas = Unit.getDesc(localMembers.objectDCS)
			local attributes = {}
			for _id, _ in pairs(datas.attributes) do
				if ATMEConst.backTargetsAttributes[_id] ~= nil then
					table.insert(attributes, ATMEConst.backTargetsAttributes[_id].value)
				end
			end
			
			return attributes
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.isEngineStarted(self)
		if localMembers.destroyed == false then
			local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
			if ATMEConst.canFly[localMembersGroup.category] == true then
				return localMembers.engineOn
			else
				return true
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.hasAttributes(self, _attributes)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "hasAttributes(attributes)"
		
		if localMembers.destroyed == false then
			if type(_attributes) == "string" or type(_attributes) == "table" then
				local err, datas = ATMEFct.extractDatasFromStringOrTable(_attributes, false)
		
				if err == false then
					return ATMEFct.hasAttributes(localMembers.objectDCS, datas)
				end
				
				ATMEFct.errorOutputCore(nil, fctDefinition, datas, false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "attributes type must be string or table", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.addEventCallback(self, _fct, _eventType)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "addEventCallback(fct, [eventType])"
		
		if localMembers.destroyed == false then
			if type(_fct) == "function" and (_eventType == nil or type(_eventType) == "string") then
				if _eventType == nil then
					for _typeName, _ in pairs(ATMEConst.coreEventTypes) do
						localMembers.eventCallbacks[ATMEDatas.currentModuleName][_typeName] = _fct
					end
				else
					local eventType = string.upper(_eventType)
					if ATMEConst.coreEventTypes[eventType] ~= nil then
						localMembers.eventCallbacks[ATMEDatas.currentModuleName][eventType] = _fct
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "eventType must be an existing event type (see manual)", false)
					end
				end
			else
				if type(_fct) ~= "function" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "fct type must be function", true)
				elseif type(_eventType) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "eventType type must be nil or string", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.removeEventCallback(self, _eventType)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "removeEventCallback([eventType])"
		
		if localMembers.destroyed == false then
			if _eventType == nil or type(_eventType) == "string" then
				if _eventType == nil then
					for _typeName, _ in pairs(ATMEConst.coreEventTypes) do
						localMembers.eventCallbacks[ATMEDatas.currentModuleName][_typeName] = nil
					end
				else
					local eventType = string.upper(_eventType)
					if ATMEConst.coreEventTypes[eventType] ~= nil then
						localMembers.eventCallbacks[ATMEDatas.currentModuleName][eventType] = nil
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "eventType must be an existing event type (see manual)", false)
					end
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "eventType type must be nil or string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.assignTarget(_target)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "assignTarget(target)"
		
		if localMembers.destroyed == false then
			if type(_target) == "string" or ATMEFct.isObjectClass(_target, ATME.C_Unit) == true then
				local unit = _target
				if type(_target) == "string" then
					unit = ATMEDatas.allAIUnitsAndPlayers[_target]
				end
				
				if unit ~= nil then
					local controller = ATMEDatas.localMembers[_ATMEUnit:getIndex()].objectDCS:getController()
					local localMembersTarget = ATMEDatas.localMembers[unit:getIndex()]
					
					controller:knowTarget(localMembersTarget.objectDCS, true, true)
				else
					ATMEFct.setError("NO_TARGET")
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "target type must be string or ATME.C_AIUnit or ATME.C_Player", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.sensorStart(self, _typeOfSensorName, _flag)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "sensorStart(typeOfSensorName, [flag])"
		
		if localMembers.destroyed == false then
			if type(_typeOfSensorName) == "string" and (type(_flag) == "number" or type(_flag) == "string" or _flag == nil) then
				if ATMEConst.typesOfSensor[_typeOfSensorName] ~= nil then
					if localMembers.targetsTracking == nil then
						localMembers.targetsTracking = {
							flag = _flag,
							unit = self,
						}
						if _typeOfSensorName ~= "ALL" then
							datas.typeOfSensorName = _typeOfSensorName
						end
						
						localMembers.targetsTracking.idTimer = timer.scheduleFunction(ATMETimers.C_Unit_targetTracking, localMembers, timer.getTime() + 1)
					else
						setError("ALREADY_TRACKING")
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "typeOfSensorName must be 'ALL', 'RADAR', 'VISUAL', 'OPTIC', 'IRST', 'RWR', 'DLINK'", false)
				end
			else
				if type(_typeOfSensorName) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "typeOfSensorName type must be string", true)
				elseif type(_flag) ~= "number" and type(_flag) ~= "string" and _flag ~= nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "flag type must be nil or string or number", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.sensorStop(self)
		if localMembers.destroyed == false then
			if localMembers.targetsTracking ~= nil then
				if localMembers.targetsTracking.idTimer ~= nil then
					timer.removeFunction(localMembers.targetsTracking.idTimer)
				end
			
				localMembers.targetsTracking = nil
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.spotActivate(self, _target, _code)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "spotActivate(target, [code])"
		
		if localMembers.destroyed == false then
			if (ATMEFct.isObjectClass(_target, ATME.C_Unit) == true or ATMEFct.isObjectClass(_target, ATME.C_StaticObject) == true or
					ATMEFct.isObjectClass(_target, ATME.C_Group) == true or ATMEFct.isObjectClass(_target, ATME.C_Zone2D) == true or 
					type(_target) == "string" or ATME.isPoint(_target) ~= POINT_BAD) and (_code == nil or type(_code) == "number") then
					 
				local point
				local targetGroupName
				local dynamicSpot = ATMEFct.isObjectClass(_target, ATME.C_StaticObject) == false and ATME.isPoint(_target) == POINT_BAD
				local datas
				local targetName
				local targetClassName
				local zone2D
				local ignoredTargets = {}
				
				if ATMEFct.isObjectClass(_target, ATME.C_StaticObject) == true or ATMEFct.isObjectClass(_target, ATME.C_Unit) == true then
					if ATMEFct.isObjectClass(_target, ATME.C_StaticObject) == true or ATMEFct.isTargetOk[_target:getClassName()](_target, ignoredTargets) == true then
						targetName = _target:getName()
						targetClassName = _target:getClassName()
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "reference must be ground unit or aircraft unit on floor", false)
						return
					end
				elseif ATMEFct.isObjectClass(_target, ATME.C_Group) == true then
					if _target:getCategoryName() ~= "SHIP" then
						if self:getCoalitionName() ~= _target:getCoalitionName() then
							targetName, targetClassName = ATMEFct.C_AIUnit_findTargetInGroup(_target, ignoredTargets)
							targetGroupName = _target:getName()
							if targetName == nil then
								-- Aucune unité trouvée. Pas d'activation
								
								return false
							end
						else
							return false
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "reference must be ground group or aircraft group", false)
						return
					end
				elseif type(_target) == "string" then
					local zone = ATMEFct.getDCSZone(_target)
					if zone ~= nil then
						zone2D = ATMEFct.C_Circle_create(nil, nil, zone.point, zone.radius)
						targetName, targetClassName = ATMEFct.C_AIUnit_findTargetInZone2D(zone2D, ATMEConst.coalitionEnemy[self:getCoalitionName()], ignoredTargets)
						
						if targetName == nil then
							-- Aucune unité trouvée. Pas d'activation

							return false
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _target .. " unknown", false)
						return
					end
				elseif ATMEFct.isObjectClass(_target, ATME.C_Zone2D) == true then
					targetName, targetClassName = ATMEFct.C_AIUnit_findTargetInZone2D(_target, ATMEConst.coalitionEnemy[self:getCoalitionName()], ignoredTargets)
					zone2D = _target
					
					if targetName == nil then
						-- Aucune unité trouvée. Pas d'activation
						
						return false
					end
				else
					point = ATMEFct.convertToPoint3D(_target)
				end
				
				-- Lasing à partir d'un drone en vol
				local lasingInAir = false
				local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
				if ATMEConst.canFly[localMembersGroup.category] == true and self:inAir() == true and ATMEConst.aircraftTypeJTAC[typeName] == true then
					lasingInAir = true
				end
				
				if localMembers.category == "GROUND" or lasingInAir == true then
					local code

					if _code ~= nil then
						code = math.floor(_code)
						if code < 1111 or code > 1788 then
							ATMEFct.errorOutputCore(nil, fctDefinition, "code must be >= 1111 and <= 1788", false)
							return
						end
					end
					
					if (self:getSpeed() == 0 and lasingInAir == false) or lasingInAir == true then
						if localMembers.spotDatas.spot ~= nil then
							localMembers.spotDatas.spot:destroy()
							localMembers.spotDatas.spot = nil
						end
						
						localMembers.spotDatas.targetName = targetName
						localMembers.spotDatas.targetClassName = targetClassName
						localMembers.spotDatas.groupName = targetGroupName
						localMembers.spotDatas.zone = zone2D
						localMembers.spotDatas.code = code
						localMembers.spotDatas.dynamicSpot = dynamicSpot
						localMembers.spotDatas.noLOS = 0
						localMembers.spotDatas.neverHaveLOS = true
						localMembers.spotDatas.ignoredTargets = ignoredTargets
						localMembers.spotDatas.ignoredTargetOn = false
						localMembers.spotDatas.fromAir = lasingInAir
						
						timer.scheduleFunction(ATMETimers.C_Unit_spot_timer, localMembers, timer.getTime() + 1)
						return true
					else
						return false
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "current unit must be ground unit or drone unit in air", false)
				end
			else
				if ATMEFct.isObjectClass(_target, ATME.C_Unit) == false and ATMEFct.isObjectClass(_target, ATME.C_StaticObject) == false and
						ATMEFct.isObjectClass(_target, ATME.C_Group) == false and ATMEFct.isObjectClass(_target, ATME.C_Zone2D) == false and 
						type(_target) ~= "string" and ATME.isPoint(_target) == POINT_BAD then
					ATMEFct.errorOutputCore(nil, fctDefinition, "target type must be string, ATME.C_Zone2D, ATME.C_Group, ATME.C_AIUnit, ATME.C_Player, ATME.C_StaticObject or table (x,y)/(x,y,z)", true)
				elseif _code ~= nil and type(_code) ~= "number" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "code type must be nil or number", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.spotDesactivate(self)
		if localMembers.destroyed == false then
			if localMembers.spotDatas.spot ~= nil then
				localMembers.spotDatas.spot.spot:destroy()
				localMembers.spotDatas.spot = nil
				ATMEDatas.spotsToReassign[localMembers.name] = nil
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.spotIgnoreTarget(self)
		if localMembers.destroyed == false then
			if localMembers.spotDatas.spot ~= nil then
				if localMembers.spotDatas.targetName ~= nil then
					localMembers.spotDatas.ignoredTargets[localMembers.spotDatas.targetName] = true
					localMembers.spotDatas.ignoredTargetOn = true
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.spotWhichIsActivated(self)
		if localMembers.destroyed == false then
			if localMembers.spotDatas.spot ~= nil then
				if localMembers.spotDatas.code ~= nil then
					return "LASER"
				else
					return  "IR"
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.spotGetDatas(self)
		if localMembers.destroyed == false then
			if localMembers.spotDatas.spot ~= nil then
				return {
					targetName = localMembers.spotDatas.targetName,
					isAIUnit = localMembers.spotDatas.targetName ~= nil and localMembers.spotDatas.dynamicSpot == true,
					isStaticObject = localMembers.spotDatas.targetName ~= nil and localMembers.spotDatas.dynamicSpot == false,
					isFixedPoint = localMembers.spotDatas.targetName == nil,
					code = localMembers.spotDatas.code,
					vector = localMembers.spotDatas.vector,
				}
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.disable(self)
		if localMembers.destroyed == false then
			ATMEDatas.unitsToDisable[localMembers.name] = self
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.getFuelRatio(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getFuel()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.getLife(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getLife()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.getLifeRatio(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getLife()/localMembers.objectDCS:getLife0()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.getCallsign(self)
		if localMembers.destroyed == false then
			local identification
			local id1
			local id2
			
			if localMembers.objectDCS:hasAttribute("Helicopters") or localMembers.objectDCS:hasAttribute("Planes") then
				local callsign = localMembers.objectDCS:getCallsign()
				if localMembers.objectDCS:getCoalition() == 2 then
					identification = string.match(callsign, "^%a*")
					id1 = string.match(callsign, "%d")
					id2 = string.match(callsign, "%d$")
				else
					identification = callsign
				end
			end
			
			return identification, id1, id2
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.getPosition(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getPoint()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.getVelocityVector(self)
		if localMembers.destroyed == false then
			local vector = ATMEFct.C_Vector3D_create(nil, nil, localMembers.objectDCS:getVelocity())
			return vector
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.getSpeed(self)
		if localMembers.destroyed == false then
			local vector = self:getVelocityVector()
			return vector:getModule(vector)
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.getHSpeed(self)
		if localMembers.destroyed == false then
			local vector = self:getVelocityVector()
			return vector:getHModule(vector)
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.getVSpeed(self)
		if localMembers.destroyed == false then
			local vector = localMembers.objectDCS:getVelocity()
			return vector.y
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.getSpeedAzimuth(self)
		if localMembers.destroyed == false then
			local vector = self:getVelocityVector()
				
			if vector:getModule() > 1 then 
				return vector:getAzimuth()
			else
				return 0
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.getRollAxisVector(self)
		if localMembers.destroyed == false then
			return ATMEFct.C_Vector3D_create(nil, nil, localMembers.objectDCS:getPosition().x)
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.getAzimuth(self)
		if localMembers.destroyed == false then
			local vector = self:getRollAxisVector()
			return vector:getAzimuth()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.getCountryName(self)
		if localMembers.destroyed == false then
			local id = localMembers.objectDCS:getCountry()
			return country.name[id]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.getCoalitionName(self)
		if localMembers.destroyed == false then
			local id = localMembers.objectDCS:getCoalition()
			return ATMEConst.coalitionNames[id]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.isHelicopter(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("Helicopters")
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.isPersonnelCarrier(self)
		if localMembers.destroyed == false then
			return ATMEConst.personnelCarrierAttributes[self:getTypeName()] ~= nil
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end


	function newUnit.isPlane(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("Planes")
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.canFly(self)
		if localMembers.destroyed == false then
			local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
			return ATMEConst.canFly[localMembersGroup.category]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.isInfantry(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("Infantry")
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.isGroundVehicle(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("Ground vehicles")
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.isAirDefence(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("Air Defence")
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.isManPad(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("MANPADS")
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.isSAMVehicle(self, _range)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "isSAMVehicle([range])"
		
		if localMembers.destroyed == false then
			if _range == nil or type(_range) == "string" then
				if _range == nil then
					return localMembers.objectDCS:hasAttribute("SR SAM") or self.localMembers.objectDCS:hasAttribute("MR SAM") or self.localMembers.objectDCS:hasAttribute("LR SAM")
				elseif _range == "SHORT" then
					return localMembers.objectDCS:hasAttribute("SR SAM")
				elseif _range == "MEDIUM" then
					return self.localMembers.objectDCS:hasAttribute("MR SAM")
				elseif _range == "LONG" then
					return self.localMembers.objectDCS:hasAttribute("LR SAM")
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "range bad value (SHORT, MEDIUM, LONG)", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "range type must be nil or string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.isSAMSiteLauncher(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("SAM LL")
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.isSAMSiteRadar(self, _category)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "isSAMSiteRadar([category])"
		
		if localMembers.destroyed == false then
			if _category == nil or type(_category) == "string" then
				if _category== nil then
					return localMembers.objectDCS:hasAttribute("SAM SR") or localMembers.objectDCS:hasAttribute("SAM TR")
				elseif _category == "SURFACE" then
					return localMembers.objectDCS:hasAttribute("SAM SR")
				elseif _category == "TRACKING" then
					return self.localMembers.objectDCS:hasAttribute("SAM TR")
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "category bad value (SURFACE, TRACKING)", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "category type must be nil or string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.isSAMSiteCommandCenter(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("SAM CC")
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.isAAA(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:hasAttribute("AAA")
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end	

	function newUnit.inAir(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:inAir()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.getMSLAltitude(self)
		if localMembers.destroyed == false then
			return ATME.getPointAltitude(self:getPosition())
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.getAGLAltitude(self)
		if localMembers.destroyed == false then
			local landHeight = ATMEFct.getSurfaceAltitude(self:getPosition())
			return ATME.getPointAltitude(self:getPosition()) - landHeight
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.explode(self, _power)
		if localMembers.destroyed == false then
			local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "explode(power)"
			
			if type(_power) == "number" then
				if _power >= 1 then
					ATME.explode(self:getPosition(), _power)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.isInZones(self, _zonesMgr)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "isInZones(zonesMgr)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zonesMgr, ATME.C_ZonesMgr) == true then
				return _zonesMgr:isInside(self) 
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zonesMgr type must be ATME.C_ZonesMgr", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end				

	function newUnit.isInDCSZone(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "isInDCSZone(name)"
		
		if localMembers.destroyed == false then
			if type(_name) == "string" then
				local zone = ATMEFct.getDCSZone(_name)
				
				if zone ~= nil then
					local vector = ATMEFct.C_Vector3D_create(nil, nil, self:getPosition(), zone.point)
					if vector:getHModule() <= zone.radius then
						return true
					else
						return false
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _name .. " unknown", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.isInZone2D(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "isInZone2D(zone)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zone, ATME.C_Zone2D) == true then
				local point = self:getPosition()
				local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
				
				local datas = {
					zone = _zone,
					actualCenter = localMembersZone.center,
					azimuth = localMembersZone.azimuth,
				}
						
				return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](point, datas)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Circle, ATME.C_PickupZone, ATME.C_Rectangle or ATME.C_Polygon", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.isInZone3D(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "isInZone3D(zone)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zone, ATME.C_Zone3D) == true then
				local point = self:getPosition()
				local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
		
				local datas = {
					zone = _zone,
					actualCenter = localMembersZone.center,
					azimuth = localMembersZone.azimuth,
				}
				
				return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](point, datas)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Sphere, ATME.C_Cylinder or ATME.C_Box", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.isPilotDead(self)
		if localMembers.destroyed == false then
			return localMembers.pilotDead
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.isPilotEjected(self)
		if localMembers.destroyed == false then
			return localMembers.pilotEjected
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.whichSide(self, _reference)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "whichSide(reference)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isReference(_reference) == true  then
				local point = ATMEFct.getPointFromReference(_reference)

				return ATMEFct.C_Player_C_AIUnit_whichSide(self, point)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "reference type must be" .. ATMEConst.errorLabelReference, true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.getNbUnitsOnBoard(self)
		if localMembers.destroyed == false then
			return localMembers.nbUnitsOnBoard
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.getOnBoardGroupNames(self)
		if localMembers.destroyed == false then
			local temp = {}
			
			for _id, _item in localMembers.onBoardGroupDatas:pairs() do
				temp[_item.name] = _id
			end
			
			return temp
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.unload(self, _param1, _param2)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "unload([whenStopped])\nunload(id, [whenStopped])"

		if localMembers.destroyed == false then	
			if (_param1 == nil and _param2 == nil) or type(_param1) == "boolean" then
				local whenStopped = false
				 if type(_param1) == "boolean" then
					whenStopped = _param1
				end
				
				ATMEFct.C_Player_C_AIUnit_unload(localMembers, nil, whenStopped)
			elseif type(_param1) == "number" and (_param2 == nil or type(_param2) == "boolean") then
				local id = math.floor(_param1)
				
				local whenStopped = _param2
				
				if _param2 == nil then
					whenStopped = false
				end
				
				ATMEFct.C_Player_C_AIUnit_unload(localMembers, id, whenStopped)
			else
				if type(_param1) ~= "boolean" and type(_param1) ~= "number" and _param2 == nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "whenStopped must be nil or boolean\nid type must be number", true)
				else
					if type(_param1) ~= "number" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "id type must be number", true)
					elseif _param2 ~= nil and type(_param2) ~= "boolean" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "whenStopped type must be nil or boolean", false)
					end
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.load(self, _radius, _groupToBoard)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "load(radius, [groupToBoard])"
		
		if localMembers.destroyed == false then
			if type(_radius) == "number" and (_groupToBoard == nil or ATMEFct.isObjectClass(_groupToBoard, ATME.C_Group) == true or type(_groupToBoard) == "string") then
				if self:isActivated() == true then
					local groupToBoard
					
					if _groupToBoard == nil then
						groupToBoard = ATMEFct.C_Player_C_AIUnit_getNearestReadyToBoardGroup(self, _radius)
					elseif type(_groupToBoard) == "string" then
						groupToBoard = ATME.C_Group.getByName(_groupToBoard)
					else
						groupToBoard = _groupToBoard
					end
					
					if groupToBoard ~= nil then
						if groupToBoard:isOnlyInfantry() == true then
							if _radius > 0 then
								if groupToBoard:isReadyToBoard() == true and groupToBoard:isBoarding() == false then
									if ATMEConst.personnelCarrierAttributes[self:getTypeName()] ~= nil then
										return ATMEFct.C_Player_C_AIUnit_load(self, groupToBoard, _radius)
									else
										ATMEFct.errorOutputCore(nil, fctDefinition, self:getName() .. " not a transport unit", false)
									end
								else
									ATMEFct.setError("INFANTRY_GROUP_NOT_READY")
								end
							else
								ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Not infantry group", false)
						end
					else
						ATMEFct.setError("NO_GROUP_FOUND")
					end
				else
					ATMEFct.setError("UNACTIVATED_UNIT")
				end
			else
				if type(_radius) ~= "number" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "radius type must be number", true)
				elseif ATMEFct.isObjectClass(_groupToBoard, ATME.C_Group) == false and _groupToBoard ~= nil then
					ATMEFct.errorOutputCore(nil, fctDefinition, "groupToBoard type must be nil or ATME.C_Group", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.getNearestReadyToBoardGroup(self, _radius)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "getNearestReadyToBoardGroup(radius)"
		
		if localMembers.destroyed == false then
			if type(_radius) == "number" then
				if _radius >= 0 then
					return ATMEFct.C_Player_C_AIUnit_getNearestReadyToBoardGroup(self, _radius)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be > 0", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "radius must be number", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.fireFlare(self, _color)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "fireFlare(color)"
		
		if localMembers.destroyed == false then
			if type(_color) == "string" then
				if ATMEConst.flareColors[string.upper(_color)]~= nil then
					if localMembers.category == "GROUND" then
						if localMembers.flare == nil then
							localMembers.flare = ATMEFct.C_Flare_create(_color, self:getPosition())
						else
							ATMEDatas.localMembers[localMembers.flare:getIndex()].point = self:getPosition()
						end
						localMembers.flare:activate()
					else
						ATMEFct.setError("NOT_GROUND_UNIT")
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a flare color or RANDOM", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.fireSmoke(self, _color)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "fireSmoke(color)"
		
		if localMembers.destroyed == false then
			if type(_color) == "string" then
				if ATMEConst.smokeColors[string.upper(_color)] ~= nil then 
					if localMembers.category == "GROUND" then
						local point = self:getPosition()
						local newPoint = {
							x = point.x + math.random(-20,20),
							y = point.y,
							z = point.z + math.random(-20,20),
						}
						
						local smoke = ATMEFct.C_Smoke_create(_color, newPoint)
						smoke:activate(false)
					else
						ATMEFct.setError("NOT_GROUND_UNIT")
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a smoke color", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.fireIlluminationBomb(self, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "fireIlluminationBomb(power)"
		
		if localMembers.destroyed == false then
			if type(_power) == "number" then
				if _power >= 1 then
					if localMembers.category == "GROUND" then
						local point = self:getPosition()
						
						point.y = point.y + 200 + math.random(0, 100)

						trigger.action.illuminationBomb(point, _power)
					else
						ATMEFct.setError("NOT_GROUND_UNIT")
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.setVoice(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "setVoice([name])"
		
		if localMembers.destroyed == false then
			if _name == nil or type(_name) == "string" then
				if _name ~= nil then
					local d, f, temp = _name:find("^(%a[%a%d_]*)")
					
					if temp == _name then
						localMembers.currentVoice = _name
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "name bad format", false)
					end
				else
					localMembers.currentVoice = ""
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be nil or string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.getVoice(self)
		if localMembers.destroyed == false then
			if localMembers.currentVoice ~= "" then
				return localMembers.currentVoice
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newUnit.beaconOn(self, _settings)
		-- BeaconSites.lua dans DCS World OpenBeta\Scripts\World\Radio : définition des types
		
		if localMembers.destroyed == false then
			local fcts = {
				["TACAN"] = function(_tokens)
					if _tokens[2].type == "TOKEN_TACAN_CHANNEL" then
						local mode = string.sub(string.upper(_tokens[2].value), -1)
						local channel = tonumber(string.sub(_tokens[2].value, 1, _tokens[2].value:len() - 1))
		
						local typeValue = 4
						local systemValue = 3
						local frequency
						
						if _tokens[1].type == "TOKEN_TYPE_TACAN_AA" then
							if localMembers.group:canFly() == true then
								channel = channel + 63

								if ATMEConst.specialAircraftTypes[self:getTypeName()] ~= nil then
									--Tanker
									
									typeValue = 4
									systemValue = 4 + ATMEConst.tacanMode[mode]
								else
									typeValue = 4
									systemValue = 13 + ATMEConst.tacanMode[mode]
								end
							else
								return true, "AA Tacan must be set on HELICOPTER or AIRPLAINE groups only"
							end
							
							frequency = (ATMEConst.tacanFrequencyOffset["Xmit"][mode] + channel) * 1000000
						else
							if mode == 'Y' then -- TACAN non AA limitation à mode X. Mode Y fonctionne mais pas de bearing pour l'instant.
								return true, "Tacan channel mode must be X only if not AA TACAN"
							end
							
							frequency = (ATMEConst.tacanFrequencyOffset["Recv"][mode] + channel) * 1000000
						end
					
						local beaconCallsign
						if #_tokens == 3 and _tokens[3].type == "TOKEN_STRING" then
							beaconCallsign = string.upper(_tokens[3].value)
						end
						
						local controller = ATMEDatas.localMembers[localMembers.group:getIndex()].objectDCS:getController()
						local command = { 
							id = 'ActivateBeacon', 
							params = { 
								type = typeValue, 
								system = systemValue, 
								callsign = beaconCallsign, 
								frequency = frequency,
							}
						}
						controller:setCommand(command)
						
						return false
					else
						return true, "Tacan channel must be >=1 and <=63, X or Y if AA TACAN, X only if not AA TACAN"
					end
				end,
				["ICLS"] = function(_tokens)
					if _tokens[2].type == "TOKEN_INTEGER" then
						local controller = ATMEDatas.localMembers[localMembers.group:getIndex()].objectDCS:getController()
						local frequency = 15415000000 + 30000000 * ((_tokens[2].value - 1) % 10)
						
						local command = { 
							id = 'ActivateBeacon', 
							params = {
								type = 131328,
								system = 16,
								frequency = frequency,
							}
						}
						controller:setCommand(command)
						
						command = { 
							id = 'ActivateBeacon', 
							params = {
								type = 131584,
								system = 17,
								frequency = frequency,
							}
						}
						controller:setCommand(command)
						
						return false
					else
						return true, "ICLS channel type must be integer"
					end
				end,
			}
			
			local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "beaconOn(settings)"
			
			if type(_settings) == "string" then
				local tokens = ATMEFct.analyzeString(_settings)
				
				if #tokens == 2 or #tokens == 3 then
					local beaconType = ATMEConst.beaconTypes[tokens[1].type]
					
					
					if fcts[beaconType] ~= nil then
						local err, errorDesc = fcts[beaconType](tokens)
						
						if err == true then
							ATMEFct.errorOutputCore(nil, fctDefinition, errorDesc, false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "settings must be 'TACAN', 'TACAN_AA' or 'ICLS'", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "settings must have 2 or 3 parameters", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "settings type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.beaconOff(self)
		if localMembers.destroyed == false then
			local controller = ATMEDatas.localMembers[localMembers.group:getIndex()].objectDCS:getController()
			local command = { 
				id = "DeactivateBeacon", 
				params = {},
			}
			
			controller:setCommand(command)
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newUnit.setFrequency(self, _modulation, _frequency)
		local fctDefinition = ATMEFct.getErrorPrefix(newUnit) .. "setFrequency(modulation, frequency)"

		if localMembers.destroyed == false then
			if type(_modulation) == "string" and type(_frequency) == "number" then
				if localMembers.category ~= "SHIP" then -- Bug DCS avec les navires -> Plantage
					local modulationValue = ATMEConst.modulationValues[_modulation]
					if modulationValue ~= nil then
						local frequencyValue = math.floor(_frequency)
						if frequencyValue > 0 then
							local controller = ATMEDatas.localMembers[localMembers.group:getIndex()].objectDCS:getController()

							local command = {
								id = "SetFrequency", 
								params = { 
									frequency = frequencyValue,
									modulation = modulationValue,							
								} 
							}
							
							controller:setCommand(command)
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "frequency type must > 0", false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "modulation must be 'AM' or 'FM'", true)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "unit can't be a ship", false)
				end
			else
				if type(_modulation) ~= "string" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "modulation type must be string", true)
				elseif type(_frequency) ~= "number" then
					ATMEFct.errorOutputCore(nil, fctDefinition, "frequency type must be number", false)
				end
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	newUnit.modules = {}
	setmetatable(newUnit.modules, ATMEDatas.C_AIUnit_MT.mtModules)
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
		localMembers.eventCallbacks[_localModuleMembers.name] = {}
		newUnit.modules[_localModuleMembers.name] = {}
	end
	
	if ATMEConst.personnelCarrierAttributes[newUnit:getTypeName()] ~= nil and newUnit:isHelicopter() == true then
		localMembers.indexPickupZoneSignal = ATMEDatas.signalCapabilityUnits:add({ 
																					object = newUnit,
																					localMembers = localMembers})
	end

	if ATMEDatas.newCreatedGroups[groupName] == nil then
		ATMEFct.C_Group_create(newUnit, _datas, _newObject)
	else
		ATMEFct.C_Group_addUnit(ATMEDatas.newCreatedGroups[groupName].group, newUnit)
	end

	localMembers.group = ATMEDatas.newCreatedGroups[groupName].group
	
	local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
	if localMembersGroup.waypoints[1].airdromeId ~= nil or localMembersGroup.waypoints[1].helipadId ~= nil then
		-- Bug ou changement de gestion des airbases ID
		if localMembersGroup.waypoints[1].airdromeId ~= nil then
			airbase = ATMEDatas.airbasesById[localMembersGroup.waypoints[1].airdromeId]:getName()
		elseif localMembersGroup.waypoints[1].helipadId ~= nil then
			airbase = ATMEDatas.airbasesById[localMembersGroup.waypoints[1].helipadId]:getName()
		end
		
		if localMembersGroup.waypoints[1].type == "TakeOffParkingHot" then
			-- Au parking moteur allumé
			
			localMembers.engineOn = true
		elseif localMembersGroup.waypoints[1].type == "TakeOff" then
			-- Sur la piste, prêt à décoler
			
			localMembers.engineOn = true
		end
	else
		-- En vol....
		
		localMembers.engineOn = true
	end
	
	localMembers.category = localMembersGroup.category
	
	ATMEFct.outputCore(nil,"Prepare unit = " .. localMembers.name .. " -> ok", 2)
end

ATMEFct.C_AIUnit_delete = function(_AIUnit)
	local localMembers = ATMEDatas.localMembers[_AIUnit:getIndex()]
	
	local flag = "DISABLED"
	if localMembers.isEnabled == true then
		flag = "EJECTION"
		if _AIUnit:isPilotDead() == true then
			flag = "DEAD"
		end
	end
	
	local localMembersGroup = ATMEDatas.localMembers[localMembers.group:getIndex()]
	for _i = #ATMEDatas.modulesNames, 1, -1 do
		local localModuleMembers = ATMEDatas.modulesLocalMembers[ATMEDatas.modulesNames[_i]]
		if localMembersGroup.onBoard == false then
			-- Les unités d'un groupe chargé dans un autre véhicule ne génèrent pas 
			-- d'appel à onDeleteAIUnitHandler
			
			if localModuleMembers.handlers.onDeleteAIUnitHandler ~= nil then
				ATMEDatas.currentModuleName = localModuleMembers.name
				
				-- Appel du module
				ATMEFct.outputCore(nil,"Appel onDeleteAIUnitHandler du module " .. localModuleMembers.name, 2)
				xpcall(function() localModuleMembers.handlers.onDeleteAIUnitHandler(_AIUnit, flag) end, function() ATMEFct.errUserHandler('onDeleteAIUnitHandler') end)
			end
		end
	end
	
	if localMembers.indexPickupZoneSignal ~= nil then
		ATMEDatas.signalCapabilityUnits:remove(localMembers.indexPickupZoneSignal)
		localMembers.indexPickupZoneSignal = nil
	end
	
	_AIUnit.modules = nil

	for _, _item in localMembers.onBoardGroupDatas:pairs() do
		-- Si un des groupes embarqués a un respawn automatique il est préparé
		local groupName = _item.groupDatas:getName()
		
		if _item.autoRespawn.activated == true then
			local respawnTime = _item.autoRespawn.time
			if _item.autoRespawn.random == true then
				respawnTime = math.random(_item.autoRespawn.minTime, _item.autoRespawn.maxTime)
			end
							
			ATMEDatas.groupsToRespawn[groupName] = {
				time = respawnTime + timer.getTime(),
				spawnDatas = _item.spawnDatas,
			}
		end
		
		ATMEDatas.groupsOnBoard[groupName] = nil
	end
	
	localMembers.onBoardGroupDatas:removeAll()
	
	if localMembers.spotDatas.spot ~= nil then
		localMembers.spotDatas.spot:destroy()
		localMembers.spotDatas.spot = nil
	end
	
	ATMEFct.C_Group_deleteUnit(_AIUnit:getGroup(), localMembers.name)
	ATMEDatas.C_AIUnit_MT:remove(_AIUnit:getIndex())
	
	
	
	ATMEDatas.AIUnits[localMembers.name] = nil
	ATMEDatas.allAIUnitsAndPlayers[localMembers.name] = nil
	localMembers.destroyed = true
	
	ATMEFct.outputCore(nil,"Delete unit = " .. localMembers.name, 2)
end

setmetatable(ATME.C_AIUnit,ATMEDatas.C_AIUnit_MT)


ATMEDatas.C_AIUnit_MT.mt.__metatable = "Private"
ATMEDatas.C_AIUnit_MT.mt.__index = ATME.C_AIUnit
ATMEDatas.C_AIUnit_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_AIUnit_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_AIUnit : Bad key " .. k, "table locked", false)
	end
end

ATMEDatas.C_AIUnit_MT.mtModules.__metatable = "Private"
ATMEDatas.C_AIUnit_MT.mtModules.__newindex = function(t, k, v)
	if ATMEDatas.modules[k] ~= nil then
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil,"ATME : ATME.C_AIUnit.modules Bad key " .. k, "table locked", false)
	end
end


----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_StaticObject : gère les objets statiques au format ATME. Pas de contructeur accessible
-- pour les modules externes. 
-- Sont exclus les objets statiques FARP et Airfieds qui sont gérés par ATME.C_Airbase.
-- Les instances de cette classe sont intégralement gérées par le module ATME_Core. La création et la destruction
-- dépendent du contexte de début de mission et des événements qui se produisent dans DCS en cours de mission.
-- Les modules externes accèdent à cette gestion indirectement par les handlers définis dans ceux-ci et les méthodes
-- accessibles de cette classe. Ceci permet de garantir une cohérence sur les listes internes ATME.
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_StaticObject_MT = {
	className = "C_StaticObject",
	mtModules = {},
	mt = {},
	
	keys = {
		["getCategoryName"] = true,
		["getLife"] = true,
		["getDCSStaticObject"] = true,
		["getDCSID"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getIndex"] = true,
		["getTypeName"] = true,
		["getPosition"] = true,
		["getAzimuth"] = true,
		["getCountryName"] = true,
		["getCoalitionName"] = true,
		["getAttributes"] = true,
		["hasAttributes"] = true,
		["isInZones"] = true,
		["isInDCSZone"] = true,
		["isInZone2D"] = true,
		["isInZone3D"] = true,
		["explode"] = true,
		["fireFlare"] = true,
		["fireSmoke"] = true,
		["fireIlluminationBomb"] = true, 
		["disable"] = true,
		["getUserPrivateDatas"] = true,
		["modules"] = true,
	},

	add = function(self, _name, _object)
		local category
		if ATMEDatas.allStaticObjectsInMission[_name] ~= nil then
			category = ATMEConst.backStaticObjectCategories[ATMEDatas.allStaticObjectsInMission[_name].category]
		else
			category = ATMEConst.backStaticObjectCategories[ATMEDatas.staticObjectsToBirth[_name].category].category
		end
	
		local item = {
			name = _name,
			objectDCS = _object,
			objectDCSId = _object:getID(),
			isEnabled = true,
			userPrivateDatas = {},						-- Point d'entrée des données privées : un par module
			tagsName = ATMEFct.decompString(_name),
			category = category,
			destroyed = false,							-- Utile lors du delete (car isExist est déjà false)
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_StaticObject = {
	getClassName = function(self)
		return ATMEDatas.C_StaticObject_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_StaticObject_MT.parentClass
	end,
	
	getByName = function(_objectName)
		if type(_objectName) == "string" then
			return ATMEDatas.staticObjects[_objectName]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_StaticObject.getByName(name)", "name must be string", false)
		end	
	end,
}


ATMEFct.C_StaticObject_create = function(_object)
	local newObject = {}
	local typeName = _object:getTypeName()
	
	local index = ATMEDatas.C_StaticObject_MT:add(_object:getName(), _object)
	local localMembers = ATMEDatas.localMembers[index]
	
	setmetatable(newObject, ATMEDatas.C_StaticObject_MT.mt)
	
	function newObject.getIndex(self)
		return index
	end
	
	function newObject.getDCSStaticObject(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.getDCSID(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCSId
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.getName(self)
		if localMembers.destroyed == false then
			return localMembers.name
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "tagGetByID(position)"
		
		if localMembers.destroyed == false then
			if type(_position) == "number" then
				local pos = math.floor(_position)
				if localMembers.tagsName ~= nil and pos >= 1 and pos <= #localMembers.tagsName then
					return localMembers.tagsName[pos]
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.tagGetCount(self)
		if localMembers.destroyed == false then
			if tagsName ~= nil then
				return #tagsName
			else
				return 0
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "tagCompare(tagsToCompare)"
		
		if localMembers.destroyed == false then
			if type(_tagsToCompare) == "string" then
				if localMembers.tagsName ~= nil then
					local ok = ATMEFct.compareTags(localMembers.tagsName, _tagsToCompare)
					
					if ok ~= nil then
						return ok
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
					end
				else
					return false
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.getCategoryName(self)
		-- Catégories objets statiques - A ne pas confondre avec catégorie DE BASE dcs des statics/cargo
		-- 		Ships
		-- 		Planes
		-- 		Helicopters
		-- 		Warehouses
		-- 		Cargos	-> Traité par la classe ATME.C_Cargo
		-- 		Unarmed (tous ce qui est véhicules terrestre)
		-- 		Fortifications
		
		if localMembers.destroyed == false then
			return localMembers.category
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.getTypeName(self)
		if localMembers.destroyed == false then
			return typeName
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.getLife(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getLife()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.getPosition(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getPoint()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newObject.getAzimuth(self)
		if localMembers.destroyed == false then
			local vector = ATMEFct.C_Vector3D_create(nil, nil, localMembers.objectDCS:getPosition().x)
			return vector:getAzimuth()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newObject.getCountryName(self)
		if localMembers.destroyed == false then
			local id = localMembers.objectDCS:getCountry()
			return country.name[id]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newObject.getCoalitionName(self)
		if localMembers.destroyed == false then
			local id = localMembers.objectDCS:getCoalition()
			return ATMEConst.coalitionNames[id]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.disable(self)
		if localMembers.destroyed == false then
			ATMEDatas.staticsToDisable[localMembers.name] = self
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.getUserPrivateDatas(self)
		if localMembers.destroyed == false then
			return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.getAttributes(self)
		if localMembers.destroyed == false then
			local datas = StaticObject.getDesc(localMembers.objectDCS)
			local attributes = {}
			for _id, _ in pairs(datas.attributes) do
				if ATMEConst.backTargetsAttributes[_id] ~= nil then
					table.insert(attributes, ATMEConst.backTargetsAttributes[_id].value)
				end
			end
			
			return attributes
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.hasAttributes(self, _attributes)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "hasAttributes(attributes)"
		
		if localMembers.destroyed == false then
			if type(_attributes) == "string" or type(_attributes) == "table" then
				local err, datas = ATMEFct.extractDatasFromStringOrTable(_attributes, false)
		
				if err == false then
					return ATMEFct.hasAttributes(localMembers.objectDCS, datas)
				end
				
				ATMEFct.errorOutputCore(nil, fctDefinition, datas, false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "attributes type must be string or table", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newObject.explode(self, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "explode(power)"
		
		if localMembers.destroyed == false then
			if type(_power) == "number" then
				if _power >= 1 then
					ATME.explode(self:getPosition(), _power)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.fireFlare(self, _color)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "fireFlare(color)"
		
		if localMembers.destroyed == false then
			if type(_color) == "string" then
				if ATMEConst.flareColors[string.upper(_color)]~= nil then
					if localMembers.flare == nil then
						localMembers.flare = ATMEFct.C_Flare_create(_color, self:getPosition())
					end
					localMembers.flare:activate()
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a flare color or RANDOM", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newObject.fireSmoke(self, _color)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "fireSmoke(color)"
		
		if localMembers.destroyed == false then
			if type(_color) == "string" then
				if ATMEConst.smokeColors[string.upper(_color)]~= nil then 
					local point = self:getPosition()
					local newPoint = {
						x = point.x + math.random(-20,20),
						y = point.y,
						z = point.z + math.random(-20,20),
					}
						
					local smoke = ATMEFct.C_Smoke_create(_color, newPoint)
					smoke:activate(false)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a smoke color", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newObject.fireIlluminationBomb(self, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "fireIlluminationBomb(power)"
		
		if localMembers.destroyed == false then
			if type(_power) == "number" then
				if _power >= 1 then
					local point = self:getPosition()
					
					point.y = point.y + 200 + math.random(0, 100)

					trigger.action.illuminationBomb(point, _power)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.isInZones(self, _zonesMgr)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "isInZones(zonesMgr)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zonesMgr, ATME.C_ZonesMgr) == true then
				return _zonesMgr:isInside(self) 
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zonesMgr type must be ATME.C_ZonesMgr", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end				
	
	function newObject.isInDCSZone(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "isInDCSZone(name)"
		
		if localMembers.destroyed == false then
			if type(_name) == "string" then
				local zone = ATMEFct.getDCSZone(_name)
				
				if zone ~= nil then
					local vector = ATMEFct.C_Vector3D_create(nil, nil, self:getPosition(), zone.point)
					if vector:getHModule() <= zone.radius then
						return true
					else
						return false
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _name .. " unknown", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.isInZone2D(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "isInZone2D(zone)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zone, ATME.C_Zone2D) == true then
				local point = self:getPosition()
				local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
				
				local datas = {
					zone = _zone,
					actualCenter = localMembersZone.center,
					azimuth = localMembersZone.azimuth,
				}
			
				return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](point, datas)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Circle, ATME.C_PickupZone, ATME.C_Rectangle or ATME.C_Polygon", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.isInZone3D(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "isInZone3D(zone)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zone, ATME.C_Zone3D) == true then
				local point = self:getPosition()
				local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
		
				local datas = {
					zone = _zone,
					actualCenter = localMembersZone.center,
					azimuth = localMembersZone.azimuth,
				}
				
				return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](point, datas)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Sphere, ATME.C_Cylinder or ATME.C_Box", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	ATMEDatas.newCreatedStatics[localMembers.name] = {
		object = newObject,
	}
	
	if ATMEDatas.staticObjectsToBirth[localMembers.name] ~= nil then
		ATMEDatas.newCreatedStatics[localMembers.name].isSpawned = ATMEDatas.staticObjectsToBirth[localMembers.name].isSpawned
		ATMEDatas.newCreatedStatics[localMembers.name].spawnCallback = ATMEDatas.staticObjectsToBirth[localMembers.name].spawnCallback
		ATMEDatas.newCreatedStatics[localMembers.name].staticCloningContext =  ATMEDatas.staticObjectsToBirth[localMembers.name].staticCloningContext
		ATMEDatas.staticObjectsToBirth[localMembers.name] = nil
	end
	
	newObject.modules = {}
	setmetatable(newObject.modules, ATMEDatas.C_StaticObject_MT.mtModules)
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
		newObject.modules[_localModuleMembers.name] = {}
	end
	
	ATMEFct.outputCore(nil,"Prepare static unit " .. localMembers.name .. " -> ok", 2)
end

ATMEFct.C_StaticObject_delete = function(_staticUnit)
	local localMembers = ATMEDatas.localMembers[_staticUnit:getIndex()]
	
	local flag = "DESTROYED"
	if localMembers.isEnabled == false then
		flag = "DISABLED"
	end
	
	for _i = #ATMEDatas.modulesNames, 1, -1 do
		local localModuleMembers = ATMEDatas.modulesLocalMembers[ATMEDatas.modulesNames[_i]]
		if localModuleMembers.handlers.onDeleteStaticObjectHandler ~= nil then
			ATMEDatas.currentModuleName = localModuleMembers.name
			
			-- Appel du module
			ATMEFct.outputCore(nil,"Appel onDeleteStaticObjectHandler du module " .. localModuleMembers.name, 2)
			xpcall(function() localModuleMembers.handlers.onDeleteStaticObjectHandler(_staticUnit, flag) end, function() ATMEFct.errUserHandler('onDeleteStaticObjectHandler') end)
		end
	end
	
	_staticUnit.modules = nil
	
	ATMEDatas.C_StaticObject_MT:remove(_staticUnit:getIndex())
	
	ATMEDatas.staticObjects[localMembers.name] = nil
	localMembers.destroyed = true
	
	ATMEFct.outputCore(nil,"Delete static unit = " .. localMembers.name, 2)
end

ATMEFct.C_StaticObject_disable = function(_staticUnit)
	local staticDCS = _staticUnit:getDCSStaticObject()
	local localMembers = ATMEDatas.localMembers[_staticUnit:getIndex()]
	localMembers.isEnabled = false
		
	staticDCS:destroy()
	ATMEFct.C_StaticObject_delete(_staticUnit)
end

setmetatable(ATME.C_StaticObject,ATMEDatas.C_StaticObject_MT)

ATMEDatas.C_StaticObject_MT.mt.__metatable = "Private"
ATMEDatas.C_StaticObject_MT.mt.__index = ATME.C_StaticObject
ATMEDatas.C_StaticObject_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_StaticObject_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_StaticObject : Bad key " .. k, "table locked", false)
	end
end

ATMEDatas.C_StaticObject_MT.mtModules.__metatable = "Private"
ATMEDatas.C_StaticObject_MT.mtModules.__newindex = function(t, k, v)
	if ATMEDatas.modules[k] ~= nil then
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil,"ATME : ATME.C_StaticObject.modules Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Cargo : gère les objets cargos au format ATME. Pas de contructeur accessible
-- pour les modules externes. 
-- Les instances de cette classe sont intégralement gérées par le module ATME_Core. La création et la destruction
-- dépendent du contexte de début de mission et des événements qui se produisent dans DCS en cours de mission.
-- Les modules externes accèdent à cette gestion indirectement par les handlers définis dans ceux-ci et les méthodes
-- accessibles de cette classe. Ceci permet de garantir une cohérence sur les listes internes ATME.
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Cargo_MT = {
	className = "C_Cargo",
	mtModules = {},
	mt = {},
	
	keys = {
		["getCategoryName"] = true,
		["getLife"] = true,
		["getDCSCargo"] = true,
		["getDCSID"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getIndex"] = true,
		["getTypeName"] = true,
		["getPosition"] = true,
		["inAir"] = true,
		["getWeight"] = true,
		["getAzimuth"] = true,
		["getCountryName"] = true,
		["getCoalitionName"] = true,
		["isInZones"] = true,
		["isInDCSZone"] = true,
		["isInZone2D"] = true,
		["isInZone3D"] = true,
		["explode"] = true,
		["fireFlare"] = true,
		["fireSmoke"] = true,
		["fireIlluminationBomb"] = true, 
		["disable"] = true,
		["getUserPrivateDatas"] = true,
		["modules"] = true,
	},

	add = function(self, _name, _cargo)
		local item = {
			name = _name,
			objectDCS = _cargo,
			objectDCSId = _cargo:getID(),
			isEnabled = true,
			userPrivateDatas = {},						-- Point d'entrée des données privées : un par module
			tagsName = ATMEFct.decompString(_name),
			category = "CARGO",
			destroyed = false,							-- Utile lors du delete (car isExist est déjà false)
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Cargo = {
	getClassName = function(self)
		return ATMEDatas.C_Cargo_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Cargo_MT.parentClass
	end,
	
	getByName = function(_cargoName)
		if type(_cargoName) == "string" then
			return ATMEDatas.cargos[_cargoName]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Cargo.getByName(name)", "name must be string", false)
		end	
	end,
}


ATMEFct.C_Cargo_create = function(_cargo)
	local newCargo = {}
	local typeName = _cargo:getTypeName()

	local index = ATMEDatas.C_Cargo_MT:add(_cargo:getName(), _cargo)
	local localMembers = ATMEDatas.localMembers[index]

	setmetatable(newCargo, ATMEDatas.C_Cargo_MT.mt)

	function newCargo.getIndex(self)
		return index
	end
	
	function newCargo.getDCSCargo(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newCargo.getDCSID(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCSId
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newCargo.getName(self)
		if localMembers.destroyed == false then
			return localMembers.name
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newCargo.tagGetByID(self, _position)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "tagGetByID(position)"
		
		if localMembers.destroyed == false then
			if type(_position) == "number" then
				local pos = math.floor(_position)
				if localMembers.tagsName ~= nil and pos >= 1 and pos <= #localMembers.tagsName then
					return localMembers.tagsName[pos]
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
		
	function newCargo.tagGetCount(self)
		if localMembers.destroyed == false then
			if tagsName ~= nil then
				return #tagsName
			else
				return 0
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newCargo.tagCompare(self, _tagsToCompare)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "tagCompare(tagsToCompare)"
		
		if localMembers.destroyed == false then
			if type(_tagsToCompare) == "string" then
				if localMembers.tagsName ~= nil then
					local ok = ATMEFct.compareTags(localMembers.tagsName, _tagsToCompare)
					
					if ok ~= nil then
						return ok
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
					end
				else
					return false
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newCargo.getCategoryName(self)
		-- Catégories objets cargos (sous ensemble static) - A ne pas confondre avec catégorie des statics/cargo
		-- 		Cargos
		
		if localMembers.destroyed == false then
			return localMembers.category
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newCargo.getTypeName(self)
		if localMembers.destroyed == false then
			return typeName
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newCargo.getLife(self)
		if localMembers.destroyed == false then
			return localMembers.cargoDCS:getLife()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newCargo.getPosition(self)
		if localMembers.destroyed == false then
			return localMembers.cargoDCS:getPoint()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newCargo.getWeight(self)
		if localMembers.destroyed == false then
			return localMembers.cargoDCS:getCargoWeight()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newCargo.inAir(self)
		if localMembers.destroyed == false then
			local pos = localMembers.cargoDCS:getPoint()
			
			local delta = pos.y - ATMEFct.getSurfaceAltitude(pos)
			
			if delta < 2 then
				return false
			else
				return true
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newCargo.getAzimuth(self)
		if localMembers.destroyed == false then
			local vector = ATMEFct.C_Vector3D_create(nil, nil, localMembers.cargoDCS:getPosition().x)
			return vector:getAzimuth()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newCargo.getCountryName(self)
		if localMembers.destroyed == false then
			local id = localMembers.cargoDCS:getCountry()
			return country.name[id]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newCargo.getCoalitionName(self)
		if localMembers.destroyed == false then
			local id = localMembers.cargoDCS:getCoalition()
			return ATMEConst.coalitionNames[id]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newCargo.disable(self)
		if localMembers.destroyed == false then
			ATMEDatas.cargosToDisable[localMembers.name] = self
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newCargo.getUserPrivateDatas(self)
		if localMembers.destroyed == false then
			return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newCargo.explode(self, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "explode(power)"
		
		if localMembers.destroyed == false then
			if type(_power) == "number" then
				if _power >= 1 then
					ATME.explode(self:getPosition(), _power)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newCargo.fireFlare(self, _color)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "fireFlare(color)"
		
		if localMembers.destroyed == false then
			if self:inAir() == false then
				if type(_color) == "string" then
					if ATMEConst.flareColors[string.upper(_color)]~= nil then
						if localMembers.flare == nil then
							localMembers.flare = ATMEFct.C_Flare_create(_color, self:getPosition())
						else
							ATMEDatas.localMembers[localMembers.flare:getIndex()].point = self:getPosition()
						end
						localMembers.flare:activate()
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a flare color or RANDOM", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", true)
				end
			else
				ATMEFct.setError("IN_AIR")
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newCargo.fireSmoke(self, _color)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "fireSmoke(color)"
		
		if localMembers.destroyed == false then
			if self:inAir() == false then
				if type(_color) == "string" then
					if ATMEConst.smokeColors[string.upper(_color)]~= nil then 
						local point = self:getPosition()
						local newPoint = {
							x = point.x + math.random(-20,20),
							y = point.y,
							z = point.z + math.random(-20,20),
						}
							
						local smoke = ATMEFct.C_Smoke_create(_color, newPoint)
						smoke:activate(false)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a smoke color", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", true)
				end
			else
				ATMEFct.setError("IN_AIR")
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newCargo.fireIlluminationBomb(self, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "fireIlluminationBomb(power)"
		
		if localMembers.destroyed == false then
			if self:inAir() == false then
				if type(_power) == "number" then
					if _power >= 1 then
						local point = self:getPosition()
						
						point.y = point.y + 200 + math.random(0, 100)

						trigger.action.illuminationBomb(point, _power)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
				end
			else
				ATMEFct.setError("IN_AIR")
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newCargo.isInZones(self, _zonesMgr)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "isInZones(zonesMgr)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zonesMgr, ATME.C_ZonesMgr) == true then
				return _zonesMgr:isInside(self) 
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zonesMgr type must be ATME.C_ZonesMgr", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end				
	
	function newCargo.isInDCSZone(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "isInDCSZone(name)"
		
		if localMembers.destroyed == false then
			if type(_name) == "string" then
				local zone = ATMEFct.getDCSZone(_name)
				
				if zone ~= nil then
					local vector = ATMEFct.C_Vector3D_create(nil, nil, self:getPosition(), zone.point)
					if vector:getHModule() <= zone.radius then
						return true
					else
						return false
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _name .. " unknown", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newCargo.isInZone2D(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "isInZone2D(zone)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zone, ATME.C_Zone2D) == true then
				local point = self:getPosition()
				local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
				
				local datas = {
					zone = _zone,
					actualCenter = localMembersZone.center,
					azimuth = localMembersZone.azimuth,
				}
			
				return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](point, datas)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Circle, ATME.C_PickupZone, ATME.C_Rectangle or ATME.C_Polygon", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newCargo.isInZone3D(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newCargo) .. "isInZone3D(zone)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zone, ATME.C_Zone3D) == true then
				local point = self:getPosition()
				local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
		
				local datas = {
					zone = _zone,
					actualCenter = localMembersZone.center,
					azimuth = localMembersZone.azimuth,
				}
				
				return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](point, datas)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Sphere, ATME.C_Cylinder or ATME.C_Box", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	ATMEDatas.newCreatedCargos[localMembers.name] = {
		cargo = newCargo,
	}
	
	if ATMEDatas.cargosToBirth[localMembers.name] ~= nil then
		ATMEDatas.newCreatedCargos[localMembers.name].isSpawned = ATMEDatas.cargosToBirth[localMembers.name].isSpawned
		ATMEDatas.newCreatedCargos[localMembers.name].spawnCallback = ATMEDatas.cargosToBirth[localMembers.name].spawnCallback
		ATMEDatas.newCreatedCargos[localMembers.name].cargoCloningContext =  ATMEDatas.cargosToBirth[localMembers.name].cargoCloningContext
		ATMEDatas.cargosToBirth[localMembers.name] = nil
	end
	
	newCargo.modules = {}
	setmetatable(newCargo.modules, ATMEDatas.C_Cargo_MT.mtModules)
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
		newCargo.modules[_localModuleMembers.name] = {}
	end
	
	ATMEFct.outputCore(nil,"Prepare Cargo " .. localMembers.name .. " -> ok", 2)
end

ATMEFct.C_Cargo_delete = function(_cargoUnit)
	local localMembers = ATMEDatas.localMembers[_cargoUnit:getIndex()]
	
	local flag = "DESTROYED"
	if localMembers.isEnabled == false then
		flag = "DISABLED"
	end
	
	for _i = #ATMEDatas.modulesNames, 1, -1 do
		local localModuleMembers = ATMEDatas.modulesLocalMembers[ATMEDatas.modulesNames[_i]]
		if localModuleMembers.handlers.onDeleteCargoHandler ~= nil then
			ATMEDatas.currentModuleName = localModuleMembers.name
			
			-- Appel du module
			ATMEFct.outputCore(nil,"Appel onDeleteCargoHandler du module " .. localModuleMembers.name.. " pour le cargo " .. _cargoUnit:getName(), 2)
			xpcall(function() localModuleMembers.handlers.onDeleteCargoHandler(_cargoUnit, flag) end, function() ATMEFct.errUserHandler('onDeleteCargoHandler') end)
		end
	end
	
	_cargoUnit.modules = nil
	
	ATMEDatas.C_Cargo_MT:remove(_cargoUnit:getIndex())
	
	ATMEDatas.cargos[localMembers.name] = nil
	localMembers.destroyed = true
	
	ATMEFct.outputCore(nil,"Delete cargo = " .. localMembers.name, 2)
end

ATMEFct.C_Cargo_disable = function(_cargoUnit)
	local cargoDCS = _cargoUnit:getDCSCargo()
	local localMembers = ATMEDatas.localMembers[_cargoUnit:getIndex()]
	localMembers.isEnabled = false
		
	cargoDCS:destroy()
	ATMEFct.C_Cargo_delete(_cargoUnit)
end

setmetatable(ATME.C_Cargo,ATMEDatas.C_Cargo_MT)

ATMEDatas.C_Cargo_MT.mt.__metatable = "Private"
ATMEDatas.C_Cargo_MT.mt.__index = ATME.C_Cargo
ATMEDatas.C_Cargo_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Cargo_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Cargo : Bad key " .. k, "table locked", false)
	end
end

ATMEDatas.C_Cargo_MT.mtModules.__metatable = "Private"
ATMEDatas.C_Cargo_MT.mtModules.__newindex = function(t, k, v)
	if ATMEDatas.modules[k] ~= nil then
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil,"ATME : ATME.C_Cargo.modules Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Weapon : gère les armes au format ATME. Pas de contructeur accessible
-- pour les modules externes.
-- Les instances de cette classe sont intégralement gérées par le module ATME_Core. La création et la destruction
-- dépendent des événements qui se produisent dans DCS en cours de mission.
-- Seules les armes activent sont référencées.
-- Ci dessous dans les fonctions de la classe, isExist permet de s'assurer de l'existance de l'arme car
-- pas de destruction détectée à l'impact si l'arme ne touche rien (bâtiment, unité....) - Bug DCS? 
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Weapon_MT = {
	className = "C_Weapon",
	
	mtModules = {},
	mt = {},
	
	keys = {
		["getDCSWeapon"] = true,
		["getIndex"] = true,
		["getName"] = true,
		["getTypeName"] = true,
		["getCoalitionName"] = true,
		["getCountryName"] = true,
		["getPosition"] = true,
		["getAzimuth"] = true,
		["getLauncher"] = true,
		["getTarget"] = true,
		["getAttributes"] = true,
		["hasAttributes"] = true,
		["explode"] = true,
		["disable"] = true,
		["getUserPrivateDatas"] = true,
		["modules"] = true,
	},
	
	add = function(self, _weapon)
		local item = {
			objectDCS = _weapon,
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Weapon = {
	getClassName = function(self)
		return ATMEDatas.C_Weapon_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Weapon_MT.parentClass
	end,
}


ATMEFct.C_Weapon_create = function(_weapon)
	local name = _weapon:getName()

	local newWeapon = {}

	local typeName = _weapon:getTypeName()
	local category
	
	setmetatable(newWeapon, ATMEDatas.C_Weapon_MT.mt)
	local index = ATMEDatas.C_Weapon_MT:add(_weapon)
	local localMembers = ATMEDatas.localMembers[index]
	
	function newWeapon.getIndex(self)
		return index
	end
	
	function newWeapon.getDCSWeapon(self)
		if localMembers.objectDCS:isExist() == true then
			return localMembers.objectDCS
		else
			ATMEFct.C_Weapon_delete(name)
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newWeapon.getName(self)
		if localMembers.objectDCS:isExist() == false then
			ATMEFct.C_Weapon_delete(name)
			ATMEFct.setError("ALREADY_DESTROYED")
		end
		
		return name
	end
	
	function newWeapon.getCountryName(self)
		if localMembers.objectDCS:isExist() == false then
			local id = localMembers.objectDCS:getCountry()
			return country.name[id]
		else
			ATMEFct.C_Weapon_delete(name)
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newWeapon.getCoalitionName(self)
		if localMembers.objectDCS:isExist() == false then
			local id = localMembers.objectDCS:getCoalition()
			return ATMEConst.coalitionNames[id]
		else
			ATMEFct.C_Weapon_delete(name)
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newWeapon.getTypeName(self)
		if localMembers.objectDCS:isExist() == false then
			return typeName
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newWeapon.getLauncher(self)
		if localMembers.objectDCS:isExist() == true then
			local launcher = localMembers.objectDCS:getLauncher()
			
			if launcher ~= nil then
				return ATMEDatas.allAIUnitsAndPlayers[launcher:getName()]
			end
		else
			ATMEFct.C_Weapon_delete(name)
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newWeapon.getTarget(self)
		if localMembers.objectDCS:isExist() == true then
			local target = localMembers.objectDCS:getTarget()
		else
			ATMEFct.C_Weapon_delete(name)
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newWeapon.getPosition(self)
		if localMembers.objectDCS:isExist() == true then
			return localMembers.objectDCS:getPoint()
		else
			ATMEFct.C_Weapon_delete(name)
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newWeapon.getAzimuth(self)
		if localMembers.objectDCS:isExist() == true then
			local vector = ATMEFct.C_Vector3D_create(nil, nil, localMembers.objectDCS:getPosition().x)
			return vector:getAzimuth()
		else
			ATMEFct.C_Weapon_delete(name)
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newWeapon.disable(self)
		if localMembers.objectDCS:isExist() == true then
			localMembers.objectDCS:destroy()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
		
		ATMEFct.C_Weapon_delete(self:getName())
	end
	
	function newWeapon.getAttributes(self)
		if localMembers.objectDCS:isExist() == true then
			local datas = Weapon.getDesc(localMembers.objectDCS)
			local attributes = {}
			for _id, _ in pairs(datas.attributes) do
		--		if ATMEConst.backTargetsAttributes[_id] ~= nil then
		--			table.insert(attributes, ATMEConst.backTargetsAttributes[_id].value)
		--		end
			end
			
			return attributes
		else
			ATMEFct.C_Weapon_delete(name)
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newWeapon.hasAttributes(self, _attributes)
		local fctDefinition = ATMEFct.getErrorPrefix(newWeapon) .. "hasAttributes(attributes)"
		
		if localMembers.objectDCS:isExist() == true then
			if type(_attributes) == "string" or type(_attributes) == "table" then
				local err, datas = ATMEFct.extractDatasFromStringOrTable(_attributes, false)
		
				if err == false then
					return ATMEFct.hasAttributes(localMembers.objectDCS, datas)
				end
				
				ATMEFct.errorOutputCore(nil, fctDefinition, datas, false)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "attributes type must be string or table", true)
			end
		else
			ATMEFct.C_Weapon_delete(name)
			ATMEFct.setError("ALREADY_DESTROYED")
			return
		end	
	end

	function newWeapon.explode(self, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newWeapon) .. "explode(power)"
		
		if localMembers.objectDCS:isExist() == true then
			if type(_power) == "number" then
				if _power >= 1 then
					ATME.explode(self:getPosition(), _power)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
			end

			ATMEDatas.weapons[name] = newWeapon
			
			ATMEFct.outputCore(nil,"Création weapon " .. name .. " -> ok", 2)
			
			return newWeapon
		else
			ATMEFct.C_Weapon_delete(name)
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newWeapon.getUserPrivateDatas(self)
		if localMembers.objectDCS:isExist() == true then
			return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
		else
			ATMEFct.C_Weapon_delete(name)
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	newWeapon.modules = {}
	setmetatable(newWeapon.modules, ATMEDatas.C_Weapon_MT.mtModules)
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
		newWeapon.modules[_localModuleMembers.name] = {}
	end
end

ATMEFct.C_Weapon_delete = function(_name)
	if ATMEDatas.weapons[_name] ~= nil then
		ATMEDatas.C_Weapon_MT:remove(ATMEDatas.weapons[_name]:getIndex())
		ATMEDatas.weapons[_name].modules = nil
		ATMEDatas.weapons[_name] = nil
		ATMEFct.outputCore(nil,"Delete weapon = " .. _name, 2)
	end
end

setmetatable(ATME.C_Weapon,ATMEDatas.C_Weapon_MT)

ATMEDatas.C_Weapon_MT.mt.__metatable = "Private"
ATMEDatas.C_Weapon_MT.mt.__index = ATME.C_Weapon
ATMEDatas.C_Weapon_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Weapon_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Weapon : Bad key " .. k, "table locked", false)
	end
end

ATMEDatas.C_Weapon_MT.mtModules.__metatable = "Private"
ATMEDatas.C_Weapon_MT.mtModules.__newindex = function(t, k, v)
	if ATMEDatas.modules[k] ~= nil then
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil,"ATME : ATME.C_Weapon_MT.modules Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_SceneryObject : gère les objects statiques positionnées sur les cartes de DCS.
-- Les instances de cette classe sont définies par un nom qui est un identifiant numérique.
-- Les instances référencées ne sont que les objets ayant été touchés par une arme mais non détruits
-- ou les objects retrouvés par ATME.C_FindContext.
--
-- A noter : getDesc et getLife non fonctionnels arrêt immédiat du script - Bug DCS
--           destroy non opérationnel, l'objet reste visible - Bug DCS
--			 Paradoxalement getDescByName fonctionne. En fait, tous les objets retournés par C_FindContext
--			 sont des "Buildings".
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_SceneryObject_MT = {
	className = "C_SceneryObject",
	
	mtModules = {},
	mt = {},
	
	keys = {
		["getDCSSceneryObject"] = true,
		["getName"] = true,
		["getTypeName"] = true,
		["getPosition"] = true,
		["explode"] = true,
		["getIndex"] = true,
		["getUserPrivateDatas"] = true,
		["fireFlare"] = true,
		["fireSmoke"] = true,
		["fireIlluminationBomb"] = true, 
		["isInZones"] = true,
		["isInDCSZone"] = true,
		["isInZone2D"] = true,
		["isInZone3D"] = true,
		["modules"] = true,
	},
	
	add = function(self, _object)
		local item = {
			objectDCS = _object,
			userPrivateDatas = {},
			destroyed = false,
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_SceneryObject = {
	getClassName = function(self)
		return ATMEDatas.C_SceneryObject_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_SceneryObject_MT.parentClass
	end,
}


ATMEFct.C_SceneryObject_create = function(_object)
	local name = _object:getName()

	local newObject = {}
	
	local typeName = _object:getTypeName()
	
	setmetatable(newObject, ATMEDatas.C_SceneryObject_MT.mt)
	local index = ATMEDatas.C_SceneryObject_MT:add(_object)
	local localMembers = ATMEDatas.localMembers[index]

	function newObject.getIndex(self)
		return index
	end
	
	function newObject.getDCSSceneryObject(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.getName(self)
		if localMembers.destroyed == false then
			return name
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.getTypeName(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getTypeName()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
		
	function newObject.getPosition(self)
		if localMembers.destroyed == false then
			return localMembers.objectDCS:getPoint()
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.explode(self, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "explode(power)"
		
		if localMembers.destroyed == false then
			if type(_power) == "number" then
				if _power >= 1 then
					ATME.explode(self:getPosition(), _power)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.fireFlare(self, _color)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "fireFlare(color)"
		
		if localMembers.destroyed == false then
			if type(_color) == "string" then
				if ATMEConst.flareColors[string.upper(_color)]~= nil then
					if localMembers.flare == nil then
						localMembers.flare = ATMEFct.C_Flare_create(_color, self:getPosition())
					end
					localMembers.flare:activate()
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a flare color or RANDOM", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newObject.fireSmoke(self, _color)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "fireSmoke(color)"
		
		if localMembers.destroyed == false then
			if type(_color) == "string" then
				if ATMEConst.smokeColors[string.upper(_color)]~= nil then 
					local point = self:getPosition()
					local newPoint = {
						x = point.x + math.random(-20,20),
						y = point.y,
						z = point.z + math.random(-20,20),
					}
						
					local smoke = ATMEFct.C_Smoke_create(_color, newPoint)
					smoke:activate(false)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "color must be a smoke color", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "color type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end

	function newObject.fireIlluminationBomb(self, _power)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "fireIlluminationBomb(power)"
		
		if localMembers.destroyed == false then
			if type(_power) == "number" then
				if _power >= 1 then
					local point = self:getPosition()
					
					point.y = point.y + 200 + math.random(0, 100)

					trigger.action.illuminationBomb(point, _power)
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "power must be >= 1", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "power type must be number", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.getUserPrivateDatas(self)
		if localMembers.destroyed == false then
			return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.isInZones(self, _zonesMgr)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "isInZones(zonesMgr)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zonesMgr, ATME.C_ZonesMgr) == true then
				return _zonesMgr:isInside(self) 
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zonesMgr type must be ATME.C_ZonesMgr", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end				
	
	function newObject.isInDCSZone(self, _name)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "isInDCSZone(name)"
		
		if localMembers.destroyed == false then
			if type(_name) == "string" then
				local zone = ATMEFct.getDCSZone(_name)
				
				if zone ~= nil then
					local vector = ATMEFct.C_Vector3D_create(nil, nil, self:getPosition(), zone.point)
					if vector:getHModule() <= zone.radius then
						return true
					else
						return false
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "DCS zone " .. _name .. " unknown", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.isInZone2D(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "isInZone2D(zone)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zone, ATME.C_Zone2D) == true then
				local point = self:getPosition()
				local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
				
				local datas = {
					zone = _zone,
					actualCenter = localMembersZone.center,
					azimuth = localMembersZone.azimuth,
				}
			
				return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](point, datas)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Circle, ATME.C_PickupZone, ATME.C_Rectangle or ATME.C_Polygon", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	function newObject.isInZone3D(self, _zone)
		local fctDefinition = ATMEFct.getErrorPrefix(newObject) .. "isInZone3D(zone)"
		
		if localMembers.destroyed == false then
			if ATMEFct.isObjectClass(_zone, ATME.C_Zone3D) == true then
				local point = self:getPosition()
				local localMembersZone = ATMEDatas.localMembers[_zone:getIndex()]
		
				local datas = {
					zone = _zone,
					actualCenter = localMembersZone.center,
					azimuth = localMembersZone.azimuth,
				}
				
				return ATMEFct[ATMEConst.indZoneConfig[_zone:getClassName()].fctIsInside](point, datas)
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "zone type must be ATME.C_Sphere, ATME.C_Cylinder or ATME.C_Box", true)
			end
		else
			ATMEFct.setError("ALREADY_DESTROYED")
		end
	end
	
	newObject.modules = {}
	setmetatable(newObject.modules, ATMEDatas.C_SceneryObject_MT.mtModules)
	
	for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
		localMembers.userPrivateDatas[_localModuleMembers.name] = {}
		newObject.modules[_localModuleMembers.name] = {}
	end
	
	ATMEDatas.sceneryObjects[name] = newObject
	
	ATMEFct.outputCore(nil,"Création objet de carte (sceneryObject) " .. name .. " -> ok", 2)
end

ATMEFct.C_SceneryObject_delete = function(_name)
	if ATMEDatas.sceneryObjects[_name] ~= nil then
		local localMembers = ATMEDatas.localMembers[ATMEDatas.sceneryObjects[_name]:getIndex()]
		localMembers.destroyed = true
		ATMEDatas.C_SceneryObject_MT:remove(ATMEDatas.sceneryObjects[_name]:getIndex())
		ATMEDatas.sceneryObjects[_name].modules = nil
		ATMEDatas.sceneryObjects[_name] = nil
		
		ATMEFct.outputCore(nil,"Delete sceneryObject = " .. _name, 2)
	end
end

setmetatable(ATME.C_SceneryObject,ATMEDatas.C_SceneryObject_MT)

ATMEDatas.C_SceneryObject_MT.mt.__metatable = "Private"
ATMEDatas.C_SceneryObject_MT.mt.__index = ATME.C_SceneryObject
ATMEDatas.C_SceneryObject_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_SceneryObject_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_SceneryObject : Bad key " .. k, "table locked", false)
	end
end

ATMEDatas.C_SceneryObject_MT.mtModules.__metatable = "Private"
ATMEDatas.C_SceneryObject_MT.mtModules.__newindex = function(t, k, v)
	if ATMEDatas.modules[k] ~= nil then
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil,"ATME : ATME.C_SceneryObject_MT.modules Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- Classe ATME.C_Race : gère les courses de joueurs. 
-- Contructeur non accessible pour les modules externes. 
-- Une course de joueurs est composées de portes définie par un côté gauche et un côté droit (sens de passage
-- de la porte par le joueur). Les temps sont gérés au 1/100è. une course est associée au module l'ayant créée
-- et donc les events ATME générés seront transmis uniquement ce module. Le description des events est données
-- dans la fonction _timerx100.
-- Le nombre maximal de portes est à 100.
----------------------------------------------------------------------------------------------------------

ATMEDatas.C_Race_MT = {
	className = "C_Race",
	mt = {},
	
	keys = {
		["getNbTurns"] = true,
		["getNbDoors"] = true,
		["getIndex"] = true,
		["getName"] = true,
		["tagGetByID"] = true,
		["tagCompare"] = true,
		["tagGetCount"] = true,
		["getDoorSides"] = true,
		["getLapAtDoor"] = true,
		["setLapAtDoor"] = true,
		["getDoorIndex"] = true,
		["displayForAllPlayers"] = true,
		["soundForAllPlayers"] = true,
		["addPlayer"] = true,
		["removePlayer"] = true,
		["removeAllPlayers"] = true,
		["isPlayerInRace"] = true,
		["getRanking"] = true,
		["setMaxAltitudeRule"] = true,
		["getMaxAltitudePenalty"] = true,
		["getMaxAltitude"] = true,
		["resetMaxAltitudeRule"] = true,
		["getPlayerNextDoorIndex"] = true,
		["getUserPrivateDatas"] = true,
	},
	
	add = function(self, _doorsIndex)
		local item = {
			missedDoorPenalty = 0,
			maxAltitude = -1,
			maxAltitudePenalty = 0,
			players = {},
			doorsIndex = _doorsIndex,
			bestPlayers = {},
			bestTurn = 
			{
				pseudo = "",
				bestTime = -1,
			},
			userPrivateDatas = {},
		}
		
		table.insert(ATMEDatas.localMembers, item)
		return #ATMEDatas.localMembers
	end,
	
	remove = function(self, _id)
		ATMEDatas.localMembers[_id].userPrivateDatas = nil
	end,
}

ATME.C_Race = {
	getClassName = function(self)
		return ATMEDatas.C_Race_MT.className
	end,
	
	getParentClass = function(self)
		return ATMEDatas.C_Race_MT.parentClass
	end,
	
	getByName = function(_raceName)
		if type(_raceName) == "string" then
			return ATMEDatas.races[_raceName]
		else
			ATMEFct.errorOutputCore(nil, "Function ATME.C_Race.getByName(name)", "name type must be string", false)
		end	
	end,
}

ATMEFct.C_Race_createEvent = function(_race, _type, _datas)
	local localMembers = ATMEDatas.localMembers[_race:getIndex()]
	
	_datas.race = _race

	local item = {
		type = _type,
		datas = _datas,
	}
	
	ATMEFct.C_Event_add(item)
end

ATMEFct.C_Race_create = function(_name, _leftSideDoorName, _rightSideDoorName, _nbTurns)
	local doors = ATMEFct.C_IndexList_create()
	local doorsIndex = {}
	
	local fctDefinition = "Constructor ATME.C_Race(name, leftSideDoorName, righSideDoorName, nbTurns)"
	
	for _i = 1, 100 do -- Identifiant de porte dans l'éditeur de mission : de #001 à #100
		local index = string.format("%03d", _i)
		if ATMEDatas.staticObjects[_leftSideDoorName .. "#" .. index] ~= nil then
			if ATMEDatas.staticObjects[_rightSideDoorName .. "#" .. index] ~= nil then
				local leftObject = ATME.C_StaticObject.getByName(_leftSideDoorName .. "#" .. index)
				local rightObject = ATME.C_StaticObject.getByName(_rightSideDoorName .. "#" .. index)
				
				local segment2D = ATMEFct.C_Segment2D_create(nil, nil, {leftObject:getPosition(), rightObject:getPosition()})
				local p1, p2, vector = segment2D:getDatas()
				if vector:getHModule() < 500 then -- Largeur de porte maximale : 500m
					local door = {
						index = _i,
						leftObject = leftObject,
						rightObject = rightObject,
						segment = {
							p1 = p1,
							p2 = p2,
							vector = vector,
						},
						lap = false,
						bestTurn = {
							pseudo = "",
							bestTime = -1,
						},
					}
					door.id = doors:add(door)
					doorsIndex[_i] = door
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "Race " .. _name .. " - Door #" .. index .. " too large (>500m)", false)
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "Race " .. _name .. " - Door #" .. index .. " no right side", false)
			end
		elseif ATMEDatas.staticObjects[_rightSideDoorName .. " #" .. index] ~= nil then
			ATMEFct.errorOutputCore(nil, fctDefinition, "Race " .. _name .. " - Door #" .. index .. " no left side", false)
		end
	end
	
	if ATMEDatas.exitATME == false then
		if doors:getCount() >= 2 then
			local newRace = {}
			local nbTurns = math.floor(_nbTurns)
			local name = _name
			local tagsName = ATMEFct.decompString(_name)
		
			local index = ATMEDatas.C_Race_MT:add(doorsIndex)
			local localMembers = ATMEDatas.localMembers[index]
			
			setmetatable(newRace, ATMEDatas.C_Race_MT.mt)
		
			function newRace.getIndex(self)
				return index
			end
			
			function newRace.getName(self)
				return name
			end
			
			function newRace.tagGetByID(self, _position)
				local fctDefinition = ATMEFct.getErrorPrefix(newRace) .. "tagGetByID(position)"
				
				if type(_position) == "number" then
					local pos = math.floor(_position)
					if tagsName ~= nil and pos >= 1 and pos <= #tagsName then
						return tagsName[pos]
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "position type must be number", true)
				end
			end
			
			function newRace.tagGetCount(self)
				if tagsName ~= nil then
					return #tagsName
				else
					return 0
				end
			end
			
			function newRace.tagCompare(self, _tagsToCompare)
				local fctDefinition = ATMEFct.getErrorPrefix(newRace) .. "tagCompare(tagsToCompare)"
				
				if type(_tagsToCompare) == "string" then
					if tagsName ~= nil then
						local ok = ATMEFct.compareTags(tagsName, _tagsToCompare)
						
						if ok ~= nil then
							return ok
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare format must be 'T1|T2|..|Tn'", false)
						end
					else
						return false
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "tagsToCompare type must be string", true)
				end
			end
		
			function newRace.getNbTurns(self)
				return nbTurns
			end
			
			function newRace.getNbDoors(self)
				return doors:getCount()
			end
			
			function newRace.getUserPrivateDatas(self)
				return localMembers.userPrivateDatas[ATMEDatas.currentModuleName]
			end
			
			function newRace.getDoorSides(self, _index)
				local fctDefinition = ATMEFct.getErrorPrefix(newRace) .. "getDoorSides(index)"
				
				if type(_index) == "number" then
					local index = math.floor(_index)
					if index > 0 and localMembers.doorsIndex[index] ~= nil then 
						local leftObject = localMembers.doorsIndex[index].leftObject
						local rightObject = localMembers.doorsIndex[index].rightObject
						
						return leftObject, rightObject
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Race " .. name .. " - Door #" .. index .. " not found", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "index type must be number", true)
				end
			end
			
			function newRace.getLapAtDoor(self, _index)
				local fctDefinition = ATMEFct.getErrorPrefix(newRace) .. "getLapAtDoor(index)"
				
				if type(_index) == "number" then
					local index = math.floor(_index)
					if index > 0 and localMembers.doorsIndex[index] ~= nil then 
						return localMembers.doorsIndex[index].lap
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Door #" .. index .. " not found", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "index type must be number", true)
				end
			end
			
			function newRace.setLapAtDoor(self, _index)
				local fctDefinition = ATMEFct.getErrorPrefix(newRace) .. "setLapAtDoor(index)"
				
				if type(_index) == "number" then
					local index = math.floor(_index)
					if index > 0 and localMembers.doorsIndex[index] ~= nil then 
						if localMembers.doorsIndex[index].id > 1 and  localMembers.doorsIndex[index].id < self:getNbDoors() then
							localMembers.doorsIndex[index].lap = true
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "Race " .. name .. " - No lap on first or last door", false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Race " .. name .. " - Door #" .. index .. " not found", false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "index type must be number", true)
				end
			end
			
			function newRace.getDoorIndex(self, _offset)
				local fctDefinition = ATMEFct.getErrorPrefix(newRace) .. "getDoorIndex(offset)"
				
				if type(_offset) == "number" then
					local offset = math.floor(_offset)
					if offset > 0 and _offset <= doors:getCount() then
						return doors:get(offset).index
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Race " .. name .. " - " .. _offset .. " must be > 0 and <= " .. doors:getCount(), false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "offset type must be number", true)
				end
			end
			
			function newRace.getPlayerNextDoorIndex(self, _player)
				local fctDefinition = ATMEFct.getErrorPrefix(newRace) .. "getPlayerNextDoorIndex(player)"
				
				if ATMEFct.isObjectClass(_player, ATME.C_Player) == true then
					if localMembers.players[_player:getName()] ~= nil then
						return self:getDoorIndex(localMembers.players[_player:getName()].nextDoor)
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "player " .. _player:getPseudo() .. " not in race " .. raceName, false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "player type must be ATME.C_Player", true)
				end
			end
			
			function newRace.displayForAllPlayers(self, _text, _duration, _clear)
				local fctDefinition = ATMEFct.getErrorPrefix(newRace) .. "displayForAllPlayers(text, duration, [clear])"

				if type(_text) == "string" and type(_duration) == "number" and
									(_clear == nil or type(_clear) == "boolean") then
					local clear = false
			
					if type(_clear) == "boolean" then
						clear = _clear
					end
					
					for _name, _player in pairs(localMembers.players) do
						local localMembersPlayer = ATMEDatas.localMembers[_player.player:getIndex()]
						trigger.action.outTextForGroup(localMembersPlayer.groupId, _text, _duration, clear)
					end
				else
					if type(_text) ~= "string" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", true)
					elseif type(_duration) ~= "number" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "duration type must be number", false)
					elseif _clear ~= nil and type(_clear) ~= "boolean" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "clear type must be nil or string", false)
					end
				end
			end
			
			function newRace.soundForAllPlayers(self, _text)
				local fctDefinition = ATMEFct.getErrorPrefix(newRace) .. "soundForAllPlayers(text)"
				
				
				if type(_text) == "string" then
					local err, playlist = ATMEFct.C_Sound_getPlaylist(_text, false)

					if err == true then
						ATMEFct.errorOutputCore(nil, fctDefinition, playlist, false)
						return
					end

					if playlist == nil then
						ATMEFct.errorOutputCore(nil, fctDefinition, "text string has no sound token", false)
						return
					end
		
					for _, _player in pairs(localMembers.players) do
						ATMEFct.C_Sound_initPlaylist({
							player = _player, 
							randomRead = false, 
							readOnce = true, 
							playlist = playlist,
							timeSound = timer.getTime(),
						})
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "text type must be string", true)
				end
			end
			
			function newRace.addPlayer(self, _player)
				local fctDefinition = ATMEFct.getErrorPrefix(newRace) .. "addPlayer(player)"
				
				if ATMEFct.isObjectClass(_player, ATME.C_Player) == true then
					local found = false
					local raceName = ""
					for _, _race in pairs(ATMEDatas.races) do
						local localMembersLoop = ATMEDatas.localMembers[_race:getIndex()]
						
						if localMembersLoop.players[_player:getName()] ~= nil then
							found = true
							raceName = _race:getName()
						end
					end
					
					if found == false then
						newPlayer = {
							player = _player,
							bestTime = -1,
							recordTime = 0,
						}
						
						table.insert(localMembers.bestPlayers, newPlayer)
						
						localMembers.players[_player:getName()] = {
							player = _player,
							startTime = 0,
							startTurnTime = 0,
							bestTurnTime = -1,
							bestTime = -1,
							penalty = 0,
							nextDoor = 1,
							currentTurn = 1,
							oldPosition = _player:getPosition(),
						}
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "player " .. _player:getPseudo() .. " already in race " .. raceName, false)
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "player type must be ATME.C_Player", true)
				end
			end
	
			function newRace.removePlayer(self, _player)
				local fctDefinition = ATMEFct.getErrorPrefix(newRace) .. "removePlayer(player)"
				if ATMEFct.isObjectClass(_player, ATME.C_Player) == true then
					if localMembers.players[_player:getName()] ~= nil then
						local index = 1
						while index <= #localMembers.bestPlayers do
							if localMembers.bestPlayers[index].player == _player then
								table.remove(localMembers.bestPlayers, index)
							else
								index = index + 1
							end
						end
						
						localMembers.players[_player:getName()] = nil
					end
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "player type must be ATME.C_Player", true)
				end
			end
	
			function newRace.removeAllPlayers(self)
				for _name, _ in pairs(localMembers.players) do
					localMembers.players[_name] = nil
				end
				
				while #localMembers.bestPlayers > 0 do
					table.remove(localMembers.bestPlayers)
				end
			end
	
			function newRace.isPlayerInRace(self, _player)
				local fctDefinition = ATMEFct.getErrorPrefix(newRace) .. "isPlayerInRace(player)"
				
				if ATMEFct.isObjectClass(_player, ATME.C_Player) == true then
					return localMembers.players[_player:getName()] ~= nil
				else
					ATMEFct.errorOutputCore(nil, fctDefinition, "player type must be ATME.C_Player", true)
				end
			end
	
			function newRace.getRanking(self)
				local ranking = {}
				
				for _, _item in ipairs(localMembers.bestPlayers) do
					local item = {
						player = _item.player,
						bestTime = _item.bestTime,
					}
					
					table.insert(ranking, item)
				end
				
				return ranking
			end
	
			function newRace.setMaxAltitudeRule(self, _maxAltitude, _penalty)
				local fctDefinition = ATMEFct.getErrorPrefix(newRace) .. "setMaxAltitudeRule(maxAltitude, penalty)"
				
				if type(_maxAltitude) == "number" and type(_penalty) == "number" then
					if _maxAltitude > 0 then
						localMembers.maxAltitude = _maxAltitude
						
						if _penalty >= 0 then
							localMembers.maxAltitudePenalty = _penalty
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "penalty must be >= 0", false)
						end
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Bad altitude max", false)
					end
				else
					if type(_maxAltitude) ~= "number" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "maxAltitude type must be number", true)
					elseif type(_penalty) ~= "number" then
						ATMEFct.errorOutputCore(nil, fctDefinition, "penalty type must be number", false)
					end
				end
			end
			
			function newRace.resetMaxAltitudeRule(self)
				localMembers.maxAltitude = -1
				localMembers.maxAltitudePenalty = 0
			end
	
			function newRace.getMaxAltitudePenalty(self)
				return localMembers.maxAltitudePenalty
			end
	
			function newRace.getMaxAltitude(self)
				return localMembers.maxAltitude
			end
			
			for _, _localModuleMembers in pairs(ATMEDatas.modulesLocalMembers) do
				localMembers.userPrivateDatas[_localModuleMembers.name] = {}
			end
			
			ATMEDatas.races[name] = newRace
			return newRace
		else
			ATMEFct.errorOutputCore(nil, fctDefinition, "Race " .. _name .. " must have two doors at least", false)
		end
	end
end

setmetatable(ATME.C_Race,ATMEDatas.C_Race_MT)

ATMEDatas.C_Race_MT.__call = function(t, _name, _leftSideDoorName, _rightSideDoorName, _nbTurns)
	local fctDefinition = "Constructor ATME.C_Race(name, leftSideDoorName, righSideDoorName, nbTurns)"
	
	if ATMEDatas.racesActivation == true then
		if type(_name) == "string" and type(_leftSideDoorName) == "string" and type(_rightSideDoorName) == "string" and type(_nbTurns) == "number" then
			local d,f = string.find(_name, "|", 1)
			if d == nil then
				if ATMEDatas.races[_name] == nil then
					if _nbTurns >= 1 then
						return ATMEFct.C_Race_create(_name, _leftSideDoorName, _rightSideDoorName, math.floor(_nbTurns))
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "Race " .. _name .. " - nbTurns must be >= 1", false)
					end
				else
					return nil
				end
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "name can't contain '|' character", false)
			end
		else
			if type(_name) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "name type must be string", false)
			elseif type(_leftSideDoorName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "leftSideDoorName type must be string", false)
			elseif type(_rightSideDoorName) ~= "string" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "rightSideDoorName type must be string", false)
			elseif type(_nbTurns) ~= "number" then
				ATMEFct.errorOutputCore(nil, fctDefinition, "nbTurns type must be number", false)
			end
		end
	else
		ATMEFct.errorOutputCore(nil, fctDefinition, "Race activation parameter of ATME.run is set to RACE_OFF (default)", false)
	end
end

ATMEDatas.C_Race_MT.mt.__metatable = "Private"
ATMEDatas.C_Race_MT.mt.__index = ATME.C_Race
ATMEDatas.C_Race_MT.mt.__newindex = function(t, k, v)
	if ATMEDatas.C_Race_MT.keys[k] ~= nil then 
		rawset(t, k, v)
	else
		ATMEFct.errorOutputCore(nil, "Function ATME.C_Race : Bad key " .. k, "table locked", false)
	end
end

----------------------------------------------------------------------------------------------------------
-- FIN DE DEFINITION DES CLASSES D'OBJETS DEFINIS DANS ATME
--
-- FONCTIONS DE GESTION ATME_CORE
----------------------------------------------------------------------------------------------------------

do
	----------------------------------------------------------------------------------------------------------
	-- Fonction de mise à jour (appelé par les 2 timers _timerMainFunction et _timer100x) des joueurs actifs :
	--		-> Tous les joueurs actifs dans la liste ATME au coup précédent sont listés
	--		-> On ajuste avec les joueurs toujours actifs (fonction DCS coalition.getPlayers)
	--		-> Les joeurs que ne sont pas actifs sont détruits de la liste ATME.
	-- Cette fonction est importante car ATME ne se base que sur l'événement DCS S_EVENT_BIRTH. En cas de changement
	-- de slot par exemple, cet événement est émis. Mais pas la destruction. 
	-- Les événements S_EVENT_PLAYER_ENTER ou S_EVENT_PLAYER_LEAVE ont buggé par le passé en multi,
	-- donc pas de gestion de ces événements pour la mise à jour interne ATME.
	--
	-- Elle permet lors dudit timer d'éviter les incohérences qui pourraient se produire dans la liste ATME joueurs.
	----------------------------------------------------------------------------------------------------------
	
	ATMEFct.patrolConditionFct = {
		["NO_COND"] = function(_patrol)
			_patrol.isArmed = true
			return true
		end,
		["MAX_TURN_COND"] = function(_patrol)
			_patrol.nbTurns = _patrol.nbTurns + 1
								
			if _patrol.nbTurns < _patrol.maxTurns then
				_patrol.isArmed = true
				return true
			else
				return false
			end
		end,
		["MAX_TIMER_COND"] = function(_patrol)
			if _patrol.maxTime >= timer.getTime() then
				_patrol.isArmed = true
				return true
			else
				return false
			end
		end,
		["RANDOM_COND"] = function(_patrol)
			if math.random() <= _patrol.maxRandom then
				_patrol.isArmed = true
				return true
			else
				return false
			end
		end,
		["FLAG_COND"] = function(_patrol)
			if ATME.getFlag(_patrol.flag) == 0 then
				_patrol.isArmed = true
				return true
			else
				return false
			end
		end,
	}

	local startMissionFcts = {
		[0] = function()
			-- Délai d'attente pour déclarer les initialisations OK : Actuellement 3 secondes pour considérer les groupes créés
			ATMEDatas.startOffset = ATMEDatas.startOffset + 1
			if ATMEDatas.startOffset == 2 then
				ATMEDatas.startMissionOk = 1
			end
		end,
		[1] = function()
			-- Appel du module onStartMission après initialisation
			for _, _name in ipairs(ATMEDatas.modulesNames) do
				local localModuleMembers = ATMEDatas.modulesLocalMembers[_name]
				if localModuleMembers.handlers.onStartMissionHandler ~= nil then
					ATMEDatas.currentModuleName = localModuleMembers.name
					
					-- Appel du module
					ATMEFct.outputCore(nil,"Appel onStartMissionHandler après initialisation des unités & joueurs du module " .. localModuleMembers.name, 2)
					xpcall(localModuleMembers.handlers.onStartMissionHandler, function() ATMEFct.errUserHandler('onStartMissionHandler') end)
				end
			end
			
			ATMEDatas.startMissionOk = 2
		end,
		[2] = function()
		end,
	}
	
	----------------------------------------------------------------------------------------------------------
	-- Fonction _timerMainFunction est appelée à l'initialisation voir plus bas, puis réarmée toutes les secondes.
	-- Cette fonction permet :
	--		- la mise à jour permanente des listes ATME des joueurs et unités AI.
	--      - la gestion des spawns multiples
	-- 		- la gestion des smokes (classe ATME.C_Smoke) qui doivent rester actifs et les tirs répétés de 
	-- 		   flares (classe ATME.C_Flare)
	-- 		- la gestion les radios transmissions permanentes affectées aux unités AI (classe ATME.AIUnit)
	--		- la gestion des sons répétés associés aux joueurs (classe ATME.C_Player)
	--      - la gestion d'entrée/sortie des zones managées (classe ATME.C_ZonesMgr)
	--      - la gestion des pickupZones et des signaux éventuels
	--		- l'appel pour chaque module du handler onTimerHandler
	--		- l'appel des onCreateHandlers suite à la création des unités et groupes, suivi éventuellement d'un onSpawnHandler
	-- 		- l'appel des fonctions de mise des joueurs
	--		- après ces appels, la liste des events actifs est remise à 0.
	--		- Traitement des event core retardés. Certains evénéments sont retardés pour garantir leur bonne prise en 
	--        par DCS.
	--      - Gestion des routes pour les patrouilles
	--
	----------------------------------------------------------------------------------------------------------
	
	local _timerMainFunction
    _timerMainFunction = function()
		if ATMEDatas.exitATME == false then
			-- Fonction des mises à jour de la liste des joueurs ... 
			-- Important pour éviter les incohérences, les timers étant asynchrones
			
			ATMEFct.outputCore(nil,"Start main timer", 2)
			
			local currentTipsCount = math.floor(timer.getTime())
			
			-- Pour les groupes créés ou spawnés en attente, on enlève un cycle et on purge les groupes détruits entre temps
			
			for _name, _ in pairs(ATMEDatas.newCreatedGroups) do
				if ATMEDatas.newCreatedGroups[_name].timerOffset > 0 then
					ATMEDatas.newCreatedGroups[_name].timerOffset = ATMEDatas.newCreatedGroups[_name].timerOffset - 1
				else
					ATMEDatas.groups[_name] = ATMEDatas.newCreatedGroups[_name].group
					local localMembersGroup = ATMEDatas.localMembers[ATMEDatas.groups[_name]:getIndex()]
							
					for _unitName, _unit in pairs(localMembersGroup.ATMEUnits) do
						ATMEDatas.allAIUnitsAndPlayers[_unitName] = _unit
						
						if _unit:getClassName() == "C_AIUnit" then
							ATMEDatas.AIUnits[_unitName] = _unit
						else
							ATMEDatas.players[_unitName] = _unit
							ATMEDatas.nbPlayers = ATMEDatas.nbPlayers + 1
						end
					end
					
					ATMEDatas.newPreparedGroups[_name] = ATMEDatas.newCreatedGroups[_name]
					ATMEDatas.newCreatedGroups[_name] = nil
				end
			end
			
			for _name, _ in pairs(ATMEDatas.newCreatedCargos) do
				ATMEDatas.cargos[_name] = ATMEDatas.newCreatedCargos[_name].cargo
				ATMEDatas.newPreparedCargos[_name] = ATMEDatas.newCreatedCargos[_name]
				ATMEDatas.newCreatedCargos[_name] = nil
			end
			
			for _name, _ in pairs(ATMEDatas.newCreatedStatics) do
				ATMEDatas.staticObjects[_name] = ATMEDatas.newCreatedStatics[_name].object
				ATMEDatas.newPreparedStatics[_name] = ATMEDatas.newCreatedStatics[_name]
				ATMEDatas.newCreatedStatics[_name] = nil
			end
			
			startMissionFcts[ATMEDatas.startMissionOk]()
			
			if ATMEDatas.startMissionOk == 2 then
				local delay = 0.5
				for _name, _item in pairs(ATMEDatas.newPreparedGroups) do
					if ATMEDatas.groups[_name]:isActivated() == true then
						local localMembersGroup = ATMEDatas.localMembers[ATMEDatas.groups[_name]:getIndex()]
						
						if localMembersGroup.justActivated ~= nil then
							localMembersGroup.justActivated = nil
						else
							ATMEFct.C_Group_CreateTrackingWaypoints(localMembersGroup, true)
							
							localMembersGroup.idTimer = timer.scheduleFunction(ATMETimers.C_Group_timerTrackingWaypoint, ATMEDatas.groups[_name], timer.getTime() + delay)
							
							for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
								local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]	
								ATMEDatas.currentModuleName = _nameModule
								
								if _item.disembark == false then
									for _, _unit in pairs(localMembersGroup.ATMEUnits) do
										if _unit:getClassName() == "C_AIUnit" and localModuleMembers.handlers.onCreateAIUnitHandler ~= nil then
											ATMEFct.outputCore(nil,"Appel onCreateAIUnitHandler du module " .. localModuleMembers.name, 2)
											xpcall(function() localModuleMembers.handlers.onCreateAIUnitHandler(_unit) end, function() ATMEFct.errUserHandler('onCreateAIUnitHandler') end)
										elseif _unit:getClassName() == "C_Player" and localModuleMembers.handlers.onCreatePlayerHandler ~= nil then
											ATMEFct.outputCore(nil,"Appel onCreatePlayerHandler du module " .. localModuleMembers.name, 2)
											xpcall(function() localModuleMembers.handlers.onCreatePlayerHandler(_unit) end, function() ATMEFct.errUserHandler('onCreatePlayerHandler') end)
										end
									end

									if localModuleMembers.handlers.onCreateGroupHandler ~= nil then
										ATMEFct.outputCore(nil,"Appel onCreateGroupHandler du module " .. localModuleMembers.name, 2)
										xpcall(function() localModuleMembers.handlers.onCreateGroupHandler(ATMEDatas.groups[_name], ATMEDatas.respawnedGroups[_name] ~= nil) end, function() ATMEFct.errUserHandler('onCreateGroupHandler') end)
									end
								end
							end

							for _unitName, _unit in pairs(localMembersGroup.ATMEUnits) do
								if ATMEConst.canFly[localMembersGroup.category] == true and _unit:isEngineStarted() == true then
									-- Emission de l'événement démarrage moteur
									local event = {
										type = "ENGINE_START",
										datas = {
											unitName = _unitName,
										},
									}
									
									ATMEFct.outputCore(nil, "Initial Start engine pour " .. _unitName, 2)
									ATMEFct.C_Event_add(event)
								end
							end

							if _item.disembark == false then
								if _item.isSpawned == true then
									for _moduleName, _callback in pairs(_item.spawnCallback) do
										ATMEDatas.currentModuleName = _moduleName
											
										ATMEFct.outputCore(nil,"Appel spawnCallback cloning context du module " .. _moduleName .. " pour le groupe " .. _name, 2)
										xpcall(function() _callback(_item.cloningContext, ATMEDatas.groups[_name]) end, function() ATMEFct.errUserHandler('spawnCallback cloning context pour le groupe ' .. _name) end)
									end
									
									for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
										local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]
										ATMEDatas.currentModuleName = _nameModule
										
										if localModuleMembers.handlers.onSpawnGroupHandler ~= nil then
											ATMEFct.outputCore(nil,"Appel de onSpawnGroupHandler du module " .. localModuleMembers.name, 2)
											xpcall(function() localModuleMembers.handlers.onSpawnGroupHandler(ATMEDatas.groups[_name]) end, function() ATMEFct.errUserHandler('onSpawnGroupHandler') end)
										end
									end
								end
							else 	-- Groupe débarqué, on prépare l'event de fin de débarquement
								ATMEFct.C_Event_add(_item.event)
							end
							
							ATMEDatas.newPreparedGroups[_name] = nil
							delay = delay + 0.1
						end
					end
				end
				
				for _name, _item in pairs(ATMEDatas.newPreparedCargos) do
					for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
						local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]
						ATMEDatas.currentModuleName = _nameModule
						
						if localModuleMembers.handlers.onCreateCargoHandler ~= nil then
							ATMEFct.outputCore(nil,"Appel onCreateCargoHandler du module " .. localModuleMembers.name .. " pour le cargo " .. _name, 2)
							xpcall(function() localModuleMembers.handlers.onCreateCargoHandler(ATMEDatas.cargos[_name]) end, function() ATMEFct.errUserHandler('onCreateCargoHandler') end)
						end
					end
					
					if _item.isSpawned == true then
						for _moduleName, _callback in pairs(_item.spawnCallback) do
							ATMEDatas.currentModuleName = _moduleName
								
							ATMEFct.outputCore(nil,"Appel spawnCallback cargo cloning context du module " .. _moduleName .. " pour le cargo " .. _name, 2)
							xpcall(function() _callback(_item.cargoCloningContext, ATMEDatas.cargos[_name]) end, function() ATMEFct.errUserHandler('spawnCallback cargo context pour le static ' .. _name) end)
						end
						
						for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
							local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]
							ATMEDatas.currentModuleName = _nameModule
							
							if localModuleMembers.handlers.onSpawnCargoHandler ~= nil then
								ATMEFct.outputCore(nil,"Appel de onSpawnCargoHandler du module " .. localModuleMembers.name, 2)
								xpcall(function() localModuleMembers.handlers.onSpawnCargoHandler(ATMEDatas.cargos[_name]) end, function() ATMEFct.errUserHandler('onSpawnCargoHandler') end)
							end
						end
					end
					
					ATMEDatas.newPreparedCargos[_name] = nil
				end
				
				for _name, _item in pairs(ATMEDatas.newPreparedStatics) do
					for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
						local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]
						ATMEDatas.currentModuleName = _nameModule
						
						if localModuleMembers.handlers.onCreateStaticObjectHandler ~= nil then
							ATMEFct.outputCore(nil,"Appel onCreateStaticObjectHandler du module " .. localModuleMembers.name .. " pour le static " .. _name, 2)
							xpcall(function() localModuleMembers.handlers.onCreateStaticObjectHandler(ATMEDatas.staticObjects[_name]) end, function() ATMEFct.errUserHandler('onCreateStaticObjectHandler') end)
						end
					end
					
					if _item.isSpawned == true then
						for _moduleName, _callback in pairs(_item.spawnCallback) do
							ATMEDatas.currentModuleName = _moduleName
								
							ATMEFct.outputCore(nil,"Appel spawnCallback static cloning context du module " .. _moduleName .. " pour le static " .. _name, 2)
							xpcall(function() _callback(_item.staticCloningContext, ATMEDatas.staticObjects[_name]) end, function() ATMEFct.errUserHandler('spawnCallback static cloning context pour le static ' .. _name) end)
						end
						
						for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
							local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]
							ATMEDatas.currentModuleName = _nameModule
							
							if localModuleMembers.handlers.onSpawnStaticObjectHandler ~= nil then
								ATMEFct.outputCore(nil,"Appel de onSpawnStaticObjectHandler du module " .. localModuleMembers.name, 2)
								xpcall(function() localModuleMembers.handlers.onSpawnStaticObjectHandler(ATMEDatas.staticObjects[_name]) end, function() ATMEFct.errUserHandler('onSpawnStaticObjectHandler') end)
							end
						end
					end
					
					ATMEDatas.newPreparedStatics[_name] = nil
				end
			end
			
			-- Affichage du nombre de joueurs
			if ATMEDatas.oldNbPlayers ~= ATMEDatas.nbPlayers then
				ATMEFct.outputCore(nil,"Nb players = " .. ATMEDatas.nbPlayers, 2)
				ATMEDatas.oldNbPlayers = ATMEDatas.nbPlayers
			end
			
			if ATMEDatas.startMissionOk == 2 then
				-- Gestion disable des groupes et unités
				ATMEFct.outputCore(nil, "Gestion disable des groupes et unités", 2)
				for _name, _unit in pairs(ATMEDatas.unitsToDisable) do
					ATMEFct.C_Player_C_AIUnit_disable(_unit)
					ATMEDatas.unitsToDisable[_name] = nil
				end
				
				for _name, _group in pairs(ATMEDatas.groupsToDisable) do
					ATMEFct.C_Group_disable(_group)
					ATMEDatas.groupsToDisable[_name] = nil
				end
				
				for _name, _object in pairs(ATMEDatas.staticsToDisable) do
					ATMEFct.C_StaticObject_disable(_object)
					ATMEDatas.staticsToDisable[_name] = nil
				end
				
				for _name, _object in pairs(ATMEDatas.cargosToDisable) do
					ATMEFct.C_Cargo_disable(_object)
					ATMEDatas.cargosToDisable[_name] = nil
				end
				
				-- Appel de la callback lors de l'affectation d'une nouvelle route
				for _name, _ in pairs(ATMEDatas.groupsWithNewRouteSet) do
					if ATMEDatas.groups[_name] ~= nil then
						local localMembersGroup = ATMEDatas.localMembers[ATMEDatas.groups[_name]:getIndex()]
						
						for _moduleName, _callback in pairs(localMembersGroup.newRouteCallback) do
							ATMEDatas.currentModuleName = _moduleName
								
							ATMEFct.outputCore(nil,"Appel de newRouteCallback du groupe " .. _name .. " pour le module " .. _moduleName, 2)
							xpcall(function() _callback(ATMEDatas.groups[_name]) end, function() ATMEFct.errUserHandler('newRouteCallback du groupe ' .. _name) end)
						end
					end
					
					ATMEDatas.groupsWithNewRouteSet[_name] = nil
				end
				
				-- Gestion du respawn automatique après destruction
				for _name, _item in pairs(ATMEDatas.groupsToRespawn) do
					if _item.time <= currentTipsCount then
						if _item.spawnDatas ~= nil then

							ATMEFct.C_CloningContext_spawn(_item.spawnDatas)
						end
						
						ATMEDatas.groupsToRespawn[_name]= nil
					end
				end
				
				-- Activation des taches missions avec délai
				for _name, _ in pairs(ATMEDatas.missionsDelayed) do
					local group = ATMEDatas.groups[_name]
						
					if group ~= nil then
						if group:isControlled() == true then
							if ATMEDatas.missionsDelayed[_name].delay > 0 then
								ATMEDatas.missionsDelayed[_name].delay = ATMEDatas.missionsDelayed[_name].delay - 1
							else
								local localMembers = ATMEDatas.localMembers[group:getIndex()]
				
								localMembers.routeUncontrolled = false
								localMembers.routeDelay = ATMEConst.basicRouteDelay
								
								ATMEFct.C_CloningContext_setMission(group, ATMEDatas.missionsDelayed[_name].waypoints, false)
						
								if ATMEDatas.missionsDelayed[_name].patrolToUpdate ~= nil then
									ATMEDatas.patrolToUpdate[_name] = ATMEDatas.missionsDelayed[_name].patrolToUpdate
								end
								
								ATMEDatas.missionsDelayed[_name] = nil
							end
						end
					else
						ATMEDatas.missionsDelayed[_name] = nil
					end
				end
				
				-- Gestion des spawns multiples
				ATMEFct.outputCore(nil, "Gestion des spawn multiples", 2)
				for _key, _item in pairs(ATMEDatas.spawnGroupDatasRepeat) do
					if _item.nbSpawn >= _item.maxQuantity then
						ATMEDatas.spawnGroupDatasRepeat[_key] = nil
					elseif _item.currentTime ~= nil then
						if _item.currentTime > 0 then
							_item.currentTime = _item.currentTime - 1
						else
							if _item.repeatRandom == true then
								_item.currentTime = math.random(_item.repeatPeriodMin, _item.repeatPeriodMax)
							else
								_item.currentTime = _item.repeatPeriodMin
							end

							_item.nbSpawn = _item.nbSpawn + 1

							_item.groupDatas:spawn(_item.arg)
						end
					end
				end
				
				-- Vérification de l'existance des unités & groupes liés aux events et
				-- récupération de l'unité/groupe à partir de son nom sinon suppression
				for _id, _event in events[activeEvents]:pairs() do
					ATMEFct.C_Event_update(_event, true)
				end
				

				-- Bug DCS contournement pour les loop radioTransmission
				-- On relance le loop si au moins un joueur s'est connecté
				--if ATMEDatas.newPlayer == true then
				--	ATMEDatas.autoRestartDCS = true
				--	ATMEDatas.newPlayer = false
				--end

				-- On fige les événements qui vont être générés et émis aux handlers.
				-- On bascule activeEvents sur le nouveau pour prendre en compte au prochain coup les événements
				-- éventuellement générés par ces mêmes handlers (OnTimer)
				local fixedEvents = activeEvents
				activeEvents = 1 - activeEvents
				
				-- Gestion des handlers des events
				ATMEFct.outputCore(nil, "Gestion des événements", 2)
				for _id, _event in events[fixedEvents]:pairs() do
					ATMEFct.C_Event_fire(_event)
				end
					
				-- Déclenchement des handlers
				ATMEFct.outputCore(nil, "Gestion des Handlers par module", 2)
				for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
					local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]
					ATMEDatas.currentModuleName = _nameModule
					
					if localModuleMembers.handlers.onTimerHandler ~= nil then
						ATMEFct.outputCore(nil,"Appel onTimerHandler du module " .. localModuleMembers.name , 2)
						xpcall(localModuleMembers.handlers.onTimerHandler, function() ATMEFct.errUserHandler('onTimerHandler') end)
					end
				end
				
				for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
					local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]
					ATMEDatas.currentModuleName = _nameModule
					
					if localModuleMembers.handlers.onUpdatePlayerHandler ~= nil then
						for _, _player in pairs(ATMEDatas.players) do
						
							-- Appel du module
							ATMEFct.outputCore(nil,"Appel onUpdatePlayerHandler du module " .. localModuleMembers.name, 2)
							xpcall(function() localModuleMembers.handlers.onUpdatePlayerHandler(_player) end, function() ATMEFct.errUserHandler('onUpdatePlayerHandler') end)
						end
					end
				end
				
				ATMEFct.C_Event_removeAll(fixedEvents)
				
				-- Traitement des patrouilles
				ATMEFct.outputCore(nil,"Traitement des patrouilles", 2)
				for _name, _patrol in pairs(ATMEDatas.patrolsToUpdate) do
					local group = ATMEDatas.groups[_name]
					local localMembersGroup = ATMEDatas.localMembers[group:getIndex()]

					if group ~= nil then
						local idCommand
						
						if ATMEConst.canFly[localMembersGroup.category] == false then
							idCommand = 'GoToWaypoint'
						else
							idCommand = 'SwitchWaypoint'
						end
						
						if _patrol.isArmed == true then
							local controller = group:getDCSGroup():getController()
							local command = {
								id = 'SwitchWaypoint', 
								params = { 
									fromWaypointIndex = _patrol.last,  
									goToWaypointIndex = _patrol.restart, 
								}, 
							}
							
							controller:setCommand(command)
							
							ATMEDatas.patrolsToUpdate[_name].isArmed = false
						end
					else
						ATMEDatas.patrolsToUpdate[_name] = nil
					end
				end
			end
			
			ATMEFct.outputCore(nil,"End main timer", 2)
			timer.scheduleFunction(_timerMainFunction, {}, timer.getTime() + 1)
		end
	end
	
	----------------------------------------------------------------------------------------------------------
	-- Fonction _timer100x est appelée à l'initialisation voir plus bas puis réarmée tous les 1/100è de seconde.
	-- Cette fonction permet de mesurer les temps au 1/100è de seconde. Elle déclenche les events ATME :
	-- 	RACE_START : quand un joueur de la course franchit le départ
	--	RACE_TURN_TIME : quand un joueur boucle un tour par un nouveau franchissement de la ligne de départ,
	--					 pour les courses à plusieurs tours
	--	RACE_PLAYER_BEST_TURN_TIME : Meilleur temps au tour du joueur qui boucle son tour
	-- 	RACE_BEST_TURN_TIME : Record du tour pour la course depuis le début de la mission
	-- 	RACE_LAP_TIME : Temps intermédiaire pour le joueur franchissant une porte définie.
	--	RACE_BEST_LAP_TIME : Meilleur temps intermédiaire pour une porte définie depuis le début de la mission
	-- 	RACE_PLAYER_TOO_HIGH : Le joueur a dépassé l'altitude maximale autorisée
	--	RACE_FINAL_TIME : Temps final pour la course pour le joueur venant de franchir l'arrivée (dernière
	--					  porte de la course).
	--	RACE_PLAYER_BEST_FINAL_TIME : Meilleur temps final pour le joueur venant de franchir l'arrivée
	--	RACE_BEST_FINAL_TIME : Meilleur temps final depuis le début de la mission. Le classement général est
	--						   recalculé à partir de ce temps.
	--
	-- Le passage d'une porte est pris en compte que si la porte précédente a été passée (franchie ou loupée)
	--
	-- ATTENTION : Il y a un décallage de temps entre la création du cor event et l'activation vers les modules.
	-- 			   On ne stocke que le nom dde l'unité joueur. La vérification de l'existance de l'unité ne se 
	-- 			   fera que lors de l'activation de l'événement dans MainTimer.
	--			   En effet, l'unité peut être détruite entre temps.
	----------------------------------------------------------------------------------------------------------
	local _timer100x
	_timer100x = function()
		if ATMEDatas.exitATME == false then
			-- Fonction des mises à jour de la liste des joueurs ... 
			-- Important pour éviter les incohérences, les timers étant asynchrones
			
			local currentTime = timer.getTime()
			for _, _race in pairs(ATMEDatas.races) do
				local localMembers = ATMEDatas.localMembers[_race:getIndex()]
				local maxAltitude = _race:getMaxAltitude()
				for _rowBestPlayer, _item in ipairs(localMembers.bestPlayers) do
					local racePlayer = localMembers.players[_item.player:getName()]
					
					local playerPosition = racePlayer.player:getPosition()
				
					local okSegment, side = ATMEFct.C_Segment2D_intersect(localMembers.doorsIndex[_race:getDoorIndex(racePlayer.nextDoor)].segment, racePlayer.oldPosition, playerPosition)

					if okSegment == true and side == -1 then
						-- Porte franchie
						-- ATMEFct.outputCore(nil, "Porte franchie", 2)
						if racePlayer.nextDoor == 1 and racePlayer.currentTurn == 1 then
							if racePlayer.player:getAGLAltitude() <= maxAltitude or maxAltitude == -1 then
							-- Franchissement de la ligne de départ
							
							racePlayer.startTime = currentTime
							racePlayer.startTurnTime = currentTime
							racePlayer.penalty = 0
							
							ATMEFct.C_Race_createEvent(_race, 
								"RACE_START", 
								{
									unitName = racePlayer.player:getName(),
								})
							else
								racePlayer.nextDoor = 0
							end
						elseif racePlayer.player:getAGLAltitude() > maxAltitude and maxAltitude ~= -1 then
							-- Test de l'altitude si on est en course (nextDoors <> 1 ou currentTurn <> 1)

							racePlayer.penalty = racePlayer.penalty + _race:getMaxAltitudePenalty()

							ATMEFct.C_Race_createEvent(_race, 
								"RACE_PLAYER_TOO_HIGH",
								{
									unitName = racePlayer.player:getName(),
									penalty = racePlayer.penalty,
								})
						end
					
						local totalPlayTime = currentTime - racePlayer.startTime + racePlayer.penalty
						
						if racePlayer.nextDoor == 1 and racePlayer.currentTurn > 1 then
							-- Franchissement de la ligne de départ et bouclage d'un tour
							local playTime = currentTime - racePlayer.startTurnTime + racePlayer.penalty
							ATMEFct.C_Race_createEvent(_race, 
									"RACE_TURN_TIME", 
									{
										unitName = racePlayer.player:getName(),
										playTime = playTime,
										totalPlayTime = totalPlayTime,
										penalty = racePlayer.penalty,
										turnNumber = racePlayer.currentTurn - 1,
									})
							
							if playTime < racePlayer.bestTurnTime or racePlayer.bestTurnTime < 0 then
								racePlayer.bestTurnTime = playTime
								ATMEFct.C_Race_createEvent(_race, 
									"RACE_PLAYER_BEST_TURN_TIME",
									{
										unitName = racePlayer.player:getName(),
										playTime = playTime,
										turnNumber = racePlayer.currentTurn - 1
									})
							end
							
							if playTime < localMembers.bestTurn.bestTime or localMembers.bestTurn.bestTime < 0 then
								localMembers.bestTurn.bestTime = playTime
								localMembers.bestTurn.pseudo = racePlayer.player:getPseudo()
								ATMEFct.C_Race_createEvent(_race, 
									"RACE_BEST_TURN_TIME",
									{
										unitName = racePlayer.player:getName(),
										playTime = playTime,
										turnNumber = racePlayer.currentTurn - 1
									})
							end

							racePlayer.startTurnTime = currentTime + racePlayer.penalty
						elseif _race:getLapAtDoor(_race:getDoorIndex(racePlayer.nextDoor)) == true then
							-- Portes avec temps intermédiaire
							local door = localMembers.doorsIndex[_race:getDoorIndex(racePlayer.nextDoor)]
							local playTime = currentTime - racePlayer.startTurnTime + racePlayer.penalty
							
							ATMEFct.C_Race_createEvent(_race, 
								"RACE_LAP_TIME", 
								{
									unitName = racePlayer.player:getName(),
									doorNumber = racePlayer.nextDoor,
									playTime = playTime,
									totalPlayTime = totalPlayTime,
									penalty = racePlayer.penalty,
									turnNumber = racePlayer.currentTurn - 1																		
								})
								
							if playTime < door.bestTurn.bestTime or door.bestTurn.bestTime < 0 then
								door.bestTurn.bestTime = playTime
								door.bestTurn.pseudo = racePlayer.player:getPseudo()
								ATMEFct.C_Race_createEvent(_race, 
									"RACE_BEST_LAP_TIME", 
									{
										unitName = racePlayer.player:getName(),
										doorNumber = racePlayer.nextDoor,
										playTime = playTime, 
										turnNumber = racePlayer.currentTurn - 1
									})
							end
						elseif racePlayer.nextDoor == _race:getNbDoors() then
							-- Franchissement de la ligne d'arrivée
							racePlayer.nextDoor = 0
							
							if racePlayer.currentTurn == _race:getNbTurns() then
								if totalPlayTime < racePlayer.bestTime or racePlayer.bestTime < 0 then
										racePlayer.bestTime = totalPlayTime
										ATMEFct.C_Race_createEvent(_race, 
											"RACE_PLAYER_BEST_FINAL_TIME", 
											{
												unitName = racePlayer.player:getName(),
												totalPlayTime = totalPlayTime,
											})
								end
								
								if totalPlayTime < _item.bestTime or _item.bestTime < 0 then
									_item.bestTime = totalPlayTime
									_item.recordTime = currentTime
									
									ATMEFct.C_Race_createEvent(_race, 
										"RACE_BEST_FINAL_TIME",
										{
											unitName = racePlayer.player:getName(),
											totalPlayTime = totalPlayTime, 
										})
								end
								
								ATMEFct.C_Race_createEvent(_race, 
									"RACE_FINAL_TIME", 
									{
										unitName = racePlayer.player:getName(),
										totalPlayTime = totalPlayTime,
										penalty = racePlayer.penalty,
									})
								
								racePlayer.currentTurn = 0
							end
							
							racePlayer.currentTurn = racePlayer.currentTurn + 1
						end
						
						racePlayer.nextDoor = racePlayer.nextDoor + 1
					end
					
					racePlayer.oldPosition = playerPosition
				end
				
				table.sort(localMembers.bestPlayers, 
								function(a, b)
									if a.bestTime >= 0 and b.bestTime >= 0 and a.bestTime ~= b.bestTime then
										return a.bestTime < b.bestTime
									elseif a.bestTime < 0 and b.bestTime >= 0 then
										return false
									elseif a.bestTime >=0 and b.bestTime < 0 then
										return true
									else
										return a.recordTime < b.recordTime
									end
								end)
			end
			-- DEBUG ATMEFct.outputCore(nil, "Current time = " .. currentTime, 1)
			-- DEBUG Penser à baisser la fréquence si activation affichage
			timer.scheduleFunction(_timer100x, {}, timer.getTime() + 0.01)
		end
	end
	
	
	----------------------------------------------------------------------------------------------------------
	-- Fonction gestion événements DCS ... essentielle ... appelée si l'événement se produit
	-- En l'état ne gère pas tous les événements mais uniquement ceux nécessaires
	----------------------------------------------------------------------------------------------------------	

	ATMEFct.fireEvent = function(_event)
		local newEvent = ATMEFct.C_Event_create(_event)
		
		ATMEFct.C_Event_update(newEvent, false)
		ATMEFct.C_Event_fire(newEvent)
	end
	
	ATMEFct.hitTargetFct = {
		[Object.Category.UNIT] = function(_event)
			local event = {
				type = "HIT_TARGET",
				datas = {
					unitName = _event.initiator:getName(),
					targetName =  _event.target:getName(),
					targetClassName = ATMEDatas.allAIUnitsAndPlayers[_event.target:getName()]:getClassName(),
					weaponName = _event.weapon:getName(),
				},
			}		
			
			ATMEFct.fireEvent(event)
			ATMEFct.outputCore(nil, "Event HIT_TARGET généré pour unit " .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
		end,
		[Object.Category.STATIC] = function(_event)
			local event = {
				type = "HIT_TARGET",
				datas = {
					unitName = _event.initiator:getName(),
					targetName =  _event.target:getName(),
					targetClassName = ATMEDatas.staticObjects[_event.target:getName()]:getClassName(),
					weaponName = _event.weapon:getName(),
				},
			}

			ATMEFct.fireEvent(event)
			ATMEFct.outputCore(nil, "Event HIT_TARGET généré pour static object " .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
		end,
		[Object.Category.CARGO] = function(_event)
			local event = {
				type = "HIT_TARGET",
				datas = {
					unitName = _event.initiator:getName(),
					targetName =  _event.target:getName(),
					targetClassName = ATMEDatas.cargos[_event.target:getName()]:getClassName(),
					weaponName = _event.weapon:getName(),
				},
			}

			ATMEFct.fireEvent(event)
			ATMEFct.outputCore(nil, "Event HIT_TARGET généré pour cargo " .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
		end,
		[Object.Category.SCENERY] = function(_event)
			if ATMEDatas.sceneryObjects[_event.target:getName()] == nil then
				ATMEFct.C_SceneryObject_create(_event.target)
			end
			
			local event = {
				type = "HIT_TARGET",
				datas = {
					unitName = _event.initiator:getName(),
					targetName =  _event.target:getName(),
					targetClassName = ATMEDatas.sceneryObjects[_event.target:getName()]:getClassName(),
					weaponName = _event.weapon:getName(),
				},
			}

			ATMEFct.fireEvent(event)
			ATMEFct.outputCore(nil, "Event HIT_TARGET généré pour scenery object" .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
		end,
	}
	
	ATMEFct.shootTargetFct = {
		[Object.Category.UNIT] = function(_event)
			local event = {
				type = "SHOOTING_START",
				datas = {
					unitName = _event.initiator:getName(),
					targetName =  _event.target:getName(),
					targetClassName = ATMEDatas.allAIUnitsAndPlayers[_event.target:getName()]:getClassName(),
				},
			}

			ATMEFct.fireEvent(event)
			ATMEFct.outputCore(nil, "Event SHOOTING_START généré pour unit " .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
		end,
		[Object.Category.STATIC] = function(_event)
			local event = {
				type = "SHOOTING_START",
				datas = {
					unitName = _event.initiator:getName(),
					targetName =  _event.target:getName(),
					targetClassName = ATMEDatas.staticObjects[_event.target:getName()]:getClassName(),
				},
			}
			
			ATMEFct.fireEvent(event)
			ATMEFct.outputCore(nil, "Event SHOOTING_START généré pour static object " .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
		end,
		[Object.Category.CARGO] = function(_event)
			local event = {
				type = "SHOOTING_START",
				datas = {
					unitName = _event.initiator:getName(),
					targetName =  _event.target:getName(),
					targetClassName = ATMEDatas.cargos[_event.target:getName()]:getClassName(),
				},
			}
			
			ATMEFct.fireEvent(event)
			ATMEFct.outputCore(nil, "Event SHOOTING_START généré pour cargo " .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
		end,
		[Object.Category.SCENERY] = function(_event)
			if ATMEDatas.sceneryObjects[_event.target:getName()] == nil then
				ATMEFct.C_SceneryObject_create(_event.target)
			end
			
			local event = {
				type = "SHOOTING_START",
				datas = {
					unitName = _event.initiator:getName(),
					targetName =  _event.target:getName(),
					targetClassName = ATMEDatas.sceneryObjects[_event.target:getName()]:getClassName(),
				},
			}
			
			ATMEFct.fireEvent(event)
			ATMEFct.outputCore(nil, "Event SHOOTING_START généré pour scenery object" .. _event.target:getName() .. " par " .. _event.initiator:getName(), 2)
		end,
	}
	
	local eventFunctions = {
		[world.event.S_EVENT_BIRTH] = {	-- Si un objet est spawné en cours de mission, on passe ici
			[Object.Category.UNIT] = function(_event)
				local unitName = _event.initiator:getName()
				
				-- Valable aussi si c'est un joueur qui arrive ou change de slot après le début de mission
						
				if ATMEDatas.allPlayerUnitsInMission[unitName] ~= nil then -- Player unit
					if ATMEDatas.players[unitName] ~= nil then
						ATMEFct.C_Player_delete(ATMEDatas.players[unitName])
						ATMEFct.outputCore(nil,"Delete Player " .. unitName .. " avant creation", 1)
					end
					--ATMEDatas.newPlayer = true
					ATMEFct.C_Player_create(_event.initiator, ATMEDatas.allPlayerUnitsInMission[unitName].datas, true)
					ATMEFct.outputCore(nil,"Player " .. unitName .. " créé", 2)
				
				elseif ATMEDatas.allAIUnitsInMission[unitName] ~= nil and ATMEDatas.allAIUnitsInMission[unitName].datas ~= nil then
					-- Unités liées à un groupe avec joueurs. Apparaissent au changement de slot.
					ATMEFct.C_AIUnit_create(_event.initiator, ATMEDatas.allAIUnitsInMission[unitName].datas, true)
					ATMEFct.outputCore(nil,"Unit " .. unitName .. " créé & associée à un groupe avec joueur", 2)
				else
					local unit = ATME.C_AIUnit.getByName(unitName)
					local isLateActivation = false
					if unit ~= nil then
						-- On verifie s'il ne s'agit pas d'une unité dont le groupe est à activation retardée
						local group = unit:getGroup()
						local groupLocalMembers = ATMEDatas.localMembers[group:getIndex()]
						
						if groupLocalMembers.isEnabled == true then
							ATMEFct.C_Airbase_delete(_event.initiator)
							ATMEFct.C_AIUnit_delete(unit)
						else   -- Groupe à activation retardée
							groupLocalMembers.nbUnitsToEnable = groupLocalMembers.nbUnitsToEnable - 1
							isLateActivation = true
							
							if groupLocalMembers.nbUnitsToEnable == 0 then
								groupLocalMembers.isEnabled = true
								ATMEFct.outputCore(nil,"Activation du groupe " .. group:getName(), 2)
							end
						end
					end
					
					-- ATTENTION deux cas :
					-- 	1*	L'événement S_EVENT_BIRTH survient pendant l'appel de la fonction addgroup
					--		pour les unités au sol et en mer ou lors de l'activation retardée d'une unité.
					-- 		Les groupes à activation retardées n'ont pas à être traités car initialisés
					--		lors du démarrage d'ATME. Parfois le groupe n'est pas encore initialisé 
					-- 		pendant l'appel de la fonction addgroup. Aussi, dans ce cas, le traitement
					-- 		se fait juste après la fonction addgroup et dans ce cas, l'unité créée est
					--		mémorisée dans ATMEDatas.unitsReadyToCreate.
					--	2*	L'événement survient après le spawn pour les groupes d'aéronefs. Dans ce cas,
					--		l'unité est mémorisée dans ATMEDatas.groupsReadyToBirth après la fonction
					--		addgroup. Le traitement se fait alors ici, lors de la création physique desactivateSpot
					--		unités du groupe.
					
					local found = false
					
					for _name, _group in pairs(ATMEDatas.groupsReadyToBirth) do
						-- L'unité à créer est un bateau avec FARP ou porte-avion

						if _event.initiator:hasAttribute("AircraftCarrier") == true then
							ATMEFct.C_Airbase_create_basic("SHIP", false, _event.initiator)
						elseif _event.initiator:hasAttribute("HelicopterCarrier") == true then
							ATMEFct.C_Airbase_create_basic("SHIP", true, _event.initiator)
						end
							
						-- Traitement du cas 2
						
						if _group.item.units[unitName] ~= nil then
							local datas = {
								uncontrolled = _group.item.uncontrolled,
								visible = _group.item.visible,
								spawnDatas = _group.item.spawnDatas,
							}
							ATMEFct.C_AIUnit_create(_event.initiator, datas, true)
							ATMEDatas.unitsReadyToBirth[unitName] = nil
													
							found = true
							
							_group.item.units[unitName] = nil
							_group.item.nbUnits = _group.item.nbUnits - 1
							
							if _group.item.nbUnits == 0 then
								ATMEDatas.newCreatedGroups[_name].isSpawned = true
								ATMEDatas.newCreatedGroups[_name].spawnCallback = _group.spawnCallback
								ATMEDatas.newCreatedGroups[_name].cloningContext = _group.cloningContext
								ATMEDatas.groupsReadyToBirth[_name] = nil
							end
						end
					end

					if found == false and isLateActivation == false then
						-- Si found vaut false, alors on est dans le cas 1

						table.insert(ATMEDatas.unitsReadyToCreate, _event.initiator)
					end
				end
			end,
			[Object.Category.CARGO] = function(_event)
				local unitName = _event.initiator:getName()
	
				if StaticObject.getByName(unitName) ~= nil then -- Cargo est un StaticObject pour DCS
					if ATMEDatas.cargos[unitName] ~= nil then
						ATMEFct.C_Cargo_delete(ATMEDatas.cargo[unitName])
					end
						
					ATMEFct.C_Cargo_create(_event.initiator)
					ATMEFct.outputCore(nil,"Cargo " .. unitName .. " créé", 2)
				end
			end,
			[Object.Category.STATIC] = function(_event)
				local unitName = _event.initiator:getName()
	
				if StaticObject.getByName(unitName) ~= nil then
					if unitName ~= "static" then -- Bug DCS certains objets spawnés ont toujours leur nom à 'static'.
						if ATMEDatas.staticObjects[unitName] ~= nil then
							ATMEFct.C_StaticObject_delete(ATMEDatas.staticObjects[unitName])
						end
						
						ATMEFct.C_StaticObject_create(_event.initiator)
						ATMEFct.outputCore(nil,"StaticObject " .. unitName .. " créé", 2)
					else
						ATMEFct.outputCore(nil,"StaticObject spawné avec un nom valant 'static' - Ignoré par ATME", 2)
					end
				end
			end,
		},
		[world.event.S_EVENT_PLAYER_ENTER_UNIT] = {
			[Object.Category.UNIT] = function(_event)
			end,
		},
		[world.event.S_EVENT_PLAYER_LEAVE_UNIT] = {
			[Object.Category.UNIT] = function(_event)
				local unitName = _event.initiator:getName()
				if ATMEDatas.players[unitName] ~= nil then
					ATMEFct.C_Player_delete(ATMEDatas.players[unitName])
					ATMEFct.outputCore(nil,"Player " .. unitName .. " Abandon pour changement de slot", 2)
				end
			end,
		},
		[world.event.S_EVENT_DEAD] = {
			[Object.Category.UNIT] = function(_event)
				local unitName = _event.initiator:getName()
				
				if ATMEDatas.AIUnits[unitName] ~= nil then
					ATMEFct.C_Airbase_delete(_event.initiator)
					ATMEFct.C_AIUnit_delete(ATMEDatas.AIUnits[unitName])
					ATMEFct.outputCore(nil,"AIUnit " .. unitName .. " détruite", 2)
				elseif ATMEDatas.players[unitName] ~= nil then
					ATMEFct.C_Player_delete(ATMEDatas.players[unitName])
					ATMEFct.outputCore(nil,"Player " .. unitName .. " détruit", 2)
				end
			end,
			[Object.Category.CARGO] = function(_event)
				local unitName = _event.initiator:getName()
	
				if ATMEDatas.cargos[unitName] ~= nil then
					ATMEFct.C_Cargo_delete(ATMEDatas.cargos[unitName])
					ATMEFct.outputCore(nil,"Cargo " .. unitName .. " détruit", 2)
				end
			end,
			[Object.Category.STATIC] = function(_event)
				local unitName = _event.initiator:getName()
	
				if ATMEDatas.staticObjects[unitName] ~= nil then
					ATMEFct.C_StaticObject_delete(ATMEDatas.staticObjects[unitName])
					ATMEFct.outputCore(nil,"StaticUnit " .. unitName .. " détruite", 2)
				end
			end,
			[Object.Category.SCENERY] = function(_event)
				if ATMEDatas.sceneryObjects[_name] ~= nil then
					ATMEFct.C_SceneryObject_delete(_event.initiator:getName())
					ATMEFct.outputCore(nil,"Scenery object " .. _event.initiator:getName() .. " détruit", 2)
				end
			end,
		},
		[world.event.S_EVENT_CRASH] = {
			[Object.Category.UNIT] = function(_event)
				local unitName = _event.initiator:getName()
				
				if ATMEDatas.AIUnits[unitName] ~= nil then
					ATMEFct.C_Airbase_delete(_event.initiator)
					ATMEFct.C_AIUnit_delete(ATMEDatas.AIUnits[unitName])
					ATMEFct.outputCore(nil,"AIUnit " .. unitName .. " détruite", 2)
				elseif ATMEDatas.players[unitName] ~= nil then
					ATMEFct.C_Player_delete(ATMEDatas.players[unitName])
					ATMEFct.outputCore(nil,"Player " .. unitName .. " détruit", 2)
				end
			end,
		},
		[world.event.S_EVENT_PILOT_DEAD] = {
			[Object.Category.UNIT] = function(_event)
				local unitName = _event.initiator:getName()
				if ATMEDatas.AIUnits[unitName] ~= nil then
					local localMembers = ATMEDatas.localMembers[ATMEDatas.AIUnits[unitName]:getIndex()]
					localMembers.pilotDead = true
					ATMEFct.outputCore(nil,"AIUnit " .. unitName .. " pilote mort", 2)
				elseif ATMEDatas.players[unitName] ~= nil then
					local localMembers = ATMEDatas.localMembers[ATMEDatas.players[unitName]:getIndex()]
					localMembers.pilotDead = true
					ATMEFct.outputCore(nil,"Player " .. unitName .. " pilote mort", 2)
				end
			end,
		},
		[world.event.S_EVENT_EJECTION] = {
			[Object.Category.UNIT] = function(_event)
				local unitName = _event.initiator:getName()
				if ATMEDatas.AIUnits[unitName] ~= nil then
					local localMembers = ATMEDatas.localMembers[ATMEDatas.AIUnits[unitName]:getIndex()]
					localMembers.pilotEjected = true
					ATMEFct.outputCore(nil,"AIUnit " .. unitName .. " pilote éjecté", 2)
				elseif ATMEDatas.players[unitName] ~= nil then
					local localMembers = ATMEDatas.localMembers[ATMEDatas.players[unitName]:getIndex()]
					localMembers.pilotEjected = true
					ATMEFct.outputCore(nil,"Player " .. unitName .. " pilote éjecté", 2)
				end
			end,
		},
		[world.event.S_EVENT_HUMAN_FAILURE] = {
			[Object.Category.UNIT] = function(_event)
				ATMEFct.outputCore(nil,"Human failure " .. _event.initiator:getName(), 2)
				
				-- Non implémenté
			end,
		},
		[world.event.S_EVENT_TAKEOFF] = {
			[Object.Category.UNIT] = function(_event)
				local event = {
					type = "TAKE_OFF",
					datas = {
						unitName = _event.initiator:getName(),
					},
				}

				ATMEFct.C_Event_add(event)
				ATMEFct.outputCore(nil, "Event TAKE_OFF généré pour " .. _event.initiator:getName(), 2)
			end,
		},
		[world.event.S_EVENT_LAND] = {
			[Object.Category.UNIT] = function(_event)
				local event = {
					type = "LAND",
					datas = {
						unitName = _event.initiator:getName(),
					},
				}

				ATMEFct.C_Event_add(event)
				
				ATMEFct.outputCore(nil, "Event LAND généré pour " .. _event.initiator:getName(), 2)
			end,
		},
		[world.event.S_EVENT_ENGINE_STARTUP] = {
			[Object.Category.UNIT] = function(_event)
				-- Bug DCS pour les avions créés en début de mission
				-- Cet événement est généré manuellement lors de la création de l'aéronef.
				-- A ce moment là ATMEDatas[ATMEDatas.allAIUnitsAndPlayers[unitName] est nil
				-- Par contre ok pour démarrage moteur postérieur.

				local unitName = _event.initiator:getName()
				
				if ATMEDatas.allAIUnitsAndPlayers[unitName] ~= nil then
					local event = {
						type = "ENGINE_START",
						datas = {
							unitName = unitName,
						},
					}
					
					ATMEFct.outputCore(nil, "Event ENGINE_START généré pour " .. unitName, 2)
					ATMEFct.C_Event_add(event)
				end
			end,
		},
		[world.event.S_EVENT_ENGINE_SHUTDOWN] = {
			[Object.Category.UNIT] = function(_event)
				local unitName = _event.initiator:getName()
				if ATMEDatas.allAIUnitsAndPlayers[unitName] ~= nil then
					local localMembers = ATMEDatas.localMembers[ATMEDatas.allAIUnitsAndPlayers[unitName]:getIndex()]
					localMembers.engineOn = false
						
					local localMembersGroup = ATMEDatas.localMembers[ATMEDatas.allAIUnitsAndPlayers[unitName]:getGroup():getIndex()]
					localMembersGroup.engineStopped = true
					local event = {
						type = "ENGINE_SHUTDOWN",
						datas = {
							unitName = unitName,
						},
					}

					ATMEFct.C_Event_add(event)
					ATMEFct.outputCore(nil, "Event ENGINE_SHUTDOWN généré pour " .. unitName, 2)
				end
			end,
		},
		[world.event.S_EVENT_BASE_CAPTURED] = {
			[Object.Category.UNIT] = function(_event)
				local unitName = _event.initiator:getName()
				
				if ATMEDatas.players[unitName] ~= nil then
					for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
						local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]
						ATMEDatas.currentModuleName = _nameModule
						if localModuleMembers.handlers.onBaseCapturedHandler ~= nil then
							local datas = {
								airbase = _event.place,
								subPlace = _event.subPlace,
							}
							
							ATMEFct.outputCore(nil,"Appel onBaseCapturedPlayerHandler du module " .. localModuleMembers.name, 2)
							
							xpcall(function() localModuleMembers.handlers.onBaseCapturedPlayerHandler(ATMEDatas.players[unitName], datas) end, function() ATMEFct.errUserHandler('onBaseCapturedPlayerHandler') end)
						end
					end
					
					ATMEFct.outputCore(nil,"Player " .. unitName .. " capture de la base : " .. _event.place:getName(), 2)
				elseif ATMEDatas.AIUnits[unitName] ~= nil then
					for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
						local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]
						ATMEDatas.currentModuleName = _nameModule
						if localModuleMembers.handlers.onBaseCapturedHandler ~= nil then
							local datas = {
								airbase = _event.place,
								subPlace = _event.subPlace,
							}
							
							ATMEFct.outputCore(nil,"Appel onBaseCapturedAIUnitHandler du module " .. localModuleMembers.name, 2)
							
							xpcall(function() localModuleMembers.handlers.onBaseCapturedAIUnitHandler(ATMEDatas.AIUnits[unitName], datas) end, function() ATMEFct.errUserHandler('onBaseCapturedAIUnitHandler') end)
						end
					end
					
					ATMEFct.outputCore(nil,"Unit " .. unitName .. " capture de la base : " .. _event.place:getName(), 2)
				end
			end,
		},
		[world.event.S_EVENT_REFUELING] = {
			[Object.Category.UNIT] = function(_event)
				local event = {
					type = "REFUELING",
					datas = {
						unitName = _event.initiator:getName(),
					},
				}
				
				ATMEFct.C_Event_add(event)
				ATMEFct.outputCore(nil, "Event REFUELING généré pour " .. _event.initiator:getName(), 2)
			end,
		},
		[world.event.S_EVENT_REFUELING_STOP] = {
			[Object.Category.UNIT] = function(_event)
				local event = {
					type = "REFUELING_STOP",
					datas = {
						unitName = _event.initiator:getName(),
					},
				}
				
				ATMEFct.C_Event_add(event)
				ATMEFct.outputCore(nil, "Event REFUELING_STOP généré pour " .. _event.initiator:getName(), 2)
			end,
		},
		[world.event.S_EVENT_SHOT] = {
			[Object.Category.UNIT] = function(_event)
				if _event.weapon ~= nil then
					local weapon = ATMEFct.C_Weapon_create(_event.weapon)
					local event = {
						type = "WEAPON_SHOT",
						datas = {
							unitName = _event.initiator:getName(),
							weaponName = _event.weapon:getName(),
						},
					}
					
					ATMEFct.fireEvent(event)
					ATMEFct.outputCore(nil, "Event WEAPON_SHOT généré pour " .. _event.initiator:getName(), 2)
				end
			end,
		},
		[world.event.S_EVENT_SHOOTING_START] = {
			[Object.Category.UNIT] = function(_event)
				if _event.target ~= nil then
					ATMEFct.shootTargetFct[ _event.target:getCategory()](_event)
					ATMEFct.outputCore(nil, "Event SHOOTING_START généré pour " .. _event.initiator:getName(), 2)
				end
			end,
		},
		[world.event.S_EVENT_SHOOTING_END] = {
			[Object.Category.UNIT] = function(_event)
				local event = {
					type = "SHOOTING_END",
					datas = {
						unitName = _event.initiator:getName(),
					},
				}
				
				ATMEFct.fireEvent(event)
				ATMEFct.outputCore(nil, "Event SHOOTING_END généré pour " .. _event.initiator:getName(), 2)
			end,
		},
		[world.event.S_EVENT_HIT] = {
			[Object.Category.UNIT] = function(_event)
				if _event.target ~= nil and _event.weapon ~= nil then
					ATMEFct.hitTargetFct[ _event.target:getCategory()](_event)
					ATMEFct.C_Weapon_delete(_event.weapon:getName())
					ATMEFct.outputCore(nil, "Event HIT généré pour " .. _event.initiator:getName(), 2)
				end
			end,
		},
		[world.event.S_EVENT_MISSION_START] = function(_event)
			-- Non pris en compte
		end,
		[world.event.S_EVENT_MISSION_END] = function(_event)
			for _, _nameModule in ipairs(ATMEDatas.modulesNames) do
				local localModuleMembers = ATMEDatas.modulesLocalMembers[_nameModule]
				ATMEDatas.currentModuleName = _nameModule
				if localModuleMembers.handlers.onEndMissionHandler ~= nil then
					ATMEFct.outputCore(nil,"Appel onEndMissionHandler du module " .. localModuleMembers.name, 2)
					xpcall(localModuleMembers.handlers.onEndMissionHandler, function() ATMEFct.errUserHandler('onEndMissionHandler') end)
				end
			end
			ATMEFct.outputCore(nil, "Fin de Mission", 2)
		end,
		[world.event.S_EVENT_MARK_ADDED] = function(_event)
			local event = {
					type = "MARK_ADDED",
					datas = {
						time = _event.time,
						id = _event.idx,
						text = _event.text,
						pos = _event.pos,
					},
			}
			
			if _event.initiator ~= nil then
				event.datas.unitName = _event.initiator:getName()
			end
			
			if _event.coalition ~= nil then
				event.datas.coalitionName = ATMEDatas.marks[_event.idx].coalitionName
			elseif _event.groupID ~= nil then
				event.datas.linkedGroupName = ATMEDatas.marks[_event.idx].groupName
			end
			
			ATMEFct.C_Event_add(event)
		end,
		[world.event.S_EVENT_MARK_CHANGE] = function(_event)
			-- Non fonctionnel si utilisation du même ID, rien ne se passe. Bug DCS?
			-- Non implémenté pour l'instant
		end,
		[world.event.S_EVENT_MARK_REMOVED] = function(_event)
			local event = {
					type = "MARK_REMOVED",
					datas = {
						time = _event.time,
						id = _event.idx,
						text = _event.text,
						pos = _event.pos,
					},
			}
			
			if _event.initiator ~= nil then
				event.datas.unitName = _event.initiator:getName()
			end
			
			if _event.coalition ~= nil then
				event.datas.coalitionName = ATMEDatas.marks[_event.idx].coalitionName
			elseif _event.groupID ~= nil then
				event.datas.linkedGroupName = ATMEDatas.marks[_event.idx].groupName
			end
			
			ATMEFct.C_Event_add(event)
			
			ATMEDatas.marks[_event.idx] = nil
		end,
	}
	
	
	function customEventHandler:onEvent(_event)
		if ATMEDatas.exitATME == true then
			world.removeEventHandler(customEventHandler)
		else
			ATMEFct.outputCore(nil,"Event " .. _event.id .. " déclenché", 2)
			
			if eventFunctions[_event.id] ~= nil then
				if type(eventFunctions[_event.id]) == "table" then
					if _event.initiator ~= nil then
						if eventFunctions[_event.id][_event.initiator:getCategory()] ~= nil then
							eventFunctions[_event.id][_event.initiator:getCategory()](_event)
						else
							ATMEFct.outputCore(nil, "Evénement non traité catégorie " .. _event.initiator:getCategory() .. " - Type " .. _event.id, 2)
						end
					else
						ATMEFct.outputVarCore(nil, "Evénement non traité Initiator à nil - Type " .. _event.id, 2)
					end
				elseif type(eventFunctions[_event.id]) == "function" then
					eventFunctions[_event.id](_event)
				else
					ATMEFct.outputCore(nil, "Evénement non traité - Type " .. _event.id, 2)
				end
			else
				ATMEFct.outputCore(nil, "Evénement inconnu - Type " .. _event.id, 2)
			end
		end
	end
	
	----------------------------------------------------------------------------------------------------------
	-- Fonction d'initialisation des objets définis dans la mission
	-- Adapté de MIST - MIssion Scripting Tools
	-- Complétée par la recherche des FARP et AirFields nécessaires à la gestion des airbases.
	----------------------------------------------------------------------------------------------------------
	local function initObjectsInMission()
		for _coalitionName, _coalitionDatas in pairs(env.mission.coalition) do
			if (_coalitionName == 'red' or _coalitionName == 'blue') and type(_coalitionDatas) == 'table' then
				if _coalitionDatas.country then
					for _country_id, _countryDatas in pairs(_coalitionDatas.country) do
						for _typeName, _datas in pairs(_countryDatas) do
							if _typeName == "helicopter" or _typeName == "ship" or _typeName == "plane" or _typeName == "vehicle" or _typeName == "static" then	
								local isStatic = _typeName == "static"
								for _, _groupDatas in pairs(_datas.group) do
									if _groupDatas ~= nil and _groupDatas.units ~= nil then
										local currentGroupName = env.getValueDictByKey(_groupDatas.name)
										local d,f = string.find(currentGroupName, "|", 1)
										
										if d == nil then
											d,f = string.find(currentGroupName:upper(), "%s+SPAWNED%s+", 1)
											if d == nil then
												if isStatic == false then
													local uncontrolled = _groupDatas.uncontrolled
													local visible
													
													if _typeName == "ship" or _typeName == "vehicle" then
														uncontrolled = false
														if _groupDatas.lateActivation == nil or _groupDatas.lateActivation == false then
															visible = true
														else
															visible = _groupDatas.visible
														end
													else
														visible = _groupDatas.lateActivation == false
													end
													
													ATMEDatas.allGroupsInMission[currentGroupName] = {
														name = currentGroupName,
														country = _countryDatas.id,
														coalition = string.upper(_coalitionName),
														groupId = _groupDatas.groupId,								-- Uniquement défini pour les groupes initiaux.
														category = ATMEConst.groupCategories[_typeName],
														type = _typeName,
														uncontrolled = uncontrolled,
														units = {},
														visible = visible,
														taskSelected = true,
														lateActivation = _groupDatas.lateActivation,
														hidden = _groupDatas.hidden,
														start_time = 0,
														task = _groupDatas.task,
														tasks = _groupDatas.tasks,
														radioSet = _groupDatas.radioSet,
														modulation = _groupDatas.modulation,
														frequency = _groupDatas.frequency,
														communication = _groupDatas.communication,
														waypoints = {},
													}
													
													ATMEDatas.allInitialGroupNames[currentGroupName] =  {
														groupName = currentGroupName,
														unitsName = {},
													}
													
													local firstPlayerFound = false
													
													for _key, _unitDatas in ipairs(_groupDatas.units) do
														local unitName = env.getValueDictByKey(_unitDatas.name)
														local d,f = string.find(unitName, "|", 1)
											
														if d == nil then
															d,f = string.find(unitName:upper(), "%s+SPAWNED%s+", 1)
															if d == nil then
																ATMEDatas.allGroupsInMission[currentGroupName].units[unitName] = {
																	name = unitName,
																	row = _key,
																	type = _unitDatas.type,
																	id = _unitDatas.unitId,
																	livery_id = _unitDatas.livery_id,
																	speed = _unitDatas.speed,
																	psi = _unitDatas.psi,
																	payload = _unitDatas.payload,
																	skill = _unitDatas.skill,
																	x = _unitDatas.x,
																	y = _unitDatas.y,
																	playerCanDrive = _unitDatas.playerCanDrive,
																	heading = _unitDatas.heading,
																	alt = _unitDatas.alt,
																	alt_type = _unitDatas.alt_type,
																	callsign = _unitDatas.callsign,
																	onboard_num = _unitDatas.onboard_num,
																	parking = _unitDatas.parking,
																	parking_id = _unitDatas.parking_id,
																}
																
																ATMEDatas.allInitialGroupNames[currentGroupName].unitsName[_key] = unitName
																
																if _unitDatas.skill == "Client" or _unitDatas.skill == "Player" then
																	if firstPlayerFound == false then
																		ATMEDatas.allPlayerUnitsInMission[unitName] = {
																			group = ATMEDatas.allGroupsInMission[currentGroupName],
																			unit = ATMEDatas.allGroupsInMission[currentGroupName].units[unitName],
																		}
																		
																		-- Test here because of DCS ME bug when changing unit to player -> possibiliy to change uncontrolled.
																		if ATMEDatas.allPlayerUnitsInMission[unitName].group.uncontrolled == true then
																			env.error("[ATME] Player " .. unitName .. " has uncontrolled attribute set to true in ME\nATME stopped", true)
																			return true
																		end
																		
																		firstPlayerFound = true
																	else
																		env.error("[ATME] Player " .. unitName .. " can't be in a group with already another player\nATME stopped", true)
																		return true
																	end
																else
																	ATMEDatas.allAIUnitsInMission[unitName] = {
																		group = ATMEDatas.allGroupsInMission[currentGroupName],
																		unit = ATMEDatas.allGroupsInMission[currentGroupName].units[unitName],
																	}
																end
															else
																env.error("[ATME] Unit name " .. unitName .. " can't contain 'Spawned' word\nATME stopped", true)
																return true
															end
														else
															env.error("[ATME] Unit name " .. unitName .. " can't contain '|' character\nATME stopped", true)
															return true
														end
													end
													
													if _groupDatas.route ~= nil and _groupDatas.route.points and #_groupDatas.route.points > 0 then
														for _pointNum, _point in ipairs(_groupDatas.route.points) do
															local pointName =  env.getValueDictByKey(_point.name)
															ATMEDatas.allGroupsInMission[currentGroupName].waypoints[#ATMEDatas.allGroupsInMission[currentGroupName].waypoints + 1] = {
																name = pointName,
																x = _point.x,
																y = _point.y,
																speed = _point.speed,
																speed_locked = _point.speed_locked,
																alt = _point.alt,
																alt_type = _point.alt_type,
																ETA = _point.ETA,
																ETA_locked = _point.ETA_locked,
																airdromeId = _point.airdromeId,
																helipadId = _point.helipadId,
																type = _point.type,
																action = _point.action,
																task = _point.task,
																properties = _point.properties,
																formation_template = _point.formation_template,
															}
														end
													end
												else 
													for _, _unitDatas in pairs(_groupDatas.units) do
														local unitName = env.getValueDictByKey(_unitDatas.name)
														local d,f = string.find(unitName, "|", 1)
											
														if d == nil then
															d,f = string.find(unitName:upper(), "%s+SPAWNED%s+", 1)
															if d == nil then
																if _unitDatas.category == "GrassAirfields" or _unitDatas.category == "Heliports" then
																	ATMEDatas.allStaticAirbasesInMission[unitName] = {
																		name = unitName,
																		type = _unitDatas.type,
																		id = _unitDatas.unitId,
																		x = _unitDatas.x,
																		y = _unitDatas.y,
																		canCargo = _unitDatas.canCargo,
																		heading = _unitDatas.heading,
																		category = _unitDatas.category,
																		coalition = string.upper(_coalitionName),
																		dead = _groupDatas.dead,
																	}
																elseif _unitDatas.category == "Cargos" then
																	ATMEDatas.allCargosInMission[unitName] = {
																		name = unitName,
																		type = _unitDatas.type,
																		x = _unitDatas.x,
																		y = _unitDatas.y,
																		canCargo = _unitDatas.canCargo,
																		heading = _unitDatas.heading,
																		category = _unitDatas.category,
																		dead = _groupDatas.dead,
																		shape_name = _unitDatas.shape_name,
																		mass = _unitDatas.mass,
																		rate =_unitDatas.rate,
																		livery_id = _unitDatas.livery_id,
																	}
																else
																	ATMEDatas.allStaticObjectsInMission[unitName] = {
																		name = unitName,
																		type = _unitDatas.type,
																		x = _unitDatas.x,
																		y = _unitDatas.y,
																		canCargo = _unitDatas.canCargo,
																		heading = _unitDatas.heading,
																		category = _unitDatas.category,
																		dead = _groupDatas.dead,
																		shape_name = _unitDatas.shape_name,
																		mass = _unitDatas.mass,
																		rate =_unitDatas.rate,
																		livery_id = _unitDatas.livery_id,
																	}
																end
															else
																env.error("[ATME] Static Unit name " .. unitName .. " can't contain 'Spawned' word\nATME stopped", true)
																return true
															end
														else
															env.error("[ATME] Static Unit name " .. unitName .. " can't contain '|' character\nATME stopped", true)
															return true
														end
													end
												end
											else
												env.error("[ATME] Group name " .. currentGroupName .. " can't contain 'Spawned' word\nATME stopped", true)
												return true
											end
										else
											env.error("[ATME] Group name " .. currentGroupName .. " can't contain '|' character\nATME stopped", true)
											return true
										end
									end
								end 
							end
						end
					end
				end
			end
		end
		
		return false
	end
	
	local _timerStartMission
	_timerStartMission = function()
		for _id, _current in pairs(world.getAirbases()) do
			if _current:hasAttribute("Airfields") == true then
				-- Bug DCS V2.5.6
				local id = _id
				if ATMEDatas.theatreName == "CAUCASUS" then
					id = id + 11
				end
				
				ATMEFct.C_Airbase_create_basic("AIRBASE", false, _current, id)
			elseif _current:hasAttribute("Helipad") == true then
				ATMEFct.C_Airbase_create_basic("HELIPAD", true, _current)
			elseif _current:hasAttribute("AircraftCarrier") == true then
				ATMEFct.C_Airbase_create_basic("SHIP", false, _current)
			elseif _current:hasAttribute("HelicopterCarrier") == true then
				ATMEFct.C_Airbase_create_basic("SHIP", true, _current)
			end
		end
	
		world.addEventHandler(customEventHandler)

		for _, _object in pairs(ATMEDatas.allStaticObjectsInMission) do
			local objectName = _object.name
			local object = StaticObject.getByName(objectName)
			
			if object ~= nil then 
				-- Bug DCS : Certains objets comme M1 Barrage Balloon sont encore avec object = nil.
				-- Non pris en compte par ATME
				ATMEFct.C_StaticObject_create(object)
			else
				ATMEFct.outputCore(nil,"Static object " .. objectName .. " ignoré par ATME : nil dans la fonction run", 2)
			end
		end
	
		for _, _object in pairs(ATMEDatas.allCargosInMission) do
			local objectName = _object.name
			local object = StaticObject.getByName(objectName)	-- Les cargos sont les staticObject pour DCS
			
			if object ~= nil then 
				-- Bug DCS : Certains objets comme M1 Barrage Balloon sont encore avec object = nil.
				-- Non pris en compte par ATME
				ATMEFct.C_Cargo_create(object)
			else
				ATMEFct.outputCore(nil,"Cargo " .. objectName .. " ignoré par ATME : nil dans la fonction run", 2)
			end
		end
	
		-- On notera que les unités avec une activation retardées sont bien créées ! 
		-- ATTENTION ... Vrai pour les unités joueurs et les unités AI
		for _name, _group in pairs(ATMEDatas.allGroupsInMission) do
			local nbUnits = 0
			
			local spawnDatas = {
				country = _group.country,
				category = _group.category,
				type = _group.type,
				initialNames = ATMEDatas.allInitialGroupNames[_name],
				datas = {
					name = _name,
					groupId = _group.groupId,
					hidden = _group.hidden,
					start_time = 0,
					task = _group.task,
					visible = false,
					taskSelected = true,
					task = _group.task,
					lateActivation = false,
					tasks = {},
					units = {},
					route = {
						points = _group.waypoints,
					},
				},
			}
			
			for _unitName, _unit in pairs(_group.units) do
				spawnDatas.datas.units[_unit.row] = {
					row = _unit.row,
					type = _unit.type,
					name = _unitName,
					skill = _unit.skill,
					x = _unit.x,
					y = _unit.y,
					alt = _unit.alt,
					alt_type = _unit.alt_type,
					playerCanDrive = _unit.playerCanDrive,
					heading = _unit.heading,
					payload = _unit.payload,
					speed = _unit.speed,
					callsign = _unit.callsign,
					onboard_num = _unit.onboard_num,
					psi = _unit.psi,
					livery_id = _unit.livery_id,
				}
			end
			
			local datas = {
				uncontrolled = _group.uncontrolled,
				visible = _group.visible,
				spawnDatas = spawnDatas,
			}

			for _, _unit in pairs(_group.units) do
				if _unit.skill ~= "Player" and _unit.skill ~= "Client" then
					local unit = Unit.getByName(_unit.name)
					if unit ~= nil then
						ATMEFct.C_AIUnit_create(unit, datas, true)
					else
						ATMEDatas.allAIUnitsInMission[_unit.name].datas = datas
					end
				else
					ATMEDatas.allPlayerUnitsInMission[_unit.name].datas = datas
				end
				
				nbUnits = nbUnits + 1
			end
			
			if _group.lateActivation == true then
				local group = ATMEDatas.newCreatedGroups[_group.name].group
				local currentLocalMembers = ATMEDatas.localMembers[group:getIndex()]
				currentLocalMembers.isEnabled = false
				currentLocalMembers.nbUnitsToEnable = nbUnits
				ATMEFct.outputCore(nil,"Activation retardee pour " .. _group.name, 2)
			end
		end

		for i = 1, 2 do
			local newPlayerUnits = coalition.getPlayers(i)
			for _, _newPlayer in ipairs(newPlayerUnits) do
				ATMEFct.C_Player_create(_newPlayer, ATMEDatas.allPlayerUnitsInMission[_newPlayer:getName()].datas, true)
			end
		end

		ATMEDatas.isRunning = true

		_timerMainFunction()

		if ATMEDatas.racesActivation == true then
			_timer100x()
		end
	end
	
	----------------------------------------------------------------------------------------------------------
	-- Fonction de démmarrage d'ATME à faire après le chargement de tous les modules. Cette fonction se met dans
	-- l'éditeur de mission par un script lua.
	----------------------------------------------------------------------------------------------------------
	ATME.run = function(_config)
		local fctDefinition = ATMEFct.getErrorPrefix() .. "run([config])"
		
		if ATMEDatas.exitATME == false then
			if _config == nil or type(_config) == "string" then
				if type(_config) == "string" then
					local tokens = ATMEFct.analyzeString(_config)
			
					if #tokens >= 1 then
						if tokens[1].type == "TOKEN_LANGUAGE" then
							if #tokens == 2 then
								if tokens[2].type == "TOKEN_RACE" then
									ATMEDatas.racesActivation = ATMEConst.raceTokenTypes[tokens[2].value]
								else
									ATMEFct.errorOutputCore(nil, fctDefinition, "config formatted string second parameter must be nil, RACE_ON or RACE_OFF", false)
									return
								end
							elseif #tokens > 2 then
								ATMEFct.errorOutputCore(nil, fctDefinition, "config formatted string must have 2 parameters max", false)
								return
							end
						else
							ATMEFct.errorOutputCore(nil, fctDefinition, "config formatted string first parameter must be FR (default), GER, RUS, EN", false)
							return
						end
						
						ATMEDatas.language = tokens[1].value
					else
						ATMEFct.errorOutputCore(nil, fctDefinition, "config formatted string must have at least 1 parameter", false)
						return
					end
				end
			elseif _config == nil then
				ATMEDatas.language = "FR"
			else
				ATMEFct.errorOutputCore(nil, fctDefinition, "config type must be nil or string", false)
				return
			end
			
			-- Permet de lancer ATME dès le début de la mission.
			if timer.getTime() < 1 then
				timer.scheduleFunction(_timerStartMission, {}, timer.getTime() + 1)
			else
				_timerStartMission()
			end
			ATMEFct.outputCore(nil, "ATME is running -> " .. ATMEDatas.language, 2)
		end
	end
	
	----------------------------------------------------------------------------------------------------------
	-- Fonction MAIN d'initialisation générale appelée une seule fois au lancement du fichier script
	----------------------------------------------------------------------------------------------------------
	do
		env.info("[ATME] - Init starts - version " .. versionATME, false)
		
		ATMEDatas.exitATME = initObjectsInMission()
		ATMEDatas.theatreName = string.upper(env.mission.theatre)
		
		env.info("[ATME] - Init ok - Theatre : " .. ATMEDatas.theatreName, false)
		
		-- Fichier script dans Racine DCS\Scripts\MissionScripting.lua, mettre sanitizeModule('lfs') en commentaire
		-- pour utilisation du gestionnaire de fichiers lua. lfs est alors non nil
		
		if lfs ~= nil then
			ATMEDatas.files = {}
			ATMEDatas.ATMEDir = lfs.writedir() .."ATME"
			
			env.info("[ATME] - lfs ok - " .. ATMEDatas.ATMEDir, false)
		end
	end
end